/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "application.h"
#include "graphics.h"
#include "gui.h"

#include "graphics.inl"
#include "gui.inl"

namespace magnum { namespace sample {
using namespace magnum::graphics;
using namespace magnum::graphics3d;
using magnum::graphics3d::Context;
using magnum::object::AutoPointer;


static AutoPointer<PointLight>	g_Light;
static AutoPointer<StringView>	g_TriangleCounter;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
GraphicsView::GraphicsView(MainWindow* win, const Rectangle<float>& rect) : OpenGL(rect) {
	setColourBits(24);
	setAlphaBits(0);
	setDepthBits(24);
	setStencilBits(0);
	m_ModelParameters = NULL;
	m_Window = win;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onAttachWindow() {
	OpenGL::onAttachWindow();

	m_Context = Context::getCurrent();
	m_Context->setCulling(Context::CullBack);

	m_MatrixStack = m_Context->getMatrixStack();

	m_Context->setClearColour(Colour<float>(0,0,0));
	m_Context->setViewport(getBounds());
	
	m_Gui = new Gui(this, m_Context);
	m_Gui->attachDirectInput();
	
	m_Timer.init(0);
	
	return onInit();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onDetachWindow() {
	onExit();
	m_Gui->release();
	return OpenGL::onDetachWindow();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onInit() {
	m_Camera.setNearPlane(10);
	m_Camera.setFarPlane(800);
	float w = getSize().getWidth();
	float h = getSize().getHeight();
	float f = max(w,h);
	w = w / f * 10.0f;
	h = h / f * 10.0f;
	m_Camera.setProjectionPlane(10.0f);
	m_Camera.setViewWindow(Rectangle<float>(-0.5f*w, -0.5f*h, w, h));

	g_Light = new PointLight();
	g_Light->setPosition(Vector3<float>(0.0f,0.0f,400.0f));
	g_Light->setAmbient(Colour<float>(0.2f,0.2f,0.2f));
	g_Light->setDiffuse(Colour<float>(1.0f,1.0f,1.0f));
	g_Light->setSpecular(Colour<float>(0.8f,0.8f,0.8f));
	
	g_TriangleCounter = new StringView(Rectangle<float>(10,480,170,40), "0 Triangles");
	g_TriangleCounter->setAlignement(View::FollowBottom);
	g_TriangleCounter->setFont(m_Gui->getFont());
	getGui()->insert(g_TriangleCounter);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onResize(const Rectangle<float>& r) {
	OpenGL::onResize(r);
	float w = r.getWidth();
	float h = r.getHeight();
	float f = max(w,h);
	w = w / f * 10.0f;
	h = h / f * 10.0f;
	m_Camera.setProjectionPlane(10.0f);
	m_Camera.setViewWindow(Rectangle<float>(-0.5f*w, -0.5f*h, w, h));
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onExit() {
	m_Model = NULL;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onHandleMessage(const Message& msg) {
	return OpenGL::onHandleMessage(msg);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GraphicsView::onRender() {
	DepthMode::selectDefault();
	BlendMode::selectDefault();

	m_Context->setClearColour(Colour<float>(0.59f, 0.59f, 1.0f));
	m_Context->clear(Context::ClearColour | Context::ClearDepth | Context::ClearStencil);
	m_Camera.select();

	m_MatrixStack->selectMode(MatrixStack::ModelView);
	m_MatrixStack->identity();

	g_Light->select(0);

	m_Context->setCulling(Context::CullBack);
	if (m_Model != NULL) {
		// Set transform matrix
		*(Matrix4<float>*)m_ModelParameters->getBuffer() = (Matrix4<float>)getTransform().getInverse();
		
		// Set animation data (if present)
		if (m_AnimationPtr != NULL)
			m_Animations[0]->animate((float)m_Timer.getCurrent() / (float)Timer::Second, m_AnimationPtr);

		// Render model
		m_Model->render(m_ModelParameters);
	}
	m_Context->renderDefered();
	
	float width = getBounds().getWidth();
	float height = getBounds().getHeight();
	GeometryRenderer::renderAxes(getTransform(), Vector2<float>(width - 32, height - 32));

	m_Gui->process();
	m_Gui->paint();
}



static void printBounds(Model* model) {
	Bounds3<float> bounds = model->getBounds();
	printf("bounds: origin = (%f,%f,%f)\n", bounds.getOrigin()[0], bounds.getOrigin()[1], bounds.getOrigin()[2]);
	printf("       axis[0] = (%f,%f,%f)\n",	bounds.getAxis(0)[0], bounds.getAxis(0)[1], bounds.getAxis(0)[2]);
	printf("       axis[1] = (%f,%f,%f)\n",	bounds.getAxis(1)[0], bounds.getAxis(1)[1], bounds.getAxis(1)[2]);
	printf("       axis[2] = (%f,%f,%f)\n",	bounds.getAxis(2)[0], bounds.getAxis(2)[1], bounds.getAxis(2)[2]);
}


static void allToPhong(MeshList& mlist) {
	AutoPointer<Shader> pshader = new PhongShader();
	AutoPointer<Shader> tshader = new TexturePhongShader();
	for (MeshList::Iterator miter = mlist.begin(); miter != mlist.end(); ++miter) {
		Appearance* app = (*miter)->getAppearance();
		if (dynamic_cast<TextureShader*>(app->getShader()))
			app->attachShader(tshader);
		else if (dynamic_cast<PlainShader*>(app->getShader()))
			app->attachShader(pshader);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GraphicsView::resetNavigator() {
	if (m_Model == NULL) {
		m_Camera.setNearPlane(10);
		m_Camera.setFarPlane(800);
		m_Gui->setTranslation(Vector3<float>(0,0,400.0f));
	}
	
	printBounds(m_Model);
	Bounds3<float> bounds = m_Model->getBounds();
	float diam = max(max(bounds.getAxis(0).maxabs(), bounds.getAxis(1).maxabs()), bounds.getAxis(2).maxabs());
	m_Gui->setTranslation(Vector3<float>(0,0,1.5f*diam) + bounds.getOrigin());
	m_Gui->setPivot(bounds.getOrigin());
	m_Gui->setMotionSpeed(0.02f*diam);
	float np = 0.02f*diam;
	float fp = 20*diam;
	m_Camera.setNearPlane(np);
	m_Camera.setFarPlane(fp);
	m_Gui->getNavigator()->reset();
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GraphicsView::setModel(Model* model) {
	if (m_ModelParameters != NULL) {
		m_ModelParameters->release();
		m_ModelParameters = NULL;
	}
	m_Model = model;
	if (m_Model == NULL)
		return;

	MeshList mlist = m_Model->getMeshes();
	m_ModelSize = 0;
	for (MeshList::Iterator miter = mlist.begin(); miter != mlist.end(); ++miter) {
		IndexMesh* imesh = dynamic_cast<IndexMesh*>(*miter);
		if (imesh == NULL)
			continue;
		if (imesh->getRange().getSize() == 0)
			m_ModelSize += imesh->getIndexBuffer()->getSize() / 3;
		else
			m_ModelSize += imesh->getRange().getSize() / 3;
	}
	
	allToPhong(mlist);

	const ParameterFormat& fmt = m_Model->getParameters();
	m_ModelParameters = ParameterBuffer::create(fmt);
	m_Animations = m_Model->getAnimations();
	if (fmt.contains(Parameters::AnimationData)) 
		m_AnimationPtr = m_ModelParameters->getPointer(fmt.getLocation(Parameters::AnimationData));
	else
		m_AnimationPtr = NULL;
	
	

	char buffer[40];
	sprintf(buffer, "Triangles: %d", m_ModelSize);
	g_TriangleCounter->setText(buffer);
}

}; };
