/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include <windows.h>

#include "window.h"
#include "application.h"
#include "graphics.h"
#include "gui.h"
#include "models.h"

#include "gui.inl"
#include "graphics.inl"

namespace magnum { namespace sample {
using magnum::application::Application;
using magnum::graphics::Dimension;
using magnum::graphics3d::Translator3DS;
using magnum::graphics3d::TranslatorIFS;
using magnum::graphics3d::TranslatorLWO;
using magnum::graphics3d::TranslatorMD2;
using magnum::graphics3d::TranslatorMD3;
using magnum::graphics3d::TranslatorMDS;
using magnum::graphics3d::TranslatorOBJ;
using magnum::graphics3d::TranslatorPLY;
using magnum::graphics3d::TranslatorOFF;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MainWindow::MainWindow(ViewerApplication* app) {
	m_Application = app;

	setName("Magnum");
	setSize(Dimension<float>(600,500));
	setMinimumSize(Dimension<float>(256,256));

	m_Graphics = new GraphicsView(this, magnum::graphics::Rectangle<float>(0,0,600,500));
	m_Graphics->setAlignement(magnum::gui::View::FollowAll);
	insert(m_Graphics);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MainWindow::~MainWindow() {
	m_Graphics->release();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool MainWindow::onRequestClose(void) {
	postMessage(Application::getInstance(), Application::QuitMessage);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool MainWindow::onHandleMessage(const Message& msg) {
	switch (msg.getCommand()) {
		case LoadModel:
			loadModel();
			return true;
		default:
			return Window::onHandleMessage(msg);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
bool MainWindow::loadModel(const char* fname) {
	AutoPointer<Model> model;
	try {
		m_Graphics->setModel(NULL);
		model = createModel<T,Model>(fname);
		m_Graphics->setModel(model);
	} catch(...) {
		return false;
	}
	
	if (m_Graphics->getModel() == NULL)
		return false;

	m_Graphics->resetNavigator();
	
	model->optimise();
	invalidate();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool MainWindow::loadModel() {
	OPENFILENAME opfn;
	char fname[512];
	fname[0] = 0;
	opfn.lStructSize = sizeof(opfn);
	opfn.hwndOwner = getWindowHandle();
	opfn.hInstance = NULL;
	opfn.lpstrFilter = 
		"All supported files\0*.3ds;*.obj;*.ply;*.lwo;*.ifs;*.md2;*.md3;*.mds;*.off\0"
		"*.3ds 3D-Studio Files\0*.3ds\0"
		"*.ifs IFS Files\0*.ifs\0"
		"*.lwo Lightwave Object Files\0*.lwo\0"
		"*.md2 Quake2 Models\0*.md2\0"
		"*.md3 Quake3 Models\0*.md3\0"
		"*.mds Return to Castle Wolfenstein\0*.mds\0"
		"*.obj Maya Wavefront Files\0*.obj\0"
		"*.off OFF File Format\0*.off\0"
		"*.ply Stanford Polygon Format\0*.ply\0"
		"All files\0*.*\0\0";
	opfn.lpstrCustomFilter = NULL;
	opfn.nMaxCustFilter = 0;
	opfn.nFilterIndex  = 0;
	opfn.lpstrFile = fname;
	opfn.nMaxFile = sizeof(fname);
	opfn.lpstrFileTitle = NULL;
	opfn.nMaxFileTitle = 0;
	opfn.lpstrInitialDir = NULL;
	opfn.lpstrTitle = "Chose File to open";
	opfn.Flags = OFN_FILEMUSTEXIST;
	opfn.nFileOffset = 0;
	opfn.nFileExtension = 0;
	opfn.lpstrDefExt = NULL;
	opfn.lpTemplateName = NULL;
	
	if (!GetOpenFileName(&opfn))
		return true;

	// Extract file extension (only 3-char supported)
	size_t len = strlen(fname);
	if (len < 4)
		return true;
	String ext(fname+len-4);
	ext.lower();
	
	if (ext == ".3ds") {
		loadModel<Translator3DS>(fname);
	} else if (ext == ".obj") {
		loadModel<TranslatorOBJ>(fname);
	} else if (ext == ".lwo") {
		loadModel<TranslatorLWO>(fname);
	} else if (ext == ".md2") {
		loadModel<TranslatorMD2>(fname);
	} else if (ext == ".md3") {
		loadModel<TranslatorMD3>(fname);
	} else if (ext == ".mds") {
		loadModel<TranslatorMDS>(fname);
	} else if (ext == ".ifs") {
		loadModel<TranslatorIFS>(fname);
	} else if (ext == ".ply") {
		loadModel<TranslatorPLY>(fname);
	} else if (ext == ".off") {
		loadModel<TranslatorOFF>(fname);
	} 
	return true;
}


}; };

