/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "graphics.h"
#include "gui.h"
#include "sky.h"

#include "graphics.inl"
#include "gui.inl"


namespace magnum { namespace sample {
using namespace magnum::graphics;
using namespace magnum::geometry;
using namespace magnum::graphics3d;
using magnum::graphics3d::Context;
using magnum::object::AutoPointer;
using magnum::io::SystemFile;
using magnum::io::Content;
using magnum::stats::Uniform;

#define numSpheres			4


static PointLight*			g_Light;
static SkyModel*			g_SkyModel;
static ParameterBuffer*		g_SkyParameters;
static GeometryModel*		g_SphereModel[numSpheres];
static Transform3<float>	g_SphereTransform[numSpheres];
static Vector<float,3>		g_SphereRotationAxis[numSpheres];
static ParameterBuffer*		g_SphereParameters[numSpheres];
static CubemapGenerator*	g_SphereRenderer[numSpheres];

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
class SphereRenderer : public CubemapGenerator {
DeclareClass(SphereRenderer, CubemapGenerator);
private:
	unsigned int			m_SphereIndex;
	MatrixStack*			m_MatrixStack;
	
public:
	SphereRenderer(CubeTexture* ,unsigned int );
	
protected:
	void					onRender(const Matrix4<float>& ,CubeTexture::Face f);
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SphereRenderer::SphereRenderer(CubeTexture* tex, unsigned int si) : CubemapGenerator(tex) {
	m_SphereIndex = si;
	m_MatrixStack = getContext()->getMatrixStack();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void SphereRenderer::onRender(const Matrix4<float>& mtx,CubeTexture::Face f) {
	getContext()->clear(Context::ClearDepth);

	// Set up light
	m_MatrixStack->identityTexture();
	m_MatrixStack->setModel(getTransform(f));
	g_Light->forceSelect(0);

	// Render background
	*(Matrix4<float>*)g_SkyParameters->getBuffer() = mtx;
	g_SkyModel->render(g_SkyParameters);

	// Setup texture matrix for reflections
	Matrix4<float> tex = mtx.getInverse();
	tex[0][3] = tex[1][3] = tex[2][3] = 0.0f;
	m_MatrixStack->setTexture(tex);

	// Render Spheres except current sphere
	for (unsigned int k = 0; k < numSpheres; k++) {
		if (k == m_SphereIndex)
			continue;
		*(Matrix4<float>*)g_SphereParameters[k]->getBuffer() = mtx * g_SphereTransform[k];
		g_SphereModel[k]->render(g_SphereParameters[k]);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onInit() {
	unsigned int i;

	m_Camera.setNearPlane(15);
	m_Camera.setFarPlane(2500);
	m_Camera.select();

	g_Light = new PointLight();
	g_Light->setPosition(Vector3<float>(300.0f,200.0f,400.0f));
	g_Light->setAmbient(Colour<float>(0.3f,0.3f,0.3f));
	g_Light->setDiffuse(Colour<float>(0.7f,0.7f,0.7f));
	g_Light->setSpecular(Colour<float>(1.0f,1.0f,1.0f));
	
	// Create Skybox
	g_SkyModel = createSky();
	g_SkyParameters = ParameterBuffer::create(g_SkyModel->getParameters());

	// Create Cubemaps
	CubeTexture* cubemap[numSpheres];
	for (i = 0; i < numSpheres; i++) {
		cubemap[i] = CubeTexture::create(Image::RGB8, 256, 256);
		g_SphereRenderer[i] = new SphereRenderer(cubemap[i],i);
	}
	
	// Create models
	AutoPointer<File> vertf = SystemFile::open("data/cubeshader.vp");
	AutoPointer<VertexProgram> vprog = VertexProgram::create(Content(vertf));
	
	AutoPointer<File> pixf = SystemFile::open("data/cubeshader.pp");
	AutoPointer<PixelProgram> pprog = PixelProgram::create(Content(pixf));

	AutoPointer<ProgrammableShader> shader = ProgrammableShader::create();
	shader->attach(vprog);
	shader->attach(pprog);
	shader->link();
	
//	AutoPointer<Shader> shader = new CubeEnvironmentShader();
	Uniform random;
	random.setSeed(0);
	for (i = 0; i < numSpheres; i++) {
		g_SphereModel[i] = new GeometryModel(GeometryModel::Sphere, random()*60.0f + 50.0f, 75);
		AutoPointer<ColourMaterial> material = new ColourMaterial();
		material->setAmbient(Colour<float>(0.1f,0.1f,0.1f));
		material->setDiffuse(Colour<float>(min(1.0f*random(),1.0f),min(1.0f*random(),1.0f),min(1.0f*random(),1.0f)));
		material->setSpecular(Colour<float>(1.0f,1.0f,1.0f));
		material->setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
		material->setShininess(30.0f);
		g_SphereModel[i]->getAppearance()->attachMaterial(material);
		g_SphereModel[i]->getAppearance()->attachTexture(cubemap[i]);
		g_SphereModel[i]->getAppearance()->attachShader(shader);
		g_SphereTransform[i].identity();
		g_SphereTransform[i].setTranslation(Vector3<float>(400.0f * random() - 200.0f, 400.0f * random() - 200.0f, 400.0f * random() - 200.0f));
		g_SphereParameters[i] = ParameterBuffer::create(g_SphereModel[i]->getParameters());
		g_SphereRotationAxis[i][0] = random();
		g_SphereRotationAxis[i][1] = random();
		g_SphereRotationAxis[i][2] = random();
		cubemap[i]->release();
	}

	// Create GUI
	StringView* sview;
	if (g_SphereRenderer[0]->getMode() == CubemapGenerator::Compatible)
		sview = new StringView(Rectangle<float>(10,480,170,40), "CopyTexture-Mode (slow)");
	else
		sview = new StringView(Rectangle<float>(10,480,170,40), "RenderTarget-Mode (fast)");
	sview->setAlignement(View::FollowBottom);
	sview->setFont(m_Gui->getFont());
	getGui()->insert(sview);

	// Initialise timer
	m_Timer.init(0);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onExit() {
	unsigned int i;
	for (i = 0; i < numSpheres; i++) {
		g_SphereRenderer[i] ->release();
		g_SphereModel[i]->release();
		g_SphereParameters[i]->release();
	}
	g_SkyModel->release();
	g_SkyParameters->release();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onHandleMessage(const Message& msg) {
	return OpenGL::onHandleMessage(msg);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GraphicsView::onRender() {
	DepthMode::selectDefault();
	BlendMode::selectDefault();

	unsigned int i;
	
	// Animate spheres
	m_Timer.update();
	float step = (float)((double)m_Timer.getElapsed() / (double)(4*Timer::Second));
	for (i = 0; i < numSpheres; i++) {
		g_SphereTransform[i].rotate(step, g_SphereRotationAxis[i]);
	}

	Transform3<float> tf = getTransform().getInverse();
	Matrix<float,3> rot = tf.getRotation();
	for (i = 0; i < numSpheres; i++) {
		// Render cubemap for each sphere
		g_SphereRenderer[i]->render((Matrix4<float>)Transform3<float>(-g_SphereTransform[i].getTranslation()));
	}

	// Select default Frame Buffer
	FrameBuffer::selectDefault();
	m_Context->clear(Context::ClearDepth);

	// Select default camera
	m_Camera.select();

	// Set up light
	m_MatrixStack->identityTexture();
	m_MatrixStack->identityModel();
	g_Light->forceSelect(0);
	
	// Render background
	*(Matrix4<float>*)g_SkyParameters->getBuffer() = (Matrix4<float>)tf;
	g_SkyModel->render(g_SkyParameters);

	// Setup texture matrix for reflections
	Matrix4<float> tex = tf.getInverse();
	tex[0][3] = tex[1][3] = tex[2][3] = 0.0f;
	m_MatrixStack->setTexture(tex);

	// Render Spheres
	for (i = 0; i < numSpheres; i++) {
		*(Matrix4<float>*)g_SphereParameters[i]->getBuffer() = (Matrix4<float>)tf * g_SphereTransform[i];
		g_SphereModel[i]->render(g_SphereParameters[i]);
	}

	m_Gui->paint();
}

}; };
