/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __APPLICATION_MESSAGE_INL
#define __APPLICATION_MESSAGE_INL

namespace magnum { namespace application {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Message::Message() {
	m_Target = NULL;
	m_References = 1;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Message::Message(int cmd) {
	m_Command = cmd;
	m_Target = NULL;
	m_References = 1;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Message::Message(Handler* hnd) {
	m_Target = hnd;
	m_References = 1;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Message::Message(Handler* hnd, int cmd) {
	m_Command = cmd;
	m_Target = hnd;
	m_References = 1;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Message::Message(const Message& msg) {
	m_Command = msg.m_Command;
	m_Target = msg.m_Target;
	m_References = 1;
	cloneParameters(msg.m_Parameters);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Message::~Message() {
	clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Message::refer() {
	m_References++;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Message::release() {
	if (m_References > 1)
		m_References--;
	else
		delete this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline int Message::getCommand() const {
	return m_Command;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Message::setCommand(int cmd) {
	m_Command = cmd;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Handler* Message::getTarget() const {
	return m_Target;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Message::setTarget(Handler* hnd) {
	m_Target = hnd;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Message::Iterator Message::begin() const {
	return m_Parameters.begin();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Message::Iterator Message::end() const {
	return m_Parameters.end();
}

	
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Message::Iterator Message::at(const Parameter* p) const {
	ASSERT(p != NULL);
	ASSERT(p->getChain() == &m_Parameters);
	return m_Parameters.at(p);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool Message::contains(const char* name) const {
	return (findParameter(name) != NULL);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Message::clear() {
	m_Parameters.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool Message::isEmpty() const {
	return m_Parameters.isEmpty();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline unsigned int Message::getSize() const {
	return m_Parameters.getSize();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const Message& Message::operator=(const Message& msg) {
	clear();
	m_Target = msg.m_Target;
	m_Command = msg.m_Command;
	cloneParameters(msg.m_Parameters);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool	Message::getPointer(const char* name,void** value) const {
	Parameter *param = findParameter(name);
	if (param == NULL)
		return false;
	*value = param->m_Pointer;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void* Message::getPointer(const char* name) const {
	Parameter *param = findParameter(name);
	ASSERT(param != NULL);
	return param->m_Pointer;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool	Message::getInteger(const char* name,int* value) const {
	Parameter *param = findParameter(name);
	if (param == NULL)
		return false;
	*value = param->m_Integer;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool	Message::getFloat(const char* name,float* value) const {
	Parameter *param = findParameter(name);
	if (param == NULL)
		return false;
	*value = param->m_Float;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool	Message::getDouble(const char* name,double* value) const {
	Parameter *param = findParameter(name);
	if (param == NULL)
		return false;
	*value = param->m_Double;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool Message::getRectangle(const char* name, Rectangle<int>* value) const {
	Parameter *param = findParameter(name);
	if (param == NULL)
		return false;
	value->setX(param->m_RectangleInt[0]);
	value->setY(param->m_RectangleInt[1]);
	value->setWidth(param->m_RectangleInt[2]);
	value->setHeight(param->m_RectangleInt[3]);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool Message::getRectangle(const char* name, Rectangle<float>* value) const {
	Parameter *param = findParameter(name);
	if (param == NULL)
		return false;
	value->setX(param->m_RectangleFloat[0]);
	value->setY(param->m_RectangleFloat[1]);
	value->setWidth(param->m_RectangleFloat[2]);
	value->setHeight(param->m_RectangleFloat[3]);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool Message::getPoint(const char* name, Point<int>* value) const {
	Parameter *param = findParameter(name);
	if (param == NULL)
		return false;
	value->setX(param->m_PointInt[0]);
	value->setY(param->m_PointInt[1]);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool Message::getPoint(const char* name, Point<float>* value) const {
	Parameter *param = findParameter(name);
	if (param == NULL)
		return false;
	value->setX(param->m_PointFloat[0]);
	value->setY(param->m_PointFloat[1]);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Message::Parameter::refer() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Message::Parameter::release() {
	delete this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const char* Message::Parameter::getName() const {
	return m_Name;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Message::Type Message::Parameter::getType() const {
	return m_Type;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline int Message::Parameter::getInteger() const {
	ASSERT(m_Type == Integer);
	return m_Integer;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline float Message::Parameter::getFloat() const {
	ASSERT(m_Type == Float);
	return m_Float;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const String& Message::Parameter::getString() const {
	ASSERT(m_Type == Text);
	return *m_String;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline double Message::Parameter::getDouble() const {
	ASSERT(m_Type == Double);
	return m_Double;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void* Message::Parameter::getPointer() const {
	ASSERT(m_Type == Pointer);
	return m_Pointer;
}

}; };
#endif
