/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/
/*
 * (C) Copyright Christopher Diggins 2005
 * (C) Copyright Pablo Aguilar 2005
 * (C) Copyright Kevlin Henney 2001
 *
 * Distributed under the Boost Software License, Version 1.0. (See
 * accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef __BASE_ANY_H
#define __BASE_ANY_H

#include <typeinfo.h>
#include "object.exception.h"

namespace magnum { namespace types {
using magnum::object::Exception;

DeclareException(BadAnyCastException, Exception);

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
class Any {
public:
	inline Any();
	inline Any(const Any& x);
	template <typename T>
	inline Any(const T& );
	inline ~Any();

	// assignment operator
	template<typename T>
	Any&			operator=(const T& );
	Any&			operator=(const Any& );
	
	// comparison operators
	template<typename T>
	inline bool		operator==(const T& ) const;
	inline bool		operator==(const Any& ) const;
	template<typename T>
	inline bool		operator!=(const T& ) const;
	inline bool		operator!=(const Any& ) const;

	// Type information
	inline const type_info& getType() const;

	// Casting
	template<typename T>
	inline const T& cast() const;

	inline bool		isEmpty() const;
	inline void		clear();
	
private:
	// function pointer table
	struct fxn_ptr_table {
		const type_info& (*get_type)();
		void (*static_delete)(void**);
		void (*clone)(void* const*, void**);
		void (*move)(void* const*,void**);
		bool (*cmp_equal)(void* const*, void* const* );
		bool (*cmp_equalT)(void* const*, void const* );
		bool (*cmp_nequal)(void* const*, void* const* );
		bool (*cmp_nequalT)(void* const*, void const* );
	};

	// static functions for small value-types
	template<bool is_small>
	struct fxns {
		template<typename T>
		struct type {
			static const type_info& get_type() {
				return typeid(T);
			}
			static void static_delete(void** x) {
				reinterpret_cast<T*>(x)->~T();
			}
			static void clone(void* const* src, void** dest) {
				new(dest) T(*reinterpret_cast<T const*>(src));
			}
			static void move(void* const* src, void** dest) {
				reinterpret_cast<T*>(dest)->~T();
				*reinterpret_cast<T*>(dest) = *reinterpret_cast<T const*>(src);
			}
			static bool cmp_equal(void* const* src, void* const* dst) {
				return *reinterpret_cast<T const*>(src) == *reinterpret_cast<T const*>(dst);
			}
			static bool cmp_equalT(void* const* src, void const* dst) {
				return *reinterpret_cast<T const*>(src) == *reinterpret_cast<T const*>(dst);
			}
			static bool cmp_nequal(void* const* src, void* const* dst) {
				return *reinterpret_cast<T const*>(src) != *reinterpret_cast<T const*>(dst);
			}
			static bool cmp_nequalT(void* const* src, void const* dst) {
				return *reinterpret_cast<T const*>(src) != *reinterpret_cast<T const*>(dst);
			}
		};
	};

	// static functions for big value-types (bigger than a void*)
	template<>
	struct fxns<false> 	{
		template<typename T>
		struct type {
			static const type_info& get_type() {
				return typeid(T);
			}
			static void static_delete(void** x) {
				delete(*reinterpret_cast<T**>(x));
			}
			static void clone(void* const* src, void** dest) {
				*dest = new T(**reinterpret_cast<T* const*>(src));
			}
			static void move(void* const* src, void** dest) {
				(*reinterpret_cast<T**>(dest))->~T();
				**reinterpret_cast<T**>(dest) = **reinterpret_cast<T* const*>(src);
			}
			static bool cmp_equal(void* const* src, void* const* dst) {
				return **reinterpret_cast<T* const*>(src) == **reinterpret_cast<T* const*>(dst);
			}
			static bool cmp_equalT(void* const* src, void const* dst) {
				return **reinterpret_cast<T* const*>(src) == *reinterpret_cast<T const*>(dst);
			}
			static bool cmp_nequal(void* const* src, void* const* dst) {
				return **reinterpret_cast<T* const*>(src) != **reinterpret_cast<T* const*>(dst);
			}
			static bool cmp_nequalT(void* const* src, void const* dst) {
				return **reinterpret_cast<T* const*>(src) != *reinterpret_cast<T const*>(dst);
			}
		};
	};

	template<typename T>
	static inline const fxn_ptr_table* getTable();

	struct empty {
		inline bool operator==(const empty& ) const {
			return true;
		}
		inline bool operator!=(const empty& ) const {
			return false;
		}
	};

private:
	// fields
	const fxn_ptr_table* m_Table;
	void*			m_Object;
};

// boost::Any-like casting
template<typename T>
T* any_cast(Any* this_);

template<typename T>
T const* any_cast(Any const* this_);

template<typename T>
T const& any_cast(Any const& this_);

}; };


#endif 