/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/
/*
 * (C) Copyright Christopher Diggins 2005
 * (C) Copyright Pablo Aguilar 2005
 * (C) Copyright Kevlin Henney 2001
 *
 * Distributed under the Boost Software License, Version 1.0. (See
 * accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */
#ifndef __BASE_ANY_INL
#define __BASE_ANY_INL

namespace magnum { namespace types {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template <typename T>
inline Any::Any(const T& x) {
	m_Table = getTable<T>();
	if (sizeof(T) <= sizeof(void*))
		new(&m_Object) T(x);
	else
		m_Object = new T(x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Any::Any() {
    m_Table = getTable<empty>();
    m_Object = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Any::Any(const Any& x) {
	x.m_Table->clone(&x.m_Object, &m_Object);
	m_Table = x.m_Table;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Any::~Any() {
    m_Table->static_delete(&m_Object);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template <typename T>
Any& Any::operator=(const T& x) {
    // are we copying between the same type?
    fxn_ptr_table* x_table = getTable<T>();
    if (m_Table == x_table) {
    // if so, we can avoid deallocating and resuse memory
	    if (sizeof(T) <= sizeof(void*))
	        // create copy on-top of m_Object pointer itself
			new(&m_Object) T(x);
	    else
	        // create copy on-top of old version
			new(m_Object) T(x);
    } else {
	    clear();
	    if (sizeof(T) <= sizeof(void*)) {
		    // create copy on-top of m_Object pointer itself
	        new(&m_Object) T(x);
	        // update m_Table pointer
			m_Table = x_table;
		} else {
			m_Object = new T(x);
			m_Table = x_table;
		}
    }
    return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<typename T>
inline bool Any::operator==(const T& t) const {
	if (m_Table->getType() != typeid(T))
		return false;
	return (m_Table->cmp_equalT(&m_Object, &t));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool	Any::operator==(const Any& any) const {
	if (m_Table != any.m_Table)
		return false;
	return (m_Table->cmp_equal(&m_Object, &any.m_Object));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<typename T>
inline bool Any::operator!=(const T& t) const {
	if (m_Table->getType() != typeid(T))
		return false;
	return (m_Table->cmp_nequalT(&m_Object, &t));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool	Any::operator!=(const Any& any) const {
	if (m_Table != any.m_Table)
		return false;
	return (m_Table->cmp_nequal(&m_Object, &any.m_Object));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const type_info& Any::getType() const {
    return m_Table->get_type();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<typename T>
const T& Any::cast() const {
	ASSERT(getType() == typeid(T));
    if (getType() != typeid(T))
		throw BadAnyCastException();
    if (sizeof(T) <= sizeof(void*))
	    return *reinterpret_cast<T const*>(&m_Object);
    else
		return *reinterpret_cast<T const*>(m_Object);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool Any::isEmpty() const {
    return m_Table == getTable<empty>();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Any::clear() {
    m_Table->static_delete(&m_Object);
    m_Table = getTable<empty>();
    m_Object = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<typename T>
inline const Any::fxn_ptr_table* Any::getTable() {
	static const bool is_small = (sizeof(T) <= sizeof(void*));
	static const fxn_ptr_table static_table = {
			fxns<is_small>::type<T>::get_type
		, fxns<is_small>::type<T>::static_delete
		, fxns<is_small>::type<T>::clone
		, fxns<is_small>::type<T>::move
		, fxns<is_small>::type<T>::cmp_equal
		, fxns<is_small>::type<T>::cmp_equalT
		, fxns<is_small>::type<T>::cmp_nequal
		, fxns<is_small>::type<T>::cmp_nequalT
	};
	return &static_table;
};



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<typename T>
T* any_cast(Any* this_) {
	ASSERT(this_->getType() == typeid(T));
	if (this_->getType() != typeid(T))
		throw BadAnyCastException();
	if (sizeof(T) <= sizeof(void*))
	    return reinterpret_cast<T*>(&this_->m_Object);
	else
		return reinterpret_cast<T*>(this_->m_Object);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<typename T>
T const* any_cast(Any const* this_) {
	return any_cast<T>(const_cast<Any*>(this_));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<typename T>
T const& any_cast(Any const& this_) {
	return *any_cast<T>(const_cast<Any*>(&this_));
}

}; };


#endif
