/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __CONTAINER_ARRAY_INL
#define __CONTAINER_ARRAY_INL

namespace magnum { namespace container {
using magnum::types::byte;
using magnum::math::min;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Array<T>::Array() {
	m_Size = 0;
	m_Capacity = 0;
	m_Array = NULL;
	m_MaxGrow = 0x7fffffff;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
 inline Array<T>::~Array() {
	if (m_Array != NULL) {
		DestructElements<T>(m_Array, m_Size);
		delete[] (byte*)m_Array;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Array<T>::Array(const Array& a) {
	m_Size = 0;
	m_Capacity = 0;
	m_Array = NULL;
	m_MaxGrow = 0x7fffffff;
	operator=(a);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline unsigned int Array<T>::getSize() const {
	 return m_Size; 
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Array<T>::isEmpty() const {
	 return (m_Size == 0); 
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
const Array<T>& Array<T>::operator=(const Array<T>& src) {
	ASSERT(this != &src);   // cannot append to itself

	setSize(src.m_Size);
	CopyElements<T>(m_Array, src.m_Array, src.m_Size);

	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Array<T>::clear() {
	if (m_Array != NULL) {
		DestructElements<T>(m_Array, m_Size);
		delete[] (byte*)m_Array;
	}
	m_Size = 0;
	m_Capacity = 0;
	m_Array = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const T& Array<T>::get(unsigned int nIndex) const {
	ASSERT(nIndex < m_Size);
	return m_Array[nIndex]; 
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline void Array<T>::set(unsigned int nIndex, const T& newElement) {
	if (nIndex >= m_Size)
		setSize(nIndex + 1);
	m_Array[nIndex] = newElement; 
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const T* Array<T>::getArray() const { 
	return (const T*)m_Array; 
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T* Array<T>::getArray() { 
	return (T*)m_Array;
}	 


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const T& Array<T>::operator[](unsigned int nIndex) const {
	ASSERT(nIndex < m_Size);
	return m_Array[nIndex]; 
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T& Array<T>::operator[](unsigned int nIndex) {
	ASSERT(nIndex < m_Size);
	return m_Array[nIndex]; 
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Array<T>::append(const T* src, unsigned int size) {
	int nOldSize = m_Size;
	setSize(m_Size + size);
	CopyElements<T>(m_Array + nOldSize, src, size);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Array<T>::append(const Array& src) {
	ASSERT(this != &src);   // cannot append to itself

	int nOldSize = m_Size;
	setSize(m_Size + src.m_Size);
	CopyElements<T>(m_Array + nOldSize, src.m_Array, src.m_Size);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline unsigned int Array<T>::append(unsigned int size) { 
	setSize(m_Size + size);
	return m_Size - size; 
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Array<T>::copy(const T* src, unsigned int size) {
	setSize(size);
	CopyElements<T>(m_Array, src, size);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Array<T>::copy(const Array& src) {
	if(this == &src);
		return;

	setSize(src.m_Size);
	CopyElements<T>(m_Array, src.m_Array, src.m_Size);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Array<T>::setSize(unsigned int nNewSize) {
	if (nNewSize == 0) 	{
		// shrink to nothing
		if (m_Array != NULL) {
			DestructElements<T>(m_Array, m_Size);
			m_Size = 0;
		}
	} else if (m_Array == NULL) {
		// create one with exact size
		m_Array = (T*) new byte[nNewSize * sizeof(T)];
		ConstructElements<T>(m_Array, nNewSize);
		m_Size = m_Capacity = nNewSize;
	} else if (nNewSize <= m_Capacity) {
		// it fits
		if (nNewSize > m_Size) {
			// initialize the new elements
			ConstructElements<T>(&m_Array[m_Size], nNewSize-m_Size);
		} else if (m_Size > nNewSize) {
			// destroy the old elements
			DestructElements<T>(&m_Array[nNewSize], m_Size-nNewSize);
		}
		m_Size = nNewSize;
	} else {
		// otherwise, grow array
		unsigned int nGrowBy = (m_Size < 4) ? 4 : ((m_Size > m_MaxGrow) ? m_MaxGrow : m_Size);
		unsigned int nNewMax;
		if (nNewSize < m_Capacity + nGrowBy)
			nNewMax = m_Capacity + nGrowBy;  // granularity
		else
			nNewMax = nNewSize;  // no slush

		ASSERT(nNewMax >= m_Capacity);  // no wrap around

		T* pNewData = (T*) new byte[nNewMax * sizeof(T)];
		ASSERT(pNewData != NULL);

		// copy new data from old
		memcpy(pNewData, m_Array, m_Size * sizeof(T));

		// construct remaining elements
		ASSERT(nNewSize > m_Size);
		ConstructElements<T>(&pNewData[m_Size], nNewSize-m_Size);

		// get rid of old stuff (note: no destructors called)
		delete[] (byte*)m_Array;
		m_Array = pNewData;
		m_Size = nNewSize;
		m_Capacity = nNewMax;
	}
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline unsigned int Array<T>::getCapacity() const {
	return m_Capacity;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Array<T>::setCapacity(unsigned int nNewSize) {
	if (nNewSize == 0) 	{
		// shrink to nothing
		if (m_Array != NULL) {
			DestructElements<T>(m_Array, m_Size);
			delete[] (byte*)m_Array;
			m_Array = NULL;
		}
		m_Size = m_Capacity = 0;
	} else if (m_Array == NULL) {
		// create one with exact size
		m_Array = (T*) new byte[nNewSize * sizeof(T)];
		m_Size = 0;
		m_Capacity = nNewSize;
	} else if (nNewSize <= m_Capacity) {
		if (m_Size > nNewSize) {
			// destroy the old elements
			DestructElements<T>(&m_Array[nNewSize], m_Size-nNewSize);
		}
		m_Size = nNewSize;
	} else {
		T* pNewData = (T*) new byte[nNewSize * sizeof(T)];
		ASSERT(pNewData != NULL);

		// copy new data from old
		m_Size = min(m_Size, nNewSize);
		memcpy(pNewData, m_Array, m_Size * sizeof(T));

		// get rid of old stuff (note: no destructors called)
		delete[] (byte*)m_Array;
		m_Array = pNewData;
		m_Capacity = nNewSize;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Array<T>::shrink() {
	if (m_Size != m_Capacity) {
		// shrink to desired size
		T* pNewData = NULL;
		if (m_Size != 0) {
			pNewData = (T*) new byte[m_Size * sizeof(T)];
			// copy new data from old
			memcpy(pNewData, m_Array, m_Size * sizeof(T));
		}

		// get rid of old stuff (note: no destructors called)
		delete[] (byte*)m_Array;
		m_Array = pNewData;
		m_Capacity = m_Size;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
bool Array<T>::contains(const T& item) const {
	for (unsigned int i = 0; i < m_Size; i++) {
		if (m_Array[i] == item)
			return true;
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline unsigned int Array<T>::insert(const T& newElement) { 
	setSize(m_Size + 1);
	m_Array[m_Size - 1] = newElement;
	return m_Size - 1; 
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Array<T>::insert(unsigned int nIndex, const T& newElement) {
	if (nIndex >= m_Size) {
		// adding after the end of the array
		setSize(nIndex + 1);   // grow so nIndex is valid
	} else {
		// inserting in the middle of the array
		int nOldSize = m_Size;
		setSize(m_Size + 1);  // grow it to new size
		// destroy intial data before copying over it
		DestructElement<T>(&m_Array[nOldSize]);
		// shift old data up to fill gap
		memmove(&m_Array[nIndex+1], &m_Array[nIndex], (nOldSize-nIndex) * sizeof(T));

		// re-init slots we copied from
		ConstructElement<T>(&m_Array[nIndex]);
	}

	// insert new value in the gap
	ASSERT(nIndex < m_Size);
	m_Array[nIndex] = newElement;
}

 
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Array<T>::remove(unsigned int nIndex) {
	ASSERT(nIndex < m_Size);

	// just remove a range
	unsigned int nMoveCount = m_Size - (nIndex + 1);
	DestructElement<T>(&m_Array[nIndex]);
	if (nMoveCount)
		memmove(&m_Array[nIndex], &m_Array[nIndex + 1], nMoveCount * sizeof(T));
	m_Size--;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Array<T>::remove(typename const ArrayIterator<T>& iter) {
	T *ptr = iter.m_Item;
	ASSERT((ptr >= m_Array) && (ptr < m_Array+m_Size));

	int nMoveCount = (int)((m_Array - ptr) + m_Size);
	DestructElement<T>(ptr);
	if (nMoveCount)
		memmove(ptr, ptr, nMoveCount * sizeof(T));
	m_Size--;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Array<T>::insert(unsigned int nIndex, const Array& pNewArray) {
	GASSERT(nStartIndex >= 0);

	if (pNewArray.m_Size > 0) {
		if (nIndex >= m_Size) {
			// adding after the end of the array
			setSize(nIndex + pNewArray.m_Size);   // grow so nIndex is valid
		} else {
			// inserting in the middle of the array
			int nOldSize = m_Size;
			setSize(m_Size + pNewArray.m_Size);  // grow it to new size
			// destroy intial data before copying over it
			DestructElements<T>(&m_Array[nOldSize], pNewArray.m_Size);
			// shift old data up to fill gap
			memmove(&m_Array[nIndex+pNewArray.m_Size], &m_pData[nIndex], (nOldSize-nIndex) * sizeof(T));

			// re-init slots we copied from
			ConstructElements<T>(&m_Array[nIndex], pNewArray.m_Size);
		}

		// insert new value in the gap
		ASSERT(nIndex +  pNewArray.m_Size <= m_Size);
		CopyElements(&m_Array[nIndex], pNewArray.m_Array, pNewArray.m_Size);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T> Array<T>::begin() const {
	return Iterator(m_Array);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T> Array<T>::end() const {
	return Iterator(m_Array + m_Size);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T> Array<T>::at(unsigned int i) const {
	return Iterator(m_Array + min(m_Size, i));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T> Array<T>::rbegin() const {
	return Iterator(m_Array + m_Size - 1);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T> Array<T>::rend() const {
	return Iterator(m_Array - 1);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T>::ArrayIterator() {
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T>::ArrayIterator(T* item) {
	m_Item = item;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T>::operator T*() {
	return m_Item;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T>::operator const T*() const {
	return m_Item;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T& ArrayIterator<T>::operator*() const {
	return *m_Item;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T* ArrayIterator<T>::operator->() const {
	return m_Item;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T>& ArrayIterator<T>::operator++() {
	ASSERT(m_Item != NULL);
	m_Item++;
	return *this;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T> ArrayIterator<T>::operator++(int ) {
	ASSERT(m_Item != NULL);
	T* item = m_Item;
	m_Item++;
	return ArrayIterator(item);
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T>& ArrayIterator<T>::operator+=(int n) {
	m_Item += n;
	return *this;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T> ArrayIterator<T>::operator+(int n) const {
	return ArrayIterator(m_Item + n);
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T>& ArrayIterator<T>::operator--() {
	ASSERT(m_Item != NULL);
	m_Item--;
	return *this;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T>	ArrayIterator<T>::operator--(int ) {
	ASSERT(m_Item != NULL);
	T* item = m_Item;
	m_Item--;
	return ArrayIterator(item);
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T>& ArrayIterator<T>::operator-=(int n) {
	m_Item -= n;
	return *this;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline ArrayIterator<T> ArrayIterator<T>::operator-(int n) const {
	return ArrayIterator(m_Item - n);
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool	ArrayIterator<T>::operator==(const ArrayIterator<T>& iter) const {
	return m_Item == iter.m_Item;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool	ArrayIterator<T>::operator<(const ArrayIterator<T>& iter) const {
	return m_Item < iter.m_Item;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool	ArrayIterator<T>::operator<=(const ArrayIterator<T>& iter) const {
	return m_Item <= iter.m_Item;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool	ArrayIterator<T>::operator>(const ArrayIterator<T>& iter) const {
	return m_Item > iter.m_Item;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool	ArrayIterator<T>::operator>=(const ArrayIterator<T>& iter) const {
	return m_Item >= iter.m_Item;
};

}; };
#endif
