/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __CONTAINER_BINARYTREE_INL
#define __CONTAINER_BINARYTREE_INL

namespace magnum { namespace container {


/*----------------------------------------------------------------------------------
Description:
	Constructs a new and empty Binary search tree.
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline BinaryTree<Key, Item>::BinaryTree() {
	m_Root = NULL;
}


/*----------------------------------------------------------------------------------
Description:
	Destroys the binary tree. All nodes will be removed from memory.
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline BinaryTree<Key, Item>::~BinaryTree() {
	clear();
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline Item BinaryTree<Key, Item>::getRoot() const {
	ASSERT(m_Root != NULL);
	return m_Root->m_Item;
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline bool BinaryTree<Key, Item>::isEmpty() const {
	return (m_Root == NULL);
}


/*----------------------------------------------------------------------------------
Description:
	Clears the binary tree, that means, that all nodes will be removed from
	memory, and the tree will be empty again.
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline void BinaryTree<Key, Item>::clear() {
	GNode *node = m_Root;
	GNode *father;

	while (node != NULL) {
		if (node->m_LSon != NULL) {
			node = node->m_LSon;
		} else if (node->m_RSon != NULL) {
			node = node->m_RSon;
		} else {
			father = node->m_Father;
			delete node;

			// delete pointer in father node to this node
			if (father != NULL) {
				if (father->m_LSon == node)
					father->m_LSon = NULL;
				else
					father->m_RSon = NULL;
			}

			node = father;
		}
	}

	m_Root = NULL;
}



/*----------------------------------------------------------------------------------
Description:
	Inserts a new item into the binary tree. If there is already an item with 
	the same key, then the item will be substituted by the new item

Parameters:
	item - item to insert
	key - key to associate with the item

Remarks:
	Only one item can be associated with one key, although more than one
	key could be associated with one item. When you insert an item with
	an already existing key, the previosu item associtaed with the key 
	will be substituted.
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline bool BinaryTree<Key, Item>::insert(const Item& item,const Key& key) {
	Node *node = m_Root;
	Node *newNode;
	
	if (node == NULL) {
		m_Root = new GNode;
		m_Root->m_LSon =
		m_Root->m_RSon =
		m_Root->m_Father = NULL;
		m_Root->m_Key = key;
		m_Root->m_Item = item;
		return true;
	}	

	for (;;) {
		if (node->m_Key == key) {
			node->m_Item = item;
			return true;
		}
		if (node->m_Key < key) {
			if (node->m_RSon == NULL) {
				newNode = node->m_RSon = new GNode;
				newNode->m_Father = node;
				newNode->m_Key = key;
				newNode->m_Item = item;
				newNode->m_LSon =
				newNode->m_RSon = NULL;
				return true;
			} else {
				node = node->m_RSon;
			}
		} else {
			if (node->m_LSon == NULL) {
				newNode = node->m_LSon = new GNode;
				newNode->m_Father = node;
				newNode->m_Key = key;
				newNode->m_Item = item;
				newNode->m_LSon =
				newNode->m_RSon = NULL;
				return true;
			} else {
				node = node->m_LSon;
			}
		}
	}
}



/*----------------------------------------------------------------------------------
Description:
	Removes an item from the binary tree. This method uses the key in order
	to determine the item to delete. When one item is associated with more
	than one key, then only one duplicate of the same items will be deleted.

Parameters:
	key - key associated to the item to delete

Returns:
	*true* when the key was found and removed. When the key was not
	found, the method will return *false*.
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline bool BinaryTree<Key, Item>::remove(const Key& key) {
	Node *node = m_Root;
	Node **father = &m_Root;

	// first search the node
	while (node != NULL) {
		if (node->m_Key == key) {
			removeNode(node, father);
			return true;
		}
		if (node->m_Key < key) {
			father = &node->m_RSon;
			node = node->m_RSon;
		} else {
			father = &node->m_LSon;
			node = node->m_LSon;
		}
	}

	return false;
}


/*----------------------------------------------------------------------------------
Description:
	Removes an item from the binary tree. This method searches the whole tree
	for the specified item by examining all nodes - this is an expensive 
	operation!

Parameters:
	item - item to delete

Returns:
	*true* when the item was found and removed. When the item was not
	found, the method will return *false*.
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline bool BinaryTree<Key, Item>::remove(const Item& item) {
	Node *node = m_Root;
	Node **father = &m_Root;

	if (node == NULL)
		return false;

	for(;;) {
		if (node->m_Item == item) {
			removeNode(node,father);
			return true;
		}

		if (node->m_LSon != NULL) {
			// walk down left nodes...
			father = &node->m_LSon;
			node = node->m_LSon;
		} else 
		if (node->m_RSon != NULL) {
			// walk down right nodes
			father = &node->m_RSon;
			node = node->m_RSon;
		} else 
		// walk up
		for(;;) {
			if (node->m_Father == NULL)
				return false;
			if (node->m_Father->m_LSon == node) {
				// Okay, we have been a left son, so walk up
				// and try the right son of the father node...
				node = node->m_Father;
				if (node->m_RSon != NULL) {
					father = &node->m_RSon;
					node = node->m_RSon;
					break;
				} 
			} else {
				node = node->m_Father;
			}
		}
	}
}


/*----------------------------------------------------------------------------------
Description:
	This method seaches the tree for the given key and saves the found item
	into item. When no appropriate item is found, the method returns false.

Parameters:
	key - key to search for in the tree
	item - item to save the fpund item into

Returns:
	*true* when they key could be found in the tree.
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline bool BinaryTree<Key, Item>::get(const Key& key, Item& item) const {
	Node *node = m_Root;

	while (node != NULL) {
		if (node->m_Key == key) {
			item = node->m_Item;
			return true;
		}
		if (node->m_Key < key) {
			node = node->m_RSon;
		} else {
			node = node->m_LSon;
		}
	}

	return false;
}



/*----------------------------------------------------------------------------------
Description:
	This method searches the tree for a key associated with an item,. This is 
	a really expensive operation, as all tree nodes have to be visited in order
	to find the key for the item. If more than one key is associated with the
	specific item, then only one key is returned.

Parameters:
	item - item to search in the tree
	key - placeholder for the key, if the item was found in the tree.

Returns:
	*true* if the item could be found inside the tree.
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline bool BinaryTree<Key, Item>::get(const Item& item, Key &key) const {
	Node *node = m_Root;

	if (node == NULL)
		return false;

	for(;;) {
		if (node->m_Item == item) {
			key = node->m_Key;
			return true;
		}

		if (node->m_LSon != NULL) {
			// walk down left nodes...
			node = node->m_LSon;
		} else 
		if (node->m_RSon != NULL) {
			// walk down right nodes
			node = node->m_RSon;
		} else 
		// walk up
		for(;;) {
			if (node->m_Father == NULL)
				return false;
			if (node->m_Father->m_LSon == node) {
				// Okay, we have been a left son, so walk up
				// and try the right son of the father node...
				node = node->m_Father;
				if (node->m_RSon != NULL) {
					node = node->m_RSon;
					break;
				} 
			} else {
				node = node->m_Father;
			}
		}
	}
}



/*----------------------------------------------------------------------------------
Description:
	This method seaches the tree for the given key and saves the found item
	into item. 

Parameters:
	key - key to search for in the tree

Returns:
	A pointer to the item within the tree, or NULL on failure.
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline Item* BinaryTree<Key, Item>::get(const Key& key) const {
	Node *node = m_Root;

	while (node != NULL) {
		if (node->m_Key == key) {
			return &node->m_Item;
		}
		if (node->m_Key < key) {
			node = node->m_RSon;
		} else {
			node = node->m_LSon;
		}
	}

	return NULL;
}



/*----------------------------------------------------------------------------------
Description:
	This method searches the tree for a key associated with an item,. This is 
	a really expensive operation, as all tree nodes have to be visited in order
	to find the key for the item. If more than one key is associated with the
	specific item, then only one key is returned.

Parameters:
	item - item to search in the tree

Returns:
	A pointer to the key in the tree, or NULL on failure
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline const Key* BinaryTree<Key, Item>::get(const Item& item) const {
	Node *node = m_Root;

	if (node == NULL)
		return NULL;

	for(;;) {
		if (node->m_Item == item) {
			return &node->m_Key;
		}

		if (node->m_LSon != NULL) {
			// walk down left nodes...
			node = node->m_LSon;
		} else 
		if (node->m_RSon != NULL) {
			// walk down right nodes
			node = node->m_RSon;
		} else 
		// walk up
		for(;;) {
			if (node->m_Father == NULL)
				return NULL;
			if (node->m_Father->m_LSon == node) {
				// Okay, we have been a left son, so walk up
				// and try the right son of the father node...
				node = node->m_Father;
				if (node->m_RSon != NULL) {
					node = node->m_RSon;
					break;
				} 
			} else {
				node = node->m_Father;
			}
		}
	}
}


/*----------------------------------------------------------------------------------
Description:
	This method checks if the given item is inside the tree. Note that this
	method is rather slow because the tree is organised for optimal search
	for keys and not for items.

Parameters:
	item - item to search in the tree.

Returns:
	*true* if the item was found in the tree
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline bool BinaryTree<Key, Item>::contains(const Item& item) const {
	Node *node = m_Root;

	if (node == NULL)
		return false;

	for(;;) {
		if (node->m_Item == item) {
			return true;
		}

		if (node->m_LSon != NULL) {
			// walk down left nodes...
			node = node->m_LSon;
		} else 
		if (node->m_RSon != NULL) {
			// walk down right nodes
			node = node->m_RSon;
		} else 
		// walk up
		for(;;) {
			if (node->m_Father == NULL)
				return false;
			if (node->m_Father->m_LSon == node) {
				// Okay, we have been a left son, so walk up
				// and try the right son of the father node...
				node = node->m_Father;
				if (node->m_RSon != NULL) {
					node = node->m_RSon;
					break;
				} 
			} else {
				node = node->m_Father;
			}
		}
	}
}


/*----------------------------------------------------------------------------------
Description:
	This method checks if the given key is inside the tree. 

Parameters:
	key - key to search in the tree.

Returns:
	*true* if the key was found in the tree
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline bool BinaryTree<Key, Item>::contains(const Key& key) const {
	Node *node = m_Root;

	while (node != NULL) {
		if (node->m_Key == key) {
			return true;
		}
		if (node->m_Key < key) {
			node = node->m_RSon;
		} else {
			node = node->m_LSon;
		}
	}

	return false;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline typename BinaryTree<Key, Item>::Iterator BinaryTree<Key, Item>::begin() const {
	Iterator iter;

	iter.m_Node = m_Tree->m_Root;
	if (iter.m_Node == NULL)
		return iter;

	while (iter.m_Node->m_LSon != NULL)
		iter.m_Node = iter.m_Node->m_LSon;

	return iter;
};


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline typename BinaryTree<Key, Item>::Iterator BinaryTree<Key, Item>::end() const {
	Iterator iter;
	
	iter.m_Node = NULL;

	return iter;
};


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class Key, class Item> 
inline void BinaryTree<Key, Item>::removeNode(typename BinaryTree<Key, Item>::Node *node, typename BinaryTree<Key, Item>::Node **father) {
	Node *cur;

	// No left son of the node? then we can simply remove 
	// the node
	if (node->m_LSon == NULL) {
		if (node->m_RSon != NULL)
			node->m_RSon->m_Father = node->m_Father;
		*father  = node->m_RSon;
		delete node;
		return;
	} 

	// No right son of the node? again, then we can simply
	// remove the node...
	if (node->m_RSon == NULL) {
		node->m_LSon->m_Father = node->m_Father;
		*father = node->m_LSon;
		delete node;
		return;
	} 

	cur = node->m_RSon;
	if (cur->m_LSon == NULL) {
		cur->m_Father = node->m_Father;
		cur->m_LSon = node->m_LSon;
		cur->m_LSon->m_Father = cur;
		*father = cur;
		delete node;
		return;
	}

	while (cur->m_LSon != NULL) 
		cur = cur->m_LSon;

	cur->m_Father->m_LSon = cur->m_RSon;
	if (cur->m_RSon != NULL)
		cur->m_RSon->m_Father = cur->m_Father;
	cur->m_Father = node->m_Father;
	cur->m_LSon = node->m_LSon;
	cur->m_LSon->m_Father = cur;
	cur->m_RSon = node->m_RSon;
	cur->m_RSon->m_Father = cur;
	*father = cur;
	delete node;
}


}; };
#endif
