/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __CONTAINER_QUEUE_INL
#define __CONTAINER_QUEUE_INL

namespace magnum { namespace container {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Queueable<T>::Queueable() {
	m_Prev = NULL;
	m_Next = NULL;
	m_Queue = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Queueable<T>::~Queueable() {
	if (m_Queue == NULL)
		return;
	
	m_Prev->m_Next = m_Next;
	m_Next->m_Prev = m_Prev;
	m_Queue->m_Size--;
	if (m_Queue->m_Head == this)
		m_Queue->m_Head = m_Next;
	if (m_Queue->m_Tail == this)
		m_Queue->m_Tail = m_Prev;

	m_Prev = NULL;
	m_Next = NULL;
	m_Queue = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Queue<T>::Queue() {
	m_Head = m_Tail = NULL;
	m_Size = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Queue<T>::~Queue() {
	clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Queue<T>::contains(const T* item) const {
	return (((Queueable<T>*)item)->m_Queue == this);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Queue<T>::insert(T* item) {
	ASSERT(((Queueable<T>*)item)->m_Queue == NULL);
	((Queueable<T>*)item)->m_Queue = this;
	((Queueable<T>*)item)->m_Next = NULL;
	((Queueable<T>*)item)->m_Prev = m_Tail;
	if (m_Tail == NULL) {
		m_Head = (Queueable<T>*)item;
	} else {
		m_Tail->m_Next = (Queueable<T>*)item;
	}
	m_Tail = (Queueable<T>*)item;
	item->refer();
	m_Size++;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Queue<T>::insertHead(T* item) {
	ASSERT(((Queueable<T>*)item)->m_Queue == NULL);
	((Queueable<T>*)item)->m_Queue = this;
	((Queueable<T>*)item)->m_Next = m_Head;
	((Queueable<T>*)item)->m_Prev = NULL;
	if (m_Head == NULL) {
		m_Tail = (Queueable<T>*)item;
	} else {
		m_Head->m_Prev = (Queueable<T>*)item;
	}
	m_Head = (Queueable<T>*)item;
	item->refer();
	m_Size++;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
bool Queue<T>::remove(T* item) {
	ASSERT(((Queueable<T>*)item)->m_Queue == this);
	if (m_Head == item) {
		m_Head = ((Queueable<T> *)item)->m_Next;
	}
	if (m_Tail == item) {
		m_Tail = ((Queueable<T> *)item)->m_Prev;
	}
	if (((Queueable<T> *)item)->m_Prev != NULL) {
		((Queueable<T> *)item)->m_Prev->m_Next
			= ((Queueable<T> *)item)->m_Next;
	}
	if (((Queueable<T> *)item)->m_Next != NULL) {
		((Queueable<T> *)item)->m_Next->m_Prev
			= ((Queueable<T> *)item)->m_Prev;
	}
	((Queueable<T> *)item)->m_Next = NULL;
	((Queueable<T> *)item)->m_Prev = NULL;
	((Queueable<T> *)item)->m_Queue = NULL;
	item->release();
	m_Size--;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Queue<T>::clear() {
	T* node;
	while (m_Head) {
		node = (T *)m_Head;
		m_Head = ((Queueable<T> *)m_Head)->m_Next;
		((Queueable<T> *)node)->m_Next = NULL;
		((Queueable<T> *)node)->m_Prev = NULL;
		((Queueable<T> *)node)->m_Queue = NULL;
		node->release();
	}

	m_Head = m_Tail = NULL;
	m_Size = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Queue<T>::isEmpty() const {
	return (m_Head == NULL);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline unsigned int Queue<T>::getSize() const {
	return m_Size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T* Queue<T>::removeHead() {
	if (m_Head == NULL)
		return NULL;

	T* item;
	item = (T *)m_Head;
	m_Head = m_Head->m_Next;
	if (m_Head == NULL)
		m_Tail = NULL;
	else
		m_Head->m_Prev = NULL;

	m_Size--;

	return item;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T* Queue<T>::peekHead() const {
	return (T *)m_Head;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T* Queue<T>::removeTail() {
	if (m_Tail == NULL)
		return NULL;

	Queueable<T>* item;
	item = m_Tail;
	m_Tail = m_Tail->m_Prev;
	if (m_Tail == NULL)
		m_Head = NULL;
	else
		m_Tail->m_Next = NULL;

	m_Size--;

	return (T*)item;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T* Queue<T>::peekTail() const {
	return (T *)m_Tail;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline typename Queue<T>::Iterator Queue<T>::begin() const {
	return Iterator(m_Head);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline typename Queue<T>::Iterator Queue<T>::end() const {
	return Iterator(NULL);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline typename Queue<T>::Iterator Queue<T>::at(T* i) const {
	return Iterator((Queueable<T>*)i);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline typename Queue<T>::Iterator Queue<T>::rbegin() const {
	return Iterator(NULL);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline typename Queue<T>::Iterator Queue<T>::rend() const {
	return Iterator(m_Tail);
}

}; };
#endif
