/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __CONTAINER_SET_INL
#define __CONTAINER_SET_INL

namespace magnum { namespace container {

/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T> 
inline Set<T>::Set() {
	m_Root = NULL;
	m_Size = 0;
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T>
inline Set<T>::~Set() {
	clear();
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T> 
inline bool Set<T>::isEmpty() const {
	return (m_Root == NULL);
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T> 
inline unsigned int Set<T>::getSize() const {
	return m_Size;
}


/*----------------------------------------------------------------------------------
Description:
	Clears the binary tree, that means, that all nodes will be removed from
	memory, and the tree will be empty again.
----------------------------------------------------------------------------------*/
template<class T> 
void Set<T>::clear() {
	SetNode<T> *node = m_Root;
	SetNode<T> *father;

	while (node != NULL) {
		if (node->m_LSon != NULL) {
			node = node->m_LSon;
		} else if (node->m_RSon != NULL) {
			node = node->m_RSon;
		} else {
			father = node->m_Father;
			node->~SetNode<T>();
			operator delete (node,m_Allocator);

			// delete pointer in father node to this node
			if (father != NULL) {
				if (father->m_LSon == node)
					father->m_LSon = NULL;
				else
					father->m_RSon = NULL;
			}

			node = father;
		}
	}

	m_Root = NULL;
	m_Size = 0;
}



/*----------------------------------------------------------------------------------
Description:
	Inserts a new item into the binary tree. If there is already an item with 
	the same key, then the item will be substituted by the new item

Parameters:
	item - item to insert
	key - key to associate with the item

Remarks:
	Only one item can be associated with one key, although more than one
	key could be associated with one item. When you insert an item with
	an already existing key, the previosu item associtaed with the key 
	will be substituted.
----------------------------------------------------------------------------------*/
template<class T> 
SetIterator<T> Set<T>::insert(const T& item) {
	SetNode<T> *node = m_Root;
	SetNode<T> *newNode;
	
	if (node == NULL) {
		m_Root = new(m_Allocator) SetNode<T>;
		m_Root->m_LSon =
		m_Root->m_RSon =
		m_Root->m_Father = NULL;
		m_Root->m_Item = item;
		m_Root->m_Rand = rand();
		m_Size++;
		return SetIterator<T>(m_Root);
	}	

	for (;;) {
		if (node->m_Item == item) {
			node->m_Item = item;
			return SetIterator<T>(node);
		}
		if (node->m_Item < item) {
			if (node->m_RSon == NULL) {
				newNode = node->m_RSon = new(m_Allocator) SetNode<T>;
				newNode->m_Father = node;
				newNode->m_Item = item;
				newNode->m_LSon =
				newNode->m_RSon = NULL;
				newNode->m_Rand = rand();
				repairUp(newNode);
				m_Size++;
				return SetIterator<T>(newNode);
			} else {
				node = node->m_RSon;
			}
		} else {
			if (node->m_LSon == NULL) {
				newNode = node->m_LSon = new(m_Allocator) SetNode<T>;
				newNode->m_Father = node;
				newNode->m_Item = item;
				newNode->m_LSon =
				newNode->m_RSon = NULL;
				newNode->m_Rand = rand();
				repairUp(newNode);
				m_Size++;
				return SetIterator<T>(newNode);
			} else {
				node = node->m_LSon;
			}
		}
	}
}



/*----------------------------------------------------------------------------------
Description:
	Removes an item from the binary tree. This method uses the key in order
	to determine the item to delete. When one item is associated with more
	than one key, then only one duplicate of the same items will be deleted.

Parameters:
	key - key associated to the item to delete

Returns:
	*true* when the key was found and removed. When the key was not
	found, the method will return *false*.
----------------------------------------------------------------------------------*/
template<class T>
bool Set<T>::remove(const T& item) {
	SetNode<T> *node = m_Root;
	SetNode<T> **father = &m_Root;

	// first search the node
	while (node != NULL) {
		if (node->m_Item == item) {
			removeNode(node, father);
			return true;
		}
		if (node->m_Item < item) {
			father = &node->m_RSon;
			node = node->m_RSon;
		} else {
			father = &node->m_LSon;
			node = node->m_LSon;
		}
	}
	
	return false;
}


/*----------------------------------------------------------------------------------
Description:
	Removes an item from the binary tree. This method uses the key in order
	to determine the item to delete. When one item is associated with more
	than one key, then only one duplicate of the same items will be deleted.

Parameters:
	key - key associated to the item to delete

Returns:
	*true* when the key was found and removed. When the key was not
	found, the method will return *false*.
----------------------------------------------------------------------------------*/
template<class T> 
bool Set<T>::remove(const typename SetIterator<T>& item) {
	SetNode<T> *node = item.m_Item;
	SetNode<T> **father;
	if (node->m_Father != NULL) {
		if (node->m_Father->m_LSon == node)
			father = &node->m_Father->m_LSon;
		else
			father = &node->m_Father->m_RSon;
	} else {
		father = &m_Root;
	}
	removeNode(node, father);
	return true;
}


/*----------------------------------------------------------------------------------
Description:
	This method checks if the given key is inside the tree. 

Parameters:
	key - key to search in the tree.

Returns:
	*true* if the key was found in the tree
----------------------------------------------------------------------------------*/
template<class T> 
inline bool Set<T>::contains(const T& item) const {
	SetNode<T> *node = m_Root;

	while (node != NULL) {
		if (node->m_Item == item) {
			return true;
		}
		if (node->m_Item < item) {
			node = node->m_RSon;
		} else {
			node = node->m_LSon;
		}
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline typename SetIterator<T> Set<T>::begin() const {
	SetNode<T> *node = m_Root;

	if (node == NULL)
		return SetIterator<T>(NULL);

	while (node->m_LSon != NULL)
		node = node->m_LSon;

	return SetIterator<T>(node);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline typename SetIterator<T> Set<T>::end() const {
	return SetIterator<T>(NULL);
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T> 
inline typename SetIterator<T> Set<T>::at(const T& item) const {
	SetNode<T> *node = m_Root;

	while (node != NULL) {
		if (node->m_Item == item) {
			iter.m_Node = node;
			return SetIterator<T>(node);
		}
		if (node->m_Item < item) {
			node = node->m_RSon;
		} else {
			node = node->m_LSon;
		}
	}

	return SetIterator<T>(NULL);
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T> 
inline bool Set<T>::operator==(const Set<T>& set) const {
	if (m_Size != set.m_Size)
		return false;

	SetIterator<T> i1(m_Root);
	SetIterator<T> i2(set.m_Root);

	while((i1.m_Node != NULL) && (i2.m_Node != NULL)) {
		if (i1->m_Item != i2->m_Item)
			return false;
		i1++;
		i2++;
	}

	return true;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T> 
inline bool Set<T>::operator!=(const Set<T>& set) const {
	if (m_Size != set.m_Size)
		return true;

	SetIterator<T> i1(m_Root);
	SetIterator<T> i2(set.m_Root);

	while((i1.m_Node != NULL) && (i2.m_Node != NULL)) {
		if (i1->m_Item != i2->m_Item)
			return true;
		i1++;
		i2++;
	}

	return false;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T> 
void Set<T>::removeNode(SetNode<T> *node, SetNode<T> **father) {
	SetNode<T> *cur;

	// No left son of the node? then we can simply remove 
	// the node
	if (node->m_LSon == NULL) {
		if (node->m_RSon != NULL)
			node->m_RSon->m_Father = node->m_Father;
		*father  = node->m_RSon;
		node->~SetNode<T>();
		operator delete (node,m_Allocator);
		m_Size--;
		return;
	} 

	// No right son of the node? again, then we can simply
	// remove the node...
	if (node->m_RSon == NULL) {
		node->m_LSon->m_Father = node->m_Father;
		*father = node->m_LSon;
		node->~SetNode<T>();
		operator delete (node,m_Allocator);
		m_Size--;
		return;
	} 

	// Move right son to node's position
	cur = node->m_RSon;
	cur->m_Father = node->m_Father;
	*father = cur;

	while (cur->m_LSon != NULL) 
		cur = cur->m_LSon;

	cur->m_LSon = node->m_LSon;
	cur->m_LSon->m_Father = cur;
	node->~SetNode<T>();
	operator delete (node, m_Allocator);
	m_Size--;
	repairUp(cur->m_LSon);
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T> 
void Set<T>::repairUp(SetNode<T> *node) {
	SetNode<T> *father, *father2;

	father = node->m_Father;
	ASSERT(father != NULL);

	for (;;) {
		// Is heap property okay?
		if (father->m_Rand >= node->m_Rand)
			return;

		// Okay we have to repair the tree
		if (father->m_LSon == node) {
			node->m_Father = father->m_Father;
			father->m_Father = node;
			father->m_LSon = node->m_RSon;
			if (father->m_LSon != NULL)
				father->m_LSon->m_Father = father;
			node->m_RSon = father;

			// Do we have to continue repairing up?
			father2 = node->m_Father;
			if (father2 == NULL) {
				m_Root = node;
				return;
			} else {
				// Set pointers to new father
				if (father2->m_LSon == father)
					father2->m_LSon = node;
				else 
					father2->m_RSon = node;
				father = father2;
			}
		} else {
			node->m_Father = father->m_Father;
			father->m_Father = node;
			father->m_RSon = node->m_LSon;
			if (father->m_RSon != NULL)
				father->m_RSon->m_Father = father;
			node->m_LSon = father;

			// Do we have to continue repairing up?
			father2 = node->m_Father;
			if (father2 == NULL) {
				m_Root = node;
				return;
			} else {
				// Set pointers to new father
				if (father2->m_LSon == father)
					father2->m_LSon = node;
				else 
					father2->m_RSon = node;
				father = father2;
			}
		}
	}
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T> unsigned int Set<T>::rand() {
	m_Rand = (m_Rand * 214013) + 2531011;
	return (m_Rand  >> 16) & 0x7fff;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T>
inline SetIterator<T>::SetIterator() {
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T>
inline SetIterator<T>::operator T*() {
	return &m_Node->m_Item;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T>
inline SetIterator<T>::operator const T*() const {
	return &m_Node->m_Item;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T>
inline T& SetIterator<T>::operator*() const {
	return m_Node->m_Item;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T>
inline T* SetIterator<T>::operator->() const {
	return &m_Node->m_Item;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T>
SetIterator<T>& SetIterator<T>::operator++() {
	ASSERT(m_Node != NULL);
	// Try to get the next value
	if (m_Node->m_RSon != NULL) {
		m_Node = m_Node->m_RSon;
		while (m_Node->m_LSon != NULL)
			m_Node = m_Node->m_LSon;
		return *this;
	}

	for (;;) {
		SetNode<T>* father = m_Node->m_Father;
		if (father == NULL) {
			m_Node = NULL;
			return *this;
		}
		if (father->m_LSon == m_Node) {
			m_Node = father;
			return *this;
		}
		m_Node = father;
	}

	return *this;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T>
inline SetIterator<T> SetIterator<T>::operator++(int ) {
	ASSERT(m_Node != NULL);
	SetNode<T>* node = m_Node;
	operator++();
	return SetIterator<T>(node);
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T>
SetIterator<T>& SetIterator<T>::operator--() {
	ASSERT(m_Node != NULL);

	// Try to get the next value
	if (m_Node->m_LSon != NULL) {
		m_Node = m_Node->m_LSon;
		while (m_Node->m_RSon != NULL)
			m_Node = m_Node->m_RSon;
		return *this;
	}

	for (;;) {
		SetNode<T>* father = m_Node->m_Father;
		if (father == NULL) {
			m_Node = NULL;
			return *this;
		}
		if (father->m_RSon == m_Node) {
			m_Node = father;
			return *this;
		}
		m_Node = father;
	}

	return *this;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T>
inline SetIterator<T> SetIterator<T>::operator--(int ) {
	ASSERT(m_Node != NULL);
	SetNode<T>* node = m_Node;
	operator--();
	return SetIterator<T>(node);
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T>
inline bool	SetIterator<T>::operator==(const SetIterator<T>& iter) const {
	return m_Node == iter.m_Node;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T>
inline bool	SetIterator<T>::operator<(const SetIterator<T>& iter) const {
	return m_Node < iter.m_Node;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T>
inline SetIterator<T>::	SetIterator(SetNode<T>* item) {
	m_Node = item;
}

}; };

#endif
