/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GEOMETRY_ALIGNEDBOX_INL
#define __GEOMETRY_ALIGNEDBOX_INL

namespace magnum { namespace geometry {
using magnum::math::max;
using magnum::math::min;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline AlignedBox<T,dim>::AlignedBox() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline AlignedBox<T,dim>::AlignedBox(const Vector<T,dim>& min, const Vector<T,dim>& max) {
	m_Minimum = min;
	m_Maximum = max;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void AlignedBox<T,dim>::setMinimum(const Vector<T,dim>& m) {
	m_Minimum = m;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Vector<T,dim>&	AlignedBox<T,dim>::getMinimum() const {
	return m_Minimum;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void AlignedBox<T,dim>::setMaximum(const Vector<T,dim>& m) {
	m_Maximum = m;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Vector<T,dim>& AlignedBox<T,dim>::getMaximum() const {
	return m_Maximum;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void AlignedBox<T,dim>::setOrigin(const Vector<T,dim>& o) {
	Vector<T,dim> v = o - (m_Minimum + m_Maximum)/T(2);
	m_Minimum += o;
	m_Maximum += o;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Vector<T,dim> AlignedBox<T,dim>::getOrigin() const {
	return (m_Minimum + m_Maximum)/T(2);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void AlignedBox<T,dim>::empty() {
	m_Maximum = m_Minimum = getOrigin();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool AlignedBox<T,dim>::isEmpty() const {
	return (m_Minimum == m_Maxium);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool AlignedBox<T,dim>::isInside(const Vector<T,dim>& v) const {
	for (unsigned int i = 0; i < dim; i++) {
		if ((v[i] < m_Minimum[i]) || (v[i] > m_Maximum[i]))
			return false;
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
T AlignedBox<T,dim>::getDistance(const Vector<T,dim>& o) const {
	// Find point within box nearest to center of sphere
	Vector<T,dim> p;
	for (unsigned int i = 0; i < dim; i++) {
		if (o[i] < m_Minimum[i]) {
			p[i] = m_Minimum[i];
		} else if (o[i] > m_Maximum[i]) {
			p[i] = m_Maximum[i];
		} else {
			p[i] = o[i];
		} 
	}
	
	return (o - p).length();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool AlignedBox<T,dim>::intersect(const Sphere<T,dim>& sphere) const {
	// Find point within box nearest to center of sphere
	Vector<T,dim> p;
	Vector<T,dim> o = sphere.getOrigin();
	for (unsigned int i = 0; i < dim; i++) {
		if (o[i] < m_Minimum[i]) {
			p[i] = m_Minimum[i];
		} else if (o[i] > m_Maximum[i]) {
			p[i] = m_Maximum[i];
		} else {
			p[i] = o[i];
		} 
	}
	
	return ((o - p).length2() <= sphere.getRadius() * sphere.getRadius());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool AlignedBox<T,dim>::intersect(const Cube<T,dim>& cube) const {
	T r = cube.getRadius();
	for(unsigned int i = 0; i < dim; i++) {
		if ((m_Maximum[i] < cube.getOrigin()[i] - r) || (m_Minimum[i] > cube.getOrigin()[i] + r))
			return false;
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool AlignedBox<T,dim>::intersect(const AlignedBox<T,dim>& box) const {
	for(unsigned int i = 0; i < dim; i++) {
		if ((m_Maximum[i] < box.m_Minimum[i]) || (m_Minimum[i] > box.m_Maximum[i]))
			return false;
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool AlignedBox<T,dim>::intersect(const Box<T,dim>& box) const {
	return box.intersect(*this);
}


/*--------------------------------------------------------------------------*/
/** 
 * Returns true if the line intersects the aligned box, else this method
 * will return false. If the is argument is not NULL, an intersection
 * point will be returned.
 * @param line The line used for intersection
 * @param is The intersection info that will hold the point of
 *        intersection. Can be NULL.
 * @return true for intersection, otherwise false
 */
template<class T, int dim> 
template<class L>
bool AlignedBox<T,dim>::intersectLinear(const L& line,Intersection<T,dim>* is) const {
	Vector<T,dim> dir = line.getDirection();
	Vector<T,dim> org = line.getOrigin();
	
	T tmin = L::MinParam;
	T tmax = L::MaxParam;
	for (unsigned int d = 0; d < dim; ++d) {
		if (dir[d] == 0.0f) {
			if (org[d] < m_Minimum[d] || org[d] > m_Maximum[d])
				return false;
			continue;
		}

		// calculate point of intersection
		T t1 = (m_Minimum[d] - org[d]) / dir[d];
		T t2 = (m_Maximum[d] - org[d]) / dir[d];
		if (t1 > t2)
			swap(t1,t2);
		
		if (t2 < tmin)
			return false;
		else if (t1 > tmin)
			tmin = t1;
		if (t1 > tmax)
			return false;
		else if (t2 < tmax)
			tmax = t2;
	}
		
	// Check if we should record the information
	if (is != NULL) {
		is->setPosition(org + tmin*dir);
		is->setNormal(is->getPosition() - m_Origin);
		is->setParameter(tmin);
	}
	
	return true;
}


/*--------------------------------------------------------------------------*/
/** 
 * Returns true if the line intersects the cube, else this method
 * will return false. If the is argument is not NULL, an intersection
 * point will be returned.
 * @param line The line used for intersection
 * @param is The intersection info that will hold the point of
 *        intersection. Can be NULL.
 * @return true for intersection, otherwise false
 */
template<class T, int dim> 
inline bool AlignedBox<T,dim>::intersect(const Line<T,dim>& line,Intersection<T,dim>* is) const {
	return intersectLinear(line, is);
}


/*--------------------------------------------------------------------------*/
/**
 * Returns true if the line segment intersects the cube, else this method
 * will return false. If the is argument is not NULL, an intersection
 * point will be returned.
 * @param seg The line segment used for intersection
 * @param is The intersection info that will hold the point of
 *        intersection. Can be NULL.
 * @return true for intersection, otherwise false
 */
template<class T, int dim> 
inline bool AlignedBox<T,dim>::intersect(const Segment<T,dim>& seg,Intersection<T,dim>* is) const {
	return intersectLinear(seg, is);
}


/*--------------------------------------------------------------------------*/
/**
 * Returns true if the ray intersects the cube, else this method
 * will return false. If the is argument is not NULL, the first intersection
 * point will be returned.
 * @param ray The ray used for intersection
 * @param is The intersection info that will hold the point of
 *        intersection. Can be NULL.
 * @return true for intersection, otherwise false
*/
template<class T, int dim> 
inline bool AlignedBox<T,dim>::intersect(const Ray<T,dim>& ray,Intersection<T,dim>* is) const {
	return intersectLinear(ray, is);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Box<T,dim> AlignedBox<T,dim>::toBox() const {
	Vector<T,dim> a[dim];
	for (unsigned int i = 0; i < dim; i++) {
		a[i].null();
		a[i][i] = (m_Maximum[i] - m_Minimum[i])/T(2);
	}
	return Box<T,dim>((m_Maximum + m_Minimum)/T(2), a);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Sphere<T,dim> AlignedBox<T,dim>::toSphere() const {
	Vector<T,dim> o = (m_Maximum[i] + m_Minimum[i])/T(2);
	T r = (m_Maximum[i] - m_Minimum[i]).length() / T(2);
	return Sphere<T,dim>(o,r);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Cube<T,dim> AlignedBox<T,dim>::toCube() const {
	Vector<T,dim> o = (m_Maximum[i] + m_Minimum[i])/T(2);
	T r = (m_Maximum[i] - m_Minimum[i]).maxabs() / T(2);
	return Cube<T,dim>(o, r);
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void AlignedBox<T,dim>::merge(const AlignedBox<T,dim>& box) {
	for (unsigned int i = 0; i < dim; i++) {
		m_Minimum[i] = min(m_Minimum[i], box.m_Minimum[i]);
		m_Maximum[i] = max(m_Maximum[i], box.m_Maximum[i]);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void AlignedBox<T,dim>::bound(const Vector<T,dim>* v,unsigned int n) {
	bound(v, sizeof(Vector<T,dim>), n);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void AlignedBox<T,dim>::bound(const Vector<T,dim>* v,unsigned int stride,unsigned int n) {
	if (n == 0) {
		m_Maximum.null();
		m_Minimum.null();
		return;	
	}
	m_Minimum = v[0];
	m_Maximum = v[0];
	
	for(unsigned int i = 0; i < n; i++) {
		for(unsigned int j = 0; j < dim; j++) {
			if (m_Minimum[j] > (*v)[j])
				m_Minimum[j] = (*v)[j];
			else if (m_Maximum[j] < (*v)[j])
				m_Maximum[j] = (*v)[j];
		}
		v = (const Vector<T,dim>*)(const void*)(((const byte*)v) + stride);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void AlignedBox<T,dim>::getVertices(Vector<T,dim>* cv) const {
	if (dim == 2) {
		cv[0][0] = m_Minimum[0];  cv[0][1] = m_Minimum[1];
		cv[1][0] = m_Maximum[0];  cv[1][1] = m_Minimum[1];
		cv[2][0] = m_Minimum[0];  cv[2][1] = m_Maximum[1];
		cv[3][0] = m_Maximum[0];  cv[3][1] = m_Maximum[1];
	} else if (dim == 3) {
		cv[0][0] = m_Minimum[0];  cv[0][1] = m_Minimum[1]; cv[0][2] = m_Minimum[2];
		cv[1][0] = m_Maximum[0];  cv[1][1] = m_Minimum[1]; cv[1][2] = m_Minimum[2];
		cv[2][0] = m_Minimum[0];  cv[2][1] = m_Maximum[1]; cv[2][2] = m_Minimum[2];
		cv[3][0] = m_Maximum[0];  cv[3][1] = m_Maximum[1]; cv[3][2] = m_Minimum[2];
		cv[4][0] = m_Minimum[0];  cv[4][1] = m_Minimum[1]; cv[4][2] = m_Maximum[2];
		cv[5][0] = m_Maximum[0];  cv[5][1] = m_Minimum[1]; cv[5][2] = m_Maximum[2];
		cv[6][0] = m_Minimum[0];  cv[6][1] = m_Maximum[1]; cv[6][2] = m_Maximum[2];
		cv[7][0] = m_Maximum[0];  cv[7][1] = m_Maximum[1]; cv[7][2] = m_Maximum[2];
	} else {
		for (i = 0; i < (1 << dim); i++) {
			for (j = 0; j < dim; j++) {
				if (i & (1 << j))
					cv[j] = m_Minimum[j];
				else
					cv[j] = m_Maximum[j];
			}
			cv++;
		}
	}	
}

}; };

#endif
