/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GEOMETRY_BINORMALGENERATOR_INL
#define __GEOMETRY_BINORMALGENERATOR_INL

namespace magnum { namespace geometry {
using magnum::math::max;
using magnum::math::min;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
BinormalGenerator<T,I>::BinormalGenerator() {
	m_Normals = NULL;
	m_NormalStride = 0;
	m_Binormals = NULL;
	m_BinormalStride = 0;
	m_Tangents = NULL;
	m_TangentStride = 0;
	m_Textures = NULL;
	m_TextureStride = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline unsigned int BinormalGenerator<T,I>::getNormalStride() const {
	return m_NormalStride;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline const Vector3<T>* BinormalGenerator<T,I>::getNormals() const {
	return m_Normals;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline void BinormalGenerator<T,I>::setNormals(const Vector3<T>* n,unsigned int s) {
	m_Normals = n;
	m_NormalStride = s;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline unsigned int BinormalGenerator<T,I>::getTextureStride() const {
	return m_TextureStride;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline const Vector2<T>* BinormalGenerator<T,I>::getTextures() const {
	return m_Textures;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline void BinormalGenerator<T,I>::setTextures(const Vector2<T>* v,unsigned int s) {
	m_Textures = v;
	m_TextureStride = s;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline unsigned int BinormalGenerator<T,I>::getBinormalStride() const {
	return m_BinormalStride;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline Vector3<T>* BinormalGenerator<T,I>::getBinormals() const {
	return m_Binormals;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline void BinormalGenerator<T,I>::setBinormals(Vector3<T>* n,unsigned int s) {
	m_Binormals = n;
	m_BinormalStride = s;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline unsigned int BinormalGenerator<T,I>::getTangentStride() const {
	return m_TangentStride;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline Vector3<T>* BinormalGenerator<T,I>::getTangents() const {
	return m_Tangents;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline void BinormalGenerator<T,I>::setTangents(Vector3<T>* n,unsigned int s) {
	m_Tangents = n;
	m_TangentStride = s;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline const Vector3<T>& BinormalGenerator<T,I>::getNormal(unsigned int i) {
	return *reinterpret_cast<const Vector3<T>*>(reinterpret_cast<const byte*>(m_Normals)+i*m_NormalStride);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline const Vector2<T>& BinormalGenerator<T,I>::getTexture(unsigned int i) {
	return *reinterpret_cast<const Vector2<T>*>(reinterpret_cast<const byte*>(m_Textures)+i*m_TextureStride);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline Vector3<T>& BinormalGenerator<T,I>::getBinormal(unsigned int i) {
	return *reinterpret_cast<Vector3<T>*>(reinterpret_cast<byte*>(m_Binormals)+i*m_BinormalStride);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
inline Vector3<T>& BinormalGenerator<T,I>::getTangent(unsigned int i) {
	return *reinterpret_cast<Vector3<T>*>(reinterpret_cast<byte*>(m_Tangents)+i*m_TangentStride);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class I>
void BinormalGenerator<T,I>::generate() {
	if (getIndexSize() == 0)
		return;
	if (getIndices() == NULL)
		return;
	if (getVertices() == NULL)
		return;
	if (m_Normals == NULL)
		return;
	if (m_Binormals == NULL)
		return;
	if (m_Tangents == NULL)
		return;
		
	unsigned int minIndex = ~0;
	unsigned int maxIndex = 0;
	
	unsigned int count = getPrimitiveCount();
	for (unsigned int i = 0; i < count; i++) {
		unsigned int vi1,vi2,vi3;
		getTriangle(i, &vi1, &vi2, &vi3);
		
		minIndex = min(vi1, min(vi2, min(vi3, minIndex)));
		maxIndex = max(vi1, max(vi2, max(vi3, maxIndex)));
	}	
	
    // first step, create 2 vectors to store the tangents for s and t direction
    Array<Vector3<T> > s_tangent;
    Array<Vector3<T> > t_tangent;
    s_tangent.setSize(maxIndex - minIndex + 1);
    t_tangent.setSize(maxIndex - minIndex + 1);
    for(unsigned i = 0; i < s_tangent.getSize(); ++i) {
        s_tangent[i].null();
        t_tangent[i].null();
    }

    // second step, iterate over all triangles and calculate the additions tangents
    for(unsigned int i = 0; i < count; ++i) {
		unsigned int vi1,vi2,vi3;
		getTriangle(i, &vi1, &vi2, &vi3);

        // fetch vertices
        Vector3<T> v1 = getVertex(vi1);
		Vector3<T> v2 = getVertex(vi2);
        Vector3<T> v3 = getVertex(vi3);

        // fetch texcoords
        Vector2<T> t1 = getTexture(vi1);
		Vector2<T> t2 = getTexture(vi2);
        Vector2<T> t3 = getTexture(vi3);
        
        Vector3<T> vedge1 = v2 - v1;
        Vector3<T> vedge2 = v3 - v1;

        Vector2<T> tedge1 = t2 - t1;
        Vector2<T> tedge2 = t3 - t1;

        T r = T(1) / (tedge1.x * tedge2.y - tedge2.x * tedge1.y);
        Vector3<T> sdir = (vedge1 * tedge2.y) - (vedge2 * tedge1.y) *r;
        Vector3<T> tdir = (vedge2 * tedge1.x) - (vedge1 * tedge2.x) *r;

        s_tangent[vi1 - minIndex] += sdir;
        s_tangent[vi2 - minIndex] += sdir;
        s_tangent[vi3 - minIndex] += sdir;

        t_tangent[vi1 - minIndex] += tdir;
        t_tangent[vi2 - minIndex] += tdir;
        t_tangent[vi3 - minIndex] += tdir;
    }
   
    // now calculate the real tangent and binormal and set them into the arrays
    for (unsigned int i = minIndex; i <= maxIndex; i++)  {
        Vector3<T> normal = getNormal(i);
        Vector3<T> s = s_tangent[i - minIndex];

        // Gram-Schmidt orthogonalize
        Vector3<T> tangent = s - normal * (normal * s);
        tangent.normalise();
        getTangent(i) = tangent;

        // Calculate handedness
        T sign = (normal % s * t_tangent[i - minIndex] < T(0)) ? T(-1) : T(1);

        // calculate binormal (ok, actually really bitangent, but who knows its correct name)
        Vector3<T> binormal = (normal % s) * sign;
        getBinormal(i) = binormal;
    }
}

}; };

#endif
