/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GEOMETRY_BOX_INL
#define __GEOMETRY_BOX_INL

namespace magnum { namespace geometry {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline Box<T,dim>::Box() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline Box<T,dim>::Box(const Vector<T,dim>& o,const Vector<T,dim>* a) {
	m_Origin = o;
	for (unsigned int i = 0; i < dim; i++) {
		m_Axes[i] = a[i];
		m_Axes2[i] = a[i]*a[i];
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void Box<T,dim>::setOrigin(const Vector<T,dim>& o) {
	m_Origin = o;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const Vector<T,dim>& Box<T,dim>::getOrigin() const {
	return m_Origin;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void	Box<T,dim>::setAxes(const Vector<T,dim>* e) {
	for (unsigned int i = 0; i < dim; i++) {
		m_Axes[i] = e[i];
		m_Axes2[i] = e[i] * e[i];
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const Vector<T,dim>* Box<T,dim>::getAxes() const {
	return m_Axes;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void	Box<T,dim>::setAxis(unsigned int i, const Vector<T,dim>& a) {
	ASSERT(i < dim);
	m_Axes[i] = a;
	m_Axes2[i] = a*a;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const Vector<T,dim>& Box<T,dim>::getAxis(unsigned int i) const {
	ASSERT(i < dim);
	return m_Axes[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const T& Box<T,dim>::getAxis2(unsigned int i) const {
	ASSERT(i < dim);
	return m_Axes2[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline bool Box<T,dim>::isEmpty() const {
	for(unsigned int i = 0; i < dim; i++) {
		if (m_Axes2[i] == T(0))
			return true;
	}
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void Box<T,dim>::empty() {
	m_Origin.null();
	for(unsigned int i = 0; i < dim; i++) {
		m_Axes[i] = Vector<T,dim>::Null;
		m_Axes2[i] = T(0);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline bool Box<T,dim>::isInside(const Vector<T,dim>& v) const {
	Vector<T,dim> p = v - m_Origin;
	for (unsigned int i = 0; i < dim; i++) {
		T val = p*m_Axes[i];
		if ((val < -m_Axes2[i]) || (val > m_Axes2[i]))
			return false;
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
T Box<T,dim>::getDistance(const Vector<T,dim>& o) const {
	// Find point within box nearest to center of sphere
	Vector<T,dim> p;
	p.null();
	for (unsigned int i = 0; i < dim; i++) {
		T t = o*m_Axes[i];
		if (t >= m_Axes2[i]) {
			p += m_Axes[i];
		} else if (t <= -m_Axes2[i]) {
			p -= m_Axes[i];
		} else {
			t /= m_Axes2[i];
			p += m_Axes[i]*t;
		}
	}

	return (o - p).length();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool Box<T,dim>::intersect(const Sphere<T,dim>& sphere) const {
	// Find point within box nearest to center of sphere
	Vector<T,dim> p;
	Vector<T,dim> o = sphere.getOrigin() - m_Origin;
	p.null();
	for (unsigned int i = 0; i < dim; i++) {
		T t = o*m_Axes[i];
		if (t >= m_Axes2[i]) {
			p += m_Axes[i];
		} else if (t <= -m_Axes2[i]) {
			p -= m_Axes[i];
		} else {
			t /= m_Axes2[i];
			p += m_Axes[i]*t;
		}
	}

	return ((o - p).length2() <= sphere.getRadius() * sphere.getRadius());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool Box<T,dim>::intersect(const Cube<T,dim>& cube) const {
	unsigned int i,j;
	T radius = cube.getRadius();
	Vector<T,dim> origin = cube.getOrigin();
	// Case 1: Project all axes onto cube axes
	for (i = 0; i < dim; i++) {
		T l = abs(m_Axes[0][i]);
		for (j = 1; j < dim; j++)
			l += abs(m_Axes[j][i]);
		if ((m_Origin[i] - l >= origin[i] + radius) 
		 || (m_Origin[i] + l <= origin[i] - radius))
			return false;
	}
	
	// Case 2: Project all cube axes onto box
	Vector<T,dim> diff = origin - m_Origin;
	for (i = 0; i < dim; i++) {
		T l = m_Axes[i].sumabs() * radius;
		T p = abs(diff * m_Axes[i]);
		if (p >= l + m_Axes2[i])
			return false;
	}
	
	if (dim == 3) {
		// Finally test mixtures of box axes and cube axes
		for (i = 0; i < 3; i++) {
			Vector<T,dim> axis;
			T l;
			axis[0] = T(0);
			axis[1] = m_Axes[i][2];
			axis[2] = -m_Axes[i][1];
			l = radius * (abs(axis[1]) + abs(axis[2]));
			l += abs(axis[1]*m_Axes[0][1] + axis[2]*m_Axes[0][2]);
			l += abs(axis[1]*m_Axes[1][1] + axis[2]*m_Axes[1][2]);
			l += abs(axis[1]*m_Axes[2][1] + axis[2]*m_Axes[2][2]);
			if (abs(axis[1]*diff[1] + axis[2]*diff[2]) >= l)
				return false;

			axis[0] = -m_Axes[i][2];
			axis[1] = T(0);
			axis[2] = m_Axes[i][0];
			l = radius * (abs(axis[0]) + abs(axis[2]));
			l += abs(axis[0]*m_Axes[0][0] + axis[2]*m_Axes[0][2]);
			l += abs(axis[0]*m_Axes[1][0] + axis[2]*m_Axes[1][2]);
			l += abs(axis[0]*m_Axes[2][0] + axis[2]*m_Axes[2][2]);
			if (abs(axis[0]*diff[0] + axis[2]*diff[2]) >= l)
				return false;

			axis[0] = m_Axes[i][1];
			axis[1] = -m_Axes[i][0];
			axis[2] = T(0);
			l = radius * (abs(axis[0]) + abs(axis[1]));
			l += abs(axis[0]*m_Axes[0][0] + axis[1]*m_Axes[0][1]);
			l += abs(axis[0]*m_Axes[1][0] + axis[1]*m_Axes[1][1]);
			l += abs(axis[0]*m_Axes[2][0] + axis[1]*m_Axes[2][1]);
			if (abs(axis[0]*diff[0] + axis[1]*diff[1]) >= l)
				return false;
		}
	} else if (dim >= 4) {
		ASSERT(false);
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool Box<T,dim>::intersect(const AlignedBox<T,dim>& box) const {
	unsigned int i,j;
	Vector<T,dim> mi = box.getMinimum();
	Vector<T,dim> ma = box.getMaximum();
	// Case 1: Project all axes onto aligned box axes
	for (i = 0; i < dim; i++) {
		T l = abs(m_Axes[0][i]);
		for (j = 1; j < dim; j++)
			l += abs(m_Axes[j][i]);
		if ((m_Origin[i] - l >= ma[i]) 
		 || (m_Origin[i] + l <= mi[i]))
			return false;
	}
	
	// Case 2: Project all aligned box axes onto box
	Vector<T,dim> diff = (mi + ma) - T(2)*m_Origin;
	Vector<T,dim> size = ma - mi;
	for (i = 0; i < dim; i++) {
		T l = abs(m_Axes[i][0] * size[0]);
		for (j = 1; j < dim; j++)	
			l += abs(m_Axes[i][j] * size[j]);
		T p = abs(diff * m_Axes[i]);
		if (p >= l + T(2)*m_Axes2[i])
			return false;
	}
	
	if (dim == 3) {
		// Finally test mixtures of box axes and cube axes
		for (i = 0; i < 3; i++) {
			Vector<T,dim> axis;
			T l,k;
			axis[0] = T(0);
			axis[1] = m_Axes[i][2];
			axis[2] = -m_Axes[i][1];
			k = abs(size[1]*abs(axis[1]) + abs(size[2]*axis[2]));
			l = abs(axis[1]*m_Axes[0][1] + axis[2]*m_Axes[0][2]);
			l += abs(axis[1]*m_Axes[1][1] + axis[2]*m_Axes[1][2]);
			l += abs(axis[1]*m_Axes[2][1] + axis[2]*m_Axes[2][2]);
			if (abs(axis[1]*diff[1] + axis[2]*diff[2]) >= k + T(2)*l)
				return false;

			axis[0] = -m_Axes[i][2];
			axis[1] = T(0);
			axis[2] = m_Axes[i][0];
			k = abs(size[0]*abs(axis[0]) + abs(size[2]*axis[2]));
			l = abs(axis[0]*m_Axes[0][0] + axis[2]*m_Axes[0][2]);
			l += abs(axis[0]*m_Axes[1][0] + axis[2]*m_Axes[1][2]);
			l += abs(axis[0]*m_Axes[2][0] + axis[2]*m_Axes[2][2]);
			if (abs(axis[0]*diff[0] + axis[2]*diff[2]) >= k + T(2)*l)
				return false;

			axis[0] = m_Axes[i][1];
			axis[1] = -m_Axes[i][0];
			axis[2] = T(0);
			k = abs(size[0]*abs(axis[0]) + abs(size[1]*axis[1]));
			l = abs(axis[0]*m_Axes[0][0] + axis[1]*m_Axes[0][1]);
			l += abs(axis[0]*m_Axes[1][0] + axis[1]*m_Axes[1][1]);
			l += abs(axis[0]*m_Axes[2][0] + axis[1]*m_Axes[2][1]);
			if (abs(axis[0]*diff[0] + axis[1]*diff[1]) >= k + T(2)*l)
				return false;
		}
	} else if (dim >= 4) {
		ASSERT(false);
	}
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool Box<T,dim>::intersect(const Box<T,dim>& box) const {
	Vector<T,dim> diff = box.m_Origin - m_Origin;
	unsigned int i,j;
	// Case 1: Project all axes onto aligned box axes
	for (i = 0; i < dim; i++) {
		T l = abs(m_Axes[i] * box.m_Axes[0]);
		for (j = 1; j < dim; j++)
			l += abs(m_Axes[i] * box.m_Axes[j]);
		T p = abs(diff * m_Axes[i]);
		if (p >= l + m_Axes2[i])
			return false;
	}
	
	// Case 2: Project all aligned box axes onto box
	for (i = 0; i < dim; i++) {
		T l = abs(m_Axes[0] * box.m_Axes[i]);
		for (j = 1; j < dim; j++)
			l += abs(m_Axes[j] * box.m_Axes[i]);
		T p = abs(diff * box.m_Axes[i]);
		if (p >= l + box.m_Axes2[i])
			return false;
	}

	if (dim == 3) {
		// Finally test mixtures of box axes and cube axes
		for (i = 0; i < 3; i++) {
			Vector<T,dim> axis;
			T l,k;
			axis = (Vector3<T>)m_Axes[i] % (Vector3<T>)box.m_Axes[0];
			k = abs(axis * box.m_Axes[1]) + abs(axis * box.m_Axes[2]);
			l = abs(axis * m_Axes[0]) + abs(axis * m_Axes[1]) + abs(axis * m_Axes[2]);
			if (abs(axis * diff) >= k + l)
				return false;

			axis = (Vector3<T>)m_Axes[i] % (Vector3<T>)box.m_Axes[1];
			k = abs(axis * box.m_Axes[0]) + abs(axis * box.m_Axes[2]);
			l = abs(axis * m_Axes[0]) + abs(axis * m_Axes[1]) + abs(axis * m_Axes[2]);
			if (abs(axis * diff) >= k + l)
				return false;

			axis = (Vector3<T>)m_Axes[i] % (Vector3<T>)box.m_Axes[2];
			k = abs(axis * box.m_Axes[0]) + abs(axis * box.m_Axes[1]);
			l = abs(axis * m_Axes[0]) + abs(axis * m_Axes[1]) + abs(axis * m_Axes[2]);
			if (abs(axis * diff) >= k + l)
				return false;
		}
	} else if (dim >= 4) {
		ASSERT(false);
	}

	return true;
}


/*--------------------------------------------------------------------------*/
/** 
 * Returns true if the line intersects the cube, else this method
 * will return false. If the is argument is not NULL, an intersection
 * point will be returned.
 * @param line The line used for intersection
 * @param is The intersection info that will hold the point of
 *        intersection. Can be NULL.
 * @return true for intersection, otherwise false
 */
template<class T, int dim> 
template<class L>
bool Box<T,dim>::intersectLinear(const L& line,Intersection<T,dim>* is) const {
	Vector<T,dim> dir = line.getDirection();
	Vector<T,dim> off = m_Origin - line.getOrigin();
	
	T tmin = L::MinParam;
	T tmax = L::MaxParam;
	for (unsigned int d = 0; d < dim; ++d) {
		T p = off * m_Axes[d];
		T l = dir * m_Axes[d];
		if (abs(l) < Limits<T>::epsilon()) {
			if (abs(p) > m_Axes2[d])
				return false;
			continue;
		}
		
		// calculate point of intersection
		T t1 = (p - m_Axes2[d]) / l;
		T t2 = (p + m_Axes2[d]) / l;
		if (t1 > t2)
			swap(t1,t2);
		
		if (t2 < tmin)
			return false;
		else if (t1 > tmin)
			tmin = t1;
		if (t1 > tmax)
			return false;
		else if (t2 < tmax)
			tmax = t2;
	}
		
	// Check if we should record the information
	if (is != NULL) {
		is->setPosition(line.getOrigin() + tmin*dir);
		is->setNormal(is->getPosition() - m_Origin);
		is->setParameter(tmin);
	}
	
	return true;
}


/*--------------------------------------------------------------------------*/
/** 
 * Returns true if the line intersects the cube, else this method
 * will return false. If the is argument is not NULL, an intersection
 * point will be returned.
 * @param line The line used for intersection
 * @param is The intersection info that will hold the point of
 *        intersection. Can be NULL.
 * @return true for intersection, otherwise false
 */
template<class T, int dim> 
inline bool Box<T,dim>::intersect(const Line<T,dim>& line,Intersection<T,dim>* is) const {
	return intersectLinear(line, is);
}


/*--------------------------------------------------------------------------*/
/**
 * Returns true if the line segment intersects the cube, else this method
 * will return false. If the is argument is not NULL, an intersection
 * point will be returned.
 * @param seg The line segment used for intersection
 * @param is The intersection info that will hold the point of
 *        intersection. Can be NULL.
 * @return true for intersection, otherwise false
 */
template<class T, int dim> 
inline bool Box<T,dim>::intersect(const Segment<T,dim>& seg,Intersection<T,dim>* is) const {
	return intersectLinear(seg, is);
}


/*--------------------------------------------------------------------------*/
/**
 * Returns true if the ray intersects the cube, else this method
 * will return false. If the is argument is not NULL, the first intersection
 * point will be returned.
 * @param ray The ray used for intersection
 * @param is The intersection info that will hold the point of
 *        intersection. Can be NULL.
 * @return true for intersection, otherwise false
*/
template<class T, int dim> 
inline bool Box<T,dim>::intersect(const Ray<T,dim>& ray,Intersection<T,dim>* is) const {
	return intersectLinear(ray, is);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
void Box<T,dim>::merge(const Cube<T,dim>& ) {
	ASSERT(false);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
void Box<T,dim>::merge(const AlignedBox<T,dim>& ) {
	ASSERT(false);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
void Box<T,dim>::merge(const Box<T,dim>& box) {
	unsigned int i,j;
	T vol1 = m_Axes2[0];
	for (i = 1; i < dim; i++)
		vol1 *= m_Axes2[i];

	T vol2 = box.m_Axes2[0];
	for (i = 1; i < dim; i++)
		vol2 *= box.m_Axes2[i];
		
	if (vol1 > vol2) {
		Vector<T,dim> mi,ma;
		for (i = 0; i < dim; i++) {
			mi[i] = (m_Origin - m_Axes[i])*m_Axes[i];
			ma[i] = (m_Origin + m_Axes[i])*m_Axes[i];
		}
		
		for (i = 0; i < dim; i++) {
			T s = m_Axes[i] * box.m_Origin;
			T r = abs(m_Axes[i] * box.m_Axes[0]);
			for (j = 0; j < dim; j++)
				r += abs(m_Axes[i] * box.m_Axes[j]);
			if (s - r < mi[i])
				mi[i] = s - r; 
			else if (s + r > ma[i])
				ma[i] = s + r; 
		}

		Vector<T,dim> center = T(0.5) * (mi + ma);
		Vector<T,dim> diff = T(0.5) * (ma - mi);
		m_Origin.null();
		for (i = 0; i < dim; i++) {
			m_Origin += center[i]/m_Axes2[i] * m_Axes[i];
			m_Axes[i] *= diff[i]/m_Axes2[i];
			m_Axes2[i] = m_Axes[i] * m_Axes[i];
		}
	} else {
		Vector<T,dim> mi,ma;
		for (i = 0; i < dim; i++) {
			mi[i] = (box.m_Origin - box.m_Axes[i])*box.m_Axes[i];
			ma[i] = (box.m_Origin + box.m_Axes[i])*box.m_Axes[i];
		}
		
		for (i = 0; i < dim; i++) {
			T s = box.m_Axes[i] * m_Origin;
			T r = abs(box.m_Axes[i] * m_Axes[0]);
			for (j = 0; j < dim; j++)
				r += abs(box.m_Axes[i] * m_Axes[j]);
			if (s - r < mi[i])
				mi[i] = s - r; 
			else if (s + r > ma[i])
				ma[i] = s + r; 
		}

		Vector<T,dim> center = T(0.5) * (mi + ma);
		Vector<T,dim> diff = T(0.5) * (ma - mi);
		m_Origin.null();
		for (i = 0; i < dim; i++) {
			m_Origin += center[i]/box.m_Axes2[i] * box.m_Axes[i];
			m_Axes[i] = diff[i]/box.m_Axes2[i] * box.m_Axes[i];
			m_Axes2[i] = m_Axes[i] * m_Axes[i];
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
Cube<T,dim> Box<T,dim>::toCube() const {
	T m, ma = T(0);
	for (unsigned int i = 0; i < dim; i++) {
		m = abs(m_Axes[0][i]);
		for (unsigned int j = 1; j < dim; j++)
			m += abs(m_Axes[j][i]);
		if (m > ma)
			ma = m;
	}
	return Cube<T,dim>(m_Origin, ma);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
void Box<T,dim>::toCube(Cube<T,dim>* cube) const {
	T m, ma = T(0);
	for (unsigned int i = 0; i < dim; i++) {
		m = abs(m_Axes[0][i]);
		for (unsigned int j = 1; j < dim; j++)
			m += abs(m_Axes[j][i]);
		if (m > ma)
			ma = m;
	}
	cube->setOrigin(m_Origin);
	cube->setRadius(ma);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
Sphere<T,dim> Box<T,dim>::toSphere() const {
	Vector<T,dim> v = m_Axes[0];
	for (unsigned int i = 1; i < dim; i++)
		v += m_Axes[i];
	return Sphere<T,dim>(m_Origin, v.length());	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
void Box<T,dim>::toSphere(Sphere<T,dim>* sphere) const {
	Vector<T,dim> v = m_Axes[0];
	for (unsigned int i = 1; i < dim; i++)
		v += m_Axes[i];
	sphere->setOrigin(m_Origin);
	sphere->setLength(v.length());	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
AlignedBox<T,dim> Box<T,dim>::toAlignedBox() const {
	Vector<T,dim> ma;
	for (unsigned int i = 0; i < dim; i++) {
		ma[i] = abs(m_Axes[0][i]);
		for (unsigned int j = 1; j < dim; j++)
			ma[i] += abs(m_Axes[j][i]);
	}
	return AlignedBox<T,dim>(m_Origin - ma, m_Origin + ma);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
void Box<T,dim>::toAlignedBox(AlignedBox<T,dim>* box) const {
	Vector<T,dim> ma;
	for (unsigned int i = 0; i < dim; i++) {
		ma[i] = abs(m_Axes[0][i]);
		for (unsigned int j = 1; j < dim; j++)
			ma[i] += abs(m_Axes[j][i]);
	}
	box->setMinimum(m_Origin - ma);
	box->setMaximum(m_Origin + ma);
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
void Box<T,dim>::getVertices(Vector<T,dim>* cv) const {
	if (dim == 2) {
		cv[0] = m_Origin - m_Axes[0] - m_Axes[1];
		cv[1] = m_Origin + m_Axes[0] - m_Axes[1];
		cv[2] = m_Origin - m_Axes[0] + m_Axes[1];
		cv[3] = m_Origin + m_Axes[0] + m_Axes[1];
	} else if (dim == 3) {
		cv[0] = m_Origin - m_Axes[0] - m_Axes[1] - m_Axes[2];
		cv[1] = m_Origin + m_Axes[0] - m_Axes[1] - m_Axes[2];
		cv[2] = m_Origin - m_Axes[0] + m_Axes[1] - m_Axes[2];
		cv[3] = m_Origin + m_Axes[0] + m_Axes[1] - m_Axes[2];
		cv[4] = m_Origin - m_Axes[0] - m_Axes[1] + m_Axes[2];
		cv[5] = m_Origin + m_Axes[0] - m_Axes[1] + m_Axes[2];
		cv[6] = m_Origin - m_Axes[0] + m_Axes[1] + m_Axes[2];
		cv[7] = m_Origin + m_Axes[0] + m_Axes[1] + m_Axes[2];
	} else {
		unsigned int i,j;
		for (i = 0; i < (1 << dim); i++) {
			*cv = m_Origin;
			for (j = 0; j < dim; j++) {
				if (i & (1 << j))
					*cv += m_Axes[j];
				else
					*cv -= m_Axes[j];
			}
			cv++;
		}
	}	
}

}; };

#endif
