/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GEOMETRY_CUBE_INL
#define __GEOMETRY_CUBE_INL

namespace magnum { namespace geometry {
using magnum::algorithm::swap;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline Cube<T,dim>::Cube() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline Cube<T,dim>::Cube(const Vector<T,dim>& o, const T& r) {
	m_Origin = o;
	m_Radius = r;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void Cube<T,dim>::setOrigin(const Vector<T,dim>& o) {
	m_Origin = o;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const Vector<T,dim>& Cube<T,dim>::getOrigin() const {
	return m_Origin;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void Cube<T,dim>::setRadius(const T& r) {
	m_Radius = r;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const T& Cube<T,dim>::getRadius() const {
	return m_Radius;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline bool Cube<T,dim>::isEmpty() const {
	return (m_Radius == T(0));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void Cube<T,dim>::empty() {
	m_Radius = T(0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline bool Cube<T,dim>::isInside(const Vector<T,dim>& v) const {
	for (unsigned int i = 0; i < dim; i++) {
		if ((v[i] < m_Origin[i] - m_Radius) || (v[i] > m_Origin[i] + m_Radius))
			return false;
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
T Cube<T,dim>::getDistance(const Vector<T,dim>& o) const {
	// Find point within cube nearest to center of sphere
	Vector<T,dim> p;
	for (unsigned int i = 0; i < dim; i++) {
		if (o[i] < m_Center[i] - m_Radius) {
			p[i] = m_Center[i] - m_Radius;
		} else if (o[i] > m_Center[i] + m_Radius) {
			p[i] = m_Center[i] + m_Radius;
		} else {
			p[i] = o[i];
		} 
	}
	
	return (o - p).getLength();	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool Cube<T,dim>::intersect(const Cube<T,dim>& cube) const {
	Vector<T,dim> diff = m_Origin - cube.m_Origin;
	T radius = m_Radius + cube.m_Radius;
	for (unsigned int i = 0; i < dim; i++) {
		if (abs(diff[i]) > radius)
			return false;
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Cube<T,dim>::intersect(const Sphere<T,dim>& sphere) const {
	// Find point within cube nearest to center of sphere
	Vector<T,dim> p;
	Vector<T,dim> o = sphere.getOrigin();
	for (unsigned int i = 0; i < dim; i++) {
		if (o[i] < m_Origin[i] - m_Radius) {
			p[i] = m_Origin[i] - m_Radius;
		} else if (o[i] > m_Origin[i] + m_Radius) {
			p[i] = m_Origin[i] + m_Radius;
		} else {
			p[i] = o[i];
		} 
	}
	
	return ((o - p).length2() <= sphere.getRadius() * sphere.getRadius());	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Cube<T,dim>::intersect(const AlignedBox<T,dim>& box) const {
	return box.intersect(*this);
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Cube<T,dim>::intersect(const Box<T,dim>& box) const {
	return box.intersect(*this);
}


/*--------------------------------------------------------------------------*/
/** 
 * Returns true if the line intersects the cube, else this method
 * will return false. If the is argument is not NULL, an intersection
 * point will be returned.
 * @param line The line used for intersection
 * @param is The intersection info that will hold the point of
 *        intersection. Can be NULL.
 * @return true for intersection, otherwise false
 */
template<class T, int dim> 
template<class L>
bool Cube<T,dim>::intersectLinear(const L& line,Intersection<T,dim>* is) const {
	Vector<T,dim> dir = line.getDirection();
	Vector<T,dim> off = m_Origin - line.getOrigin();
	
	T tmin = L::MinParam;
	T tmax = L::MaxParam;
	for (unsigned int d = 0; d < dim; ++d) {
		if (dir[d] == 0.0f) {
			if (abs(off[d]) > m_Radius)
				return false;
			continue;
		}

		// calculate point of intersection
		T t1 = (off[d] - m_Radius) / dir[d];
		T t2 = (off[d] + m_Radius) / dir[d];
		if (t1 > t2)
			swap(t1,t2);
		
		if (t2 < tmin)
			return false;
		else if (t1 > tmin)
			tmin = t1;
		if (t1 > tmax)
			return false;
		else if (t2 < tmax)
			tmax = t2;
	}
		
	// Check if we should record the information
	if (is != NULL) {
		is->setPosition(line.getOrigin() + tmin*dir);
		is->setNormal(is->getPosition() - m_Origin);
		is->setParameter(tmin);
	}
	
	return true;
}


/*--------------------------------------------------------------------------*/
/** 
 * Returns true if the line intersects the cube, else this method
 * will return false. If the is argument is not NULL, an intersection
 * point will be returned.
 * @param line The line used for intersection
 * @param is The intersection info that will hold the point of
 *        intersection. Can be NULL.
 * @return true for intersection, otherwise false
 */
template<class T, int dim> 
inline bool Cube<T,dim>::intersect(const Line<T,dim>& line,Intersection<T,dim>* is) const {
	return intersectLinear(line, is);
}


/*--------------------------------------------------------------------------*/
/**
 * Returns true if the line segment intersects the cube, else this method
 * will return false. If the is argument is not NULL, an intersection
 * point will be returned.
 * @param seg The line segment used for intersection
 * @param is The intersection info that will hold the point of
 *        intersection. Can be NULL.
 * @return true for intersection, otherwise false
 */
template<class T, int dim> 
inline bool Cube<T,dim>::intersect(const Segment<T,dim>& seg,Intersection<T,dim>* is) const {
	return intersectLinear(seg, is);
}


/*--------------------------------------------------------------------------*/
/**
 * Returns true if the ray intersects the cube, else this method
 * will return false. If the is argument is not NULL, the first intersection
 * point will be returned.
 * @param ray The ray used for intersection
 * @param is The intersection info that will hold the point of
 *        intersection. Can be NULL.
 * @return true for intersection, otherwise false
*/
template<class T, int dim> 
inline bool Cube<T,dim>::intersect(const Ray<T,dim>& ray,Intersection<T,dim>* is) const {
	return intersectLinear(ray, is);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Box<T,dim> Cube<T,dim>::toBox() const {
	Vector<T,dim> a[dim];
	for (unsigned int i = 0; i < dim; i++) {
		a[i].null();
		a[i][i] = m_Radius;
	}
	return Box<T,dim>(m_Origin, a);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Sphere<T,dim> Cube<T,dim>::toSphere() const {
	return Sphere<T,dim>(m_Origin, m_Radius*sqrt(dim));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
AlignedBox<T,dim> Cube<T,dim>::toAlignedBox() const {
	Vector<T,dim> v;
	for (unsigned int i = 0; i < dim; i++)
		v[i] = m_Radius;
	return AlignedBox<T,dim>(m_Origin - v, m_Origin + v);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Cube<T,dim>::bound(const Vector<T,dim>* v,unsigned int n) {
	bound(v, sizeof(Vector<T,dim>), n);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Cube<T,dim>::bound(const Vector<T,dim>* v,unsigned int stride,unsigned int n) {
	Vector<T,dim> minimum = v[0];
	Vector<T,dim> maximum = v[0];
	
	for(unsigned int i = 0; i < n; i++) {
		for(unsigned int j = 0; j < dim; j++) {
			if (minimum[j] > (*v)[j])
				minimum[j] = (*v)[j];
			else if (maximum[j] < (*v)[j])
				maximum[j] = (*v)[j];
		}
		v = (const Vector<T,dim>*)(const void*)(((const byte*)v) + stride);
	}

	Vector<T,dim> diff = (maximum - minimum) / T(2);
	m_Origin = (maximum + minimum) / T(2);
	m_Radius = diff.max();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Cube<T,dim>::getVertices(Vector<T,dim>* cv) const {
	v = m_Origin;
	if (dim == 2) {
		cv[0][0] = v[0] + radius;  cv[0][1] = v[1] + radius;
		cv[1][0] = v[0] - radius;  cv[1][1] = v[1] + radius;
		cv[2][0] = v[0] + radius;  cv[2][1] = v[1] - radius;
		cv[3][0] = v[0] - radius;  cv[3][1] = v[1] - radius;
	} else if (dim == 3) {
		cv[0][0] = v[0] + radius;  cv[0][1] = v[1] + radius;  cv[0][2] = v[2] + radius;
		cv[1][0] = v[0] - radius;  cv[1][1] = v[1] + radius;  cv[1][2] = v[2] + radius;
		cv[2][0] = v[0] + radius;  cv[2][1] = v[1] - radius;  cv[2][2] = v[2] + radius;
		cv[3][0] = v[0] - radius;  cv[3][1] = v[1] - radius;  cv[3][2] = v[2] + radius;
		cv[4][0] = v[0] + radius;  cv[4][1] = v[1] + radius;  cv[4][2] = v[2] - radius;
		cv[5][0] = v[0] - radius;  cv[5][1] = v[1] + radius;  cv[5][2] = v[2] - radius;
		cv[6][0] = v[0] + radius;  cv[6][1] = v[1] - radius;  cv[6][2] = v[2] - radius;
		cv[7][0] = v[0] - radius;  cv[7][1] = v[1] - radius;  cv[7][2] = v[2] - radius;
	} else {
		for (i = 0; i < (1 << dim); i++) {
			cv[i] = v;
			for (j = 0; j < dim; j++) {
				if (i & (1 << j))
					cv[i][j] -= radius;
				else
					cv[i][j] += radius;
			}
		}
	}	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Cube2<T>::Cube2() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Cube2<T>::Cube2(const Vector2<T>& v,const T& t) {
	m_Origin[0] = v.x;
	m_Origin[1] = v.y;
	m_Radius = t;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Cube2<T>::Cube2(const Vector<T,2>& v,const T& r) : Cube<T,2>(v,r) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Cube2<T>::Cube2(const Cube<T,2>& s) : Cube<T,2>(s) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Cube3<T>::Cube3() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Cube3<T>::Cube3(const Vector3<T>& v,const T& r) {
	m_Origin[0] = v.x;
	m_Origin[1] = v.y;
	m_Origin[2] = v.z;
	m_Radius = t;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Cube3<T>::Cube3(const Vector<T,3>& v,const T& r) : Cube<T,3>(v,r) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Cube3<T>::Cube3(const Cube<T,3>& s) : Cube<T,3>(s) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Cube4<T>::Cube4() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Cube4<T>::Cube4(const Vector4<T>& v,const T& r) {
	m_Origin[0] = v.x;
	m_Origin[1] = v.y;
	m_Origin[2] = v.z;
	m_Origin[3] = v.w;
	m_Radius = t;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Cube4<T>::Cube4(const Vector<T,4>& v,const T& r) : Cube<T,4>(v,r) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Cube4<T>::Cube4(const Cube<T,4>& s) : Cube<T,4>(s) {
}


}; };

#endif
