/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GEOMETRY_DIRECTIONBOXCONSTRAINT_INL
#define __GEOMETRY_DIRECTIONBOXCONSTRAINT_INL

namespace magnum { namespace geometry {
using namespace magnum::math;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline DirectionBoxConstraint3<T>::DirectionBoxConstraint3() {
	m_Center.null();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline void DirectionBoxConstraint3<T>::setCenter(const Vector3<T>& c) {
	m_Center = c;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Vector3<T>& DirectionBoxConstraint3<T>::getCenter() const {
	return m_Center;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Vector3<T>& DirectionBoxConstraint3<T>::getDirection() const {
	return m_Dir;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Vector2<float>& DirectionBoxConstraint3<T>::getDirectionRange() const {
	return m_DirRange;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Vector3<T>& DirectionBoxConstraint3<T>::getUp() const {
	return m_Up;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline float DirectionBoxConstraint3<T>::getUpRange() const {
	return m_UpRange;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline void DirectionBoxConstraint3<T>::setDirection(const Vector3<T>& d) {
	m_Dir = d;
	m_Dir.normalise();
	m_Right = m_Dir % m_Up;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline void DirectionBoxConstraint3<T>::setDirection(const Vector3<T>& d,const Vector2<float>& r) {
	m_Dir = d;
	m_Dir.normalise();
	m_Right = m_Dir % m_Up;
	m_DirRange = r;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline void DirectionBoxConstraint3<T>::setUp(const Vector3<T>& u) {
	m_Up = u;
	m_Up.normalise();
	m_Right = m_Dir % m_Up;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline void DirectionBoxConstraint3<T>::setUp(const Vector3<T>& u,float r) {
	m_Up = u;
	m_Up.normalise();
	m_Right = m_Dir % m_Up;
	m_UpRange = r;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
bool DirectionBoxConstraint3<T>::onConstrain(Transform<T,3>& mtx) {
	bool changed = false;
	Vector3<T> mright = mtx.getAxis(0);
	Vector3<T> mup = mtx.getAxis(1);
	Vector3<T> mdir = mtx.getAxis(2);
	
	// Save length of vectors
	T ldir = mdir.length();
	T lright = mright.length();
	T lup = mup.length();
	
	// Project mdir into m_Right-_Dir-Plane
	Vector3<T> tmp = mdir - (mdir * m_Up) * m_Up;
	tmp.normalise();
	// Get angle between tmp and m_Up
	float alpha = acos(tmp * m_Dir)*sign(tmp * m_Right);

	// Project mdir into m_Right-_Dir-Plane
	tmp = mdir - (mdir * m_Right) * m_Right;
	tmp.normalise();
	// Get angle between tmp and m_Up
	float beta = acos(tmp * m_Dir)*sign(tmp * m_Up);
	
	// Clip angles
	if (alpha < -m_DirRange[0]) {
		alpha = -m_DirRange[0];
		changed = true;
	} else if (alpha > m_DirRange[0]) {
		alpha = m_DirRange[0];
		changed = true;
	}

	if (beta < -m_DirRange[1]) {
		beta = -m_DirRange[1];
		changed = true;
	} else if (beta > m_DirRange[1]) {
		beta = m_DirRange[1];
		changed = true;
	}
	
	if (changed) {
		// Construct new mdir
		Vector3<T> w1 = cos(alpha)*m_Dir + sin(alpha)*m_Right;
		Vector3<T> w2 = cos(beta)*w1 + sin(beta)*m_Up;
		mdir = w2*ldir;
		
		// Ajdust mup and mright
		mup = mup - (mup * w2) * w2;
		mup.normalise();
		mup *= lup;
		
		mright = mup % mdir;
		mright.normalise();
		mright *= lright;
	}

	// Construct new up vector
	Vector3<T> newUp = m_Up - (m_Up * mdir) / ldir*mdir;
	newUp.normalise();
	
	// Get Angle between mup and m_Up;
	float angle = acos((mup * newUp) / lup);
	if (abs(angle) > m_UpRange) {
		changed = true;
		
		// Clip angle
		if (angle < -m_UpRange)
			angle = -m_UpRange;
		else
			angle = m_UpRange;

		// Create orthonormal system from mup and m_Up;
		Vector3<T> base = mup - (mup * newUp) * newUp;
		base.normalise();
		
		// Construct new mup
		Vector3<T> tmp = cos(angle)*newUp + sin(angle)*base;
		mup = tmp*lup;
		
		// Ajdust mup and mright
		mdir = mdir - (mdir * tmp) * tmp;
		mdir.normalise();
		mdir *= ldir;
		
		mright = mup % mdir;
		mright.normalise();
		mright *= lright;
	}	


	if (changed) {
		Transform3<T> inv(mtx.getRotation().getInverse());
        inv.setTranslation( m_Center - inv * m_Center);
        Vector3<T> trans = inv*mtx.getTranslation();

		mtx.setAxis(0, mright);
		mtx.setAxis(1, mup);
		mtx.setAxis(2, mdir);
        mtx.setTranslation(mtx.getRotation() * (trans - m_Center) + m_Center);
    }

	return changed;
}

}; };

#endif
