/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GEOMETRY_DIRECTIONCONECONSTRAINT_INL
#define __GEOMETRY_DIRECTIONCONECONSTRAINT_INL

namespace magnum { namespace geometry {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline DirectionConeConstraint3<T>::DirectionConeConstraint3() {
	m_Center.null();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline void DirectionConeConstraint3<T>::setCenter(const Vector3<T>& c) {
	m_Center = c;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Vector3<T>& DirectionConeConstraint3<T>::getCenter() const {
	return m_Center;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Vector3<T>& DirectionConeConstraint3<T>::getDirection() const {
	return m_Dir;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline float DirectionConeConstraint3<T>::getDirectionRange() const {
	return m_DirRange;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Vector3<T>& DirectionConeConstraint3<T>::getUp() const {
	return m_Up;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline float DirectionConeConstraint3<T>::getUpRange() const {
	return m_UpRange;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline void DirectionConeConstraint3<T>::setDirection(const Vector3<T>& d) {
	m_Dir = d;
	m_Dir.normalize();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline void DirectionConeConstraint3<T>::setDirection(const Vector3<T>& d,float r) {
	m_Dir = d;
	m_Dir.normalize();
	m_DirRange = r;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline void DirectionConeConstraint3<T>::setUp(const Vector3<T>& u) {
	m_Up = u;
	m_Up.normalize();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline void DirectionConeConstraint3<T>::setUp(const Vector3<T>& u,float r) {
	m_Up = u;
	m_Up.normalize();
	m_UpRange = r;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
bool DirectionConeConstraint3<T>::onConstrain(Transform<T,3>& mtx) {
	bool changed = false;
	Vector3<T> mright = mtx.getAxis(0);
	Vector3<T> mup = mtx.getAxis(1);
	Vector3<T> mdir = mtx.getAxis(2);
	
	// Save length of vectors
	T ldir = mdir.length();
	T lright = mright.length();
	T lup = mup.length();
	
	// Get angle between mdir and m_Dir
	float angle = acos(mdir * m_Dir / ldir);
	
	// If angle does not lie within constraints, adjust mdir
	if (abs(angle) > m_DirRange) {
		changed = true;
		
		// Clip angle
		if (angle < -m_DirRange)
			angle = -m_DirRange;
		else
			angle = m_DirRange;

		// Create orthonormal system from mdir and _Dir;
		Vector3<T> base = mdir - (mdir * m_Dir) * m_Dir;
		base.normalize();
		
		// Construct new mdir
		Vector3<T> tmp = cos(angle)*m_Dir + sin(angle)*base;
		mdir = tmp*ldir;
		
		// Ajdust mup and mright
		mup = (mup - (mup * tmp) * tmp);
		mup.normalize();
		mup *= lup;
		
		mright = mup % mdir;
		mright.normalize();
		mright *= lright;
	}

	// Construct new up vector
	Vector3<T> newUp = m_Up - (m_Up * mdir)/ldir*mdir;
	newUp.normalize();
	
	// Get Angle between mup and _Up;
	angle = acos((mup * newUp) / lup);
	if (abs(angle) > m_UpRange) {
		changed = true;
		
		// Clip angle
		if (angle < -m_UpRange)
			angle = -m_UpRange;
		else
			angle = m_UpRange;

		// Create orthonormal system from mup and _Up;
		Vector3<T> base = mup - (mup * newUp) * newUp;
		base.normalize();
		
		// Construct new mup
		Vector3<T> tmp = cos(angle)*newUp + sin(angle)*base;
		mup = tmp*lup;
		
		// Ajdust mup and mright
		mdir = (mdir - (mdir * tmp)*tmp);
		mdir.normalize();
		mdir *= ldir;
		
		mright = mup % mdir;
		mright.normalize();
		mright *= lright;
	}	

	if (changed) {
		Transform3<T> inv(mtx.getRotation().getInverse());
        inv.setTranslation( m_Center - inv * m_Center);
        Vector3<T> trans = inv*mtx.getTranslation();

		mtx.setAxis(0, mright);
		mtx.setAxis(1, mup);
		mtx.setAxis(2, mdir);
        mtx.setTranslation(mtx.getRotation() * (trans - m_Center) + m_Center);
    }
	
	return changed;
}


}; };

#endif
