/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GEOMETRY_LINEAR_INL
#define __GEOMETRY_LINEAR_INL

namespace magnum { namespace geometry {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline Linear<T,dim>::Linear() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline Linear<T,dim>::Linear(const Vector<T,dim>& o,const Vector<T,dim>& d) :
	m_Origin(o), m_Direction(d) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void	Linear<T,dim>::setOrigin(const Vector<T,dim>& o) {
	m_Origin = o;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const Vector<T,dim>& Linear<T,dim>::getOrigin() const {
	return m_Origin;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void	Linear<T,dim>::setDirection(const Vector<T,dim>& d) {
	m_Direction = d;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const Vector<T,dim>& Linear<T,dim>::getDirection() const {
	return m_Direction;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline Vector<T,dim> Linear<T,dim>::operator()(const T& t) const {
	return m_Origin + t*m_Direction;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline T Linear<T,dim>::getDistance2(const Vector<T,dim>& v) const {
    Vector<T,dim> diff = v - m_Origin;
    T t = diff * m_Direction / m_Direction.length2();
    diff -= t * m_Direction;
    return diff.length2();	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline T Linear<T,dim>::getDistance(const Vector<T,dim>& v) const {
    Vector<T,dim> diff = v - m_Origin;
    T t = diff * m_Direction / m_Direction.length2();
    diff -= t * m_Direction;
    return diff.length();	
}


/*--------------------------------------------------------------------------*/
/** Transforms the current linear object (line, ray or segment) by
 * the given matrix. This transformation includes the affine part, thus
 * the matrix has dimension dim+1.
 * \param mtx Affine linear transformation matrix
 * \return Reference to the transformed object.
 */
template<class T, int dim>
Linear<T,dim>& Linear<T,dim>::transform(const Matrix<T,dim+1>& mtx) {
	if (dim == 2) {
		T x,y;
		x = mtx[0][0]*m_Origin[0] + mtx[0][1]*m_Origin[1] + mtx[0][2];
		y = mtx[1][0]*m_Origin[0] + mtx[1][1]*m_Origin[1] + mtx[1][2];
		m_Origin[0] = x;
		m_Origin[1] = y;
		x = mtx[0][0]*m_Direction[0] + mtx[0][1]*m_Direction[1];
		y = mtx[1][0]*m_Direction[0] + mtx[1][1]*m_Direction[1];
		m_Direction[0] = x;
		m_Direction[1] = y;
	} else if (dim == 3) {
		T x,y,z;
		x = mtx[0][0]*m_Origin[0] + mtx[0][1]*m_Origin[1] + mtx[0][2]*m_Origin[2] + mtx[0][3];
		y = mtx[1][0]*m_Origin[0] + mtx[1][1]*m_Origin[1] + mtx[1][2]*m_Origin[2] + mtx[1][3];
		z = mtx[2][0]*m_Origin[0] + mtx[2][1]*m_Origin[1] + mtx[2][2]*m_Origin[2] + mtx[2][3];
		m_Origin[0] = x;
		m_Origin[1] = y;
		m_Origin[2] = z;
		x = mtx[0][0]*m_Direction[0] + mtx[0][1]*m_Direction[1] + mtx[0][2]*m_Direction[2];
		y = mtx[1][0]*m_Direction[0] + mtx[1][1]*m_Direction[1] + mtx[1][2]*m_Direction[2];
		z = mtx[2][0]*m_Direction[0] + mtx[2][1]*m_Direction[1] + mtx[2][2]*m_Direction[2];
		m_Direction[0] = x;
		m_Direction[1] = y;
		m_Direction[2] = z;
	} else {
		Vector<T,dim> tmp;
		for (unsigned int i = 0; i < dim; i++) {
			tmp[i] = mtx[i][dim];
			for (unsigned int j = 0; j < dim; j++)
				tmp[i] += mtx[i][j]*m_Origin[j];
		}
		m_Origin = tmp;

		for (unsigned int i = 0; i < dim; i++) {
			tmp[i] = mtx[i][0]*m_Direction[0];
			for (unsigned int j = 1; j < dim; j++)
				tmp[i] += mtx[i][j]*m_Direction[j];
		}
		m_Direction = tmp;
	}
	
	return *this;
}


/*--------------------------------------------------------------------------*/
/** Transforms the current linear object (line, ray or segment) by
 * the given transformation. This transformation includes the affine part, thus
 * the matrix has dimension dim+1.
 * \param t Affine linear transformation
 * \return Reference to the transformed object.
 */
template<class T, int dim>
inline Linear<T,dim>& Linear<T,dim>::transform(const Transform<T,dim>& t) {
	return transform(t.toMatrix());
}

}; };

#endif
