/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GEOMETRY_SPACE_H
#define __GEOMETRY_SPACE_H

#include "geometry.cube.h"
#include "geometry.bounds.h"
#include "geometry.line.h"
#include "geometry.ray.h"
#include "geometry.segment.h"
#include "container.iterator.h"
#include "container.allocator.h"

namespace magnum { namespace geometry {
using magnum::container::IteratorBase;
using magnum::container::Allocator;
using magnum::math::Vector;
using magnum::object::Object;
using magnum::types::String;

template<class T, int dim, class Item> class SpaceIterator;
template<class T, int dim, class Base> class Placeable;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
class Space : public Object {
DeclareClass(Space, Object);
friend class SpaceIterator<T,dim,Item>;
friend class Placeable<T,dim,Item>;
private:
	struct Node {
	public:
		Node*					m_Parent;
		Cube<T,dim>				m_Cube;
		Placeable<T,dim,Item>*	m_Leafs;
		unsigned int			m_ChildTypes;
		unsigned int			m_Index;
		union {
			Node*					m_ChildNode[1 << dim];
			Placeable<T,dim,Item>*	m_ChildItem[1 << dim];
		};
	public:
		inline Node();

		inline void				setOrigin(const Vector<T,dim>& );
		inline const Vector<T,dim>&	getOrigin() const;

		inline void				setRadius(const T& );
		inline const T&			getRadius() const;
	};

public:
	typedef SpaceIterator<T,dim,Item>		Iterator;

public:
	Space();
	~Space();

	Iterator				insert(Item* );
	void					remove(Item* );
	inline void				remove(const Iterator& );
	void					update(Item* );
	bool					contains(const Item* ) const;
	void					clear();

	inline unsigned int		getSize() const;
	inline bool				isEmpty() const;

	inline const Cube<T,dim>& getBounds() const;

	inline Iterator			begin() const;
	inline Iterator			at(const Item* ) const;
	inline Iterator			end() const;

	template<class V, class F> 
	void					intersect(const V& ,F& ) const;
	template<class F> 
	void					intersect(const Vector<T,dim>& ,F& ) const;
	template<class F> 
	inline void				intersect(const Line<T,dim>& ,F& ) const;
	template<class F> 
	inline void				intersect(const Ray<T,dim>& ,F& ) const;
	template<class F> 
	inline void				intersect(const Segment<T,dim>& ,F& ) const;

	inline void				setMaxDepth(unsigned int );
	inline unsigned int		getMaxDepth() const;

private:
	template<class L,class F> 
	void					intersectLinear(const L& ,F& ) const;

	void					createTree(Item* );
	void					expandTree(T neededradius);
	void					expandChild(Node* node, unsigned int index);
	void					collapseNode(Node* node);

	Placeable<T,dim,Item>*	getFirstItem() const;

	template<class F> 
	void					intersectAll(F& ,const Node* ) const;

	inline unsigned int		getIndex(const Vector<T,dim>& icenter, const Vector<T,dim>& scenter) const;
	inline Vector<T,dim>	getCenter(const Vector<T,dim>& oldcenter, T radius, unsigned int index) const;

private:
	Allocator<Node>			m_Allocator;
	Node					m_Root;
	unsigned int			m_Size;
	unsigned int			m_MaxDepth;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Base> 
class Placeable {
friend class Space<T,dim,Base>;
friend class SpaceIterator<T,dim,Base>;
public:
	inline Placeable();
	inline Placeable(const Placeable& );
	~Placeable();
	
	inline Space<T,dim,Base>*	getSpace() const;
	
	inline const Bounds<T,dim>& getBounds() const;
	inline void					setBounds(const Bounds<T,dim>& );
	
	inline const Placeable&		operator=(const Placeable& );

private:
	Placeable*					getNextNode() const;

private:
	typename Space<T,dim,Base>::Node* m_Node;
	Space<T,dim,Base>*			m_Space;
	Bounds<T,dim>				m_Bounds;
	AlignedBox<T,dim>			m_Box;
	Placeable*					m_Next;
	Placeable*					m_Prev;
	unsigned int				m_Index;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
class SpaceIterator : public IteratorBase<SpaceIterator<T,dim,Item> > {
friend class Space<T,dim,Item>;
private:
	inline SpaceIterator(Placeable<T,dim,Item>* item);

public:
	inline SpaceIterator();
	inline					operator Item*() const;
	inline Item*			operator->() const;
	inline Item&			operator*() const;

	inline SpaceIterator&	operator++();
	inline SpaceIterator	operator++(int );
	inline SpaceIterator&	operator--();
	inline SpaceIterator	operator--(int );

	inline bool				operator==(const SpaceIterator& iter) const;
	inline bool				operator!=(const SpaceIterator& iter) const;
	inline bool				operator<(const SpaceIterator& iter) const;
	inline bool				operator>(const SpaceIterator& iter) const;
	inline bool				operator<=(const SpaceIterator& iter) const;
	inline bool				operator>=(const SpaceIterator& iter) const;

private:
	Placeable<T,dim,Item>*	m_Item;
};

}; };

#endif
