/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GEOMETRY_SPHERE_INL
#define __GEOMETRY_SPHERE_INL

#include "geometry.spherelimiter.h"

namespace magnum { namespace geometry {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline Sphere<T,dim>::Sphere() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline Sphere<T,dim>::Sphere(const Vector<T,dim>& o, const T& r) {
	m_Origin = o;
	m_Radius = r;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void Sphere<T,dim>::setOrigin(const Vector<T,dim>& o) {
	m_Origin = o;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const Vector<T,dim>& Sphere<T,dim>::getOrigin() const {
	return m_Origin;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void Sphere<T,dim>::setRadius(const T& r) {
	m_Radius = r;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const T& Sphere<T,dim>::getRadius() const {
	return m_Radius;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void Sphere<T,dim>::empty() {
	m_Radius = T(0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline bool Sphere<T,dim>::isEmpty() const {
	return (m_Radius == T(0.0));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline bool Sphere<T,dim>::isInside(const Vector<T,dim>& v) const {
	return ((v - m_Origin).length2() <= m_Radius * m_Radius);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline T Sphere<T,dim>::getDistance(const Vector<T,dim>& v) const {
	return ((v - m_Origin).length() - m_Radius);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Sphere<T,dim>::merge(const Sphere<T,dim>& s) {
	Vector<T,dim> diff = s.m_Origin - m_Origin;
	T dist = diff.length();
	if (s.m_Radius >= m_Radius) {
		// check if this is contained in s
		if (dist + m_Radius <= s.m_Radius) {
			m_Origin = s.m_Origin;
			m_Radius = s.m_Radius;
			return;
		}
	} else {
		if (dist + s.m_Radius <= m_Radius)
			return;
	}
	
	// We really need to merge these two
	T diam = dist + m_Radius + s.m_Radius;
	m_Origin = (m_Origin + s.m_Origin + (s.m_Radius - m_Radius)/dist * diff) / T(2);
	m_Radius = diam / T(2);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Sphere<T,dim>::intersect(const Sphere<T,dim>& s) const {
	Vector<T,dim> diff = s.m_Origin - m_Origin;
	T dist = diff.length();
	return (dist <= m_Radius + s.m_Radius);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Sphere<T,dim>::intersect(const Box<T,dim>& box) const {
	return box.intersect(*this);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Sphere<T,dim>::intersect(const AlignedBox<T,dim>& box) const {
	return box.intersect(*this);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Sphere<T,dim>::intersect(const Cube<T,dim>& cube) const {
	return cube.intersect(*this);
}


/*--------------------------------------------------------------------------*/
/** Intersects the sphere with a line. If the second parameter is not
 * NULL, the Intersection class will hold additional information.
 */
template<class T, int dim> 
bool Sphere<T,dim>::intersect(const Line<T,dim>& line,Intersection<T,dim>* is) const {
	Vector<T,dim> so = line.getOrigin() - m_Origin;
	T d2 = line.getDirection() * line.getDirection();
	T p = line.getDirection() * so;
	T p2 = p*p;
	T q = d2 * (so * so - m_Radius*m_Radius);
	if (q > p2)
		return false;
	if (is == NULL)
		return true;
		
	T s = sqrt(p2 - q);
	T t = (p + s) / (-d2);
	is->setPosition(line.getOrigin() + t*line.getDirection());
	is->setNormal(is->getPosition() - m_Origin);
	is->setParemeter(t);
	return true;
}


/*--------------------------------------------------------------------------*/
/** Intersects the sphere with a line sgment. Even if the line segment 
 * completely lies within the sphere, this method will return true. So
 * it won't be useful for raytracing or such things.
 */
template<class T, int dim> 
bool Sphere<T,dim>::intersect(const Segment<T,dim>& seg,Intersection<T,dim>* is) const {
	Vector<T,dim> so = seg.getOrigin() - m_Origin;
	T d2 = seg.getDirection() * seg.getDirection();
	T p = seg.getDirection() * so;
	T p2 = p*p;
	T q = d2 * (so * so - m_Radius*m_Radius);
	if (q > p2)
		return false;
		
	T s = sqrt(p2 - q);
	T t1 = (p - s) / (-d2);
	T t2 = (p + s) / (-d2); // t1 > t2
	if (t1 < T(0))
		return false;
	if (t2 > T(1))
		return false;
		
	if (is == NULL)		
		return true;

	T t = (t2 < T(0)) ? T(0) : t2; // holds the correct solution		
	is->setPosition(seg.getOrigin() + t*seg.getDirection());
	is->setNormal(is->getPosition() - m_Origin);
	is->setParemeter(t);
	return true;
}



/*--------------------------------------------------------------------------*/
/**
 */
template<class T, int dim> 
bool Sphere<T,dim>::intersect(const Ray<T,dim>& ray,Intersection<T,dim>* is) const {
	Vector<T,dim> so = ray.getOrigin() - m_Origin;
	T d2 = ray.getDirection() * ray.getDirection();
	T p = ray.getDirection() * so;
	T p2 = p*p;
	T q = d2 * (so * so - m_Radius*m_Radius);
	if (q > p2)
		return false;
		
	T s = sqrt(p2 - q);
	T t1 = (p - s) / (-d2);
	T t2 = (p + s) / (-d2); // t1 > t2
	if (t1 < T(0))
		return false;
		
	if (is == NULL)		
		return true;
		
	T t = (t2 < T(0)) ? T(0) : t2; // holds the correct solution		
	is->setPosition(ray.getOrigin() + t*ray.getDirection());
	is->setNormal(is->getPosition() - m_Origin);
	is->setParameter(t);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Cube<T,dim> Sphere<T,dim>::toCube() const {
	return Cube<T,dim>(m_Origin, m_Radius);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Box<T,dim> Sphere<T,dim>::toBox() const {
	Box<T,dim> box;
	box.setOrigin(m_Origin);
	for (unsigned int i = 0; i < dim; i++) {
		Vector<T,dim> v;
		v.null()
		v[i] = m_Radius;
		box.setAxis(i, v);
	}
	return box;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
AlignedBox<T,dim> Sphere<T,dim>::toAlignedBox() const {
	Vector<T,dim> mi, ma;
	mi = ma = m_Origin;
	for (unsigned int i = 0; i < dim; i++) {
		mi[i] -= m_Radius;
		ma[i] += m_Radius;
	}
	return AlignedBox<T,dim>(mi, ma);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Sphere<T,dim>::construct(const Vector<T,dim>& p0) {
	m_Origin = p0;
	m_Radius = T(0);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Sphere<T,dim>::construct(const Vector<T,dim>& p0, const Vector<T,dim>& p1) {
	m_Origin = (p0 + p1)/T(2);
	m_Radius = (m_Origin - p0).length();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Sphere<T,dim>::construct(const Vector<T,dim>& p0, const Vector<T,dim>& p1, const Vector<T,dim>& p2) {
    Vector<T,3> a = p0 - p2;
    Vector<T,3> b = p1 - p2;
    T aa = a * a;
    T ab = a * b;
    T bb = b * b;
    T det = aa*bb - ab*ab;

    if (abs(det) == T(0)) {
        m_Origin.null();
		m_Radius = T(0); // Infinity!
		return false;
    }

    T hid = T(0.5) / det;
    T u0 = hid * bb * (aa - ab);
    T u1 = hid * aa * (bb - ab);
    T u2 = T(1) - u0 - u1;
    m_Origin = u0*p0 + u1*p1 + u2*p2;
	m_Radius = (u0*a + u1*b).length();
	return true;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Sphere<T,dim>::construct(const Vector<T,dim>& p0, const Vector<T,dim>& p1, 
		const Vector<T,dim>& p2, const Vector<T,dim>& p3) {
    Vector<T,dim> v0 = p0 - p3;
    Vector<T,dim> v1 = p1 - p3;
    Vector<T,dim> v2 = p2 - p3;

    T mtx[3][3];
    mtx[0][0] = v0 * v0;
    mtx[0][1] = v0 * v1;
    mtx[0][2] = v0 * v2;
    mtx[1][0] = mtx[0][1];
    mtx[1][1] = v1 * v1;
    mtx[1][2] = v1 * v2;
    mtx[2][0] = mtx[0][2];
    mtx[2][1] = mtx[1][2];
    mtx[2][2] = v2 * v2;

    T B[3];
    B[0] = ((T)0.5)*mtx[0][0];
    B[1] = ((T)0.5)*mtx[1][1];
    B[2] = ((T)0.5)*mtx[2][2];

    T invmtx[3][3];
    invmtx[0][0] = mtx[1][1]*mtx[2][2] - mtx[1][2]*mtx[2][1];
    invmtx[0][1] = mtx[0][2]*mtx[2][1] - mtx[0][1]*mtx[2][2];
    invmtx[0][2] = mtx[0][1]*mtx[1][2] - mtx[0][2]*mtx[1][1];
    invmtx[1][0] = mtx[1][2]*mtx[2][0] - mtx[1][0]*mtx[2][2];
    invmtx[1][1] = mtx[0][0]*mtx[2][2] - mtx[0][2]*mtx[2][0];
    invmtx[1][2] = mtx[0][2]*mtx[1][0] - mtx[0][0]*mtx[1][2];
    invmtx[2][0] = mtx[1][0]*mtx[2][1] - mtx[1][1]*mtx[2][0];
    invmtx[2][1] = mtx[0][1]*mtx[2][0] - mtx[0][0]*mtx[2][1];
    invmtx[2][2] = mtx[0][0]*mtx[1][1] - mtx[0][1]*mtx[1][0];
    T det = mtx[0][0]*invmtx[0][0] + mtx[0][1]*invmtx[1][0] + mtx[0][2]*invmtx[2][0];

    if ( det == T(0) ) {
        m_Origin.null();
		m_Radius = T(0); // INfinity!
		return false;
	}

    T invdet = T(1)/det;
    int i, j;
    for (i = 0; i < 3; i++) {
        for (j = 0; j < 3; j++)
            invmtx[i][j] *= invdet;
    }
    
    T U[4];
    for (i = 0; i < 3; i++) {
        U[i] = 0.0f;
        for (j = 0; j < 3; j++)
            U[i] += invmtx[i][j]*B[j];
    }

    U[3] = (T)1.0 - U[0] - U[1] - U[2];
    
    m_Origin = U[0]*p0 + U[1]*p1 + U[2]*p2 + U[3]*p3;
    m_Radius = (U[0]*v0 + U[1]*v1 + U[2]*v2).length();

    return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Sphere<T,dim>::construct(const Vector<T,dim>* p, unsigned int count) {
	if (count == 1)
		return construct(P[0]);
	if (count > dim+1)
		return false;

	Vector<T,dim> kp[dim];
	Vector<T,dim> x;
	unsigned int i,j;
	
	for(i = 0; i < count-1; i++) {
		kp[i] = p[i] - p[count-1];
		T t = kp[i]*kp[i];
		m_LinearSystem(i,i) = t;
		x[i] = t * T(0.5);
		for (j = 0; j < i; j++) {
			m_LinearSystem(i,j) =
			m_LinearSystem(j,i) = kp[i]*kp[j];
		}
	}

	// Try to solve the system
	if (!m_LinearSystem.solve((T*)x);) {
        m_Origin.null();
		m_Radius = T(0); // INfinity!
		return false;
	}

	// Sum up the center by using the barycentric coordinates in x
	// and calc the radius by r^2 = |u0*A+u1*B+u2*C|^2
	Vector<T,dim> center;
	Vector<T,dim> radius;
	T xn;
	center = x[0]*p[0];
	radius = x[0]*kp[0];
	xn = T(1);
	for (i = 1; i < count-1; i++) {
		center += x[i]*p[i];
		radius += x[i]*kp[i];
		xn -= x[i];
	}
	center += xn*p[i];

	m_Origin = center;
	m_Radius = radius.length();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline bool	Sphere<T,dim>::bound(const Vector<T,dim>* v,unsigned int n) {
	return bound(v, sizeof(Vector<T,dim>), n);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool Sphere<T,dim>::bound(const Vector<T,dim>* points, unsigned int stride, unsigned int count) {
	m_Radius = T(0);
	m_Origin.null();
	
	if (count == 0)
		return true;
		
	SphereLimiter<T,dim> bnd;
	if (!bnd.limit(points, stride, count)) {
		return false;
	}

	m_Origin = bnd.getSphere().m_Origin;
	m_Radius = sqrt(bnd.getSphere().m_Radius);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Sphere2<T>::Sphere2() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Sphere2<T>::Sphere2(const Vector2<T>& v,const T& t) {
	m_Origin[0] = v.x;
	m_Origin[1] = v.y;
	m_Radius = t;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Sphere2<T>::Sphere2(const Vector<T,2>& v,const T& r) : Sphere<T,2>(v,r) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Sphere2<T>::Sphere2(const Sphere<T,2>& s) : Sphere<T,2>(s) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Sphere2<T>::bound(const Vector2<T>* v,unsigned int n) {
	return Sphere<T,2>::bound((const Vector<T,2>*)(const void*)v, sizeof(Vector<T,2>), n);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Sphere2<T>::bound(const Vector2<T>* v,unsigned int stride,unsigned int n) {
	return Sphere<T,2>::bound((const Vector<T,2>*)(const void*)v, stride, n);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Sphere3<T>::Sphere3() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Sphere3<T>::Sphere3(const Vector3<T>& v,const T& r) {
	m_Origin[0] = v.x;
	m_Origin[1] = v.y;
	m_Origin[2] = v.z;
	m_Radius = t;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Sphere3<T>::Sphere3(const Vector<T,3>& v,const T& r) : Sphere<T,3>(v,r) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Sphere3<T>::Sphere3(const Sphere<T,3>& s) : Sphere<T,3>(s) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Sphere3<T>::bound(const Vector3<T>* v,unsigned int n) {
	return Sphere<T,3>::bound((const Vector<T,3>*)(const void*)v, sizeof(Vector<T,3>), n);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Sphere3<T>::bound(const Vector3<T>* v,unsigned int stride,unsigned int n) {
	return Sphere<T,3>::bound((const Vector<T,3>*)(const void*)v, stride, n);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Sphere4<T>::Sphere4() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Sphere4<T>::Sphere4(const Vector4<T>& v,const T& r) {
	m_Origin[0] = v.x;
	m_Origin[1] = v.y;
	m_Origin[2] = v.z;
	m_Origin[3] = v.w;
	m_Radius = t;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Sphere4<T>::Sphere4(const Vector<T,4>& v,const T& r) : Sphere<T,4>(v,r) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Sphere4<T>::Sphere4(const Sphere<T,4>& s) : Sphere<T,4>(s) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Sphere4<T>::bound(const Vector4<T>* v,unsigned int n) {
	return Sphere<T,4>::bound((const Vector<T,4>*)(const void*)v, sizeof(Vector<T,4>), n);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Sphere4<T>::bound(const Vector4<T>* v,unsigned int stride,unsigned int n) {
	return Sphere<T,4>::bound((const Vector<T,4>*)(const void*)v, stride, n);
}

}; };

#endif
