/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GEOMETRY_SPHERELIMITER_INL
#define __GEOMETRY_SPHERELIMITER_INL

namespace magnum { namespace geometry {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
const Sphere<T,dim>& SphereLimiter<T,dim>::getSphere() const {
	return m_Sphere;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool SphereLimiter<T,dim>::limit(const Vector<T,dim>* points, unsigned int stride, unsigned int count) {
	if (count == 0)
		return false;
		
	m_Points = points;
	m_Stride = stride;

	m_Permutation.setSize(count);
	m_Permutation();

	m_SupportSize = 1;
	m_Support[0] = m_Permutation[0];

	m_Sphere.setRadius(T(0));
	m_Sphere.setOrigin(getPoint(m_Permutation[0]));
	
	for (unsigned int j = 3; j <= dim; j++)
		m_LinearSystem[j].setSize(j);
	
	unsigned int i = 1;
	unsigned int idx;
	while(i < count) {
		idx = m_Permutation[i];
		if (supports(idx)) {
			i++;
			continue;
		}
		if (contains(idx)) {
			i++;
			continue;
		}
		
		update(idx);
		i = 0;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline bool SphereLimiter<T,dim>::limit(const Vector<T,dim>* points, unsigned int count) {
	return limit(points, sizeof(Vector<T,dim>), count);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline bool SphereLimiter<T,dim>::contains(unsigned int idx) const {
	return ((getPoint(idx) - m_Sphere.getOrigin()).length2() <= m_Sphere.getRadius());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const Vector<T,dim>& SphereLimiter<T,dim>::getPoint(unsigned int idx) const {
	return *(Vector<T,dim>*)(((byte*)m_Points) + idx*m_Stride);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline bool SphereLimiter<T,dim>::supports(unsigned int idx) const {
	for (unsigned i = 0; i < m_SupportSize; i++) {
		if (m_Support[i] == idx)
			return true;
	}
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
void SphereLimiter<T,dim>::update(unsigned int idx) {
	// Loop over all possible support sizes
	if (updateSphere1(idx))
		return;
	if ((dim >= 2) && updateSphere2(idx))
		return;
	if ((dim >= 3) && updateSphere3(idx))
		return;
	for (unsigned int size = 4; size < m_SupportSize; size++) {
		if (updateSphereN(size, idx))
			return;
	} 
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool SphereLimiter<T,dim>::updateSphere1(unsigned int idx) {
	unsigned int i, j;
	for (i = 0; i < m_SupportSize; i++) {
		exactSphere2(m_Support[i], idx);
		for (j = 0; j < m_SupportSize; j++) {
			if (j == i)
				continue;
			if ((m_TempSphere.getOrigin() - getPoint(m_Support[j])).length2() > m_TempSphere.getRadius())
				break;
		}
		if (j == m_SupportSize)
			goto found;
	}
	
	return false;
	
found:	
	m_Sphere.setOrigin(m_TempSphere.getOrigin());
	m_Sphere.setRadius(max(m_Sphere.getRadius(), m_TempSphere.getRadius()));
	m_SupportSize = 2;
	m_Support[0] = m_Support[i];
	m_Support[1] = idx;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool SphereLimiter<T,dim>::updateSphere2(unsigned int idx) {
	unsigned int i0, i1, j;
	for (i0 = 1; i0 < m_SupportSize; i0++) {
		for (i1 = 0; i1 < i0; i1++) {
			if (!exactSphere3(m_Support[i0], m_Support[i1], idx))
				continue;
			for (j = 0; j < m_SupportSize; j++) {
				if ((j == i0) || (j == i1))
					continue;
				if ((m_TempSphere.getOrigin() - getPoint(m_Support[j])).length2() > m_TempSphere.getRadius())
					break;
			}
			if (j == m_SupportSize)
				goto found;
		}
	}
	
	return false;
	
found:	
	m_Sphere.setOrigin(m_TempSphere.getOrigin());
	m_Sphere.setRadius(max(m_Sphere.getRadius(), m_TempSphere.getRadius()));
	m_SupportSize = 3;
	i0 = m_Support[i0];
	i1 = m_Support[i1];
	m_Support[0] = i0;
	m_Support[1] = i1;
	m_Support[2] = idx;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool SphereLimiter<T,dim>::updateSphere3(unsigned int idx) {
	unsigned int i0, i1, i2, j;
	for (i0 = 2; i0 < m_SupportSize; i0++) {
		for (i1 = 1; i1 < i0; i1++) {
			for (i2 = 0; i2 < i1; i2++) {
				if (!exactSphere4(m_Support[i0], m_Support[i1], m_Support[i2], idx))
					continue;
				for (j = 0; j < m_SupportSize; j++) {
					if ((j == i0) || (j == i1) || (j == i2))
						continue;
					if ((m_TempSphere.getOrigin() - getPoint(m_Support[j])).length2() > m_TempSphere.getRadius())
						break;
				}
				if (j == m_SupportSize)
					goto found;
			}
		}
	}
	
	return false;
	
found:	
	m_Sphere.setOrigin(m_TempSphere.getOrigin());
	m_Sphere.setRadius(max(m_Sphere.getRadius(), m_TempSphere.getRadius()));
	m_SupportSize = 4;
	i0 = m_Support[i0];
	i1 = m_Support[i1];
	i2 = m_Support[i2];
	m_Support[0] = i0;
	m_Support[1] = i1;
	m_Support[2] = i2;
	m_Support[3] = idx;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool SphereLimiter<T,dim>::updateSphereN(unsigned int size, unsigned int idx) {
	ASSERT(size <= m_SupportSize);
	Tuple<unsigned int, dim> support;
	Tuple<unsigned int, dim+1> psupport;
	unsigned int i,j;

	for(i = 0; i < size; i++) {
		support[i] = i;
		psupport[i] = m_Support[i];
	}
	psupport[i] = idx;

	for (;;) {
		if (exactSphereN(psupport, size+1)) {
			for (j = 0; j < m_SupportSize; j++) {
				if (support.contains(j))
					continue;
				if ((m_TempSphere.getOrigin() - getPoint(m_Support[j])).length2() > m_TempSphere.getRadius())
					break;
			}
			if (j == m_SupportSize)
				goto found;
		}
		
		for (i = size-1; i >= 0; i--) {
			if (support[i] < m_SupportSize - size + i) {
				j = support[i];
				for(;i < size; i++) {
					j++;
					support[i] = j;
					psupport[i] = m_Support[j];
				}
				break;
			}
		}
		
		if (i == (unsigned int)-1)
			return false;
	}
	
found:	
	m_Sphere.setOrigin(m_TempSphere.getOrigin());
	m_Sphere.setRadius(max(m_Sphere.getRadius(), m_TempSphere.getRadius()));
	m_SupportSize = size+1;
	memcpy((unsigned int *)m_Support, (unsigned int*)psupport, (size+1)*sizeof(unsigned int));
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool SphereLimiter<T,dim>::exactSphere2(unsigned int i0, unsigned int i1) {
	const Vector<T,dim>& p0 = getPoint(i0);
	const Vector<T,dim>& p1 = getPoint(i1);
	Vector<T,dim> center = (p0 + p1)/T(2);
	m_TempSphere.setOrigin(center);
	m_TempSphere.setRadius((center - p0).length2());
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool SphereLimiter<T,dim>::exactSphere3(unsigned int i0, unsigned int i1, unsigned int i2) {
	const Vector<T,dim>& p0 = getPoint(i0);
	const Vector<T,dim>& p1 = getPoint(i1);
	const Vector<T,dim>& p2 = getPoint(i2);

    Vector<T,3> a = p0 - p2;
    Vector<T,3> b = p1 - p2;
    T aa = a * a;
    T ab = a * b;
    T bb = b * b;
    T det = aa*bb - ab*ab;

    if (abs(det) == T(0))
		return false;

    T hid = T(0.5) / det;
    T u0 = hid * bb * (aa - ab);
    T u1 = hid * aa * (bb - ab);
    T u2 = T(1) - u0 - u1;
    m_TempSphere.setOrigin(u0*p0 + u1*p1 + u2*p2);
	m_TempSphere.setRadius((u0*a + u1*b).length2());
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool SphereLimiter<T,dim>::exactSphere4(unsigned int i0, unsigned int i1, unsigned int i2, unsigned int i3) {
	const Vector<T,dim>& p0 = getPoint(i0);
	const Vector<T,dim>& p1 = getPoint(i1);
	const Vector<T,dim>& p2 = getPoint(i2);
	const Vector<T,dim>& p3 = getPoint(i3);
    Vector<T,dim> v0 = p0 - p3;
    Vector<T,dim> v1 = p1 - p3;
    Vector<T,dim> v2 = p2 - p3;

    T mtx[3][3];
    mtx[0][0] = v0 * v0;
    mtx[0][1] = v0 * v1;
    mtx[0][2] = v0 * v2;
    mtx[1][0] = mtx[0][1];
    mtx[1][1] = v1 * v1;
    mtx[1][2] = v1 * v2;
    mtx[2][0] = mtx[0][2];
    mtx[2][1] = mtx[1][2];
    mtx[2][2] = v2 * v2;

    T B[3];
    B[0] = ((T)0.5)*mtx[0][0];
    B[1] = ((T)0.5)*mtx[1][1];
    B[2] = ((T)0.5)*mtx[2][2];

    T invmtx[3][3];
    invmtx[0][0] = mtx[1][1]*mtx[2][2] - mtx[1][2]*mtx[2][1];
    invmtx[0][1] = mtx[0][2]*mtx[2][1] - mtx[0][1]*mtx[2][2];
    invmtx[0][2] = mtx[0][1]*mtx[1][2] - mtx[0][2]*mtx[1][1];
    invmtx[1][0] = mtx[1][2]*mtx[2][0] - mtx[1][0]*mtx[2][2];
    invmtx[1][1] = mtx[0][0]*mtx[2][2] - mtx[0][2]*mtx[2][0];
    invmtx[1][2] = mtx[0][2]*mtx[1][0] - mtx[0][0]*mtx[1][2];
    invmtx[2][0] = mtx[1][0]*mtx[2][1] - mtx[1][1]*mtx[2][0];
    invmtx[2][1] = mtx[0][1]*mtx[2][0] - mtx[0][0]*mtx[2][1];
    invmtx[2][2] = mtx[0][0]*mtx[1][1] - mtx[0][1]*mtx[1][0];
    T det = mtx[0][0]*invmtx[0][0] + mtx[0][1]*invmtx[1][0] + mtx[0][2]*invmtx[2][0];

    if ( det == T(0) )
		return false;

    T invdet = T(1)/det;
    int i, j;
    for (i = 0; i < 3; i++) {
        for (j = 0; j < 3; j++)
            invmtx[i][j] *= invdet;
    }
    
    T U[4];
    for (i = 0; i < 3; i++) {
        U[i] = 0.0f;
        for (j = 0; j < 3; j++)
            U[i] += invmtx[i][j]*B[j];
    }

    U[3] = (T)1.0 - U[0] - U[1] - U[2];
    
    m_TempSphere.setOrigin(U[0]*p0 + U[1]*p1 + U[2]*p2 + U[3]*p3);
    m_TempSphere.setRadius((U[0]*v0 + U[1]*v1 + U[2]*v2).length2());
    return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool SphereLimiter<T,dim>::exactSphereN(unsigned int* idc, unsigned int count) {
	ASSERT((1 < count) && (count <= dim+1));

	const Vector<T,dim>* p[dim+1];
	Vector<T,dim> kp[dim];
	Vector<T,dim> x;
	unsigned int i,j;
	
	p[count-1] = &getPoint(idc[count-1]);

	// Build linear system to solve
	LinearSystem<T>& ls = m_LinearSystem[count-1];
	for(i = 0; i < count-1; i++) {
		p[i] = &getPoint(idc[i]);
		kp[i] = *p[i] - *p[count-1];
		T t = kp[i]*kp[i];
		ls(i,i) = t;
		x[i] = t * T(0.5);
		for (j = 0; j < i; j++) {
			ls(i,j) =
			ls(j,i) = kp[i]*kp[j];
		}
	}

	// Try to solve the system
	if (!ls.solve((T*)x))
		return false;

	// Sum up the center by using the barycentric coordinates in x
	// and calc the radius by r^2 = |u0*A+u1*B+u2*C|^2
	Vector<T,dim> center;
	Vector<T,dim> radius;
	T xn;
	center = x[0] * *p[0];
	radius = x[0] * kp[0];
	xn = T(1);
	for (i = 1; i < count-1; i++) {
		center += x[i] * *p[i];
		radius += x[i] * kp[i];
		xn -= x[i];
	}
	center += xn * *p[i];

	m_TempSphere.setOrigin(center);
	m_TempSphere.setRadius(radius.length2());
	return true;
}


}; };

#endif
