/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GEOMETRY_TRANSFORM_H
#define __GEOMETRY_TRANSFORM_H

#include "math.matrix.h"
#include "math.vector.h"
#include "math.vector2.h"
#include "math.vector3.h"
#include "math.vector4.h"

namespace magnum { namespace geometry {
using magnum::math::Matrix;
using magnum::math::Matrix3;
using magnum::math::Matrix4;
using magnum::math::Vector;
using magnum::math::Vector2;
using magnum::math::Vector3;
using magnum::math::Vector4;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> class Transform {
public:
	enum Type {
		Id,
		Scale,
		Translation,
		Orthogonal,
		Rigid,
		Congruent,
		Affine,
	};

protected:
	Matrix<T,dim+1>		m_Matrix;
	Type				m_Type;
	T					m_Scale;

public:
	static const Transform<T,dim>	Identity;

public:
	inline Transform();
	inline Transform(const Matrix<T,dim+1>& );
	inline Transform(const Matrix<T,dim>& );
	inline Transform(const Vector<T,dim>& );
	inline Transform(const Matrix<T,dim>& ,const Vector<T,dim>& );

	inline Type					getType() const;
	inline T					getScale() const;
								
	inline Transform<T,dim>		operator-() const;
								
	inline Transform<T,dim>		operator+(const Transform<T,dim>& ) const;
	inline Transform<T,dim>		operator+(const Matrix<T,dim+1>& ) const;
	inline Transform<T,dim>		operator+(const Vector<T,dim>& ) const;
	inline Transform<T,dim>		operator-(const Transform<T,dim>& ) const;
	inline Transform<T,dim>		operator-(const Matrix<T,dim+1>& ) const;
	inline Transform<T,dim>		operator-(const Vector<T,dim>& ) const;
	inline Transform<T,dim>		operator*(const Transform<T,dim>& ) const;
	inline Transform<T,dim>		operator*(const Matrix<T,dim+1>& ) const;
	inline Transform<T,dim>		operator*(const Matrix<T,dim>& ) const;
	inline Transform<T,dim>		operator*(const T& ) const;
	inline Transform<T,dim>		operator/(const T& ) const;
								
	inline Vector<T,dim>		operator*(const Vector<T,dim>& v) const;
	inline Vector<T,dim+1>		operator*(const Vector<T,dim+1>& v) const;

	inline const Transform<T,dim>&	operator+=(const Transform<T,dim>& );
	inline const Transform<T,dim>&	operator+=(const Matrix<T,dim+1>& );
	inline const Transform<T,dim>&	operator+=(const Vector<T,dim>& );
	inline const Transform<T,dim>&	operator-=(const Transform<T,dim>& );
	inline const Transform<T,dim>&	operator-=(const Matrix<T,dim+1>& );
	inline const Transform<T,dim>&	operator-=(const Vector<T,dim>& );
	inline const Transform<T,dim>&	operator*=(const Transform<T,dim>& );
	inline const Transform<T,dim>&	operator*=(const Matrix<T,dim+1>& );
	inline const Transform<T,dim>&	operator*=(const Matrix<T,dim>& );
	inline const Transform<T,dim>&	operator*=(const T& );
	inline const Transform<T,dim>&	operator/=(const T& );

	inline const Transform<T,dim>&	operator=(const Matrix<T,dim>& );
	inline const Transform<T,dim>&	operator=(const Matrix<T,dim+1>& );

	inline const T*				operator[](unsigned int ) const;
	inline T*					operator[](unsigned int );
	inline const T&				operator()(unsigned int, unsigned int ) const;
	inline T&					operator()(unsigned int, unsigned int );

	inline const Transform<T,dim>&	mult(const Transform<T,dim>& );
	inline const Transform<T,dim>&	mult(const Matrix<T,dim+1>& );
	inline const Transform<T,dim>&	mult(const Matrix<T,dim>& );

	inline const Transform<T,dim>&	rmult(const Transform<T,dim>& );
	inline const Transform<T,dim>&	rmult(const Matrix<T,dim+1>& );
	inline const Transform<T,dim>&	rmult(const Matrix<T,dim>& );

	inline Transform<T,dim>		getInverse() const;
								
	inline Transform<T,dim>&	invert();
	inline Transform<T,dim>&	identity();
								
	inline void					setTranslation(const Vector<T,dim>& );
	inline Vector<T,dim>		getTranslation() const;
								
	void						setRotation(const Matrix<T,dim>& );
	Matrix<T,dim>				getRotation() const;

	inline void					setTransform(const Matrix<T,dim+1>& );
	inline const Matrix<T,dim+1>& getTransform() const;
	
	inline void					setAxis(unsigned int ,const Vector<T,dim>& );
	inline Vector<T,dim>		getAxis(unsigned int ) const;
								
	inline Transform<T,dim>&	translate(const Vector<T,dim>& );
	inline Transform<T,dim>&	rtranslate(const Vector<T,dim>& );
	Transform<T,dim>&			scale(const Vector<T,dim>& );
	Transform<T,dim>&			scale(const T& );
	
	inline Matrix<T,dim+1>&		toMatrix();
	inline const Matrix<T,dim+1>& toMatrix() const;
								
	inline						operator Matrix<T,dim+1>&();
	inline						operator const Matrix<T,dim+1>&() const;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> class Transform2 : public Transform<T,2> {
public:
	inline Transform2() {};
	inline Transform2(const Transform<T,2>& t) : Transform<T,2>(t) {};
	inline Transform2(const Matrix<T,3>& m) : Transform<T,2>(m) {};
	inline Transform2(const Matrix<T,2>& m) : Transform<T,2>(m) {};
	inline Transform2(const Vector<T,2>& v) : Transform<T,2>(v) {};
	inline Transform2(const Vector2<T>& v)  : Transform<T,2>((Vector<T,2>)v) {};
	inline Transform2(const Matrix<T,2>& m,const Vector<T,2>& v) : Transform<T,2>(m,v) {};
	inline Transform2(const Matrix<T,2>& m,const Vector2<T>& v)  : Transform<T,2>(m,(Vector<T,2>)v) {};

	inline void					setRotation(const Matrix<T,2>& );
	void						setRotation(const T& );
	Transform2<T>&				rotate(const T& );
	Transform2<T>&				rrotate(const T& );

	inline const Transform2<T>&	operator=(const Transform<T,2>& );
	inline const Transform2<T>&	operator=(const Matrix<T,2>& );
	inline const Transform2<T>&	operator=(const Matrix<T,3>& );

	inline Matrix3<T>&			toMatrix3();
	inline const Matrix3<T>&	toMatrix3() const;

	inline						operator Matrix3<T>&();
	inline						operator const Matrix3<T>&() const;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> class Transform3 : public Transform<T,3> {
public:
	inline Transform3() {};
	inline Transform3(const Transform<T,3>& t) : Transform<T,3>(t) {};
	inline Transform3(const Matrix<T,4>& m) : Transform<T,3>(m) {};
	inline Transform3(const Matrix<T,3>& m) : Transform<T,3>(m) {};
	inline Transform3(const Vector<T,3>& v) : Transform<T,3>(v) {};
	inline Transform3(const Vector3<T>& v)  : Transform<T,3>((Vector<T,3>)v) {};
	inline Transform3(const Matrix<T,3>& m,const Vector<T,3>& v) : Transform<T,3>(m,v) {};
	inline Transform3(const Matrix<T,3>& m,const Vector3<T>& v)  : Transform<T,3>(m,(Vector<T,3>)v) {};
	
	inline void					setRotation(const Matrix<T,3>& );
	void						setRotation(const T&, const Vector<T,3>& );
	Transform3<T>&				rotate(const T&, const Vector<T,3>& );
	Transform3<T>&				rrotate(const T&, const Vector<T,3>& );

	void						setRotationX(const T& );
	void						setRotationY(const T& );
	void						setRotationZ(const T& );

	Transform3<T>&				rotateX(const T& );
	Transform3<T>&				rotateY(const T& );
	Transform3<T>&				rotateZ(const T& );

	Transform3<T>&				rrotateX(const T& );
	Transform3<T>&				rrotateY(const T& );
	Transform3<T>&				rrotateZ(const T& );

	inline const Transform3<T>&	operator=(const Transform<T,3>& );
	inline const Transform3<T>&	operator=(const Matrix<T,3>& );
	inline const Transform3<T>&	operator=(const Matrix<T,4>& );

	inline Matrix4<T>&			toMatrix4();
	inline const Matrix4<T>&	toMatrix4() const;

	inline						operator Matrix4<T>&();
	inline						operator const Matrix4<T>&() const;
};

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> class Transform4 : public Transform<T,4> {
public:
	inline Transform4() {};
	inline Transform4(const Transform<T,4>& t) : Transform<T,4>(t) {};
	inline Transform4(const Matrix<T,5>& m) : Transform<T,4>(m) {};
	inline Transform4(const Matrix<T,4>& m) : Transform<T,4>(m) {};
	inline Transform4(const Vector<T,4>& v) : Transform<T,4>(v) {};
	inline Transform4(const Vector4<T>& v)  : Transform<T,4>((Vector<T,4>)v) {};
	inline Transform4(const Matrix<T,4>& m,const Vector<T,4>& v) : Transform<T,4>(m,v) {};
	inline Transform4(const Matrix<T,4>& m,const Vector4<T>& v)  : Transform<T,4>(m,(Vector<T,4>)v) {};

	inline const Transform4<T>&	operator=(const Transform<T,4>& );
	inline const Transform4<T>&	operator=(const Matrix<T,4>& );
	inline const Transform4<T>&	operator=(const Matrix<T,5>& );
};

}; };

#endif
