/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GEOMETRY_TRANSFORM_INL
#define __GEOMETRY_TRANSFORM_INL

namespace magnum { namespace geometry {
using magnum::math::_MatrixIdentity;
using magnum::math::Matrix2;
using magnum::math::Matrix3;
using magnum::math::Matrix4;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
const Transform<T,dim> Transform<T,dim>::Identity = Transform<T,dim>(_MatrixIdentity<T,dim+1>());


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim>::Transform() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim>::Transform(const Matrix<T,dim+1>& m) : m_Matrix(m) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim>::Transform(const Matrix<T,dim>& m) {
	for (unsigned int i = 0; i < dim; i++) {
		for (unsigned int j = 0; j < dim; j++)
			m_Matrix[i][j] = m[i][j];
		m_Matrix[i][dim] = m_Matrix[dim][i] = T(0);
	}
	m_Matrix[dim][dim] = T(1);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim>::Transform(const Matrix<T,dim>& m, const Vector<T,dim>& v) {
	for (unsigned int i = 0; i < dim; i++) {
		for (unsigned int j = 0; j < dim; j++)
			m_Matrix[i][j] = m[i][j];
		m_Matrix[i][dim] = v[i];
		m_Matrix[dim][i] = T(0);
	}
	m_Matrix[dim][dim] = T(1);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim>::Transform(const Vector<T,dim>& t) {
	unsigned int i,j;
	for (i = 0; i < dim; i++) {
		for (j = 0; j < dim; j++)
			m_Matrix[i][j] = T(0);
		m_Matrix[i][i] = T(1);
		m_Matrix[i][dim] = t[i];
	}
	for (j = 0; j < dim; j++)
		m_Matrix[dim][j] = T(0.0);
	m_Matrix[dim][j] = T(1.0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline typename Transform<T,dim>::Type Transform<T,dim>::getType() const {
	return m_Type;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline T Transform<T,dim>::getScale() const {
	return m_Scale;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim> Transform<T,dim>::operator-() const {
	return Transform<T,dim>(-m_Matrix);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim> Transform<T,dim>::operator+(const Transform<T,dim>& t) const {
	return Transform<T,dim>(m_Matrix + t.m_Matrix);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim> Transform<T,dim>::operator+(const Matrix<T,dim+1>& m) const {
	return Transform<T,dim>(m_Matrix + m);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim> Transform<T,dim>::operator+(const Vector<T,dim>& v) const {
	Transform<T,dim> t(m_Matrix);
	for (unsigned int i = 0; i < dim; i++)
		t.m_Matrix[i][3] += v[i];
	return t;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim> Transform<T,dim>::operator-(const Transform<T,dim>& t) const {
	return Transform<T,dim>(m_Matrix - t.m_Matrix);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim> Transform<T,dim>::operator-(const Matrix<T,dim+1>& m) const {
	return Transform<T,dim>(m_Matrix - m);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim> Transform<T,dim>::operator-(const Vector<T,dim>& v) const {
	Transform<T,dim> t(m_Matrix);
	for (unsigned int i = 0; i < dim; i++)
		t.m_Matrix[i][3] -= v[i];
	return t;
}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim> Transform<T,dim>::operator*(const Transform<T,dim>& t) const {
	return Transform<T,dim>(m_Matrix * t.m_Matrix);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim> Transform<T,dim>::operator*(const Matrix<T,dim+1>& m) const {
	return Transform<T,dim>(m_Matrix * m);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim> Transform<T,dim>::operator*(const Matrix<T,dim>& m) const {
	Transform<T,dim> t;
	if (dim == 2) {
		t.m_Matrix[0][0] = m_Matrix[0][0]*m[0][0] + m_Matrix[0][1]*m[1][0];
		t.m_Matrix[0][1] = m_Matrix[0][0]*m[0][1] + m_Matrix[0][1]*m[1][1];
		t.m_Matrix[0][2] = m_Matrix[0][2]; 
		t.m_Matrix[1][0] = m_Matrix[1][0]*m[0][0] + m_Matrix[1][1]*m[1][0];
		t.m_Matrix[1][1] = m_Matrix[1][0]*m[0][1] + m_Matrix[1][1]*m[1][1];
		t.m_Matrix[1][2] = m_Matrix[1][2]; 
		t.m_Matrix[2][0] = t.m_Matrix[2][1] = T(0);
		t.m_Matrix[2][2] = T(1);
		return t;
	}
	if (dim == 3) {
		t.m_Matrix[0][0] = m_Matrix[0][0]*m[0][0] + m_Matrix[0][1]*m[1][0] + m_Matrix[0][2]*m[2][0];
		t.m_Matrix[0][1] = m_Matrix[0][0]*m[0][1] + m_Matrix[0][1]*m[1][1] + m_Matrix[0][2]*m[2][1];
		t.m_Matrix[0][2] = m_Matrix[0][0]*m[0][2] + m_Matrix[0][1]*m[1][2] + m_Matrix[0][2]*m[2][2];
		t.m_Matrix[0][3] = m_Matrix[0][3]; 

		t.m_Matrix[1][0] = m_Matrix[1][0]*m[0][0] + m_Matrix[1][1]*m[1][0] + m_Matrix[1][2]*m[2][0];
		t.m_Matrix[1][1] = m_Matrix[1][0]*m[0][1] + m_Matrix[1][1]*m[1][1] + m_Matrix[1][2]*m[2][1];
		t.m_Matrix[1][2] = m_Matrix[1][0]*m[0][2] + m_Matrix[1][1]*m[1][2] + m_Matrix[1][2]*m[2][2];
		t.m_Matrix[1][3] = m_Matrix[1][3]; 

		t.m_Matrix[2][0] = m_Matrix[2][0]*m[0][0] + m_Matrix[2][1]*m[1][0] + m_Matrix[2][2]*m[2][0];
		t.m_Matrix[2][1] = m_Matrix[2][0]*m[0][1] + m_Matrix[2][1]*m[1][1] + m_Matrix[2][2]*m[2][1];
		t.m_Matrix[2][2] = m_Matrix[2][0]*m[0][2] + m_Matrix[2][1]*m[1][2] + m_Matrix[2][2]*m[2][2];
		t.m_Matrix[2][3] = m_Matrix[1][3]; 
		
		t.m_Matrix[3][0] = t.m_Matrix[3][1] = t.m_Matrix[3][2] = T(0);
		t.m_Matrix[3][3] = T(1);
		return t;
	}
	for (unsigned int i = 0; i < dim; i++) {
		for (unsigned int j = 0; j < dim; j++) {
			T tmp = m_Matrix[i][0]*m[0][j];
			for (unsigned int k = 1; k < dim; k++)
				tmp += m_Matrix[i][k]*m[k][j];
			t.m_Matrix[i][j] = tmp;
		}
		t.m_Matrix[i][dim] = m_Matrix[i][dim];
		t.m_Matrix[dim][i] = T(0);
	}
	t.m_Matrix[dim][dim] = T(1);
	return t;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim> Transform<T,dim>::operator*(const T& t) const {
	return Transform<T,dim>(m_Matrix * t);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim> Transform<T,dim>::operator/(const T& t) const {
	return Transform<T,dim>(m_Matrix / t);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Vector<T,dim> Transform<T,dim>::operator*(const Vector<T,dim>& v) const {
	Vector<T,dim> result;
	if (dim == 2) {
		result[0] = m_Matrix[0][0]*v[0] + m_Matrix[0][1]*v[1] + m_Matrix[0][2];
		result[1] = m_Matrix[1][0]*v[0] + m_Matrix[1][1]*v[1] + m_Matrix[1][2];
		return result;
	}
	if (dim == 3) {
		result[0] = m_Matrix[0][0]*v[0] + m_Matrix[0][1]*v[1] + m_Matrix[0][2]*v[2] + m_Matrix[0][3];
		result[1] = m_Matrix[1][0]*v[0] + m_Matrix[1][1]*v[1] + m_Matrix[1][2]*v[2] + m_Matrix[1][3];
		result[2] = m_Matrix[2][0]*v[0] + m_Matrix[2][1]*v[1] + m_Matrix[2][2]*v[2] + m_Matrix[2][3];
		return result;
	}
	for (unsigned int i = 0; i < dim; i++) {
		T tmp = m_Matrix[i][dim];
		for (unsigned int j = 0; j < dim; j++) 
			tmp += m_Matrix[i][j]*v[j];
		result[i] = tmp;	
	}
	return result;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Vector<T,dim+1> Transform<T,dim>::operator*(const Vector<T,dim+1>& v) const {
	Vector<T,dim+1> result;
	if (dim == 2) {
		result[0] = m_Matrix[0][0]*v[0] + m_Matrix[0][1]*v[1] + m_Matrix[0][2]*v[2];
		result[1] = m_Matrix[1][0]*v[0] + m_Matrix[1][1]*v[1] + m_Matrix[1][2]*v[2];
		result[2] = m_Matrix[2][0]*v[0] + m_Matrix[2][1]*v[1] + m_Matrix[2][2]*v[2];
		return result;
	}
	if (dim == 3) {
		result[0] = m_Matrix[0][0]*v[0] + m_Matrix[0][1]*v[1] + m_Matrix[0][2]*v[2] + m_Matrix[0][3]*v[3];
		result[1] = m_Matrix[1][0]*v[0] + m_Matrix[1][1]*v[1] + m_Matrix[1][2]*v[2] + m_Matrix[1][3]*v[3];
		result[2] = m_Matrix[2][0]*v[0] + m_Matrix[2][1]*v[1] + m_Matrix[2][2]*v[2] + m_Matrix[2][3]*v[3];
		result[3] = m_Matrix[3][0]*v[0] + m_Matrix[3][1]*v[1] + m_Matrix[3][2]*v[2] + m_Matrix[3][3]*v[3];
		return result;
	}
	for (unsigned int i = 0; i < dim+1; i++) {
		T tmp = m_Matrix[i][0]*v[0];
		for (unsigned int j = 1; j < dim+1; j++) 
			tmp += m_Matrix[i][j]*v[j];
		result[i] = tmp;	
	}
	return result;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::operator+=(const Transform<T,dim>& t) {
	m_Matrix += t.m_Matrix;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::operator+=(const Matrix<T,dim+1>& m) {
	m_Matrix += m;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::operator+=(const Vector<T,dim>& v) {
	for (unsigned int i = 0; i < dim; i++)
		m_Matrix[i][dim] += v[i];
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::operator-=(const Transform<T,dim>& t) {
	m_Matrix -= t.m_Matrix;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::operator-=(const Matrix<T,dim+1>& m) {
	m_Matrix -= m;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::operator-=(const Vector<T,dim>& v) {
	for (unsigned int i = 0; i < dim; i++)
		m_Matrix[i][dim] -= v[i];
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::operator*=(const Transform<T,dim>& t) {
	m_Matrix *= t.m_Matrix;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::operator*=(const Matrix<T,dim+1>& m) {
	m_Matrix *= m;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::operator*=(const Matrix<T,dim>& m) {
	return rmult(m);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::rmult(const Transform<T,dim>& t) {
	m_Matrix.rmult(t.m_Matrix);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::rmult(const Matrix<T,dim+1>& m) {
	m_Matrix.rmult(m);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::rmult(const Matrix<T,dim>& mtx) {
	if (dim == 2) {
		T r[4];
		r[0] = m_Matrix[0][0]*mtx[0][0] + m_Matrix[0][1]*mtx[1][0];
		r[1] = m_Matrix[0][0]*mtx[0][1] + m_Matrix[0][1]*mtx[1][1];
		r[2] = m_Matrix[1][0]*mtx[0][0] + m_Matrix[1][1]*mtx[1][0];
		r[3] = m_Matrix[1][0]*mtx[0][1] + m_Matrix[1][1]*mtx[1][1];
		m_Matrix[0][0] = r[0];
		m_Matrix[0][1] = r[1];
		m_Matrix[1][0] = r[2];
		m_Matrix[1][1] = r[3];
		return *this;
	}
	if (dim == 3) {
		T r[3];
		r[0] = m_Matrix[0][0]*mtx[0][0] + m_Matrix[0][1]*mtx[1][0] + m_Matrix[0][2]*mtx[2][0];
		r[1] = m_Matrix[0][0]*mtx[0][1] + m_Matrix[0][1]*mtx[1][1] + m_Matrix[0][2]*mtx[2][1];
		r[2] = m_Matrix[0][0]*mtx[0][2] + m_Matrix[0][1]*mtx[1][2] + m_Matrix[0][2]*mtx[2][2];
		m_Matrix[0][0] = r[0];
		m_Matrix[0][1] = r[1];
		m_Matrix[0][2] = r[2];
		r[0] = m_Matrix[1][0]*mtx[0][0] + m_Matrix[1][1]*mtx[1][0] + m_Matrix[1][2]*mtx[2][0];
		r[1] = m_Matrix[1][0]*mtx[0][1] + m_Matrix[1][1]*mtx[1][1] + m_Matrix[1][2]*mtx[2][1];
		r[2] = m_Matrix[1][0]*mtx[0][2] + m_Matrix[1][1]*mtx[1][2] + m_Matrix[1][2]*mtx[2][2];
		m_Matrix[1][0] = r[0];
		m_Matrix[1][1] = r[1];
		m_Matrix[1][2] = r[2];
		r[0] = m_Matrix[2][0]*mtx[0][0] + m_Matrix[2][1]*mtx[1][0] + m_Matrix[2][2]*mtx[2][0];
		r[1] = m_Matrix[2][0]*mtx[0][1] + m_Matrix[2][1]*mtx[1][1] + m_Matrix[2][2]*mtx[2][1];
		r[2] = m_Matrix[2][0]*mtx[0][2] + m_Matrix[2][1]*mtx[1][2] + m_Matrix[2][2]*mtx[2][2];
		m_Matrix[2][0] = r[0];
		m_Matrix[2][1] = r[1];
		m_Matrix[2][2] = r[2];
		return *this;
	}	
	T r[dim];
	for (unsigned int i = 0; i < dim; i++) {
		for (unsigned int j = 0; j < dim; j++) {
			T tmp = m_Matrix[i][0]*mtx[0][j];
			for (unsigned int k = 1; k < dim; k++)
				tmp += m_Matrix[i][k] * mtx[k][j];
			r[j] = tmp;
		}
		for (unsigned int j = 0; j < dim; j++)
			m_Matrix[i][j] = r[j];
	}
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::mult(const Transform<T,dim>& t) {
	m_Matrix.lmult(t.m_Matrix);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::mult(const Matrix<T,dim+1>& m) {
	m_Matrix.lmult(m);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::mult(const Matrix<T,dim>& mtx) {
	if (dim == 2) {
		T r[4];
		r[0] = mtx[0][0]*m_Matrix[0][0] + mtx[0][1]*m_Matrix[1][0];
		r[1] = mtx[0][0]*m_Matrix[0][1] + mtx[0][1]*m_Matrix[1][1];
		r[2] = mtx[1][0]*m_Matrix[0][0] + mtx[1][1]*m_Matrix[1][0];
		r[3] = mtx[1][0]*m_Matrix[0][1] + mtx[1][1]*m_Matrix[1][1];
		m_Matrix[0][0] = r[0];
		m_Matrix[0][1] = r[1];
		m_Matrix[1][0] = r[2];
		m_Matrix[1][1] = r[3];
		return *this;
	}
	if (dim == 3) {
		T r[3];
		r[0] = mtx[0][0]*m_Matrix[0][0] + mtx[0][1]*m_Matrix[1][0] + mtx[0][2]*m_Matrix[2][0];
		r[1] = mtx[1][0]*m_Matrix[0][0] + mtx[1][1]*m_Matrix[1][0] + mtx[1][2]*m_Matrix[2][0];
		r[2] = mtx[2][0]*m_Matrix[0][0] + mtx[2][1]*m_Matrix[1][0] + mtx[2][2]*m_Matrix[2][0];
		m_Matrix[0][0] = r[0];
		m_Matrix[1][0] = r[1];
		m_Matrix[2][0] = r[2];
		r[0] = mtx[0][0]*m_Matrix[0][1] + mtx[0][1]*m_Matrix[1][1] + mtx[0][2]*m_Matrix[2][1];
		r[1] = mtx[1][0]*m_Matrix[0][1] + mtx[1][1]*m_Matrix[1][1] + mtx[1][2]*m_Matrix[2][1];
		r[2] = mtx[2][0]*m_Matrix[0][1] + mtx[2][1]*m_Matrix[1][1] + mtx[2][2]*m_Matrix[2][1];
		m_Matrix[0][1] = r[0];
		m_Matrix[1][1] = r[1];
		m_Matrix[2][1] = r[2];
		r[0] = mtx[0][0]*m_Matrix[0][2] + mtx[0][1]*m_Matrix[1][2] + mtx[0][2]*m_Matrix[2][2];
		r[1] = mtx[1][0]*m_Matrix[0][2] + mtx[1][1]*m_Matrix[1][2] + mtx[1][2]*m_Matrix[2][2];
		r[2] = mtx[2][0]*m_Matrix[0][2] + mtx[2][1]*m_Matrix[1][2] + mtx[2][2]*m_Matrix[2][2];
		m_Matrix[0][2] = r[0];
		m_Matrix[1][2] = r[1];
		m_Matrix[2][2] = r[2];
		r[0] = mtx[0][0]*m_Matrix[0][3] + mtx[0][1]*m_Matrix[1][3] + mtx[0][2]*m_Matrix[2][3];
		r[1] = mtx[1][0]*m_Matrix[0][3] + mtx[1][1]*m_Matrix[1][3] + mtx[1][2]*m_Matrix[2][3];
		r[2] = mtx[2][0]*m_Matrix[0][3] + mtx[2][1]*m_Matrix[1][3] + mtx[2][2]*m_Matrix[2][3];
		m_Matrix[0][3] = r[0];
		m_Matrix[1][3] = r[1];
		m_Matrix[2][3] = r[2];
		return *this;
	}	
	T r[dim];
	for (unsigned int i = 0; i < dim; i++) {
		for (unsigned int j = 0; j < dim; j++) {
			T tmp = mtx[j][0]*m_Matrix[0][i];
			for (unsigned int k = 1; k < dim; k++)
				tmp += mtx[j][k] * m_Matrix[k][i];
			r[j] = tmp;
		}
		for (unsigned int j = 0; j < dim; j++)
			m_Matrix[j][i] = r[j];
	}
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::operator=(const Matrix<T,dim>& mtx) {
	for (unsigned int i = 0; i < dim-1; i++) {
		for (unsigned int j = 0; j < dim-1; j++)
			m_Matrix[i][j] = mtx[i][j];
		m_Matrix[i][j] = T(0);
	}
	for (unsigned int j = 0; j < dim-1; j++)
		m_Matrix[i][j] = T(0);
	m_Matrix[i][j] = T(1);	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::operator=(const Matrix<T,dim+1>& mtx) {
	m_Matrix = mtx;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::operator*=(const T& t) {
	m_Matrix *= t;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Transform<T,dim>& Transform<T,dim>::operator/=(const T& t) {
	m_Matrix /= t;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const T* Transform<T,dim>::operator[](unsigned int i) const {
	return m_Matrix[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline T* Transform<T,dim>::operator[](unsigned int i) {
	return m_Matrix[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const T& Transform<T,dim>::operator()(unsigned int i,unsigned int j) const {
	return m_Matrix[i][j];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline T& Transform<T,dim>::operator()(unsigned int i,unsigned int j) {
	return m_Matrix[i][j];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim> Transform<T,dim>::getInverse() const {
	return Transform<T,dim>(m_Matrix.getInverse());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim>& Transform<T,dim>::invert() {
	m_Matrix.invert();
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim>& Transform<T,dim>::identity() {
	m_Matrix.identity();
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Transform<T,dim>::setTranslation(const Vector<T,dim>& v) {
	for (unsigned int i = 0; i < dim; i++)
		m_Matrix[i][dim] = v[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Vector<T,dim> Transform<T,dim>::getTranslation() const {
	Vector<T,dim> v;
	for (unsigned int i = 0; i < dim; i++)
		v[i] = m_Matrix[i][dim];
	return v;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Transform<T,dim>::setRotation(const Matrix<T,dim>& mtx) {
	for (unsigned int i = 0; i < dim; i++) {
		for (unsigned int j = 0; j < dim; j++)
			m_Matrix[i][j] = mtx[i][j];
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Matrix<T,dim> Transform<T,dim>::getRotation() const {
	Matrix<T,dim> mtx;
	for (unsigned int i = 0; i < dim; i++) {
		for (unsigned int j = 0; j < dim; j++)
			mtx[i][j] = m_Matrix[i][j];
	}
	return mtx;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Transform<T,dim>::setTransform(const Matrix<T,dim+1>& mtx) {
	m_Matrix = mtx;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Matrix<T,dim+1>& Transform<T,dim>::getTransform() const {
	return m_Matrix;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Transform<T,dim>::setAxis(unsigned int i,const Vector<T,dim>& v) {
	ASSERT(i < dim);
	for (unsigned int j = 0; j < dim; j++)
		m_Matrix(j,i) = v[j];
	m_Matrix(dim,i) = T(0);	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Vector<T,dim> Transform<T,dim>::getAxis(unsigned int i) const {
	Vector<T,dim> v;
	for (unsigned int j = 0; j < dim; j++)
		v[j] = m_Matrix(j,i);
	return v;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim>& Transform<T,dim>::translate(const Vector<T,dim>& v) {
	for (unsigned int i = 0; i < dim; i++)
		m_Matrix[i][dim] += v[i];
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim>& Transform<T,dim>::rtranslate(const Vector<T,dim>& v) {
	Vector<T,dim> tmp = *this * v;
	for (unsigned int i = 0; i < dim; i++)
		m_Matrix[i][dim] = v[i];
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Transform<T,dim>& Transform<T,dim>::scale(const Vector<T,dim>& v) {
	for (unsigned int i = 0; i < dim; i++) {
		for (unsigned int j = 0; j < dim+1; j++)
			m_Matrix[i][j] *= v[i];
	}
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Transform<T,dim>& Transform<T,dim>::scale(const T& s) {
	for (unsigned int i = 0; i < dim; i++) {
		for (unsigned int j = 0; j < dim+1; j++)
			m_Matrix[i][j] *= s;
	}
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim>::operator Matrix<T,dim+1>&() {
	return m_Matrix;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Transform<T,dim>::operator const Matrix<T,dim+1>&() const {
	return m_Matrix;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Matrix<T,dim+1>& Transform<T,dim>::toMatrix() {
	return m_Matrix;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Matrix<T,dim+1>& Transform<T,dim>::toMatrix() const {
	return m_Matrix;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Transform2<T>& Transform2<T>::operator=(const Transform<T,2>& t) {
	Transform<T,2>::operator=(t); 
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Transform2<T>& Transform2<T>::operator=(const Matrix<T,2>& mtx) {
	Transform<T,2>::operator=(mtx); 
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Transform2<T>& Transform2<T>::operator=(const Matrix<T,3>& mtx) {
	Transform<T,2>::operator=(mtx); 
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline void Transform2<T>::setRotation(const Matrix<T,2>& mtx) {
	Transform<T,2>::setRotation(mtx);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Transform2<T>::setRotation(const T& a) {
	T c = T(cos(a));
	T s = T(sin(a));
	m_Matrix[0][0] = c
	m_Matrix[0][1] = s;
	m_Matrix[1][0] = -s;
	m_Matrix[1][1] = c;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Transform2<T>& Transform2<T>::rotate(const T& a) {
	T c = T(cos(a));
	T s = T(sin(a));

	//m_Matrix = mtx * m_Matrix;
	T r[2];
	r[0] = c*m_Matrix[0][0] + s*m_Matrix[1][0];
	r[1] = -s*m_Matrix[0][0] + c*m_Matrix[1][0];
	m_Matrix[0][0] = r[0];
	m_Matrix[1][0] = r[1];
	m_Matrix[2][0] = T(0);

	r[0] = c*m_Matrix[0][1] + s*m_Matrix[1][1];
	r[1] = -s*m_Matrix[0][1] + c*m_Matrix[1][1];
	m_Matrix[0][1] = r[0];
	m_Matrix[1][1] = r[1];
	m_Matrix[2][1] = T(0);
	
	r[0] = c*m_Matrix[0][2] + s*m_Matrix[1][2];
	r[1] = -s*m_Matrix[0][2] + c*m_Matrix[1][2];
	m_Matrix[0][2] = r[0];
	m_Matrix[1][2] = r[1];
	m_Matrix[2][2] = T(1);

	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Transform2<T>& Transform2<T>::rrotate(const T& a) {
	T c = T(cos(a));
	T s = T(sin(a));

	T r[2];
	r[0] = m_Matrix[0][0]*c - m_Matrix[0][1]*s;
	r[1] = m_Matrix[0][0]*s + m_Matrix[0][1]*c;
	m_Matrix[0][0] = r[0];
	m_Matrix[0][1] = r[1];
	m_Matrix[0][2] = T(0);
	r[0] = m_Matrix[1][0]*c - m_Matrix[1][1]*s;
	r[1] = m_Matrix[1][0]*s + m_Matrix[1][1]*c;
	m_Matrix[1][0] = r[0];
	m_Matrix[1][1] = r[1];
	m_Matrix[1][2] = T(0);
	
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Matrix3<T>& Transform2<T>::toMatrix3() {
	return *(Matrix3<T>*)&m_Matrix;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Matrix3<T>& Transform2<T>::toMatrix3() const {
	return *(const Matrix3<T>*)&m_Matrix;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Transform2<T>::operator Matrix3<T>&() {
	return *(Matrix3<T>*)&m_Matrix;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Transform2<T>::operator const Matrix3<T>&() const {
	return *(const Matrix3<T>*)&m_Matrix;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Transform3<T>& Transform3<T>::operator=(const Matrix<T,3>& mtx) {
	Transform<T,3>::operator=(mtx); 
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Transform3<T>& Transform3<T>::operator=(const Matrix<T,4>& mtx) {
	Transform<T,3>::operator=(mtx); 
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Transform3<T>& Transform3<T>::operator=(const Transform<T,3>& t) {
	Transform<T,3>::operator=(t); 
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline void Transform3<T>::setRotation(const Matrix<T,3>& mtx) {
	Transform<T,3>::setRotation(mtx);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Transform3<T>::setRotation(const T& a, const Vector<T,3>& v) {
	T c = T(cos(a));
	T _c = T(1.0) - c;
	T s = T(sin(a));
	T _s = T(1.0) - s;
	T n = v.length();
	T x = v[0]/n;
	T y = v[1]/n;
	T z = v[2]/n;

	m_Matrix[0][0] = x*x*_c + c;
	m_Matrix[0][1] = x*y*_c - z*s;
	m_Matrix[0][2] = x*z*_c + y*s;
	m_Matrix[1][0] = y*x*_c + z*s;
	m_Matrix[1][1] = y*y*_c + c;
	m_Matrix[1][2] = y*z*_c - x*s;
	m_Matrix[2][0] = x*z*_c - y*s;
	m_Matrix[2][1] = y*z*_c + x*s;
	m_Matrix[2][2] = z*z*_c + c;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Transform3<T>& Transform3<T>::rotate(const T& a, const Vector<T,3>& v) {
	T c = T(cos(a));
	T _c = T(1.0) - c;
	T s = T(sin(a));
	T _s = T(1.0) - s;
	T n = v.length();
	T x = v[0]/n;
	T y = v[1]/n;
	T z = v[2]/n;

	Matrix<T,3> mtx;
	mtx[0][0] = x*x*_c + c;
	mtx[0][1] = x*y*_c - z*s;
	mtx[0][2] = x*z*_c + y*s;
	mtx[1][0] = y*x*_c + z*s;
	mtx[1][1] = y*y*_c + c;
	mtx[1][2] = y*z*_c - x*s;
	mtx[2][0] = x*z*_c - y*s;
	mtx[2][1] = y*z*_c + x*s;
	mtx[2][2] = z*z*_c + c;
	
	//m_Matrix = mtx * m_Matrix;
	T r[4];
	r[0] = mtx[0][0]*m_Matrix[0][0] + mtx[0][1]*m_Matrix[1][0] + mtx[0][2]*m_Matrix[2][0];
	r[1] = mtx[1][0]*m_Matrix[0][0] + mtx[1][1]*m_Matrix[1][0] + mtx[1][2]*m_Matrix[2][0];
	r[2] = mtx[2][0]*m_Matrix[0][0] + mtx[2][1]*m_Matrix[1][0] + mtx[2][2]*m_Matrix[2][0];
	m_Matrix[0][0] = r[0];
	m_Matrix[1][0] = r[1];
	m_Matrix[2][0] = r[2];
	m_Matrix[3][0] = T(0);
	r[0] = mtx[0][0]*m_Matrix[0][1] + mtx[0][1]*m_Matrix[1][1] + mtx[0][2]*m_Matrix[2][1];
	r[1] = mtx[1][0]*m_Matrix[0][1] + mtx[1][1]*m_Matrix[1][1] + mtx[1][2]*m_Matrix[2][1];
	r[2] = mtx[2][0]*m_Matrix[0][1] + mtx[2][1]*m_Matrix[1][1] + mtx[2][2]*m_Matrix[2][1];
	m_Matrix[0][1] = r[0];
	m_Matrix[1][1] = r[1];
	m_Matrix[2][1] = r[2];
	m_Matrix[3][1] = T(0);
	r[0] = mtx[0][0]*m_Matrix[0][2] + mtx[0][1]*m_Matrix[1][2] + mtx[0][2]*m_Matrix[2][2];
	r[1] = mtx[1][0]*m_Matrix[0][2] + mtx[1][1]*m_Matrix[1][2] + mtx[1][2]*m_Matrix[2][2];
	r[2] = mtx[2][0]*m_Matrix[0][2] + mtx[2][1]*m_Matrix[1][2] + mtx[2][2]*m_Matrix[2][2];
	m_Matrix[0][2] = r[0];
	m_Matrix[1][2] = r[1];
	m_Matrix[2][2] = r[2];
	m_Matrix[3][2] = T(0);
	r[0] = mtx[0][0]*m_Matrix[0][3] + mtx[0][1]*m_Matrix[1][3] + mtx[0][2]*m_Matrix[2][3];
	r[1] = mtx[1][0]*m_Matrix[0][3] + mtx[1][1]*m_Matrix[1][3] + mtx[1][2]*m_Matrix[2][3];
	r[2] = mtx[2][0]*m_Matrix[0][3] + mtx[2][1]*m_Matrix[1][3] + mtx[2][2]*m_Matrix[2][3];
	m_Matrix[0][3] = r[0];
	m_Matrix[1][3] = r[1];
	m_Matrix[2][3] = r[2];
	m_Matrix[3][3] = T(1);

	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Transform3<T>& Transform3<T>::rrotate(const T& a, const Vector<T,3>& v) {
	T c = T(cos(a));
	T _c = T(1.0) - c;
	T s = T(sin(a));
	T _s = T(1.0) - s;
	T n = v.length();
	T x = v[0]/n;
	T y = v[1]/n;
	T z = v[2]/n;

	Matrix<T,3> mtx;
	mtx[0][0] = x*x*_c + c;
	mtx[0][1] = x*y*_c - z*s;
	mtx[0][2] = x*z*_c + y*s;
	mtx[1][0] = y*x*_c + z*s;
	mtx[1][1] = y*y*_c + c;
	mtx[1][2] = y*z*_c - x*s;
	mtx[2][0] = x*z*_c - y*s;
	mtx[2][1] = y*z*_c + x*s;
	mtx[2][2] = z*z*_c + c;
	
	//m_Matrix = m_Matrix * mtx;
	T r[4];
	r[0] = m_Matrix[0][0]*mtx[0][0] + m_Matrix[0][1]*mtx[1][0] + m_Matrix[0][2]*mtx[2][0];
	r[1] = m_Matrix[0][0]*mtx[0][1] + m_Matrix[0][1]*mtx[1][1] + m_Matrix[0][2]*mtx[2][1];
	r[2] = m_Matrix[0][0]*mtx[0][2] + m_Matrix[0][1]*mtx[1][2] + m_Matrix[0][2]*mtx[2][2];
	m_Matrix[0][0] = r[0];
	m_Matrix[0][1] = r[1];
	m_Matrix[0][2] = r[2];
	r[0] = m_Matrix[1][0]*mtx[0][0] + m_Matrix[1][1]*mtx[1][0] + m_Matrix[1][2]*mtx[2][0];
	r[1] = m_Matrix[1][0]*mtx[0][1] + m_Matrix[1][1]*mtx[1][1] + m_Matrix[1][2]*mtx[2][1];
	r[2] = m_Matrix[1][0]*mtx[0][2] + m_Matrix[1][1]*mtx[1][2] + m_Matrix[1][2]*mtx[2][2];
	m_Matrix[1][0] = r[0];
	m_Matrix[1][1] = r[1];
	m_Matrix[1][2] = r[2];
	r[0] = m_Matrix[2][0]*mtx[0][0] + m_Matrix[2][1]*mtx[1][0] + m_Matrix[2][2]*mtx[2][0];
	r[1] = m_Matrix[2][0]*mtx[0][1] + m_Matrix[2][1]*mtx[1][1] + m_Matrix[2][2]*mtx[2][1];
	r[2] = m_Matrix[2][0]*mtx[0][2] + m_Matrix[2][1]*mtx[1][2] + m_Matrix[2][2]*mtx[2][2];
	m_Matrix[2][0] = r[0];
	m_Matrix[2][1] = r[1];
	m_Matrix[2][2] = r[2];
	
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Transform3<T>::setRotationX(const T& a) {
	T c = T(cos(a));
	T s = T(sin(a));
	
	m_Matrix[0][0] = T(1);
	m_Matrix[0][1] = T(0);
	m_Matrix[0][2] = T(0);
	m_Matrix[1][0] = T(0);
	m_Matrix[1][1] = c;
	m_Matrix[1][2] = s;
	m_Matrix[2][0] = T(0);
	m_Matrix[2][1] = -s;
	m_Matrix[2][2] = c;
	m_Matrix[3][0] = T(0);
	m_Matrix[3][1] = T(0);
	m_Matrix[3][2] = T(0);
	m_Matrix[3][3] = T(1);
}

	
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Transform3<T>::setRotationY(const T& a){
	T c = T(cos(a));
	T s = T(sin(a));
	
	m_Matrix[0][0] = c;
	m_Matrix[0][1] = T(0);
	m_Matrix[0][2] = -s;
	m_Matrix[1][0] = T(0);
	m_Matrix[1][1] = T(1);
	m_Matrix[1][2] = T(0);
	m_Matrix[2][0] = s;
	m_Matrix[2][1] = T(0);
	m_Matrix[2][2] = c;
	m_Matrix[3][0] = T(0);
	m_Matrix[3][1] = T(0);
	m_Matrix[3][2] = T(0);
	m_Matrix[3][3] = T(1);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void Transform3<T>::setRotationZ(const T& a) {
	T c = T(cos(a));
	T s = T(sin(a));
	
	m_Matrix[0][0] = c;
	m_Matrix[0][1] = s;
	m_Matrix[0][2] = T(0);
	m_Matrix[1][0] = -s;
	m_Matrix[1][1] = c;
	m_Matrix[1][2] = T(0);
	m_Matrix[2][0] = T(0);
	m_Matrix[2][1] = T(0);
	m_Matrix[2][2] = T(1);
	m_Matrix[3][0] = T(0);
	m_Matrix[3][1] = T(0);
	m_Matrix[3][2] = T(0);
	m_Matrix[3][3] = T(1);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Transform3<T>& Transform3<T>::rotateX(const T& a) {
	T c = T(cos(a));
	T s = T(sin(a));

	T r[4];
	r[1] = c*m_Matrix[1][0] + s*m_Matrix[2][0];
	r[2] = -s*m_Matrix[1][0] + c*m_Matrix[2][0];
	m_Matrix[1][0] = r[1];
	m_Matrix[2][0] = r[2];
	m_Matrix[3][0] = T(0);
	r[1] = c*m_Matrix[1][1] + s*m_Matrix[2][1];
	r[2] = -s*m_Matrix[1][1] +c*m_Matrix[2][1];
	m_Matrix[1][1] = r[1];
	m_Matrix[2][1] = r[2];
	m_Matrix[3][1] = T(0);
	r[1] = c*m_Matrix[1][2] + s*m_Matrix[2][2];
	r[2] = -s*m_Matrix[1][2] + c*m_Matrix[2][2];
	m_Matrix[1][2] = r[1];
	m_Matrix[2][2] = r[2];
	m_Matrix[3][2] = T(0);
	r[1] = c*m_Matrix[1][3] + s*m_Matrix[2][3];
	r[2] = -s*m_Matrix[1][3] + c*m_Matrix[2][3];
	m_Matrix[1][3] = r[1];
	m_Matrix[2][3] = r[2];
	m_Matrix[3][3] = T(1);

	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Transform3<T>& Transform3<T>::rotateY(const T& a) {
	T c = T(cos(a));
	T s = T(sin(a));

	//m_Matrix = mtx * m_Matrix;
	T r[4];
	r[0] = c*m_Matrix[0][0] - s*m_Matrix[2][0];
	r[2] = s*m_Matrix[0][0] + c*m_Matrix[2][0];
	m_Matrix[0][0] = r[0];
	m_Matrix[2][0] = r[2];
	m_Matrix[3][0] = T(0);
	r[0] = c*m_Matrix[0][1] - s*m_Matrix[2][1];
	r[2] = s*m_Matrix[0][1] + c*m_Matrix[2][1];
	m_Matrix[0][1] = r[0];
	m_Matrix[2][1] = r[2];
	m_Matrix[3][1] = T(0);
	r[0] = c*m_Matrix[0][2] - s*m_Matrix[2][2];
	r[2] = s*m_Matrix[0][2] + c*m_Matrix[2][2];
	m_Matrix[0][2] = r[0];
	m_Matrix[2][2] = r[2];
	m_Matrix[3][2] = T(0);
	r[0] = c*m_Matrix[0][3] - s*m_Matrix[2][3];
	r[2] = s*m_Matrix[0][3] + c*m_Matrix[2][3];
	m_Matrix[0][3] = r[0];
	m_Matrix[2][3] = r[2];
	m_Matrix[3][3] = T(1);

	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Transform3<T>& Transform3<T>::rotateZ(const T& a) {
	T c = T(cos(a));
	T s = T(sin(a));

	T r[4];
	r[0] = c*m_Matrix[0][0] + s*m_Matrix[1][0];
	r[1] = -s*m_Matrix[0][0] + c*m_Matrix[1][0];
	m_Matrix[0][0] = r[0];
	m_Matrix[1][0] = r[1];
	m_Matrix[3][0] = T(0);
	r[0] = c*m_Matrix[0][1] + s*m_Matrix[1][1];
	r[1] = -s*m_Matrix[0][1] + c*m_Matrix[1][1];
	m_Matrix[0][1] = r[0];
	m_Matrix[1][1] = r[1];
	m_Matrix[3][1] = T(0);
	r[0] = c*m_Matrix[0][2] + s*m_Matrix[1][2];
	r[1] = -s*m_Matrix[0][2] + c*m_Matrix[1][2];
	m_Matrix[0][2] = r[0];
	m_Matrix[1][2] = r[1];
	m_Matrix[3][2] = T(0);
	r[0] = c*m_Matrix[0][3] + s*m_Matrix[1][3];
	r[1] = -s*m_Matrix[0][3] + c*m_Matrix[1][3];
	m_Matrix[0][3] = r[0];
	m_Matrix[1][3] = r[1];
	m_Matrix[3][3] = T(1);

	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Transform3<T>& Transform3<T>::rrotateX(const T& a) {
	T c = T(cos(a));
	T s = T(sin(a));

	T r[4];
	r[1] = m_Matrix[0][1]*c - m_Matrix[0][2]*s;
	r[2] = m_Matrix[0][1]*s + m_Matrix[0][2]*c;
	m_Matrix[0][1] = r[1];
	m_Matrix[0][2] = r[2];
	r[1] = m_Matrix[1][1]*c - m_Matrix[1][2]*s;
	r[2] = m_Matrix[1][1]*s + m_Matrix[1][2]*c;
	m_Matrix[1][1] = r[1];
	m_Matrix[1][2] = r[2];
	r[1] = m_Matrix[2][1]*c - m_Matrix[2][2]*s;
	r[2] = m_Matrix[2][1]*s + m_Matrix[2][2]*c;
	m_Matrix[2][1] = r[1];
	m_Matrix[2][2] = r[2];

	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Transform3<T>& Transform3<T>::rrotateY(const T& a) {
	T c = T(cos(a));
	T s = T(sin(a));

	T r[4];
	r[0] = m_Matrix[0][0]*c + m_Matrix[0][2]*s;
	r[2] = -m_Matrix[0][0]*s + m_Matrix[0][2]*c;
	m_Matrix[0][0] = r[0];
	m_Matrix[0][2] = r[2];
	r[0] = m_Matrix[1][0]*c + m_Matrix[1][2]*s;
	r[2] = -m_Matrix[1][0]*s + m_Matrix[1][2]*c;
	m_Matrix[1][0] = r[0];
	m_Matrix[1][2] = r[2];
	r[0] = m_Matrix[2][0]*c + m_Matrix[2][2]*s;
	r[2] = -m_Matrix[2][0]*s + m_Matrix[2][2]*c;
	m_Matrix[2][0] = r[0];
	m_Matrix[2][2] = r[2];

	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Transform3<T>& Transform3<T>::rrotateZ(const T& a) {
	T c = T(cos(a));
	T s = T(sin(a));

	T r[4];
	r[0] = m_Matrix[0][0]*c - m_Matrix[0][1]*s;
	r[1] = m_Matrix[0][0]*s + m_Matrix[0][1]*c;
	m_Matrix[0][0] = r[0];
	m_Matrix[0][1] = r[1];
	r[0] = m_Matrix[1][0]*c - m_Matrix[1][1]*s;
	r[1] = m_Matrix[1][0]*s + m_Matrix[1][1]*c;
	m_Matrix[1][0] = r[0];
	m_Matrix[1][1] = r[1];
	r[0] = m_Matrix[2][0]*c - m_Matrix[2][1]*s;
	r[1] = m_Matrix[2][0]*s + m_Matrix[2][1]*c ;
	m_Matrix[2][0] = r[0];
	m_Matrix[2][1] = r[1];

	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Matrix4<T>& Transform3<T>::toMatrix4() {
	return *(Matrix4<T>*)&m_Matrix;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Matrix4<T>& Transform3<T>::toMatrix4() const {
	return *(const Matrix4<T>*)&m_Matrix;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Transform3<T>::operator Matrix4<T>&() {
	return *(Matrix4<T>*)&m_Matrix;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Transform3<T>::operator const Matrix4<T>&() const {
	return *(const Matrix4<T>*)&m_Matrix;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Transform4<T>& Transform4<T>::operator=(const Matrix<T,4>& mtx) {
	Transform<T,4>::operator=(mtx); 
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Transform4<T>& Transform4<T>::operator=(const Matrix<T,5>& mtx) {
	Transform<T,4>::operator=(mtx); 
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Transform4<T>& Transform4<T>::operator=(const Transform<T,4>& t) {
	Transform<T,4>::operator=(t); 
	return *this;
}

}; };

#endif
