/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GRAPHICS3D_ANIMATION_INL
#define __GRAPHICS3D_ANIMATION_INL

#include "graphics3d.parameterbuffer.h"

namespace magnum { namespace graphics3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline String Animation::getName() const {
	return m_Name;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Animation::setName(const char* name) {
	m_Name = name;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Animation::setStart(float s) {
	m_Start = s;
	m_Length = m_End - m_Start;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline float Animation::getStart() const {
	return m_Start;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Animation::setEnd(float e) {
	m_End = e;
	m_Length = m_End - m_Start;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline float Animation::getEnd() const {
	return m_End;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline float Animation::getLength() const {
	return m_Length;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Animation::setDirection(Animation::Direction d) {
	m_Direction = d;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Animation::Direction Animation::getDirection() const {
	return m_Direction;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Animation::setRepeat(Animation::Repeat r) {
	m_Repeat = r;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Animation::Repeat Animation::getRepeat() const {
	return m_Repeat;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Animation::setInterpolation(Animation::Interpolation i) {
	m_Interpolation = i;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Animation::Interpolation Animation::getInterpolation() const {
	return m_Interpolation;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const ParameterFormat& Animation::getParameters() const {
	return m_Parameters;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline String Animation::getLocation() const {
	return m_Location;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Animation::setLocation(const char* name) {
	m_Location = name;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Animation::publishParameters(const ParameterFormat& fmt) {
	m_Parameters = fmt;
	m_Parameters.shrink();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Animation::animate(float t,void* data) const {
	onAnimate(t,data);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Animation::animate(float t,ParameterBuffer* buf) const {
	ParameterFormat::Location l = buf->getFormat()->getLocation(m_Location);
	onAnimate(t,buf->getPointer(l));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Animation::animate(float t,ParameterBuffer* buf,ParameterFormat::Location l) const {
	onAnimate(t,buf->getPointer(l));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
const T* Animation::getKeyframe(const Array<T>& frames,float t) const {
	ASSERT(frames.getSize() > 0);
	const T* lo = frames.begin();
	const T* hi = frames.end();
	const T* mid;
	size_t num = hi - lo;
	size_t half;
	while(lo < hi) {
        if (half = num / 2) {
            mid = lo + half;
            if (mid->getTime() > t) {
				hi = mid;
                num = half; // num & 1 ? half : half-1;
            } else {
                lo = mid; // + 1;
                num = num & 1 ? half+1 : half;
            }
        } else {
            return lo; // mid->getTime() < t ? mid : lo;
		}
	}
	return lo;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
const T* Animation::getKeyframe(const Array<T>& frames,float t,const T* last) const {
	ASSERT(frames.getSize() > 0);
	if (t >= last->getTime()) {
		const T* next = last + 1;
		if (next == frames.end())
			return last;
		if (t < next->getTime())
			return last;
		
		const T* nextnext = next + 1;
		if (nextnext == frames.end())
			return next;
		if (t < nextnext->getTime())
			return next;
		return getKeyframe(frames,t);
	} else {
		const T* prev = last - 1;
		if (prev < &frames[0])
			return last;
		if (t >= prev->getTime())
			return prev;
		return getKeyframe(frames,t);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
void Animation::sortKeyframes(Array<T>& frames) {
	if (frames.getSize() == 0) {
		setStart(0.0f);
		setEnd(0.0f);
		setRepeat(Single);
		return;
	}
	if (frames.getSize() == 1) {
		setStart(frames[0].getTime());
		setEnd(frames[0].getTime());
		setRepeat(Single);
		frames[0].setDuration(FLT_MAX);
		return;
	}
	sort(frames.begin(), frames.end());
	setStart(frames[0].getTime());
	setEnd(frames[frames.getSize()-1].getTime());
	
	float time = getStart();
	unsigned int i;
	for (i = 0; i < frames.getSize()-1; i++) {
		frames[i].setDuration(frames[i+1].getTime() - time);
		time = frames[i+1].getTime();
	}
	frames[i].setDuration(FLT_MAX);
}

}; };

#endif
