/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GRAPHICS3D_BOUNDS_INL
#define __GRAPHICS3D_BOUNDS_INL

namespace magnum { namespace graphics3d {
	
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
bool fitSphere(VertexBuffer* vb, Sphere<T,3>* sphere) {
	ASSERT(vb != NULL);
	ASSERT(sphere != NULL);

	// Reset sphere
	sphere->empty();
	
	// Check if vertex buffer has correct attributes
	if ((vb->getAttributes() & Attributes::Vertex) == 0)
		return false;
	
	// Search attribute
	const VertexFormat* vfmt;
	vfmt = &vb->getFormat();
	unsigned int c;
	for (c = 0; c < vfmt->getSize(); c++) {
		if (vfmt->getAttribute(c) == Attributes::Vertex)
			goto attribFound;
	}
	
	return false;
	
attribFound:
	// Save some data
	unsigned int size = vb->getSize();
	unsigned int offset = vfmt->getOffset(c);
	unsigned int stride = vfmt->getStride();
	Types::Type type = vfmt->getType(c);
	
	// Lock vertex data	
	const void* data = vb->lock(VertexBuffer::ReadOnly);
	data = (const void*)((const byte*)data + offset);
	if (data == NULL)
		return false;

	switch(type) {
		case Types::Byte3:
		case Types::Byte4:
			Sphere<byte,3> bSphere;
			bSphere.bound((const Vector<byte,3>*)data, stride, size);
			sphere->setOrigin((Vector<T,3>)bSphere.getOrigin());
			sphere->setRadius((T)bSphere.getRadius());
			break;

		case Types::Integer3:
		case Types::Integer4:
			Sphere<int,3> iSphere;
			iSphere.bound((const Vector<int,3>*)data, stride, size);
			sphere->setOrigin((Vector<T,3>)iSphere.getOrigin());
			sphere->setRadius((T)iSphere.getRadius());
			break;

		case Types::Float3:
		case Types::Float4:
			Sphere<float,3> fSphere;
			fSphere.bound((const Vector<float,3>*)data, stride, size);
			sphere->setOrigin((Vector<T,3>)fSphere.getOrigin());
			sphere->setRadius((T)fSphere.getRadius());
			break;

		case Types::Double3:
		case Types::Double4:
			Sphere<double,3> dSphere;
			dSphere.bound((const Vector<double,3>*)data, stride, size);
			sphere->setOrigin((Vector<T,3>)dSphere.getOrigin());
			sphere->setRadius((T)dSphere.getRadius());
			break;
			
		default:
			vb->unlock();
			return false;
	}
		
	vb->unlock();
	return true;
}

		
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
bool fitAlignedBox(VertexBuffer* vb, AlignedBox<T,3>* box) {
	ASSERT(vb != NULL);
	ASSERT(box != NULL);

	// Reset box
	box->empty();
	
	// Check if vertex buffer has correct attributes
	if ((vb->getAttributes() & Attributes::Vertex) == 0)
		return false;
	
	// Search attribute
	const VertexFormat* vfmt;
	vfmt = &vb->getFormat();
	unsigned int c;
	for (c = 0; c < vfmt->getSize(); c++) {
		if (vfmt->getAttribute(c) == Attributes::Vertex)
			goto attribFound;
	}
	
	return false;
	
attribFound:
	// Save some data
	unsigned int size = vb->getSize();
	unsigned int offset = vfmt->getOffset(c);
	unsigned int stride = vfmt->getStride();
	Types::Type type = vfmt->getType(c);
	
	// Lock vertex data	
	const void* data = vb->lock(VertexBuffer::ReadOnly);
	data = (const void*)((const byte*)data + offset);
	if (data == NULL)
		return false;

	switch(type) {
		case Types::Byte3:
		case Types::Byte4: {
			AlignedBox<byte,3> bBox;
			bBox.bound((const Vector<byte,3>*)data, stride, size);
			box->setMinimum((Vector<T,3>)bBox.getMinimum());
			box->setMaximum((Vector<T,3>)bBox.getMaximum());
			break;
		}

		case Types::Integer3:
		case Types::Integer4: {
			AlignedBox<int,3> iBox;
			iBox.bound((const Vector<int,3>*)data, stride, size);
			box->setMinimum((Vector<T,3>)iBox.getMinimum());
			box->setMaximum((Vector<T,3>)iBox.getMaximum());
			break;
		}

		case Types::Float3:
		case Types::Float4: {
			AlignedBox<float,3> fBox;
			fBox.bound((const Vector<float,3>*)data, stride, size);
			box->setMinimum((Vector<T,3>)fBox.getMinimum());
			box->setMaximum((Vector<T,3>)fBox.getMaximum());
			break;
		}

		case Types::Double3:
		case Types::Double4: {
			AlignedBox<double,3> dBox;
			dBox.bound((const Vector<double,3>*)data, stride, size);
			box->setMinimum((Vector<T,3>)dBox.getMinimum());
			box->setMaximum((Vector<T,3>)dBox.getMaximum());
			break;
		}
			
		default:
			vb->unlock();
			return false;
	}
		
	vb->unlock();
	return true;
}



}; };

#endif
