/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GRAPHICS3D_CONTEXT_H
#define __GRAPHICS3D_CONTEXT_H

#include "container.chain.h"
#include "system.thread.h"
#include "graphics3d.colourmaterial.h"
#include "graphics3d.framebuffer.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.depthmode.h"
#include "graphics3d.polygonmode.h"
#include "graphics3d.plainshader.h"
#include "graphics3d.texture.h"
#include "graphics3d.mesh.h"
#include "graphics3d.matrixstack.h"
#include "graphics3d.lightstack.h"
#include "graphics3d.capabilities.h"
#include "graphics3d.deferqueue.h"


namespace magnum { namespace graphics3d {
using magnum::object::Object;
using magnum::types::String;
using magnum::system::Thread;
using magnum::container::Chain;
using magnum::container::Chainable;

class IndexBuffer;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
class Context : public Object, private Chainable<Context>, private magnum::object::Noncopyable {
DeclareClass(Context, Object);
friend class Resource;
friend class IndexBuffer;
friend class FrameBuffer;
friend class Shader;
friend class Material;
friend class StencilMode;
friend class BlendMode;
friend class PolygonMode;
friend class DepthMode;
friend class Texture;
friend class Texture1D;
friend class Texture2D;
friend class Texture3D;
friend class DepthTexture;
friend class CubeTexture;
public:
	enum CullMode {
		CullFront = 1,
		CullBack = 2,
		CullNone = 3
	};

	enum ClearFlag {
		ClearColour = GL_COLOR_BUFFER_BIT,
		ClearDepth = GL_DEPTH_BUFFER_BIT,
		ClearStencil = GL_STENCIL_BUFFER_BIT,
		ClearAccum = GL_ACCUM_BUFFER_BIT
	};
	

private:
	Context();

public:
	~Context();
	
	static Context*			create();
	Context*				share() const;

	inline static Context*	getCurrent();
	bool					makeCurrent();
	
	inline const Dimension<int>& getSize() const;

	inline const FrameBuffer* getFrameBuffer() const;

	inline const Shader*	getShader() const;
	inline const Material*	getMaterial() const;
	inline const BlendMode*	getBlendMode() const;
	inline const PolygonMode* getPolygonMode() const;
	inline const DepthMode*	getDepthMode() const;
	inline const StencilMode* getStencilMode() const;
	inline const Texture*	getTexture(unsigned int = 0) const;
	inline unsigned int		getTextureTarget(unsigned int = 0) const;

	inline Capabilities*	getCapabilities();
	inline const Capabilities* getCapabilities() const;
	inline LightStack*		getLightStack();
	inline const LightStack* getLightStack() const;
	inline MatrixStack*		getMatrixStack();
	inline const MatrixStack* getMatrixStack() const;

	inline void				flush();
	inline void				swap();

	// Depth Buffer
	inline void				setDepthRange(float , float);
	inline void				setDepthRange(const Range<float>& );
	inline void				setDepthTest(DepthMode::TestFunction  );
	inline void				setDepthWrite(bool );

	// Stencil Buffer							
	inline void				setStencilAction(StencilMode::Action ,StencilMode::Action ,StencilMode::Action );
	inline void				setStencilFunction(StencilMode::TestFunction , unsigned int ref, unsigned int mask);
	inline void				setStencilWriteMask(unsigned int = 0xffffffff);

	// Blending
	inline void				setBlendFunction(BlendMode::BlendFunction ,BlendMode::BlendFunction );
	inline void				setBlendEquation(BlendMode::Equation );
	inline void				setBlendColour(const Colour<float>& );

	// Colour/Alpha writing
	inline void				setColourMask(bool r, bool g, bool b, bool a);
	inline void				setColourMask(bool r);

	// Clearing stuff
	inline void				setClearColour(const Colour<float>& );
	inline Colour<float>	getClearColour() const;
	inline void				setClearAccum(const Colour<float>& );
	inline Colour<float>	getClearAccum() const;
	inline void				setClearDepth(float );
	inline float			getClearDepth() const;
	inline void				setClearStencil(unsigned int );
	inline unsigned int		getClearStencil() const;
	inline void				clear(unsigned int c);

	// Viewport
	inline void				setViewport(const Rectangle<float>& );
	inline Rectangle<float> getViewport() const;

	// Clipping / Scissoring
	inline void				setClipRectangle(const Rectangle<int>& );
	inline Rectangle<int>	getClipRectangle() const;
	inline void				setClipping(bool );
	inline bool				getClipping() const;

	// Culling
	inline void				setCulling(CullMode );
	inline CullMode			getCulling() const;

	inline void				setPolygonOffset(float ,float );

	// Image copying
	void					copyTo(Texture1D* ,unsigned int level, const Point<int>& ,const Rectangle<int>& ) const;
	void					copyTo(Texture2D* ,unsigned int level, const Point<int>& ,const Rectangle<int>& ) const;
	void					copyTo(CubeTexture* ,CubeTexture::Face ,unsigned int level, const Point<int>& ,const Rectangle<int>& ) const;
	void					copyTo(Image* img, const Point<int>& ,const Rectangle<int>& ) const;
	void					copyFrom(const Image* img, const Point<int>& ,const Rectangle<int>& );

	// Defered rendering for transparencies
	inline void				defer(const Mesh* ,const Matrix<float,4>& );
	inline void				renderDefered();

	inline void				renderVertices(Mesh::Primitive p, unsigned int start, unsigned int size);
	inline void				renderVertices(Mesh::Primitive p, unsigned int size);
	inline void				renderIndices(Mesh::Primitive p, unsigned int start, unsigned int size);
	inline void				renderIndices(Mesh::Primitive p, unsigned int size);
	
#if defined(__FORWINDOWS__)
	inline HDC				getDeviceHandle() const;
	inline HWND				getWindowHandle() const;
#endif

private:
	static Context*			searchCurrent();
	static Context*			createCurrent();

private:
	class Startup {
	public:
		Startup();
	};
	friend Startup;

	static Startup			g_Startup;
	static Chain<Context>	g_Contexts;

#if defined(__NATIVETLS__)
	static __declspec(thread) Context*	g_Current;
#else
	static Thread::Storage	g_Current;
#endif

	FrameBuffer				m_DefaultFrameBuffer;
	const FrameBuffer*		m_FrameBuffer;
	unsigned int			m_Attributes;
	PlainShader				m_DefaultShader;
	const Shader*			m_Shader;
	ColourMaterial			m_DefaultMaterial;
	const Material*			m_Material;
	StencilMode				m_DefaultStencilMode;
	const StencilMode*		m_StencilMode;
	BlendMode				m_DefaultBlendMode;
	const BlendMode*		m_BlendMode;
	DepthMode				m_DefaultDepthMode;
	const DepthMode*		m_DepthMode;
	PolygonMode				m_DefaultPolygonMode;
	const PolygonMode*		m_PolygonMode;

	MatrixStack				m_MatrixStack;
	LightStack				m_LightStack;
	Capabilities			m_Capabilities;
	DeferQueue				m_DeferQueue;

	// Texture data
	const Texture*			m_Texture[Texture::MaxTextures];
	unsigned int			m_OglTextureTarget[Texture::MaxTextures];

	// This is a dirty hack, because Index buffer are not part of the internal
	// OpenGL state, except for IBOs.
	void*					m_Indices;
	unsigned int			m_IndexType;
	unsigned int			m_IndexSize;
	
	// Size of current framebuffer
	Dimension<int>			m_Size;

#if defined(__FORWINDOWS__)
	HDC						m_DeviceHandle;
	HWND					m_WindowHandle;
	HGLRC					m_OglContext;
#endif
};

}; };


#endif
