/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GRAPHICS3D_CONTEXT_INL
#define __GRAPHICS3D_CONTEXT_INL


namespace magnum { namespace graphics3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Context* Context::getCurrent() {
#if defined(__NATIVETLS__)
	Context* c = g_Current;
#else
	Context* c = (Context *)Thread::getTLS(g_Current);
#endif

	if (c == NULL)
		return searchCurrent();
	ASSERT(c->m_OglContext == wglGetCurrentContext());

	return c;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const FrameBuffer* Context::getFrameBuffer() const {
	return m_FrameBuffer;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const Shader* Context::getShader() const {
	return m_Shader;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const Material* Context::getMaterial() const {
	return m_Material;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const BlendMode*	Context::getBlendMode() const {
	return m_BlendMode;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const PolygonMode* Context::getPolygonMode() const {
	return m_PolygonMode;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const DepthMode*	Context::getDepthMode() const {
	return m_DepthMode;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const StencilMode* Context::getStencilMode() const {
	return m_StencilMode;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const Texture* Context::getTexture(unsigned int i) const {
	ASSERT(i < Texture::MaxTextures);
	return m_Texture[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline unsigned int Context::getTextureTarget(unsigned int i) const {
	ASSERT(i < Texture::MaxTextures);
	return m_OglTextureTarget[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline LightStack* Context::getLightStack() {
	return &m_LightStack;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const LightStack* Context::getLightStack() const {
	return &m_LightStack;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline MatrixStack* Context::getMatrixStack() {
	return &m_MatrixStack;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const MatrixStack* Context::getMatrixStack() const {
	return &m_MatrixStack;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Capabilities* Context::getCapabilities() {
	return &m_Capabilities;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const Capabilities* Context::getCapabilities() const {
	return &m_Capabilities;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Context::flush() {
	renderDefered();
	glFlush();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Context::swap() {
	renderDefered();
	SwapBuffers(m_DeviceHandle);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const Dimension<int>& Context::getSize() const {
	return m_Size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Context::setColourMask(bool r, bool g, bool b, bool a) {
	glColorMask(r,g,b,a);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Context::setColourMask(bool r) {
	glColorMask(r,r,r,r);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Context::setViewport(const Rectangle<float>& r) {
	glViewport((int)r.getX(), m_Size.getHeight() - ((int)r.getY() + (int)r.getHeight()), (int)r.getWidth(), (int)r.getHeight());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Rectangle<float> Context::getViewport() const {
	GLfloat f[4];
	glGetFloatv(GL_VIEWPORT, f);
	return Rectangle<float>(f[0], (float)m_Size.getHeight() - (f[1] + f[3]), f[2], f[3]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::setClipRectangle(const Rectangle<int>& r) {
	glScissor(r.getX(), m_Size.getHeight() - (r.getY() + r.getHeight()), r.getWidth(), r.getHeight());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Rectangle<int> Context::getClipRectangle() const {
	GLint f[4];
	glGetIntegerv(GL_SCISSOR_BOX, f);
	return Rectangle<int>(f[0], m_Size.getHeight() - (f[1] + f[3]), f[2], f[3]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::setClipping(bool s) {
	if (s)
		glEnable(GL_SCISSOR_TEST);
	else
		glDisable(GL_SCISSOR_TEST);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool Context::getClipping() const {
	GLboolean b;
	glGetBooleanv(GL_SCISSOR_TEST, &b);
	return (b == GL_TRUE);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Context::setCulling(CullMode m) {
	if (m == CullFront) {
		glEnable(GL_CULL_FACE);
		glCullFace(GL_FRONT);
	} else if (m == CullBack) {
		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);
	} else {
		glDisable(GL_CULL_FACE);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Context::CullMode Context::getCulling() const {
	if (glIsEnabled(GL_CULL_FACE) == GL_FALSE)
		return CullNone;
	int m;
	glGetIntegerv(GL_CULL_FACE_MODE, &m);
	if (m == GL_FRONT)
		return CullFront;
	else
		return CullBack;
}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Context::setClearColour(const Colour<float>& c) {
	glClearColor(c.R, c.G, c.B, c.A);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Colour<float> Context::getClearColour() const {
	GLfloat color[4];
	glGetFloatv(GL_COLOR_CLEAR_VALUE, color);
	return Colour<float>(color[0], color[1], color[2], color[3]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Context::setClearAccum(const Colour<float>& c) {
	glClearAccum(c.R, c.G, c.B, c.A);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Colour<float> Context::getClearAccum() const {
	GLfloat color[4];
	glGetFloatv(GL_ACCUM_CLEAR_VALUE, color);
	return Colour<float>(color[0], color[1], color[2], color[3]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Context::setClearDepth(float d) {
	glClearDepth(d);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline float Context::getClearDepth() const {
	GLfloat val;
	glGetFloatv(GL_DEPTH_CLEAR_VALUE, &val);
	return val;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Context::setClearStencil(unsigned int s) {
	glClearStencil(s);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline unsigned int Context::getClearStencil() const {
	GLint val;
	glGetIntegerv(GL_STENCIL_CLEAR_VALUE, &val);
	return val;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Context::clear(unsigned int c) {
	glClear(c);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Context::setPolygonOffset(float m,float r) {
	ASSERT(glPolygonOffsetEXT);
	if ((m == 0) && (r == 0.0f)) {
		glDisable(GL_POLYGON_OFFSET_FILL);
	} else {
		glPolygonOffsetEXT(m, r);
		glEnable(GL_POLYGON_OFFSET_FILL);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::defer(const Mesh* m, const Matrix<float,4>& mtx) {
	m_DeferQueue.insert(m,mtx);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::renderDefered() {
	m_DeferQueue.render();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::renderVertices(Mesh::Primitive p, unsigned int start, unsigned int size) {
	glDrawArrays(p, start, size);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::renderVertices(Mesh::Primitive p, unsigned int size) {
	glDrawArrays(p, 0, size);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::renderIndices(Mesh::Primitive p, unsigned int start, unsigned int size) {
	glDrawElements(p, size, m_IndexType, (byte *)m_Indices + start*m_IndexSize);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::renderIndices(Mesh::Primitive p, unsigned int size) {
	glDrawElements(p, size, m_IndexType, (byte *)m_Indices);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::setDepthRange(float min, float max) {
	m_DepthMode = NULL;
	glDepthRange(min, max);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::setDepthRange(const Range<float>& r) {
	m_DepthMode = NULL;
	glDepthRange(r.getStart(), r.getEnd());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::setDepthTest(DepthMode::TestFunction t) {
	m_DepthMode = NULL;
	if (t == DepthMode::Disable) {
		glDisable(GL_DEPTH_TEST);
	} else {
		glEnable(GL_DEPTH_TEST);
		glDepthFunc(t);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::setDepthWrite(bool b) {
	m_DepthMode = NULL;
	glDepthMask(b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::setStencilAction(StencilMode::Action fail,StencilMode::Action zfail,StencilMode::Action pass) {
	m_StencilMode = NULL;
	glStencilOp(fail, zfail, pass);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::setStencilFunction(StencilMode::TestFunction t,unsigned int ref, unsigned int mask) {
	m_StencilMode = NULL;
	if (t == StencilMode::Disable) {
		glDisable(GL_STENCIL_TEST);
	} else {
		glEnable(GL_STENCIL_TEST);
		glStencilFunc(t, ref, mask);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::setStencilWriteMask(unsigned int mask) {
	m_StencilMode = NULL;
	glStencilMask(mask);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::setBlendFunction(BlendMode::BlendFunction src, BlendMode::BlendFunction dst) {
	m_BlendMode = NULL;
	if ((src == BlendMode::One) && (dst == BlendMode::Zero)) {
		glDisable(GL_BLEND);
	} else {
		glEnable(GL_BLEND);
		glBlendFunc(src ,dst);
	}
}
						

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Context::setBlendEquation(BlendMode::Equation e) {
	ASSERT(glBlendEquationEXT != NULL);
	m_BlendMode = NULL;
	glBlendEquationEXT(e);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	Context::setBlendColour(const Colour<float>& col) {
	ASSERT(glBlendColorEXT != NULL);
	m_BlendMode = NULL;
	glBlendColorEXT(col.R, col.G, col.B, col.A);
}


#if defined(__FORWINDOWS__)
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline HWND Context::getWindowHandle() const {
	return m_WindowHandle;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline HDC Context::getDeviceHandle() const {
	return m_DeviceHandle;
}
#endif

}; };


#endif
