/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GRAPHICS3D_MATRIXSTACK_INL
#define __GRAPHICS3D_MATRIXSTACK_INL

namespace magnum {namespace graphics3d {


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline MatrixStack::MatrixStack() {
	m_Mode = (Mode)0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::setProjection(const Matrix<float,4>& mat) {
	if (m_Mode != Projection) {
		m_Mode = Projection;
		glMatrixMode(GL_PROJECTION);
	}
#if __FOROPENGL__ >= 13
	glLoadTransposeMatrixf(&mat[0][0]);
#else
	if (glLoadTransposeMatrixf)
		glLoadTransposeMatrixf(&mat[0][0]);
	else
		glLoadMatrixf(&mat.transpose()[0][0]);
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Matrix4<float> MatrixStack::getProjection(void) const {
	Matrix4<float> mat;
#if __FOROPENGL__ >= 13
	glGetFloatv(GL_TRANSPOSE_PROJECTION_MATRIX, &mat[0][0]);
	return mat;
#else
	if (glLoadTransposeMatrixf) {
		glGetFloatv(GL_TRANSPOSE_PROJECTION_MATRIX, &mat[0][0]);
		return mat;
	} else {
		glGetFloatv(GL_PROJECTION_MATRIX, &mat[0][0]);
		return mat.transpose();
	}
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::identityProjection() {
	if (m_Mode != Projection) {
		m_Mode = Projection;
		glMatrixMode(GL_PROJECTION);
	}
	glLoadIdentity();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::orthoProjection(float left, float top, float right, float bottom, float znear, float zfar) {
	if (m_Mode != Projection) {
		m_Mode = Projection;
		glMatrixMode(GL_PROJECTION);
	}
	glOrtho(left, right, bottom, top, znear, zfar);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::orthoProjection(const Rectangle<float>& rect, float znear, float zfar) {
	if (m_Mode != Projection) {
		m_Mode = Projection;
		glMatrixMode(GL_PROJECTION);
	}
	glOrtho(rect.getX(), rect.getX() + rect.getWidth(), rect.getY() + rect.getHeight(), rect.getY(), znear, zfar);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::pushProjection() {
	if (m_Mode != Projection) {
		m_Mode = Projection;
		glMatrixMode(GL_PROJECTION);
	}
	glPushMatrix();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::popProjection() {
	if (m_Mode != Projection) {
		m_Mode = Projection;
		glMatrixMode(GL_PROJECTION);
	}
	glPopMatrix();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::setTexture(const Matrix<float,4>& mat) {
	selectMode(Texture);
#if __FOROPENGL__ >= 13
	glLoadTransposeMatrixf(&mat[0][0]);
#else
	if (glLoadTransposeMatrixf)
		glLoadTransposeMatrixf(&mat[0][0]);
	else
		glLoadMatrixf(&mat.transpose()[0][0]);
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::identityTexture() {
	selectMode(Texture);
	glLoadIdentity();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::pushTexture() {
	selectMode(Texture);
	glPushMatrix();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::popTexture() {
	selectMode(Texture);
	glPopMatrix();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::setModel(const Matrix<float,4>& mat) {
	if (m_Mode != ModelView) {
		m_Mode = ModelView;
		glMatrixMode(GL_MODELVIEW);
	}
#if __FOROPENGL__ >= 13
	glLoadTransposeMatrixf(&mat[0][0]);
#else
	if (glLoadTransposeMatrixf)
		glLoadTransposeMatrixf(&mat[0][0]);
	else
		glLoadMatrixf(&mat.transpose()[0][0]);
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Matrix4<float> MatrixStack::getModel(void) const {
	Matrix4<float> mat;
#if __FOROPENGL__ >= 13
	glGetFloatv(GL_TRANSPOSE_MODELVIEW_MATRIX, &mat[0][0]);
	return mat;
#else
	if (glLoadTransposeMatrixf) {
		glGetFloatv(GL_TRANSPOSE_MODELVIEW_MATRIX, &mat[0][0]);
		return mat;
	} else {
		glGetFloatv(GL_MODELVIEW_MATRIX, &mat[0][0]);
		return mat.transpose();
	}
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::multModel(const Matrix<float,4>& mat) {
	if (m_Mode != ModelView) {
		m_Mode = ModelView;
		glMatrixMode(GL_MODELVIEW);
	}
#if __FOROPENGL__ >= 13
	glMultTransposeMatrixf(&mat[0][0]);
#else
	if (glMultTransposeMatrixf)
		glMultTransposeMatrixf(&mat[0][0]);
	else
		glMultMatrixf(&mat.transpose()[0][0]);
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::translateModel(const Vector3<float>& v) {
	if (m_Mode != ModelView) {
		m_Mode = ModelView;
		glMatrixMode(GL_MODELVIEW);
	}
	glTranslatef(v.x, v.y, v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::scaleModel(const Vector3<float>& v) {
	if (m_Mode != ModelView) {
		m_Mode = ModelView;
		glMatrixMode(GL_MODELVIEW);
	}
	glScalef(v.x, v.y, v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::identityModel() {
	if (m_Mode != ModelView) {
		m_Mode = ModelView;
		glMatrixMode(GL_MODELVIEW);
	}
	glLoadIdentity();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::pushModel() {
	if (m_Mode != ModelView) {
		m_Mode = ModelView;
		glMatrixMode(GL_MODELVIEW);
	}
	glPushMatrix();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::popModel() {
	if (m_Mode != ModelView) {
		m_Mode = ModelView;
		glMatrixMode(GL_MODELVIEW);
	}
	glPopMatrix();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	MatrixStack::selectMode(Mode m) {
	if (m_Mode != m)
		forceSelectMode(m);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline MatrixStack::Mode MatrixStack::getMode() const {
	glGetIntegerv(GL_MATRIX_MODE, (int *)&m_Mode);
	return m_Mode;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	MatrixStack::set(const Matrix<float,4>& mat) {
#if __FOROPENGL__ >= 13
	glLoadTransposeMatrixf(&mat[0][0]);
#else
	if (glLoadTransposeMatrixf)
		glLoadTransposeMatrixf(&mat[0][0]);
	else
		glLoadMatrixf(&mat.transpose()[0][0]);
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Matrix4<float> MatrixStack::get(void) const {
	Matrix4<float> mat;
#if __FOROPENGL__ >= 13
	glGetFloatv(GL_TRANSPOSE_MODELVIEW_MATRIX, &mat[0][0]);
	return mat;
#else
	if (glLoadTransposeMatrixf) {
		glGetFloatv(GL_TRANSPOSE_MODELVIEW_MATRIX, &mat[0][0]);
		return mat;
	} else {
		glGetFloatv(GL_MODELVIEW_MATRIX, &mat[0][0]);
		return mat.transpose();
	}
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	MatrixStack::mult(const Matrix<float,4>& mat) {
#if __FOROPENGL__ >= 13
	glMultTransposeMatrixf(&mat[0][0]);
#else
	if (glMultTransposeMatrixf)
		glMultTransposeMatrixf(&mat[0][0]);
	else
		glMultMatrixf(&mat.transpose()[0][0]);
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	MatrixStack::push() {
	glPushMatrix();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	MatrixStack::pop() {
	glPopMatrix();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::translate(const Vector3<float>& v) {
	glTranslatef(v.x, v.y, v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::scale(const Vector3<float>& v) {
	glScalef(v.x, v.y, v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::identity() {
	glLoadIdentity();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::rotate(float angle, const Vector3<float>& vec) {
	glRotatef(angle*180.0f/(float)M_PI, vec.x, vec.y, vec.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::ortho(float left, float top, float right, float bottom, float znear, float zfar) {
	glOrtho(left, right, bottom, top, znear, zfar);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void MatrixStack::ortho(const Rectangle<float>& rect, float znear, float zfar) {
	glOrtho(rect.getX(), rect.getX() + rect.getWidth(), rect.getY() + rect.getHeight(), rect.getY(), znear, zfar);
}

}; };

#endif
