/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GRAPHICS3D_MESHSTRIPPER_INL
#define __GRAPHICS3D_MESHSTRIPPER_INL

namespace magnum {namespace graphics3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void TriangleStripper<T>::clear() {
	m_Mesh.setSize(0);
	m_Mesh.clear();
	m_Edges.clear();
	m_Strips.clear();
	m_MarkedTriangles.clear();
	m_RemovedTriangles.clear();
	m_CacheSize = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void TriangleStripper<T>::insert(const T* idx,unsigned int size) {
	unsigned int numTris = size/3;
	unsigned int meshStart = m_Mesh.getSize();
	unsigned int edgeStart = m_Edges.getSize();

	m_Mesh.setSize(meshStart + numTris);
	m_Edges.setSize(edgeStart + numTris*3);
	Array<TriangleEdge>::Iterator edge = m_Edges.at(edgeStart);
	Graph<Triangle, Edge>::Iterator tri = m_Mesh.at(meshStart);

	for (unsigned int i = 0; i < numTris; i++) {
		const T* iptr = &idx[i*3]; 
		(*tri)->v0 = iptr[0];
		(*tri)->v1 = iptr[1];
		(*tri)->v2 = iptr[2];
		tri++;

		*edge++ = TriangleEdge(Edge(iptr[0], iptr[1]),i);
		*edge++ = TriangleEdge(Edge(iptr[1], iptr[2]),i);
		*edge++ = TriangleEdge(Edge(iptr[2], iptr[0]),i);
	}

}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void TriangleStripper<T>::connect(const typename TriangleStripper<T>::TriangleEdge& edge) {
	Array<TriangleEdge>::Iterator iter;
	Edge e(edge.getKey().second, edge.getKey().first);

	iter = findfirst(m_Edges.begin(), m_Edges.end(), e);
	while ((iter != m_Edges.end()) && (iter->getKey() == e)) {
		m_Mesh.insert(edge.getItem(), iter->getItem(), edge.getKey());		
		iter++;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void TriangleStripper<T>::finish() {
	unsigned int numTris = m_Mesh.getSize();

	// Sort edges
	sort(m_Edges.begin(), m_Edges.end());

	// build connectivity graph
	unsigned int i;
	for (i = 0; i < numTris; i++) {
		Graph<Triangle, Edge>::Node& node = m_Mesh[i];
		connect(TriangleEdge(Edge(node->v0, node->v1), i));
		connect(TriangleEdge(Edge(node->v1, node->v2), i));
		connect(TriangleEdge(Edge(node->v2, node->v0), i));
	}

	// build triangle degreees
	m_Degrees.setCapacity(numTris);
	PriorityHeap<int>::Iterator phiter;
	for (i = 0; i < numTris; i++) {
		phiter = m_Degrees.insert(i, -(int)m_Mesh[i].getSize());
		m_Mesh[i]->m_Degree = phiter;
	}

	// init cache

	// beginn stripping
	m_StripSize = 0;
	while(!m_Degrees.isEmpty()) {
		Array<T>* strip = findStrip();
		m_StripSize += strip->getSize() + 4;
		m_Strips.insert(strip);		
		strip->release();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Array<T>* TriangleStripper<T>::strip() {
	Array<T>* strip = new Array<T>();

	// merge strips
	strip->setSize(m_StripSize);
	unsigned int index = 0;
	unsigned int i,j;
	for (i = 0; i < m_Strips.getSize(); i++) {
		Array<T>* curstrip = m_Strips[i];

		if (index > 0) {
			(*strip)[index++] = (*strip)[index - 1];
			(*strip)[index++] = (*curstrip)[0];
			if (index & 1)
				(*strip)[index++] = (*curstrip)[0];
		}

		for (j = 0; j < curstrip->getSize(); j++) {
			(*strip)[index++] = (*curstrip)[j];
		}
	}
	strip->setSize(index);

	return strip;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Array<T>* TriangleStripper<T>::findStrip() {
	ASSERT(m_Degrees.getSize() > 0)
	int neighbours = m_Degrees.top().priority();
	
	// Check for singleton
	if (neighbours == 0) {
		unsigned int triangle = m_Degrees.peek();

		// we cannot do anything in that case
		// except for returning a tiny singleton strip
		return buildSingleton(triangle);
	}

	const Graph<Triangle, Edge>::Edge* bestedge;
	unsigned int bestfirst;
	int bestsize = 0;

	// Only perform aggressive search strategies in the case of
	// 1 or 2 neighbour. For triangles with more than 2 neighbour
	// simply chose the next one in the heap.
	if (neighbours < -1) {
		unsigned int triangle = m_Degrees.peek();
		int stripsize;

		// Check all possible neighbours
		Graph<Triangle, Edge>::Node::Iterator iter, end;
		iter = m_Mesh[triangle].begin();
		end = m_Mesh[triangle].end();

		// Check for singlteon
		ASSERT (iter != end);

		// If no singleton check all neighbours
		while(iter != end) {
			stripsize = getStripValue(triangle, *iter);

			// Check for new size record
			if (stripsize > bestsize) {
				bestfirst = triangle;
				bestedge = iter;
				bestsize = stripsize;
			}

			++iter;
		}
		
		// Now build the strip!
		ASSERT(bestsize > 1);
		return buildStrip(bestfirst, *bestedge);
	}

	// Search longest strip 
	PriorityHeap<int>::Iterator hend = m_Degrees.end();
	m_RemovedTriangles.setSize(0);
	
	while(true) {
		PriorityHeap<int>::Iterator htop = m_Degrees.top();
		if (htop == hend)
			break;
		if (htop.priority() != neighbours)
			break;
			
		unsigned int priority = htop.priority();
		unsigned int triangle = htop.item();
		int stripsize;

		m_Degrees.remove();
		m_RemovedTriangles.insert(Pair<int,int>(triangle,priority));

		// Check all possible neighbours
		Graph<Triangle, Edge>::Node::Iterator iter, end;
		iter = m_Mesh[triangle].begin();
		end = m_Mesh[triangle].end();

		// Check for singlteon
		ASSERT (iter != end);

		// If no singleton check all neighbours
		while(iter != end) {
			stripsize = getStripValue(triangle, *iter);

			// Check for new size record
			if (stripsize > bestsize) {
				bestfirst = triangle;
				bestedge = iter;
				bestsize = stripsize;
			}

			++iter;
		}
	}
	
	// Reinsert all triangles
	Array<Pair<int,int> >::Iterator iter;
	PriorityHeap<int>::Iterator phiter;
	for (iter = m_RemovedTriangles.begin(); iter != m_RemovedTriangles.end(); ++iter) {
		phiter = m_Degrees.insert(iter->first, iter->second);
		m_Mesh[iter->first]->m_Degree = phiter;
	}

	// Now build the strip!
	ASSERT(bestsize > 1);
	return buildStrip(bestfirst, *bestedge);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
int TriangleStripper<T>::getStripValue(unsigned int triangle, const typename Graph<Triangle, Edge>::Edge& ge) {
	T idx[3];
	int size = 1;
	int penalty = 0;
	const Graph<Triangle, Edge>::Edge* edge = &ge;

	// Insert first triangle and mark it as used
	m_MarkedTriangles.setSize(0);
	m_MarkedTriangles.insert(triangle);
	m_Mesh[triangle].mark();

	// Get triangle edge that connects the two triangles
	Edge triedge = **edge;
	idx[0] = 0;
	idx[1] = triedge.first;
	idx[2] = triedge.second;

	while(true) {
		unsigned int oldtriangle = triangle;

		// get next connected triangle
		triangle = edge->getEnd();
		m_MarkedTriangles.insert(triangle);
		m_Mesh[triangle].mark();

		// increment size
		size++;

		// Get number of singleton sgenerated
		penalty += 2*countSingletons(triangle);

		// get next admissible edge (there is only one possibility)
		idx[0] = idx[1];
		idx[1] = idx[2];
		idx[2] = getConnection(triangle, triedge);
		if (size & 1) {
			triedge.first = idx[1];
			triedge.second = idx[2];
		} else {
			triedge.first = idx[2];
			triedge.second = idx[1];
		}

		// check if current triangle has such an edge
		Graph<Triangle, Edge>::Node::Iterator iter, end, alt;
		iter = m_Mesh[triangle].begin();
		alt = end = m_Mesh[triangle].end();
		for( ;iter != end; ++iter) {
			if (m_Mesh[iter->getEnd()].isMarked())
				continue;
			if (**iter == triedge)
				break;
			alt = iter;
		}

		// did we find a triangle?
		if (iter != end) {
			edge = iter;
		} else if (alt != end) {
			edge = alt;
			T a = idx[0];
			idx[0] = idx[1];
			idx[1] = idx[2];
			idx[2] = a;
			penalty++;
			if (size & 1) {
				triedge.first = idx[1];
				triedge.second = idx[2];
			} else {
				triedge.first = idx[2];
				triedge.second = idx[1];
			}
		} else {
			break;
		}

	}

	for (unsigned int i = 0; i < m_MarkedTriangles.getSize(); i++)
		m_Mesh[m_MarkedTriangles[i]].unmark();

	return 2*size - penalty;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Array<T>* TriangleStripper<T>::buildStrip(unsigned int triangle, const typename Graph<Triangle, Edge>::Edge& ge) {
	T idx[3];
	Array<T>* strip = new Array<T>();
	const Graph<Triangle, Edge>::Edge* edge = &ge;

	// Get triangle edge that connects the two triangles
	Edge triedge = **edge;

	// insert first three indices
	if (Edge(m_Mesh[triangle]->v0, m_Mesh[triangle]->v1) == triedge) {
		strip->insert(m_Mesh[triangle]->v2);
		strip->insert(m_Mesh[triangle]->v0);
		strip->insert(m_Mesh[triangle]->v1);
		idx[0] = m_Mesh[triangle]->v2;
		idx[1] = m_Mesh[triangle]->v0;
		idx[2] = m_Mesh[triangle]->v1;
	} else if (Edge(m_Mesh[triangle]->v1, m_Mesh[triangle]->v2) == triedge) {
		strip->insert(m_Mesh[triangle]->v0);
		strip->insert(m_Mesh[triangle]->v1);
		strip->insert(m_Mesh[triangle]->v2);
		idx[0] = m_Mesh[triangle]->v0;
		idx[1] = m_Mesh[triangle]->v1;
		idx[2] = m_Mesh[triangle]->v2;
	} else if (Edge(m_Mesh[triangle]->v2, m_Mesh[triangle]->v0) == triedge) {
		strip->insert(m_Mesh[triangle]->v1);
		strip->insert(m_Mesh[triangle]->v2);
		strip->insert(m_Mesh[triangle]->v0);
		idx[0] = m_Mesh[triangle]->v1;
		idx[1] = m_Mesh[triangle]->v2;
		idx[2] = m_Mesh[triangle]->v0;
	}

	bool odd = true;
	while(true) {
		// get next connected triangle
		triangle = edge->getEnd();

		// Remove previous triangle
		removeTriangle(edge->getStart());

		odd = !odd;

		// get next admissible edge (there is only one possibility)
		idx[0] = idx[1];
		idx[1] = idx[2];
		idx[2] = getConnection(triangle, triedge);
		if (odd) {
			triedge.first = idx[1];
			triedge.second = idx[2];
		} else {
			triedge.first = idx[2];
			triedge.second = idx[1];
		}

		// Insert index into strip
		strip->insert(idx[2]);

		// check if current triangle has such an edge
		Graph<Triangle, Edge>::Node::Iterator iter, end, alt;
		iter = m_Mesh[triangle].begin();
		alt = end = m_Mesh[triangle].end();
		for( ;iter != end; iter++) {
			ASSERT(!m_Mesh[iter->getEnd()].isMarked());
			if (**iter == triedge)
				break;
			alt = iter;
		}

		// did we find a triangle?
		if (iter != end) {
			edge = iter;
		} else if (alt != end) {
			edge = alt;
			strip->insert(idx[2]);
			strip->insert(idx[0]);
			T a = idx[0];
			idx[0] = idx[1];
			idx[1] = idx[2];
			idx[2] = a;
			if (odd) {
				triedge.first = idx[1];
				triedge.second = idx[2];
			} else {
				triedge.first = idx[2];
				triedge.second = idx[1];
			}
		} else {
			break;
		}
	}

	// Remove last triangle
	removeTriangle(triangle);

	return strip;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
unsigned int TriangleStripper<T>::countSingletons(unsigned int triangle) const {
	int count = 0;
	Graph<Triangle, Edge>::Node::Iterator iter, end;
	iter = m_Mesh[triangle].begin();
	end = m_Mesh[triangle].end();
	for( ;iter != end; ++iter) {
		if (m_Mesh[iter->getEnd()].isMarked())
			continue;
		if (m_Mesh[iter->getEnd()].getSize() == 1)
			count++;
	}

	return count;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
T TriangleStripper<T>::getConnection(unsigned int tri, typename const TriangleStripper<T>::Edge& edge) const {
	// get next admissible edge (there is only one possibility)
	if (Edge(m_Mesh[tri]->v1, m_Mesh[tri]->v0) == edge) {
		return m_Mesh[tri]->v2;
	} else if (Edge(m_Mesh[tri]->v2, m_Mesh[tri]->v1) == edge) {
		return m_Mesh[tri]->v0;
	} else if (Edge(m_Mesh[tri]->v0, m_Mesh[tri]->v2) == edge) {
		return m_Mesh[tri]->v1;
	} else {
		T v0 = m_Mesh[tri]->v0;
		T v1 = m_Mesh[tri]->v1;
		T v2 = m_Mesh[tri]->v2;
		ASSERT(false);
		return 0;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Array<T>* TriangleStripper<T>::buildSingleton(unsigned int triangle) {
	Array<T>* strip = new Array<T>();
	strip->insert(m_Mesh[triangle]->v0);
	strip->insert(m_Mesh[triangle]->v1);
	strip->insert(m_Mesh[triangle]->v2);
	removeTriangle(triangle);
	return strip;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline void TriangleStripper<T>::removeTriangle(unsigned int i) {
	Graph<Triangle, Edge>::Node* node = &m_Mesh[i];
	Graph<Triangle, Edge>::Node::Iterator iter;
	
	while ((iter = node->begin()) != node->end()) {
		unsigned int end = iter->getEnd();
		unsigned int start = iter->getStart();
		m_Mesh.remove(end, start);
		m_Mesh.remove(start, end);
		ASSERT(!m_Mesh[end].isMarked());
		m_Degrees.change(m_Mesh[end]->m_Degree, -(int)m_Mesh[end].getSize());
	}
	
	m_Degrees.remove((*node)->m_Degree);
	node->mark();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline void TriangleStripper<T>::setCacheSize(unsigned int s) {
	m_CacheSize = s;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline unsigned int TriangleStripper<T>::getCacheSize() const {
	return m_CacheSize;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool TriangleStripper<T>::findCache(unsigned int i) {
	return m_Cache.contains(i);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline void TriangleStripper<T>::clearCache() {
	m_Cache.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline void TriangleStripper<T>::insertCache(unsigned int i) {
	m_Cache.remove(0);
	m_Cache.insert(i);
}


}; };

#endif