/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GRAPHICS3D_OPTIMISER_H
#define __GRAPHICS3D_OPTIMISER_H

#include "object.pointer.h"
#include "graphics3d.mesh.h"
#include "graphics3d.texture.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.compiler.h"
#include "graphics3d.meshstripper.h"
#include "graphics3d.meshsplitter.h"
#include "container.array.h"
#include "container.tuple.h"

namespace magnum {namespace graphics3d {
using magnum::object::Pointer;
using magnum::container::Array;
using magnum::container::Tuple;

class Shader;
class Material;
class Mode;
class RenderStream;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
class Optimiser : public Object,private magnum::object::Noncopyable {
DeclareClass(Optimiser, Object);
public:
	enum Strategy {
		SortShader = 0x0001,
		SortMaterial = 0x0002,
		SortMode = 0x0004,
		SortBuffer = 0x0008,
		MergeBuffer = 0x0010
	};

public:
    Optimiser();
    ~Optimiser();

	inline void			setStrategy(Strategy );
	inline Strategy		getStrategy() const;
						
    inline void			preselectVertexBuffer(VertexBuffer* , unsigned int attribs = 0xffffffff);
    inline void			preselectIndexBuffer(IndexBuffer* );
    inline void			preselectMaterial(Material* );
    inline void			preselectTexture(Texture* ,unsigned int );
    inline void			preselectShader(Shader* );
    inline void			preselectStencilMode(StencilMode* );
    inline void			preselectBlendMode(BlendMode* );
    inline void			preselectDepthMode(DepthMode* );
    inline void			preselectPolygonMode(PolygonMode* );
    void				preselectAppearance(Appearance* );

    inline void			selectVertexBuffer(VertexBuffer* , unsigned int attribs = 0xffffffff);
    inline void			selectIndexBuffer(IndexBuffer* );
    inline void			selectMaterial(Material* );
    inline void			selectTexture(Texture* ,unsigned int );
    inline void			selectShader(Shader* );
    inline void			selectStencilMode(StencilMode* );
    inline void			selectBlendMode(BlendMode* );
    inline void			selectDepthMode(DepthMode* );
    inline void			selectPolygonMode(PolygonMode* );
    void				selectAppearance(Appearance* );

    inline void			deselectMaterial();
    inline void			deselectTexture(unsigned int );
    inline void			deselectShader();
    inline void			deselectStencilMode();
    inline void			deselectBlendMode();
    inline void			deselectDepthMode();
    inline void			deselectPolygonMode();
    void				deselectAppearance();

    void				renderMesh(Mesh* );
    void				renderVertices(Mesh::Primitive, unsigned int start, unsigned int size);
    void				renderIndices(Mesh::Primitive, unsigned int start, unsigned int size);
						
    void				clear();
	void				finish();

	MeshList			createMeshes(bool fullMats = true);
	RenderStream*		createStream();

private:
    struct Job {
        Pointer<Mesh>			m_Mesh;
        Pointer<Material>		m_Material;
        Pointer<Shader>			m_Shader;
		Pointer<Texture>		m_Textures[Texture::MaxTextures];
		Pointer<BlendMode>		m_BlendMode;
		Pointer<DepthMode>		m_DepthMode;
		Pointer<StencilMode>	m_StencilMode;
		Pointer<PolygonMode>	m_PolygonMode;
        Pointer<VertexBuffer>	m_VertexBuffers[Attributes::MaxAttributes];
        Pointer<IndexBuffer>	m_IndexBuffer;
        unsigned int			m_Start;
        unsigned int			m_Count;
        Mesh::Primitive			m_Primitive;
    };

private:
	static int			cmpJobs(const void* ,const void* );

    void				renderVertexMesh(VertexMesh* );
    void				renderIndexMesh(IndexMesh* );
    void				insertMesh(Mesh* );

	void				mergeJobs();
	void				mergeBuffers();
	void				splitBuffers();

	Tuple<VertexBuffer*,VertexBuffer::MaxBuffers>	collectVertexBuffers(const Pointer<VertexBuffer>* );
	unsigned int		collectAttributes(const Tuple<VertexBuffer*,Attributes::MaxAttributes>& );
	unsigned int		collectAttributes(const Tuple<VertexBuffer*,Attributes::MaxAttributes>& ,const VertexBuffer*);

private:
    Array<Job>			m_Jobs;

	Compiler			m_Compiler;
	MeshSplitter		m_Splitter;
	MeshStripper		m_Stripper;
						
	Strategy			m_Strategy;

	Texture*			m_PreselectedTextures[Texture::MaxTextures];						
    Material*			m_PreselectedMaterial;
    Shader*				m_PreselectedShader;
    StencilMode*		m_PreselectedStencilMode;
    BlendMode*			m_PreselectedBlendMode;
    DepthMode*			m_PreselectedDepthMode;
    PolygonMode*		m_PreselectedPolygonMode;
    VertexBuffer*		m_PreselectedVertexBuffers[Attributes::MaxAttributes];
    IndexBuffer*		m_PreselectedIndexBuffer;

	Texture*			m_CurrentTextures[Texture::MaxTextures];						
    Material*			m_CurrentMaterial;
    Shader*				m_CurrentShader;
    StencilMode*		m_CurrentStencilMode;
    BlendMode*			m_CurrentBlendMode;
    DepthMode*			m_CurrentDepthMode;
    PolygonMode*		m_CurrentPolygonMode;
    VertexBuffer*		m_CurrentVertexBuffers[Attributes::MaxAttributes];
    IndexBuffer*		m_CurrentIndexBuffer;
};

}; };

#endif
