/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GRAPHICS3D_TRANSLATOR_MDS_H
#define __GRAPHICS3D_TRANSLATOR_MDS_H

#include "io.translator.h"
#include "container.map.h"
#include "container.array.h"
#include "graphics.colour.h"
#include "graphics3d.translator.h"

namespace magnum { namespace graphics3d {
using magnum::object::Object;
using magnum::types::String;
using magnum::math::Matrix4;
using magnum::graphics::Colour;
using magnum::container::Map;
using magnum::container::Array;
using magnum::types::uint32;
using magnum::types::int32;
using magnum::types::uint16;
using magnum::types::int16;

class BoneModel;
class IndexMesh;
class Shader;
class WeightBuffer;
class Skeleton;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
class TranslatorMDS : public ModelTranslator {
DeclareClass(TranslatorMDS, ModelTranslator);
public:
	TranslatorMDS();

protected:
	Model*				onDecode(InputStream* ,size_t size = Autosize);
	size_t				onEncode(OutputStream* ,const Model* );
	bool				onIdentify(File* );

private:
	enum {
		Magic			= 0x5753444d,
	};
	
	struct Header {
		uint32  		magic;          // "MDSW"
		uint32  		version;        // 0x0004
		char    		name[64];       // Model name eg. "body.mds"
		float   		lodScale;       // LOD Scale
		float   		lodBias;        // LOD Bias
		uint32  		numFrames;      // Number of animation frames
		uint32  		numBones;       // Number of bones
		uint32  		offsetFrames;   // Offset of animation frames
		uint32  		offsetBones;    // Offset of bones (see MDSBone)
		uint32  		torsoParent;    // Index of torso parent bone
		uint32  		numSurfaces;    // Number of surfaces
		uint32  		offsetSurfaces; // Offset of surfaces
		uint32  		numTags;        // Number of tags
		uint32  		offsetTags;     // Offset of tags
		uint32  		offsetEnd;      // Offset of end of file
	};
	
	struct Bone {
		char   			name[64];    // Bone name
		int32 			parentIndex; // Bone parent index (-1 if root)
		float  			torsoWeight; // 0.0 to 1.0
		float  			parentDist;  // Distance from parent bone
		uint32 			flags;       // Bit 0 is set if bone is a tag
	};

	struct Tag {
		char   			name[64];    // Name of tag
		float  			torsoWeight; // 0.0, 0.3
		uint32 			boneIndex;   // Index of bone this tag is attached to
	};
	
	struct CompressedBoneFrame {
		int16  			angles[4];      // Defines the bone orientation
		int16	  		offsetAngles[2]; // Defines direction of the bone
	};

	struct Frame {
		Vector3<float>	bboxMin;      // Bounding box min
		Vector3<float>	bboxMax;      // Bounding box max
		Vector3<float>	localOrigin;  // Local origin
		float     		radius;       // Radius of bounding sphere
		Vector3<float>	parentOffset; // Offset of parent bone from origin
		//MDSCompressedBoneFrame  bones[numBones];
	};
	
	struct FrameExt : public Frame {
		Array<CompressedBoneFrame> bones;
	};

	struct Surface {
		uint32  		ident;           // Always 0x0008
		char    		name[64];        // Name of surface
		char    		shader[64];      // Name of shader
		uint32  		shaderIndex;     // Used in game only
		int32  			minLod;          // Minimum Lod level
		int32  			offsetHeader;    // Offset of this surface header
		uint32  		numVertices;     // Number of vertices
		uint32  		offsetVertices;  // Offset of vertices
		uint32  		numTriangles;    // Number of triangles
		uint32  		offsetTriangles;  // Offset of triangles
		uint32  		offsetCollapseMap; // Offset of the collapse map 
		uint32  		numBoneRefs;     // Number of bone references
		uint32  		offsetBoneRefs;  // Offset of bone refs
		uint32  		offsetEnd;       // Offset of the end of this
	};

	struct StaticVertex {
		Vector3<float>	normal;      // Vertex normal vector.
		Vector2<float> 	texCoords;	// Texture (s,t) co-ordinates
	};

	struct Vertex : StaticVertex {
		uint32    		numWeights;		// Number of bone weights affecting this vertex
		uint32    		fixedParent;	// Stay equidistant from this parent
		float     		fixedDist;		// Fixed distance from parent
		//MDSWeight		weights[numWeights]  // List of weights follows
	};
	
private:
	size_t				readTags(unsigned int );
	size_t				readBones(Skeleton* ,unsigned int );
	size_t				readSurfaces(BoneModel* ,WeightBuffer* ,unsigned int );
	size_t				readBoneRefs(unsigned int );	
	size_t				readFrames(unsigned int ,unsigned int );
	size_t				readCollapseMap(unsigned int );
	size_t				readTriangles(IndexMesh* ,unsigned int ,unsigned int );
	size_t				readVertices(WeightBuffer* ,unsigned int );
	
	void				createAnimations(BoneModel* );
	void				createVertices(BoneModel* );

private:
	InputStream*		m_In;
	List<FrameExt>		m_Frames;
	List<Surface>		m_Surfaces;
	List<Tag>			m_Tags;
	Array<StaticVertex>	m_Vertices;
};

	
}; };

#endif