/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GRAPHICS3D_VERTEXBUFFER_INL
#define __GRAPHICS3D_VERTEXBUFFER_INL

namespace magnum { namespace graphics3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void* VertexBuffer::lock(LockFlag f) {
	ASSERT((f >= ReadOnly) && (f <= Discard));
	ASSERT(!m_Locked);
	if (m_OglBuffer != 0) {
		glBindBufferARB(GL_ARRAY_BUFFER, m_OglBuffer);
		m_Buffer = glMapBufferARB(GL_ARRAY_BUFFER, g_OglAcces[f]);
		if (m_Buffer)
			m_Locked = true;
	} else {
		m_Locked = true;
	}
	return m_Buffer;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	VertexBuffer::unlock() {
	ASSERT(m_Locked);
	if (m_OglBuffer != 0) {
		glBindBufferARB(GL_ARRAY_BUFFER, m_OglBuffer);
		glUnmapBufferARB(GL_ARRAY_BUFFER);
		m_Buffer = NULL;
	}
	m_Locked = false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool VertexBuffer::isLocked() const {
	return m_Locked;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void* VertexBuffer::bindBuffer() const {
	if (m_OglBuffer) {
		glBindBufferARB(GL_ARRAY_BUFFER, m_OglBuffer);
		return NULL;
	} else {
		if (glBindBufferARB)
			glBindBufferARB(GL_ARRAY_BUFFER, NULL);
		return m_Buffer;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void VertexBuffer::selectVertices(unsigned int i) const {
	ASSERT(i < VertexFormat::MaxComponents);
	const VertexFormat::Component* cmp = &m_Format[i];
	glVertexPointer(cmp->m_OglSize, cmp->m_OglType, getStride(), (byte *)bindBuffer() + cmp->m_Offset);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void VertexBuffer::selectNormals(unsigned int i) const {
	ASSERT(i < VertexFormat::MaxComponents);
	const VertexFormat::Component* cmp = &m_Format[i];
	ASSERT(cmp->m_OglSize == 3);
	glNormalPointer(cmp->m_OglType, getStride(), (byte *)bindBuffer() + cmp->m_Offset);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	VertexBuffer::selectColours(unsigned int i) const {
	ASSERT(i < VertexFormat::MaxComponents);
	const VertexFormat::Component* cmp = &m_Format[i];
	ASSERT((cmp->m_OglSize == 3) || (cmp->m_OglSize == 4)); 
	glColorPointer(cmp->m_OglSize, cmp->m_OglType, getStride(), (byte *)bindBuffer() + cmp->m_Offset);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	VertexBuffer::selectEdges(unsigned int i) const {
	ASSERT(i < VertexFormat::MaxComponents);
	const VertexFormat::Component* cmp = &m_Format[i];
	ASSERT(cmp->m_OglSize == 1); 
	glEdgeFlagPointer(getStride(), (byte *)bindBuffer() + cmp->m_Offset);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	VertexBuffer::selectTexturCoords(unsigned int i, unsigned int stage) const {
	ASSERT(i < VertexFormat::MaxComponents);
	const VertexFormat::Component* cmp = &m_Format[i];
	ASSERT((cmp->m_OglSize >= 1) || (cmp->m_OglSize <= 4));
	glClientActiveTextureARB(GL_TEXTURE0_ARB + stage);
	glTexCoordPointer(cmp->m_OglSize, cmp->m_OglType, getStride(), (byte *)bindBuffer() + cmp->m_Offset);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	VertexBuffer::selectFogCoords(unsigned int i) const {
	ASSERT(i < VertexFormat::MaxComponents);
	const VertexFormat::Component* cmp = &m_Format[i];
	ASSERT(cmp->m_OglSize == 1);
	glFogCoordPointerEXT(cmp->m_OglType, getStride(), (byte *)bindBuffer() + cmp->m_Offset);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void VertexBuffer::selectVertices() const {
	ASSERT(m_Attributes & Attributes::Vertex);
	ASSERT((m_Mapping[Attributes::VertexBit].m_OglSize >= 3) 
		&& (m_Mapping[Attributes::VertexBit].m_OglSize <= 4)); 
	byte* base = (byte *)bindBuffer();
	glVertexPointer(m_Mapping[Attributes::VertexBit].m_OglSize, 
					m_Mapping[Attributes::VertexBit].m_OglType, 
					getStride(), 
					base + m_Mapping[Attributes::VertexBit].m_Offset);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void VertexBuffer::selectNormals() const {
	ASSERT(m_Attributes & Attributes::Normal);
	ASSERT(m_Mapping[Attributes::NormalBit].m_OglSize == 3);
	byte* base = (byte *)bindBuffer();
	glNormalPointer(m_Mapping[Attributes::NormalBit].m_OglType, 
					getStride(), 
					base + m_Mapping[Attributes::NormalBit].m_Offset);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void VertexBuffer::selectColours() const {
	ASSERT(m_Attributes & Attributes::Colour);
	ASSERT((m_Mapping[Attributes::ColourBit].m_OglSize == 3) 
		|| (m_Mapping[Attributes::ColourBit].m_OglSize == 4)); 
	byte* base = (byte *)bindBuffer();
	glColorPointer(m_Mapping[Attributes::ColourBit].m_OglSize, 
					m_Mapping[Attributes::ColourBit].m_OglType, 
					getStride(), 
					base + m_Mapping[Attributes::ColourBit].m_Offset);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void VertexBuffer::selectFogCoords() const {
	ASSERT(m_Attributes & Attributes::Fog);
	ASSERT(m_Mapping[Attributes::FogBit].m_OglSize == 1);
	byte* base = (byte *)bindBuffer();
	glFogCoordPointerEXT(m_Mapping[Attributes::FogBit].m_OglType, 
					getStride(), 
					base + m_Mapping[Attributes::FogBit].m_Offset);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void VertexBuffer::selectEdges() const {
	ASSERT(m_Attributes & Attributes::Edge);
	ASSERT(m_Mapping[Attributes::EdgeBit].m_OglSize == 1);
	byte* base = (byte *)bindBuffer();
	glEdgeFlagPointer(getStride(), base + m_Mapping[Attributes::EdgeBit].m_Offset);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void	VertexBuffer::selectTexturCoords(unsigned int stage) const {
	ASSERT(m_Attributes & (Attributes::Texture << stage));
	ASSERT((m_Mapping[Attributes::TextureBit + stage].m_OglSize >= 1) 
		|| (m_Mapping[Attributes::TextureBit + stage].m_OglSize <= 4));
	byte* base = (byte *)bindBuffer();
	glClientActiveTextureARB(GL_TEXTURE0_ARB + stage);
	glTexCoordPointer(m_Mapping[Attributes::TextureBit + stage].m_OglSize, 
					m_Mapping[Attributes::TextureBit + stage].m_OglType, 
					getStride(), 
					base + m_Mapping[Attributes::TextureBit + stage].m_Offset);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline unsigned int VertexBuffer::getSize() const {
	return m_Size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline unsigned int VertexBuffer::getStride() const {
	return m_Format.getStride();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline VertexBuffer::Usage VertexBuffer::getUsage() const {
	return m_Usage;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline const VertexFormat& VertexBuffer::getFormat() const {
	return m_Format;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline unsigned int VertexBuffer::getOffset(unsigned int i) const {
	return m_Format.getOffset(i);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline unsigned int VertexBuffer::getSize(unsigned int i) const {
	return m_Format.getSize(i);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline unsigned int VertexBuffer::getCount(unsigned int i) const {
	return m_Format.getCount(i);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Types::Type VertexBuffer::getType(unsigned int i) const {
	return m_Format.getType(i);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Attributes::Attribute VertexBuffer::getAttribute(unsigned int i) const {
	return m_Format.getAttribute(i);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline unsigned int	VertexBuffer::getAttributes() const {
	return m_Attributes;
}

}; };

#endif
