/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __GUI_WINDOW_H
#define __GUI_WINDOW_H

#include "application.looper.h"
#include "input.keyboard.h"
#include "input.character.h"
#include "input.mouse.h"
#include "system.signal.h"
#include "graphics.dimension.h"
#include "graphics.rectangle.h"
#include "graphics.point.h"
#include "graphics.palette.h"
#include "container.chain.h"
#include "gui.view.h"


# if defined (__FORX11__)
#	include <X11/Xlib.h>
#	include <X11/Xutil.h>
#	include <X11/Xresource.h>
# endif


namespace magnum { namespace system {
class Thread;
}; };

namespace magnum { namespace types {
class String;
}; };


namespace magnum { namespace gui {
using namespace magnum::system;
using namespace magnum::container;
using magnum::types::byte;
using magnum::types::word;
using magnum::types::dword;
using magnum::system::Signal;
using magnum::application::Handler;
using magnum::application::Looper;
using magnum::application::Message;
using magnum::input::Keyboard;
using magnum::input::Character;
using magnum::input::Mouse;
using magnum::graphics::Palette;
using magnum::graphics::Rectangle;
using magnum::graphics::Dimension;
using magnum::graphics::Point;

class Context;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
class Window : public Looper {
DeclareClass(Window, Looper);
friend class View;
public:
	typedef magnum::container::Chain<View>::Iterator	Iterator;

public:
	enum {
		MSG_OPEN_WINDOW  = 0x70000001,
		MSG_CLOSE_WINDOW = 0x70000002,
		MSG_INVALIDATE_VIEW = 0x70000003,
		MSG_RESIZE_VIEW = 0x70000004
	};

	Window();
	~Window();

	bool					open();
	bool					close();

	inline bool				isOpen() const;

	void					setName(const char* );
	inline String			getName(void) const;

	void					setLocation(const Point<float>& point);
	inline Point<float>		getLocation(void) const;
	void					setSize(const Dimension<float>& dim);
	inline Dimension<float> getSize(void) const;

	inline void				setMinimumSize(const Dimension<float>& min);
	inline Dimension<float> getMinimumSize(void) const;
	inline void				setMaximumSize(const Dimension<float>& min);
	inline Dimension<float> getMaximumSize(void) const;

	bool					insert(View* view);
	bool					remove(View* view);
	bool					remove(const Iterator& iter);
	bool					clear();

	inline Iterator			begin() const;
	inline Iterator			at(View* view) const;
	inline Iterator			end() const;
	
	inline bool				insertHandler(Handler* );
	inline bool				removeHandler(Handler* );
	
	bool					invalidate(const Rectangle<float>& rect);
	bool					invalidate(void); 
							
	inline View*			getFocus() const;
	bool					setFocus(View* focus);
	bool					releaseFocus(View* focus);
	bool					releaseFocus();
	
	bool					setPalette(const Palette<byte>& pal);
	bool					getPalette(Palette<byte>& pal) const;
	bool					realizePalette();
	bool					unrealizePalette();
	
#if defined(__FORWINDOWS__)
	inline HWND				getWindowHandle() const;
	inline HDC				getDeviceHandle() const;
#endif


protected:
	virtual bool			onHandleMessage(const Message& msg);

	// Creation, open and close methods
	virtual bool			onCreate(void);
	virtual bool			onDestroy(void);
	virtual bool			onRequestClose(void);

	// Activation / Deactivation
    virtual bool			onActivate(void);
    virtual bool			onDeactivate(void);
	
	// Window resizing and movement	
	virtual bool			onMove(const Point<float>& );
	virtual bool			onResize(const Rectangle<float>& );

	// Paint method
	virtual bool			onPaint(Context*, const Rectangle<float>& );

	// Keyboard input
	virtual bool			onKeyDown(Keyboard::Key );
	virtual bool			onKeyUp(Keyboard::Key );
	virtual bool			onChar(Character::Code ,Character::Modifier );
	
	// Mouse input
	virtual bool			onMouseMove(const Point<float>& ,Mouse::Button );
	virtual bool			onMouseButton(const Point<float>& ,Mouse::Button );
	virtual bool			onMouseWheel(float ,Mouse::Button );

private:
	bool					sysWindowCreate();
	bool					sysWindowDestroy();
	bool					sysWindowShow();

	// Target System specific stuff
#if defined(__FORWINDOWS__)
public:
	static LRESULT CALLBACK windowProc(HWND hwnd, UINT Message,WPARAM wParam, LPARAM lParam);
#elif defined(__FORX11__)
	virtual bool			dispatchEvent(const XEvent& event);
#endif

private:
	Chain<View>				m_Views;
	Rectangle<float>		m_Frame;
	Rectangle<float>		m_OldFrame;
	Dimension<float>		m_MinSize;
	Dimension<float>		m_MaxSize;
	String					m_Title;
	View*					m_Focus;
	bool					m_Open;
	Signal					m_Signal;
	Context					m_Context;

	// Target System specific stuff
#if defined(__FORWINDOWS__)
	HWND					m_WindowHandle;
	HDC						m_DeviceHandle;
	DWORD					m_WindowStyle;
	LOGPALETTE				m_Palette;
	PALETTEENTRY			m_PaletteColours[256];
	HPALETTE				m_PaletteHandle;
#elif defined(__FORX11__)
	Window					m_WindowHandle;
    C						m_DeviceHandle;
    Colormap				m_PaletteHandle;
    XColor*					m_PaletteColours;
#endif
};

}; };
#endif
