/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "io.memfile.h"
#include "io.exception.h"

#include "io.memfile.inl"
#include "io.inputstream.inl"
#include "io.outputstream.inl"
#include "io.filestream.inl"


namespace magnum { namespace io {
using magnum::math::min;
using magnum::math::max;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MemoryFile::MemoryFile(unsigned int bs) : File(File::Readable | File::Writeable) {
	m_Position = 0;
	m_Size = 0;
	m_Blocksize = bs;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MemoryFile::~MemoryFile() {
	while(m_Blocks.getSize() > 0) {
		void* data = m_Blocks[m_Blocks.getSize() - 1];
		delete[] (byte *)data;
		m_Blocks.remove(m_Blocks.getSize() - 1);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t MemoryFile::onWrite(const void *buf, size_t size) {
	size_t posInCurrentBlock = m_Position % m_Blocksize;
	size_t freeInCurrentBlock = (m_Blocksize - posInCurrentBlock) % m_Blocksize;
	size_t bytesToWrite;
	size_t bytesWritten = 0;

	bytesToWrite = min(freeInCurrentBlock, size);
	if (bytesToWrite > 0) {
		memcpy(((byte *)m_Blocks[(unsigned int)(m_Position / m_Blocksize)]) + posInCurrentBlock, buf, bytesToWrite);
		m_Position += bytesToWrite;
		m_Size = max(m_Size, m_Position);
		size -= bytesToWrite;
		buf = ((byte *)buf) + bytesToWrite;
		bytesWritten += bytesToWrite;
	}

	if (m_Position + size > m_Size)
		append(m_Position + size - m_Size);

	while(size > 0) {
		bytesToWrite = min(m_Blocksize, size);
		memcpy(m_Blocks[(unsigned int)(m_Position / m_Blocksize)], buf, bytesToWrite);
		m_Position += bytesToWrite;
		size -= bytesToWrite;
		buf = ((byte *)buf) + bytesToWrite;
		bytesWritten += bytesToWrite;
	}

	return bytesWritten;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t MemoryFile::onRead(void *buf, size_t size) {
	if (size == 0)
		return 0;
	if (m_Position >= m_Size)
		return EndOfFile;
		
	size_t posInCurrentBlock = m_Position % m_Blocksize;
	size_t leftInCurrentBlock = (m_Blocksize - m_Position) % m_Blocksize;
	size_t bytesToRead;
	size_t bytesRead = 0;

	size = min(size, m_Size - m_Position);
	bytesToRead = min(leftInCurrentBlock, size);
	if (bytesToRead > 0) {
		memcpy(buf, ((byte *)m_Blocks[(unsigned int)(m_Position / m_Blocksize)]) + posInCurrentBlock, bytesToRead);
		m_Position += bytesToRead;
		size -= bytesToRead;
		buf = ((byte *)buf) + bytesToRead;
		bytesRead += bytesToRead;
	}

	while(size > 0) {
		bytesToRead = min(m_Blocksize, size);
		memcpy(buf, m_Blocks[(unsigned int)(m_Position / m_Blocksize)], bytesToRead);
		m_Position += bytesToRead;
		size -= bytesToRead;
		buf = ((byte *)buf) + bytesToRead;
		bytesRead += bytesToRead;
	}

	return bytesRead;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool MemoryFile::onFlush() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int64 MemoryFile::onSetPosition(long where,PositionMode mode) {
	switch(mode) {
		case Absolute:
			if (where < 0)
				throw IllegalPositionException();
			if (where > (long)m_Size)
				throw IllegalPositionException();
			m_Position = where;
			break;
		case Current:
			if (m_Position + where < 0)
				throw IllegalPositionException();
			if (m_Position + where > m_Size)
				throw IllegalPositionException();
			m_Position += where;
			break;
		case End:
			if (where > 0)
				throw IllegalPositionException();
			if (m_Size + where < 0)
				throw IllegalPositionException();
			m_Position = m_Size + where;
		default:
			throw IllegalPositionModeException();
	}

	return m_Position;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int64 MemoryFile::onGetPosition() const {
	return m_Position;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool MemoryFile::onRewind() {
	m_Position = 0;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int64 MemoryFile::onGetSize() const {
	return m_Size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int64 MemoryFile::onSetSize(int64 newSize) {
	if (newSize < m_Size)
		truncate((size_t)(m_Size - newSize));
	if (newSize > m_Size)
		append((size_t)(newSize - m_Size));
	return m_Size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool MemoryFile::append(size_t addSize) {
	size_t freeInLastBlock = m_Blocks.getSize()*m_Blocksize - m_Size;
	if (addSize <= freeInLastBlock) {
		m_Size += addSize;
		return true;
	}

	addSize -= freeInLastBlock;
	m_Size += freeInLastBlock;

	for(;;) {
		void *newBlock = new byte[m_Blocksize];
		if (newBlock == NULL)
			throw DiskFullException("Not enough free memory for requested operation.");
		m_Blocks.insert(newBlock);
		if (addSize > (int64)m_Blocksize) {
			m_Size += m_Blocksize;
			addSize -= m_Blocksize;
		} else {
			m_Size += addSize;
			return true;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool MemoryFile::truncate(size_t delSize) {
	size_t usedInLastBlock = (m_Blocksize - (m_Size % m_Blocksize)) % m_Blocksize;
	if (delSize < usedInLastBlock) {
		m_Size -= delSize;
		return true;
	}

	delSize -= usedInLastBlock;
	m_Size -= usedInLastBlock;
	void* data = m_Blocks[m_Blocks.getSize() - 1];
	delete[] (byte *)data;
	m_Blocks.remove(m_Blocks.getSize() - 1);

	for(;;) {
		if (delSize >= (int64)m_Blocksize) {
			data = m_Blocks[m_Blocks.getSize() - 1];
			delete[] (byte *)data;
			m_Blocks.remove(m_Blocks.getSize() - 1);
			m_Size -= m_Blocksize;
		} else {
			m_Size -= delSize;
			return true;
		}
	}
}

}; };
