/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __MATH_LIMITS_H
#define __MATH_LIMITS_H

#include <float.h>

#include "math.consts.h"

namespace magnum { namespace math {

/*--------------------------------------------------------------------------*/
/**
 */
template<class T> class Limits {
public:
	static const bool	hasInfinity = false;
	static const bool	hasNaN = false;
	static const bool	isExact = false;
	static const bool	isBounded = true;
	static const bool	isSigned = true;
	static const bool	isInteger = true;
	static const int	minExponent = 0;
	static const int	maxExponent = 0;
	static inline T		epsilon() { return T(0); };
	static inline T		infinity() { return T(0); };
	static inline T		NaN() { return T(0); };
	static inline T		min() { return T(0); };
	static inline T		max() { return T(0); };
	static inline T		isNaN(const T& ) { return false; };
	static inline T		isInfinity(const T& ) { return false; };
};


/*--------------------------------------------------------------------------*/
/**
 */
template<class T> class IntegerLimits {
public:
	static const bool	hasInfinity = false;
	static const bool	hasNaN = false;
	static const bool	isExact = true;
	static const bool	isBounded = true;
	static const bool	isSigned = true;
	static const bool	isInteger = true;
	static const int	minExponent = 0;
	static const int	maxExponent = 0;
	static inline T		epsilon() { return T(1); };
	static inline T		infinity() { return T(0); };
	static inline T		NaN() { return T(0); };
	static inline bool	isNaN(const T& ) { return false; };
	static inline bool	isInfinity(const T& ) { return false; };
};


/*--------------------------------------------------------------------------*/
/**
 */
template<class T> class FloatLimits {
public:
	static const bool	hasInfinity = true;
	static const bool	hasNaN = true;
	static const bool	isExact = false;
	static const bool	isBounded = true;
	static const bool	isSigned = true;
	static const bool	isInteger = false;
};


/*--------------------------------------------------------------------------*/
/**
 */
template<> class Limits<signed int> : public IntegerLimits<signed int> {
public:
	static inline signed int	min() { return -0x7fffffff; };
	static inline signed int	max() { return 0x7fffffff; };
};


/*--------------------------------------------------------------------------*/
/**
 */
template<> class Limits<unsigned int> : public IntegerLimits<unsigned int> {
public:
	static const bool	isSigned = false;
	static inline unsigned int min() { return 0; };
	static inline unsigned int max() { return 0xffffffff; };
};


/*--------------------------------------------------------------------------*/
/**
 */
template<> class Limits<signed short> : public IntegerLimits<signed short> {
public:
	static inline signed short	min() { return -0x7fff; };
	static inline signed short	max() { return 0x7fff; };
};


/*--------------------------------------------------------------------------*/
/**
 */
template<> class Limits<unsigned short> : public IntegerLimits<unsigned short> {
public:
	static const bool		isSigned = false;
	static inline unsigned short min() { return 0; };
	static inline unsigned short max() { return 0xffff; };
};


/*--------------------------------------------------------------------------*/
/**
 */
template<> class Limits<float> : public FloatLimits<float> {
public:
	static const int		minExponent = FLT_MIN_EXP;
	static const int		maxExponent = FLT_MAX_EXP;
	static inline float		epsilon() { return FLT_EPSILON; };
	static inline float		min() { return FLT_MIN; };
	static inline float		max() { return FLT_MAX; };
	static inline float		NaN() { return *(float*)_quietNan; };
	static inline float		infinity() { return *(float*)_infinity; };
	static inline bool		isNaN(float f) { return (f != f); };
	static inline bool		isInfinity(float f) { return (f == infinity() || f == -infinity()); };
	
private:
	static const unsigned char _quietNan[4];	
	static const unsigned char _signalingNan[4];
	static const unsigned char _infinity[4];
};


/*--------------------------------------------------------------------------*/
/**
 */
template<> class Limits<double> : public FloatLimits<double> {
public:
	static const int		minExponent = DBL_MIN_EXP;
	static const int		maxExponent = DBL_MAX_EXP;
	static inline double	epsilon() { return DBL_EPSILON; };
	static inline double	min() { return DBL_MIN; };
	static inline double	max() { return DBL_MAX; };
	static inline double	NaN() { return *(double*)_quietNan; };
	static inline double	infinity() { return *(double*)_infinity; };
	static inline bool		isNaN(double f) { return (f != f); };
	static inline bool		isInfinity(double f) { return (f == infinity() || f == -infinity()); };

private:
	static const unsigned char _quietNan[8];	
	static const unsigned char _signalingNan[8];	
	static const unsigned char _infinity[8];	
};


/*--------------------------------------------------------------------------*/
/**
 */
template<> class Limits<long double> : public FloatLimits<long double> {
public:
	static const int		minExponent = LDBL_MIN_EXP;
	static const int		maxExponent = LDBL_MAX_EXP;
	static inline long double epsilon() { return LDBL_EPSILON; };
	static inline long double min() { return LDBL_MIN; };
	static inline long double max() { return LDBL_MAX; };
	static inline long double NaN() { return *(double*)_quietNan; };
	static inline long double infinity() { return *(double*)_infinity; };
	static inline bool		isNaN(long double f) { return (f != f); };
	static inline bool		isInfinity(long double f) { return (f == infinity() || f == -infinity()); };

private:
	static const unsigned char _quietNan[8];	
	static const unsigned char _signalingNan[8];	
	static const unsigned char _infinity[8];	
};

}; };

#endif
