/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __MATH_LINEARSYSTEM_INL
#define __MATH_LINEARSYSTEM_INL

namespace magnum { namespace math {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> LinearSystem<T>::LinearSystem(unsigned int size) {
	m_Inverse = NULL;
	m_Size = size;
	if (m_Size > 0) {
		m_System = new T[m_Size * m_Size];
	} else {
		m_System = NULL;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> LinearSystem<T>::LinearSystem(const LinearSystem<T>& l) {
	m_Inverse = NULL;
	m_System = NULL;
	m_Size = l.m_Size;
	if (m_Size > 0) {
		m_System = new T[m_Size * m_Size];
		CopyElements(m_System, l.m_System, m_Size*m_Size);
		if (l.m_Inverse) {
			m_Inverse = new T[m_Size*m_Size];
			CopyElements(m_Inverse, l.m_Inverse, m_Size*m_Size);
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> LinearSystem<T>::~LinearSystem() {
	if (m_System != NULL)
		delete[] m_System;
	if (m_Inverse != NULL)
		delete[] m_Inverse;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline void LinearSystem<T>::setSize(unsigned int size) {
	if (m_Size == size)
		return;
	if (m_System != NULL)
		delete[] m_System;
	if (m_Inverse != NULL)
		delete[] m_Inverse;
	
	m_Inverse = NULL;
	m_Size = size;
	if (m_Size > 0) {
		m_System = new T[m_Size * m_Size];
	} else {
		m_System = NULL;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline unsigned int LinearSystem<T>::getSize() const {
	return m_Size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
template<int dim> 
inline LinearSystem<T>::LinearSystem(const Matrix<T,dim>& mtx) {
	m_Inverse = NULL;
	m_Size = dim;
	m_System = new T[dim*dim];
	CopyElements(m_System, &mtx[0][0], dim*dim);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
template<int dim>
inline bool LinearSystem<T>::solve(const Vector<T,dim>& b,Vector<T,dim>* x) const {
	ASSERT(m_Size == dim);
	return LinearSolver<T>::solve(dim, m_System, (const T*)b, (T*)x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
template<int dim>
inline LinearSystem<T>& LinearSystem<T>::operator=(const Matrix<T,dim>& mtx) {
	setSize(dim);
	CopyElements(m_System, &mtx[0][0], dim*dim);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
template<int dim>
void LinearSystem<T>::apply(const Vector<T,dim>& b,Vector<T,dim>* x) const {
	ASSERT(m_Inverse != NULL);
	ASSERT(m_Size == dim);
	for (unsigned int i = 0; i < dim; i++) {
		T tmp = T(0);
		for (unsigned int j = 0; j < dim; j++)
			tmp += m_Inverse[i*dim + j] * b[j];
		x[i] = tmp;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool LinearSystem<T>::solve(const T* b,T* x) const {
	ASSERT(m_Size > 0);
	return LinearSolver<T>::solve(m_Size, m_System, b, x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool LinearSystem<T>::solve(T* b) const {
	ASSERT(m_Size > 0);
	return LinearSolver<T>::solve(m_Size, m_System, b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> void LinearSystem<T>::apply(const T* b,T* x) const {
	ASSERT(m_Inverse != NULL);
	for (unsigned int i = 0; i < m_Size; i++) {
		T tmp = T(0);
		for (unsigned int j = 0; j < m_Size; j++)
			tmp += m_Inverse[i*m_Size + j] * b[j];
		x[i] = tmp;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline void LinearSystem<T>::apply(T* b) const;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool LinearSystem<T>::invert() {
	if (m_Inverse == NULL)
		m_Inverse = new T[m_Size*m_Size];

	return LinearSolver<T>::invert(m_Size, m_System, m_Inverse);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const T& LinearSystem<T>::operator()(unsigned int i, unsigned int j) const {
	ASSERT(i < m_Size);
	ASSERT(j < m_Size);
	return m_System[i*m_Size + j];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T& LinearSystem<T>::operator()(unsigned int i, unsigned int j) {
	ASSERT(i < m_Size);
	ASSERT(j < m_Size);
	return m_System[i*m_Size + j];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const T* LinearSystem<T>::operator[](unsigned int i) const {
	ASSERT(i < m_Size);
	return &m_System[i*m_Size];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T* LinearSystem<T>::operator[](unsigned int i) {
	ASSERT(i < m_Size);
	return &m_System[i*m_Size];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> LinearSystem<T>& LinearSystem<T>::operator=(const LinearSystem<T>& l) {
	setSize(l.m_Size);
	if (m_Size > 0) {
		CopyElements(m_System, l.m_System, m_Size*m_Size);
		if (l.m_Inverse) {
			m_Inverse = new T[m_Size*m_Size];
			CopyElements(m_Inverse, l.m_Inverse, m_Size*m_Size);
		}
	}
}

}; };

#endif
