/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __MATH_POLYNOMIAL_INL
#define __MATH_POLYNOMIAL_INL

namespace magnum { namespace math {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T>::Polynomial(unsigned int d) {
	m_Degree = d;
	m_Coefficients = new T[d+1];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T>::Polynomial(const Polynomial<S,T>& p) {
	m_Degree = p.m_Degree;
	m_Coefficients = new T[m_Degree+1];
	CopyElements(m_Coefficients, p.m_Coefficients, m_Degree+1);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T>::~Polynomial() {
	delete[] m_Coefficients;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline unsigned int Polynomial<S,T>::getDegree() const {
	return m_Degree;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
unsigned int Polynomial<S,T>::getEffectiveDegree() const {
	unsigned int i;
	for (i = m_Degree; i > 0; i--)
		if (m_Coefficients[i] != T(0))
			return i;
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline void Polynomial<S,T>::setDegree(unsigned int d) {
	if (m_Degree == d)
		return;
	delete[] m_Coefficients;
	m_Degree = d;
	m_Coefficients = new T[m_Degree + 1];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T>& Polynomial<S,T>::operator=(const Polynomial<S,T>& p) {
	setDegree(p.m_Degree);
	CopyElements(m_Coefficients, p.m_Coefficients, m_Degree+1);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline T Polynomial<S,T>::operator() (const S& x) const {
	T f = T(0);
	for (unsigned int i = m_Degree; i > 0; i--) {
		f = (f + m_Coefficients[i])*x;
	}
	f += m_Coefficients[0];

	return f;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T> Polynomial<S,T>::operator+ (const Polynomial<S,T>& p) const {
	unsigned int i;
	if (m_Degree >= p.m_Degree) {
		Polynomial<S,T> poly(m_Degree);
		for (i = 0; i <= p.m_Degree; i++)
			poly.m_Coefficients[i] = m_Coefficients[i] + p.m_Coefficients[i];
		for (i = p.m_Degree + 1; i <= m_Degree; i++)
			poly.m_Coefficients[i] = m_Coefficients[i];
		return poly;
	} else {
		Polynomial<S,T> poly(p.m_Degree);
		for (i = 0; i <= m_Degree; i++)
			poly.m_Coefficients[i] = m_Coefficients[i] + p.m_Coefficients[i];
		for (i = m_Degree + 1; i <= p.m_Degree; i++)
			poly.m_Coefficients[i] = p.m_Coefficients[i];
		return poly;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T> Polynomial<S,T>::operator- (const Polynomial<S,T>& p) const {
	unsigned int i;
	if (m_Degree >= p.m_Degree) {
		Polynomial<S,T> poly(m_Degree);
		for (i = 0; i <= p.m_Degree; i++)
			poly.m_Coefficients[i] = m_Coefficients[i] - p.m_Coefficients[i];
		for (i = p.m_Degree + 1; i <= m_Degree; i++)
			poly.m_Coefficients[i] = m_Coefficients[i];
		return poly;
	} else {
		Polynomial<S,T> poly(p.m_Degree);
		for (i = 0; i <= m_Degree; i++)
			poly.m_Coefficients[i] = m_Coefficients[i] - p.m_Coefficients[i];
		for (i = m_Degree + 1; i <= p.m_Degree; i++)
			poly.m_Coefficients[i] = - p.m_Coefficients[i];
		return poly;
	}
}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T> Polynomial<S,T>::operator* (const Polynomial<S,T>& p) const {
	unsigned int da = getEffectiveDegree();
	unsigned int db = p.getEffectiveDegree();
	Polynomial poly(da + db);

	for (unsigned int k = 0; k <= da+db; k++)
		poly.m_Coefficients[k] = T(0);
	for (unsigned int i = 0; i <= da; i++)
		for (unsigned int j = 0; j <= db; j++)
			poly.m_Coefficients[i+j] += m_Coefficients[i]*p.m_Coefficients[j];

	return poly;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T> Polynomial<S,T>::operator+ (const T& t) const {
	Polynomial<S,T> poly(*this);
	poly.m_Coefficients[0] += t;
	return poly;
} 


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T> Polynomial<S,T>::operator- (const T& t) const {
	Polynomial<S,T> poly(*this);
	poly.m_Coefficients[0] -= t;
	return poly;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T> Polynomial<S,T>::operator* (const S& s) const {
	Polynomial<S,T> poly(m_Degree);
	for (unsigned int i = 0; i <= m_Degree; i++)
		poly.m_Coefficients[i] = m_Coefficients[i] * s;
	return poly;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T> Polynomial<S,T>::operator/ (const S& s) const {
	Polynomial<S,T> poly(m_Degree);
	for (unsigned int i = 0; i <= m_Degree; i++)
		poly.m_Coefficients[i] = m_Coefficients[i] / s;
	return poly;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T> Polynomial<S,T>::operator- () const {
	Polynomial<S,T> poly(m_Degree);
	for (unsigned int i = 0; i <= m_Degree; i++)
		poly.m_Coefficients[i] = -m_Coefficients[i];
	return poly;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
Polynomial<S,T>& Polynomial<S,T>::operator += (const Polynomial<S,T>& p) {
	unsigned int i;
	if (m_Degree >= p.m_Degree) {
		for (i = 0; i <= p.m_Degree; i++)
			m_Coefficients[i] += p.m_Coefficients[i];
	} else {
		T* poly = new T[p.m_Degree];
		for (i = 0; i <= m_Degree; i++)
			poly[i] = m_Coefficients[i] + p.m_Coefficients[i];
		for (i = m_Degree + 1; i <= p.m_Degree; i++)
			poly[i] = p.m_Coefficients[i];
		delete[] m_Coefficients;
		m_Coefficients = poly;
	}

	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
Polynomial<S,T>& Polynomial<S,T>::operator -= (const Polynomial<S,T>& p) {
	unsigned int i;
	if (m_Degree >= p.m_Degree) {
		for (i = 0; i <= p.m_Degree; i++)
			m_Coefficients[i] -= p.m_Coefficients[i];
	} else {
		T* poly = new T[p.m_Degree];
		for (i = 0; i <= m_Degree; i++)
			poly[i] = m_Coefficients[i] - p.m_Coefficients[i];
		for (i = m_Degree + 1; i <= p.m_Degree; i++)
			poly[i] = -p.m_Coefficients[i];
		delete[] m_Coefficients;
		m_Coefficients = poly;
	}

	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T>& Polynomial<S,T>::operator *= (const Polynomial<S,T>& p);


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T>& Polynomial<S,T>::operator += (const T& t) {
	m_Coefficients[0] += t;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T>& Polynomial<S,T>::operator -= (const T& t) {
	m_Coefficients[0] -= t;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
Polynomial<S,T>& Polynomial<S,T>::operator *= (const S& s) {
	for (unsigned int i = 0; i <= m_Degree; i++)
		m_Coefficients[i] *= s;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
Polynomial<S,T>& Polynomial<S,T>::operator /= (const S& s) {
	for (unsigned int i = 0; i <= m_Degree; i++)
		m_Coefficients[i] /= s;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
Polynomial<S,T> Polynomial<S,T>::derive() const {
	if (m_Degree == 0) {
		Polynomial<S,T> dv(0);
		dv[0] = T(0);
		return dv;
	}

	Polynomial<S,T> dv(m_Degree - 1);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T>::operator const T*() const {
	return m_Coefficients;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T>::operator T*() {
	return m_Coefficients;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
bool Polynomial<S,T>::operator==(const Polynomial<S,T>& p) const {
	unsigned int i;
	if (m_Degree <= p.m_Degree) {
		for (i = 0; i <= m_Degree; i++) {
			if (m_Coefficients[i] != p.m_Coefficients[i])
				return false;
		}
		for (; i <= p.m_Degree; i++) {
			if (p.m_Coefficients[i] != T(0))
				return false;
		}
	} else {
		for (i = 0; i <= p.m_Degree; i++) {
			if (m_Coefficients[i] != p.m_Coefficients[i])
				return false;
		}
		for (; i <= m_Degree; i++) {
			if (m_Coefficients[i] != T(0))
				return false;
		}
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
bool Polynomial<S,T>::operator!=(const Polynomial<S,T>& p) const {
	unsigned int i;
	if (m_Degree <= p.m_Degree) {
		for (i = 0; i <= m_Degree; i++) {
			if (m_Coefficients[i] != p.m_Coefficients[i])
				return true;
		}
		for (; i <= p.m_Degree; i++) {
			if (p.m_Coefficients[i] != T(0))
				return true;
		}
	} else {
		for (i = 0; i <= p.m_Degree; i++) {
			if (m_Coefficients[i] != p.m_Coefficients[i])
				return true;
		}
		for (; i <= m_Degree; i++) {
			if (m_Coefficients[i] != T(0))
				return true;
		}
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
bool Polynomial<S,T>::operator<=(const Polynomial<S,T>& p) const {
	int da = getEffectiveDegree(), db = p.getEffectiveDegree();
	if (da < db)
		return true;
	if (da > db)
		return false;

	for (int i = da; i >= 0; i--) {
		if (m_Coefficients[i] < p.m_Coefficients[i])
			return true;
		if (m_Coefficients[i] > p.m_Coefficients[i])
			return false;
	}
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
bool Polynomial<S,T>::operator>=(const Polynomial<S,T>& p) const {
	int da = getEffectiveDegree(), db = p.getEffectiveDegree();
	if (da > db)
		return true;
	if (da < db)
		return false;

	for (int i = da; i >= 0; i--) {
		if (m_Coefficients[i] > p.m_Coefficients[i])
			return true;
		if (m_Coefficients[i] < p.m_Coefficients[i])
			return false;
	}
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
bool Polynomial<S,T>::operator<(const Polynomial<S,T>& p) const {
	int da = getEffectiveDegree(), db = p.getEffectiveDegree();
	if (da > db)
		return true;
	if (da < db)
		return false;

	for (int i = da; i >= 0; i--) {
		if (m_Coefficients[i] > p.m_Coefficients[i])
			return true;
		if (m_Coefficients[i] < p.m_Coefficients[i])
			return false;
	}
	
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
bool Polynomial<S,T>::operator>(const Polynomial<S,T>& p) const {
	int da = getEffectiveDegree(), db = p.getEffectiveDegree();
	if (da > db)
		return true;
	if (da < db)
		return false;

	for (int i = da; i >= 0; i--) {
		if (m_Coefficients[i] > p.m_Coefficients[i])
			return true;
		if (m_Coefficients[i] < p.m_Coefficients[i])
			return false;
	}
	
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline Polynomial<S,T> operator*(const S& s,Polynomial<S,T>& p) {
	Polynomial<S,T> poly(p.getDegree());
	for (unsigned int i = 0; i <= p.getDegree(); i++)
		poly[i] = s * p[i];
	return poly;
}

}; };

#endif
