/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __MATH_QUATERNION_INL
#define __MATH_QUATERNION_INL

namespace magnum { namespace math {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T>::Quaternion() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T>::Quaternion(const T& iw, const T& ix, const T& iy, const T& iz) {
	w = iw;
	x = ix;
	y = iy;
	z = iz;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::operator+(const Quaternion<T>& q) const {
	return Quaternion(w + q.w, x + q.x, y + q.y, z + q.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::operator-(const Quaternion<T>& q) const {
	return Quaternion(w - q.w, x - q.x, y - q.y, z - q.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::operator*(const Quaternion<T>& q) const {
	return Quaternion(
		w*q.w - x*q.x - y*q.y - z*q.z,
		w*q.x + x*q.w + y*q.z - z*q.y,
		w*q.y + y*q.w + z*q.x - x*q.z,
		w*q.z + z*q.w + x*q.y - y*q.x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::operator*(const T& t) const {
	return Quaternion(w * t, x * t, y * t, z * t);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::operator/(const T& t) const {
	return Quaternion(w / t, x / t, y / t, z / t);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::operator-() const {
	return Quaternion(-w, -x, -y, -z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline T Quaternion<T>::dot(const Quaternion<T>& q) const {
	return w*q.w + x*q.x + y*q.y + z*q.z;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline T Quaternion<T>::length2() const {
	return w*w + x*x + y*y + z*z;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline T Quaternion<T>::length() const {
	return T(sqrt(w*w + x*x + y*y + z*z));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T>& Quaternion<T>::normalise() {
	T l = w*w + x*x + y*y + z*z;
	if (l <= T(0))
		return *this;
	l = T(1.0f / sqrt(l));
	w *= l;
	x *= l;
	y *= l;
	z *= l;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Quaternion<T>& Quaternion<T>::operator= (const Quaternion<T>& q) {
	w = q.w;
	x = q.x;
	y = q.y;
	z = q.z;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Quaternion<T>& Quaternion<T>::operator+=(const Quaternion<T>& q) {
	w += q.w;
	x += q.x;
	y += q.y;
	z += q.z;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Quaternion<T>& Quaternion<T>::operator-=(const Quaternion<T>& q) {
	w -= q.w;
	x -= q.x;
	y -= q.y;
	z -= q.z;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Quaternion<T>& Quaternion<T>::operator*=(const T& t) {
	w *= t;
	x *= t;
	y *= t;
	z *= t;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Quaternion<T>& Quaternion<T>::operator/=(const T& t) {
	w /= t;
	x /= t;
	y /= t;
	z /= t;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Quaternion<T>& Quaternion<T>::operator*=(const Quaternion<T>& q) {
	T nw = w*q.w - x*q.x - y*q.y - z*q.z;
	T nx = w*q.x + x*q.w + y*q.z - z*q.y;
	T ny = w*q.y + y*q.w + z*q.x - x*q.z;
	T nz = w*q.z + z*q.w + x*q.y - y*q.x;
	w = nw;
	x = nx;
	y = ny;
	z = nz;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline bool Quaternion<T>::operator==(const Quaternion<T>& q) const {
	return ((x == q.x) && (y == q.y) && (z == q.z) && (w == q.w)):
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline bool Quaternion<T>::operator!=(const Quaternion<T>& q) const {
	return ((x != q.x) || (y != q.y) || (z != q.z) || (w != q.w)):
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline bool Quaternion<T>::operator<(const Quaternion& q) const {
	if (w < q.w) return true;
	if (w > q.w) return false;
	if (x < q.x) return true;
	if (x > q.x) return false;
	if (y < q.y) return true;
	if (y > q.y) return false;
	return (z < q.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline bool Quaternion<T>::operator>(const Quaternion& q) const {
	if (w > q.w) return true;
	if (w < q.w) return false;
	if (x > q.x) return true;
	if (x < q.x) return false;
	if (y > q.y) return true;
	if (y < q.y) return false;
	return (z > q.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline bool Quaternion<T>::operator<=(const Quaternion& q) const {
	if (w < q.w) return true;
	if (w > q.w) return false;
	if (x < q.x) return true;
	if (x > q.x) return false;
	if (y < q.y) return true;
	if (y > q.y) return false;
	return (z <= q.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline bool Quaternion<T>::operator>=(const Quaternion& q) const {
	if (w > q.w) return true;
	if (w < q.w) return false;
	if (x > q.x) return true;
	if (x < q.x) return false;
	if (y > q.y) return true;
	if (y < q.y) return false;
	return (z >= q.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T>::operator const T*() const {
	return &x;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T>::operator T*() {
	return &x;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Matrix<T,3> Quaternion<T>::toMatrix() const {
    T tx  = ((T)2.0)*x;
    T ty  = ((T)2.0)*y;
    T tz  = ((T)2.0)*z;
    T twx = tx*w;
    T twy = ty*w;
    T twz = tz*w;
    T txx = tx*x;
    T txy = ty*x;
    T txz = tz*x;
    T tyy = ty*y;
    T tyz = tz*y;
    T tzz = tz*z;

	Matrix<T,3> mtx;

    mtx(0,0) = ((T)1.0) - (tyy+tzz);
    mtx(0,1) = txy - twz;
    mtx(0,2) = txz + twy;
    mtx(1,0) = txy + twz;
    mtx(1,1) = ((T)1.0) - (txx+tzz);
    mtx(1,2) = tyz - twx;
    mtx(2,0) = txz - twy;
    mtx(2,1) = tyz + twx;
    mtx(2,2) = ((T)1.0) - (txx+tyy);
	return mtx;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline void Quaternion<T>::toMatrix(Matrix<T,3>* mtx) const {
    T tx  = ((T)2.0)*x;
    T ty  = ((T)2.0)*y;
    T tz  = ((T)2.0)*z;
    T twx = tx*w;
    T twy = ty*w;
    T twz = tz*w;
    T txx = tx*x;
    T txy = ty*x;
    T txz = tz*x;
    T tyy = ty*y;
    T tyz = tz*y;
    T tzz = tz*z;

    (*mtx)(0,0) = ((T)1.0) - (tyy+tzz);
    (*mtx)(0,1) = txy - twz;
    (*mtx)(0,2) = txz + twy;
    (*mtx)(1,0) = txy + twz;
    (*mtx)(1,1) = ((T)1.0) - (txx+tzz);
    (*mtx)(1,2) = tyz - twx;
    (*mtx)(2,0) = txz - twy;
    (*mtx)(2,1) = tyz + twx;
    (*mtx)(2,2) = ((T)1.0) - (txx+tyy);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
void Quaternion<T>::fromMatrix(const Matrix<T,3>& mtx) {
    // Algorithm in Ken Shoemake's article in 1987 SIGGRAPH course notes
    // article "Quaternion Calculus and Fast Animation".
    T trace = mtx(0,0) + mtx(1,1) + mtx(2,2);
    T root;

    if ( trace > (T)0 ) {
        // |w| > 1/2, may as well choose w > 1/2
        root = sqrt(trace + (T)1);  // 2w
        w = ((T)0.5) * root;
        root = ((T)0.5)/root;  // 1/(4w)
        x = (mtx(2,1) - mtx(1,2)) * root;
        y = (mtx(0,2) - mtx(2,0)) * root;
        z = (mtx(1,0) - mtx(0,1)) * root;
        return;
    } 
    // |w| <= 1/2
    if ( mtx(1,1) > mtx(0,0) ) {
		if ( mtx(2,2) > mtx(1,1) ) {
			root = sqrt(mtx(2,2) - mtx(0,0) - mtx(1,1) + (T)1);
			z = ((T)0.5)*root;
			root = ((T)0.5)/root;
			w = (mtx(1,0) - mtx(0,1)) * root;
			x = (mtx(0,2) + mtx(2,0)) * root;
			y = (mtx(1,2) + mtx(2,1)) * root;
		} else {
			root = sqrt(mtx(1,1) - mtx(2,2) - mtx(0,0) + (T)1);
			y = ((T)0.5)*root;
			root = ((T)0.5)/root;
			w = (mtx(0,2) - mtx(2,0)) * root;
			z = (mtx(2,1) + mtx(1,2)) * root;
			x = (mtx(0,1) + mtx(1,0)) * root;
		}
    } else {
		if ( mtx(2,2) > mtx(0,0) ) {
			root = sqrt(mtx(2,2) - mtx(0,0) - mtx(1,1) + (T)1);
			z = ((T)0.5)*root;
			root = ((T)0.5)/root;
			w = (mtx(1,0) - mtx(0,1)) * root;
			x = (mtx(0,2) + mtx(2,0)) * root;
			y = (mtx(1,2) + mtx(2,1)) * root;
		} else {
			root = sqrt(mtx(0,0) - mtx(1,1) - mtx(2,2) + (T)1);
			x = ((T)0.5)*root;
			root = ((T)0.5)/root;
			w = (mtx(2,1) - mtx(1,2)) * root;
			y = (mtx(1,0) + mtx(0,1)) * root;
			z = (mtx(2,0) + mtx(0,2)) * root;
		}
    }
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template <class T>
void Quaternion<T>::toAxisAngle(Vector3<T>* v, T* angle) const {
    T l2 = x*x + y*y + z*z;
    if ( l2 > T(0) ) {
        *angle = T(2) * acos(w);
        T il = T(1) / sqrt(l2);
        v->x = x*il;
        v->y = y*il;
        v->z = z*il;
    } else {
        // angle is 0 (mod 2*pi), so any axis will do
        *angle = (T)0;
        v->x = (T)1;
        v->y = (T)0;
        v->z = (T)0;
    }
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template <class T>
inline void Quaternion<T>::fromAxisAngle(const Vector3<T>& axis, T angle) {
    // assert:  axis[] is unit length
    T ha = ((T)0.5)*angle;
    T si = (T)sin(ha);
    w = (T)cos(ha);
    x = si*axis.x;
    y = si*axis.y;
    z = si*axis.z;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::getInverse() const {
	T in = w*w + x*x + y*y + z*z;
	in = T(1) / in;
    return Quaternion<T>(w*in, -x*in, -y*in, -z*in);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T>& Quaternion<T>::invert() {
	T in = w*w + x*x + y*y + z*z;
	in = T(1) / in;
    w *= in;
    x *= -in;
    y *= -in;
    z *= -in;
    return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::getConjugate() const {
	return Quaternion<T>(w, -x, -y, -z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T>& Quaternion<T>::conjugate() {
	x = -x;
	y = -y;
	z = -z;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector<T,3> operator*(const Quaternion<T>& q,const Vector<T,3>& v) {
    Matrix<T,3> mtx;
    q.toMatrix(&mtx);
    return mtx * v;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector3<T> operator*(const Quaternion<T>& q,const Vector3<T>& v) {
    Matrix<T,3> mtx;
    q.toMatrix(&mtx);
    return mtx * v;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Quaternion<T> operator*(const T& t,const Quaternion<T>& q) {
	return Quaternion<T>(t*q.w, t*q.x, t*q.y, t*q.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> slerp(const T& t,const Quaternion<T>& p, const Quaternion<T>& q) {
	T d = p.dot(q);
	if (abs(1.0f - d) < 0.0001f)
		return t*(q - p) + p;
		
    if (d < 0.0f) {
		T angle = acos(-d);
		T s = T(1) / sin(angle);
		T c0 = sin((T(1) - t)*angle) * s;
		T c1 = sin(t*angle) * s;
		return c0*p - c1*q;
    } else {
	    T angle = acos(d);
	    T s = T(1) / sin(angle);
		T c0 = sin((T(1) - t)*angle) * s;
		T c1 = sin(t*angle) * s;
		return c0*p + c1*q;
    }
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> nlerp(const T& t,const Quaternion<T>& p, const Quaternion<T>& q) {
	Quaternion<T> result;		

    if (p.dot(q) < 0.0f)
		result = t*(q + p) - p;
	else
		result = t*(q - p) + p;
	result.normalise();
	return result;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> squad(const T& t, const Quaternion<T>& q0, const Quaternion<T>& a0, const Quaternion<T>& a1, const Quaternion<T>& q1) {
    Quaternion<T> slerpP = slerp(t,q0,q1);
    Quaternion<T> slerpQ = slerp(t,a0,a1);
    T s = T(2.0)*t*(T(1.0)-t);
    return slerp(s,slerpP,slerpQ);
}

}; };

#endif
