/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __MATH_TRIGONOMETRICPOLYNOMIAL_INL
#define __MATH_TRIGONOMETRICPOLYNOMIAL_INL

namespace magnum { namespace math {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T>::TrigonometricPolynomial(unsigned int d) {
	m_Degree = d;
	m_Coefficients = new T[2*d+1];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T>::TrigonometricPolynomial(const TrigonometricPolynomial<S,T>& p) {
	m_Degree = p.m_Degree;
	m_Coefficients = new T[2*m_Degree+1];
	CopyElements(m_Coefficients, p.m_Coefficients, 2*m_Degree+1);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T>::~TrigonometricPolynomial() {
	delete[] m_Coefficients;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline unsigned int TrigonometricPolynomial<S,T>::getDegree() const {
	return m_Degree;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
unsigned int TrigonometricPolynomial<S,T>::getEffectiveDegree() const {
	unsigned int i;
	for (i = 2*m_Degree; i > 0; i--)
		if (m_Coefficients[i] != T(0))
			return i;
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline void TrigonometricPolynomial<S,T>::setDegree(unsigned int d) {
	if (m_Degree == d)
		return;
	delete[] m_Coefficients;
	m_Degree = d;
	m_Coefficients = new T[2*m_Degree + 1];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T>& TrigonometricPolynomial<S,T>::operator=(const TrigonometricPolynomial<S,T>& p) {
	setDegree(p.m_Degree);
	CopyElements(m_Coefficients, p.m_Coefficients, 2*m_Degree+1);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline T TrigonometricPolynomial<S,T>::operator() (const S& x) const {
	T f = S(0.5) * m_Coefficients[0];
	for (unsigned int i = 1; i <= m_Degree; i++) {
		f += S(cos(S(i)*x)) * m_Coefficients[2*i-1] + S(sin(S(i)*x)) * m_Coefficients[2*i];
	}

	return f;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T> TrigonometricPolynomial<S,T>::operator+ (const TrigonometricPolynomial<S,T>& p) const {
	unsigned int i;
	if (m_Degree >= p.m_Degree) {
		TrigonometricPolynomial<S,T> poly(m_Degree);
		for (i = 0; i <= 2*p.m_Degree; i++)
			poly.m_Coefficients[i] = m_Coefficients[i] + p.m_Coefficients[i];
		for (i = 2*p.m_Degree + 1; i <= 2*m_Degree; i++)
			poly.m_Coefficients[i] = m_Coefficients[i];
		return poly;
	} else {
		TrigonometricPolynomial<S,T> poly(p.m_Degree);
		for (i = 0; i <= 2*m_Degree; i++)
			poly.m_Coefficients[i] = m_Coefficients[i] + p.m_Coefficients[i];
		for (i = 2*m_Degree + 1; i <= 2*p.m_Degree; i++)
			poly.m_Coefficients[i] = p.m_Coefficients[i];
		return poly;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T> TrigonometricPolynomial<S,T>::operator- (const TrigonometricPolynomial<S,T>& p) const {
	unsigned int i;
	if (m_Degree >= p.m_Degree) {
		TrigonometricPolynomial<S,T> poly(m_Degree);
		for (i = 0; i <= 2*p.m_Degree; i++)
			poly.m_Coefficients[i] = m_Coefficients[i] - p.m_Coefficients[i];
		for (i = 2*p.m_Degree + 1; i <= 2*m_Degree; i++)
			poly.m_Coefficients[i] = m_Coefficients[i];
		return poly;
	} else {
		TrigonometricPolynomial<S,T> poly(p.m_Degree);
		for (i = 0; i <= 2*m_Degree; i++)
			poly.m_Coefficients[i] = m_Coefficients[i] - p.m_Coefficients[i];
		for (i = 2*m_Degree + 1; i <= 2*p.m_Degree; i++)
			poly.m_Coefficients[i] = - p.m_Coefficients[i];
		return poly;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T> TrigonometricPolynomial<S,T>::operator+ (const T& t) const {
	TrigonometricPolynomial<S,T> poly(*this);
	poly.m_Coefficients[0] += t;
	return poly;
} 


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T> TrigonometricPolynomial<S,T>::operator- (const T& t) const {
	TrigonometricPolynomial<S,T> poly(*this);
	poly.m_Coefficients[0] -= t;
	return poly;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T> TrigonometricPolynomial<S,T>::operator* (const S& s) const {
	TrigonometricPolynomial<S,T> poly(m_Degree);
	for (unsigned int i = 0; i <= 2*m_Degree; i++)
		poly.m_Coefficients[i] = m_Coefficients[i] * s;
	return poly;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T> TrigonometricPolynomial<S,T>::operator/ (const S& s) const {
	TrigonometricPolynomial<S,T> poly(m_Degree);
	for (unsigned int i = 0; i <= 2*m_Degree; i++)
		poly.m_Coefficients[i] = m_Coefficients[i] / s;
	return poly;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T> TrigonometricPolynomial<S,T>::operator- () const {
	TrigonometricPolynomial<S,T> poly(m_Degree);
	for (unsigned int i = 0; i <= 2*m_Degree; i++)
		poly.m_Coefficients[i] = -m_Coefficients[i];
	return poly;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
TrigonometricPolynomial<S,T>& TrigonometricPolynomial<S,T>::operator += (const TrigonometricPolynomial<S,T>& p) {
	unsigned int i;
	if (m_Degree >= p.m_Degree) {
		for (i = 0; i <= 2*p.m_Degree; i++)
			m_Coefficients[i] += p.m_Coefficients[i];
	} else {
		T* poly = new T[p.m_Degree];
		for (i = 0; i <= 2*m_Degree; i++)
			poly[i] = m_Coefficients[i] + p.m_Coefficients[i];
		for (i = 2*m_Degree + 1; i <= 2*p.m_Degree; i++)
			poly[i] = p.m_Coefficients[i];
		delete[] m_Coefficients;
		m_Coefficients = poly;
	}

	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
TrigonometricPolynomial<S,T>& TrigonometricPolynomial<S,T>::operator -= (const TrigonometricPolynomial<S,T>& p) {
	unsigned int i;
	if (m_Degree >= p.m_Degree) {
		for (i = 0; i <= 2*p.m_Degree; i++)
			m_Coefficients[i] -= p.m_Coefficients[i];
	} else {
		T* poly = new T[p.m_Degree];
		for (i = 0; i <= 2*m_Degree; i++)
			poly[i] = m_Coefficients[i] - p.m_Coefficients[i];
		for (i = 2*m_Degree + 1; i <= 2*p.m_Degree; i++)
			poly[i] = -p.m_Coefficients[i];
		delete[] m_Coefficients;
		m_Coefficients = poly;
	}

	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T>& TrigonometricPolynomial<S,T>::operator += (const T& t) {
	m_Coefficients[0] += t;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T>& TrigonometricPolynomial<S,T>::operator -= (const T& t) {
	m_Coefficients[0] -= t;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T>& TrigonometricPolynomial<S,T>::operator *= (const S& s) {
	for (unsigned int i = 0; i <= 2*m_Degree; i++)
		m_Coefficients[i] *= s;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T>& TrigonometricPolynomial<S,T>::operator /= (const S& s) {
	for (unsigned int i = 0; i <= 2*m_Degree; i++)
		m_Coefficients[i] /= s;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
TrigonometricPolynomial<S,T> TrigonometricPolynomial<S,T>::derive() const {
	if (m_Degree == 0) {
		TrigonometricPolynomial<S,T> dv(0);
		dv[0] = T(0);
		return dv;
	}

	TrigonometricPolynomial<S,T> dv(m_Degree - 1);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T>::operator const T*() const {
	return m_Coefficients;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T>::operator T*() {
	return m_Coefficients;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
bool TrigonometricPolynomial<S,T>::operator==(const TrigonometricPolynomial<S,T>& p) const {
	unsigned int i;
	if (m_Degree <= p.m_Degree) {
		for (i = 0; i <= 2*m_Degree; i++) {
			if (m_Coefficients[i] != p.m_Coefficients[i])
				return false;
		}
		for (; i <= 2*p.m_Degree; i++) {
			if (p.m_Coefficients[i] != T(0))
				return false;
		}
	} else {
		for (i = 0; i <= 2*p.m_Degree; i++) {
			if (m_Coefficients[i] != p.m_Coefficients[i])
				return false;
		}
		for (; i <= 2*m_Degree; i++) {
			if (m_Coefficients[i] != T(0))
				return false;
		}
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
bool TrigonometricPolynomial<S,T>::operator!=(const TrigonometricPolynomial<S,T>& p) const {
	unsigned int i;
	if (m_Degree <= p.m_Degree) {
		for (i = 0; i <= 2*m_Degree; i++) {
			if (m_Coefficients[i] != p.m_Coefficients[i])
				return true;
		}
		for (; i <= 2*p.m_Degree; i++) {
			if (p.m_Coefficients[i] != T(0))
				return true;
		}
	} else {
		for (i = 0; i <= 2*p.m_Degree; i++) {
			if (m_Coefficients[i] != p.m_Coefficients[i])
				return true;
		}
		for (; i <= 2*m_Degree; i++) {
			if (m_Coefficients[i] != T(0))
				return true;
		}
	}

	return false;
}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class S,class T> 
inline TrigonometricPolynomial<S,T> operator*(const S& s,TrigonometricPolynomial<S,T>& p) {
	TrigonometricPolynomial<S,T> poly(p.getDegree());
	for (unsigned int i = 0; i <= 2*p.getDegree(); i++)
		poly[i] = s * p[i];
	return poly;
}


}; };

#endif
