/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __MEDIA_NODE_H
#define __MEDIA_NODE_H

#include "object.h"
#include "system.thread.h"
#include "system.mutex.h"
#include "container.chain.h"
#include "media.looper.h"
#include "media.pin.h"

namespace magnum { namespace media {
using namespace magnum::container;
using magnum::object::Object;
using magnum::system::Mutex;
using magnum::system::Thread;
using magnum::container::Chain;
using magnum::types::String;

class Device;
class Buffer;
class Graph;
class Format;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
class Node : public Looper, public Chainable<Node> {
DeclareClass(Node, Looper);
friend class Graph;
friend class Pin;
friend class InputPin;
friend class OutputPin;
friend class Buffer;
public:
	static const char DefaultInput[];
	static const char DefaultOutput[];
	static const char DefaultFilter[];
	static const char DefaultNode[];

	enum Type {
		Input = 1,
		Producer = 1,
		Output = 2,
		Consumer = 2,
		Filter = 3
	};

	enum State {
		Unknown = 0,
		Running,
		Stopped,
		Paused,
		Starting,
		Stopping,
	};

public:
	Node(const char* name, Type t);
	Node(Device* dev, const char* name, Type t);
	~Node();

	InputPin*			getInputPin(const char* which = NULL);
	OutputPin*			getOutputPin(const char* which = NULL);
	inline const InputPinList& getInputPins() const;
	inline const OutputPinList& getOutputPins() const;

	inline Device*		getDevice() const;
	inline Graph*		getGraph() const;
	inline String		getName() const;
	inline State		getState() const;
	
	bool				start();
	bool				stop();
	bool				pause();
	bool				resume();
	bool				reset();

protected:
	// Looper methods
	virtual bool		onLoop();
	virtual void		onHandleMessage(const Message& );

	// (Dis)Connection methods
	virtual bool		onConnect(Pin* src);
	virtual bool		onDisconnect(Pin* src);

	// Buffer routines
	virtual void		onDestroyBuffer(Buffer* );
	virtual bool		onReceiveBuffer(InputPin* ,Buffer* );
	virtual bool		onRequestBuffer(OutputPin* ,Buffer* );
	virtual void		onEndOfInput(InputPin* );
	virtual void		onEndOfOutput(OutputPin* );

	// Start/Stop/Pause - this is system and device specific
	virtual bool		onStart();
	virtual bool		onStop();
	virtual bool		onPause();
	virtual bool		onResume();
	virtual bool		onReset();
	
	// Format methods
	virtual bool		onSetFormat(Pin* ,const Format& );
	virtual bool		onAdjustFormat(Pin* ,Format& );

	void				endOfInput(InputPin* );
	void				endOfOutput(OutputPin* );
	inline void			requestBuffer(InputPin* );
	inline void			requestBuffer(InputPin* ,Buffer* );
	inline void			requestBuffer(OutputPin* );
	inline void			requestBuffer(OutputPin* ,Buffer* );
	bool				postBuffer(OutputPin* ,Buffer* );
	bool				sendBuffer(OutputPin* ,Buffer* );
	
private:
	bool				waitForState(State );

private:
	Graph*				m_Graph;
	Device*				m_Device;
	String				m_Name;
	Chain<InputPin>		m_InputPins;
	Chain<OutputPin>	m_OutputPins;
	InputPin*			m_DefaultInput;
	OutputPin*			m_DefaultOutput;
	State				m_State;
	Type				m_Type;
};


}; };

#endif
