/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __PHYSICS_BODY_INL
#define __PHYSICS_BODY_INL

namespace magnum { namespace physics {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Body::Body(Domain* l) {
	m_Body = dBodyCreate(l->m_World);
	l->m_Bodies.insert(this);
	release();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Body::~Body() {
	dBodyDestroy(m_Body);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setPosition(float x,float y,float z) {
	dBodySetPosition(m_Body, x, y, z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setPosition(const Vector3<float>& v) {
	dBodySetPosition(m_Body, v.x, v.y, v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Vector3<float> Body::getPosition() const {
	const float* pos = dBodyGetPosition(m_Body);
	return Vector3<float>(pos[0], pos[1], pos[2]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setRotation(const Matrix3<float>& mtx) {
	dBodySetRotation(m_Body, mtx[0]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Matrix3<float> Body::getRotation() const {
	return Matrix3<float>(dBodyGetRotation(m_Body));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setTransform(const Matrix4<float>& mtx) {
	float tmp[3][3];
	tmp[0][0] = mtx[0][0];
	tmp[0][1] = mtx[0][1];
	tmp[0][2] = mtx[0][2];
	tmp[1][0] = mtx[1][0];
	tmp[1][1] = mtx[1][1];
	tmp[1][2] = mtx[1][2];
	tmp[2][0] = mtx[2][0];
	tmp[2][1] = mtx[2][1];
	tmp[2][2] = mtx[2][2];
	dBodySetRotation(m_Body, &tmp[0][0]);
	dBodySetPosition(m_Body, mtx[0][3], mtx[1][3], mtx[2][3]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Matrix4<float> Body::getTransform() const {
	const float* pos = dBodyGetPosition(m_Body);
	const float* rot = dBodyGetRotation(m_Body);

	Matrix4<float> mat;
	mat[0][0] = rot[0];
	mat[0][1] = rot[1];
	mat[0][2] = rot[2];
	mat[1][0] = rot[3];
	mat[1][1] = rot[4];
	mat[1][2] = rot[5];
	mat[2][0] = rot[6];
	mat[2][1] = rot[7];
	mat[2][2] = rot[8];
	mat[0][3] = pos[0];
	mat[1][3] = pos[0];
	mat[2][3] = pos[0];
	mat[3][3] = 1.0f;
	mat[3][0] = mat[3][1] = mat[3][2] = 0.0f;
	return mat;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setLinearVelocity(float x,float y,float z) {
	dBodySetLinearVel(m_Body, x, y, z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setLinearVelocity(const Vector3<float>& v) {
	dBodySetLinearVel(m_Body, v.x, v.y, v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Vector3<float> Body::getLinearVelocity() const {
	const float* vel = dBodyGetLinearVel(m_Body);
	return Vector3<float>(vel[0], vel[1], vel[2]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setAngularVelocity(float x,float y,float z) {
	dBodySetAngularVel(m_Body, x, y, z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setAngularVelocity(const Vector3<float>& v) {
	dBodySetAngularVel(m_Body, v.x, v.y, v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Vector3<float> Body::getAngularVelocity() const {
	const float* vel = dBodyGetAngularVel(m_Body);
	return Vector3<float>(vel[0], vel[1], vel[2]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setMass(const Mass& ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Mass Body::getMass() const {
	return Mass();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setForce(float x,float y,float z) {
	dBodySetForce(m_Body, x, y, z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setForce(const Vector3<float>& v) {
	dBodySetForce(m_Body, v.x, v.y, v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::addForce(float x,float y,float z) {
	dBodyAddForce(m_Body, x, y, z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::addForce(const Vector3<float>& v) {
	dBodyAddForce(m_Body, v.x, v.y, v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Vector3<float> Body::getForce() const {
	const float* pos = dBodyGetForce(m_Body);
	return Vector3<float>(pos[0], pos[1], pos[2]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setTorque(float x,float y,float z) {
	dBodySetTorque(m_Body, x, y, z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setTorque(const Vector3<float>& v) {
	dBodySetTorque(m_Body, v.x, v.y, v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::addTorque(float x,float y,float z) {
	dBodyAddTorque(m_Body, x, y, z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::addTorque(const Vector3<float>& v) {
	dBodyAddTorque(m_Body, v.x, v.y, v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Vector3<float> Body::getTorque() const {
	const float* pos = dBodyGetTorque(m_Body);
	return Vector3<float>(pos[0], pos[1], pos[2]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::enable() {
	dBodyEnable(m_Body);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::disable() {
	dBodyDisable(m_Body);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool Body::isEnabled() const {
	return (dBodyIsEnabled(m_Body) == 1);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void Body::setGravity(bool b) {
	dBodySetGravityMode(m_Body, b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline bool Body::getGravity() const {
	return (dBodyGetGravityMode(m_Body) == 1);
}

}; };

#endif
