/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __PHYSICS_RELAXATOR_INL
#define __PHYSICS_RELAXATOR_INL

namespace magnum { namespace physics {


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Relaxator<T,dim>::Relaxator() {
	m_Gravity.null();
	m_Constant = T(0);
	m_Damping = T(0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Relaxator<T,dim>::clear() {
	m_Weights.clear();
	m_Springs.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Relaxator<T,dim>::setSize(unsigned int i) {
	m_Weights.clear();
	m_Springs.clear();
	m_Weights.setSize(i);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline unsigned int Relaxator<T,dim>::getSize() const {
	return m_Weights.getSize();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Relaxator<T,dim>::isEmpty() const {
	return m_Weights.isEmpty();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Weight<T,dim>& Relaxator<T,dim>::operator[](unsigned int i) {
	return m_Weights[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Weight<T,dim>& Relaxator<T,dim>::operator[](unsigned int i) const {
	return m_Weights[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline typename Relaxator<T,dim>::Iterator Relaxator<T,dim>::begin() const {
	return m_Weights.begin();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline typename Relaxator<T,dim>::Iterator Relaxator<T,dim>::end() const {
	return m_Weights.end();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline typename Relaxator<T,dim>::Iterator Relaxator<T,dim>::at(unsigned int i) const {
	return m_Weights.at(i);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Spring<T,dim>* Relaxator<T,dim>::connect(unsigned int i,unsigned int j,T c) {
	ASSERT(!isConnected(i,j));
	Spring<T,dim>* spring = m_Springs.append();
	spring->m_Constant = c;
	spring->m_Left = &m_Weights[i];
	spring->m_Right = &m_Weights[j];
	spring->m_LeftNext = m_Weights[i].m_Springs;
	spring->m_RightNext = m_Weights[j].m_Springs;
	m_Weights[i].m_Springs = m_Weights[j].m_Springs = spring;
	return spring;
} 


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Relaxator<T,dim>::disconnect(unsigned int i,unsigned int j) {
	Weight<T,dim>* ba = &m_Weights[i];
	Weight<T,dim>* bb = &m_Weights[j];

	// Find spring in body A
	Spring<T,dim>* sprA = ba->m_Springs;
	Spring<T,dim>* lastA = NULL;
	Spring<T,dim>* nextA;
	while(true) {
		if (sprA == NULL)
			return;
		if (sprA->m_Left == ba)
			nextA = sprA->m_LeftNext;
		else
			nextA = sprA->m_RightNext;
		if ((sprA->m_Left == bb) || (sprA->m_Right == bb))
			break;
		lastA = sprA;
		sprA = nextA;
	}

	// Find spring in body B
	Spring<T,dim>* sprB = bb->m_Springs;
	Spring<T,dim>* lastB = NULL;
	Spring<T,dim>* nextB;
	while(true) {
		ASSERT(sprB != NULL);
		if (sprB->m_Left == bb)
			nextB = sprB->m_LeftNext;
		else
			nextB = sprB->m_RightNext;
		if ((sprB->m_Left == ba) || (sprB->m_Right == ba))
			break;
		lastB = sprB;
		sprB = nextB;
	}

	ASSERT(sprA == sprB);
	if (lastA == NULL) {
		ba->m_Springs = nextA;
	} else {
		if (lastA->m_Left == ba)
			lastA->m_LeftNext = nextA;
		else
			lastA->m_RightNext = nextA;
	}

	if (lastB == NULL) {
		bb->m_Springs = nextB;
	} else {
		if (lastB->m_Left == bb)
			lastB->m_LeftNext = nextB;
		else
			lastB->m_RightNext = nextB;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Relaxator<T,dim>::isConnected(unsigned int i,unsigned int j) const {
	const Weight<T,dim>* ba = &m_Weights[i];
	const Weight<T,dim>* bb = &m_Weights[j];

	// Find spring in body A
	Spring<T,dim>* spr = ba->m_Springs;
	while(true) {
		if (spr == NULL)
			return false;
		if ((spr->m_Left == bb) || (spr->m_Right == bb))
			return true;
		if (spr->m_Left == ba)
			spr = spr->m_LeftNext;
		else
			spr = spr->m_RightNext;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Relaxator<T,dim>::setGravity(const Vector<T,dim>& g) {
	m_Gravity = g;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Vector<T,dim>& Relaxator<T,dim>::getGravity() const {
	return m_Gravity;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Relaxator<T,dim>::setConstant(const T& c) {
	m_Constant = c;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const T& Relaxator<T,dim>::getConstant() const {
	return m_Constant;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Relaxator<T,dim>::setDamping(const T& c) {
	m_Damping = c;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const T& Relaxator<T,dim>::getDamping() const {
	return m_Damping;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Relaxator<T,dim>::reset() {
	// Reset all springs
	List<Spring<T,dim> >::Iterator iter;
	for (iter = m_Springs.begin(); iter != m_Springs.end(); ++iter) {
		T length = (iter->m_Left->m_Position - iter->m_Right->m_Position).length();
		iter->m_Length = iter->m_NaturalLength = length;
		iter->m_Force.null();
	}
	
	// Reset all weights
	Iterator witer;
	for (witer = m_Weights.begin(); witer != m_Weights.end(); ++witer) {
		witer->m_Velocity.null();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Relaxator<T,dim>::relax(T time) {
	relax(defaultFunctor, time);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
template<class F>
void Relaxator<T,dim>::relax(F& functor, T time) {
	List<Spring<T,dim> >::Iterator siter;

	if (m_Constant != T(0)) {
		// Process all springs
		T constime = time*m_Constant;
		for (siter = m_Springs.begin(); siter != m_Springs.end(); ++siter) {
			// Calculate force between the two ends
			Vector<T,dim> diff = siter->m_Left->m_Position - siter->m_Right->m_Position;
			T length = diff.length();
			T factor = constime * (length - siter->m_NaturalLength) / length;
			Vector<T,dim> force =  factor * diff;
			
			// Apply force to both ends
			siter->m_Left->m_Velocity -= force;
			siter->m_Right->m_Velocity += force;
	
			// Save length and force
			siter->m_Length = length;
			siter->m_Force = force;
		}
	} else {
		// Process all springs
		for (siter = m_Springs.begin(); siter != m_Springs.end(); ++siter) {
			// Calculate force between the two ends
			Vector<T,dim> diff = siter->m_Left->m_Position - siter->m_Right->m_Position;
			T length = diff.length();
			Vector<T,dim> force = time * siter->m_Constant * (siter->m_NaturalLength - length) / length * diff;
			
			// Apply force to both ends
			Weight<T,dim>* w;
			w = siter->m_Left;
			w->m_Velocity += force / w->m_Mass;
			w = siter->m_Right;
			w->m_Velocity -= force / w->m_Mass;
	
			// Save length and force
			siter->m_Length = length;
			siter->m_Force = force;
		}
	}
 
	// Process all weights
	Array<Weight<T,dim> >::Iterator witer;
	Vector<T,dim> gravity = time * m_Gravity;
	T damp = exp(-time*m_Damping);
	for (witer = m_Weights.begin(); witer != m_Weights.end(); ++witer) {
		if (witer->isFixed())
			continue;
		witer->m_Velocity *= damp;
		witer->m_Velocity += gravity;
		functor((Weight<T,dim>*)witer, time);
		// witer->m_Position += time * witer->m_Velocity;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Relaxator<T,dim>::defaultFunctor(Weight<T,dim>* w, const T& time) {
	w->m_Position += time * w->m_Velocity;
}

}; };

#endif
