/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#ifndef __SYSTEM_TIMESTAMP_INL
#define __SYSTEM_TIMESTAMP_INL


namespace magnum { namespace system {

/*----------------------------------------------------------------------------
Description:
	Constructs a new Timestamp object and sets it to the current System time
	in UTC time format.
----------------------------------------------------------------------------*/
inline Timestamp::Timestamp() : Object() {
	getCurrentTime();
}


/*----------------------------------------------------------------------------
Description:
	Constructs a new Timestamp object and sets the number of 100-nanosecond 
	intervals since January 1, 1601 to the value passed

Parameters:
	time - number of 100-nanosecond intervals since January 1, 1601
----------------------------------------------------------------------------*/
inline Timestamp::Timestamp(const uint64 time) : Object() {
	m_Time = time;
}


/*----------------------------------------------------------------------------
Description:
	Sets the Timestamp object to the passed parameter.

Parameters:
	time - number of 100-nanosecond intervals since January 1, 1601
----------------------------------------------------------------------------*/
inline void Timestamp::setTime(const uint64 time) {
	m_Time = time;
}


/*----------------------------------------------------------------------------
Description:
	Returns the UTC time represented by this Timestamp object.

Returns:
	number of 100-nanosecond intervals since January 1, 1601
----------------------------------------------------------------------------*/
inline uint64 Timestamp::getTime() const {
	return m_Time;
}


/*----------------------------------------------------------------------------
Description:
	Returns *true* when the current Timestamp object represents a time
	before the time represented by the Timestamp object passed.

Parameters:
	time - Timestamp object to compare

Returns:
	*true* when the current timestamp is before time.
----------------------------------------------------------------------------*/
inline bool Timestamp::isBefore(const Timestamp& time) const {
	return m_Time < time.m_Time;
}


/*----------------------------------------------------------------------------
Description:
	Returns *true* when the current Timestamp object represents a time
	after the time represented by the Timestamp object passed.

Parameters:
	time - Timestamp object to compare

Returns:
	*true* when the current timestamp is after time.
----------------------------------------------------------------------------*/
inline bool Timestamp::isAfter(const Timestamp& time) const {
	return m_Time > time.m_Time;
}


/*----------------------------------------------------------------------------
Description:
	Returns *true* when the current Timestamp object represents a time
	before the time represented by the Timestamp object passed.

Parameters:
	time - Timestamp object to compare

Returns:
	*true* when the current timestamp is before time.
----------------------------------------------------------------------------*/
inline bool Timestamp::operator< (const Timestamp& time) const {
	return m_Time < time.m_Time;
}


/*----------------------------------------------------------------------------
Description:
	Returns *true* when the current Timestamp object represents a time
	after the time represented by the Timestamp object passed.

Parameters:
	time - Timestamp object to compare

Returns:
	*true* when the current timestamp is after time.
----------------------------------------------------------------------------*/
inline bool Timestamp::operator> (const Timestamp& time) const {
	return m_Time > time.m_Time;
}


/*----------------------------------------------------------------------------
Description:
	Returns *true* when the current Timestamp object represents a time
	before or equal to the time represented by the Timestamp object passed.

Parameters:
	time - Timestamp object to compare

Returns:
	*true* when the current timestamp is before or equal to time.
----------------------------------------------------------------------------*/
inline bool Timestamp::operator<= (const Timestamp& time) const {
	return m_Time <= time.m_Time;
}


/*----------------------------------------------------------------------------
Description:
	Returns *true* when the current Timestamp object represents a time
	after or equal to the time represented by the Timestamp object passed.

Parameters:
	time - Timestamp object to compare

Returns:
	*true* when the current timestamp is after or equal to time.
----------------------------------------------------------------------------*/
inline bool Timestamp::operator>= (const Timestamp& time) const {
	return m_Time >= time.m_Time;
}


/*----------------------------------------------------------------------------
Description:
	Compares two time-stamps

Returns:
	*true* when this time equals to the time represemted by time
----------------------------------------------------------------------------*/
inline bool Timestamp::operator== (const Timestamp& time) const {
	return m_Time == time.m_Time;
}


/*----------------------------------------------------------------------------
Description:
	Compares two time-stamps

Returns:
	*true* when this time does not equal to the time represemted by time
----------------------------------------------------------------------------*/
inline bool Timestamp::operator!= (const Timestamp& time) const {
	return m_Time != time.m_Time;
}


/*----------------------------------------------------------------------------
Description:
	Return the difference between two time-stamp objects.

Parameters:
	time - timestamp to subtract from the current time stamp

Returns:
	The signed difference between both timestamps.
----------------------------------------------------------------------------*/
inline int64 Timestamp::operator- (const Timestamp& time) const {
	return m_Time - time.m_Time;
}


/*----------------------------------------------------------------------------
Description:
	Adds an interval to the current time.

Parameters:
	time - number of 100 nanosecond intervals to add

Returns:
	A new time-stamp object with the interval added
----------------------------------------------------------------------------*/
inline Timestamp Timestamp::operator+ (const int64& time) const {
	return Timestamp(m_Time + time);
}

}; };
#endif
