/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "application.messenger.h"

#include "application.message.inl"
#include "application.messenger.inl"
#include "container.chain.inl"
#include "container.iterator.inl"
#include "container.allocator.inl"
#include "container.queue.inl"
#include "system.thread.inl"
#include "object.inl"


namespace magnum { namespace application {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Messenger::Messenger() {
	m_Thread = NULL;
	m_Handlers.insert(this);
	m_Messenger = this;
	m_PreferredHandler = this;
	release();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Messenger::~Messenger() {
	refer();
	m_Handlers.remove(this);
	m_Messenger = NULL;

	while(!m_Handlers.isEmpty()) {
		Chain<Handler>::Iterator iter = m_Handlers.begin();
		m_Handlers.remove(&(*iter));
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Messenger::clear() {
	m_HandlerMutex.lock();
	refer();
	while(!m_Handlers.isEmpty()) {
		Chain<Handler>::Iterator iter = m_Handlers.begin();
		iter->m_Messenger = NULL;
		m_Handlers.remove(&(*iter));
	}
	m_Handlers.insert(this);
	m_Messenger = this;
	m_PreferredHandler = this;
	release();
	m_HandlerMutex.unlock();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Messenger::insert(Handler* hnd) {
	ASSERT(hnd != this);
	ASSERT(hnd->m_Messenger == NULL);
	if (hnd == this)
		return false;
	m_HandlerMutex.lock();
	m_Handlers.insert(hnd);
	m_HandlerMutex.unlock();
	hnd->m_Messenger = this;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Messenger::remove(Handler* hnd) {
	ASSERT(hnd != this);
	ASSERT(hnd->m_Messenger == this);
	if (hnd == this)
		return false;
	hnd->m_Messenger = NULL;
	m_HandlerMutex.lock();
	m_Handlers.remove(hnd);
	if (m_PreferredHandler == hnd)
		m_PreferredHandler = this;
	m_HandlerMutex.unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Messenger::postMessage(const Message& m) {
	Handler* hnd = m.getTarget();
	
	// Check if handler belongs to this messenger
	if (hnd == NULL) {
		hnd = m_PreferredHandler;
	} else if (hnd->m_Messenger != this) {
		ASSERT(hnd->m_Messenger != NULL);
		if (hnd->m_Messenger->m_Thread == NULL)
			return false;
		return hnd->m_Messenger->onPostMessage(new Message(m));
	}

	// Deliver message within this messenger
	if (m_Thread == NULL)
		return false;

	Message *msg = new Message(m);
	msg->setTarget(hnd);
	return onPostMessage(msg);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Messenger::postMessage(int cmd) {
	if (m_Thread == NULL)
		return false;

	Message *msg = new Message(m_PreferredHandler, cmd);
	return onPostMessage(msg);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Messenger::postMessage(Handler* target, const Message& m) {
	ASSERT(target != NULL);
	ASSERT(target->m_Messenger != NULL);
	ASSERT((m.getTarget() == NULL) || (m.getTarget() == target))
	
	// Get appropriate messenger
	Messenger* msgr = target->m_Messenger;

	// Deliver message to correct messenger
	if (msgr->m_Thread == NULL)
		return false;

	Message* msg = new Message(m);
	msg->setTarget(target);
	return msgr->onPostMessage(msg);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Messenger::postMessage(Handler* target, int cmd) {
	ASSERT(target != NULL);
	ASSERT(target->m_Messenger != NULL);
	
	// Get appropriate messenger
	Messenger* msgr = target->m_Messenger;

	// Deliver message within this messenger
	if (msgr->m_Thread == NULL)
		return false;

	Message *msg = new Message(target, cmd);
	return msgr->onPostMessage(msg);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Messenger::sendMessage(const Message& m) {
	Handler* hnd = m.getTarget();
	Messenger* msgr;

	// Get appropriate messenger and handler
	if (hnd == NULL) {
		hnd = m_PreferredHandler;
		msgr = this;
	} else {
		ASSERT(hnd->m_Messenger != NULL);
		msgr = hnd->m_Messenger;
	}

	// Deliver message within this messenger
	if (msgr->m_Thread == Thread::getCurrent()) {
		hnd->handleMessage(m);
		return true;
	} else {
		if (msgr->m_Thread == NULL)
			return false;
		Message *msg = new Message(m);
		msg->setTarget(hnd);
		return msgr->onSendMessage(msg);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Messenger::sendMessage(int cmd) {
	if (m_Thread == NULL)
		return false;

	if (m_Thread == Thread::getCurrent()) {
		Message msg(this, cmd);
		handleMessage(msg);
		return true;
	} else {
		Message *msg = new Message(m_PreferredHandler, cmd);
		return onSendMessage(msg);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Messenger::sendMessage(Handler* target, const Message& m) {
	ASSERT(target != NULL);
	ASSERT(target->m_Messenger != NULL);
	ASSERT((m.getTarget() == NULL) || (m.getTarget() == target))

	// Get appropriate messenger
	Messenger* msgr = target->m_Messenger;

	// Deliver message within this messenger
	if (msgr->m_Thread == Thread::getCurrent()) {
		target->handleMessage(m);
		return true;
	} else {
		if (msgr->m_Thread == NULL)
			return false;
		Message *msg = new Message(m);
		msg->setTarget(target);
		return msgr->onSendMessage(msg);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Messenger::sendMessage(Handler* target, int cmd) {
	ASSERT(target != NULL);
	ASSERT(target->m_Messenger != NULL);

	// Get appropriate messenger
	Messenger* msgr = target->m_Messenger;

	// Deliver message within this messenger
	if (msgr->m_Thread == Thread::getCurrent()) {
		Message msg(target, cmd);
		target->handleMessage(msg);
		return true;
	} else {
		if (msgr->m_Thread == NULL)
			return false;
		Message *msg = new Message(target, cmd);
		return msgr->onSendMessage(msg);
	}
}



}; };
