/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "application.performer.h"
#include "application.task.h"

#include "object.inl"
#include "base.string.inl"
#include "system.thread.inl"
#include "system.mutex.inl"
#include "system.barrier.inl"
#include "system.signal.inl"
#include "application.performer.inl"
#include "application.task.inl"

namespace magnum { namespace application {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Performer::Performer() : m_PerformerThread(this) {
	m_Thread = NULL;
	m_Barrier.setWaiters(1);
	m_SyncBarrier.setWaiters(1);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Performer::~Performer() {
	if (m_Thread != NULL) {
		stop();
		join();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Performer::insert(Task* t) {
	if (t->getPerformer() != NULL)
		return false;

	m_ChainMutex.lock();
	m_Tasks.insert(t);
	m_Barrier.setWaiters(m_Tasks.getSize() + 1);
	m_SyncBarrier.setWaiters(m_Tasks.getSize() + 1);
	m_ChainMutex.unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Performer::remove(Task* t) {
	if (t->getPerformer() != this)
		return;
		
	m_ChainMutex.lock();
	m_Tasks.remove(t);
	m_Barrier.setWaiters(m_Tasks.getSize() + 1);
	m_SyncBarrier.setWaiters(m_Tasks.getSize() + 1);
	m_ChainMutex.unlock();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Performer::clear() {
	m_ChainMutex.lock();
	m_Tasks.clear();
	m_Barrier.setWaiters(m_Tasks.getSize() + 1);
	m_SyncBarrier.setWaiters(m_Tasks.getSize() + 1);
	m_ChainMutex.unlock();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Performer::loop() {
	if (m_Thread != NULL)
		return false;
		
	m_Thread = Thread::getCurrent();
	m_StopSignal.reset();
	m_PauseSignal.raise();
	
	onStart();
	
	// Start all Threads
	for (Iterator iter = m_Tasks.begin(); iter != m_Tasks.end(); ++iter) {
		iter->start();
	}	
	
	do {
		// Wait for all threads to reach barrier
		m_Barrier.wait();

		// synchronize all
		for (Iterator iter = m_Tasks.begin(); iter != m_Tasks.end(); ++iter) {
			iter->onSync();
		}
		
		// check for pause mode
		m_PauseSignal.wait();
		
		onLoop();
		
		// Continue
		m_SyncBarrier.wait();
	} while(!m_StopSignal.isRaised());
	
	// Wait for all threads to stop
	for (Iterator iter = m_Tasks.begin(); iter != m_Tasks.end(); ++iter) {
		iter->join();
	}
	
	onEnd();
	
	m_Thread = NULL;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Performer::step() {
	if (m_Thread != NULL)
		return false;

	m_Thread = Thread::getCurrent();

	// Wait for all threads to reach barrier
	m_Barrier.wait();

	// synchronize all
	for (Iterator iter = m_Tasks.begin(); iter != m_Tasks.end(); ++iter) {
		iter->onSync();
	}

	onLoop();

	// Continue
	m_SyncBarrier.wait();

	m_Thread = NULL;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Performer::start() {
	if (m_Thread != NULL)
		return false;
		
	m_PerformerThread.start();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Performer::stop() {
	m_StopSignal.raise();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Performer::join() {
	if (m_Thread == NULL)
		return;
	m_Thread->join();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Performer::suspend() {
	m_PauseSignal.reset();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Performer::resume() {
	m_PauseSignal.raise();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool PerformerThread::run() {
	ASSERT(m_Performer != NULL);
	m_Performer->loop();
	return true;
}

}; };
