/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.h"
#include "base.string.h"
#include "object.class.h"
#include "math.h"

#include "object.inl"
#include "base.string.inl"
#include "object.class.inl"

#include <stdio.h>

extern "C" {
int __cdecl _input(FILE *, const unsigned char *, va_list);
};

namespace magnum { namespace types {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
String::String(const String& str) {
	m_Length = str.m_Length;
	if (m_Length > 0) {
		allocBuffer();
		memcpy(m_Buffer, str.m_Buffer, m_Length+1);
	} else {
		m_BufferSize = 0;
		m_Buffer = NULL;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
String::String(const char* str) {
	ASSERT(str != NULL);
	m_Length = (unsigned int)strlen(str);
	if (m_Length > 0) {
		allocBuffer();
		memcpy(m_Buffer, str, m_Length+1);
	} else {
		m_BufferSize = 0;
		m_Buffer = NULL;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
String::String(char str) {
	m_Length = 1;
	allocBuffer();
	m_Buffer[0] = str;
	m_Buffer[1] = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
String::~String() {
	if (m_Buffer) {
		delete[] m_Buffer;
		m_Buffer = NULL;
	}
	m_BufferSize = 0;
	m_Length = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool String::startsWith(const String& str, unsigned int from) const {
	unsigned int len = str.m_Length;
	if ((from + len) > m_Length)
		return false;
	return (memcmp(m_Buffer + from, str.m_Buffer, len) == 0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool String::startsWith(const char*str, unsigned int from) const {
	ASSERT(str != NULL);
	unsigned int len = (unsigned int)strlen(str);
	if ((from + len) > m_Length)
		return false;
	return (memcmp(m_Buffer + from, str, len) == 0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int String::indexOf(const String& str, unsigned int from) const {
	if (from >= m_Length)
		return -1;
	char *pos = strstr(m_Buffer + from, str.m_Buffer);
	return pos == NULL ? -1 : (int)(pos - m_Buffer);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int String::indexOf(const char* str, unsigned int from) const {
	ASSERT(str != NULL);
	if (from >= m_Length)
		return -1;
	char *pos = strstr(m_Buffer + from, str);
	return pos == NULL ? -1 : (int)(pos - m_Buffer);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int String::indexOf(char c, unsigned int from) const {
	if (from >= m_Length)
		return -1;
	char *pos = strchr(m_Buffer + from, c);
	return pos == NULL ? -1 : (int)(pos - m_Buffer);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int String::lastIndexOf(const char* str , unsigned int from) const {
	unsigned int lencmp = (unsigned int)strlen(str);
	unsigned int len = m_Length;

	if ((from + lencmp) > len)
		return -1;

	char *pos = m_Buffer + len - lencmp;
	for (int i = len - from; i > 0; i--) {
		if (memcmp(pos, str, lencmp) == 0)
			return (int)(pos - m_Buffer);
		pos--;
	}

	return -1;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int String::lastIndexOf(char c, unsigned int from) const {
	unsigned int len = m_Length; 
	if (len <= from)
		return -1;

	char *pos = m_Buffer + len - 1;
	for (int i = len - from; i > 0; i--) {
		if (*pos == c)
			return (int)(pos - m_Buffer);
		pos--;
	}

	return -1;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
String String::toLower() const {
	if (m_Length == NULL)
		return String();
	String str(*this);
#if defined(__FORWINDOWS__)
	_strlwr(str.m_Buffer);
#else
	char* c;
	for(c = str.m_Buffer; *c != 0; c++)
		*c = tolower(*c);
#endif
	return str;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
String String::toUpper() const {
	if (m_Length == NULL)
		return String();
	String str(*this);
#if defined(__FORWINDOWS__)
	_strupr(str.m_Buffer);
#else
	char* c;
	for(c = str.m_Buffer; *c != 0; c++)
		*c = toupper(*c);
#endif
	return str;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
String& String::lower() {
	if (m_Length == NULL)
		return *this;
#if defined(__FORWINDOWS__)
	_strlwr(m_Buffer);
#else
	char* c;
	for(c = m_Buffer; *c != 0; c++)
		*c = tolower(*c);
#endif
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
String& String::upper() {
	if (m_Length == NULL)
		return *this;
#if defined(__FORWINDOWS__)
	_strupr(m_Buffer);
#else
	char* c;
	for(c = m_Buffer; *c != 0; c++)
		*c = toupper(*c);
#endif
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
const String& String::operator+=(char c) {
	if (m_Buffer == NULL) {
		m_Length = 1;
		allocBuffer();
		m_Buffer[0] = c;
		m_Buffer[1] = 0;
		return *this;
	}
	if ((m_Length + 1) < m_BufferSize) {
		m_Buffer[m_Length] = c;
		m_Length++;
		m_Buffer[m_Length] = 0;
		return *this;
	}
	
	int oldLength = m_Length;
	char* oldBuffer = m_Buffer;

	m_Length++;
	allocBuffer();
	memcpy(m_Buffer, oldBuffer, oldLength);
	m_Buffer[oldLength] = c;
	m_Buffer[oldLength+1] = 0;
	delete[] oldBuffer;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
const String& String::operator+=(const String& str) {
	if (str.m_Length == 0)
		return *this;
	if (m_Buffer == NULL) {
		m_Length = str.m_Length;
		allocBuffer();
		memcpy(m_Buffer, str.m_Buffer, m_Length + 1);
		return *this;
	}
	if ((m_Length + str.m_Length) < m_BufferSize) {
		memcpy(m_Buffer + m_Length, str.m_Buffer, str.m_Length+1);
		m_Length += str.m_Length;
		return *this;
	}
	
	int oldLength = m_Length;
	char* oldBuffer = m_Buffer;

	m_Length += str.m_Length;
	allocBuffer();
	memcpy(m_Buffer, oldBuffer, oldLength);
	memcpy(m_Buffer+oldLength, str.m_Buffer, str.m_Length+1);
	delete[] oldBuffer;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
const String& String::operator=(const String& str) {
	if (&str == this)
		return *this;
	if (str.m_Length == 0) {
		delete[] m_Buffer;
		m_BufferSize = 0;
		m_Buffer = NULL;
		m_Length = 0;
		return *this;
	}
	if (str.m_Length < m_BufferSize) {
		m_Length = str.m_Length;
		memcpy(m_Buffer, str.m_Buffer, m_Length + 1);
		return *this;
	}  
	
	delete[] m_Buffer;
	m_Length = str.m_Length;
	allocBuffer();
	memcpy(m_Buffer, str.m_Buffer, m_Length + 1);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
const String& String::operator+=(const char* str) {
	ASSERT(str != NULL);

	unsigned int strLen = (unsigned int)strlen(str);
	if (strLen == 0)
		return *this;

	if (m_Buffer == NULL) {
		m_Length = strLen;
		allocBuffer();
		memcpy(m_Buffer, str, strLen+1);
		return *this;
	} 
	if ((m_Length + strLen) < m_BufferSize) {
		memcpy(m_Buffer + m_Length, str, strLen+1);
		m_Length += strLen;
		return *this;
	}
	
	int oldLength = m_Length;
	char* oldBuffer = m_Buffer;
	m_Length += strLen;
	allocBuffer();
	memcpy(m_Buffer, oldBuffer, oldLength);
	memcpy(m_Buffer + oldLength, str, strLen + 1);
	delete[] oldBuffer;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
const String& String::operator=(const char* str) {
	ASSERT(str != NULL);

	m_Length = (unsigned int)strlen(str);
	if (m_Length == 0) {
		delete[] m_Buffer;
		m_BufferSize = 0;
		m_Buffer = NULL;
		return *this;
	}
	if (m_Length < m_BufferSize) {
		memcpy(m_Buffer, str, m_Length + 1);
		return *this;
	}  
	
	delete[] m_Buffer;
	allocBuffer();
	memcpy(m_Buffer, str, m_Length + 1);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
const String& String::operator=(char str) {
	m_Length = 1;
	if (m_Length >= m_BufferSize) {
		delete[] m_Buffer;
		allocBuffer();
	}
	
	m_Buffer[0] = str;
	m_Buffer[1] = 0;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void String::sprintf(const char* fmt, ...) {
	va_list arglist;
	va_start(arglist, fmt);

	// Estimate size of resulting string
	m_Length = _vscprintf(fmt, arglist);
	if (m_Buffer != NULL) {
		if (m_BufferSize <= m_Length) {
			delete[] m_Buffer;
			allocBuffer();
		}
	} else {
		allocBuffer();
	}
	vsprintf(m_Buffer, fmt, arglist);
	ASSERT(m_Length == strlen(m_Buffer));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int String::sscanf(const char* fmt, ...) {
    va_list arglist;
    FILE str;
    FILE *infile = &str;

    va_start(arglist, fmt);

    infile->_flag = _IOREAD|_IOSTRG|_IOMYBUF;
    infile->_ptr = infile->_base = (char *) m_Buffer;
    infile->_cnt = m_Length;;
    return (_input(infile,(const unsigned char*)fmt,arglist));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void String::allocate(unsigned int len) {
	if (m_BufferSize > len) {
		m_Buffer[0] = 0;
		return;
	}
	
	if (m_Buffer != NULL)
		delete[] m_Buffer;

	unsigned int size;
	len++;
	if (len < 64) {
		size = 64;
	} else {
		size = 0x10000000;
		while ((size & len) == 0)
			size >>= 1;
		int add = (size >> 2);
		while (size < len)
			size += add;
	}

	m_BufferSize = size;	
	m_Length = 0;
	m_Buffer = new char[m_BufferSize];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void String::allocBuffer() {
	if (m_Length == 0) {
		m_BufferSize = 0;
		m_Buffer = NULL;
		return;
	}

	unsigned int size;
	if (m_Length < 64) {
		size = 64;
	} else {
		size = 0x10000000;
		while ((size & m_Length) == 0)
			size >>= 1;
		int add = (size >> 2);
		while (size <= m_Length)
			size += add;
	}

	m_BufferSize = size;	
	m_Buffer = new char[m_BufferSize];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
String operator+(const String& str1, const char* str2) {
	ASSERT(str2 != NULL);

	unsigned int strLen = (unsigned int)strlen(str2);
	if (strLen == 0)
		return str1;

	String result;
	result.m_Length = str1.m_Length + strLen;
	result.allocBuffer();
	if (str1.m_Length > 0)
		memcpy(result.m_Buffer, str1.m_Buffer, str1.m_Length);
	if (strLen > 0)
		memcpy(result.m_Buffer + str1.m_Length, str2, strLen);
	result.m_Buffer[result.m_Length] = 0;
	return result;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
String operator+(const char* str1, const String& str2) {
	ASSERT(str1 != NULL);

	unsigned int strLen = (unsigned int)strlen(str1);
	if (strLen == 0)
		return str2;

	String result;
	result.m_Length = str2.m_Length + strLen;
	result.allocBuffer();
	if (strLen > 0)
		memcpy(result.m_Buffer, str1, strLen);
	if (str2.m_Length > 0)
		memcpy(result.m_Buffer + strLen, str2.m_Buffer, str2.m_Length);
	result.m_Buffer[result.m_Length] = 0;
	return result;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
String operator+(const String& str1, const String& str2) {
	String result;
	result.m_Length = str1.m_Length + str2.m_Length;
	result.allocBuffer();
	if (str1.m_Length > 0)
		memcpy(result.m_Buffer, str1.m_Buffer, str1.m_Length);
	if (str2.m_Length > 0)
		memcpy(result.m_Buffer + str1.m_Length, str2.m_Buffer, str2.m_Length);
	result.m_Buffer[result.m_Length] = 0;
	return result;
}

}; };
