/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics.image.h"
#include "graphics.palette.h"

#include "object.inl"
#include "base.string.inl"
#include "graphics.image.inl"
#include "graphics.rectangle.inl"
#include "graphics.dimension.inl"
#include "graphics.colour.inl"
#include "graphics.palette.inl"


namespace magnum { namespace graphics {


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image::Image() {
	m_Locked = false;
	m_Type = Unknown;
	m_Size.setWidth(0);
	m_Size.setHeight(0);
	m_Linesize = 0;
	m_Next = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image::Image(Type type, int w, int h) {
	m_Locked = false;
	m_Type = type;
	m_Size.setWidth(w);
	m_Size.setHeight(h);
	m_Linesize = calcLinesize();
	m_Next = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image::Image(Type type, const Dimension<int>& size) {
	m_Locked = false;
	m_Type = type;
	m_Size = size;
	m_Linesize = calcLinesize();
	m_Next = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image::~Image() {
	ASSERT(!m_Locked);
	if (m_Next != NULL)
		m_Next->release();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int Image::calcLinesize() {
	switch (m_Type) {
		case Unknown:
			m_Linesize = 0;
			break;
		case RGBA8:
		case BGRA8:
			m_Linesize = m_Size.getWidth()*4;
			break;
		case RGB5:
		case R5G6B5:
			m_Linesize = m_Size.getWidth()*2;
			break;
		case RGB8:
		case BGR8:
			m_Linesize = m_Size.getWidth()*3;
			break;
		case RGB12:
			m_Linesize = m_Size.getWidth()*5;
			break;
		case RGB16:
		case BGR16:
		case RGB16f:
		case BGR16f:
			m_Linesize = m_Size.getWidth()*6;
			break;
		case RGBA16:
		case BGRA16:
		case RGBA16f:
		case BGRA16f:
			m_Linesize = m_Size.getWidth()*8;
			break;
		case RGB32f:
		case BGR32f:
			m_Linesize = m_Size.getWidth()*12;
			break;
		case RGBA32f:
		case BGRA32f:
			m_Linesize = m_Size.getWidth()*16;
			break;
		case GR8:
			m_Linesize = m_Size.getWidth();
			break;
		case GR16:
		case GRA8:
		case GR16f:
			m_Linesize = m_Size.getWidth()*2;
			break;
		case GRA16:
		case GRA16f:
			m_Linesize = m_Size.getWidth()*4;
			break;
		case Index8:
			m_Linesize = m_Size.getWidth();
			break;
		case DUDV8:
			m_Linesize = m_Size.getWidth()*2;
			break;
		case DUDV16:
			m_Linesize = m_Size.getWidth()*4;
			break;
		default:
			ASSERT(false);
			m_Linesize = 0;
	}

	return m_Linesize;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Image::copyPixels(const Image* img) {
	byte* src, *dst;
	int i,lsize,height;

	if (img->getType() != getType())
		return false;

	src = (byte *)((Image *)img)->lock();
	if (src == NULL) {
		return false;
	}
	dst = (byte *)lock();
	if (dst == NULL) {
		((Image *)img)->unlock();
		return false;
	}

	lsize = min(img->getLinesize(), getLinesize());
	height = min(img->getHeight(), getHeight());
	for (i = 0; i < height; i++) {
		memcpy(dst, src, lsize);
		src += img->getLinesize();
		dst += getLinesize();
	}

	((Image *)img)->unlock();
	unlock();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Image::copyPalette(const Image* src) {
	if ((getType() != Image::Index8) || (src->getType() != Image::Index8))
		return false;

	Palette<byte> pal;
	src->getPalette(&pal);
	setPalette(pal);

	return true;
}

}; };
