/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "io.exception.h"
#include "graphics.bitmap.h"
#include "graphics.translator.bmp.h"

#include "object.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "io.file.inl"
#include "io.translator.inl"
#include "graphics.translator.inl"
#include "graphics.colour.inl"
#include "graphics.palette.inl"
#include "graphics.image.inl"
#include "graphics.dimension.inl"

namespace magnum { namespace graphics {


#pragma pack(1)


struct BmpHeader {
	word	bfID;
	dword 	bfSize;
	word	bfReserved1;
	word	bfReserved2;
	dword	bfOffBits;
};


struct BitmapInfoHeader {
	dword 	biSize;
	dword 	biWidth;
	dword 	biHeight;
	word	biPlanes;
	word	biBitCount;
	dword 	biCompression;
	dword 	biSizeImage;
	dword 	biXPelsPerMeter;
	dword 	biYPelsPerMeter;
	dword 	biClrUsed;
	dword 	biClrImportant;
};



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image* TranslatorBMP::onDecode(InputStream* in, size_t size) {
	Image* img;
	Image::Type type;
	BmpHeader baseHeader;
	BitmapInfoHeader* infoHeader;
	dword bmpType;
	int compression, colcount;
	int pos;

	if (size == Autosize)
		size = (size_t)in->available();

	// Read base header...
	in->read(&baseHeader, sizeof(baseHeader));

	// Read in BitmapInfo header and evaluate data
	in->read(&bmpType, sizeof(dword));
	infoHeader = (BitmapInfoHeader* )(void *)new byte[bmpType];
	infoHeader->biSize = bmpType;
	in->read(&(infoHeader->biWidth), bmpType - sizeof(dword));
	pos = sizeof(baseHeader) + bmpType;

	// Get image type
	switch (infoHeader->biBitCount) {
		case 1:
		case 2:
		case 4:
		case 8:
			type = Image::Index8;
			break;
		case 15:
			type = Image::RGB5;
			break;
		case 16:
			type = Image::R5G6B5;
			break;
		case 24:
			type = Image::BGR8;
			break;
		case 32:
			type = Image::BGRA8;
			break;
		default:
			throw magnum::io::TranslatorFormatException();
	}

	if (bmpType >= 40) {
		if (infoHeader->biClrUsed)
			colcount = infoHeader->biClrUsed;
		else
			colcount = 1 << infoHeader->biBitCount;
		compression = infoHeader->biCompression;
	} else {
		colcount = 1 << infoHeader->biBitCount;
		compression = 0;
	}

	// Create new image in memory...
	img = new Bitmap(type, infoHeader->biWidth, infoHeader->biHeight);
	if (img == NULL)
		throw magnum::io::TranslatorMemoryException();

	// Read in Palette
	if ((infoHeader->biBitCount <= 8) && (colcount > 0)) {
		dword *cols;
		dword *curcol;

		cols = new dword[colcount];
		in->read(cols, sizeof(dword) * colcount);
		pos += sizeof(dword) * colcount;

		Palette<byte> pal;
		curcol = cols;
		for (int i = 0; i < min(colcount, 256); i++) {
			pal[i] = Colour<byte>(((byte *)curcol)[2], ((byte *)curcol)[1], ((byte *)curcol)[0]);
			curcol++;
		}

		img->setPalette(pal);
		delete[] cols;
	}

	// Skip to compressed data
	in->skip(baseHeader.bfOffBits - pos);
	pos = baseHeader.bfOffBits;
	byte* cdata = new byte[(size_t)(size - pos)];
	in->read(cdata, (size_t)(size-pos));

	if ((compression != 0) && (infoHeader->biBitCount == 4)) {
		decodeRle4(img, cdata);
	} else if ((compression != 0) && (infoHeader->biBitCount == 8)) {
		decodeRle8(img, cdata);
	} else if ((compression == 0) && (infoHeader->biBitCount == 1)) {
		decodeRgb1(img, cdata);
	} else if ((compression == 0) && (infoHeader->biBitCount == 4)) {
		decodeRgb4(img, cdata);
	} else if ((compression == 0) && (infoHeader->biBitCount == 8)) {
		decodeRgb8(img, cdata);
	} else if ((compression == 0) && (infoHeader->biBitCount == 15)) {
		decodeRgb16(img, cdata);
	} else if ((compression == 0) && (infoHeader->biBitCount == 16)) {
		decodeRgb16(img, cdata);
	} else if ((compression == 0) && (infoHeader->biBitCount == 24)) {
		decodeRgb24(img, cdata);
	} else {
		throw magnum::io::TranslatorFormatException();
	}

	delete[] (byte *)(void *)infoHeader;
	delete[] cdata;
	return img;
}


/* ---------------------------------------------------------------------------
--------------------------------------------------------------------------- */
void TranslatorBMP::decodeRle4(Image* img, const byte* in) {
	byte c1,c2,c3,ca,cb,quit;
	int xx,i,j;
	byte *data = (byte *)img->lock();
	int linesize = img->getLinesize();
	data = data + linesize * (img->getHeight() - 1); 

	xx = 0;
	quit = 1;

	while (quit) {
		c1 = *in++;
		c2 = *in++;

		if (c1 == 0x00) {
			/* keine repeating Group */
			switch (c2) {
				case 0x00 :
					/* neue zeile */
					xx = 0;
					data = data - (xx + linesize);
					break;

				case 0x01 :
					/* ende der bitmap */
					quit = 0;
					break;

				case 0x02 :
					/* gotoxy */

					i = *in++;
					j = *in++;
					xx += i;
					data = data + (i - j * linesize);
					break;

				default :
					/* eine literal group */
					for (i=0; i<c2; i++) {

						if (!(i & 0x01)) 
							c3 = *in++;

						ca = (c3 & 0xf0) >> 4;
						cb = (c3 & 0x0f);

						if (i&1)
							*data++ = cb;
						else
							*data++ = ca;

						xx++;
					}

					if ((c2+1) & 2)
						in++; 
					break;
			}
		} else {
			ca = (c2 & 0xf0) >> 4;
			cb = (c2 & 0x0f);

			for (i=0; i<c1; i++) {
				if (i & 1)
					*data++ = cb;
				else
					*data++ = ca;
				xx++;
			}
		}
	}

	img->unlock();
}


/* ---------------------------------------------------------------------------
--------------------------------------------------------------------------- */
void TranslatorBMP::decodeRle8(Image* img, const byte* in) {
	byte c1,c2,quit;
	int xx,i,j;
	byte *data = (byte *)img->lock();
	int linesize = img->getLinesize();
	data = data + linesize * (img->getHeight() - 1); 

	xx = 0;
	quit = 1;

	while (quit) {
		c1 = *in++;
		c2 = *in++;

		if (c1 == 0x00) {
			/* keine repeating Group */
			switch (c2) {
				case 0x00 :
					/* neue zeile */
					data = data - (xx + linesize);
					xx = 0;
					break;

				case 0x01 :
					/* ende der bitmap */
					quit = 0;
					break;

				case 0x02 :
					/* gotoxy */

					i = *in++;
					j = *in++;
					xx += i;
					data = data + (i - j * linesize);
					break;

				default :
					/* eine literal group */
					memcpy(data, in, c2);
					in += c2;
					data += c2;
					xx += c2;

					if ((c2 & 0x01) != 0)
						in++;
					break;
			}

		} else {
			/* eine repeating group */
			memset(data, c2, c1);
			xx += c1;
			data += c1;
		}
	}

	img->unlock();
}


/* ---------------------------------------------------------------------------
--------------------------------------------------------------------------- */
void TranslatorBMP::decodeRgb1(Image* img, const byte* in) {
	byte c1,c2,c3;
	int xx,yy,i,j,xpix,xpad;

	byte* data = (byte *)img->lock();
	int linesize = img->getLinesize();
	int width = img->getWidth();

	xx = 0;
	yy = img->getHeight() - 1;
	data = data + linesize * yy;

	// Picture ist nicht gecrunched
	xpix = width >> 3;
	if ( (width - (xpix << 3)) != 0 )
		xpix++;	// berlauf 1 byte dazu

	xpad = ((xpix + 3) & 0xFFFC) - xpix;

	for (yy = 0; yy < img->getHeight(); yy++) {
		xx = 0;
		for (j=0; j<xpix; j++) {
			// 1 bit per pixel
			c1 = *in++;
			c2 = 0x80;
			c3 = 7;

			for (i=0; i<8; i++) {
				*data++ = (c1 & c2) >> c3;
				xx++;
				if (xx >= width) 
					i = 8;			
				c2 = (c2 >> 1);
				c3--;
			}
		}
		data -= linesize + xx;
		in += xpad;
	}

	img->unlock();
}



/* ---------------------------------------------------------------------------
--------------------------------------------------------------------------- */
void TranslatorBMP::decodeRgb4(Image* img, const byte* in) {
	byte c1,ca,cb;
	int xx,yy,i,j,xpix,xpad;

	byte* data = (byte *)img->lock();
	int linesize = img->getLinesize();
	int width = img->getWidth();

	xx = 0;
	yy = img->getHeight() - 1;
	data = data + linesize * yy;

	// Picture ist nicht gecrunched
	xpix = (width + 1) >> 1;
	xpad = (((width + 7 ) & 0xfff8) >> 1) - xpix;

	for (yy = 0; yy < img->getHeight(); yy++) {
		xx = 0;
		for (j=0; j<xpix; j++) {
			// 4 bit per pixel
			c1 = *in++;
			ca = (c1 & 0xf0) >>4;
			cb = (c1 & 0x0f);
			for (i=0; i<2; i++) {
				if ((i & 0x01) == 0)
					*data++ = ca;
				else
					*data++ = cb;
				xx++;
				if (xx >= width)
					i = 2;	
			}
		}
		data -= linesize + xx;
		in += xpad;
	}

	img->unlock();
}


/* ---------------------------------------------------------------------------
--------------------------------------------------------------------------- */
void TranslatorBMP::decodeRgb8(Image* img, const byte* in) {
	int xx,yy,xpix,xpad;

	byte* data = (byte *)img->lock();
	int linesize = img->getLinesize();
	int width = img->getWidth();

	xx = 0;
	yy = img->getHeight() - 1;
	data = data + linesize * yy;

	// Picture ist nicht gecrunched
	xpix = width;
	xpad = ((width + 3) & 0xFFFC) - xpix;

	for (yy = 0; yy < img->getHeight(); yy++) {
		memcpy(data, in, xpix);
		in += xpix + xpad;
		data -= linesize;
	}

	img->unlock();
}


/* ---------------------------------------------------------------------------
--------------------------------------------------------------------------- */
void TranslatorBMP::decodeRgb16(Image* img, const byte* in) {
	int xx,yy,xpix,xpad;

	byte* data = (byte *)img->lock();
	int linesize = img->getLinesize();
	int width = img->getWidth();

	xx = 0;
	yy = img->getHeight() - 1;
	data = data + linesize * yy;

	// Picture ist nicht gecrunched
	xpix = 2*width;
	xpad = ((xpix + 3) & 0xFFFC) - xpix;

	for (yy = 0; yy < img->getHeight(); yy++) {
		memcpy(data, in, xpix);
		in += xpix + xpad;
		data -= linesize;
	}

	img->unlock();
}


/* ---------------------------------------------------------------------------
--------------------------------------------------------------------------- */
void TranslatorBMP::decodeRgb24(Image* img, const byte* in) {
	int xx,yy,xpix,xpad;

	byte* data = (byte *)img->lock();
	int linesize = img->getLinesize();
	int width = img->getWidth();

	xx = 0;
	yy = img->getHeight() - 1;
	data = data + linesize * yy;

	xpix = 3*width;
	xpad = ((xpix + 3) & 0xFFFC) - xpix;

	for (yy = 0; yy < img->getHeight(); yy++) {
		memcpy(data, in, xpix);
		in += xpix + xpad;
		data -= linesize;
	}

	img->unlock();
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorBMP::onEncode(OutputStream* out, const Image* img) {
	return 0;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorBMP::onIdentify(File* in) {
	word id;

	try {
		int64 oldpos;
		oldpos = in->getPosition();	
		in->read(&id, sizeof(id));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if ((word)id != (word)('B' | ((word)'M')<<8))
		return false;

	return true;
} 


}; };
