/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "io.exception.h"
#include "object.pointer.h"
#include "graphics.bitmap.h"
#include "graphics.translator.dds.h"

#include "object.inl"
#include "object.exception.inl"
#include "object.pointer.inl"
#include "base.string.inl"
#include "io.file.inl"
#include "graphics.translator.inl"
#include "graphics.palette.inl"
#include "graphics.colour.inl"
#include "graphics.image.inl"
#include "graphics.dimension.inl"


namespace magnum { namespace graphics {
using magnum::object::Pointer;

#define MCHAR2(a, b) (a | (b << 8))
#define MCHAR4(a, b, c, d) (a | (b << 8) | (c << 16) | (d << 24))

#define FLAGS_CAPS					0x00000001
#define FLAGS_HEIGHT				0x00000002
#define FLAGS_WIDTH					0x00000004
#define FLAGS_PITCH					0x00000008
#define FLAGS_PIXELFORMAT			0x00001000
#define FLAGS_MIPMAPCOUNT			0x00020000
#define FLAGS_LINEARSIZE			0x00080000
#define FLAGS_DEPTH					0x00800000

#define CAPS1_TEXTURE				0x00001000
#define CAPS1_MIPMAP				0x00400000
#define CAPS1_COMPLEX				0x00000008

#define CAPS2_CUBEMAP				0x00000200
#define CAPS2_CUBEMAP_POSITIVE_X	0x00000400
#define CAPS2_CUBEMAP_NEGATIVE_X	0x00000800
#define CAPS2_CUBEMAP_POSITIVE_Y	0x00001000
#define CAPS2_CUBEMAP_NEGATIVE_Y	0x00002000
#define CAPS2_CUBEMAP_POSITIVE_Z	0x00004000
#define CAPS2_CUBEMAP_NEGATIVE_Z	0x00008000
#define CAPS2_VOLUME				0x00200000


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image* TranslatorDDS::onDecode(InputStream* in, size_t size) {
	m_Input = in;
	
	if (size == Autosize)
		size = (size_t)m_Input->available();
	
	m_Input->read(&m_Header, sizeof(m_Header));
	if (m_Header.ddsIdentifier != MCHAR4('D', 'D', 'S', ' '))
		throw magnum::io::TranslatorFormatException();

	if (!(m_Header.flags & FLAGS_HEIGHT) || !(m_Header.flags & FLAGS_WIDTH)
		|| !(m_Header.flags & FLAGS_CAPS) || !(m_Header.flags & FLAGS_PIXELFORMAT))
		throw magnum::io::TranslatorFormatException();
	
	// Get width and height
	m_Width = m_Header.width;
	m_Height = m_Header.height;
	
	// Check if mipmap count is available
	if (m_Header.flags & FLAGS_MIPMAPCOUNT) {
		m_MipMaps = m_Header.m_MipMaps;
		if (m_MipMaps == 0)
			m_MipMaps = 1; 
	} else {
		m_MipMaps = 1;
	}
	
	// Check if depth (for volume textures) is available
	if (m_Header.flags & FLAGS_DEPTH) {
		m_Depth = m_Header.depth;
		if (m_Depth <= 0)
			m_Depth = 1;
	} else {
		m_Depth = 1;
	}

	// Get pixel format
	switch(m_Header.pixelFourCC){
		case MCHAR4('D', 'X', 'T', '1'):
			m_Type = Image::DXT1;
			break;
		case MCHAR4('D', 'X', 'T', '3'):
			m_Type = Image::DXT3;
			break;
		case MCHAR4('D', 'X', 'T', '5'):
			m_Type = Image::DXT5;
			break;
		case 20:
			m_Type = Image::BGR8;
			break;
		case 21:
			m_Type = Image::BGRA8;
			break;
		case 22:
			m_Type = Image::BGRA8;
			break;
		case 23:
			m_Type = Image::R5G6B5;
			break;
		case 24:
			m_Type = Image::RGB5;
			break;
		case 36:
			m_Type = Image::BGRA16;
			break;
		case 41:
			m_Type = Image::Index8;
			break;
		default:
			switch(m_Header.pixelBpp){
				case 16:
					m_Type = Image::R5G6B5;
					break;
				case 24:
					m_Type = Image::BGR8;
					break;
				case 32:
					m_Type = Image::BGRA8;
					break;
				default:
					throw magnum::io::TranslatorFormatException();
			}
	}
	
	// Read m_Input all maps
	if ((m_Header.caps1 & CAPS1_COMPLEX) && (m_Header.caps2 & CAPS2_CUBEMAP))
		return readCubeTexture();
	else if ((m_Header.caps1 & CAPS1_COMPLEX) && (m_Header.caps2 & CAPS2_VOLUME))
		return readTexture3D();
	else
		return readTexture2D();
}

		
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image* TranslatorDDS::readCubeTexture() {
	Image* faces[6];
	if (m_Header.caps2 & CAPS2_CUBEMAP_POSITIVE_X)
		faces[0] = readTexture2D();
	else
		faces[0] = NULL;
	if (m_Header.caps2 & CAPS2_CUBEMAP_NEGATIVE_X)
		faces[1] = readTexture2D();
	else
		faces[1] = NULL;
	if (m_Header.caps2 & CAPS2_CUBEMAP_POSITIVE_Y)
		faces[2] = readTexture2D();
	else
		faces[2] = NULL;
	if (m_Header.caps2 & CAPS2_CUBEMAP_NEGATIVE_Y)
		faces[3] = readTexture2D();
	else
		faces[3] = NULL;
	if (m_Header.caps2 & CAPS2_CUBEMAP_POSITIVE_Z)
		faces[4] = readTexture2D();
	else
		faces[4] = NULL;
	if (m_Header.caps2 & CAPS2_CUBEMAP_NEGATIVE_Z)
		faces[5] = readTexture2D();
	else
		faces[5] = NULL;
		
	Image* root = NULL;
	Image* img;
	Image* tmp;
	for (unsigned int i = 0; i < m_MipMaps; ++i) {
		for (unsigned int j = 0; j < 6; ++j) {
			if (faces[j] == NULL)
				continue;
				
			tmp = faces[j];
			faces[j] = tmp->getNext();
			if (faces[j] != NULL) {
				faces[j]->refer();
				tmp->detachNext();
			}
				
			if (root == NULL) {
				root = tmp;
				img = root;
			} else {
				img->attachNext(tmp);
				img = tmp;
				tmp->release();
			}
		}
	}
	
	return root;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image* TranslatorDDS::readTexture3D() {
	Image* root = NULL;
	Image* img = NULL;
	
	unsigned int maps = m_MipMaps;
	unsigned int width = m_Width;
	unsigned int height = m_Height;
	unsigned int depth = m_Depth;
	
	for (unsigned int j = 0; j < maps; ++j) {
		for (unsigned int i = 0; i < depth; ++i) {
			Image* nextImg = new Bitmap(m_Type, width, height);
			if (root != NULL) {
				img->attachNext(nextImg);
				nextImg->release();
			} else {
				root = nextImg;
			}
			img = nextImg;

			void* data = img->lock();
			int readsize = img->getLinesize() * img->getHeight();
			m_Input->read(data, readsize);
			img->unlock();
		}

		if (width > 1)
			width >>= 1;
		if (height > 1)
			height >>= 1;
		if (depth > 1)
			depth >>= 1;
	}

	return root;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image* TranslatorDDS::readTexture2D() {
	Image* root = NULL;
	Image* img = NULL;
	
	unsigned int maps = m_MipMaps;
	unsigned int width = m_Width;
	unsigned int height = m_Height;
	
	while (maps > 0) {
		Image* nextImg = new Bitmap(m_Type, width, height);
		if (root != NULL) {
			img->attachNext(nextImg);
			nextImg->release();
		} else {
			root = nextImg;
		}
		img = nextImg;

		void* data = img->lock();
		int readsize = img->getLinesize() * img->getHeight();
		m_Input->read(data, readsize);
		img->unlock();

		maps--;
		if (width > 1)
			width >>= 1;
		if (height > 1)
			height >>= 1;
	}

	return root;
}

 
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorDDS::onEncode(OutputStream* out, const Image* img) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorDDS::onIdentify(File* in) {
	dword id;

	try {
		int64 oldpos;
		oldpos = in->getPosition();	
		in->read(&id, sizeof(id));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if (id != MCHAR4('D', 'D', 'S', ' '))
		return false;

	return true;
} 


}; };
