/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "io.exception.h"
#include "io.datainput.h"
#include "graphics.bitmap.h"
#include "graphics.translator.hdr.h"

#include "object.inl"
#include "object.exception.inl"
#include "object.autopointer.inl"
#include "base.string.inl"
#include "io.datainput.inl"
#include "io.translator.inl"
#include "io.file.inl"
#include "graphics.image.inl"
#include "graphics.translator.inl"
#include "graphics.colour.inl"
#include "graphics.palette.inl"
#include "graphics.dimension.inl"


namespace magnum { namespace graphics {
using magnum::io::DataInput;
using magnum::object::AutoPointer;

/* offsets to red, green, and blue components in a data (float) pixel */
#define RGBE_DATA_RED    0
#define RGBE_DATA_GREEN  1
#define RGBE_DATA_BLUE   2
/* number of floats per pixel */
#define RGBE_DATA_SIZE   3


/*----------------------------------------------------------------------------
standard conversion from float pixels to rgbe pixels
note: you can remove the "inline"s if your compiler complains about it
----------------------------------------------------------------------------*/
static inline void float2rgbe(unsigned char rgbe[4], float rgb[4]) {
  	float v;
  	int e;

  	v = rgb[0];
  	if (rgb[1] > v) 
  		v = rgb[1];
  	if (rgb[2] > v) 
  		v = rgb[2];
  	if (v < 1e-32) {
		rgbe[0] = rgbe[1] = rgbe[2] = rgbe[3] = 0;
	} else {
    	v = (float)frexp(v,&e) * 256.0f/v;
    	rgbe[0] = (unsigned char) (rgb[0] * v);
    	rgbe[1] = (unsigned char) (rgb[1] * v);
    	rgbe[2] = (unsigned char) (rgb[2] * v);
    	rgbe[3] = (unsigned char) (e + 128);
	}
}


/*----------------------------------------------------------------------------
standard conversion from rgbe to float pixels
note: Ward uses ldexp(col+0.5,exp-(128+8)).  However we wanted pixels
      in the range [0,1] to map back into the range [0,1].
----------------------------------------------------------------------------*/
static inline void rgbe2float(float rgb[4], unsigned char rgbe[4]) {
  	if (rgbe[3]) {   /*nonzero pixel*/
  	  	float f = (float)ldexp(1.0,rgbe[3]-(int)(128+8));
  	  	rgb[0] = rgbe[0] * f;
  	  	rgb[1] = rgbe[1] * f;
  	  	rgb[2] = rgbe[2] * f;
  	} else {
  		rgb[0] = rgb[1] = rgb[2] = 0.0;
  	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image* TranslatorHDR::onDecode(InputStream* in,size_t size) {
	m_Input = in;
	if (size == Autosize)
		m_Size = (size_t)in->available();
	else
		m_Size = size;
		
	readHeader();
	AutoPointer<Image> img = new Bitmap(Image::RGB32f, m_Width, m_Height);
	m_Buffer = new byte[m_Width*4];
	float* data = (float*)img->lock();
	readPixelsRLE(data);
	img->unlock();
	delete[] m_Buffer;
	
	img->refer();
	return img;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorHDR::onEncode(OutputStream* ,const Image* ) {
	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorHDR::onIdentify(File* ) {
	return true;
}


/*----------------------------------------------------------------------------
minimal header reading.  modify if you want to parse more information
----------------------------------------------------------------------------*/
void TranslatorHDR::readHeader() {
	String str;
	DataInput in(m_Input);
  	int found_format;
  	float tempf;

  	found_format = 0;
  	m_Valid = 0;
  	m_ProgramType = "";
  	m_Gamma = 1.0f;
  	m_Exposure = 1.0f;

	str = in.readLine();
  	if ((str[0] != '#') || (str[1] != '?'))
  		throw magnum::io::TranslatorFormatException();

 	m_Valid |= RGBE_VALID_PROGRAMTYPE;
 	m_ProgramType = str.substring(2);
  	
  	for(;;) {
		str = in.readLine();  		
  		if ((str[0] == 0) || (str[0] == '\n'))
	  		throw magnum::io::TranslatorFormatException("TranslatorHDR: No FORMAT specifier found");
  		else if (str == "FORMAT=32-bit_rle_rgbe\n")
  			break;       /* format found so break out of loop */
  		else if (str.sscanf("GAMMA=%g",&tempf) == 1) {
  			m_Gamma = tempf;
  			m_Valid |= RGBE_VALID_GAMMA;
  		} else if (str.sscanf("EXPOSURE=%g",&tempf) == 1) {
  			m_Exposure = tempf;
  			m_Valid |= RGBE_VALID_EXPOSURE;
  		}
  	}
  	str = in.readLine();
  	if (str != "\n")
  		throw magnum::io::TranslatorFormatException("TranslatorHDR: Missing blank line after FORMAT specifier");
  	str = in.readLine();
  	if (str.sscanf("-Y %d +X %d",&m_Height,&m_Width) < 2)
  		throw magnum::io::TranslatorFormatException("TranslatorHDR: Missing image size specifier");
  		
  	m_Size -= (size_t)in.getBytesRead();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorHDR::readPixels(float* data,unsigned int numpixels) {
	unsigned char rgbe[4];
	unsigned int toRead;

	while(numpixels > 0) {
		if (numpixels > m_Width)
			toRead = m_Width;
		else
			toRead = numpixels;
		numpixels -= toRead;
			
		if (m_Input->read(m_Buffer, sizeof(rgbe)*toRead) != sizeof(rgbe)*toRead)
	  		throw magnum::io::TranslatorFormatException();
	  		
	  	while(toRead > 0) {
			rgbe2float(data,rgbe);
			data += RGBE_DATA_SIZE;
			toRead--;
		}
	}
}

      
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorHDR::readPixelsRLE(float *data) {
  	unsigned char rgbe[4], *ptr, *ptr_end;
  	unsigned int i, count;
  	unsigned char buf[2];
  	unsigned int num_scanlines = m_Height;
  	
  	if ((m_Width < 8)||(m_Width > 0x7fff))
  		/* run length encoding is not allowed so read flat*/
  		return readPixels(data, m_Width * m_Height);
  	/* read in each successive scanline */
  	while(num_scanlines > 0) {
  		if (m_Input->read(rgbe,sizeof(rgbe)) != sizeof(rgbe))
	  		throw magnum::io::TranslatorFormatException();

  		if ((rgbe[0] != 2) || (rgbe[1] != 2) || (rgbe[2] & 0x80)) {
  			/* this file is not run length encoded */
  			rgbe2float(data,rgbe);
  			data += RGBE_DATA_SIZE;
  			return readPixels(data, m_Width*num_scanlines-1);
  		}
  		if ((((int)rgbe[2])<<8 | rgbe[3]) != m_Width)
	  		throw magnum::io::TranslatorFormatException("TranslatorHDR: Wrong scanline width");
  	  
  		ptr = &m_Buffer[0];
  		/* read each of the four channels for the scanline into the buffer */
  		for(i = 0; i < 4; i++) {
  			ptr_end = &m_Buffer[(i+1)*m_Width];
  			while(ptr < ptr_end) {
				if (m_Input->read(buf,2) != 2)
			  		throw magnum::io::TranslatorFormatException();

				if (buf[0] > 128) {
					/* a run of the same value */
					count = buf[0]-128;
					if ((count == 0) || ((int)count > ptr_end - ptr))
				  		throw magnum::io::TranslatorFormatException("TranslatorHDR: Bad scanline data");

					while(count-- > 0)
						*ptr++ = buf[1];
				} else {
					/* a non-run */
	  				count = buf[0];
	  				if ((count == 0) || ((int)count > ptr_end - ptr))
				  		throw magnum::io::TranslatorFormatException("TranslatorHDR: Bad scanline data");

	  				*ptr++ = buf[1];
	  				if (--count > 0) {
	  					if (m_Input->read(ptr,count) != count)
					  		throw magnum::io::TranslatorFormatException();
	  					ptr += count;
	  				}
				}
  			}
  		}
  		
    	/* now convert data from buffer into floats */
    	for(i = 0; i < m_Width; i++) {
    	  	rgbe[0] = m_Buffer[i];
    	  	rgbe[1] = m_Buffer[i+m_Width];
    	  	rgbe[2] = m_Buffer[i+2*m_Width];
    	  	rgbe[3] = m_Buffer[i+3*m_Width];
    	  	rgbe2float(data,rgbe);
    	  	data += RGBE_DATA_SIZE;
    	}
    	num_scanlines--;
	}
}

}; };
