/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "io.exception.h"
#include "graphics.bitmap.h"
#include "graphics.translator.jpg.h"

#include "object.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "io.translator.inl"
#include "io.file.inl"
#include "container.chain.inl"
#include "graphics.translator.inl"
#include "graphics.palette.inl"
#include "graphics.image.inl"


#include <setjmp.h>


namespace magnum { namespace graphics {


#define INPUT_BUF_SIZE  (1<<14)	/* choose an efficiently fread'able size */

#define MakeID(a,b,c,d) (((long)((a)&0xff)<<24)|((long)((b)&0xff)<<16)|((long)((c)&0xff)<<8)|((long)((d)&0xff)))


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorJPG::initSource (j_decompress_ptr cinfo) {
	SourceMgr* mgr = (SourceMgr*)cinfo->src;
	mgr->start_of_file = true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
boolean TranslatorJPG::fillSourceBuffer (j_decompress_ptr cinfo) {
	SourceMgr* src = (SourceMgr*) cinfo->src;
	size_t nbytes;

	nbytes = src->infile->read(src->buffer, INPUT_BUF_SIZE);

	if (nbytes <= 0) {
		if (src->start_of_file)	/* Treat empty input file as fatal error */
			ERREXIT(cinfo, JERR_INPUT_EMPTY);
		WARNMS(cinfo, JWRN_JPEG_EOF);
		/* Insert a fake EOI marker */
		src->buffer[0] = (JOCTET) 0xFF;
		src->buffer[1] = (JOCTET) JPEG_EOI;
		nbytes = 2;
	}

	src->pub.next_input_byte = (unsigned char*)src->buffer;
	src->pub.bytes_in_buffer = nbytes;
	src->start_of_file = false;

	return TRUE;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorJPG::skipSourceData (j_decompress_ptr cinfo, long num_bytes) {
	SourceMgr* src = (SourceMgr*) cinfo->src;

	/* Just a dumb implementation for now.  Could use fseek() except
	 * it doesn't work on pipes.  Not clear that being smart is worth
	 * any trouble anyway --- large skips are infrequent.
	 */
	if (num_bytes > 0) {
		while (num_bytes > (long) src->pub.bytes_in_buffer) {
			num_bytes -= (long) src->pub.bytes_in_buffer;
			(void) fillSourceBuffer(cinfo);
		/* note we assume that fillSourceBuffer will never return FALSE,
         * so suspension need not be handled.
		 */
	    }
		src->pub.next_input_byte += (size_t) num_bytes;
		src->pub.bytes_in_buffer -= (size_t) num_bytes;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorJPG::termSource (j_decompress_ptr cinfo) {
  /* no work necessary here */
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorJPG::errorExit (j_common_ptr cinfo) {
	ErrorMgr* myerr = (ErrorMgr*) cinfo->err;
	(*cinfo->err->output_message) (cinfo);
	longjmp(myerr->setjmp_buffer, 1);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorJPG::TranslatorJPG() {
	m_Decompressor.err = jpeg_std_error(&m_ErrorMgr.pub);
	m_ErrorMgr.pub.error_exit = errorExit;
	jpeg_create_decompress(&m_Decompressor);

	m_Decompressor.src = &m_Source.pub;
	m_Source.buffer = new byte[INPUT_BUF_SIZE];
	m_Source.pub.init_source = initSource;
	m_Source.pub.fill_input_buffer = fillSourceBuffer;
	m_Source.pub.skip_input_data = skipSourceData;
	m_Source.pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
	m_Source.pub.term_source = termSource;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorJPG::~TranslatorJPG() {
	jpeg_destroy_decompress(&m_Decompressor);
	delete[] m_Source.buffer;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image* TranslatorJPG::onDecode(InputStream* in, size_t size) {
	int rowStride;
	Image* img = NULL;
	byte* imgBuf = NULL;
	byte** rowTable = NULL;
	int i;

	if (setjmp(m_ErrorMgr.setjmp_buffer)) {
		if (rowTable != NULL)	
			delete[] rowTable;
		if (img != NULL)
			img->release();
		throw magnum::io::TranslatorFormatException();
	}

	// Reset some source parameters
	m_Source.infile = in;
	m_Source.pub.bytes_in_buffer = 0; 
	m_Source.pub.next_input_byte = NULL; 

	jpeg_read_header(&m_Decompressor, TRUE);
	m_Decompressor.out_color_space = JCS_RGB;

	jpeg_start_decompress(&m_Decompressor);

	rowStride = m_Decompressor.output_width * m_Decompressor.output_components;
	rowTable = new byte*[m_Decompressor.output_height];

	img = new Bitmap(Image::RGB8, m_Decompressor.output_width, m_Decompressor.output_height);
	imgBuf = (byte *)img->lock();

	for (i = 0; i < (int)m_Decompressor.output_height; i++) {
		rowTable[i] = imgBuf + i*rowStride;
	}

	while (m_Decompressor.output_scanline < m_Decompressor.output_height) {
		jpeg_read_scanlines(&m_Decompressor, (unsigned char **)rowTable+m_Decompressor.output_scanline, m_Decompressor.output_height);
	}
	jpeg_finish_decompress(&m_Decompressor);

	img->unlock();
	delete[] rowTable;

	return img;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorJPG::onEncode(OutputStream* out, const Image* img) {
	return 0;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorJPG::onIdentify(File* in) {
	dword id;

	try {
		int64 oldpos;
		oldpos = in->getPosition();	
		in->setPosition(6, File::Current);
		in->read(&id, sizeof(id));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if (id != MakeID('F','I','F','J'))
		return false;

	return true;
} 


}; };
