/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "io.exception.h"
#include "graphics.bitmap.h"
#include "graphics.translator.pcx.h"

#include "object.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "io.translator.inl"
#include "io.file.inl"
#include "graphics.translator.inl"
#include "graphics.colour.inl"
#include "graphics.palette.inl"
#include "graphics.image.inl"
#include "graphics.dimension.inl"


namespace magnum { namespace graphics {

#define DCX_Magic	       		0x3ade68b1L
#define SCR_Magic	       		0x0cd
#define PCX_Magic             	10
#define PCX_Coding            	1
#define PCX_Version           	5
#define PCX_Pal_Type          	0
#define PCX_Bitplanes         	1

typedef struct {
		byte 	m_Magic;
		byte 	m_Version;
		byte 	m_PackType;
		byte 	m_ColorDepth;
		word   	m_X1,m_Y1,m_X2,m_Y2;
		word   	m_XResolution;
		word   	m_YResolution;
		byte 	m_EgaPalette[48];
		byte 	m_Reserved;
		byte 	m_Bitplanes;
		word   	m_BytesPerLine;
		word   	m_PaletteType;
		byte 	m_Dummy[58];
} PcxHeader;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image* TranslatorPCX::onDecode(InputStream* in, size_t size) {
	int bsize,i,w,h;
	byte *sbuffer,*pbuffer,*srcptr,*dstptr,col;
	PcxHeader phead;
	Image *img;

	if (size == Autosize)
		size = (size_t)in->available();

	if (in->read(&phead, sizeof(phead)) != sizeof(phead))
		throw magnum::io::TranslatorFormatException();
 
	if (phead.m_Magic != PCX_Magic) 
		throw magnum::io::TranslatorFormatException();
	
	w = lsb2word(phead.m_X2) - lsb2word(phead.m_X1) + 1;
	h = lsb2word(phead.m_Y2) - lsb2word(phead.m_Y1) + 1;
	if ((w < 0) || (w > 16384) || (h < 0) || (h > 16384))
		throw magnum::io::TranslatorFormatException();

	bsize = w*h;
	img = new Bitmap(Image::Index8, w, h);
	if (img == NULL)
		throw magnum::io::TranslatorMemoryException();

	size -= 768 + sizeof(PcxHeader);
	sbuffer = new byte[(size_t)size];
	pbuffer = new byte[768];
	if ((sbuffer == NULL) || (pbuffer == NULL))
		throw magnum::io::TranslatorMemoryException();

	in->read(sbuffer, (size_t)size);
	in->read(pbuffer, 768);

	srcptr = sbuffer;
	dstptr = (byte *)img->lock();

	while(bsize>0) {
		if (*srcptr>192) {
			i=(int)((*srcptr++) & 63);
			col=*srcptr++;
			bsize-=(int)i;
			while (i) {
				*dstptr++=col;
				i--;
			}
		} else {
			*dstptr++=*srcptr++;
			bsize--;
		}
	}

	img->unlock();

	Palette<byte> pal;
	for (i=0; i<256; i++)
		pal.setColour(Colour<byte>(pbuffer[i*3],pbuffer[i*3+1],pbuffer[i*3+2]),i);

	img->setPalette(pal);

	delete[] pbuffer;
	delete[] sbuffer;

	return img;
}


 
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorPCX::onEncode(OutputStream* out, const Image* img) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorPCX::onIdentify(File* in) {
	PcxHeader head;

	try {
		int64 oldpos;
		oldpos = in->getPosition();	
		in->read(&head, sizeof(head));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if (head.m_Magic != PCX_Magic)
		return false;
	if ((head.m_ColorDepth != 8) &&
		(head.m_ColorDepth != 1) &&
		(head.m_ColorDepth != 2) &&
		(head.m_ColorDepth != 4))
		return false;
	if (head.m_Version > PCX_Version)
		return false;
	if (head.m_PackType != PCX_Coding)
		return false;

	return true;
} 


}; };
