/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "io.exception.h"
#include "graphics.bitmap.h"
#include "graphics.translator.psd.h"

#include "object.autopointer.inl"
#include "graphics.image.inl"
#include "graphics.bitmap.inl"
#include "io.translator.inl"
#include "io.file.inl"
#include "io.inputstream.inl"


namespace magnum { namespace graphics {
using magnum::object::AutoPointer;
using magnum::types::msb2dword;
using magnum::types::msb2word;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorPSD::TranslatorPSD() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorPSD::~TranslatorPSD() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorPSD::onEncode(OutputStream* ,const Image* ) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorPSD::onIdentify(File* in) {
	char signature[4];

	try {
		int64 oldpos;
		oldpos = in->getPosition();	
		in->read(signature, 4);
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if (signature[0] == '8' &&
		signature[1] == 'B' &&
		signature[2] == 'P' &&
		signature[3] == 'S')
		return true;

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image* TranslatorPSD::onDecode(InputStream* file,size_t size) {
	file->read(&m_Header, sizeof(Header));

	m_Header.m_Version = msb2word(m_Header.m_Version);
	m_Header.m_Channels = msb2word(m_Header.m_Channels);
	m_Header.m_Height = msb2word(m_Header.m_Height);
	m_Header.m_Width = msb2word(m_Header.m_Width);
	m_Header.m_Depth = msb2word(m_Header.m_Depth);
	m_Header.m_Mode = msb2word(m_Header.m_Mode);

	if (m_Header.m_Signature[0] != '8' ||
		m_Header.m_Signature[1] != 'B' ||
		m_Header.m_Signature[2] != 'P' ||
		m_Header.m_Signature[3] != 'S')
		throw magnum::io::TranslatorFormatException();

	if (m_Header.m_Version != 1)
		throw magnum::io::TranslatorFormatException();

	if (m_Header.m_Mode != 3 || m_Header.m_Depth != 8)
		throw magnum::io::TranslatorFormatException();

	// skip color mode data
	uint32 l;
	file->read(&l, sizeof(uint32));
	l = msb2dword(l);
	if (file->skip(l) != l)
		throw magnum::io::TranslatorFormatException();

	// skip image resources
	file->read(&l, sizeof(uint32));
	l = msb2dword(l);
	if (file->skip(l) != l)
		throw magnum::io::TranslatorFormatException();

	// skip layer & mask
	file->read(&l, sizeof(uint32));
	l = msb2dword(l);
	if (file->skip(l) != l)
		throw magnum::io::TranslatorFormatException();

	// read image data
	uint16 compressionType;
	file->read(&compressionType, sizeof(uint16));
	compressionType = msb2word(compressionType);
	if (compressionType != 1 && compressionType != 0)
		throw magnum::io::TranslatorFormatException();

	// create image data block
	AutoPointer<Image> img = new Bitmap(Image::RGBA8, m_Header.m_Width, m_Header.m_Height);

	if (compressionType == 0)
		readRaw(file, img); // RAW image data
	else
		readRLE(file, img); // RLE compressed data

	img->refer();
	return img;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPSD::readRaw(InputStream *file, Image* img) {
	byte* tmp = new byte[(unsigned int)m_Header.m_Width * (unsigned int)m_Header.m_Height];

	for (unsigned int channel=0; channel < m_Header.m_Channels && channel < 3; ++channel) {
		unsigned int size = m_Header.m_Width * m_Header.m_Height;
		if (file->read(tmp, size) != size) {
			delete[] tmp;
			throw magnum::io::TranslatorFormatException();
		}

		int shift = getShift(channel);
		if (shift != -1) {
			dword* data = (dword*)img->lock();
			uint32 mask = ~(0xff << shift);

			for (unsigned int x=0; x<m_Header.m_Width; ++x) {
				for (unsigned int y=0; y<m_Header.m_Height; ++y) {
					unsigned int index = x + y*m_Header.m_Width;
					data[index] = (data[index] & mask) | (tmp[index] << shift);
				}
			}
			img->unlock();
		}
	}
		
	delete [] tmp;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPSD::readRLE(InputStream *file, Image* img) {
	/*	If the compression code is 1, the image data
		starts with the byte counts for all the scan lines in the channel
		(LayerBottom LayerTop), with each count stored as a two
		byte value. The RLE compressed data follows, with each scan line
		compressed separately. The RLE compression is the same compres-sion
		algorithm used by the Macintosh ROM routine PackBits, and
		the TIFF standard.
		If the Layers Size, and therefore the data, is odd, a pad byte will
		be inserted at the end of the row.
	*/

	/*
	A pseudo code fragment to unpack might look like this:

	Loop until you get the number of unpacked bytes you are expecting:
		Read the next source byte into n.
		If n is between 0 and 127 inclusive, copy the next n+1 bytes literally.
		Else if n is between -127 and -1 inclusive, copy the next byte -n+1
		times.
		Else if n is -128, noop.
	Endloop

	In the inverse routine, it is best to encode a 2-byte repeat run as a replicate run
	except when preceded and followed by a literal run. In that case, it is best to merge
	the three runs into one literal run. Always encode 3-byte repeats as replicate runs.
	That is the essence of the algorithm. Here are some additional rules:
	- Pack each row separately. Do not compress across row boundaries.
	- The number of uncompressed bytes per row is defined to be (ImageWidth + 7)
	/ 8. If the uncompressed bitmap is required to have an even number of bytes per
	row, decompress into word-aligned buffers.
	- If a run is larger than 128 bytes, encode the remainder of the run as one or more
	additional replicate runs.
	When PackBits data is decompressed, the result should be interpreted as per com-pression
	type 1 (no compression).
	*/

	byte* tmp = new byte[(unsigned int)m_Header.m_Width * (unsigned int)m_Header.m_Height];
	uint16 *rleCount= new uint16 [(unsigned int)m_Header.m_Height * m_Header.m_Channels];

	unsigned int size=0;      
      
	for (unsigned int y = 0; y < (unsigned int)m_Header.m_Height * m_Header.m_Channels; ++y) {
		if (file->read(&rleCount[y], sizeof(uint16)) != sizeof(uint16)) {
			delete [] tmp;
			delete [] rleCount;
			throw magnum::io::TranslatorFormatException();
		}

		rleCount[y] = msb2word(rleCount[y]);
		size += rleCount[y];
	}

	byte *buf = new byte[size];       
	if (file->read(buf, size) != size) {
		delete [] rleCount;
		delete [] buf;
		delete [] tmp;
		throw magnum::io::TranslatorFormatException();
	}

	uint16 *rcount = rleCount; 
	
	int8 rh;
	uint16 bytesRead;
	byte* dest;
	byte *pBuf = buf;

	// decompress packbit rle 

	for (unsigned int channel = 0; channel < m_Header.m_Channels; channel++) {
		for (unsigned int y = 0; y < m_Header.m_Height; ++y, ++rcount) {
			bytesRead=0;
			dest = &tmp[y*m_Header.m_Width];
			
			while (bytesRead < *rcount) {
				rh = *pBuf++;
				++bytesRead;
				
				if (rh >= 0) {
					++rh;
					while (rh--) { 
						*dest = *pBuf++;
						++bytesRead;
						++dest;
					} 
				} else /*if (rh > -128)*/ {
					rh = -rh +1;
					while (rh--) {
						*dest = *pBuf;
						++dest;
					}

					++pBuf;
					++bytesRead;
				}
			}
		}
		
		int shift = getShift(channel);
		if (shift != -1) {
			dword* data = (dword*)img->lock();
			uint32 mask = ~(0xff << shift);

			for (unsigned int x=0; x<m_Header.m_Width; ++x) {
				for (unsigned int y=0; y<m_Header.m_Height; ++y) {
					unsigned int index = x + y*m_Header.m_Width;
					data[index] = (data[index] & mask) | (tmp[index] << shift);
				}
			}
			img->unlock();
		}
	}
       
	delete [] rleCount;
	delete [] buf;
	delete [] tmp;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int TranslatorPSD::getShift(int channelNr) {
	switch(channelNr) {
		case 0:
			return 0;  // red
		case 1:
			return 8;   // green
		case 2:
			return 16;   // blue
		case 3:
			return m_Header.m_Channels == 4 ? 24 : -1;	// ?
		case 4:
			return 24;  // alpha
		default:
			return -1;
	}
}


}; };

