/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "graphics3d.exception.h"
#include "graphics3d.appearance.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.normalgenerator.h"
#include "graphics3d.binormalgenerator.h"
#include "graphics3d.context.h"

#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "geometry.attributegenerator.inl"
#include "geometry.binormalgenerator.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.normalgenerator.inl"
#include "graphics3d.binormalgenerator.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.mesh.inl"

namespace magnum { namespace graphics3d {
using magnum::object::AutoPointer;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void BinormalGenerator::process(IndexMesh* mesh) {
	ASSERT(mesh != NULL);
	ASSERT(mesh->getIndexBuffer() != NULL);
	
	// First check if there are already normals
	bool hasNormals = false;
	for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		VertexBuffer* vb = mesh->getVertexBuffer(i);
		if (vb == NULL)
			continue;
			
		const VertexFormat& vfmt = vb->getFormat();
		for (unsigned int j = 0; j < vfmt.getSize(); j++) {
			if (vfmt[j].getAttribute() == Attributes::Normal) {
				hasNormals = true;
				break;
			}
		}
	}
	
	if (!hasNormals) {
		NormalGenerator generator;
		generator.process(mesh);
	}

	// Now create binormals	
	IndexBuffer* ib = mesh->getIndexBuffer();
	if (ib->getType() == IndexBuffer::Index8)
		processType<uint8>(mesh);
	else if (ib->getType() == IndexBuffer::Index16)
		processType<uint16>(mesh);
	else if (ib->getType() == IndexBuffer::Index16)
		processType<uint32>(mesh);
	else
		throw UnsupportedFormatException();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class I>
void BinormalGenerator::processType(IndexMesh* mesh) {
	ASSERT(mesh != NULL);
	ASSERT(mesh->getIndexBuffer() != NULL);

	// Get index buffer
	IndexBuffer* ib = mesh->getIndexBuffer();
	
	// Search vertex buffer with vertex and normal information
	VertexBuffer* vvb = NULL;
	unsigned int voffset, vstride;
	VertexBuffer* nvb = NULL;
	unsigned int noffset, nstride;
	VertexBuffer* bvb = NULL;
	unsigned int boffset, bstride;
	VertexBuffer* tvb = NULL;
	unsigned int toffset, tstride;
	for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		VertexBuffer* vb = mesh->getVertexBuffer(i);
		if (vb == NULL)
			continue;
			
		const VertexFormat& vfmt = vb->getFormat();
		for (unsigned int j = 0; j < vfmt.getSize(); j++) {
			if (vfmt[j].getAttribute() == Attributes::Vertex) {
				vvb = vb;
				voffset = vfmt[j].getOffset();
				vstride = vb->getStride();
			} else if (vfmt[j].getAttribute() == Attributes::Normal) {
				nvb = vb;
				noffset = vfmt[j].getOffset();
				nstride = vb->getStride();
			} else if (vfmt[j].getAttribute() == Attributes::Binormal) {
				bvb = vb;
				boffset = vfmt[j].getOffset();
				bstride = vb->getStride();
			} else if (vfmt[j].getAttribute() == Attributes::Tangent) {
				tvb = vb;
				toffset = vfmt[j].getOffset();
				tstride = vb->getStride();
			}
		}
	}
	
	// Check if we found at least the vertices
	if (vvb == NULL)
		throw UnsupportedFormatException();
	if (nvb == NULL)
		throw UnsupportedFormatException();
		
	// If there is no normal buffer, create one
	VertexFormat fmt;
	if (bvb == NULL)
		fmt.insert(Types::Float3, Attributes::Binormal);
	if (tvb == NULL)
		fmt.insert(Types::Float3, Attributes::Tangent);

	// Create buffer for binormals and tangents if neccessary
	if (fmt.getSize() > 0) {
		AutoPointer<VertexBuffer> vb = VertexBuffer::create(fmt, vvb->getSize(), VertexBuffer::Static);
		for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
			if (mesh->getVertexBuffer(i) == NULL) {
				mesh->attachVertexBuffer(vb, i);
				break;
			}
		}
		
		// Assign vertex buffers
		if (bvb == NULL) {
			bvb = vb;
			bstride = fmt.getStride();
			boffset = 0;
			if (tvb == NULL) {
				tvb = vb;
				tstride = fmt.getStride();
				toffset = sizeof(Vector3<float>);
			}
		} else {
			tvb = vb;
			tstride = fmt.getStride();
			toffset = 0;
		}
	}

	// Allocate new normal generator
	magnum::geometry::BinormalGenerator<float, I> generator;
	
	// Setup primitive
	switch(mesh->getPrimitive()) {
		case Mesh::TriangleFan:
			generator.setPrimitive(magnum::geometry::AttributeGenerator<float,I>::TriangleFan);
			break;
		case Mesh::TriangleList:
			generator.setPrimitive(magnum::geometry::AttributeGenerator<float,I>::TriangleList);
			break;
		case Mesh::TriangleStrip:
			generator.setPrimitive(magnum::geometry::AttributeGenerator<float,I>::TriangleStrip);
			break;
		default:
			throw UnsupportedPrimitiveException();
	}

	// Setup all buffers
	const I* idxs = (const I*)ib->lock(IndexBuffer::ReadOnly);
	generator.setIndices(idxs, ib->getSize());
	
	// Lock all buffers
	Map<VertexBuffer*, byte*> locks;
	locks.insert(vvb, (byte*)vvb->lock(VertexBuffer::ReadWrite));
	if (!locks.contains(nvb))
		locks.insert(nvb, (byte*)nvb->lock(VertexBuffer::ReadWrite));
	if (!locks.contains(tvb))
		locks.insert(tvb, (byte*)bvb->lock(VertexBuffer::ReadWrite));
	if (!locks.contains(bvb))
		locks.insert(bvb, (byte*)bvb->lock(VertexBuffer::ReadWrite));
	
	const Vector3<float>* verts = (const Vector3<float>*)(locks[vvb] + voffset);
	generator.setVertices(verts, vstride);
	const Vector3<float>* normals = (const Vector3<float>*)(locks[nvb] + noffset);
	generator.setNormals(normals, nstride);
	Vector3<float>* binormals = (Vector3<float>*)(locks[bvb] + boffset);
	generator.setBinormals(binormals, bstride);
	Vector3<float>* tangents = (Vector3<float>*)(locks[tvb] + toffset);
	generator.setTangents(tangents, tstride);
	
	// Create normals!
	generator.generate();
	
	// Unlock buffers
	ib->unlock();
	vvb->unlock();
	if (nvb->isLocked())
		nvb->unlock();
	if (tvb->isLocked())
		tvb->unlock();
	if (bvb->isLocked())
		bvb->unlock();
}

}; };
