/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics3d.bonemodel.h"
#include "graphics3d.exception.h"
#include "graphics3d.mesh.h"
#include "graphics3d.meshgroup.h"
#include "graphics3d.material.h"
#include "graphics3d.shader.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.appearance.h"
#include "graphics3d.renderstream.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.weightbuffer.h"
#include "graphics3d.context.h"
#include "graphics3d.bounds.h"
#include "graphics3d.parameter.h"
#include "graphics3d.parameterbuffer.h"
#include "graphics3d.skeleton.h"
#include "graphics3d.boneanimation.h"

#include "object.pointer.inl"
#include "math.matrix.inl"
#include "math.vector.inl"
#include "container.array.inl"
#include "container.list.inl"
#include "container.allocator.inl"
#include "container.iterator.inl"
#include "geometry.bounds.inl"
#include "geometry.sphere.inl"
#include "geometry.box.inl"
#include "geometry.alignedbox.inl"
#include "graphics3d.joint.inl"
#include "graphics3d.model.inl"
#include "graphics3d.bonemodel.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.renderstream.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.matrixstack.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.weightbuffer.inl"
#include "graphics3d.parameterbuffer.inl"
#include "graphics3d.parameterformat.inl"
#include "graphics3d.bounds.inl"
#include "graphics3d.animation.inl"
#include "graphics3d.skeleton.inl"
#include "graphics3d.boneanimation.inl"


namespace magnum { namespace graphics3d {
using magnum::geometry::AlignedBox3;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void BoneModel::finish() {
	if (m_Skeleton == NULL)
		throw IncompleteModelException();
	if (m_Weights == NULL)
		throw IncompleteModelException();
	
	// Create parameter format
	ParameterFormat fmt;
	fmt.insert(Parameters::ModelView, Types::Matrix4f);
	fmt.insert(Parameters::AnimationData, Types::Struct);
	for (Skeleton::Iterator iter = m_Skeleton->begin(); iter != m_Skeleton->end(); ++iter) {
		fmt.insert(Parameters::BoneMatrix, Types::Matrix4f);
	}
	publishParameters(fmt);
	
	// Publish all bones as joints
	unsigned int joint = sizeof(Matrix4<float>);
	for (Skeleton::Iterator iter = m_Skeleton->begin(); iter != m_Skeleton->end(); ++iter) {
		publishJoint(iter->getName(), joint);
		joint += sizeof(Matrix4<float>);
	}	

	// Create additional vertex buffers for interpolated frames
	VertexFormat vfmt;
	vfmt.set(0, Types::Float3, Attributes::Vertex);
	vfmt.set(1, Types::Float3, Attributes::Normal);
	m_Vertices = VertexBuffer::create(vfmt, m_Weights->getSize(), VertexBuffer::Stream);

	// Create bounding box
	// Transform frame
	Array<Matrix4<float> > mtx;
	mtx.setSize(m_Skeleton->getSize());
	for (unsigned int i = 0; i < mtx.getSize(); i++)
		mtx[i].identity();

	Array<Vector3<float> > verts;
	verts.setSize(2*m_Weights->getSize());		
	m_Weights->transform(&verts[0], &mtx[0]);
	AlignedBox3<float> box;
	box.bound(reinterpret_cast<Vector<float,3>*>(&verts[0]), 2*sizeof(Vector3<float>), m_Weights->getSize());
	setBounds(Bounds3<float>(box));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void BoneModel::onRender(const ParameterBuffer* pbuf) const {
	ASSERT(pbuf != NULL);
	MatrixStack* ms = getContext()->getMatrixStack();
	ms->selectMode(MatrixStack::ModelView);
	ms->set(*(const Matrix4<float>*)pbuf->getBuffer());

	// Get pointers to animation data
	const void* anim_data = (const byte*)pbuf->getBuffer() + 64;

	// Transform frame
	Vector3<float>* dst = (Vector3<float>*)m_Vertices->lock(VertexBuffer::Discard);
	m_Weights->transform(dst, (const Matrix4<float>*)anim_data);
	m_Vertices->unlock();

	// Select vertex buffer based on current anim
	m_Vertices->select();
	
	// Render all meshes in this meshgroup
	MeshList::Iterator mend = getMeshes().end();
	for (MeshList::Iterator miter = getMeshes().begin(); miter != mend; ++miter) {
		(*miter)->render();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void BoneModel::onTransform(const ParameterBuffer* pbuf,Matrix4<float>* dst,Joint::Handle hnd) const {
	*dst = *(const Matrix4<float>*)((const byte*)pbuf->getBuffer() + (unsigned int)hnd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool BoneModel::onOptimize() {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void BoneModel::onClear() {
	m_Weights.setNull();
	m_Skeleton.setNull();
	m_Vertices.setNull();
	Model::onClear();
}

}; };
