/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "system.thread.h"
#include "graphics3d.context.h"
#include "graphics3d.vertexformat.h"
#include "graphics3d.texture2d.h"

#include "object.inl"
#include "base.string.inl"
#include "system.thread.inl"
#include "container.iterator.inl"
#include "container.chain.inl"
#include "container.allocator.inl"
#include "graphics.rectangle.inl"
#include "graphics3d.context.inl"
#include "graphics3d.lightstack.inl"
#include "graphics3d.matrixstack.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.fixedshader.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.framebuffer.inl"
#include "graphics3d.material.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.texture2d.inl"
#include "graphics3d.cubetexture.inl"


namespace magnum { namespace graphics3d {

Context::Startup Context::g_Startup;
Chain<Context> Context::g_Contexts;

#if defined(__NATIVETLS__)
	__declspec(thread) Context*	Context::g_Current = NULL;
#else
	Thread::Storage	Context::g_Current;
#endif


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Context::Startup::Startup() {
#if defined(__NATIVETLS__)

#else
	Context::g_Current = Thread::createTLS();
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Context::Context() : 
	m_DefaultShader(FixedShader::Plain),
	m_DefaultFrameBuffer(0), 
	m_DeferQueue(this),
	m_FrameBuffer(&m_DefaultFrameBuffer),
	m_Shader(NULL),
	m_Material(NULL),
	m_StencilMode(NULL),
	m_BlendMode(NULL),
	m_DepthMode(NULL),
	m_PolygonMode(NULL),
	m_Attributes(0) {
	
	for (int i = 0; i < Texture::MaxTextures; i++) {
		m_Texture[i] = NULL;
		m_OglTextureTarget[i] = GL_TEXTURE_2D;
	}
	
	static bool glew = false;
	if (!glew) {
		glewExperimental = GL_FALSE;
		glewInit();
		glew = true;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Context::~Context() {
	wglDeleteContext(m_OglContext);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Context* Context::create() {
	Context* c = new Context();
	if (c == NULL)
		return NULL;

	// Here we should create a device context etc...


	g_Contexts.insert(c);

	return c;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Context* Context::share() const {
	Context* context = new Context();
	context->m_DeviceHandle = m_DeviceHandle;
	context->m_WindowHandle = m_WindowHandle;
	context->m_OglContext = wglCreateContext(m_DeviceHandle);
	if (context->m_OglContext == NULL) {
		delete context;
		return NULL;
	}
	
	wglShareLists(m_OglContext, context->m_OglContext);
	
	g_Contexts.insert(context);
	
	return context;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Context::makeCurrent() {
#if defined(__FORWINDOWS__)
	BOOL ret = wglMakeCurrent(m_DeviceHandle, m_OglContext);
	if (ret == FALSE)
		return false;
		
	RECT rect;
	GetWindowRect(m_WindowHandle, &rect);
	m_DefaultFrameBuffer.m_Size.setWidth(rect.right - rect.left);
	m_DefaultFrameBuffer.m_Size.setHeight(rect.bottom - rect.top);
	if (m_FrameBuffer == &m_DefaultFrameBuffer) 
		m_Size = m_DefaultFrameBuffer.getSize();
#endif

#if defined(__NATIVETLS__)
	g_Current = this;
#else
	Thread::setTLS(g_Current, this);
#endif

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Context* Context::searchCurrent() {
#if defined(__FORWINDOWS__)	
	HGLRC glc = wglGetCurrentContext();
	HDC dc = wglGetCurrentDC();
#elif defined(__FORUNIX__)
#endif

	magnum::container::Chain<Context>::Iterator iter = g_Contexts.begin();
	magnum::container::Chain<Context>::Iterator end = g_Contexts.end();
	while (iter != end) {
		if (glc == iter->m_OglContext)
			break;
		iter++;
	}

	// Didn't we find the thread?
	if (iter == end)
		return createCurrent();

#if defined(__NATIVETLS__)
	g_Current = &(*iter);
#else
	Thread::setTLS(g_Current, &(*iter));
#endif

	return &(*iter);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Context* Context::createCurrent() {
	// Create a new Thread object for the current thread...
	Context *c = new Context();

#if defined(__FORWINDOWS__)	
	HGLRC glc = wglGetCurrentContext();
	HDC dc = wglGetCurrentDC();
	c->m_DeviceHandle = dc;
	c->m_OglContext = glc;
	c->m_WindowHandle = WindowFromDC(dc);

	RECT rect;
	GetWindowRect(c->m_WindowHandle, &rect);
	c->m_DefaultFrameBuffer.m_Size.setWidth(rect.right - rect.left);
	c->m_DefaultFrameBuffer.m_Size.setHeight(rect.bottom - rect.top);
	c->m_Size = c->m_DefaultFrameBuffer.getSize();
#elif defined(__FORUNIX__)
#endif

	g_Contexts.insert(c);

#if defined(__NATIVETLS__)
	g_Current = c;
#else
	Thread::setTLS(g_Current, c);
#endif

	return c;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::copyTo(Texture2D* tex,unsigned int level, const Point<int>& p,const Rectangle<int>& r) const {
	ASSERT(tex != NULL);
	ASSERT(level < tex->getLevels());

	GLint oldtex;
	glGetIntegerv(GL_TEXTURE_BINDING_2D, &oldtex);
	glBindTexture(GL_TEXTURE_2D, tex->m_OglTexture);
	glCopyTexSubImage2D(GL_TEXTURE_2D, level, 
		p.getX(), tex->getHeight() - (p.getY() + r.getHeight()), 
		r.getX(), m_Size.getHeight() - (r.getY() + r.getHeight()), 
		r.getWidth(), r.getHeight());

	glBindTexture(GL_TEXTURE_2D, oldtex);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::copyTo(CubeTexture* tex,CubeTexture::Face face, unsigned int level, const Point<int>& p,const Rectangle<int>& r) const {
	ASSERT(tex != NULL);
	ASSERT(level < tex->getLevels());

	GLint oldtex;
	glGetIntegerv(GL_TEXTURE_BINDING_CUBE_MAP, &oldtex);
	glBindTexture(GL_TEXTURE_CUBE_MAP, tex->m_OglTexture);
	glCopyTexSubImage2D(face, level, 
		p.getX(), tex->getHeight() - (p.getY() + r.getHeight()), 
		r.getX(), m_Size.getHeight() - (r.getY() + r.getHeight()), 
		r.getWidth(), r.getHeight());

	glBindTexture(GL_TEXTURE_CUBE_MAP, oldtex);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::copyTo(Image* img, const Point<int>& p,const Rectangle<int>& r) const {
	ASSERT(img != NULL);
	ASSERT((img->getWidth() - p.getX()) >= r.getWidth());
	ASSERT((img->getHeight() - p.getX()) >= r.getHeight());

	int type;
	int format;
	int psize;

	switch (img->getType()) {
		case Image::RGB8:
			type = GL_UNSIGNED_BYTE;
			format = GL_RGB;
			psize = 3;
			break;
		case Image::RGBA8:
			type = GL_UNSIGNED_BYTE;
			format = GL_RGBA;
			psize = 4;
			break;
		case Image::BGR8:
			type = GL_UNSIGNED_BYTE;
			format = GL_BGR_EXT;
			psize = 3;
			break;
		case Image::BGRA8:
			type = GL_UNSIGNED_BYTE;
			format = GL_BGRA_EXT;
			psize = 4;
			break;
		case Image::Index8:
			type = GL_UNSIGNED_BYTE;
			format = 1;
			psize = 1;
			break;
		default:
			return;
	}

	void* data = img->lock();
	glPixelStorei(GL_PACK_SWAP_BYTES , false);
	glPixelStorei(GL_PACK_ROW_LENGTH, img->getWidth());
	glPixelStorei(GL_PACK_SKIP_PIXELS, p.getX());
	glPixelStorei(GL_PACK_SKIP_ROWS, img->getHeight() - (p.getY() + r.getHeight()));
	glPixelStorei(GL_PACK_ALIGNMENT, 1); 
	glReadPixels(r.getX(), m_Size.getHeight() - (r.getY() + r.getHeight()), 
		r.getWidth(), r.getHeight(), 
		format, type, data);

	// Now we still have to flip the image
	int i;
	int len = psize * r.getWidth();
	int lsize = img->getLinesize();
	byte* upos = (byte*)data + p.getY() * lsize + p.getX()*psize;
	byte* lpos = upos + (r.getHeight() - 1) * lsize;
	byte flipbuffer[4096*4];
	for (i = 0; i < (r.getHeight() >> 1); i++) {
		memcpy(flipbuffer, lpos, len);
		memcpy(lpos, upos, len);
		memcpy(upos, flipbuffer, len);
		lpos -= lsize;
		upos += lsize;
	}

	img->unlock();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::copyFrom(const Image* img, const Point<int>& p,const Rectangle<int>& r) {
	ASSERT(img != NULL);

	void* data = ((Image *)img)->lock();
	glPixelZoom(1.0f, -1.0f);
	glPixelStorei(GL_UNPACK_SWAP_BYTES , false);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, img->getWidth());
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, p.getX());
	glPixelStorei(GL_UNPACK_SKIP_ROWS, img->getHeight() - (p.getY() + r.getHeight()));
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); 

	int format, type;
	switch (img->getType()) {
		case Image::RGB8:
			format = GL_RGB;
			type = GL_UNSIGNED_BYTE;
			break;
		case Image::Index8:
			format = GL_COLOR_INDEX;
			type = GL_UNSIGNED_BYTE;
			break;
		default:
			format = GL_COLOR_INDEX;
			type = GL_UNSIGNED_BYTE;
	}

	GLint mmode;
	glGetIntegerv(GL_MATRIX_MODE, &mmode);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glRasterPos2i(r.getX(), m_Size.getHeight() - (r.getY() + r.getHeight()));
	glDrawPixels(r.getWidth(), r.getHeight(), format, type, data);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glMatrixMode(mmode);
	((Image *)img)->unlock();
}

}; };
