/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "graphics3d.context.h"
#include "graphics3d.cubemapgenerator.h"
#include "graphics3d.exception.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "math.matrix.inl"
#include "graphics.point.inl"
#include "graphics.rectangle.inl"
#include "graphics3d.context.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.cubemapgenerator.inl"
#include "graphics3d.cubetexture.inl"
#include "graphics3d.framebuffer.inl"
#include "graphics3d.rendertarget.inl"


namespace magnum { namespace graphics3d {
using magnum::object::AutoPointer;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
CubemapGenerator::CubemapGenerator() {
	// Initialize camera	
	m_Camera.setNearPlane(5);
	m_Camera.setProjectionPlane(5);
	m_Camera.setFarPlane(5000);
	m_Camera.setViewWindow(Rectangle<float>(-5.0f,-5.0f,10.0f,10.0f));

	// Set framebuffer pointers to 0
	for (unsigned int i = 0; i <6; i++)
		m_FrameBuffer[i] = NULL;
	m_Mode = Auto;

	// Initializse Cubemap transforms.
	initTransforms();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
CubemapGenerator::CubemapGenerator(CubeTexture* tex,Mode mode) {
	// Initialize camera	
	m_Camera.setNearPlane(5);
	m_Camera.setProjectionPlane(5);
	m_Camera.setFarPlane(5000);
	m_Camera.setViewWindow(Rectangle<float>(-5.0f,-5.0f,10.0f,10.0f));

	// Set framebuffer pointers to 0
	for (unsigned int i = 0; i <6; i++)
		m_FrameBuffer[i] = NULL;
	m_Mode = mode;

	// Initializse Cubemap transforms.
	initTransforms();

	// Attach texture
	attachTexture(tex, mode);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
CubemapGenerator::~CubemapGenerator() {
	detachTexture();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void CubemapGenerator::attachTexture(CubeTexture* tex,Mode mode) {
	ASSERT(tex != NULL);
	if ((m_Texture == tex) && (m_Mode == mode))
		return;
	
	// Remove old texture and frame buffers
	detachTexture();
	
	m_Texture = tex;
	unsigned int twidth = tex->getWidth();
	unsigned int theight = tex->getHeight();
	
	// Get mode
	switch (mode) {
		case Compatible:
			m_Mode = Compatible;
			break;
		case Fast:
			if (!RenderTarget::isSupported() || !FrameBuffer::isSupported())
				throw UnsupportedFeatureException();
			m_Mode = Fast;
			break;	
		case Auto:
			if (RenderTarget::isSupported() && FrameBuffer::isSupported())
				m_Mode = Fast;
			else
				m_Mode = Compatible;
			break;
		default:
			throw InvalidModeException();
	}

	// Create framebuffers
	if (m_Mode == Fast) {
		AutoPointer<RenderTarget> rt_px = RenderTarget::create(m_Texture, CubeTexture::PositiveX);
		AutoPointer<RenderTarget> rt_nx = RenderTarget::create(m_Texture, CubeTexture::NegativeX);
		AutoPointer<RenderTarget> rt_py = RenderTarget::create(m_Texture, CubeTexture::PositiveY);
		AutoPointer<RenderTarget> rt_ny = RenderTarget::create(m_Texture, CubeTexture::NegativeY);
		AutoPointer<RenderTarget> rt_pz = RenderTarget::create(m_Texture, CubeTexture::PositiveZ);
		AutoPointer<RenderTarget> rt_nz = RenderTarget::create(m_Texture, CubeTexture::NegativeZ);
		AutoPointer<RenderTarget> rt_depth = RenderTarget::create(RenderTarget::Depth24, twidth, theight);
		m_FrameBuffer[0] = FrameBuffer::create();
		m_FrameBuffer[0]->attach(rt_px, RenderTarget::ColourBuffer);
		m_FrameBuffer[0]->attach(rt_depth, RenderTarget::DepthBuffer);
		m_FrameBuffer[1] = FrameBuffer::create();
		m_FrameBuffer[1]->attach(rt_nx, RenderTarget::ColourBuffer);
		m_FrameBuffer[1]->attach(rt_depth, RenderTarget::DepthBuffer);
		m_FrameBuffer[2] = FrameBuffer::create();
		m_FrameBuffer[2]->attach(rt_py, RenderTarget::ColourBuffer);
		m_FrameBuffer[2]->attach(rt_depth, RenderTarget::DepthBuffer);
		m_FrameBuffer[3] = FrameBuffer::create();
		m_FrameBuffer[3]->attach(rt_ny, RenderTarget::ColourBuffer);
		m_FrameBuffer[3]->attach(rt_depth, RenderTarget::DepthBuffer);
		m_FrameBuffer[4] = FrameBuffer::create();
		m_FrameBuffer[4]->attach(rt_pz, RenderTarget::ColourBuffer);
		m_FrameBuffer[4]->attach(rt_depth, RenderTarget::DepthBuffer);
		m_FrameBuffer[5] = FrameBuffer::create();
		m_FrameBuffer[5]->attach(rt_nz, RenderTarget::ColourBuffer);
		m_FrameBuffer[5]->attach(rt_depth, RenderTarget::DepthBuffer);
	} 
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void CubemapGenerator::detachTexture() {
	for (unsigned int i = 0; i < 6; i++) {
		if (m_FrameBuffer[i] != NULL) {
			m_FrameBuffer[i]->release();
			m_FrameBuffer[i] = NULL;
		}
	}
	m_Texture = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void CubemapGenerator::render(const Matrix4<float>& mtx) {
	m_Camera.select();
	if (m_Mode == Compatible) {
		Point<float> p(0,0);
		Rectangle<float> r(0,0,(float)m_Texture->getWidth(),(float)m_Texture->getHeight());
		Context* c = getContext();
		Rectangle<float> oldviewport = c->getViewport();
		c->setViewport(r);
		onRender(m_Transform[0]*mtx, CubeTexture::PositiveX);
		c->copyTo(m_Texture, CubeTexture::PositiveX, 0, p, r);
		onRender(m_Transform[1]*mtx, CubeTexture::NegativeX);
		c->copyTo(m_Texture, CubeTexture::NegativeX, 0, p, r);
		onRender(m_Transform[2]*mtx, CubeTexture::PositiveY);
		c->copyTo(m_Texture, CubeTexture::PositiveY, 0, p, r);
		onRender(m_Transform[3]*mtx, CubeTexture::NegativeY);
		c->copyTo(m_Texture, CubeTexture::NegativeY, 0, p, r);
		onRender(m_Transform[4]*mtx, CubeTexture::PositiveZ);
		c->copyTo(m_Texture, CubeTexture::PositiveZ, 0, p, r);
		onRender(m_Transform[5]*mtx, CubeTexture::NegativeZ);
		c->copyTo(m_Texture, CubeTexture::NegativeZ, 0, p, r);
		c->setViewport(oldviewport);
	} else {
		const FrameBuffer* oldfb = getContext()->getFrameBuffer();
		m_FrameBuffer[0]->select();
		onRender(m_Transform[0]*mtx, CubeTexture::PositiveX);
		m_FrameBuffer[1]->select();
		onRender(m_Transform[1]*mtx, CubeTexture::NegativeX);
		m_FrameBuffer[2]->select();
		onRender(m_Transform[2]*mtx, CubeTexture::PositiveY);
		m_FrameBuffer[3]->select();
		onRender(m_Transform[3]*mtx, CubeTexture::NegativeY);
		m_FrameBuffer[4]->select();
		onRender(m_Transform[4]*mtx, CubeTexture::PositiveZ);
		m_FrameBuffer[5]->select();
		onRender(m_Transform[5]*mtx, CubeTexture::NegativeZ);
		oldfb->select();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void CubemapGenerator::onRender(const Matrix4<float>& ,CubeTexture::Face ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void CubemapGenerator::initTransforms() {
	// Initializse Cubemap transforms.
	m_Transform[0].null();
	m_Transform[0][2][0] = -1.0f;
	m_Transform[0][1][1] = -1.0f;
	m_Transform[0][0][2] = -1.0f;
	m_Transform[0][3][3] = 1.0f;
	m_Transform[1].null();
	m_Transform[1][2][0] = 1.0f;
	m_Transform[1][1][1] = -1.0f;
	m_Transform[1][0][2] = 1.0f;
	m_Transform[1][3][3] = 1.0f;

	m_Transform[2].null();
	m_Transform[2][0][0] = 1.0f;
	m_Transform[2][2][1] = -1.0f;
	m_Transform[2][1][2] = 1.0f;
	m_Transform[2][3][3] = 1.0f;
	m_Transform[3].null();
	m_Transform[3][0][0] = 1.0f;
	m_Transform[3][2][1] = 1.0f;
	m_Transform[3][1][2] = -1.0f;
	m_Transform[3][3][3] = 1.0f;

	m_Transform[4].null();
	m_Transform[4][0][0] = 1.0f;
	m_Transform[4][1][1] = -1.0f;
	m_Transform[4][2][2] = -1.0f;
	m_Transform[4][3][3] = 1.0f;
	m_Transform[5].null();
	m_Transform[5][0][0] = -1.0f;
	m_Transform[5][1][1] = -1.0f;
	m_Transform[5][2][2] = 1.0f;
	m_Transform[5][3][3] = 1.0f;
}

}; };

