/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "system.thread.h"
#include "graphics.filter.resize.h"
#include "graphics3d.cubetexture.h"
#include "graphics3d.exception.h"
#include "graphics3d.context.h"
#include "graphics3d.vertexformat.h"

#include "object.inl"
#include "base.string.inl"
#include "object.exception.inl"
#include "system.thread.inl"
#include "io.filter.inl"
#include "container.chain.inl"
#include "graphics.image.inl"
#include "graphics.dimension.inl"
#include "graphics.rectangle.inl"
#include "graphics.filter.resize.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.cubetexture.inl"
#include "graphics3d.context.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.vertexformat.inl"


namespace magnum { namespace graphics3d {
using magnum::graphics::FilterResize;

GLenum CubeTexture::g_Faces[6] = {
	GL_TEXTURE_CUBE_MAP_POSITIVE_X,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
	GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
	GL_TEXTURE_CUBE_MAP_POSITIVE_Z,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
};

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline static int roundUp2n(int in) {
	int out = 1;
	while (in > 1) {
		if (in & 1)
			in++;
		in >>= 1;
		out <<= 1;
	}

	return out;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
static Dimension<int> adjustSize(const Dimension<int>& in) {
	Dimension<int> size;

	size.setWidth(roundUp2n(in.getWidth()));
	size.setHeight(roundUp2n(in.getHeight()));

	// Check for min/max size etc

	return size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
CubeTexture* CubeTexture::create(Type t,Dimension<int>& d, int levels) {
	int type = mapInputType(t);
	int iformat = mapInputFormat(t);
	int oformat = mapOutputFormat(t);
	if ((oformat == -1) || (iformat == -1) || (type == -1))
		throw TypeException();

	Dimension<int> size = adjustSize(d);

	GLuint tex;
	glGetError();

	glGenTextures(1, &tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();
		
	glBindTexture(GL_TEXTURE_CUBE_MAP, tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	if (levels == GenerateAll)
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, TRUE);
	else
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, FALSE);

	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 0, oformat, size.getWidth(), size.getHeight(), 0, iformat, type, NULL);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X, 0, oformat, size.getWidth(), size.getHeight(), 0, iformat, type, NULL);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y, 0, oformat, size.getWidth(), size.getHeight(), 0, iformat, type, NULL);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, 0, oformat, size.getWidth(), size.getHeight(), 0, iformat, type, NULL);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z, 0, oformat, size.getWidth(), size.getHeight(), 0, iformat, type, NULL);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 0, oformat, size.getWidth(), size.getHeight(), 0, iformat, type, NULL);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	CubeTexture* texture = new CubeTexture(tex);

	return texture;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
CubeTexture* CubeTexture::create(Type t,unsigned int w, unsigned int h, int levels) {
	return create(t, Dimension<int>(w,h), levels);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
CubeTexture* CubeTexture::create(const Image* img, int levels) {
	// Map image format to OpenGL formats
	int type = mapInputType(img->getType());
	int iformat = mapInputFormat(img->getType());
	int oformat = mapOutputFormat(img->getType());
	if ((oformat == -1) || (iformat == -1) || (type == -1))
		throw TypeException();
	
	// Setup transfer mode
	if (img->getType() == Image::Index8) {
		glPixelTransferi(GL_MAP_COLOR, true);
		glPixelTransferi(GL_MAP_STENCIL, false);
		glPixelTransferi(GL_INDEX_SHIFT, 0);
		glPixelTransferi(GL_INDEX_OFFSET, 0);
	} else {
		glPixelTransferi(GL_MAP_COLOR, false);
		glPixelTransferi(GL_MAP_STENCIL, false);
	}

	glPixelStorei(GL_UNPACK_SWAP_BYTES , false);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); 

	Dimension<int> size = adjustSize(img->getSize());

	unsigned int tex;
	glGetError();

	glGenTextures(1, &tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();
		
	glBindTexture(GL_TEXTURE_CUBE_MAP, tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	if (levels == GenerateAll) {
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, TRUE);
		levels = 1;
	} else if (levels == CopyAll) {
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, FALSE);
		levels = 256;
	}

	Image* root;

	// Check if we need to resize the image
	if (size != img->getSize()) {
		FilterResize filter(size);
		root = filter << img;
	} else {
		root = (Image *)img;
		root->refer();
	}
	img = root;

	int level = 0;
	while ((img != NULL) && (level < levels)) {
		int ls = img->getLinesize();

		// Transfer image into texture memory
		for (unsigned int i = 0; i < 6; i++) {
			if (img == NULL)
				throw TextureException();
			byte* src = (byte *)((Image *)img)->lock();
			glTexImage2D(g_Faces[i], level, oformat, img->getWidth(), img->getHeight(), 0, iformat, type, src);
			((Image *)img)->unlock();
			img = img->getNext();
		}

		level++;
	}

	// Release any reference to original image
	root->release();

	// Some final error checking
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	CubeTexture* texture = new CubeTexture(tex);

	return texture;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
CubeTexture::CubeTexture(unsigned int tex) : Texture(tex, GL_TEXTURE_CUBE_MAP) {
	int tmp;

	m_Locked = false;
	glGetTexLevelParameteriv(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 0, GL_TEXTURE_WIDTH, &tmp);
	m_Size.setWidth(tmp);
	glGetTexLevelParameteriv(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 0, GL_TEXTURE_HEIGHT, &tmp);
	m_Size.setHeight(tmp);

	m_WrapS = ClampEdge;
	m_WrapT = ClampEdge;
	m_WrapU = ClampEdge;
	glTexParameteri(m_OglTarget, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(m_OglTarget, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(m_OglTarget, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(m_OglTarget, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(m_OglTarget, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
CubeTexture::~CubeTexture() {
	glDeleteTextures(1, &m_OglTexture);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool CubeTexture::update(const Image* img, int level) {
	ASSERT(img != NULL);
	
	// Map image format to OpenGL formats
	int type = mapInputType(img->getType());
	int iformat = mapInputFormat(img->getType());
	if ((iformat == -1) || (type == -1))
		throw TypeException();
	
	// Setup transfer mode
	if (img->getType() == Image::Index8) {
		glPixelTransferi(GL_MAP_COLOR, true);
		glPixelTransferi(GL_MAP_STENCIL, false);
		glPixelTransferi(GL_INDEX_SHIFT, 0);
		glPixelTransferi(GL_INDEX_OFFSET, 0);
	} else {
		glPixelTransferi(GL_MAP_COLOR, false);
		glPixelTransferi(GL_MAP_STENCIL, false);
	}

	glPixelStorei(GL_UNPACK_SWAP_BYTES , false);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); 

	int oldtex;
	glGetIntegerv(GL_TEXTURE_BINDING_CUBE_MAP, &oldtex);
	glBindTexture(GL_TEXTURE_CUBE_MAP, m_OglTexture);

	// Transfer image into texture memory
	for (unsigned int i = 0; i < 6; i++) {
		if (img == NULL)
			throw TextureException();
		byte* src = (byte *)((Image *)img)->lock();
		glTexSubImage2D(g_Faces[i], level, 0, 0, img->getWidth(), img->getHeight(), iformat, type, src);
		((Image *)img)->unlock();
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool CubeTexture::copyFrom(const Image* img, const Point<int>& p,const Rectangle<int>& r) {
	ASSERT(img != NULL);
	
	int type = GL_UNSIGNED_BYTE;

	// Map image format to OpenGL formats
	int iformat = mapInputFormat(img->getType());
	if (iformat == -1)
		throw TypeException();
	
	// Setup transfer mode
	if (img->getType() == Image::Index8) {
		glPixelTransferi(GL_MAP_COLOR, true);
		glPixelTransferi(GL_MAP_STENCIL, false);
		glPixelTransferi(GL_INDEX_SHIFT, 0);
		glPixelTransferi(GL_INDEX_OFFSET, 0);
	} else {
		glPixelTransferi(GL_MAP_COLOR, false);
		glPixelTransferi(GL_MAP_STENCIL, false);
	}

	glPixelStorei(GL_UNPACK_SWAP_BYTES , false);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, img->getWidth());
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, r.getX());
	glPixelStorei(GL_UNPACK_SKIP_ROWS, r.getY());
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); 

	glBindTexture(GL_TEXTURE_CUBE_MAP, m_OglTexture);

	int level = 0;

	// Transfer image into texture memory
	byte* src = (byte *)((Image *)img)->lock();
	glTexSubImage2D(GL_TEXTURE_CUBE_MAP, level, p.getX(), p.getY(), r.getWidth(), r.getHeight(), iformat, type, src);
	((Image *)img)->unlock();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool CubeTexture::copyFrom(const Image* img, const Point<int>& p) {
	Rectangle<int> rect;
	rect.setOrigin(0,0);
	rect.setSize(img->getSize());
	return copyFrom(img, p, rect);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void* CubeTexture::lock(Face face,int level, const Rectangle<int>& , LockFlag ) {
	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void CubeTexture::unlock() {

}


}; };
