/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics3d.context.h"
#include "graphics3d.framebuffer.h"
#include "graphics3d.exception.h"
#include "graphics3d.texture1d.h"
#include "graphics3d.texture2d.h"
#include "graphics3d.vertexformat.h"

#include "object.inl"
#include "base.string.inl"
#include "container.iterator.inl"
#include "container.array.inl"
#include "graphics.point.inl"
#include "graphics.rectangle.inl"
#include "graphics.dimension.inl"
#include "graphics.image.inl"
#include "graphics3d.framebuffer.inl"
#include "graphics3d.rendertarget.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.texture1d.inl"
#include "graphics3d.texture2d.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.context.inl"


namespace magnum { namespace graphics3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool FrameBuffer::isSupported() {
	return (glGenFramebuffersEXT != NULL);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
FrameBuffer* FrameBuffer::create() {
	GLuint buffer;
	if (glGenFramebuffersEXT == NULL)
		throw UnsupportedFeatureException();
	glGenFramebuffersEXT(1, &buffer);
	if (glGetError() != GL_NO_ERROR)
		throw MemoryException();
	
	return new FrameBuffer(buffer);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
FrameBuffer::FrameBuffer(GLuint buffer) {
	m_OglFramebuffer = buffer;
	m_Size.setWidth(0);
	m_Size.setHeight(0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
FrameBuffer::~FrameBuffer() {
	detachAll();
		
	if (m_OglFramebuffer != 0) {
		glDeleteFramebuffersEXT(1, &m_OglFramebuffer);
		m_OglFramebuffer = 0;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool FrameBuffer::isValid() const {
	if (m_OglFramebuffer == 0)
		return true;
	if (glCheckFramebufferStatusEXT == 0)
		return false;
	GLenum status;
	GLint oldfb;
	glGetIntegerv(GL_FRAMEBUFFER_BINDING_EXT, &oldfb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_BINDING_EXT, m_OglFramebuffer);
	status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	glBindFramebufferEXT(GL_FRAMEBUFFER_BINDING_EXT, oldfb);
	return (status != GL_FRAMEBUFFER_COMPLETE_EXT);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameBuffer::select() const {
	Context* context = getContext();
	FrameBuffer* fb = const_cast<FrameBuffer*>(context->m_FrameBuffer);
	if (fb == this)
		return;

	// Save old viewport
	if (fb != NULL) {
		GLint vp[4];
		glGetIntegerv(GL_VIEWPORT, vp);
		fb->m_Viewport.setOrigin(vp[0],vp[1]);
		fb->m_Viewport.setWidth(vp[2]);
		fb->m_Viewport.setHeight(vp[3]);
	}

	// Set new FrameBuffer in context
	context->m_FrameBuffer = this;
	context->m_Size = m_Size;

	if ((context->m_FrameBuffer == &context->m_DefaultFrameBuffer) && (glBindFramebufferEXT == NULL))
		return;
	ASSERT(glBindFramebufferEXT != NULL);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_OglFramebuffer);
	glViewport(m_Viewport.getOrigin().getX(), m_Viewport.getOrigin().getY(), m_Viewport.getWidth(), m_Viewport.getHeight());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameBuffer::forceSelect() const {
	Context* context = getContext();
	FrameBuffer* fb = const_cast<FrameBuffer*>(context->m_FrameBuffer);
	
	// Save old viewport
	if (fb != NULL) {
		GLint vp[4];
		glGetIntegerv(GL_VIEWPORT, vp);
		fb->m_Viewport.setOrigin(vp[0],vp[1]);
		fb->m_Viewport.setWidth(vp[2]);
		fb->m_Viewport.setHeight(vp[3]);
	}

	// Set new FrameBuffer in context
	context->m_FrameBuffer = this;
	context->m_Size = m_Size;

	if ((context->m_FrameBuffer == &context->m_DefaultFrameBuffer) && (glBindFramebufferEXT == NULL))
		return;
	ASSERT(glBindFramebufferEXT != NULL);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_OglFramebuffer);
	glViewport(m_Viewport.getOrigin().getX(), m_Viewport.getOrigin().getY(), m_Viewport.getWidth(), m_Viewport.getHeight());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameBuffer::deselect() const {
	Context* context = getContext();
	if (context->m_FrameBuffer != this)
		return;
		
	// Save old viewport
	GLint vp[4];
	glGetIntegerv(GL_VIEWPORT, vp);
	const_cast<FrameBuffer*>(this)->m_Viewport.setOrigin(vp[0],vp[1]);
	const_cast<FrameBuffer*>(this)->m_Viewport.setWidth(vp[2]);
	const_cast<FrameBuffer*>(this)->m_Viewport.setHeight(vp[3]);
	
	// Set new FrameBuffer in context
	FrameBuffer* fb = &context->m_DefaultFrameBuffer;
	context->m_FrameBuffer = fb;
	context->m_Size = fb->m_Size;

	// Bind default frame buffer
	if (glBindFramebufferEXT == NULL)
		return;
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glViewport(fb->m_Viewport.getOrigin().getX(), fb->m_Viewport.getOrigin().getY(), fb->m_Viewport.getWidth(), fb->m_Viewport.getHeight());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameBuffer::selectDefault() {
	Context* context = getContext();
	FrameBuffer* fb = const_cast<FrameBuffer*>(context->m_FrameBuffer);
	
	if (fb == &context->m_DefaultFrameBuffer)
		return;

	// Save old viewport
	if (context->m_FrameBuffer != NULL) {
		GLint vp[4];
		glGetIntegerv(GL_VIEWPORT, vp);
		fb->m_Viewport.setOrigin(vp[0],vp[1]);
		fb->m_Viewport.setWidth(vp[2]);
		fb->m_Viewport.setHeight(vp[3]);
	}
	
	// Set new FrameBuffer in context
	fb = &context->m_DefaultFrameBuffer;
	context->m_FrameBuffer = fb;
	context->m_Size = fb->m_Size;

	// Bind default frame buffer
	if (glBindFramebufferEXT == NULL)
		return;
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glViewport(fb->m_Viewport.getOrigin().getX(), fb->m_Viewport.getOrigin().getY(), fb->m_Viewport.getWidth(), fb->m_Viewport.getHeight());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameBuffer::attach(RenderTarget* rt,RenderTarget::Attachment att) {
	ASSERT(rt != NULL);
	rt->refer();
	for (Iterator iter = m_Bindings.begin(); iter != m_Bindings.end(); ++iter) {
		if (iter->getAttachment() == att) {
			att = iter->getAttachment();
			iter->getTarget()->release();			
			m_Bindings.remove(iter);
			break;		
		}
	}
	m_Bindings.insert(Binding(rt,att));
	rt->onAttach(this, att);
	m_Size = rt->getSize();
	m_Viewport.setOrigin(0,0);
	m_Viewport.setSize(m_Size);
	if (glGetError() != GL_NO_ERROR)
		throw FrameBufferException();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameBuffer::detach(RenderTarget* rt) {
	ASSERT(rt != NULL);
	RenderTarget::Attachment att;
	Iterator iter;
	for (iter = m_Bindings.begin(); iter != m_Bindings.end(); ++iter) {
		if (iter->getTarget() == rt) {
			att = iter->getAttachment();
			m_Bindings.remove(iter);
			break;		
		}
	}
	if (iter == end())
		return;
	rt->onDetach(this, att);
	rt->release();
	if (m_Bindings.isEmpty()) {
		m_Size.setWidth(0);
		m_Size.setHeight(0);
	} else {
		m_Size = m_Bindings[0].getTarget()->getSize();
	}
	m_Viewport.setOrigin(0,0);
	m_Viewport.setSize(m_Size);
	if (glGetError() != GL_NO_ERROR)
		throw FrameBufferException();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameBuffer::detachAll() {
	for (Iterator iter = m_Bindings.begin(); iter != m_Bindings.end(); ++iter) {
		iter->getTarget()->onDetach(this, iter->getAttachment());
	}
	m_Bindings.clear();
	m_Size.setWidth(0);
	m_Size.setHeight(0);
	m_Viewport.setOrigin(0,0);
	m_Viewport.setSize(m_Size);
}

}; };
