/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics3d.framemodel.h"
#include "graphics3d.mesh.h"
#include "graphics3d.meshgroup.h"
#include "graphics3d.material.h"
#include "graphics3d.shader.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.appearance.h"
#include "graphics3d.renderstream.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.context.h"
#include "graphics3d.bounds.h"
#include "graphics3d.parameter.h"
#include "graphics3d.parameterbuffer.h"
#include "graphics3d.frameanimation.h"

#include "object.pointer.inl"
#include "math.matrix.inl"
#include "math.vector.inl"
#include "container.array.inl"
#include "container.list.inl"
#include "container.allocator.inl"
#include "container.iterator.inl"
#include "geometry.bounds.inl"
#include "geometry.sphere.inl"
#include "geometry.box.inl"
#include "geometry.alignedbox.inl"
#include "graphics3d.model.inl"
#include "graphics3d.joint.inl"
#include "graphics3d.framemodel.inl"
#include "graphics3d.meshgroup.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.renderstream.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.matrixstack.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.parameterbuffer.inl"
#include "graphics3d.parameterformat.inl"
#include "graphics3d.bounds.inl"
#include "graphics3d.animation.inl"
#include "graphics3d.frameanimation.inl"


namespace magnum { namespace graphics3d {
using magnum::geometry::AlignedBox3;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
FrameModel::FrameModel() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
FrameModel::~FrameModel() {
	clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::finish() {
	// Create parameter format
	ParameterFormat fmt;
	fmt.insert(Parameters::ModelView, Types::Matrix4f);
	fmt.insert(Parameters::AnimationData, Types::Struct);
	for (List<Group>::Iterator iter = m_Groups.begin(); iter != m_Groups.end(); iter++) {
		fmt.insert(Parameters::BlendIndex0, Types::Integer);
		fmt.insert(Parameters::BlendIndex1, Types::Integer);
		fmt.insert(Parameters::BlendWeight, Types::Float);
	}
	publishParameters(fmt);

	// Publish meshes
	clearMeshes();
	for (List<Group>::Iterator giter = m_Groups.begin(); giter != m_Groups.end(); giter++) {
		MeshGroup::Iterator miter;
		for (miter = giter->m_MeshGroup->begin(); miter != giter->m_MeshGroup->end(); miter++)
			publishMesh(*miter);
	}

	// Create additional vertex buffers for interpolated frames
	VertexFormat vfmt;
	vfmt.set(0, Types::Float3, Attributes::Vertex);
	vfmt.set(1, Types::Float3, Attributes::Normal);
	for (List<Group>::Iterator iter = m_Groups.begin(); iter != m_Groups.end(); iter++) {
		if (iter->m_Frames.getSize() == 0)
			continue;
		unsigned int size = iter->m_Frames[0]->getSize();
		iter->m_Vertices = VertexBuffer::create(vfmt, size, VertexBuffer::Stream);
		iter->m_Vertices->release();
	}
		
	// Create bounding box
	onUpdateBounds(true);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Array<MeshGroup*> FrameModel::getGroups() const {
	magnum::container::List<Group>::Iterator giter, gend;
	Array<MeshGroup*> groups;

	for (giter = m_Groups.begin(); giter != m_Groups.end(); giter++) {
		groups.insert(giter->m_MeshGroup);
	}
	
	return groups;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Array<VertexBuffer*> FrameModel::getFrames(const MeshGroup* group) const {
	magnum::container::List<Group>::Iterator giter;
	Array<VertexBuffer*> frames;

	for (giter = m_Groups.begin(); giter != m_Groups.end(); giter++) {
		if (giter->m_MeshGroup != group)
			continue;

		magnum::container::Array<Pointer<VertexBuffer> >::Iterator fiter;
		for (fiter = giter->m_Frames.begin(); fiter != giter->m_Frames.end(); fiter++) {
			frames.insert(*fiter);
		}
	}
	
	return frames;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::insertGroup(MeshGroup* g) {
	Group group;
	group.m_MeshGroup = g;
	m_Groups.insert(group);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::removeGroup(MeshGroup* mg) {
	magnum::container::List<Group>::Iterator giter, gend;
	gend = m_Groups.end();

	for (giter = m_Groups.begin(); giter != gend; giter++) {
		if (giter->m_MeshGroup == mg) {
			m_Groups.remove(giter);
			return;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::insertFrame(MeshGroup* mg,VertexBuffer* vb,unsigned int frame) {
	magnum::container::List<Group>::Iterator giter, gend;
	gend = m_Groups.end();

	for (giter = m_Groups.begin(); giter != gend; giter++) {
		if (giter->m_MeshGroup == mg) {
			giter->m_Frames.set(frame, vb);
			return;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::removeFrame(MeshGroup* mg,VertexBuffer* vb) {
	magnum::container::List<Group>::Iterator giter, gend;
	gend = m_Groups.end();

	for (giter = m_Groups.begin(); giter != gend; giter++) {
		if (giter->m_MeshGroup == mg)
			break;
	}

	if (giter == gend)
		return;

	for (unsigned int i = 0; i < giter->m_Frames.getSize(); i++) {
		if (giter->m_Frames[i] == vb)
			giter->m_Frames[i] = NULL;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::removeFrame(MeshGroup* mg, unsigned int i) {
	magnum::container::List<Group>::Iterator giter, gend;
	gend = m_Groups.end();

	for (giter = m_Groups.begin(); giter != gend; giter++) {
		if (giter->m_MeshGroup == mg) {
			giter->m_Frames[i] = NULL;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::onClear() {
	m_Groups.clear();
	Model::onClear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool FrameModel::onUpdateBounds(bool deep) {
	if (!deep)
		return Model::onUpdateBounds(false);
		
	// If a recursive box calculation was forced, we use all frames
	AlignedBox3<float> box;
	box.setMaximum(Vector3<float>::Null);
	box.setMinimum(Vector3<float>::Null);

	for (List<Group>::Iterator iter = m_Groups.begin(); iter != m_Groups.end(); ++iter) {
		for (unsigned int i = 0; i < iter->m_Frames.getSize(); i++) {
			VertexBuffer* vb = iter->m_Frames[i];
			AlignedBox3<float> tmpbox;
			if (fitAlignedBox(vb, &tmpbox))
				box.merge(tmpbox);
		}
	}
	setBounds(Bounds3<float>(box));
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::onRender(const ParameterBuffer* pbuf) const {
	ASSERT(pbuf != NULL);
	MatrixStack* ms = getContext()->getMatrixStack();
	ms->selectMode(MatrixStack::ModelView);
	ms->set(*(const Matrix4<float>*)pbuf->getBuffer());

	// Get pointers to animation data
	const unsigned int* anim_idx = (const unsigned int*)((byte*)pbuf->getBuffer() + 64);
	const float* anim_weight = (const float*)((byte*)pbuf->getBuffer() + 72);

	// Loop over all mesh groups (each can have a different anim)	
	magnum::container::List<Group>::Iterator giter, gend;
	gend = m_Groups.end();
	for (giter = m_Groups.begin(); giter != gend; ++giter) {
		// Interpolate frames
		interpolate(giter, anim_idx[0], anim_idx[1], *anim_weight);

		// Select vertex buffer based on current anim
		giter->m_Vertices->select();
		
		// Render all meshes in this meshgroup
		MeshGroup::Iterator mend = giter->m_MeshGroup->end();
		for (MeshGroup::Iterator miter = giter->m_MeshGroup->begin(); miter != mend; ++miter) {
			(*miter)->render();
		}
		
		// Increment animation data
		anim_idx += 3;
		anim_weight += 3;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::interpolate(FrameModel::Group* group, unsigned int idx0, unsigned int idx1, float weight) const {
	if (weight == 1.0f) {
		VertexBuffer* vb = group->m_Frames[idx0];
		const void* src = vb->lock(VertexBuffer::ReadOnly);
		group->m_Vertices->setBuffer(0, src, vb->getSize());
		vb->unlock();
	} else {
		VertexBuffer* vb0 = group->m_Frames[idx0];
		VertexBuffer* vb1 = group->m_Frames[idx1];
		const float* src0 = (float*)vb0->lock(VertexBuffer::ReadOnly);
		const float* src1 = (float*)vb1->lock(VertexBuffer::ReadOnly);
		float* dst = (float* )group->m_Vertices->lock(VertexBuffer::Discard);
		for (unsigned int i = 0; i < vb0->getSize(); ++i) {
			dst[0] = weight * (src0[0] - src1[0]) + src1[0];
			dst[1] = weight * (src0[1] - src1[1]) + src1[1];
			dst[2] = weight * (src0[2] - src1[2]) + src1[2];
			dst[3] = weight * (src0[3] - src1[3]) + src1[3];
			dst[4] = weight * (src0[4] - src1[4]) + src1[4];
			dst[5] = weight * (src0[5] - src1[5]) + src1[5];
			src0 += 6;
			src1 += 6;
			dst += 6;
		}
		group->m_Vertices->unlock();
		vb0->unlock();
		vb1->unlock();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool FrameModel::onOptimize() {
	return false;
}

}; };
