/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "graphics3d.geometrymodel.h"
#include "graphics3d.exception.h"
#include "graphics3d.context.h"
#include "graphics3d.meshbuilder.h"
#include "graphics3d.parameter.h"
#include "graphics3d.parameterbuffer.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"

#include "object.autopointer.inl"
#include "container.array.inl"
#include "math.vector2.inl"
#include "math.vector3.inl"
#include "geometry.cube.inl"
#include "geometry.bounds.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.meshbuilder.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.parameterbuffer.inl"
#include "graphics3d.parameterformat.inl"


namespace magnum { namespace graphics3d {
using magnum::geometry::Cube;
using magnum::object::AutoPointer;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
GeometryModel::GeometryModel(GeometryModel::Type t, float size, unsigned int tess) {
	switch(t) {
		case Sphere:
			createSphere(size, tess);
			break;
		case Box:
			createBox(size, tess);
			break;
		default:
			ASSERT(false);
			throw TypeException();
	}

	m_Mesh.attachAppearance(&m_Appearance);
	m_Appearance.attachMaterial(&m_Material);
	m_Appearance.attachShader(&m_Shader);
	m_Material.setAmbient(Colour<float>(0.03f,0.03f,0.1f));
	m_Material.setDiffuse(Colour<float>(0.2f,0.2f,0.5f));
	m_Material.setEmission(Colour<float>(0.0f,0.0f,0.0f));
	m_Material.setSpecular(Colour<float>(0.0f,0.0f,0.0f));

	// Publish parameters
	ParameterFormat fmt;
	fmt.insert(Parameters::ModelView, Types::Matrix4f);
	publishParameters(fmt);

	// Publish meshes
	publishMesh(&m_Mesh);

	Cube<float,3> cube;
	cube.setOrigin(Vector<float,3>::Null);
	cube.setRadius(size);
	setBounds(Bounds3<float>(cube));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
GeometryModel::~GeometryModel() {
	clearMeshes();
	m_Mesh.detachAppearance();
	m_Appearance.detachMaterial();
	m_Appearance.detachShader();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryModel::createSphere(float r, unsigned int tess) {
	unsigned int i,j;

	MeshBuilder builder;
	builder.setPrimitive(Mesh::TriangleStrip);
	builder.setAttributes(Attributes::Vertex | Attributes::Normal | Attributes::Texture);
	
	// Top row	
    for (i=0; i<=tess; i++) {
        builder.normal(0.0f,-1.0f,0.0f);
        builder.texCoord(i/(float)tess, 0.0f);
        builder.vertex(0.0f,-r,0.0f);
    }

	// ALl middle rows
	for (j=1; j<tess/2; j++) {
      	float theta1 = 2* j * (float)M_PI / tess - (float)M_PI_2;
      	
      	float cost1 = cos(theta1);
      	float sint1 = sin(theta1);

      	for (i=0; i<=tess; i++) {
			Vector3<float> e,p;
         	float theta2 = 2* i * (float)M_PI / tess;
         	float cost2 = cos(theta2);
         	float sint2 = sin(theta2);

         	e.x = cost1 * cost2;
         	e.y = sint1;
         	e.z = cost1 * sint2;
         	p = r*e;

         	builder.normal(e.x,e.y,e.z);
         	builder.texCoord(i/(float)tess, 2*(j+1)/(float)tess);
         	builder.vertex(p.x,p.y,p.z);
      	}
   }

	// Bottom row	
    for (i=0; i<=tess; i++) {
        builder.normal(0.0f,1.0f,0.0f);
        builder.texCoord(i/(float)tess, 1.0f);
        builder.vertex(0.0f,r,0.0f);
    }
    
    // Create indices
    for (j = 0; j < tess/2; j++) {
		unsigned int base1 = j*(tess+1);
		unsigned int base2 = (j+1)*(tess+1);
      	for (i=0; i<=tess; i++) {
			builder.index(base1+i);
			builder.index(base2+i);
		}
		builder.index(base2+tess);
		builder.index(base2+tess);
    }
   
	builder.finish();
	AutoPointer<IndexMesh> mesh = (IndexMesh*)builder.build();
	ASSERT(mesh != NULL);
	m_Mesh.setPrimitive(Mesh::TriangleStrip);
	m_Mesh.attachIndexBuffer(mesh->getIndexBuffer());
	m_Mesh.attachVertexBuffer(mesh->getVertexBuffer());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryModel::createBox(float size, unsigned int tess) {
	VertexFormat vfmt;
	vfmt.insert(Types::Float3, Attributes::Vertex);
	vfmt.insert(Types::Float3, Attributes::Normal);
	vfmt.insert(Types::Float3, Attributes::Texture);
	
	AutoPointer<VertexBuffer> vb = VertexBuffer::create(vfmt, 8, VertexBuffer::Static);
	ASSERT(vb != NULL);
	
	struct BoxVertex {
		Vector3<float>	position;
		Vector3<float>	normal;
		Vector3<float>	texture;
	};
	BoxVertex* vertex = (BoxVertex*)vb->lock(VertexBuffer::Discard);
	#define addVertex(a,b,c) \
		vertex->texture.x = a; \
		vertex->position.x = a * size; \
		vertex->texture.y = b; \
		vertex->position.y = b * size; \
		vertex->texture.z = -(c); \
		vertex->position.z = c * size; \
		vertex->normal = -vertex->position; \
		vertex->normal.normalise(); \
		vertex++;
	addVertex(-1.0f,-1.0f,-1.0f);
	addVertex(-1.0f,-1.0f, 1.0f);
	addVertex(-1.0f, 1.0f,-1.0f);
	addVertex(-1.0f, 1.0f, 1.0f);
	addVertex( 1.0f,-1.0f,-1.0f);
	addVertex( 1.0f,-1.0f, 1.0f);
	addVertex( 1.0f, 1.0f,-1.0f);
	addVertex( 1.0f, 1.0f, 1.0f);
	#undef addVertex
	vb->unlock();

	AutoPointer<IndexBuffer> ib = IndexBuffer::create(IndexBuffer::Index16, 36, IndexBuffer::Static);
	ASSERT(ib != NULL);
	int16* id = (int16*)ib->lock(IndexBuffer::Discard);
	#define addIndex(a,b,c,d) \
		*id++ = b; \
		*id++ = a; \
		*id++ = c; \
		*id++ = c; \
		*id++ = a; \
		*id++ = d;
	addIndex(2,3,1,0);
	addIndex(4,5,7,6);
	addIndex(0,1,5,4);
	addIndex(6,7,3,2);
	addIndex(4,6,2,0);		
	addIndex(1,3,7,5);
	#undef addIndex		
	ib->unlock();
	
	m_Mesh.setPrimitive(Mesh::TriangleList);
	m_Mesh.attachIndexBuffer(ib);
	m_Mesh.attachVertexBuffer(vb);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryModel::onRender(const ParameterBuffer* pbuf) const {
	ASSERT(pbuf != NULL);
	MatrixStack* ms = getContext()->getMatrixStack();
	ms->selectMode(MatrixStack::ModelView);
	ms->set(*(const Matrix4<float>*)pbuf->getBuffer());

	m_Mesh.render();
}


}; };

