/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "math.vector3.h"
#include "graphics3d.context.h"
#include "graphics3d.opengl.h"
#include "graphics3d.geometryrenderer.h"

#include "math.vector.inl"
#include "math.vector2.inl"
#include "math.vector3.inl"
#include "math.matrix.inl"
#include "geometry.line.inl"
#include "geometry.ray.inl"
#include "geometry.segment.inl"
#include "geometry.sphere.inl"
#include "geometry.cube.inl"
#include "geometry.alignedbox.inl"
#include "geometry.box.inl"
#include "geometry.frustum.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.geometryrenderer.inl"

#ifdef near
#undef near
#endif

#ifdef far
#undef far
#endif

namespace magnum { namespace graphics3d {
using magnum::math::Vector3;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryRenderer::renderAxes(const Matrix<float,4>& mtx) {
	// Setup render state
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	Shader::selectDefault();

	DepthMode::selectDefault();
	PolygonMode::selectDefault();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryRenderer::renderAxes(const Matrix<float,4>& mtx,const Vector2<float>& pos) {
	Matrix3<float> mtx3;
	mtx3[0][0] = mtx[0][0]; mtx3[0][1] = mtx[0][1]; mtx3[0][2] = mtx[0][2];
	mtx3[1][0] = mtx[1][0]; mtx3[1][1] = mtx[1][1]; mtx3[1][2] = mtx[1][2];
	mtx3[2][0] = mtx[2][0]; mtx3[2][1] = mtx[2][1]; mtx3[2][2] = mtx[2][2];
	renderAxes(mtx3, pos);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryRenderer::renderAxes(const Matrix<float,3>& mtx,const Vector2<float>& pos) {
	// Setup render state
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	Shader::selectNone();

	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0,viewport[2],viewport[3],0,-1,1);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	// Draw axes	
	float x,y;
	glBegin(GL_LINES);
	// x-Axis
	x = 20.0f*(mtx[0][0] + 0.5f*mtx[0][2]);
	y = -20.0f*(mtx[0][1] - 0.5f*mtx[0][2]);
	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
	glVertex2f(pos.x, pos.y);
	glVertex2f(pos.x + x, pos.y + y);

	// y-Axis
	x = 20.0f*(mtx[1][0] + 0.5f*mtx[1][2]);
	y = -20.0f*(mtx[1][1] - 0.5f*mtx[1][2]);
	glColor4f(0.0f, 1.0f, 0.0f, 1.0f);
	glVertex2f(pos.x, pos.y);
	glVertex2f(pos.x + x, pos.y + y);
	
	// z-Axis
	x = 20.0f*(mtx[2][0] + 0.5f*mtx[2][2]);
	y = -20.0f*(mtx[2][1] - 0.5f*mtx[2][2]);
	glColor4f(0.0f, 0.0f, 1.0f, 1.0f);
	glVertex2f(pos.x, pos.y);
	glVertex2f(pos.x + x, pos.y + y);
	glEnd();

	// Restore state
	glPopMatrix();
	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );
		
	DepthMode::selectDefault();
	PolygonMode::selectDefault();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryRenderer::render(const Line<float, 3>& ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryRenderer::render(const Segment<float, 3>& ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryRenderer::render(const Ray<float, 3>& ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryRenderer::render(const Sphere<float, 3>& sphere, unsigned int n) {
   	unsigned int i,j;
   	float theta1, theta2, theta3;
   	Vector3<float> e,p;
   	Vector3<float> c = sphere.getOrigin();
   	float r = sphere.getRadius();

   	if (n < 4 || r <= 0) {
   	   	glBegin(GL_POINTS);
   	   	glVertex3f(c.x,c.y,c.z);
   	   	glEnd();
   	   	return;
   	}

   	for (j=0; j<n/2; j++) {
   	   	theta1 = 2* j * (float)M_PI / n - (float)M_PI_2;
   	   	theta2 = 2* (j + 1) * (float)M_PI / n - (float)M_PI_2;
   	   	
   	   	float cost1 = cos(theta1);
   	   	float sint1 = sin(theta1);
   	   	float cost2 = cos(theta2);
   	   	float sint2 = sin(theta2);

   	   	glBegin(GL_TRIANGLE_STRIP);
   	   	for (i=0; i<=n; i++) {
   	   	   	theta3 = 2* i * (float)M_PI / n;
   	   	   	float cost3 = cos(theta3);
   	   	   	float sint3 = sin(theta3);

   	   	   	e.x = cost1 * cost3;
   	   	   	e.y = sint1;
   	   	   	e.z = cost1 * sint3;
   	   	   	p = c + r*e;

   	   	   	glNormal3f(e.x,e.y,e.z);
   	   	   	glTexCoord2f(i/(float)n, 2*(j+1)/(float)n);
   	   	   	glVertex3f(p.x,p.y,p.z);

   	   	   	e.x = cost2 * cost3;
   	   	   	e.y = sint2;
   	   	   	e.z = cost2 * sint3;
   	   	   	p = c + r*e;

   	   	   	glNormal3f(e.x,e.y,e.z);
   	   	   	glTexCoord2f(i/(float)n, 2*j/(float)n);
   	   	   	glVertex3f(p.x,p.y,p.z);
   	   	}
   	   	glEnd();
   	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryRenderer::render(const Cube<float, 3>& cube) {
	Vector3<float> o = cube.getOrigin();
	float r = cube.getRadius();
	
	glBegin(GL_QUADS);
	// Face 1
	glNormal3f(-1.0f, 0.0f, 0.0f);
	glVertex3f(o.x - r, o.y - r, o.z - r);
	glVertex3f(o.x - r, o.y - r, o.z + r);
	glVertex3f(o.x - r, o.y + r, o.z + r);
	glVertex3f(o.x - r, o.y + r, o.z - r);

	// Face 2
	glNormal3f(+1.0f, 0.0f, 0.0f);
	glVertex3f(o.x + r, o.y - r, o.z - r);
	glVertex3f(o.x + r, o.y + r, o.z - r);
	glVertex3f(o.x + r, o.y + r, o.z + r);
	glVertex3f(o.x + r, o.y - r, o.z + r);

	// Face	3
	glNormal3f(0.0f, -1.0f, 0.0f);
	glVertex3f(o.x - r, o.y - r, o.z - r);
	glVertex3f(o.x + r, o.y - r, o.z - r);
	glVertex3f(o.x + r, o.y - r, o.z + r);
	glVertex3f(o.x - r, o.y - r, o.z + r);

	// Face	4
	glNormal3f(0.0f, 1.0f, 0.0f);
	glVertex3f(o.x - r, o.y + r, o.z - r);
	glVertex3f(o.x - r, o.y + r, o.z + r);
	glVertex3f(o.x + r, o.y + r, o.z + r);
	glVertex3f(o.x + r, o.y + r, o.z - r);

	// Face	5
	glNormal3f(0.0f, 0.0f, -1.0f);
	glVertex3f(o.x - r, o.y - r, o.z - r);
	glVertex3f(o.x - r, o.y + r, o.z - r);
	glVertex3f(o.x + r, o.y + r, o.z - r);
	glVertex3f(o.x + r, o.y - r, o.z - r);

	// Face	6
	glNormal3f(0.0f, 0.0f, 1.0f);
	glVertex3f(o.x - r, o.y - r, o.z + r);
	glVertex3f(o.x + r, o.y - r, o.z + r);
	glVertex3f(o.x + r, o.y + r, o.z + r);
	glVertex3f(o.x - r, o.y + r, o.z + r);
	glEnd();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryRenderer::render(const AlignedBox<float, 3>& box) {
	Vector3<float> mi = box.getMinimum();
	Vector3<float> ma = box.getMaximum();

	glBegin(GL_QUADS);

	// Face 1
	glNormal3f(-1.0f, 0.0f, 0.0f);
	glVertex3f(mi.x, mi.y, mi.z);
	glVertex3f(mi.x, mi.y, ma.z);
	glVertex3f(mi.x, ma.y, ma.z);
	glVertex3f(mi.x, ma.y, mi.z);

	// Face 2
	glNormal3f(1.0f, 0.0f, 0.0f);
	glVertex3f(ma.x, mi.y, mi.z);
	glVertex3f(ma.x, ma.y, mi.z);
	glVertex3f(ma.x, ma.y, ma.z);
	glVertex3f(ma.x, mi.y, ma.z);

	// Face	3
	glNormal3f(0.0f, -1.0f, 0.0f);
	glVertex3f(mi.x, mi.y, mi.z);
	glVertex3f(ma.x, mi.y, mi.z);
	glVertex3f(ma.x, mi.y, ma.z);
	glVertex3f(mi.x, mi.y, ma.z);
	
	// Face	4
	glNormal3f(0.0f, 1.0f, 0.0f);
	glVertex3f(mi.x, ma.y, mi.z);
	glVertex3f(mi.x, ma.y, ma.z);
	glVertex3f(ma.x, ma.y, ma.z);
	glVertex3f(ma.x, ma.y, mi.z);

	// Face	5
	glNormal3f(0.0f, 0.0f, -1.0f);
	glVertex3f(mi.x, mi.y, mi.z);
	glVertex3f(mi.x, ma.y, mi.z);
	glVertex3f(ma.x, ma.y, mi.z);
	glVertex3f(ma.x, mi.y, mi.z);

	// Face	6
	glNormal3f(0.0f, 0.0f, 1.0f);
	glVertex3f(mi.x, mi.y, ma.z);
	glVertex3f(ma.x, mi.y, ma.z);
	glVertex3f(ma.x, ma.y, ma.z);
	glVertex3f(mi.x, ma.y, ma.z);
	glEnd();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryRenderer::render(const Box<float, 3>& box) {
	Vector3<float> o = box.getOrigin();
	Vector3<float> v[4];
	Vector3<float> n;

	glBegin(GL_QUADS);
	
	// Face 1 OK
	v[0] = o - box.getAxis(0) - box.getAxis(1) - box.getAxis(2);
	v[1] = o - box.getAxis(0) + box.getAxis(1) - box.getAxis(2);
	v[2] = o + box.getAxis(0) + box.getAxis(1) - box.getAxis(2);
	v[3] = o + box.getAxis(0) - box.getAxis(1) - box.getAxis(2);
	n = ((Vector3<float>)box.getAxis(1)) % ((Vector3<float>)box.getAxis(0));
	n.normalise();
	glNormal3f(n.x, n.y, n.z);
	glVertex3f(v[0].x, v[0].y, v[0].z);
	glVertex3f(v[1].x, v[1].y, v[1].z);
	glVertex3f(v[2].x, v[2].y, v[2].z);
	glVertex3f(v[3].x, v[3].y, v[3].z);

	// Face 2 OK!
	v[0] = o - box.getAxis(0) - box.getAxis(1) - box.getAxis(2);
	v[1] = o - box.getAxis(0) - box.getAxis(1) + box.getAxis(2);
	v[2] = o - box.getAxis(0) + box.getAxis(1) + box.getAxis(2);
	v[3] = o - box.getAxis(0) + box.getAxis(1) - box.getAxis(2);
	n = ((Vector3<float>)box.getAxis(2)) % ((Vector3<float>)box.getAxis(1));
	n.normalise();
	glNormal3f(n.x, n.y, n.z);
	glVertex3f(v[0].x, v[0].y, v[0].z);
	glVertex3f(v[1].x, v[1].y, v[1].z);
	glVertex3f(v[2].x, v[2].y, v[2].z);
	glVertex3f(v[3].x, v[3].y, v[3].z);
	
	// Face 3 OK
	v[0] = o - box.getAxis(0) - box.getAxis(1) - box.getAxis(2);
	v[1] = o + box.getAxis(0) - box.getAxis(1) - box.getAxis(2);
	v[2] = o + box.getAxis(0) - box.getAxis(1) + box.getAxis(2);
	v[3] = o - box.getAxis(0) - box.getAxis(1) + box.getAxis(2);
	n = ((Vector3<float>)box.getAxis(0)) % ((Vector3<float>)box.getAxis(2));
	n.normalise();
	glNormal3f(n.x, n.y, n.z);
	glVertex3f(v[0].x, v[0].y, v[0].z);
	glVertex3f(v[1].x, v[1].y, v[1].z);
	glVertex3f(v[2].x, v[2].y, v[2].z);
	glVertex3f(v[3].x, v[3].y, v[3].z);

	// Face 4 OK
	v[0] = o + box.getAxis(0) + box.getAxis(1) + box.getAxis(2);
	v[1] = o + box.getAxis(0) - box.getAxis(1) + box.getAxis(2);
	v[2] = o + box.getAxis(0) - box.getAxis(1) - box.getAxis(2);
	v[3] = o + box.getAxis(0) + box.getAxis(1) - box.getAxis(2);
	n = ((Vector3<float>)box.getAxis(1)) % ((Vector3<float>)box.getAxis(2));
	n.normalise();
	glNormal3f(n.x, n.y, n.z);
	glVertex3f(v[0].x, v[0].y, v[0].z);
	glVertex3f(v[1].x, v[1].y, v[1].z);
	glVertex3f(v[2].x, v[2].y, v[2].z);
	glVertex3f(v[3].x, v[3].y, v[3].z);

	// Face 5 
	v[0] = o + box.getAxis(0) + box.getAxis(1) + box.getAxis(2);
	v[1] = o + box.getAxis(0) + box.getAxis(1) - box.getAxis(2);
	v[2] = o - box.getAxis(0) + box.getAxis(1) - box.getAxis(2);
	v[3] = o - box.getAxis(0) + box.getAxis(1) + box.getAxis(2);
	n = ((Vector3<float>)box.getAxis(2)) % ((Vector3<float>)box.getAxis(0));
	n.normalise();
	glNormal3f(n.x, n.y, n.z);
	glVertex3f(v[0].x, v[0].y, v[0].z);
	glVertex3f(v[1].x, v[1].y, v[1].z);
	glVertex3f(v[2].x, v[2].y, v[2].z);
	glVertex3f(v[3].x, v[3].y, v[3].z);

	// Face 6
	v[0] = o + box.getAxis(0) + box.getAxis(1) + box.getAxis(2);
	v[1] = o - box.getAxis(0) + box.getAxis(1) + box.getAxis(2);
	v[2] = o - box.getAxis(0) - box.getAxis(1) + box.getAxis(2);
	v[3] = o + box.getAxis(0) - box.getAxis(1) + box.getAxis(2);
	n = ((Vector3<float>)box.getAxis(0)) % ((Vector3<float>)box.getAxis(1));
	n.normalise();
	glNormal3f(n.x, n.y, n.z);
	glVertex3f(v[0].x, v[0].y, v[0].z);
	glVertex3f(v[1].x, v[1].y, v[1].z);
	glVertex3f(v[2].x, v[2].y, v[2].z);
	glVertex3f(v[3].x, v[3].y, v[3].z);

	glEnd();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GeometryRenderer::render(const Frustum3<float>& frst) {
	Vector3<float> o = frst.getOrigin();
	Vector3<float> d = frst.getDirection();
	Vector3<float> n, v[4];
	Vector3<float> pn, pf;
	float near = frst.getNear();
	float far = frst.getFar();
	float frac = far / near;

	glBegin(GL_QUADS);

	n = frst.getOffset()[0]*frst.getAxis(0) + frst.getOffset()[1]*frst.getAxis(1);
	pn = o + near*d + n;
	pf = o + far*d + frac*n;
	
	// Near plane
	v[0] = pn - frst.getAxis(0) + frst.getAxis(1); 
	v[1] = pn + frst.getAxis(0) + frst.getAxis(1);
	v[2] = pn + frst.getAxis(0) - frst.getAxis(1);
	v[3] = pn - frst.getAxis(0) - frst.getAxis(1);
	n = -d;
	n.normalise();
	glNormal3f(n.x, n.y, n.z);
	glVertex3f(v[0].x, v[0].y, v[0].z);
	glVertex3f(v[1].x, v[1].y, v[1].z);
	glVertex3f(v[2].x, v[2].y, v[2].z);
	glVertex3f(v[3].x, v[3].y, v[3].z);

	// Far plane
	v[0] = pf - frac * (frst.getAxis(0) + frst.getAxis(1));
	v[1] = pf + frac * (frst.getAxis(0) - frst.getAxis(1));
	v[2] = pf + frac * (frst.getAxis(0) + frst.getAxis(1));
	v[3] = pf - frac * (frst.getAxis(0) - frst.getAxis(1));
	n = d;
	n.normalise();
	glNormal3f(n.x, n.y, n.z);
	glVertex3f(v[0].x, v[0].y, v[0].z);
	glVertex3f(v[1].x, v[1].y, v[1].z);
	glVertex3f(v[2].x, v[2].y, v[2].z);
	glVertex3f(v[3].x, v[3].y, v[3].z);
	
	// Left
	v[0] = pn - frst.getAxis(0) - frst.getAxis(1); 
	v[1] = pf - frac * (frst.getAxis(0) + frst.getAxis(1));
	v[2] = pf - frac * (frst.getAxis(0) - frst.getAxis(1));
	v[3] = pn - frst.getAxis(0) + frst.getAxis(1);
	n = (v[1] - v[0]) % (v[2] - v[0]);
	n.normalise();
	glNormal3f(n.x, n.y, n.z);
	glVertex3f(v[0].x, v[0].y, v[0].z);
	glVertex3f(v[1].x, v[1].y, v[1].z);
	glVertex3f(v[2].x, v[2].y, v[2].z);
	glVertex3f(v[3].x, v[3].y, v[3].z);
	
	// Right
	v[0] = pn + frst.getAxis(0) + frst.getAxis(1); 
	v[1] = pf + frac * (frst.getAxis(0) + frst.getAxis(1));
	v[2] = pf + frac * (frst.getAxis(0) - frst.getAxis(1));
	v[3] = pn + frst.getAxis(0) - frst.getAxis(1);
	n = (v[1] - v[0]) % (v[2] - v[0]);
	n.normalise();
	glNormal3f(n.x, n.y, n.z);
	glVertex3f(v[0].x, v[0].y, v[0].z);
	glVertex3f(v[1].x, v[1].y, v[1].z);
	glVertex3f(v[2].x, v[2].y, v[2].z);
	glVertex3f(v[3].x, v[3].y, v[3].z);
	
	// Upper
	v[0] = pn + frst.getAxis(0) - frst.getAxis(1);
	v[1] = pf + frac * (frst.getAxis(0) - frst.getAxis(1));
	v[2] = pf - frac * (frst.getAxis(0) + frst.getAxis(1));
	v[3] = pn - frst.getAxis(0) - frst.getAxis(1);
	n = (v[1] - v[0]) % (v[2] - v[0]);
	n.normalise();
	glNormal3f(n.x, n.y, n.z);
	glVertex3f(v[0].x, v[0].y, v[0].z);
	glVertex3f(v[1].x, v[1].y, v[1].z);
	glVertex3f(v[2].x, v[2].y, v[2].z);
	glVertex3f(v[3].x, v[3].y, v[3].z);
	
	// Lower
	v[0] = pn - frst.getAxis(0) + frst.getAxis(1);
	v[1] = pf - frac * (frst.getAxis(0) - frst.getAxis(1));
	v[2] = pf + frac * (frst.getAxis(0) + frst.getAxis(1));
	v[3] = pn + frst.getAxis(0) + frst.getAxis(1);
	n = (v[1] - v[0]) % (v[2] - v[0]);
	n.normalise();
	glNormal3f(n.x, n.y, n.z);
	glVertex3f(v[0].x, v[0].y, v[0].z);
	glVertex3f(v[1].x, v[1].y, v[1].z);
	glVertex3f(v[2].x, v[2].y, v[2].z);
	glVertex3f(v[3].x, v[3].y, v[3].z);
	
	glEnd();
}

}; };

