/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics3d.context.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexformat.h"

#include "object.inl"
#include "base.string.inl"
#include "system.assembler.inl"
#include "graphics3d.context.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.vertexformat.inl"

namespace magnum { namespace graphics3d {

const int IndexBuffer::g_OglAcces[] = {
	0,
	GL_READ_ONLY,	// ReadOnly = 1,
	GL_WRITE_ONLY,	// WriteOnly = 2,
	GL_READ_WRITE,	// ReadWrite = 3,
	GL_WRITE_ONLY	// Discard = 4
};

const int IndexBuffer::g_OglUsage[] = {
	0,
	GL_STATIC_DRAW,  // Static = 1,
	GL_DYNAMIC_DRAW, // Dynamic = 2,
	GL_STREAM_DRAW,	 // Stream = 3
	GL_DYNAMIC_DRAW, // System = 4,
};

const int IndexBuffer::g_OglType[] = {
	0,
	GL_UNSIGNED_BYTE,
	GL_UNSIGNED_SHORT,
	0,
	GL_UNSIGNED_INT };


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
IndexBuffer* IndexBuffer::create(Type t, unsigned int count, Usage cf) {
	if ((t != Index8) && (t != Index16) && (t != Index32))
		return NULL;

	IndexBuffer* ib = new IndexBuffer();
	if (ib == NULL)
		return NULL;

	ib->m_Type = t;
	ib->m_Size = count;
	ib->m_OglUsage = g_OglUsage[cf];
	ib->m_OglType = g_OglType[t];

	// Try to create Buffer Object
	if ((glBindBufferARB != NULL) && (cf != System)) {
		glGetError();
		glGenBuffersARB(1, &ib->m_OglBuffer);
		if (glGetError() != GL_NO_ERROR) {
			ib->m_OglBuffer = NULL;
			goto sysmem;
		}
			
		glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER, ib->m_OglBuffer);
		glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER, ((int)t) * count, NULL, ib->m_OglUsage);
		if (glGetError() != GL_NO_ERROR) {
			glDeleteBuffersARB(1, &ib->m_OglBuffer);
			ib->m_OglBuffer = NULL;
			goto sysmem;
		}
		
		ib->m_Usage = cf;
		return ib;
	} 

sysmem:	
	// Fall back to simple memory buffer
	try {
		ib->m_Buffer = (void *)new byte[((int)t) * count];
		ib->m_Usage = System;
	} catch (...) {
		delete ib;
		return NULL;
	}

	if (ib->m_Buffer == NULL) {
		delete ib;
		return NULL;
	}

	return ib;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
IndexBuffer::~IndexBuffer() {
	if (m_OglBuffer != 0) {
		glDeleteBuffersARB(1, &m_OglBuffer);
	} else {
		if (m_Buffer != NULL)
			delete[] (byte *)m_Buffer;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
IndexBuffer::IndexBuffer() {
	m_Locked = false;
	m_Buffer = NULL;
	m_Size = 0;
	m_Type = 0;
	m_OglBuffer = 0;
	m_OglUsage = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void IndexBuffer::setBuffer(unsigned int offset, const void* data, unsigned int size) {
	ASSERT(offset < m_Size);
	ASSERT(offset + size <= m_Size);
	ASSERT(!m_Locked);
	offset *= m_Type;
	size *= m_Type;
	if (m_OglBuffer) {
		glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER, m_OglBuffer);
		glBufferSubDataARB(GL_ELEMENT_ARRAY_BUFFER, offset, size, data);
	} else {
		memcpy(((byte *)m_Buffer) + offset, data, size);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void IndexBuffer::getBuffer(unsigned int offset, void* data, unsigned int size) const {
	ASSERT(offset < m_Size);
	ASSERT(offset + size <= m_Size);
	ASSERT(!m_Locked);
	offset *= m_Type;
	size *= m_Type;
	if (m_OglBuffer) {
		glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER, m_OglBuffer);
		glGetBufferSubDataARB(GL_ELEMENT_ARRAY_BUFFER, offset, size, data);
	} else {
		memcpy(data, ((byte *)m_Buffer) + offset, size);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void IndexBuffer::compileSelect(Assembler& as) const {
	Context* c = getContext();
	ptrdiff_t offset;
	void* buffer;
	
	if (m_OglBuffer) {
		as.asmCall(glBindBufferARB, (GLenum)GL_ELEMENT_ARRAY_BUFFER, (GLuint)m_OglBuffer);
		buffer = NULL;;
	} else {
		if (glBindBufferARB)
			as.asmCall(glBindBufferARB, (GLenum)GL_ELEMENT_ARRAY_BUFFER, (GLuint)0);
		buffer = m_Buffer;
	}

	offset = (byte*)&c->m_IndexType - (byte*)c;
	as.asmLoad(Assembler::T0, (int)m_OglType);
	as.asmStore(Assembler::R0, (int)offset, Assembler::T0);

	offset = (byte*)&c->m_IndexSize - (byte*)c;
	as.asmLoad(Assembler::T0, (int)m_Type);
	as.asmStore(Assembler::R0, (int)offset, Assembler::T0);

	offset = (byte*)&c->m_Indices - (byte*)c;
	as.asmLoad(Assembler::T0, (int)(intptr_t)buffer);
	as.asmStore(Assembler::R0, (int)offset, Assembler::T0);
}

}; };
