/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.context.h"
#include "graphics3d.material.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.shader.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.appearance.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "base.string.inl"
#include "system.thread.inl"
#include "math.vector.inl"
#include "math.vector3.inl"
#include "math.matrix.inl"
#include "math.range.inl"
#include "container.iterator.inl"
#include "container.map.inl"
#include "geometry.trianglemesh.inl"
#include "geometry.bounds.inl"
#include "geometry.sphere.inl"
#include "geometry.box.inl"
#include "geometry.alignedbox.inl"
#include "geometry.collider.inl"
#include "geometry.meshcollider.inl"
#include "geometry.intersection.inl"
#include "geometry.linear.inl"
#include "geometry.line.inl"
#include "geometry.ray.inl"
#include "geometry.segment.inl"
#include "container.array.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.context.inl"
#include "graphics3d.material.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.pixelprogram.inl"
#include "graphics3d.vertexprogram.inl"
#include "graphics3d.appearance.inl"


namespace magnum { namespace graphics3d {
using magnum::object::AutoPointer;
using magnum::math::Vector;
using magnum::container::Map;
using magnum::geometry::Bounds;
using magnum::geometry::AlignedBox3;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void IndexMesh::triangulateTriangleStrip(TriangleMesh3<float>* mesh, const T* indices, unsigned int numIndices) {
	mesh->setIndices((numIndices-2)*3);
	T a,b,c;
	a = *indices++;
	b = *indices++;
	unsigned int tri = 0;
	for (unsigned int i = 0; i < numIndices - 2; i++) {
		c = *indices++;
		if ((a != b) && (a != c) && (b != c))
			mesh->setTriangle(tri++, a,b,c);
		if (i & 1) {
			b = c;
		} else {
			a = c;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void IndexMesh::triangulateTriangleList(TriangleMesh3<float>* mesh, const T* indices, unsigned int numIndices) {
	mesh->setIndices(numIndices);
	T a,b,c;
	for (unsigned int i = 0; i < numIndices/3; i++) {
		a = *indices++;
		b = *indices++;
		c = *indices++;
		mesh->setTriangle(i, a,b,c);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void IndexMesh::triangulateTriangleFan(TriangleMesh3<float>* mesh, const T* indices, unsigned int numIndices) {
	mesh->setIndices((numIndices-2)*3);
	T a,b,c;
	a = *indices++;
	b = *indices++;
	for (unsigned int i = 0; i < numIndices - 2; i++) {
		c = *indices++;
		mesh->setTriangle(i, a,b,c);
		b = c;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
void IndexMesh::triangulateQuadList(TriangleMesh3<float>* mesh, const T* indices, unsigned int numIndices) {
	mesh->setIndices(numIndices/4*3);
	T a,b,c;
	for (unsigned int i = 0; i < numIndices/4; i++) {
		a = indices[0];
		b = indices[1];
		c = indices[2];
		mesh->setTriangle(2*i, a,b,c);
		a = indices[0];
		b = indices[2];
		c = indices[3];
		mesh->setTriangle(2*i+1, a,b,c);
	}
}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool IndexMesh::hasRange() const {
	IndexBuffer* ib = getIndexBuffer();
	if (ib == NULL)
		return false;
		
	unsigned int numIndices = ib->getSize();
	return ((getRange().getSize() > 0) && ((getRange().getStart() != 0) || (getRange().getSize() != numIndices)));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
bool IndexMesh::copyIndices(MeshCollider3<float>* collider, Map<unsigned int,unsigned int>& idxmap) const {
	// Copy indices
	IndexBuffer* ib = getIndexBuffer();
	if (ib == NULL)
		return false;
		
	unsigned int numIndices = ib->getSize();
	void* indices = ib->lock(IndexBuffer::ReadOnly);
	if (indices == NULL)
		return false;

	// Adjust for subrange
	if (hasRange()) {
		numIndices = getRange().getSize();
		if (ib->getType() == IndexBuffer::Index8)
			indices = (void *)((byte *)indices + getRange().getStart());
		else if (ib->getType() == IndexBuffer::Index16)
			indices = (void *)((word *)indices + getRange().getStart());
		else if (ib->getType() == IndexBuffer::Index32)
			indices = (void *)((dword *)indices + getRange().getStart());
	}

	// Allocate space for indices
	collider->setIndices(numIndices);

	const T* idx = (const T*)indices;
	if (hasRange()) {
		// Copy indices with remapping
		for (unsigned int i = 0; i < numIndices; ++i) {
			unsigned int curidx = *idx++;
			Map<unsigned int,unsigned int>::Iterator mi = idxmap.at(curidx); 
			if (mi == idxmap.end()) {
				collider->setIndex(i, idxmap.getSize());
				idxmap.insert(curidx, idxmap.getSize());
			} else {
				collider->setIndex(i, mi.item());
			}
		}
	} else {
		// Copy indices directly
		for (unsigned int i = 0; i < numIndices; ++i)
			collider->setIndex(i, *idx++);
	}

	ib->unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
bool IndexMesh::copyVertices(MeshCollider3<float>* collider, Map<unsigned int,unsigned int>& idxmap) const {
	// Find correct attribute and offset
	VertexBuffer* vb;
	unsigned int offset;
	unsigned int type;
	unsigned int count;
	if (!searchVertices(&vb, &offset, &count, &type))
		return false;

	// Get offset, stride etc	
	unsigned int numVertices = vb->getSize();
	unsigned int incVertices = vb->getStride();
	const void* vertices = vb->lock(VertexBuffer::ReadOnly);
	if (vertices == NULL)
		return false;
	vertices = (void*)((byte *)vertices + offset);

	// Copy vertices
	if (hasRange()) {
		// copy vertices with remapping
		collider->setVertices(idxmap.getSize());
		for (Map<unsigned int, unsigned int>::Iterator iter = idxmap.begin(); iter != idxmap.end(); ++iter) {
			const T* vertex = (const T*)((const byte *)vertices + iter.key()*incVertices);
			float x = (float)vertex[0];
			float y = (count >= 2) ? (float)vertex[1] : 0.0f;
			float z = (count >= 3) ? (float)vertex[2] : 0.0f;
			collider->setVertex(iter.item(), Vector3<float>(x,y,z));
		}
	} else {
		// Copy vertices directly
		collider->setVertices(numVertices);
		for (unsigned int i = 0; i < numVertices; i++) {
			float x = (float)((const T*)vertices)[0];
			float y = (count >= 2) ? (float)((const T*)vertices)[1] : 0.0f;
			float z = (count >= 3) ? (float)((const T*)vertices)[2] : 0.0f;
			vertices = (const void*)((const byte *)vertices + incVertices);
			collider->setVertex(i, Vector3<float>(x,y,z));
		}
	} 

	vb->unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void IndexMesh::onRender() const {
	ASSERT(m_IndexBuffer != NULL);

	selectVertexBuffers();
	selectIndexBuffer();

	if (getAppearance() != NULL)
		getAppearance()->select();

	if (getRange().getSize()) 
		getContext()->renderIndices(getPrimitive(), getRange().getStart(), getRange().getSize());
	else
		getContext()->renderIndices(getPrimitive(), m_IndexBuffer->getSize());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool IndexMesh::onUpdateCollider() {
	// Check for existence of index buffer
	if (getIndexBuffer() == NULL)
		return NULL;

	// Find correct attribute and offset
	VertexBuffer* vb;
	unsigned int offset;
	unsigned int type;
	unsigned int count;
	if (!searchVertices(&vb, &offset, &count, &type))
		return false;

	AutoPointer<MeshCollider3<float> > collider = new MeshCollider3<float>();
	collider->setPrimitive((MeshCollider3<float>::Primitive)getPrimitive());

	// Supply a map that might be used for remapping
	Map<unsigned int,unsigned int> idxmap;
	
	// Copy indices
	switch(getIndexBuffer()->getType()) {
		case IndexBuffer::Index8: 
			if (!copyIndices<uint8>(collider, idxmap))
				return NULL;
			break;
		case IndexBuffer::Index16:
			if (!copyIndices<uint16>(collider, idxmap))
				return NULL;
			break;
		case IndexBuffer::Index32:
			if (!copyIndices<uint32>(collider, idxmap))
				return NULL;
			break;
		default:
			break;
	}

	// Copy vertices
	switch(type) {
		case Types::Byte3:
		case Types::Byte4:
			if (!copyVertices<int8>(collider, idxmap))
				return NULL;
			break;
		case Types::Short3:
		case Types::Short4:
			if (!copyVertices<int16>(collider, idxmap))
				return NULL;
			break;
		case Types::Integer3:
		case Types::Integer4:
			if (!copyVertices<int32>(collider, idxmap))
				return NULL;
			break;
		case Types::Float3:
		case Types::Float4:
			if (!copyVertices<float>(collider, idxmap))
				return NULL;
			break;
		case Types::Double3:
		case Types::Double4:
			if (!copyVertices<double>(collider, idxmap))
				return NULL;
			break;
		default:
			break;
	}

	// Finally attach newly created collider
	attachCollider(collider);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool IndexMesh::onUpdateBounds() {
	// Check for existence of index buffer
	if (getIndexBuffer() == NULL)
		return NULL;

	// Find correct attribute and offset
	VertexBuffer* vb;
	unsigned int offset;
	unsigned int type;
	unsigned int count;
	if (!searchVertices(&vb, &offset, &count, &type))
		return false;
	
	// Get offset, stride etc	
	unsigned int numVertices = vb->getSize();
	unsigned int incVertices = vb->getStride();
	void* vertices = vb->lock(VertexBuffer::ReadOnly);
	if (vertices == NULL)
		return NULL;

	// Adjust buffer for offset and range	
	vertices = (void*)((byte *)vertices + offset);
//	if (getRange().getSize() > 0) {
//		vertices = (void*)((byte *)vertices + getRange().getStart()*incVertices);
//		numVertices = getRange().getSize();	
//	}
	
	switch(type) {
		case Types::Byte3:
		case Types::Byte4: {
			AlignedBox3<byte> box;
			box.bound((Vector<byte,3>*)vertices, incVertices, numVertices);
			setBounds(AlignedBox3<float>(Vector<float,3>(box.getMinimum()), Vector<float,3>(box.getMaximum())));
			break;		
		}
		case Types::Short3:
		case Types::Short4: {
			AlignedBox3<short> box;
			box.bound((Vector<short,3>*)vertices, incVertices, numVertices);
			setBounds(AlignedBox3<float>(Vector<float,3>(box.getMinimum()), Vector<float,3>(box.getMaximum())));
			break;		
		}
		case Types::Integer3:
		case Types::Integer4: {
			AlignedBox3<int> box;
			box.bound((Vector<int,3>*)vertices, incVertices, numVertices);
			setBounds(AlignedBox3<float>(Vector<float,3>(box.getMinimum()), Vector<float,3>(box.getMaximum())));
			break;		
		}
		case Types::Float3:
		case Types::Float4: {
			AlignedBox3<float> box;
			box.bound((Vector<float,3>*)vertices, incVertices, numVertices);
			setBounds(AlignedBox3<float>(box.getMinimum(), box.getMaximum()));
			break;		
		}
		case Types::Double3:
		case Types::Double4: {
			AlignedBox3<double> box;
			box.bound((Vector<double,3>*)vertices, incVertices, numVertices);
			setBounds(AlignedBox3<float>(Vector<float,3>(box.getMinimum()), Vector<float,3>(box.getMaximum())));
			break;		
		}
		
		default:
			vb->unlock();
			return false;
	}
	
	vb->unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TriangleMesh3<float>* IndexMesh::onTriangulate() const {
	// Check for existence of index buffer
	if (getIndexBuffer() == NULL)
		return NULL;

	// Find correct attribute and offset
	VertexBuffer* vb;
	unsigned int offset;
	unsigned int type;
	unsigned int count;
	if (!searchVertices(&vb, &offset, &count, &type))
		return NULL;

	// Copy vertices
	unsigned int numVertices = vb->getSize();
	unsigned int incVertices = vb->getStride();
	void* vertices = vb->lock(VertexBuffer::ReadOnly);
	if (vertices == NULL)
		return NULL;
		
	vertices = (void*)((byte *)vertices + offset);
	
	// Create Triangle mesh
	TriangleMesh3<float>* mesh = new TriangleMesh3<float>();

#define copyVertex(T) \
		float x = (float)((T*)vertices)[0]; \
		float y = (count >= 2) ? (float)((T*)vertices)[1] : 0.0f; \
		float z = (count >= 3) ? (float)((T*)vertices)[2] : 0.0f; \
		vertices = (void*)((byte *)vertices + incVertices); \
		mesh->setVertex(i, Vector3<float>(x,y,z));
#define copyVertices(T) \
		mesh->setVertices(numVertices); \
		for (unsigned int i = 0; i < numVertices; i++) { \
			copyVertex(T); \
		}
	switch(type) {
		case Types::Byte3:
		case Types::Byte4:
			copyVertices(int8);
			break;
		case Types::UByte3:
		case Types::UByte4:
			copyVertices(uint8);
			break;
		case Types::Short3:
		case Types::Short4:
			copyVertices(int16);
			break;
		case Types::UShort3:
		case Types::UShort4:
			copyVertices(uint16);
			break;
		case Types::Int3:
		case Types::Int4:
			copyVertices(int32);
			break;
		case Types::UInt3:
		case Types::UInt4:
			copyVertices(uint32);
			break;
		case Types::Float3:
		case Types::Float4:
			copyVertices(float);
			break;
		case Types::Double3:
		case Types::Double4:
			copyVertices(double);
			break;
		default:
			break;
	}
	vb->unlock();

	// Copy indices
	IndexBuffer* ib = getIndexBuffer();
	unsigned int numIndices = ib->getSize();
	void* indices = ib->lock(IndexBuffer::ReadOnly);
	if (indices == NULL)
		return NULL;
		
	// Adjust for subrange
	if (hasRange()) {
		numIndices = getRange().getSize();
		if (ib->getType() == IndexBuffer::Index8)
			indices = (void *)((byte *)indices + getRange().getStart());
		else if (ib->getType() == IndexBuffer::Index16)
			indices = (void *)((word *)indices + getRange().getStart());
		else if (ib->getType() == IndexBuffer::Index32)
			indices = (void *)((dword *)indices + getRange().getStart());
	}

	switch(getPrimitive()) {
		case TriangleList:
			if (ib->getType() == IndexBuffer::Index8)
				triangulateTriangleList(mesh, (byte*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index16)
				triangulateTriangleList(mesh, (word*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index32)
				triangulateTriangleList(mesh, (dword*)indices, numIndices);
			break;
		case TriangleStrip:
			if (ib->getType() == IndexBuffer::Index8)
				triangulateTriangleStrip(mesh, (byte*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index16)
				triangulateTriangleStrip(mesh, (word*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index32)
				triangulateTriangleStrip(mesh, (dword*)indices, numIndices);
			break;
		case TriangleFan:
			if (ib->getType() == IndexBuffer::Index8)
				triangulateTriangleFan(mesh, (byte*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index16)
				triangulateTriangleFan(mesh, (word*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index32)
				triangulateTriangleFan(mesh, (dword*)indices, numIndices);
			break;
		case QuadStrip:
			ASSERT(false);
			break;
		case QuadList:
			if (ib->getType() == IndexBuffer::Index8)
				triangulateQuadList(mesh, (byte*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index16)
				triangulateQuadList(mesh, (word*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index32)
				triangulateQuadList(mesh, (dword*)indices, numIndices);
			break;
		default:
			break;
	}
	ib->unlock();

	return mesh;
}


}; };
