/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics3d.context.h"
#include "graphics3d.exception.h"
#include "graphics3d.meshconverter.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.meshstripper.h"
#include "graphics3d.appearance.h"
#include "graphics3d.material.h"
#include "graphics3d.shader.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.texture.h"

#include "object.inl"
#include "object.pointer.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.meshconverter.inl"
#include "graphics3d.meshstripper.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.polygonmode.inl"

#include <typeinfo.h>

namespace magnum { namespace graphics3d {
using magnum::types::byte;
using magnum::types::word;
using magnum::types::dword;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Mesh* MeshConverter::onProcess(const Mesh* mesh) {
	ASSERT(mesh != NULL);
	if (typeid(*mesh) == typeid(VertexMesh))
		mesh = createIndexMesh(static_cast<const VertexMesh*>(mesh));
	else if (typeid(*mesh) != typeid(IndexMesh))
		throw UnsupportedMeshException();
		
	// Check if mesh has already correct format
	if (m_Primitive == mesh->getPrimitive()) {
		const_cast<Mesh*>(mesh)->refer();
		return const_cast<Mesh*>(mesh);
	}
	
	// Check if mesh has an index buffer
	const IndexMesh* imesh = static_cast<const IndexMesh*>(mesh);
	if (imesh->getIndexBuffer() == NULL) {
		IndexMesh* newmesh = cloneMesh(imesh);
		newmesh->setPrimitive(m_Primitive);
		return newmesh;
	}
	
	// Check other cases
	IndexBuffer* ib = imesh->getIndexBuffer();
	switch (ib->getType()) {
		case IndexBuffer::Index8:
			return convert<byte>(imesh);
		case IndexBuffer::Index16:
			return convert<word>(imesh);
		case IndexBuffer::Index32:
			return convert<dword>(imesh);
		default:
			throw UnsupportedFormatException();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
IndexMesh* MeshConverter::convert(const IndexMesh* mesh) {
	// Check other cases
	if (m_Primitive == Mesh::TriangleList && mesh->getPrimitive() == Mesh::TriangleStrip)
		return convertTS2TL<T>(static_cast<const IndexMesh*>(mesh));
	if (m_Primitive == Mesh::TriangleList && mesh->getPrimitive() == Mesh::QuadList)
		return convertQL2TL<T>(static_cast<const IndexMesh*>(mesh));
	if (m_Primitive == Mesh::TriangleList && mesh->getPrimitive() == Mesh::QuadStrip)
		return convertQS2TL<T>(static_cast<const IndexMesh*>(mesh));
	if (m_Primitive == Mesh::TriangleStrip && mesh->getPrimitive() == Mesh::TriangleList)
		return convertTS2TL<T>(static_cast<const IndexMesh*>(mesh));
		
	throw UnsupportedPrimitiveException();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
IndexMesh* MeshConverter::convertTL2TS(const IndexMesh* src) {
	MeshStripper stripper;
	return stripper.process(src);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
IndexMesh* MeshConverter::convertTS2TL(const IndexMesh* srcmesh) {
	IndexMesh* dstmesh = cloneMesh(srcmesh);
	dstmesh->setPrimitive(Mesh::TriangleList);
	
	IndexBuffer* srcib = srcmesh->getIndexBuffer();
	ASSERT(srcib != NULL);
	
	IndexBuffer* dstib = IndexBuffer::create(srcib->getType(), (srcib->getSize()-2)*3, IndexBuffer::Static);
	if (dstib == NULL)
		throw MemoryException();
	dstmesh->attachIndexBuffer(dstib);
	
	T* src = (T*)srcib->lock();
	T* dst = (T*)dstib->lock();
	for (unsigned int i = 0; i < srcib->getSize()-2; i++) {
		if (i & 1) {
			*dst++ = src[0];
			*dst++ = src[2];
			*dst++ = src[1];
			src++;
		} else {
			*dst++ = src[0];
			*dst++ = src[1];
			*dst++ = src[2];
			src++;
		}
	}
	srcib->unlock();
	dstib->unlock();
	dstib->release();
	return dstmesh;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
IndexMesh* MeshConverter::convertQL2TS(const IndexMesh* im) {
	IndexMesh mesh = convertQL2TL<T>(im);
	MeshStripper stripper;
	im = stripper.process(mesh);
	mesh->release();
	return im;
}	


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
IndexMesh* MeshConverter::convertQL2TL(const IndexMesh* srcmesh) {
	IndexMesh* dstmesh = cloneMesh(srcmesh);
	dstmesh->setPrimitive(Mesh::TriangleList);
	
	IndexBuffer* srcib = srcmesh->getIndexBuffer();
	ASSERT(srcib != NULL);
	
	IndexBuffer* dstib = IndexBuffer::create(srcib->getType(), srcib->getSize()*3/4, IndexBuffer::Static);
	if (dstib == NULL)
		throw MemoryException();
	dstmesh->attachIndexBuffer(dstib);
	
	T* src = (T*)srcib->lock();
	T* dst = (T*)dstib->lock();
	for (unsigned int i = 0; i < srcib->getSize()/4; i++) {
		*dst++ = src[0];
		*dst++ = src[1];
		*dst++ = src[2];

		*dst++ = src[0];
		*dst++ = src[2];
		*dst++ = src[3];
		src += 4;
	}
	srcib->unlock();
	dstib->unlock();
	dstib->release();
	return dstmesh;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
IndexMesh* MeshConverter::convertQS2TS(const IndexMesh* im) {
	IndexMesh* dstmesh = cloneMesh(srcmesh);
	dstmesh->setPrimitive(Mesh::TriangleList);
	
	IndexBuffer* srcib = srcmesh->getIndexBuffer();
	ASSERT(srcib != NULL);
	
	IndexBuffer* dstib = IndexBuffer::create(srcib->getType(), srcib->getSize(), IndexBuffer::Static);
	if (dstib == NULL)
		throw MemoryException();
	dstmesh->attachIndexBuffer(dstib);
	
	T* src = (T*)srcib->lock();
	T* dst = (T*)dstib->lock();
	*dst++ = *src++;
	*dst++ = *src++;
	for (unsigned int i = 0; i < (srcib->getSize()-2)/2; i++) {
		if (i & 1) {
			*dst++ = src[0];
			*dst++ = src[1];
			src += 2;
		} else {
			*dst++ = src[1];
			*dst++ = src[0];
			src += 2;
		}
	}
	srcib->unlock();
	dstib->unlock();
	dstib->release();
	return dstmesh;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
IndexMesh* MeshConverter::convertQS2TL(const IndexMesh* srcmesh) {
	IndexMesh* dstmesh = cloneMesh(srcmesh);
	dstmesh->setPrimitive(Mesh::TriangleList);
	
	IndexBuffer* srcib = srcmesh->getIndexBuffer();
	ASSERT(srcib != NULL);
	
	IndexBuffer* dstib = IndexBuffer::create(srcib->getType(), (srcib->getSize()-2)/2*3, IndexBuffer::Static);
	if (dstib == NULL)
		throw MemoryException();
	dstmesh->attachIndexBuffer(dstib);
	
	T* src = (T*)srcib->lock();
	T* dst = (T*)dstib->lock();
	for (unsigned int i = 0; i < (srcib->getSize()-2)/2; i++) {
		if (i & 1) {
			*dst++ = src[0];
			*dst++ = src[2];
			*dst++ = src[1];

			*dst++ = src[0];
			*dst++ = src[3];
			*dst++ = src[2];
			src += 2;
		} else {
			*dst++ = src[0];
			*dst++ = src[1];
			*dst++ = src[2];

			*dst++ = src[0];
			*dst++ = src[2];
			*dst++ = src[3];
			src += 2;
		}
	}
	srcib->unlock();
	dstib->unlock();
	dstib->release();
	return dstmesh;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
IndexMesh* MeshConverter::createIndexMesh(const VertexMesh* ) {
	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
IndexMesh* MeshConverter::cloneMesh(const IndexMesh* src) {
	IndexMesh* clone = new IndexMesh();
	clone->setPrimitive(src->getPrimitive());
	clone->setBounds(src->getBounds());
	if (src->getAppearance() != NULL)
		clone->attachAppearance(src->getAppearance());
	for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; ++i) {
		if (src->getVertexBuffer(i) != NULL)
			clone->attachVertexBuffer(src->getVertexBuffer(i),i);
	}
	return clone;
}


}; };
