/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics3d.exception.h"
#include "graphics3d.context.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.meshsplitter.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "container.iterator.inl"
#include "container.allocator.inl"
#include "container.map.inl"
#include "container.array.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"

#include <typeinfo.h>

namespace magnum { namespace graphics3d {
using magnum::object::AutoPointer;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MeshSplitter::MeshSplitter() {
	m_VertexSize = 0x10000;
	m_IndexSize = 0x40000;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MeshList MeshSplitter::split(const Mesh* mesh) {
	ASSERT(mesh != NULL);
	if (typeid(*mesh) == typeid(IndexMesh))
		return split(dynamic_cast<const IndexMesh*>(mesh));
	else if (typeid(*mesh) == typeid(VertexMesh))
		return split(dynamic_cast<const VertexMesh*>(mesh));
	else
		throw UnsupportedMeshException();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MeshList MeshSplitter::split(const IndexMesh* mesh) {
	ASSERT(mesh != NULL);
	MeshList result;
	
	unsigned int pos = 0;
	unsigned int size = 0;
	if (mesh->getRange().getSize()) {
		pos = mesh->getRange().getStart();
		size = mesh->getRange().getSize();
	} else if (mesh->getIndexBuffer()) {
		size = mesh->getIndexBuffer()->getSize();
	}

	// Nothing to do?
	if (size == 0)
		return result;

	// Try to get a reasonable blocksize
	unsigned int blocks = (size + m_IndexSize - 1) / m_IndexSize;
	unsigned int blocksize = size / blocks;

	// Now the real processing starts here
	unsigned int newsize;
	IndexMesh* newmesh;
	bool first = true;
	while(size > 0) {
		newmesh = cloneMesh(mesh);
		result.insert(newmesh);
		switch(mesh->getPrimitive()) {
			case Mesh::TriangleList:
				newsize = min(size,blocksize - (blocksize % 3));
				newsize = copyIndexBuffer(newmesh, mesh, pos, newsize);
				pos += newsize;
				size -= newsize;
				break;
			case Mesh::TriangleStrip:
				newsize = min(size,blocksize & 0xfffffff8);
				newsize = copyIndexBuffer(newmesh, mesh, pos, newsize + (first ? 0 : 2));
				pos += newsize-2;
				size -= newsize;
				break;
			case Mesh::QuadList:
				newsize = min(size,blocksize & 0xfffffff8);
				newsize = copyIndexBuffer(newmesh, mesh, pos, newsize);
				pos += newsize;
				size -= newsize;
				break;
			case Mesh::QuadStrip:
				newsize = min(size,blocksize & 0xfffffff8);
				newsize = copyIndexBuffer(newmesh, mesh, pos, newsize + (first ? 0 : 2));
				pos += newsize-2;
				size -= newsize;
				break;
			case Mesh::LineList:
				newsize = min(size,blocksize & 0xfffffff8);
				newsize = copyIndexBuffer(newmesh, mesh, pos, newsize);
				pos += newsize;
				size -= newsize;
				break;
			case Mesh::LineStrip:
				newsize = min(size,blocksize & 0xfffffff8);
				newsize = copyIndexBuffer(newmesh, mesh, pos, newsize + (first ? 0 : 1));
				pos += newsize-1;
				size -= newsize;
				break;
			case Mesh::LineLoop:
			default:
				throw UnsupportedPrimitiveException();
		}
		first = false;
	}


	return result;	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MeshList MeshSplitter::split(const VertexMesh* mesh) {
	ASSERT(mesh != NULL);
	MeshList result;

	unsigned int pos = 0;
	unsigned int size = 0;
	if (mesh->getRange().getSize()) {
		pos = mesh->getRange().getStart();
		size = mesh->getRange().getSize();
	} else {
		for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; ++i) {
			VertexBuffer* vb = mesh->getVertexBuffer(i);
			if (vb == NULL)
				continue;
			size = vb->getSize();
			break;			
		}
	}

	// Nothing to do?
	if (size == 0)
		return result;

	// Try to get a reasonable blocksize
	unsigned int blocks = (size + m_VertexSize - 1) / m_VertexSize;
	unsigned int blocksize = size / blocks;

	// Now the real processing starts here
	unsigned int newsize;
	VertexMesh* newmesh;
	bool first = true;
	while(size > 0) {
		newmesh = cloneMesh(mesh);
		result.insert(newmesh);
		switch(mesh->getPrimitive()) {
			case Mesh::TriangleList:
				newsize = min(size,blocksize - (blocksize % 3));
				copyVertexBuffer(newmesh, mesh, pos, newsize);
				pos += newsize;
				size -= newsize;
				break;
			case Mesh::TriangleStrip:
				newsize = min(size,blocksize & 0xfffffff8);
				copyVertexBuffer(newmesh, mesh, pos, newsize + (first ? 0 : 2));
				pos += newsize-2;
				size -= newsize;
				break;
			case Mesh::QuadList:
				newsize = min(size,blocksize & 0xfffffff8);
				copyVertexBuffer(newmesh, mesh, pos, newsize);
				pos += newsize;
				size -= newsize;
				break;
			case Mesh::QuadStrip:
				newsize = min(size,blocksize & 0xfffffff8);
				copyVertexBuffer(newmesh, mesh, pos, newsize + (first ? 0 : 2));
				pos += newsize-2;
				size -= newsize;
				break;
			case Mesh::LineList:
				newsize = min(size,blocksize & 0xfffffff8);
				copyVertexBuffer(newmesh, mesh, pos, newsize);
				pos += newsize;
				size -= newsize;
				break;
			case Mesh::LineStrip:
				newsize = min(size,blocksize & 0xfffffff8);
				copyVertexBuffer(newmesh, mesh, pos, newsize + (first ? 0 : 1));
				pos += newsize-1;
				size -= newsize;
				break;
			case Mesh::LineLoop:
			default:
				throw UnsupportedPrimitiveException();
		}
		first = false;
	}

	return result;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
VertexMesh* MeshSplitter::cloneMesh(const VertexMesh* mesh) {
	VertexMesh* newmesh = new VertexMesh();
	newmesh->setPrimitive(mesh->getPrimitive());
	if (mesh->getAppearance())
		newmesh->attachAppearance(mesh->getAppearance());
		
	return newmesh;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
IndexMesh* MeshSplitter::cloneMesh(const IndexMesh* mesh) {
	IndexMesh* newmesh = new IndexMesh();
	newmesh->setPrimitive(mesh->getPrimitive());
	if (mesh->getAppearance())
		newmesh->attachAppearance(mesh->getAppearance());
		
	return newmesh;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshSplitter::copyVertexBuffer(VertexMesh* dst,const VertexMesh* src,unsigned int pos, unsigned int size) {
	for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; ++i) {
		VertexBuffer* srcvb = src->getVertexBuffer(i);
		if (srcvb == NULL)
			continue;
			
		const VertexFormat& fmt = srcvb->getFormat();
		AutoPointer<VertexBuffer> dstvb = VertexBuffer::create(fmt, size, VertexBuffer::Static);
		void* srcmem = srcvb->lock(VertexBuffer::ReadOnly);
		srcmem = ((byte *)srcmem) + pos*fmt.getStride();
		void* dstmem = dstvb->lock(VertexBuffer::Discard);
		memcpy(dstmem, srcmem, size*fmt.getStride());
		srcvb->unlock();
		dstvb->unlock();
		dst->attachVertexBuffer(dstvb,i);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int MeshSplitter::copyIndexBuffer(IndexMesh* dst,const IndexMesh* src,unsigned int pos, unsigned int size) {
	IndexBuffer* srcib = src->getIndexBuffer();
	switch (srcib->getType()) {
		case IndexBuffer::Index8:
			return copyIndexBufferT<byte>(dst, src, pos, size);
		case IndexBuffer::Index16:
			return copyIndexBufferT<word>(dst, src, pos, size);
		case IndexBuffer::Index32:
			return copyIndexBufferT<dword>(dst, src, pos, size);
		default:
			throw UnsupportedFormatException();
	}
}
	

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
unsigned int MeshSplitter::copyIndexBufferT(IndexMesh* dst,const IndexMesh* src,unsigned int pos, unsigned int size) {
	Map<T,T> indexmap;
	
	// First collect all indices and check that the resulting vertex buffer
	// won't get too large
	IndexBuffer* srcib = src->getIndexBuffer();
	T* srcidx = ((T*)srcib->lock(IndexBuffer::ReadOnly)) + pos;
	for (unsigned int i = 0; i < size; i++) {
		if (!indexmap.contains(srcidx[i])) {
			indexmap.insert(srcidx[i],indexmap.getSize());
			if (indexmap.getSize() > m_VertexSize) {
				// Check if we are on an acceptable boundary (trilist, quadlist)
				if ((i+1) % 12)
					continue;
				size = i+1;
				break;
			}
		}
	}
	srcib->unlock();
	
	// Now create a new index buffer
	AutoPointer<IndexBuffer> dstib = IndexBuffer::create(srcib->getType(), size, IndexBuffer::Static);
	dst->attachIndexBuffer(dstib);
	srcidx = ((T*)srcib->lock(IndexBuffer::ReadOnly)) + pos;
	T* dstidx = (T*)dstib->lock(IndexBuffer::Discard);
	for (unsigned int i = 0; i < size; i++) {
		dstidx[i] = indexmap[srcidx[i]];
	}
	srcib->unlock();
	dstib->unlock();

	// Finally create new vertex buffers
	for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; ++i) {
		VertexBuffer* srcvb = src->getVertexBuffer(i);
		if (srcvb == NULL)
			continue;
			
		const VertexFormat& fmt = srcvb->getFormat();
		AutoPointer<VertexBuffer> dstvb = VertexBuffer::create(fmt, indexmap.getSize(), VertexBuffer::Static);
		dst->attachVertexBuffer(dstvb,i);
		void* srcmem = srcvb->lock(VertexBuffer::ReadOnly);
		void* dstmem = dstvb->lock(VertexBuffer::Discard);
		
		unsigned int stride = fmt.getStride();
		for (Map<T,T>::Iterator iter = indexmap.begin(); iter != indexmap.end(); ++iter) {
			void* srcvtx = ((byte*)srcmem) + stride*iter.key();
			void* dstvtx = ((byte*)dstmem) + stride*iter.item();
			memcpy(dstvtx, srcvtx, stride);
		}
		
		srcvb->unlock();
		dstvb->unlock();
	}
	
	return size;
}

}; };
