/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics3d.context.h"
#include "graphics3d.exception.h"
#include "graphics3d.meshstripper.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.material.h"
#include "graphics3d.shader.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.appearance.h"

#include "graphics3d.meshstripper.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.material.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.polygonmode.inl"
#include "container.array.inl"
#include "container.allocator.inl"
#include "container.iterator.inl"
#include "container.compound.inl"
#include "container.pair.inl"
#include "container.list.inl"
#include "container.priorityheap.inl"
#include "algorithm.sort.inl"
#include "algorithm.find.inl"
#include "algorithm.swap.inl"
#include "graph.graph.inl"
#include "graph.edge.inl"
#include "graph.node.inl"
#include "object.pointer.inl"


namespace magnum {namespace graphics3d {
using magnum::types::byte;
using magnum::types::word;
using magnum::types::dword;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> static IndexMesh* stripListMesh(T* _t, const IndexMesh* mesh) {
	ASSERT(mesh != NULL);
	Array<T>* strip;
	T* list;
	unsigned int size;

	IndexBuffer* oldbuf = mesh->getIndexBuffer();
	ASSERT(oldbuf != NULL);
	list = (T*)oldbuf->lock(IndexBuffer::ReadOnly);
	if (mesh->getRange().getSize() > 0) {
		list += mesh->getRange().getStart();
		size = mesh->getRange().getSize();
	} else {
		size = oldbuf->getSize();
	}

	TriangleStripper<T> stripper;
	stripper.insert(list, size);
	stripper.finish();
	strip = stripper.strip();

	oldbuf->unlock();

	IndexBuffer* newbuf = IndexBuffer::create((IndexBuffer::Type)sizeof(T), strip->getSize(), IndexBuffer::Static);
	newbuf->setBuffer(0, strip->getArray(), strip->getSize());
	IndexMesh* newmesh = new IndexMesh();
	newmesh->attachIndexBuffer(newbuf);
	newmesh->setRange(0, strip->getSize());
	newmesh->setPrimitive(Mesh::TriangleStrip);
	if (mesh->getAppearance() != NULL)
		newmesh->attachAppearance(mesh->getAppearance());
	for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (mesh->getVertexBuffer(i) != NULL)
			newmesh->attachVertexBuffer(mesh->getVertexBuffer(i),i);
	}
	return newmesh;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Mesh* MeshStripper::onProcess(const Mesh* mesh) {
	ASSERT(mesh != NULL);

	// Does the mesh already have the correct primitive type?
	if (mesh->getPrimitive() == Mesh::TriangleStrip) {
		const_cast<Mesh*>(mesh)->refer();
		return const_cast<Mesh*>(mesh);
	}
	
	// We currently only support stripping of triangle lists
	if (mesh->getPrimitive() != Mesh::TriangleList) {
		const_cast<Mesh*>(mesh)->refer();
		return const_cast<Mesh*>(mesh);
	}

	// Make sure the mesh is of indexed type
	const IndexMesh* imesh = dynamic_cast<const IndexMesh*>(mesh);
	if (imesh == NULL) {
		const_cast<Mesh*>(mesh)->refer();
		return const_cast<Mesh*>(mesh);
	}

	// Get Indexbuffer of mesh - if it does not have one, we are done
	IndexBuffer* idxbuf = imesh->getIndexBuffer();
	if (idxbuf == NULL) {
		const_cast<Mesh*>(mesh)->refer();
		return const_cast<Mesh*>(mesh);
	}

	switch(idxbuf->getType()) {
		case IndexBuffer::Index8:
			return stripListMesh((byte*)0, imesh);
		case IndexBuffer::Index16:
			return stripListMesh((word*)0, imesh);
		case IndexBuffer::Index32:
			return stripListMesh((dword*)0, imesh);
		default:
			throw UnsupportedFormatException();
	}
}


}; };
