/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics3d.model.h"
#include "graphics3d.parameter.h"
#include "graphics3d.parameterbuffer.h"
#include "graphics3d.intersection.h"

#include "object.inl"
#include "object.pointer.inl"
#include "math.vector.inl"
#include "math.vector3.inl"
#include "math.matrix.inl"
#include "container.iterator.inl"
#include "container.list.inl"
#include "container.array.inl"
#include "container.allocator.inl"
#include "geometry.bounds.inl"
#include "geometry.box.inl"
#include "geometry.alignedbox.inl"
#include "geometry.collider.inl"
#include "geometry.linear.inl"
#include "geometry.ray.inl"
#include "geometry.intersection.inl"
#include "graphics3d.parameterbuffer.inl"
#include "graphics3d.parameterformat.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.joint.inl"
#include "graphics3d.model.inl"
#include "graphics3d.intersection.inl"


namespace magnum {namespace graphics3d {
using magnum::geometry::AlignedBox;
using magnum::math::Limits;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Model::onClear() {
	clearParameters();
	clearJoints();
	clearBounds();
	clearAnimations();
	clearMeshes();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Model::unpublishAnimation(Animation* m) {
	for (unsigned int i = 0; i < m_Animations.getSize(); i++) {
		if (m_Animations[i] == m) {
			m_Animations.remove(i);
			m->release();
			return;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Model::clearAnimations() {
	for (AnimationList::Iterator iter = m_Animations.begin(); iter != m_Animations.end(); ++iter)
		(*iter)->release();
	m_Animations.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Model::unpublishJoint(const Joint& j) {
	for (JointList::Iterator iter = m_Joints.begin(); iter != m_Joints.end(); ++iter) {
		if (iter->getHandle() != j.getHandle())
			continue;
		if (strncmp(iter->getName(), j.getName(), 64))
			continue;
		m_Joints.remove(iter);
		return;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Model::unpublishJoint(const char* n) {
	for (JointList::Iterator iter = m_Joints.begin(); iter != m_Joints.end(); ++iter) {
		if (strncmp(iter->getName(), n, 64))
			continue;
		m_Joints.remove(iter);
		return;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Model::unpublishJoint(Joint::Handle hnd) {
	for (JointList::Iterator iter = m_Joints.begin(); iter != m_Joints.end(); ++iter) {
		if (iter->getHandle() != hnd)
			continue;
		m_Joints.remove(iter);
		return;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Model::publishParameters(const ParameterFormat& fmt) {
	m_Parameters = fmt;
	m_Parameters.shrink();
	
	// Get location of ModelView for intersection tests
	ParameterFormat::Iterator iter;
	if ((iter =  fmt.at(Parameters::ModelView)) != fmt.end())
		m_ModelMatrix = iter->getLocation();
	else if ((iter =  fmt.at(Parameters::Model)) != fmt.end())
		m_ModelMatrix = iter->getLocation();
	else
		m_ModelMatrix = -1;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Model::clearParameters() {
	m_Parameters.clear();
	m_ModelMatrix = -1;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Model::unpublishMesh(Mesh* m) {
	for (unsigned int i = 0; i < m_Meshes.getSize(); i++) {
		if (m_Meshes[i] == m) {
			m_Meshes.remove(i);
			m->release();
			return;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Model::clearMeshes() {
	for (MeshList::Iterator iter = m_Meshes.begin(); iter != m_Meshes.end(); ++iter)
		(*iter)->release();
	m_Meshes.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Model::onTransform(const ParameterBuffer* ,Matrix4<float>* ,Joint::Handle ) const {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Model::onOptimise() {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Model::onUpdateBounds(bool deep) {
	// Create bounds
	Array<Vector<float,3> > verts;
	verts.setSize(getMeshes().getSize() * 8);
	unsigned int i = 0;
	for (MeshList::Iterator iter = getMeshes().begin(); iter != getMeshes().end(); iter++) {
		Mesh* mesh = *iter;
		if (deep)
			mesh->updateBounds();
		mesh->getBounds().getVertices(&verts[i]);
		i += 8;
	}
	
	AlignedBox<float,3> abox;
	abox.bound(&verts[0], getMeshes().getSize() * 8);
	setBounds(Bounds3<float>(abox));
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Model::onUpdateCollider(bool deep) {
	for (MeshList::Iterator iter = getMeshes().begin(); iter != getMeshes().end(); iter++) {
		Mesh* mesh = *iter;
		if (deep || (mesh->getCollider() == NULL))
			mesh->updateCollider();
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Model::onIntersect(const ParameterBuffer* buf,const Ray<float,3>& oray,Intersection* is) const {
	if (m_ModelMatrix == -1)
		return false;
		
	// Get model-view matrix
	const Matrix4<float>& mtx = *static_cast<const Matrix4<float>*>(buf->getPointer(m_ModelMatrix));
	Matrix4<float> invmtx = mtx.getInverse();
	
	// Copy ray and backtransform it from world-space into model-space
	Ray<float,3> ray = oray;
	ray.transform(invmtx);
	
	if (is == NULL) {
		// Iterate over all meshes
		for (MeshList::Iterator iter = getMeshes().begin(); iter != getMeshes().end(); iter++) {
			Mesh* mesh = *iter;
			Collider<float,3>* coll = mesh->getCollider();
			if (coll == NULL)
				continue;
			if (coll->intersect(ray))
				return true;
		}
		return false;			
	}

	// Initialize intersection (which is not NULL at this point)
	is->setParameter(Limits<float>::infinity());

	// Iterate over all meshes
	for (MeshList::Iterator iter = getMeshes().begin(); iter != getMeshes().end(); iter++) {
		Mesh* mesh = *iter;
		Collider<float,3>* coll = mesh->getCollider();
		if (coll == NULL)
			continue;

		magnum::geometry::Intersection<float,3> tmp;
		if (coll->intersect(ray,&tmp)) {
			if (tmp.getParameter() < is->getParameter()) {
				is->setModel(this);
				is->setMesh(mesh);
				is->setPosition(mtx * tmp.getPosition());
				is->setNormal(tmp.getNormal());
				is->setParameter(tmp.getParameter());
			}
		}
	}
	
	return (is->getParameter() < Limits<float>::infinity());
}

}; };
