/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "graphics3d.exception.h"
#include "graphics3d.appearance.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.normalgenerator.h"
#include "graphics3d.context.h"

#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "geometry.attributegenerator.inl"
#include "geometry.normalgenerator.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.normalgenerator.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.mesh.inl"

namespace magnum { namespace graphics3d {
using magnum::object::AutoPointer;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void NormalGenerator::process(IndexMesh* mesh) {
	ASSERT(mesh != NULL);
	ASSERT(mesh->getIndexBuffer() != NULL);
	
	IndexBuffer* ib = mesh->getIndexBuffer();
	if (ib->getType() == IndexBuffer::Index8)
		processType<uint8>(mesh);
	else if (ib->getType() == IndexBuffer::Index16)
		processType<uint16>(mesh);
	else if (ib->getType() == IndexBuffer::Index16)
		processType<uint32>(mesh);
	else
		throw UnsupportedFormatException();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class I>
void NormalGenerator::processType(IndexMesh* mesh) {
	ASSERT(mesh != NULL);
	ASSERT(mesh->getIndexBuffer() != NULL);

	// Get index buffer
	IndexBuffer* ib = mesh->getIndexBuffer();
	
	// Search vertex buffer with vertex and normal information
	VertexBuffer* vvb = NULL;
	unsigned int voffset, vstride;
	VertexBuffer* nvb = NULL;
	unsigned int noffset, nstride;
	for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		VertexBuffer* vb = mesh->getVertexBuffer(i);
		if (vb == NULL)
			continue;
			
		const VertexFormat& vfmt = vb->getFormat();
		for (unsigned int j = 0; j < vfmt.getSize(); j++) {
			if (vfmt[j].getAttribute() == Attributes::Vertex) {
				vvb = vb;
				voffset = vfmt[j].getOffset();
				vstride = vb->getStride();
			} else if (vfmt[j].getAttribute() == Attributes::Normal) {
				nvb = vb;
				noffset = vfmt[j].getOffset();
				nstride = vb->getStride();
			}
		}
	}
	
	// Check if we found at least the vertices
	if (vvb == NULL)
		throw UnsupportedFormatException();
		
	// If there is no normal buffer, create one
	if (nvb == NULL) {
		VertexFormat fmt;
		fmt.set(0, Types::Float3, Attributes::Normal);
		AutoPointer<VertexBuffer> normals = VertexBuffer::create(fmt, vvb->getSize(), VertexBuffer::Static);
		for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
			if (mesh->getVertexBuffer(i) == NULL) {
				mesh->attachVertexBuffer(normals, i);
				break;
			}
		}

		nvb = normals;
		nstride = sizeof(Vector3<float>);
		noffset = 0;
	}

	// Allocate new normal generator
	magnum::geometry::NormalGenerator<float, I> generator;
	
	// Setup primitive
	switch(mesh->getPrimitive()) {
		case Mesh::TriangleFan:
			generator.setPrimitive(magnum::geometry::AttributeGenerator<float,I>::TriangleFan);
			break;
		case Mesh::TriangleList:
			generator.setPrimitive(magnum::geometry::AttributeGenerator<float,I>::TriangleList);
			break;
		case Mesh::TriangleStrip:
			generator.setPrimitive(magnum::geometry::AttributeGenerator<float,I>::TriangleStrip);
			break;
		default:
			throw UnsupportedPrimitiveException();
	}

	// Setup all buffers
	const I* idxs = (const I*)ib->lock(IndexBuffer::ReadOnly);
	generator.setIndices(idxs, ib->getSize());
	if (vvb != nvb) {
		const Vector3<float>* verts = (const Vector3<float>*)((byte*)vvb->lock(VertexBuffer::ReadOnly) + voffset);
		generator.setVertices(verts, vstride);
		Vector3<float>* normals = (Vector3<float>*)((byte*)nvb->lock(VertexBuffer::WriteOnly) + noffset);
		generator.setNormals(normals, nstride);
	} else {
		const byte* buf = (byte*)vvb->lock(VertexBuffer::ReadWrite);
		generator.setVertices((const Vector3<float>*)(buf + voffset));
		generator.setNormals((Vector3<float>*)(buf + noffset));
	}
	
	// Create normals!
	generator.generate();
	
	// Unlock buffers
	ib->unlock();
	vvb->unlock();
	if (nvb->isLocked())
		nvb->unlock();
}

}; };
