/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "graphics3d.optimiser.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.material.h"
#include "graphics3d.shader.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.appearance.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.context.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.vertexmesh.h"
#include "graphics3d.renderstream.h"
#include "graphics3d.meshmerger.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "base.string.inl"
#include "math.range.inl"
#include "container.array.inl"
#include "container.map.inl"
#include "container.iterator.inl"
#include "container.tuple.inl"
#include "container.allocator.inl"
#include "container.chain.inl"
#include "io.filter.inl"
#include "graphics3d.optimiser.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.renderstream.inl"
#include "graphics3d.meshmerger.inl"
#include "graphics3d.meshstripper.inl"
#include "graphics3d.meshsplitter.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.material.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.pixelprogram.inl"
#include "graphics3d.vertexprogram.inl"

#include <typeinfo.h>
#include <stdlib.h>

namespace magnum {namespace graphics3d {
using namespace magnum::container;
using namespace magnum::math;
using magnum::object::AutoPointer;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Optimiser::Optimiser() {
	m_Strategy = SortShader;
	clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Optimiser::~Optimiser() {
	clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::clear() {
	unsigned int i;

	m_CurrentMaterial = NULL;
	m_CurrentShader = NULL;
	m_CurrentBlendMode = NULL;
	m_CurrentPolygonMode = NULL;
	m_CurrentStencilMode = NULL;
	m_CurrentDepthMode = NULL;
	m_CurrentIndexBuffer = NULL;
	for (i = 0; i < Attributes::MaxAttributes; i++)
		m_CurrentVertexBuffers[i] = NULL;
	for (i = 0; i < Texture::MaxTextures; i++)
		m_CurrentTextures[i] = NULL;

	m_PreselectedMaterial = NULL;
	m_PreselectedShader = NULL;
	m_PreselectedBlendMode = NULL;
	m_PreselectedPolygonMode = NULL;
	m_PreselectedStencilMode = NULL;
	m_PreselectedDepthMode = NULL;
	m_PreselectedIndexBuffer = NULL;
	for (i = 0; i < Attributes::MaxAttributes; i++)
		m_PreselectedVertexBuffers[i] = NULL;
	for (i = 0; i < Texture::MaxTextures; i++)
		m_PreselectedTextures[i] = NULL;

	m_Jobs.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::renderMesh(Mesh* mesh) {
	ASSERT(mesh != NULL);
	if (typeid(*mesh) == typeid(IndexMesh)) {
		renderIndexMesh(static_cast<IndexMesh*>(mesh));
	} else if (typeid(*mesh) == typeid(VertexMesh)) {
		renderVertexMesh(static_cast<VertexMesh*>(mesh));
	} else {
		insertMesh(mesh);
	}
}		
		
		
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::renderIndexMesh(IndexMesh* imesh) {
	if (imesh->getIndexBuffer() == 0)
		return;

	// This is a little	trick for making sure that any newly created
	// mesh will be released
	AutoPointer<IndexMesh> mesh = imesh;
	mesh->refer();

	// Check if we should strip the mesh
//	if (mesh->getPrimitive() == Mesh::TriangleList)
//		mesh = static_cast<IndexMesh*>(m_Stripper.process(mesh));

	// Get size of rendered data
	unsigned int size = mesh->getRange().getSize();
	if (size == 0)
		size = mesh->getIndexBuffer()->getSize();
	if (size == 0)
		return;
		
	// Check if mesh is too large
	if (size <= m_Splitter.getIndexSize()) {
		insertMesh(mesh);
		return;
	}
	
	MeshList list = m_Splitter.split(mesh);
	for (MeshList::Iterator iter = list.begin(); iter != list.end(); ++iter) {
		insertMesh(*iter);
		(*iter)->release();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::renderVertexMesh(VertexMesh* mesh) {
	// Get size of rendered data
	unsigned int size = mesh->getRange().getSize();
	if (size == 0) {
		for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; ++i) {
			if (mesh->getVertexBuffer(i) != NULL) {
				size = mesh->getVertexBuffer(i)->getSize();
				break;
			}
		}
	}
	if (size == 0)
		return;

	// Check if mesh is too large
	if (size <= m_Splitter.getVertexSize()) {
		insertMesh(mesh);
		return;
	}
	
	MeshList list = m_Splitter.split(mesh);
	for (MeshList::Iterator iter = list.begin(); iter != list.end(); ++iter) {
		insertMesh(*iter);
		(*iter)->release();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::insertMesh(Mesh* mesh) {
	Job job;
	unsigned int i, j, attribs;
	VertexBuffer* vb;
	Appearance* app;

	// Get type of mesh:
	if (typeid(*mesh) == typeid(IndexMesh)) {
		IndexMesh* imesh = (IndexMesh*)mesh;
		job.m_Mesh = NULL;
		job.m_IndexBuffer = imesh->getIndexBuffer();
		job.m_Start = imesh->getRange().getStart();
		job.m_Count = imesh->getRange().getSize();
		if (job.m_Count == 0)
			job.m_Count = imesh->getIndexBuffer()->getSize();
		for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
			vb = imesh->getVertexBuffer(i);
			if (vb == NULL)
				continue;
			attribs = vb->getAttributes();
			for (j = 0; j < Attributes::MaxAttributes; j++) {
				if ((attribs & (1 << j)) == 0)
					continue;
				job.m_VertexBuffers[j] = vb;
			}
		}
	} else if (typeid(*mesh) == typeid(VertexMesh)) {
		VertexMesh* vmesh = (VertexMesh*)mesh;
		job.m_Mesh = NULL;
		job.m_IndexBuffer = NULL;
		job.m_Start = vmesh->getRange().getStart();
		job.m_Count = vmesh->getRange().getSize();
		if (job.m_Count == 0) {
			for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
				vb = vmesh->getVertexBuffer(i);
				if (vb != NULL) {
					job.m_Count = vb->getSize();
					break;
				}
			}
		}
		for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
			vb = vmesh->getVertexBuffer(i);
			if (vb == NULL)
				continue;
			attribs = vb->getAttributes();
			for (j = 0; j < Attributes::MaxAttributes; j++) {
				if ((attribs & (1 << j)) == 0)
					continue;
				job.m_VertexBuffers[j] = vb;
			}
		}
	} else {
		job.m_Start = 0;
		job.m_Count = 0;
		job.m_Mesh = mesh;
		job.m_IndexBuffer = NULL;
		for (i = 0; i < Attributes::MaxAttributes; i++) 
			job.m_VertexBuffers[i] = NULL;
	}

	job.m_Primitive = mesh->getPrimitive();
	app = mesh->getAppearance();
	if (app != NULL) {
		job.m_Material = app->getMaterial();
		job.m_Shader = app->getShader();
		job.m_BlendMode = app->getBlendMode();
		job.m_PolygonMode = app->getPolygonMode();
		job.m_StencilMode = app->getStencilMode();
		job.m_DepthMode = app->getDepthMode();
		for (i = 0; i < Texture::MaxTextures; i++)
			job.m_Textures[i] = app->getTexture(i);
	} else {
		job.m_Material = NULL;
		job.m_Shader = NULL;
		job.m_BlendMode = NULL;
		job.m_PolygonMode = NULL;
		job.m_StencilMode = NULL;
		job.m_DepthMode = NULL;
		for (i = 0; i < Texture::MaxTextures; i++)
			job.m_Textures[i] = NULL;
	}

	if (job.m_Material == NULL)
		job.m_Material = m_CurrentMaterial;
	else
		m_CurrentMaterial = job.m_Material;

	if (job.m_BlendMode == NULL)
		job.m_BlendMode = m_CurrentBlendMode;
	else
		m_CurrentBlendMode = job.m_BlendMode;

	if (job.m_PolygonMode == NULL)
		job.m_PolygonMode = m_CurrentPolygonMode;
	else
		m_CurrentPolygonMode = job.m_PolygonMode;

	if (job.m_StencilMode == NULL)
		job.m_StencilMode = m_CurrentStencilMode;
	else
		m_CurrentStencilMode = job.m_StencilMode;

	if (job.m_DepthMode == NULL)
		job.m_DepthMode = m_CurrentDepthMode;
	else
		m_CurrentDepthMode = job.m_DepthMode;

	if (job.m_Shader == NULL)
		job.m_Shader = m_CurrentShader;
	else
		m_CurrentShader = job.m_Shader;

	if (job.m_IndexBuffer != NULL)
		m_CurrentIndexBuffer = job.m_IndexBuffer;

	for (i = 0; i < Attributes::MaxAttributes; i++) {
		if (job.m_VertexBuffers[i] != NULL)
			m_CurrentVertexBuffers[i] = job.m_VertexBuffers[i];
	}

	for (i = 0; i < Texture::MaxTextures; i++) {
		if (job.m_Textures[i] != NULL)
			m_CurrentTextures[i] = job.m_Textures[i];
	}

	m_Jobs.insert(job);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::renderVertices(Mesh::Primitive p, unsigned int start, unsigned int size) {
	Job job;
	unsigned int i;
	unsigned int attribs;

	job.m_Primitive = p;
	job.m_Start = start;
	job.m_Count = size;
	job.m_Mesh = NULL;
	job.m_IndexBuffer = NULL;
	job.m_Shader = m_CurrentShader;
	job.m_Material = m_CurrentMaterial;
	job.m_BlendMode = m_CurrentBlendMode;
	job.m_PolygonMode = m_CurrentPolygonMode;
	job.m_StencilMode = m_CurrentStencilMode;
	job.m_DepthMode = m_CurrentDepthMode;

	if (m_CurrentShader != NULL)
		attribs = m_CurrentShader->getAttributes();
	else
		attribs = 0xffffffff;

	for (i = 0; i < Attributes::MaxAttributes; i++) {
		if (attribs & (1 << i))
			job.m_VertexBuffers[i] = m_CurrentVertexBuffers[i];
		else
			job.m_VertexBuffers[i] = NULL;
	}

	for (i = 0; i < Texture::MaxTextures; i++) {
		job.m_Textures[i] = m_CurrentTextures[i];
	}

	m_Jobs.insert(job);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::renderIndices(Mesh::Primitive p, unsigned int start, unsigned int size) {
	Job job;
	unsigned int i;
	unsigned int attribs;

	job.m_Primitive = p;
	job.m_Start = start;
	job.m_Count = size;
	job.m_Mesh = NULL;
	job.m_IndexBuffer = m_CurrentIndexBuffer;
	job.m_Shader = m_CurrentShader;
	job.m_Material = m_CurrentMaterial;
	job.m_BlendMode = m_CurrentBlendMode;
	job.m_PolygonMode = m_CurrentPolygonMode;
	job.m_StencilMode = m_CurrentStencilMode;
	job.m_DepthMode = m_CurrentDepthMode;

	if (m_CurrentShader != NULL)
		attribs = m_CurrentShader->getAttributes();
	else
		attribs = 0xffffffff;

	for (i = 0; i < Attributes::MaxAttributes; i++) {
		if (attribs & (1 << i))
			job.m_VertexBuffers[i] = m_CurrentVertexBuffers[i];
		else
			job.m_VertexBuffers[i] = NULL;
	}

	for (i = 0; i < Texture::MaxTextures; i++) {
		job.m_Textures[i] = m_CurrentTextures[i];
	}

	m_Jobs.insert(job);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::selectAppearance(Appearance* app) {
	ASSERT(app != NULL)
	if (app->getShader() != NULL)
		selectShader(app->getShader());
	if (app->getMaterial() != NULL)
		selectMaterial(app->getMaterial());
	if (app->getBlendMode() != NULL)
		selectBlendMode(app->getBlendMode());
	if (app->getDepthMode() != NULL)
		selectDepthMode(app->getDepthMode());
	if (app->getPolygonMode() != NULL)
		selectPolygonMode(app->getPolygonMode());
	if (app->getStencilMode() != NULL)
		selectStencilMode(app->getStencilMode());
	for (unsigned int i = 0; i < Texture::MaxTextures; i++) {
		if (app->getTexture(i) != NULL)
			selectTexture(app->getTexture(i), i);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::deselectAppearance() {
	deselectShader();
	deselectMaterial();
	deselectBlendMode();
	deselectPolygonMode();
	deselectDepthMode();
	deselectStencilMode();
	for (unsigned int i = 0; i < Texture::MaxTextures; i++) {
		deselectTexture(i);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::preselectAppearance(Appearance* app) {
	ASSERT(app != NULL)
	if (app->getShader() != NULL)
		preselectShader(app->getShader());
	if (app->getMaterial() != NULL)
		preselectMaterial(app->getMaterial());
	if (app->getBlendMode() != NULL)
		preselectBlendMode(app->getBlendMode());
	if (app->getDepthMode() != NULL)
		preselectDepthMode(app->getDepthMode());
	if (app->getPolygonMode() != NULL)
		preselectPolygonMode(app->getPolygonMode());
	if (app->getStencilMode() != NULL)
		preselectStencilMode(app->getStencilMode());
	for (unsigned int i = 0; i < Texture::MaxTextures; i++) {
		if (app->getTexture(i) != NULL)
			preselectTexture(app->getTexture(i), i);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::finish() {
	// Split up index buffers that are used for more than one mesh
	splitBuffers();

	// Now sort jobs
	qsort(m_Jobs.getArray(), m_Jobs.getSize(), sizeof(Job), cmpJobs);

	// Merge buffers
	mergeBuffers();

	// Merge jobs
	mergeJobs();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
RenderStream* Optimiser::createStream() {
	// Loop through all jobs again
	unsigned int i, j;
    const Material* curMaterial = m_PreselectedMaterial;
    const Shader* curShader = m_PreselectedShader;
    const StencilMode* curStencilMode = m_PreselectedStencilMode;
    const BlendMode* curBlendMode = m_PreselectedBlendMode;
    const DepthMode* curDepthMode = m_PreselectedDepthMode;
    const PolygonMode* curPolygonMode = m_PreselectedPolygonMode;
    const IndexBuffer* curIndexBuffer = m_PreselectedIndexBuffer;
    const VertexBuffer* curVertexBuffers[Attributes::MaxAttributes];
	const Texture* curTextures[Texture::MaxTextures];
	unsigned int curStart = 0xffffffff, curEnd = 0xffffffff;
	for (i = 0; i < Texture::MaxTextures; i++)
		curTextures[i] = m_PreselectedTextures[i];
	for (i = 0; i < Attributes::MaxAttributes; i++)
		curVertexBuffers[i] = m_PreselectedVertexBuffers[i];

	m_Compiler.clear();

	for (i = 0; i < m_Jobs.getSize(); i++) {
		Job* job = &m_Jobs[i];
		
		// Is this job a complete mesh? If so, we need some 
		// special care
		if (job->m_Mesh != NULL) {
			continue;
		}
		
		// Compare all appearance attributes of this job
		// with current settings
		if (job->m_BlendMode != curBlendMode) {
			curBlendMode = job->m_BlendMode;
			if (curBlendMode != NULL)
				m_Compiler.selectBlendMode(curBlendMode);
			else
				m_Compiler.deselectBlendMode();
		}
		if (job->m_DepthMode != curDepthMode) {
			curDepthMode = job->m_DepthMode;
			if (curDepthMode != NULL)
				m_Compiler.selectDepthMode(curDepthMode);
			else
				m_Compiler.deselectDepthMode();
		}
		if (job->m_PolygonMode != curPolygonMode) {
			curPolygonMode = job->m_PolygonMode;
			if (curPolygonMode != NULL)
				m_Compiler.selectPolygonMode(curPolygonMode);
			else
				m_Compiler.deselectPolygonMode();
		}
		if (job->m_StencilMode != curStencilMode) {
			curStencilMode = job->m_StencilMode;
			if (curStencilMode != NULL)
				m_Compiler.selectStencilMode(curStencilMode);
			else
				m_Compiler.deselectStencilMode();
		}
		if (job->m_Shader != curShader) {
			curShader = job->m_Shader;
			if (curShader != NULL)
				m_Compiler.selectShader(curShader);
			else
				m_Compiler.deselectShader();
		}
		if (job->m_Material != curMaterial) {
			curMaterial = job->m_Material;
			if (curMaterial != NULL)
				m_Compiler.selectMaterial(curMaterial);
		}
		if (job->m_IndexBuffer != curIndexBuffer) {
			curIndexBuffer = job->m_IndexBuffer;
			if (curIndexBuffer != NULL)
				m_Compiler.selectIndexBuffer(curIndexBuffer);
		}
		// Bind all textures
		for (j = 0; j < Texture::MaxTextures; j++) {
			if (job->m_Textures[j] != curTextures[j]) {
				curTextures[j] = job->m_Textures[j];
				if (curTextures[j] != NULL)
					m_Compiler.selectTexture(curTextures[j], j);
			}
		}

		// Check if a shader is present (no shader - no rendering)
		if (curShader == NULL)
			continue;

		// collect vertex attributes of this job
		Map<const VertexBuffer*, unsigned int> vbAttribs;
		Map<const VertexBuffer*, unsigned int>::Iterator iter;
		unsigned int attribs = curShader->getAttributes();
		for (j = 0; j < Attributes::MaxAttributes; j++) {
			const VertexBuffer* vb = job->m_VertexBuffers[j];
			if ((attribs & (1 << j)) == 0)
				continue;
			if (vb == NULL)
				continue;
			if (vb != curVertexBuffers[j]) {
				curVertexBuffers[j] = vb;
				iter = vbAttribs.at(vb);
				if (iter == vbAttribs.end()) {
					vbAttribs.insert(vb, 1 << j);
				} else {
					*iter = *iter | (1 << j);
				}
			}
		}
		
		// Select all vertex buffers that need to be bound
		for (iter = vbAttribs.begin(); iter != vbAttribs.end(); iter++) {
			m_Compiler.selectVertexBuffer(iter.key(), iter.item());
		}

		// Render mesh!
		if (curIndexBuffer != NULL) {
			m_Compiler.renderIndices(job->m_Primitive, job->m_Start, job->m_Count);
		} else {
			m_Compiler.renderVertices(job->m_Primitive, job->m_Start, job->m_Count);
		}
	}

	// Deselect active shader - this is sort of a hack	
	m_Compiler.deselectShader();
	m_Compiler.finish();

	return m_Compiler.compile();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MeshList Optimiser::createMeshes(bool complete) {
	MeshList mlist;
	
	// Loop through all jobs again
	unsigned int i, j;
    const Material* curMaterial = m_PreselectedMaterial;
    const Shader* curShader = m_PreselectedShader;
    const StencilMode* curStencilMode = m_PreselectedStencilMode;
    const BlendMode* curBlendMode = m_PreselectedBlendMode;
    const DepthMode* curDepthMode = m_PreselectedDepthMode;
    const PolygonMode* curPolygonMode = m_PreselectedPolygonMode;
    const IndexBuffer* curIndexBuffer = m_PreselectedIndexBuffer;
    const VertexBuffer* curVertexBuffers[Attributes::MaxAttributes];
	const Texture* curTextures[Texture::MaxTextures];
	unsigned int curStart = 0xffffffff, curEnd = 0xffffffff;
	for (i = 0; i < Texture::MaxTextures; i++)
		curTextures[i] = m_PreselectedTextures[i];
	for (i = 0; i < Attributes::MaxAttributes; i++)
		curVertexBuffers[i] = m_PreselectedVertexBuffers[i];

	for (i = 0; i < m_Jobs.getSize(); i++) {
		Job* job = &m_Jobs[i];
		
		// Is this job a complete mesh? If so, we need some 
		// special care
		if (job->m_Mesh != NULL) {
			mlist.insert(job->m_Mesh);
			job->m_Mesh->refer();
			continue;
		}
		
		// CReate new appearance for this mesh
		AutoPointer<Appearance> app = new Appearance();
		
		// Compare all appearance attributes of this job
		// with current settings
		if ((job->m_BlendMode != curBlendMode) || complete){
			curBlendMode = job->m_BlendMode;
			if (curBlendMode != NULL)
				app->attachBlendMode(const_cast<BlendMode*>(curBlendMode));
			else
				app->detachBlendMode();
		}
		if ((job->m_DepthMode != curDepthMode) || complete) {
			curDepthMode = job->m_DepthMode;
			if (curDepthMode != NULL)
				app->attachDepthMode(const_cast<DepthMode*>(curDepthMode));
			else
				app->detachDepthMode();
		}
		if ((job->m_PolygonMode != curPolygonMode) || complete) {
			curPolygonMode = job->m_PolygonMode;
			if (curPolygonMode != NULL)
				app->attachPolygonMode(const_cast<PolygonMode*>(curPolygonMode));
			else
				app->defaultPolygonMode();
		}
		if ((job->m_StencilMode != curStencilMode) || complete) {
			curStencilMode = job->m_StencilMode;
			if (curStencilMode != NULL)
				app->attachStencilMode(const_cast<StencilMode*>(curStencilMode));
			else
				app->detachStencilMode();
		}
		if ((job->m_Shader != curShader) || complete) {
			curShader = job->m_Shader;
			if (curShader != NULL)
				app->attachShader(const_cast<Shader*>(curShader));
			else
				app->detachShader();
		}
		if ((job->m_Material != curMaterial) || complete) {
			curMaterial = job->m_Material;
			if (curMaterial != NULL)
				app->attachMaterial(const_cast<Material*>(curMaterial));
			else
				app->detachMaterial();
		}
		for (j = 0; j < Texture::MaxTextures; j++) {
			if ((job->m_Textures[j] != curTextures[j]) || complete) {
				curTextures[j] = job->m_Textures[j];
				if (curTextures[j] != NULL)
					app->attachTexture(const_cast<Texture*>(curTextures[j]), j);
				else
					app->detachTexture(j);
			}
		}
		
		curIndexBuffer = job->m_IndexBuffer;

		// Check if a shader is present (no shader - no rendering)
		if (curShader == NULL)
			continue;

		// collect vertex attributes of this job
		Map<const VertexBuffer*, unsigned int> vbAttribs;
		Map<const VertexBuffer*, unsigned int>::Iterator iter;
		unsigned int attribs = curShader->getAttributes();
		for (j = 0; j < Attributes::MaxAttributes; j++) {
			const VertexBuffer* vb = job->m_VertexBuffers[j];
			if (vb == NULL) 
				continue;
			if ((attribs & (1 << j)) == 0)
				continue;
			if ((vb != curVertexBuffers[j]) || complete) {
				curVertexBuffers[j] = vb;
				iter = vbAttribs.at(vb);
				if (iter == vbAttribs.end()) {
					vbAttribs.insert(vb, 1 << j);
				} else {
					*iter = *iter | (1 << j);
				}
			}
		}
		
		// Create mesh;
		VertexMesh* mesh = NULL;
		if (curIndexBuffer != NULL) {
			IndexMesh* imesh = new IndexMesh();
			imesh->attachIndexBuffer(const_cast<IndexBuffer*>(curIndexBuffer));
			mesh = imesh;
		} else {
			mesh = new VertexMesh();
		}
		mesh->attachAppearance(app);
		mesh->setPrimitive(job->m_Primitive);
		mesh->setRange(job->m_Start, job->m_Count);
		
		// Attach all vertex buffers that need to be bound
		j = 0;
		for (iter = vbAttribs.begin(); iter != vbAttribs.end(); iter++) {
			mesh->attachVertexBuffer(const_cast<VertexBuffer*>(iter.key()), j);
			j++;
		}
		
		// Update bounding volume information
		mesh->updateBounds();
		
		mlist.insert(mesh);
	}

	return mlist;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::mergeJobs() {
	// Loop through all jobs again
    const Material* curMaterial = NULL;
    const Shader* curShader = NULL;
    const Mode* curMode = NULL;
    const BlendMode* curBlendMode = NULL;
    const PolygonMode* curPolygonMode = NULL;
    const DepthMode* curDepthMode = NULL;
    const StencilMode* curStencilMode = NULL;
    const IndexBuffer* curIndexBuffer = NULL;
    const VertexBuffer* curVertexBuffers[Attributes::MaxAttributes];
	const Texture* curTextures[Texture::MaxTextures];
	unsigned int curStart = 0xffffffff, curEnd = 0xffffffff;
	Mesh::Primitive curPrimitive = Mesh::Unknown;
	memset(curVertexBuffers, 0, sizeof(curVertexBuffers));
	memset(curTextures, 0, sizeof(curTextures));

	unsigned int i, j;
	bool equal;
	for (i = 0; i < m_Jobs.getSize(); i++) {
		if (m_Jobs[i].m_Mesh != NULL) {
			continue;
		}

		equal = true;
		if (m_Jobs[i].m_Primitive != curPrimitive) {
			curPrimitive = m_Jobs[i].m_Primitive;
			equal = false;
		}
		if (m_Jobs[i].m_BlendMode != curBlendMode) {
			curBlendMode = m_Jobs[i].m_BlendMode;
			equal = false;
		}
		if (m_Jobs[i].m_DepthMode != curDepthMode) {
			curDepthMode = m_Jobs[i].m_DepthMode;
			equal = false;
		}
		if (m_Jobs[i].m_PolygonMode != curPolygonMode) {
			curPolygonMode = m_Jobs[i].m_PolygonMode;
			equal = false;
		}
		if (m_Jobs[i].m_StencilMode != curStencilMode) {
			curStencilMode = m_Jobs[i].m_StencilMode;
			equal = false;
		}
		if (m_Jobs[i].m_Shader != curShader) {
			curShader = m_Jobs[i].m_Shader;
			equal = false;
		}
		if (m_Jobs[i].m_Material != curMaterial) {
			curMaterial = m_Jobs[i].m_Material;
			equal = false;
		}
		if (m_Jobs[i].m_IndexBuffer != curIndexBuffer) {
			curIndexBuffer = m_Jobs[i].m_IndexBuffer;
			equal = false;
		}
		for (j = 0; j < Attributes::MaxAttributes; j++) {
			if (m_Jobs[i].m_VertexBuffers[j] != curVertexBuffers[j]) {
				curVertexBuffers[j] = m_Jobs[i].m_VertexBuffers[j];
				equal = false;
			}
		}
		for (j = 0; j < Texture::MaxTextures; j++) {
			if (m_Jobs[i].m_Textures[j] != curTextures[j]) {
				curTextures[j] = m_Jobs[i].m_Textures[j];
				equal = false;
			}
		}

		// Currently we cannot handle Triangle Strips
		if (curPrimitive == Mesh::TriangleStrip)
			equal = false;

		// Check if we can merge two subsequent jobs...
		if ((m_Jobs[i].m_Start == curEnd) && equal) {
			m_Jobs[i-1].m_Count += m_Jobs[i].m_Count;
			curEnd += m_Jobs[i].m_Count;
			m_Jobs.remove(i);
			i--;
			continue;
		}

		curStart = m_Jobs[i].m_Start;
		curEnd = m_Jobs[i].m_Start + m_Jobs[i].m_Count;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::mergeBuffers() {
	MeshMerger merger;
	Tuple<VertexBuffer*, VertexBuffer::MaxBuffers> vbt;
	IndexBuffer* ib;
	
	merger.setVertexSize(m_Splitter.getVertexSize());
	merger.setIndexSize(m_Splitter.getIndexSize());

	// Collect buffer infos
	unsigned int i, j;
	for (i = 0; i < m_Jobs.getSize(); i++) {
		if (m_Jobs[i].m_Mesh != NULL)
			continue;

		vbt = collectVertexBuffers(m_Jobs[i].m_VertexBuffers);
		ib = m_Jobs[i].m_IndexBuffer;
		merger.add(vbt, ib);
	}

	// merge buffers
	merger.finish();

	// Walk through all jobs and replace buffers
	for (i = 0; i < m_Jobs.getSize(); i++) {
		if (m_Jobs[i].m_Mesh != NULL)
			continue;

		vbt = collectVertexBuffers(m_Jobs[i].m_VertexBuffers);
		ib = m_Jobs[i].m_IndexBuffer;

		MeshMerger::VertexMapping vmap = merger.getMapping(vbt);
		for (j = 0; j < Attributes::MaxAttributes; j++) {
			if (m_Jobs[i].m_VertexBuffers[j] != NULL) 
				m_Jobs[i].m_VertexBuffers[j] = vmap.m_Buffer;
		}

		MeshMerger::IndexMapping imap = merger.getMapping(ib);
		if (m_Jobs[i].m_IndexBuffer != NULL) {
			m_Jobs[i].m_IndexBuffer = imap.m_Buffer;
			m_Jobs[i].m_Start += imap.m_Start;
		} else {
			m_Jobs[i].m_Start += vmap.m_Start;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimiser::splitBuffers() {
	unsigned int i;

	// Walk through all jobs and replace buffers
	for (i = 0; i < m_Jobs.getSize(); i++) {
		if (m_Jobs[i].m_Mesh != NULL)
			continue;

		if (m_Jobs[i].m_IndexBuffer == NULL)
			continue;
		
		IndexBuffer* src = m_Jobs[i].m_IndexBuffer;
		if (m_Jobs[i].m_Count == src->getSize())
			continue;
			
		unsigned int size = m_Jobs[i].m_Count;
		unsigned int start = m_Jobs[i].m_Start;
		AutoPointer<IndexBuffer> dst = IndexBuffer::create(src->getType(), size, IndexBuffer::System);
		ASSERT(dst != NULL);
		if (dst == NULL)
			continue;
		
		// Copy index buffer
		void* srcdata = src->lock(IndexBuffer::ReadOnly);
		srcdata = (byte *)srcdata + start * src->getType();
		void* dstdata = dst->lock(IndexBuffer::Discard);
		memcpy(dstdata, srcdata, size*src->getType());
		dst->unlock();
		src->unlock();

		m_Jobs[i].m_IndexBuffer = dst;
		m_Jobs[i].m_Start = 0;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int Optimiser::cmpJobs(const void* a,const void* b) {
	unsigned int i;
	Job* ja = (Job*)a;
	Job* jb = (Job*)b;

	// check shader
	if (ja->m_Shader != jb->m_Shader)
		return (int)((intptr_t)(void *)ja->m_Shader - (intptr_t)(void *)jb->m_Shader);

	// check textures
	for (i = 0; i < Texture::MaxTextures; i++) {
		if (ja->m_Textures[i] == jb->m_Textures[i])
			continue;
		return (int)((intptr_t)(void*)ja->m_Textures[i] - (intptr_t)(void*)jb->m_Textures[i]);
	}

	// check mode
	if (ja->m_BlendMode != jb->m_BlendMode)
		return (int)((intptr_t)(void *)ja->m_BlendMode - (intptr_t)(void *)jb->m_BlendMode);

	// check mode
	if (ja->m_StencilMode != jb->m_StencilMode)
		return (int)((intptr_t)(void *)ja->m_StencilMode - (intptr_t)(void *)jb->m_StencilMode);

	// check mode
	if (ja->m_DepthMode != jb->m_DepthMode)
		return (int)((intptr_t)(void *)ja->m_DepthMode - (intptr_t)(void *)jb->m_DepthMode);

	// check mode
	if (ja->m_PolygonMode != jb->m_PolygonMode)
		return (int)((intptr_t)(void *)ja->m_PolygonMode - (intptr_t)(void *)jb->m_PolygonMode);

	// check material
	if (ja->m_Material != jb->m_Material)
		return (int)((intptr_t)(void *)ja->m_Material - (intptr_t)(void *)jb->m_Material);

	// check index buffer
	if (ja->m_IndexBuffer != jb->m_IndexBuffer)
		return (int)((intptr_t)(void *)ja->m_IndexBuffer - (intptr_t)(void *)jb->m_IndexBuffer);
	
	// check vertex buffers
	for (i = 0; i < Attributes::MaxAttributes; i++) {
		if (ja->m_VertexBuffers[i] == jb->m_VertexBuffers[i])
			continue;
		return (int)((intptr_t)(void*)ja->m_VertexBuffers[i] - (intptr_t)(void*)jb->m_VertexBuffers[i]);
	}

	// check ranges
	if (ja->m_Start != jb->m_Start)
		return ja->m_Start - jb->m_Start;

	// check primitives
	if (ja->m_Primitive != jb->m_Primitive)
		return ja->m_Primitive - jb->m_Primitive;

	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Tuple<VertexBuffer*,VertexBuffer::MaxBuffers> Optimiser::collectVertexBuffers(const Pointer<VertexBuffer>* vb) {
	Tuple<VertexBuffer*,VertexBuffer::MaxBuffers> vbt;
	unsigned int i,j;

	for (j = 0; j < VertexBuffer::MaxBuffers; j++)
		vbt[j] = NULL;

	for (j = 0; j < Attributes::MaxAttributes; j++) {
		if (vb[j] == NULL) 
			continue;
		if (vbt.contains(vb[j]))
			continue;
		for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
			if (vbt[i] == NULL) {
				vbt[i] = vb[j];
				break;
			}
		}
	}

	return vbt;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int Optimiser::collectAttributes(const Tuple<VertexBuffer*,Attributes::MaxAttributes>& vbt) {
	unsigned int i, attribs = 0;

	for (i = 0; i < Attributes::MaxAttributes; i++) {
		if(vbt[i] != NULL) {
			attribs |= (1 << i);
		}
	}

	return attribs;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int Optimiser::collectAttributes(const Tuple<VertexBuffer*,Attributes::MaxAttributes>& vbt,const VertexBuffer* vb) {
	unsigned int i, attribs = 0;

	for (i = 0; i < Attributes::MaxAttributes; i++) {
		if(vbt[i] == vb) {
			attribs |= (1 << i);
		}
	}

	return attribs;
}

}; };
