/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "system.thread.h"
#include "graphics3d.phongshader.h"
#include "graphics3d.context.h"
#include "graphics3d.vertexformat.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.exception.h"

#include "object.inl"
#include "object.pointer.inl"
#include "base.string.inl"
#include "system.thread.inl"
#include "system.assembler.inl"
#include "container.array.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.fixedshader.inl"
#include "graphics3d.context.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.constant.inl"


namespace magnum { namespace graphics3d {

// works only with one point light!
static const char phong_vp_prg[] =
"!!ARBvp1.0\n"
"OPTION ARB_position_invariant;\n"
"\n"
"PARAM lightPos  = state.light[0].position;\n"
"PARAM m[4]={state.matrix.modelview};\n"
"PARAM mvinv[4]={state.matrix.modelview.invtrans};\n"
"\n"
"TEMP EyeCamPos, EyeNormal, EyeVertex;\n"
"\n"
"#Vertex position in eye space...\n"
"DP4    EyeVertex.x, m[0], vertex.position;\n"
"DP4    EyeVertex.y, m[1], vertex.position;\n"
"DP4    EyeVertex.z, m[2], vertex.position;\n"
"\n"
"#Transform the camera pos to eye space.\n"
"DP4    EyeCamPos.x, m[0], -m[3];\n"
"DP4    EyeCamPos.y, m[1], -m[3];\n"
"DP4    EyeCamPos.z, m[2], -m[3];\n"
"\n"
"#Transform the normal to eye space.\n"
"DP3    EyeNormal.x, mvinv[0], vertex.normal;\n"
"DP3    EyeNormal.y, mvinv[1], vertex.normal;\n"
"DP3    EyeNormal.z, mvinv[2], vertex.normal;\n"
"\n"
"#Normalize\n"
"DP3 EyeNormal.w, EyeNormal, EyeNormal;\n"
"RSQ EyeNormal.w, EyeNormal.w;\n"
"MUL EyeNormal, EyeNormal, EyeNormal.w;\n"
"\n"
"SUB    result.texcoord[1], state.light[0].position, EyeVertex;\n"
"SUB    result.texcoord[2], EyeCamPos, EyeVertex;\n"
"MOV result.texcoord[3], EyeNormal;\n"
"END\n";

static const char phong_fp_prg[] =
"!!ARBfp1.0\n"
"TEMP lightVec, viewVec, reflVec, normal, attenuation, Len, finalCol, lightContrib, tex;\n"
"PARAM two = {2.0, 2.0, 2.0, 2.0};\n"
"\n"
"#Normalize normal, lightvec and viewvec.\n"
"DP3        Len.w, fragment.texcoord[1], fragment.texcoord[1];\n"
"RSQ        lightVec.w, Len.w;\n"
"MUL        lightVec.xyz, fragment.texcoord[1], lightVec.w;\n"
"\n"
"DP3        viewVec.w, fragment.texcoord[2], fragment.texcoord[2];\n"
"RSQ        viewVec.w, viewVec.w;\n"
"MUL        viewVec.xyz, fragment.texcoord[2], viewVec.w;\n"
"\n"
"DP3        normal.w, fragment.texcoord[3], fragment.texcoord[3];\n"
"RSQ        normal.w, normal.w;\n"
"MUL        normal.xyz, fragment.texcoord[3], normal.w;\n"
"\n"
"#Calculate attenuation.\n"
"MAD        attenuation, state.light[0].attenuation.z, Len.w, state.light[0].attenuation.x;\n"
"RCP        Len, lightVec.w;\n"
"MAD        attenuation, Len.w, state.light[0].attenuation.y, attenuation.x;\n"
"RCP        attenuation.x, attenuation.x;\n"
"\n"
"#Diffuse\n"
"DP3_SAT	lightContrib.x, normal, lightVec;\n"
"ABS		lightContrib.x, lightContrib.x;\n"
"\n"
"#Specular\n"
"DP3        reflVec, lightVec, normal;\n"
"MUL        reflVec, reflVec, two;\n"
"MAD        reflVec, reflVec, normal, -lightVec;\n"
"\n"
"DP3_SAT	lightContrib.y, reflVec, viewVec;\n"
"ABS		lightContrib.y, lightContrib.y;\n"
"MOV        lightContrib.w, state.material.shininess.x;\n"
"\n"
"#Accelerates lighting computations\n"
"LIT_SAT    lightContrib, lightContrib;\n"
"\n"
"MAD        finalCol, lightContrib.y, state.lightprod[0].diffuse, state.lightprod[0].ambient;\n"
"\n"
"MAD        finalCol, lightContrib.z, state.lightprod[0].specular, finalCol;\n"
"MAD        finalCol, finalCol, attenuation.x, state.material.emission;\n"
"ADD        result.color.xyz, finalCol, state.lightmodel.scenecolor;\n"
"MOV        result.color.w, state.material.diffuse.w;\n"
"END\n";

static const char phong_fp_prg_twosided[] =
"!!ARBfp1.0\n"
"TEMP lightVec, viewVec, reflVec, normal, attenuation, Len, finalCol, lightContrib, tex;\n"
"PARAM two = {2.0, 2.0, 2.0, 2.0};\n"
"\n"
"#Normalize normal, lightvec and viewvec.\n"
"DP3        Len.w, fragment.texcoord[1], fragment.texcoord[1];\n"
"RSQ        lightVec.w, Len.w;\n"
"MUL        lightVec.xyz, fragment.texcoord[1], lightVec.w;\n"
"\n"
"DP3        viewVec.w, fragment.texcoord[2], fragment.texcoord[2];\n"
"RSQ        viewVec.w, viewVec.w;\n"
"MUL        viewVec.xyz, fragment.texcoord[2], viewVec.w;\n"
"\n"
"DP3        normal.w, fragment.texcoord[3], fragment.texcoord[3];\n"
"RSQ        normal.w, normal.w;\n"
"MUL        normal.xyz, fragment.texcoord[3], normal.w;\n"
"\n"
"#Calculate attenuation.\n"
"MAD        attenuation, state.light[0].attenuation.z, Len.w, state.light[0].attenuation.x;\n"
"RCP        Len, lightVec.w;\n"
"MAD        attenuation, Len.w, state.light[0].attenuation.y, attenuation.x;\n"
"RCP        attenuation.x, attenuation.x;\n"
"\n"
"#Diffuse\n"
"DP3		lightContrib.x, normal, lightVec;\n"
"ABS		lightContrib.x, lightContrib.x;\n"
"\n"
"#Specular\n"
"DP3        reflVec, lightVec, normal;\n"
"MUL        reflVec, reflVec, two;\n"
"MAD        reflVec, reflVec, normal, -lightVec;\n"
"\n"
"DP3		lightContrib.y, reflVec, viewVec;\n"
"ABS		lightContrib.y, lightContrib.y;\n"
"MOV        lightContrib.w, state.material.shininess.x;\n"
"\n"
"#Accelerates lighting computations\n"
"LIT_SAT    lightContrib, lightContrib;\n"
"\n"
"MAD        finalCol, lightContrib.y, state.lightprod[0].diffuse, state.lightprod[0].ambient;\n"
"\n"
"MAD        finalCol, lightContrib.z, state.lightprod[0].specular, finalCol;\n"
"MAD        finalCol, finalCol, attenuation.x, state.material.emission;\n"
"ADD        result.color.xyz, finalCol, state.lightmodel.scenecolor;\n"
"MOV        result.color.w, state.material.diffuse.w;\n"
"END\n";

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
PhongShader::PhongShader(Features f) : FixedShader(f) {
	publishAttributes(Attributes::Vertex | Attributes::Normal);

	// Check if all extensions are present
	const char *extensions = (const char *)(glGetString( GL_EXTENSIONS));
	if (extensions == NULL || strstr(extensions, "GL_ARB_fragment_program") == NULL)
		goto fallback;
	
	glGetError();
	glGenProgramsARB(1, &m_FragmentProgram);
	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, m_FragmentProgram);
	if (f & FixedShader::TwoSided)
		glProgramStringARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, 
			(GLsizei)strlen(phong_fp_prg_twosided), phong_fp_prg_twosided);
	else
		glProgramStringARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, 
			(GLsizei)strlen(phong_fp_prg), phong_fp_prg);
	if (glGetError() != GL_NO_ERROR)
		goto fallback;

	glGenProgramsARB(1, &m_VertexProgram);
	glBindProgramARB(GL_VERTEX_PROGRAM_ARB, m_VertexProgram);
	glProgramStringARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, (GLsizei)strlen(phong_vp_prg), phong_vp_prg);
	if (glGetError() != GL_NO_ERROR)
		goto fallback;

	return;
	
fallback:	
	m_FragmentProgram = 0;
	m_VertexProgram = 0;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void PhongShader::onSelect() const {
	// Fallback to standard OpenGL 1.2 stuff
	if (m_FragmentProgram == 0) {
		glEnable(GL_LIGHTING);
		glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
		glLightModeli(GL_LIGHT_MODEL_TWO_SIDE , (getFeatures() & TwoSided) ? GL_TRUE : GL_FALSE);
		glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
		return;
	}

	glBindProgramARB(GL_VERTEX_PROGRAM_ARB,m_VertexProgram);
	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, m_FragmentProgram);
	glEnable(GL_VERTEX_PROGRAM_ARB);
	glEnable(GL_FRAGMENT_PROGRAM_ARB);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void PhongShader::onDeselect() const {
	// Fallback to standard OpenGL 1.2 stuff
	if (m_FragmentProgram == 0) {
		glDisable(GL_LIGHTING);
		return;
	}

	glDisable(GL_VERTEX_PROGRAM_ARB);
	glDisable(GL_FRAGMENT_PROGRAM_ARB);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void PhongShader::onCompileSelect(Assembler& as) const {
	// Fallback to standard OpenGL 1.2 stuff
	if (m_FragmentProgram == 0) {
		as.asmCall(glEnable,(GLenum)GL_LIGHTING);
		as.asmCall(glLightModeli,(GLenum)GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
		as.asmCall(glLightModeli,(GLenum)GL_LIGHT_MODEL_TWO_SIDE , (getFeatures() & TwoSided) ? GL_TRUE : GL_FALSE);
		as.asmCall(glLightModeli,(GLenum)GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
		return;
	}

	as.asmCall(glBindProgramARB,(GLenum)GL_VERTEX_PROGRAM_ARB,m_VertexProgram);
	as.asmCall(glBindProgramARB,(GLenum)GL_FRAGMENT_PROGRAM_ARB, m_FragmentProgram);
	as.asmCall(glEnable,(GLenum)GL_VERTEX_PROGRAM_ARB);
	as.asmCall(glEnable,(GLenum)GL_FRAGMENT_PROGRAM_ARB);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void PhongShader::onCompileDeselect(Assembler& as) const {
	// Fallback to standard OpenGL 1.2 stuff
	if (m_FragmentProgram == 0) {
		as.asmCall(glDisable,(GLenum)GL_LIGHTING);
		return;
	}

	as.asmCall(glDisable,(GLenum)GL_VERTEX_PROGRAM_ARB);
	as.asmCall(glDisable,(GLenum)GL_FRAGMENT_PROGRAM_ARB);
}

}; };
