/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "system.thread.h"
#include "graphics3d.programshader.h"
#include "graphics3d.context.h"
#include "graphics3d.vertexformat.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.exception.h"

#include "object.inl"
#include "object.pointer.inl"
#include "system.thread.inl"
#include "system.assembler.inl"
#include "container.iterator.inl"
#include "container.array.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.programshader.inl"
#include "graphics3d.shaderprogram.inl"
#include "graphics3d.vertexprogram.inl"
#include "graphics3d.pixelprogram.inl"
#include "graphics3d.context.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.constant.inl"


namespace magnum { namespace graphics3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ProgrammableShader* ProgrammableShader::create() {
	return new ProgrammableShader();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ProgrammableShader::ProgrammableShader() {
	if (glCreateProgram == 0)
		throw UnsupportedFeatureException();
	m_OglProgramHandle = glCreateProgram();
	if (glGetError() != GL_NO_ERROR)
		throw InvalidOperationException();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ProgrammableShader::~ProgrammableShader() {
	if (m_OglProgramHandle != 0)
		glDeleteProgram(m_OglProgramHandle);
	m_OglProgramHandle = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool ProgrammableShader::attach(ShaderProgram* prg) {
	if (m_Programs.contains(prg))
		return true;
	glAttachShader(m_OglProgramHandle, prg->m_OglShaderHandle);
	if (glGetError() !=  GL_NO_ERROR)
		return false;
	m_Programs.insert(prg);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool ProgrammableShader::detach(ShaderProgram* prg) {
	Iterator iter;
	for (iter = m_Programs.begin(); iter != m_Programs.end(); ++iter) {
		if (*iter == prg)
			break;
	}
	if (iter == m_Programs.end())
		return true;
	glDetachShader(m_OglProgramHandle, prg->m_OglShaderHandle);
	if (glGetError() !=  GL_NO_ERROR)
		return false;
	m_Programs.remove(iter);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void ProgrammableShader::clear() {
	for (Iterator iter = m_Programs.begin(); iter != m_Programs.end(); ++iter) {
		glDetachShader(m_OglProgramHandle, (*iter)->m_OglShaderHandle);
	}
	m_Programs.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void ProgrammableShader::link() {
	glLinkProgram(m_OglProgramHandle);
	if (glGetError() != GL_NO_ERROR)
		throw ShaderLinkerException();
		
	// Get current shader
	GLint oldprg;
	glGetIntegerv(GL_CURRENT_PROGRAM, &oldprg);
		
	// Get all attributes
	GLint ac,al,i,size;
	GLenum type;
	glGetProgramiv(m_OglProgramHandle, GL_ACTIVE_ATTRIBUTES, &ac);
	glGetProgramiv(m_OglProgramHandle, GL_ACTIVE_ATTRIBUTE_MAX_LENGTH, &al);
	al++;
	unsigned int att = 0;
	char* name = new char[al];
	for (i = 0; i < ac; i++) {
		glGetActiveAttrib(m_OglProgramHandle, i, al, NULL, &size, &type, name);
		att |= mapAttribute(name);
	}
	delete[] name;
	publishAttributes(att);
	
	// Get all uniforms	
	GLint uc,ul,loc;
	glUseProgram(m_OglProgramHandle);
	glGetProgramiv(m_OglProgramHandle, GL_ACTIVE_UNIFORMS, &uc);
	glGetProgramiv(m_OglProgramHandle, GL_ACTIVE_UNIFORM_MAX_LENGTH, &ul);
	ul++;
	name = new char[ul];
	GLint tex = 0;
	for (i = 0; i < uc; i++) {
		glGetActiveUniform(m_OglProgramHandle, i, ul, NULL, &size, &type, name);
		loc = glGetUniformLocation(m_OglProgramHandle, name);
		// Bind texture to unit (if neccesary)
		if ((type == GL_SAMPLER_1D) || (type == GL_SAMPLER_2D) || (type == GL_SAMPLER_3D)
		|| (type == GL_SAMPLER_CUBE) || (type == GL_SAMPLER_1D_SHADOW) || (type == GL_SAMPLER_2D_SHADOW))
			glUniform1i(loc, tex++);
	}
	delete[] name;
	
	glUseProgram(oldprg);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool ProgrammableShader::isValid() const {
	glValidateProgram(m_OglProgramHandle);
	GLint val;
	glGetProgramiv(m_OglProgramHandle, GL_VALIDATE_STATUS, &val);
	return (val == GL_TRUE);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void ProgrammableShader::onCompileSelect(Assembler& as) const {
	as.asmCall(glUseProgram, m_OglProgramHandle);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void ProgrammableShader::onCompileDeselect(Assembler& as) const {
	as.asmCall(glUseProgram, (unsigned int)0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void ProgrammableShader::onSelect() const {
	glUseProgram(m_OglProgramHandle);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void ProgrammableShader::onDeselect() const {
	glUseProgram(0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int ProgrammableShader::mapAttribute(const char* lwrname) {
	String name = String(lwrname).toLower();
	if (strcmp(name, "gl_vertex"))
		return Attributes::Vertex;
	else if (strcmp(name, "gl_normal"))
		return Attributes::Normal;
	else if (strcmp(name, "gl_color"))
		return Attributes::Colour1;
	else if (strcmp(name, "gl_secondarycolor"))
		return Attributes::Colour2;
	else if (strcmp(name, "gl_multitexcoord0"))
		return Attributes::Texture0;
	else if (strcmp(name, "gl_multitexcoord1"))
		return Attributes::Texture1;
	else if (strcmp(name, "gl_multitexcoord2"))
		return Attributes::Texture2;
	else if (strcmp(name, "gl_multitexcoord3"))
		return Attributes::Texture3;
	else if (strcmp(name, "gl_multitexcoord4"))
		return Attributes::Texture4;
	else if (strcmp(name, "gl_multitexcoord5"))
		return Attributes::Texture5;
	else if (strcmp(name, "gl_multitexcoord6"))
		return Attributes::Texture6;
	else if (strcmp(name, "gl_multitexcoord7"))
		return Attributes::Texture7;
	else if (strcmp(name, "gl_fogcoord"))
		return Attributes::Fog;
	return 0;
}

}; };
