/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "container.pair.h"
#include "container.array.h"
#include "container.map.h"
#include "geometry.trianglemesh.h"
#include "math.vector4.h"
#include "graphics3d.exception.h"
#include "graphics3d.projectionvolume.h"
#include "graphics3d.mesh.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.material.h"
#include "graphics3d.shader.h"
#include "graphics3d.context.h"

#include "object.inl"
#include "base.string.inl"
#include "container.pair.inl"
#include "container.array.inl"
#include "container.map.inl"
#include "container.iterator.inl"
#include "container.allocator.inl"
#include "math.vector.inl"
#include "math.vector3.inl"
#include "math.vector4.inl"
#include "math.range.inl"
#include "geometry.trianglemesh.inl"
#include "graphics3d.context.inl"
#include "graphics3d.material.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.projectionvolume.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.resource.inl"

namespace magnum { namespace graphics3d {
using namespace magnum::container;
using magnum::math::Vector4;
using magnum::math::Vector;


typedef Pair<unsigned int, unsigned int> Ipair;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
static void insertEdge(Map<Ipair,Ipair>* edges, unsigned int a,unsigned int b, unsigned int tri) {
	Ipair curEdge;
	Map<Ipair,Ipair>::Iterator iter;
	if (a < b) {
		curEdge = Ipair(a, b);
		iter = edges->at(curEdge);
		if (iter != edges->end()) {
//			ASSERT(iter->first == ~0);
			iter->first = tri;
		} else {
			edges->insert(curEdge, Ipair(tri,~0));
		}
	} else {
		curEdge = Ipair(b, a);
		iter = edges->at(curEdge);
		if (iter != edges->end()) { 
//			ASSERT(iter->second == ~0);
			iter->second = tri;
		} else {
			edges->insert(curEdge, Ipair(~0,tri));
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ProjectionVolume::ProjectionVolume() {
	m_VertexFormat.insert(Types::Float3, Attributes::Vertex);
	m_VertexBuffer = NULL;
	m_IndexBuffer = NULL;
	m_EdgeCount = 0;
	m_FaceCount = 0;
	m_VertexCount = 0;
	m_IndexRange = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ProjectionVolume::~ProjectionVolume() {
	if (m_VertexBuffer != NULL)
		m_VertexBuffer->release();
	if (m_IndexBuffer != NULL)
		m_IndexBuffer->release();
	m_VertexBuffer = NULL;
	m_IndexBuffer = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void ProjectionVolume::clear() {
	m_TriangleMesh.clear();
	m_Edges.clear();
	m_Signs.clear();
	m_Normals.clear();
	m_Vertices.clear();
	if (m_VertexBuffer != NULL)
		m_VertexBuffer->release();
	if (m_IndexBuffer != NULL)
		m_IndexBuffer->release();

	m_VertexBuffer = NULL;
	m_IndexBuffer = NULL;
	m_EdgeCount = 0;
	m_FaceCount = 0;
	m_VertexCount = 0;
	m_IndexRange = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool ProjectionVolume::add(const Mesh* mesh) {
	TriangleMesh3<float>* tmesh = mesh->triangulate();
	if (tmesh == NULL)
		return false;

	m_TriangleMesh.append(tmesh);
	tmesh->release();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool ProjectionVolume::add(const TriangleMesh3<float>* mesh) {
	m_TriangleMesh.append(mesh);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void ProjectionVolume::finish() {
	unsigned int i, i1, i2, i3;
	Vector3<float> v1,v2,v3,e1,e2;
	Vector3<unsigned int> triangle;
	Vector4<float> normal;

	// Optimize the triangle mesh;
//	m_TriangleMesh.optimize(0.0f);

	// Setup a map
	Map<Ipair,Ipair> edges;
	Map<Ipair,Ipair>::Iterator iter, end;

	// Allocate space for normal list
	m_FaceCount = m_TriangleMesh.getSize();
	if (m_FaceCount == 0)
		return;
	m_VertexCount = m_TriangleMesh.getVertices().getSize();
	if (m_VertexCount == 0)
		return;

	// Allocate space for normal list
	m_Normals.setSize(m_FaceCount);
	m_Vertices.setSize(m_VertexCount);
	m_Signs.setSize(m_FaceCount);

	// Copy Vertices;
	m_TriangleMesh.getVertices((Vector<float,3>*)m_Vertices.getArray(), m_VertexCount);

	// Process all faces
	for (i = 0; i < m_FaceCount; i++) {
		// Extract traingle information
		triangle = m_TriangleMesh.getIndices(i);
		i1 = triangle[0];
		i2 = triangle[1];
		i3 = triangle[2];
		v1 = m_TriangleMesh.getVertex(i1);
		v2 = m_TriangleMesh.getVertex(i2);
		v3 = m_TriangleMesh.getVertex(i3);

		// Calc normal for face
		e1 = v2 - v1;
		e2 = v3 - v1;
		normal = Vector4<float>(e1 % e2);
		normal.w = normal.xyz() * v1;

		m_Normals[i] = normal;
//		if (normal.length() < 0.1f)
//			continue;

		// Set normal index for all edges of this tri
		insertEdge(&edges, i1, i2, i);
		insertEdge(&edges, i2, i3, i);
		insertEdge(&edges, i3, i1, i);
	}

	// Create Edgebuffer
	m_EdgeCount = edges.getSize();
	m_Edges.setSize(m_EdgeCount);

	// Copy all edges into edge-buffer
	i = 0;
	end = edges.end();
	for (iter = edges.begin(); iter != end; iter++) {
		if ((iter->first == ~0) || (iter->second == ~0)) {
			m_EdgeCount--;
			continue;
		}
		m_Edges[i].m_Normal[0] = iter.item().first;
		m_Edges[i].m_Normal[1] = iter.item().second;
		m_Edges[i].m_Vertex[0] = iter.key().first;
		m_Edges[i].m_Vertex[1] = iter.key().second;
		i++;
	}

	// Create VertexBuffer
	m_VertexBuffer = VertexBuffer::create(m_VertexFormat, 2*m_VertexCount, VertexBuffer::Dynamic);
	m_VertexBuffer->setBuffer(0, m_TriangleMesh.getVertices().getArray(), m_VertexCount);
	m_IndexBuffer = IndexBuffer::create(IndexBuffer::Index16, m_EdgeCount*6, IndexBuffer::Dynamic);

	// Clear Trianglemesh
	m_TriangleMesh.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void ProjectionVolume::update(const Vector3<float>& lightpos, float lsize) {
	unsigned int i, n1, n2, v1, v2;
	int8* signs = m_Signs.getArray();
	Vector4<float> normal;
	Vector3<float> vec, dir;
	uint16* indices;
	Vector3<float>* vertices; 
	unsigned int faces = 0;
	unsigned int sign;
	float dist;

	if ((m_EdgeCount == 0) || (m_FaceCount == 0) || (m_VertexCount == 0))
		return;

	indices = (uint16*)m_IndexBuffer->lock(IndexBuffer::Discard);
	vertices = (Vector3<float>*)m_VertexBuffer->lock(VertexBuffer::WriteOnly);
	vertices = vertices + m_VertexCount;

	// Calc normal signs
	for (i = 0; i < m_FaceCount; i++) {
		normal = m_Normals[i];
		float val = normal.xyz() * lightpos - normal.w;

		if (val < 0)
			*signs = 0;
		else
			*signs = 1;
		signs++;
	}

	// Create indices
	for (i = 0; i < m_EdgeCount; i++) {
		n1 = m_Edges[i].m_Normal[0];
		n2 = m_Edges[i].m_Normal[1];

		if ((m_Signs[n1] ^ m_Signs[n2]) == 0)
			continue;
		// We need the sign later...
		sign = m_Signs[n1];

		v1 = m_Edges[i].m_Vertex[0];
		v2 = m_Edges[i].m_Vertex[1];

		// Calc extruded edge from v1,v2
		vec = m_Vertices[v1];
		dir = vec - lightpos;
		dist = dir.length();
		if (dist > lsize)
			continue;
		vec = vec + (lsize - dist)/dist*dir;
		vertices[v1] = vec;

		vec = m_Vertices[v2];
		dir = vec - lightpos;
		dist = dir.length();
		if (dist > lsize)
			continue;
		vec = vec + (lsize - dist)/dist*dir;
		vertices[v2] = vec;

		// Output the correct quad (in relation to the sign)
		if (sign) {
			// Output first triangle
			*indices = (uint16)v1;
			indices++;
			*indices = (uint16)v2;
			indices++;
			*indices = (uint16)v2 + m_VertexCount;
			indices++;
			faces++;

			// Output second triangle
			*indices = (uint16)v1;
			indices++;
			*indices = (uint16)v2 + m_VertexCount;
			indices++;
			*indices = (uint16)v1 + m_VertexCount;
			indices++;
			faces++;
		} else {
			// Output first triangle
			*indices = (uint16)v2;
			indices++;
			*indices = (uint16)v1;
			indices++;
			*indices = (uint16)v2 + m_VertexCount;
			indices++;
			faces++;

			// Output second triangle
			*indices = (uint16)v1;
			indices++;
			*indices = (uint16)v1 + m_VertexCount;
			indices++;
			*indices = (uint16)v2 + m_VertexCount;
			indices++;
			faces++;
		}

	}

	m_VertexBuffer->unlock();
	m_IndexBuffer->unlock();
	m_IndexRange = faces*3;
}


}; };
