/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics3d.rendertarget.h"
#include "graphics3d.exception.h"
#include "graphics3d.texture1d.h"
#include "graphics3d.texture2d.h"
#include "graphics3d.texture3d.h"
#include "graphics3d.cubetexture.h"

#include "object.inl"
#include "object.exception.inl"
#include "object.pointer.inl"
#include "graphics.dimension.inl"
#include "graphics3d.context.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.rendertarget.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.texture1d.inl"
#include "graphics3d.texture2d.inl"
#include "graphics3d.texture3d.inl"
#include "graphics3d.cubetexture.inl"

#include <typeinfo.h>

namespace magnum { namespace graphics3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool RenderTarget::isSupported() {
	return (glGenRenderbuffersEXT != NULL);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
RenderTarget::RenderTarget(GLuint buffer) {
	ASSERT(buffer != 0);
	GLint p;
	m_OglRenderbuffer = buffer;
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, buffer);
	glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, GL_RENDERBUFFER_WIDTH_EXT, &p);
	m_Size.setWidth(p);
	glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, GL_RENDERBUFFER_HEIGHT_EXT, &p);
	m_Size.setHeight(p);
	glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, GL_RENDERBUFFER_INTERNAL_FORMAT_EXT, &p);
	m_Type = (Type)p;
	m_ZOffset = 0;
	m_OglTexture = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
RenderTarget::RenderTarget(Texture* tex,unsigned int slice) {
	ASSERT(tex != NULL);
	m_OglRenderbuffer = 0;
	m_Texture = tex;
	m_OglTexture = tex->m_OglTexture;
	m_Type = mapType(tex->getType());
	m_ZOffset = 0;
	if (typeid(*tex) == typeid(Texture1D)) {
		m_Size.setWidth(((Texture1D*)tex)->getSize());
		m_Size.setHeight(0);
	} else if (typeid(*tex) == typeid(Texture2D)) {
		m_Size = ((Texture2D*)tex)->getSize();
	} else if (typeid(*tex) == typeid(Texture3D)) {
		m_Size.setWidth(((Texture3D*)tex)->getWidth());		
		m_Size.setHeight(((Texture3D*)tex)->getHeight());
		m_ZOffset = slice;
	} else if (typeid(*tex) == typeid(CubeTexture)) {
		m_Size = ((CubeTexture*)tex)->getSize();
		m_ZOffset = slice;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
RenderTarget::~RenderTarget() {
	if (m_OglRenderbuffer != 0) {
		ASSERT(glDeleteRenderbuffersEXT != NULL);
		glDeleteRenderbuffersEXT(1, &m_OglRenderbuffer);
		m_OglRenderbuffer = 0;
	}
	if (m_Texture != NULL) {
		m_Texture->release();
		m_Texture = NULL;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
RenderTarget* RenderTarget::create(RenderTarget::Type type,const Dimension<int>& dim) {
	return create(type, dim.getWidth(), dim.getHeight());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
RenderTarget* RenderTarget::create(RenderTarget::Type type,unsigned int w,unsigned int h) {
	if (glGenRenderbuffersEXT == NULL)
		throw UnsupportedFeatureException();
		
	GLuint buffer;
	glGenRenderbuffersEXT(1, &buffer);
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, buffer);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, type, w, h);	
	
	return new RenderTarget(buffer);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
RenderTarget* RenderTarget::create(Texture2D* tex) {
	if (glGenRenderbuffersEXT == NULL)
		throw UnsupportedFeatureException();
		
	if (mapType(tex->getType()) == -1)
		throw UnsupportedFormatException();
		
	return new RenderTarget(tex);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
RenderTarget* RenderTarget::create(CubeTexture* tex,CubeTexture::Face face) {
	if (glGenRenderbuffersEXT == NULL)
		throw UnsupportedFeatureException();

	if (mapType(tex->getType()) == -1)
		throw UnsupportedFormatException();
		
	return new RenderTarget(tex,face);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void RenderTarget::onAttach(FrameBuffer* fb,RenderTarget::Attachment att) {
	GLint oldbuffer;
	glGetIntegerv(GL_FRAMEBUFFER_BINDING_EXT, &oldbuffer);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb->m_OglFramebuffer);
	if (m_OglRenderbuffer != 0) {
		glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, att, GL_RENDERBUFFER_EXT, m_OglRenderbuffer);
	} else if (typeid(*m_Texture) == typeid(Texture1D)){
		glFramebufferTexture1DEXT(GL_FRAMEBUFFER_EXT, att, GL_TEXTURE_1D, m_OglTexture, 0);
	} else if (typeid(*m_Texture) == typeid(Texture2D)){
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, att, GL_TEXTURE_2D, m_OglTexture, 0);
	} else if (typeid(*m_Texture) == typeid(Texture3D)){
		glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT, att, GL_TEXTURE_3D, m_OglTexture, 0, m_ZOffset);
	} else if (typeid(*m_Texture) == typeid(CubeTexture)){
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, att, m_ZOffset, m_OglTexture, 0);
	}
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, oldbuffer);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void RenderTarget::onDetach(FrameBuffer* fb,RenderTarget::Attachment att) {
	GLint oldbuffer;
	glGetIntegerv(GL_FRAMEBUFFER_BINDING_EXT, &oldbuffer);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb->m_OglFramebuffer);
	if (m_OglRenderbuffer != 0) {
		glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, att, GL_RENDERBUFFER_EXT, 0);
	} else if (typeid(*m_Texture) == typeid(Texture1D)){
		glFramebufferTexture1DEXT(GL_FRAMEBUFFER_EXT, att, GL_TEXTURE_1D, 0, 0);
	} else if (typeid(*m_Texture) == typeid(Texture2D)){
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, att, GL_TEXTURE_2D, 0, 0);
	} else if (typeid(*m_Texture) == typeid(Texture3D)){
		glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT, att, GL_TEXTURE_3D, 0, 0, 0);
	} else if (typeid(*m_Texture) == typeid(CubeTexture)){
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, att, m_ZOffset, 0, 0);
	}
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, oldbuffer);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
RenderTarget::Type RenderTarget::mapType(Image::Type t) {
	switch(t) {
		case Image::RGB8:
			return RGB8;
		case Image::RGBA8:
			return RGBA8;
		case Image::RGB16:
			return RGB16;
		case Image::RGBA16:
			return RGBA16;
		case Image::Depth16:
			return Depth16;
		case Image::Depth24:
			return Depth24;
		default:
			return (Type)-1;
	}
}
}; };
