/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "system.thread.h"
#include "graphics3d.shader.h"
#include "graphics3d.context.h"
#include "graphics3d.vertexformat.h"

#include "object.inl"
#include "object.pointer.inl"
#include "system.thread.inl"
#include "system.assembler.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.context.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.constant.inl"


namespace magnum { namespace graphics3d {


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Shader::forceSelect() const {
	Context* context = getContext();
	const Shader* cur = context->m_Shader;
	context->m_Shader = this;
	if (cur != NULL) {
//		if (typeid(*cur) != typeid(*this))
			cur->onDeselect();
	}
	selectAttributes();
	onSelect();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Shader::selectDefault() {
	Context* context = getContext();
	const Shader* cur = context->m_Shader;
	if (cur == &context->m_DefaultShader)
		return;
	context->m_Shader = &context->m_DefaultShader;
	if (cur != NULL) 
		cur->onDeselect();
	context->m_Shader->selectAttributes();
	context->m_Shader->onSelect();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Shader::compileSelect(Assembler& as) const {
	Context* context = getContext();
	int offset = (int)((byte*)&context->m_Shader - (byte*)context);
	as.asmCall((void (Shader::*)() const)&Shader::selectAttributes, this);
	as.asmLoad(Assembler::T0, (int)(intptr_t)this);
	as.asmStore(Assembler::R0, offset, Assembler::T0);

	onCompileSelect(as);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Shader::compileDeselect(Assembler& as) const {
	Context* context = getContext();
	int offset = (int)((byte*)&context->m_Shader - (byte*)context);
	as.asmLoad(Assembler::T0, NULL);
	as.asmStore(Assembler::R0, offset, Assembler::T0);
	as.asmCall(&Shader::selectNone);

	onCompileDeselect(as);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Shader::onCompileSelect(Assembler& as) const {
	as.asmCall((void (Shader::*)() const)&Shader::onSelect, this);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Shader::onCompileDeselect(Assembler& as) const {
	as.asmCall((void (Shader::*)() const)&Shader::onDeselect, this);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Shader::onSelect() const {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Shader::onDeselect() const {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Shader::selectNone() {
	Context* context = getContext();
	if (context->m_Shader != NULL) {
		context->m_Shader->onDeselect();
		context->m_Shader = NULL;
	}

	unsigned int disable = context->m_Attributes;
	context->m_Attributes = 0;

	if (disable & Attributes::Vertex)
		glDisableClientState(GL_VERTEX_ARRAY);
	if (disable & Attributes::Normal)
		glDisableClientState(GL_NORMAL_ARRAY);
	if (disable & Attributes::Colour)
		glDisableClientState(GL_COLOR_ARRAY);
	if (disable & Attributes::Edge)
		glDisableClientState(GL_EDGE_FLAG_ARRAY);
	if (disable & Attributes::Fog)
		glDisableClientState(GL_FOG_COORD_ARRAY);
	if (disable & Attributes::Weights)
		glDisableClientState(GL_WEIGHT_ARRAY_ARB);

	for (unsigned int i = 0; i < Texture::MaxTextures; i++) {
		if (disable & (Attributes::Texture0 << i)) {
			glClientActiveTextureARB(GL_TEXTURE0_ARB + i);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Shader::selectAttributes() const {
	Context* context = getContext();
	if (context->m_Attributes == m_Attributes)
		return;

	unsigned int enable = m_Attributes & ~context->m_Attributes;
	unsigned int disable = !m_Attributes & context->m_Attributes;
	context->m_Attributes = m_Attributes;

	glClientActiveTextureARB(GL_TEXTURE0_ARB);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	if (disable & Attributes::Vertex)
		glDisableClientState(GL_VERTEX_ARRAY);
	else if (enable & Attributes::Vertex)
		glEnableClientState(GL_VERTEX_ARRAY);

	if (disable & Attributes::Normal)
		glDisableClientState(GL_NORMAL_ARRAY);
	else if (enable & Attributes::Normal)
		glEnableClientState(GL_NORMAL_ARRAY);

	if (disable & Attributes::Colour)
		glDisableClientState(GL_COLOR_ARRAY);
	else if (enable & Attributes::Colour)
		glEnableClientState(GL_COLOR_ARRAY);

	if (disable & Attributes::Edge)
		glDisableClientState(GL_EDGE_FLAG_ARRAY);
	else if (enable & Attributes::Edge)
		glEnableClientState(GL_EDGE_FLAG_ARRAY);

	if (disable & Attributes::Fog)
		glDisableClientState(GL_FOG_COORD_ARRAY);
	else if (enable & Attributes::Fog)
		glEnableClientState(GL_FOG_COORD_ARRAY);

	if (disable & Attributes::Weights)
		glDisableClientState(GL_WEIGHT_ARRAY_ARB);
	else if (enable & Attributes::Weights)
		glEnableClientState(GL_WEIGHT_ARRAY_ARB);

	for (unsigned int i = 0; i < Texture::MaxTextures; i++) {
		if (disable & (Attributes::Texture0 << i)) {
			glClientActiveTextureARB(GL_TEXTURE0_ARB + i);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		} else if (enable & (Attributes::Texture0 << i)) {
			glClientActiveTextureARB(GL_TEXTURE0_ARB + i);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		}
	}
}


}; };
