/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics3d.staticmodel.h"
#include "graphics3d.mesh.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.material.h"
#include "graphics3d.shader.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.appearance.h"
#include "graphics3d.context.h"
#include "graphics3d.optimiser.h"
#include "graphics3d.parameter.h"
#include "graphics3d.parameterbuffer.h"

#include "base.string.inl"
#include "object.inl"
#include "object.pointer.inl"
#include "math.matrix.inl"
#include "math.vector.inl"
#include "container.iterator.inl"
#include "container.allocator.inl"
#include "container.tuple.inl"
#include "container.array.inl"
#include "container.list.inl"
#include "container.pair.inl"
#include "system.program.inl"
#include "geometry.bounds.inl"
#include "geometry.sphere.inl"
#include "geometry.box.inl"
#include "geometry.alignedbox.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.joint.inl"
#include "graphics3d.model.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.context.inl"
#include "graphics3d.deferqueue.inl"
#include "graphics3d.deferedmesh.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.optimiser.inl"
#include "graphics3d.matrixstack.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.renderstream.inl"
#include "graphics3d.parameterbuffer.inl"
#include "graphics3d.parameterformat.inl"


namespace magnum { namespace graphics3d {
using namespace magnum::container;
using magnum::geometry::AlignedBox;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
const Joint& StaticModel::insertJoint(const char* n,const Matrix4<float>& mtx) {
	Matrix4<float>* m = m_Joints.insert(mtx);
	return publishJoint(n,reinterpret_cast<Joint::Handle>(m));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void StaticModel::removeJoint(const char* n) {
	const JointList& list = getJoints();
	for (JointList::Iterator iter = list.begin(); iter != list.end(); ++iter) {
		if (strncmp(iter->getName() ,n,64))
			continue;
		removeJoint(*iter);
	}
}	


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void StaticModel::removeJoint(const Joint& j) {
	m_Joints.remove(reinterpret_cast<Matrix4<float>*>(j.getHandle()));
	unpublishJoint(j);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void StaticModel::finish() {
	// Recreate parameter format
	ParameterFormat fmt;
	fmt.insert(Parameters::ModelView, Types::Matrix4f);
	publishParameters(fmt);

	// Early out for empty models	
	if (getMeshes().getSize() == 0)
		return;

	// Create bounds
	Array<Vector<float,3> > verts;
	verts.setSize(getMeshes().getSize() * 8);
	unsigned int i = 0;
	for (MeshList::Iterator iter = getMeshes().begin(); iter != getMeshes().end(); ++iter) {
		(*iter)->getBounds().getVertices(&verts[i]);
		i += 8;
	}
	
	AlignedBox<float,3> abox;
	abox.bound(&verts[0], getMeshes().getSize() * 8);
	setBounds(Bounds3<float>(abox));		
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void StaticModel::onClear() {
	Model::onClear();
	m_Joints.clear();
	
	for (MeshList::Iterator iter = m_Transparents.begin(); iter != m_Transparents.end(); ++iter) {
		(*iter)->release();
	}
	m_Transparents.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void StaticModel::onRender(const ParameterBuffer* pbuf) const {
	ASSERT(pbuf != NULL);
	Context* context = getContext();
	MatrixStack* ms = context->getMatrixStack();
	const Matrix4<float>* mtx = (const Matrix4<float>*)pbuf->getBuffer();
	ms->selectMode(MatrixStack::ModelView);
	ms->set(*mtx);

	if (isOptimised()) {
		m_Stream->execute();
		for (MeshList::Iterator iter = m_Transparents.begin(); iter != m_Transparents.end(); ++iter)
			context->defer(*iter, *mtx);
	} else {
		MeshList::Iterator end = getMeshes().end();
		for (MeshList::Iterator iter = getMeshes().begin(); iter != end; ++iter) {
			Mesh* mesh = *iter;
			Appearance* app = mesh->getAppearance();
			if ((app != NULL) && (app->isTransparent()))
				context->defer(mesh, *mtx);
			else
				mesh->render();
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void StaticModel::onTransform(const ParameterBuffer* pbuf, Matrix4<float>* dst, Joint::Handle hnd) const {
	*dst = *(const Matrix4<float>*)hnd;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool StaticModel::onOptimise() {
	// Clear list of transparent meshes
	m_Transparents.clear();

	// Create optimiser object
	Optimiser optimiser;

	// Insert all meshes into optimiser
	for (MeshList::Iterator iter = getMeshes().begin(); iter != getMeshes().end(); iter++) {
		Mesh* mesh = *iter;
		Appearance* app = mesh->getAppearance();
		if ((app != NULL) && (app->isTransparent())) {
			mesh->refer();
			m_Transparents.insert(mesh);	
		} else {
			optimiser.renderMesh(mesh);
		}
	}

	optimiser.finish();

	// Create render stream
	m_Stream = optimiser.createStream();
	if (m_Stream == NULL)
		return false;
	m_Stream->release();
	
	// Republish mesh information
	clearMeshes();
	MeshList mlist = optimiser.createMeshes();
	for (MeshList::Iterator iter = mlist.begin(); iter != mlist.end(); ++iter) {
		publishMesh(*iter);
		(*iter)->release();
	}
	for (MeshList::Iterator iter = m_Transparents.begin(); iter != m_Transparents.end(); ++iter) {
		publishMesh(*iter);
	}

	return true;
}


}; };
