/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics.image.h"
#include "graphics3d.context.h"
#include "graphics3d.texture.h"
#include "graphics3d.vertexformat.h"

#include "object.inl"
#include "base.string.inl"
#include "system.thread.inl"
#include "system.assembler.inl"
#include "graphics.palette.inl"
#include "graphics.colour.inl"
#include "graphics3d.context.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.vertexformat.inl"


namespace magnum { namespace graphics3d {
using magnum::graphics::Image;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Texture::Texture(unsigned int tex, unsigned int target) {
	int tmp;

	m_OglTexture = tex;
	m_OglTarget = target;
	m_MagFilter = Nearest;
	m_MinFilter = Nearest;
	m_WrapS = Repeat;
	m_WrapT = Repeat;
	m_WrapU = Repeat;
	m_MinLod = -1000.0f;
	m_MaxLod = 1000.0f;
	m_LodBias = 0.0f;

	glBindTexture(m_OglTarget, tex);
	if (m_OglTarget == GL_TEXTURE_CUBE_MAP)
		glGetTexLevelParameteriv(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 0, GL_TEXTURE_INTERNAL_FORMAT, &tmp);
	else
		glGetTexLevelParameteriv(m_OglTarget, 0, GL_TEXTURE_INTERNAL_FORMAT, &tmp);
	switch (tmp) {
		case 1: m_Type = Image::Index8; break;
		case GL_ALPHA8: m_Type = Image::A8; break;
		case GL_ALPHA16: m_Type = Image::A16; break;
		case GL_ALPHA16F_ARB: m_Type = Image::A16f; break;
		case GL_ALPHA32F_ARB: m_Type = Image::A32f; break;
		case GL_RGB8: m_Type = Image::RGB8; break;
		case GL_RGBA8: m_Type = Image::RGBA8; break;
		case GL_RGB16: m_Type = Image::RGB16; break;
		case GL_RGBA16: m_Type = Image::RGBA16; break;
		case GL_RGB16F_ARB: m_Type = Image::RGB16f; break;
		case GL_RGBA16F_ARB: m_Type = Image::RGBA16f; break;
		case GL_RGB32F_ARB: m_Type = Image::RGB32f; break;
		case GL_RGBA32F_ARB: m_Type = Image::RGBA32f; break;
		default: m_Type = Image::Unknown;
	}
	switch (m_OglTarget) {
		case GL_TEXTURE_1D: m_OglBinding = GL_TEXTURE_BINDING_1D; break;
		case GL_TEXTURE_2D: m_OglBinding = GL_TEXTURE_BINDING_2D; break;
		case GL_TEXTURE_3D: m_OglBinding = GL_TEXTURE_BINDING_3D; break;
		case GL_TEXTURE_CUBE_MAP: m_OglBinding = GL_TEXTURE_BINDING_CUBE_MAP; break;
		default: ASSERT(false);
	}
	glGetTexParameteriv(m_OglTarget, GL_TEXTURE_MAX_LEVEL, &tmp);
	m_Levels = tmp+1;
	
	glTexParameteri(m_OglTarget, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(m_OglTarget, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(m_OglTarget, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(m_OglTarget, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(m_OglTarget, GL_TEXTURE_WRAP_R, GL_REPEAT);
	glTexParameterf(m_OglTarget, GL_TEXTURE_MIN_LOD, m_MinLod);
	glTexParameterf(m_OglTarget, GL_TEXTURE_MAX_LOD, m_MaxLod);

	GLint tu;
	glGetIntegerv(GL_ACTIVE_TEXTURE, &tu);
	tu -= GL_TEXTURE0;	
	getContext()->m_Texture[tu] = this;
	getContext()->m_OglTextureTarget[tu] = m_OglTarget;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Texture::~Texture() {
	glDeleteTextures(1, &m_OglTexture);

	// Remove texture from all contexts!
	Context* context = getContext();
	if (context == NULL)
		return;
	for (int i = 0; i < MaxTextures; i++) {
		if (context->m_Texture[i] == this)
			context->m_Texture[i] = NULL;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Texture::forceSelect(unsigned int idx) const {
	ASSERT(idx < MaxTextures);
	Context* c = getContext();
	c->m_Texture[idx] = this;
	c->m_OglTextureTarget[idx] = m_OglTarget;
	glActiveTexture(GL_TEXTURE0 + idx);
	glBindTexture(m_OglTarget, m_OglTexture);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Texture::compileSelect(Assembler& as, unsigned int stage) const {
	as.asmCall((void (Texture::*)(unsigned int ) const)&Texture::forceSelect, this, stage);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Texture::compileDeselect(Assembler& as, unsigned int stage) const {
	Context* context = getContext();
	int offset = (int)((byte*)&context->m_Texture[stage] - (byte*)context);
	as.asmLoad(Assembler::T0, NULL);
	as.asmStore(Assembler::R0, offset, Assembler::T0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Texture::generateMipmaps() {
	if (glGenerateMipmapEXT == NULL)
		return;
	glGenerateMipmapEXT(m_OglTarget);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Texture::setFilter(Filter min, Filter mag) {
	m_MinFilter = min;
	m_MagFilter = mag;
	int oldtex;
	glGetIntegerv(m_OglBinding, &oldtex);
	glBindTexture(m_OglTarget, m_OglTexture);
	glTexParameteri(m_OglTarget, GL_TEXTURE_MIN_FILTER, min);
	glTexParameteri(m_OglTarget, GL_TEXTURE_MAG_FILTER, mag);
	glBindTexture(m_OglTarget, oldtex);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Texture::setWrapping(Wrapping s, Wrapping t, Wrapping u) {
	m_WrapS = s;
	m_WrapT = t;
	m_WrapU = u;
	int oldtex;
	glGetIntegerv(m_OglBinding, &oldtex);
	glBindTexture(m_OglTarget, m_OglTexture);
	glTexParameteri(m_OglTarget, GL_TEXTURE_WRAP_S, s);
	glTexParameteri(m_OglTarget, GL_TEXTURE_WRAP_T, t);
	glTexParameteri(m_OglTarget, GL_TEXTURE_WRAP_R, u);
	glBindTexture(m_OglTarget, oldtex);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Texture::setLod(float mi, float ma) {
	m_MinLod = mi;
	m_MaxLod = ma;
	int oldtex;
	glGetIntegerv(m_OglBinding, &oldtex);
	glBindTexture(m_OglTarget, m_OglTexture);
	glTexParameterf(m_OglTarget, GL_TEXTURE_MIN_LOD, mi);
	glTexParameterf(m_OglTarget, GL_TEXTURE_MAX_LOD, ma);
	glBindTexture(m_OglTarget, oldtex);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Texture::setLodBias(float bias) {
	m_LodBias = bias;
	int oldtex;
	glGetIntegerv(m_OglBinding, &oldtex);
	glBindTexture(m_OglTarget, m_OglTexture);
	glTexParameterf(m_OglTarget, GL_TEXTURE_LOD_BIAS, bias);
	glBindTexture(m_OglTarget, oldtex);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Texture::setupPalette(const Palette<byte>& pal) {
	float mapr[256], mapg[256], mapb[256];
	for (unsigned int i = 0; i < 256; i++) {
		mapr[i] = pal[i].R / 255.0f;
		mapg[i] = pal[i].G / 255.0f;
		mapb[i] = pal[i].B / 255.0f;
	}
	glPixelMapfv(GL_PIXEL_MAP_I_TO_R, 256, mapr);
	glPixelMapfv(GL_PIXEL_MAP_I_TO_G, 256, mapg);
	glPixelMapfv(GL_PIXEL_MAP_I_TO_B, 256, mapb);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int Texture::mapOutputFormat(Type t) {
	switch (t) {
		case Image::BGRA8:
			return GL_RGBA8;
		case Image::BGR8:
			return GL_RGB8;
		case Image::RGBA8:
			return GL_RGBA8;
		case Image::RGB8:
			return GL_RGB8;
		case Image::R5G6B5:
			return GL_RGB5;
		case Image::RGB5:
			return GL_RGB5;
		case Image::RGB16:
			return GL_RGB16;
		case Image::RGBA16:
			return GL_RGBA16;
		case Image::BGR16:
			return GL_RGB16;
		case Image::BGRA16:
			return GL_RGBA16;
		case Image::RGBA16f:
			return GL_RGBA16F_ARB;
		case Image::RGB16f:
			return GL_RGB16F_ARB;
		case Image::RGBA32f:
			return GL_RGBA32F_ARB;
		case Image::RGB32f:
			return GL_RGB32F_ARB;
		case Image::Index8:
			return GL_RGB8;;
		case Image::A8:
			return GL_ALPHA8;
		case Image::A16:
			return GL_ALPHA16;
		case Image::A16f:
			return GL_ALPHA16F_ARB;
		case Image::A32f:
			return GL_ALPHA32F_ARB;
		case Image::DUDV8:
			return GL_DU8DV8_ATI;
		case Image::Depth16:
			return GL_DEPTH_COMPONENT16;
		case Image::Depth24:
			return GL_DEPTH_COMPONENT24;
		default:
			return -1;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int Texture::mapInputFormat(Type t) {
	switch (t) {
		case Image::BGRA8:
			return GL_BGRA_EXT;
		case Image::BGR8:
			return GL_BGR_EXT;
		case Image::RGBA8:
			return GL_RGBA;
		case Image::RGB8:
			return GL_RGB;
		case Image::R5G6B5:
			return GL_RGB;
		case Image::RGB5:
			return GL_RGB;
		case Image::RGB16:
			return GL_RGB;
		case Image::RGBA16:
			return GL_RGBA;
		case Image::RGB16f:
			return GL_RGB;
		case Image::RGBA16f:
			return GL_RGBA;
		case Image::RGB32f:
			return GL_RGB;
		case Image::RGBA32f:
			return GL_RGBA;
		case Image::BGR16:
			return GL_BGR_EXT;
		case Image::BGRA16:
			return GL_BGRA_EXT;
		case Image::Index8:
			return GL_COLOR_INDEX;
		case Image::DUDV8:
			return GL_DUDV_ATI;
		case Image::A8:
			return GL_ALPHA;
		case Image::A16:
			return GL_ALPHA;
		case Image::A16f:
			return GL_ALPHA;
		case Image::A32f:
			return GL_ALPHA;
		case Image::Depth16:
			return GL_DEPTH_COMPONENT;
		case Image::Depth24:
			return GL_DEPTH_COMPONENT;
		default:
			return -1;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int Texture::mapInputType(Type t) {
	switch (t) {
		case Image::BGRA8:
			return GL_UNSIGNED_BYTE;
		case Image::BGR8:
			return GL_UNSIGNED_BYTE;
		case Image::RGBA8:
			return GL_UNSIGNED_BYTE;
		case Image::RGB8:
			return GL_UNSIGNED_BYTE;
		case Image::R5G6B5:
			return GL_UNSIGNED_SHORT_5_6_5;
		case Image::RGB5:
			return GL_UNSIGNED_SHORT_5_5_5_1;
		case Image::RGB16:
			return GL_UNSIGNED_SHORT;
		case Image::RGBA16:
			return GL_UNSIGNED_SHORT;
		case Image::BGR16:
			return GL_UNSIGNED_SHORT;
		case Image::BGRA16:
			return GL_UNSIGNED_SHORT;
		case Image::RGB16f:
			return GL_FLOAT;
		case Image::RGBA16f:
			return GL_FLOAT;
		case Image::RGB32f:
			return GL_FLOAT;
		case Image::RGBA32f:
			return GL_FLOAT;
		case Image::Index8:
			return GL_UNSIGNED_BYTE;
		case Image::DUDV8:
			return GL_BYTE;
		case Image::A8:
			return GL_UNSIGNED_BYTE;
		case Image::A16:
			return GL_UNSIGNED_SHORT;
		case Image::Depth16:
			return GL_FLOAT;
		case Image::Depth24:
			return GL_FLOAT;
		default:
			return -1;
	}
}

}; };
