/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics.filter.resize.h"
#include "graphics3d.texture1d.h"
#include "graphics3d.exception.h"
#include "graphics3d.context.h"

#include "io.filter.inl"
#include "container.chain.inl"
#include "graphics.image.inl"
#include "graphics.filter.resize.inl"
#include "graphics3d.texture1d.inl"
#include "graphics3d.context.inl"


namespace magnum { namespace graphics3d {
using magnum::graphics::FilterResize;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline static int roundUp2n(int in) {
	int out = 1;
	while (in > 1) {
		if (in & 1)
			in++;
		in >>= 1;
		out <<= 1;
	}

	return out;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
static unsigned int adjustSize(const unsigned int& in) {
	unsigned int size;

	size = roundUp2n(in);

	// Check for min/max size etc

	return size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Texture1D* Texture1D::create(Type t,unsigned int d, int levels) {
	int level = 0;
	unsigned int size;

	int oformat = mapOutputFormat(t);
	if (oformat == -1)
		throw TypeException();

	size = adjustSize(d);

	unsigned int tex;
	glGetError();

	glGenTextures(1, &tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();
		
	glBindTexture(GL_TEXTURE_1D, tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	if (levels == GenerateAll)
		glTexParameteri(GL_TEXTURE_1D, GL_GENERATE_MIPMAP, TRUE);

	glTexImage1D(GL_TEXTURE_1D, level, oformat, size, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	Texture1D* texture = new Texture1D(tex);

	return texture;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Texture1D* Texture1D::create(const Image* img, int levels) {
	// Map image format to OpenGL formats
	int type = mapInputType(img->getType());
	int iformat = mapInputFormat(img->getType());
	int oformat = mapOutputFormat(img->getType());
	if ((oformat == -1) || (iformat == -1) || (type == -1))
		throw TypeException();
	
	// Setup transfer mode
	if (img->getType() == Image::Index8) {
		glPixelTransferi(GL_MAP_COLOR, true);
		glPixelTransferi(GL_MAP_STENCIL, false);
		glPixelTransferi(GL_INDEX_SHIFT, 0);
		glPixelTransferi(GL_INDEX_OFFSET, 0);
	} else {
		glPixelTransferi(GL_MAP_COLOR, false);
		glPixelTransferi(GL_MAP_STENCIL, false);
	}

	glPixelStorei(GL_UNPACK_SWAP_BYTES , false);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); 

	int size = adjustSize(img->getWidth());

	unsigned int tex;
	glGetError();

	glGenTextures(1, &tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();
		
	glBindTexture(GL_TEXTURE_1D, tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	if (levels == GenerateAll) {
		glTexParameteri(GL_TEXTURE_1D, GL_GENERATE_MIPMAP, TRUE);
		levels = 1;
	} else if (levels == CopyAll) {
		glTexParameteri(GL_TEXTURE_1D, GL_GENERATE_MIPMAP, FALSE);
		levels = 256;
	}

	Image* root;

	// Check if we need to resize the image
	if ((size != img->getWidth()) || (img->getHeight() != 1)) {
		FilterResize filter(Dimension<int>(size,1));
		root = filter << img;
	} else {
		root = (Image *)img;
		root->refer();
	}
	img = root;

	int level = 0;
	while ((img != NULL) && (level < levels)) {
		// Transfer image into texture memory
		byte* src = (byte *)((Image *)img)->lock();
		glTexImage1D(GL_TEXTURE_1D, level, oformat, img->getWidth(), 0, iformat, type, src);
		((Image *)img)->unlock();
		img = img->getNext();
		level++;
	}

	// Release any reference to original image
	root->release();

	// Some final error checking
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	Texture1D* texture = new Texture1D(tex);

	return texture;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Texture1D::Texture1D(unsigned int tex) : Texture(tex, GL_TEXTURE_1D) {
	int tmp;

	m_Locked = false;
	glBindTexture(GL_TEXTURE_1D, tex);
	glGetTexLevelParameteriv(GL_TEXTURE_1D, 0, GL_TEXTURE_WIDTH, &tmp);
	m_Size = tmp;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void* Texture1D::lock(int level, unsigned int , LockFlag ) {
	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Texture1D::unlock() {

}


}; };
