/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "system.thread.h"
#include "graphics.filter.resize.h"
#include "graphics3d.texture2d.h"
#include "graphics3d.exception.h"
#include "graphics3d.context.h"
#include "graphics3d.vertexformat.h"

#include "object.inl"
#include "base.string.inl"
#include "object.exception.inl"
#include "system.thread.inl"
#include "io.filter.inl"
#include "container.chain.inl"
#include "graphics.image.inl"
#include "graphics.dimension.inl"
#include "graphics.rectangle.inl"
#include "graphics.filter.resize.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.texture2d.inl"
#include "graphics3d.context.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.vertexformat.inl"


namespace magnum { namespace graphics3d {
using magnum::graphics::FilterResize;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline static int roundUp2n(int in) {
	int out = 1;
	while (in > 1) {
		if (in & 1)
			in++;
		in >>= 1;
		out <<= 1;
	}

	return out;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
static Dimension<int> adjustSize(const Dimension<int>& in) {
	Dimension<int> size;

	size.setWidth(roundUp2n(in.getWidth()));
	size.setHeight(roundUp2n(in.getHeight()));

	// Check for min/max size etc

	return size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Texture2D* Texture2D::create(Type t,const Dimension<int>& d, int levels) {
	int level = 0;
	
	// Map image format to OpenGL formats
	int type = mapInputType(t);
	int iformat = mapInputFormat(t);
	int oformat = mapOutputFormat(t);
	if ((oformat == -1) || (iformat == -1) || (type == -1))
		throw TypeException();

	Dimension<int> size = adjustSize(d);

	unsigned int tex;
	glGetError();

	glGenTextures(1, &tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();
		
	glBindTexture(GL_TEXTURE_2D, tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	if (levels == GenerateAll)
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, TRUE);
	else
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, FALSE);

	glTexImage2D(GL_TEXTURE_2D, level, oformat, size.getWidth(), size.getHeight(), 0, iformat, type, NULL);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	Texture2D* texture = new Texture2D(tex);

	return texture;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Texture2D* Texture2D::create(Type t,unsigned int w, unsigned int h, int levels) {
	return create(t, Dimension<int>(w,h), levels);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Texture2D* Texture2D::create(const Image* img, int levels) {
	// Map image format to OpenGL formats
	int type = mapInputType(img->getType());
	int iformat = mapInputFormat(img->getType());
	int oformat = mapOutputFormat(img->getType());
	if ((oformat == -1) || (iformat == -1) || (type == -1))
		throw TypeException();
	
	// Setup transfer mode
	if (img->getType() == Image::Index8) {
		glPixelTransferi(GL_MAP_COLOR, true);
		glPixelTransferi(GL_MAP_STENCIL, false);
		glPixelTransferi(GL_INDEX_SHIFT, 0);
		glPixelTransferi(GL_INDEX_OFFSET, 0);
		
		Palette<byte> pal;
		img->getPalette(&pal);
		setupPalette(pal);
	} else {
		glPixelTransferi(GL_MAP_COLOR, false);
		glPixelTransferi(GL_MAP_STENCIL, false);
	}

	glPixelStorei(GL_UNPACK_SWAP_BYTES , false);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); 

	Dimension<int> size = adjustSize(img->getSize());

	unsigned int tex;
	glGetError();

	glGenTextures(1, &tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();
		
	glBindTexture(GL_TEXTURE_2D, tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	if (levels == GenerateAll) {
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, TRUE);
		levels = 1;
	} else if (levels == CopyAll) {
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, FALSE);
		levels = 256;
	}

	Image* root;

	// Check if we need to resize the image
	if (size != img->getSize()) {
		FilterResize filter(size);
		root = filter << img;
	} else {
		root = const_cast<Image*>(img);
		root->refer();
	}
	img = root;

	int level = 0;
	while ((img != NULL) && (level < levels)) {
		int ls = img->getLinesize();

		// Transfer image into texture memory
		byte* src = (byte *)((Image *)img)->lock();
		glTexImage2D(GL_TEXTURE_2D, level, oformat, img->getWidth(), img->getHeight(), 0, iformat, type, src);
		((Image *)img)->unlock();

		img = img->getNext();
		level++;
	}

	// Release any reference to original image
	root->release();

	// Some final error checking
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	Texture2D* texture = new Texture2D(tex);

	return texture;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Texture2D::Texture2D(unsigned int tex) : Texture(tex, GL_TEXTURE_2D) {
	int tmp;

	m_Locked = false;
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &tmp);
	m_Size.setWidth(tmp);
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &tmp);
	m_Size.setHeight(tmp);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Texture2D::update(const Image* img, int level) {
	ASSERT(img != NULL);
	
	// Map image format to OpenGL formats
	int type = mapInputType(img->getType());
	int iformat = mapInputFormat(img->getType());
	if ((iformat == -1) || (type == -1))
		throw TypeException();
	
	// Setup transfer mode
	if (img->getType() == Image::Index8) {
		glPixelTransferi(GL_MAP_COLOR, true);
		glPixelTransferi(GL_MAP_STENCIL, false);
		glPixelTransferi(GL_INDEX_SHIFT, 0);
		glPixelTransferi(GL_INDEX_OFFSET, 0);

		Palette<byte> pal;
		img->getPalette(&pal);
		setupPalette(pal);
	} else {
		glPixelTransferi(GL_MAP_COLOR, false);
		glPixelTransferi(GL_MAP_STENCIL, false);
	}

	glPixelStorei(GL_UNPACK_SWAP_BYTES , false);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, img->getWidth());
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); 

	int oldtex;
	glGetIntegerv(GL_TEXTURE_BINDING_2D, &oldtex);
	glBindTexture(GL_TEXTURE_2D, m_OglTexture);

	// Transfer image into texture memory
	byte* src = (byte *)((Image *)img)->lock();
	glTexSubImage2D(GL_TEXTURE_2D, level, 0, 0, img->getWidth(), img->getHeight(), iformat, type, src);
	((Image *)img)->unlock();

	glBindTexture(GL_TEXTURE_2D, oldtex);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Texture2D::copyFrom(const Image* img, const Point<int>& p,const Rectangle<int>& r) {
	ASSERT(img != NULL);
	
	// Map image format to OpenGL formats
	int iformat = mapInputFormat(img->getType());
	int type = mapInputType(img->getType());
	if ((iformat == -1) || (type == -1))
		throw TypeException();
	
	// Setup transfer mode
	if (img->getType() == Image::Index8) {
		glPixelTransferi(GL_MAP_COLOR, true);
		glPixelTransferi(GL_MAP_STENCIL, false);
		glPixelTransferi(GL_INDEX_SHIFT, 0);
		glPixelTransferi(GL_INDEX_OFFSET, 0);

		Palette<byte> pal;
		img->getPalette(&pal);
		setupPalette(pal);
	} else {
		glPixelTransferi(GL_MAP_COLOR, false);
		glPixelTransferi(GL_MAP_STENCIL, false);
	}

	glPixelStorei(GL_UNPACK_SWAP_BYTES , false);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, img->getWidth());
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, r.getX());
	glPixelStorei(GL_UNPACK_SKIP_ROWS, r.getY());
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); 

	int oldtex;
	glGetIntegerv(GL_TEXTURE_BINDING_2D, &oldtex);
	glBindTexture(GL_TEXTURE_2D, m_OglTexture);

	int level = 0;

	// Transfer image into texture memory
	byte* src = (byte *)((Image *)img)->lock();
	glTexSubImage2D(GL_TEXTURE_2D, level, p.getX(), p.getY(), r.getWidth(), r.getHeight(), iformat, type, src);
	((Image *)img)->unlock();

	glBindTexture(GL_TEXTURE_2D, oldtex);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Texture2D::copyFrom(const Image* img, const Point<int>& p) {
	Rectangle<int> rect;
	rect.setOrigin(0,0);
	rect.setSize(img->getSize());
	return copyFrom(img, p, rect);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void* Texture2D::lock(int level, const Rectangle<int>& , LockFlag ) {
	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Texture2D::unlock() {

}


}; };
