/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "graphics3d.exception.h"
#include "graphics3d.context.h"
#include "graphics3d.texture3d.h"

#include "object.inl"
#include "base.string.inl"
#include "object.exception.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.texture3d.inl"
#include "graphics3d.context.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.texture3d.inl"


namespace magnum { namespace graphics3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline static unsigned int roundUp2n(unsigned int in) {
	unsigned int out = 1;
	while (in > 1) {
		if (in & 1)
			in++;
		in >>= 1;
		out <<= 1;
	}

	return out;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
static void adjustSize(unsigned int& w,unsigned int &h, unsigned int&d) {
	w = roundUp2n(w);
	h = roundUp2n(h);
	d = roundUp2n(d);

	// Check for min/max size etc
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
static unsigned int calcDepth(const Image* img) {
	ASSERT(img != NULL);
	Dimension<int> size = img->getSize();
	unsigned int depth = 0;
	while(img != NULL) {
		if (img->getSize() != size)
			return depth;
		depth++;
		img = img->getNext();
	}
	return depth;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Texture3D* Texture3D::create(const Image* img, int levels) {
	// Map image format to OpenGL formats
	int type = mapInputType(img->getType());
	int iformat = mapInputFormat(img->getType());
	int oformat = mapOutputFormat(img->getType());
	if ((oformat == -1) || (iformat == -1) || (type == -1))
		throw TypeException();
	
	// Setup transfer mode
	if (img->getType() == Image::Index8) {
		glPixelTransferi(GL_MAP_COLOR, true);
		glPixelTransferi(GL_MAP_STENCIL, false);
		glPixelTransferi(GL_INDEX_SHIFT, 0);
		glPixelTransferi(GL_INDEX_OFFSET, 0);
		
		Palette<byte> pal;
		img->getPalette(&pal);
		setupPalette(pal);
	} else {
		glPixelTransferi(GL_MAP_COLOR, false);
		glPixelTransferi(GL_MAP_STENCIL, false);
	}

	glPixelStorei(GL_UNPACK_SWAP_BYTES , false);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); 

	// get texture dimension
	unsigned int width = img->getWidth();
	unsigned int height = img->getHeight();
	unsigned int depth = calcDepth(img);
	adjustSize(width, height, depth);

	// Create new texture object
	unsigned int tex;
	glGetError();
	glGenTextures(1, &tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();
		
	glBindTexture(GL_TEXTURE_3D, tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	if (levels == GenerateAll) {
		glTexParameteri(GL_TEXTURE_3D, GL_GENERATE_MIPMAP, TRUE);
		levels = 1;
	} else if (levels == CopyAll) {
		glTexParameteri(GL_TEXTURE_3D, GL_GENERATE_MIPMAP, FALSE);
		levels = 256;
	}

	// Create texture	
	glTexImage3D(GL_TEXTURE_3D, 0, oformat, width, height, depth, 0, iformat, type, NULL);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	// Copy all texture maps
	for (unsigned int i = 0; i < depth; i++) {
		ASSERT(img != NULL);
		const void* data = const_cast<Image*>(img)->lock();
		glTexSubImage3D(GL_TEXTURE_3D, 0,0,0,i,	img->getWidth(), img->getHeight(), 1, iformat, type, data);
		const_cast<Image*>(img)->unlock();
		img = img->getNext();
	}

	return new Texture3D(tex);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Texture3D* Texture3D::create(Type t,unsigned int width, unsigned int height, unsigned int depth) {
	int level = 0;
	
	// Map image format to OpenGL formats
	int type = mapInputType(t);
	int iformat = mapInputFormat(t);
	int oformat = mapOutputFormat(t);
	if ((oformat == -1) || (iformat == -1) || (type == -1))
		throw TypeException();

	adjustSize(width,height,depth);

	unsigned int tex;
	glGetError();

	glGenTextures(1, &tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();
		
	glBindTexture(GL_TEXTURE_3D, tex);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	glTexImage3D(GL_TEXTURE_3D, level, oformat, width, height, depth, 0, iformat, type, NULL);
	if(glGetError() != GL_NO_ERROR)
		throw TextureException();

	return new Texture3D(tex);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Texture3D::Texture3D(unsigned int tex) : Texture(tex, GL_TEXTURE_3D) {
	int tmp;

	m_Locked = false;
	glGetTexLevelParameteriv(GL_TEXTURE_3D, 0, GL_TEXTURE_WIDTH, &tmp);
	m_Size.setWidth(tmp);
	glGetTexLevelParameteriv(GL_TEXTURE_3D, 0, GL_TEXTURE_HEIGHT, &tmp);
	m_Size.setHeight(tmp);
	glGetTexLevelParameteriv(GL_TEXTURE_3D, 0, GL_TEXTURE_DEPTH, &tmp);
	m_Depth = tmp;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Texture3D::update(const Image* img, int level) {
	ASSERT(img != NULL);
	
	// Map image format to OpenGL formats
	int type = mapInputType(img->getType());
	int iformat = mapInputFormat(img->getType());
	if ((iformat == -1) || (type == -1))
		throw TypeException();
	
	// Setup transfer mode
	if (img->getType() == Image::Index8) {
		glPixelTransferi(GL_MAP_COLOR, true);
		glPixelTransferi(GL_MAP_STENCIL, false);
		glPixelTransferi(GL_INDEX_SHIFT, 0);
		glPixelTransferi(GL_INDEX_OFFSET, 0);

		Palette<byte> pal;
		img->getPalette(&pal);
		setupPalette(pal);
	} else {
		glPixelTransferi(GL_MAP_COLOR, false);
		glPixelTransferi(GL_MAP_STENCIL, false);
	}

	glPixelStorei(GL_UNPACK_SWAP_BYTES , false);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, img->getWidth());
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); 

	int oldtex;
	glGetIntegerv(GL_TEXTURE_BINDING_3D, &oldtex);
	glBindTexture(GL_TEXTURE_3D, m_OglTexture);

	// Transfer image into texture memory
	byte* src = (byte *)((Image *)img)->lock();
	glTexSubImage3D(GL_TEXTURE_3D, 0, 0, 0, level, img->getWidth(), img->getHeight(), 1, iformat, type, src);
	((Image *)img)->unlock();

	glBindTexture(GL_TEXTURE_3D, oldtex);

	return true;
}


}; };
