/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "io.exception.h"
#include "graphics3d.translator.ifs.h"
#include "graphics3d.staticmodel.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.context.h"
#include "graphics3d.appearance.h"
#include "graphics3d.plainshader.h"
#include "graphics3d.colourmaterial.h"
#include "graphics3d.texturemanager.h"
#include "graphics3d.shadermanager.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "base.string.inl"
#include "container.array.inl"
#include "io.inputstream.inl"
#include "io.file.inl"
#include "io.translator.inl"
#include "math.vector3.inl"
#include "graphics3d.staticmodel.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.context.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.colourmaterial.inl"

#define MCHAR2(a, b) (a | (b << 8))
#define MCHAR4(a, b, c, d) (a | (b << 8) | (c << 16) | (d << 24))


namespace magnum { namespace graphics3d {
using magnum::object::AutoPointer;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorIFS::TranslatorIFS() {
	m_VertexCount = 0;
	m_FaceCount = 0;
	
	m_Input = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
String TranslatorIFS::readString() {
	dword len;
	m_Input->read(&len, sizeof(len));

	char buffer[100];
		
	if (m_Input->read(buffer, len) != len)
		throw magnum::io::TranslatorFormatException();
	buffer[len] = 0;

	return String(buffer);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorIFS::readHeader() {
	// Read header
	if (readString() != "IFS")
		throw magnum::io::TranslatorFormatException();
	float version;
	m_Input->read(&version, sizeof(version));
	if (version != 1.0f)
		throw magnum::io::TranslatorFormatException();

	// Read in name
	m_Name = readString();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorIFS::readVertices() {
	// Read vertex header
	String str = readString();
	if(str != "VERTICES")
		throw magnum::io::TranslatorFormatException();

	// Read vertex buffer
	m_Input->read(&m_VertexCount, sizeof(m_VertexCount));
	m_Vertices.setSize(m_VertexCount);
	m_Input->read(m_Vertices.getArray(), m_VertexCount*sizeof(Vector3<float>));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorIFS::readIndices() {
	// Read index header
	String str = readString();
	if(str != "TRIANGLES")
		throw magnum::io::TranslatorFormatException();

	m_Input->read(&m_FaceCount, sizeof(m_FaceCount));
	m_Indices.setSize(m_FaceCount*3);
	m_Input->read(m_Indices.getArray(), 3*m_FaceCount*sizeof(unsigned int));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorIFS::createNormals() {
	// Create normals
	m_Normals.setSize(m_VertexCount);
	unsigned int i;
	for(i = 0; i < m_VertexCount; i++) 
		m_Normals[i].null();
		
	for (i = 0; i < m_FaceCount; i++) {
		int vi1,vi2,vi3;
		vi1 = m_Indices[3*i];
		vi2 = m_Indices[3*i+1];
		vi3 = m_Indices[3*i+2];

		Vector3<float> v1 = m_Vertices[vi1];
		Vector3<float> v2 = m_Vertices[vi2];
		Vector3<float> v3 = m_Vertices[vi3];
		Vector3<float> normal = (v2-v1) % (v3-v1);

		m_Normals[vi1] += normal;
		m_Normals[vi2] += normal;
		m_Normals[vi3] += normal;
	}
	for (i = 0; i < m_VertexCount; i++) {
		float l = m_Normals[i].length();
		if (l > 0.0f)
			m_Normals[i] /= l;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Model* TranslatorIFS::onDecode(InputStream* in,size_t size) {
	m_Input = in;
	
	readHeader();
	readVertices();
	readIndices();
	createNormals();
	
	// Create vertex and index buffers
	VertexFormat fmt;
	fmt.set(0, Types::Float3, Attributes::Vertex);
	fmt.set(1, Types::Float3, Attributes::Normal);
	AutoPointer<VertexBuffer> vb = VertexBuffer::create(fmt, m_VertexCount, VertexBuffer::Static);
	Vector3<float>* vbmem = (Vector3<float>*)vb->lock(VertexBuffer::Discard);
	for (unsigned int i = 0; i < m_VertexCount; i++) {
		vbmem[2*i] = 300.0f*m_Vertices[i];
		vbmem[2*i + 1] = m_Normals[i];
	}
	vb->unlock();
	
	AutoPointer<IndexBuffer> ib = IndexBuffer::create(IndexBuffer::Index32, m_FaceCount*3, IndexBuffer::Static);
	ib->setBuffer(0, m_Indices.getArray(), m_FaceCount*3);

	// Release memory
	m_Vertices.clear();
	m_Indices.clear();
	m_Normals.clear();

	// Create appearance for mesh
	AutoPointer<Appearance> appearance = new Appearance();
	AutoPointer<Shader> shader = new PlainShader();
	appearance->attachShader(shader);
	
	// Create material
	AutoPointer<ColourMaterial> material = new ColourMaterial();
	material->setAmbient(Colour<float>(0.2f, 0.2f, 0.2f));
	material->setDiffuse(Colour<float>(0.7f, 0.7f, 0.5f));
	material->setSpecular(Colour<float>(1.0f, 1.0f, 1.0f));
	material->setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
	material->setShininess(20.0f);
	appearance->attachMaterial(material);

	// Create new mesh
	AutoPointer<IndexMesh> mesh = new IndexMesh();
	mesh->setPrimitive(Mesh::TriangleList);
	mesh->attachAppearance(appearance);
	mesh->attachIndexBuffer(ib);
	mesh->attachVertexBuffer(vb,0);
	mesh->updateBounds();

	// Create model
	StaticModel* model = new StaticModel();
	model->setName(m_Name);
	model->insertMesh(mesh);
	model->finish();
	
	return model;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorIFS::onEncode(OutputStream* ,const Model* ) {
	ASSERT(false);
	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorIFS::onIdentify(File* in) {
	dword id;

	try {
		int64 oldpos;
		oldpos = in->getPosition();	
		in->read(&id, sizeof(id));
		in->read(&id, sizeof(id));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if ((dword)id != MCHAR4('I','F','S',0))
		return false;

	return true;
}

}; };
