/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "container.multimap.h"
#include "io.exception.h"
#include "graphics3d.context.h"
#include "graphics3d.appearance.h"
#include "graphics3d.colourmaterial.h"
#include "graphics3d.texture.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.plainshader.h"
#include "graphics3d.textureshader.h"
#include "graphics3d.texturemanager.h"
#include "graphics3d.shadermanager.h"
#include "graphics3d.exception.h"
#include "graphics3d.opengl.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.vertexmesh.h"
#include "graphics3d.staticmodel.h"
#include "graphics3d.translator.lwo.h"

#include "object.inl"
#include "object.pointer.inl"
#include "base.string.inl"
#include "io.file.inl"
#include "io.translator.inl"
#include "container.iterator.inl"
#include "container.allocator.inl"
#include "container.map.inl"
#include "container.multimap.inl"
#include "container.array.inl"
#include "container.cache.inl"
#include "math.vector3.inl"
#include "graphics.colour.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.context.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.translator.inl"
#include "graphics3d.translator.lwo.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.model.inl"
#include "graphics3d.staticmodel.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.material.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.texturemanager.inl"
#include "graphics3d.shadermanager.inl"


namespace magnum { namespace graphics3d {
using magnum::types::msb2dword;
using magnum::types::msb2word;
using magnum::container::Multimap;

#define LWO_ID(a,b,c,d) (((a) << 24) | ((b) << 16) | ((c) << 8) | (d))
#define LWO_ID_FORM		LWO_ID('F','O','R','M')
#define LWO_ID_LWO2		LWO_ID('L','W','O','2')
#define LWO_ID_LAYR		LWO_ID('L','A','Y','R')
#define LWO_ID_PNTS		LWO_ID('P','N','T','S')
#define LWO_ID_VMAP		LWO_ID('V','M','A','P')
#define LWO_ID_TXUV		LWO_ID('T','X','U','V')
#define LWO_ID_VMAD		LWO_ID('V','M','A','D')
#define LWO_ID_POLS		LWO_ID('P','O','L','S')
#define LWO_ID_PTAG		LWO_ID('P','T','A','G')
#define LWO_ID_FACE		LWO_ID('F','A','C','E')
#define LWO_ID_SURF		LWO_ID('S','U','R','F')
#define LWO_ID_BLOK		LWO_ID('B','L','O','K')
#define LWO_ID_SMAN		LWO_ID('S','M','A','N')
#define LWO_ID_TAGS		LWO_ID('T','A','G','S')
#define LWO_ID_BBOX		LWO_ID('B','B','O','X')
#define LWO_ID_CLIP		LWO_ID('C','L','I','P')


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline int TranslatorLWO::readShort() {
	word ret;
	if (m_In->read(&ret,2) != 2)
			throw magnum::io::TranslatorFormatException();
	return msb2word(ret);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline int TranslatorLWO::readInt() {
	dword ret;
	if (m_In->read(&ret,4) != 4)
		throw magnum::io::TranslatorFormatException();
	return msb2dword(ret);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline float TranslatorLWO::readFloat() {
	dword ret;
	if (m_In->read(&ret,4) != 4)
		throw magnum::io::TranslatorFormatException();
	ret = msb2dword(ret);
	return *(float*)(void*)&ret;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline int TranslatorLWO::readString(String& str) {
	str.clear();
	int len = 0;
	for(;;) {
		char c; 
		m_In->read(&c,1);
		len++;
		if(c == '\0') 
			break;
		str.append(c);
	}
	if (len % 2) {
		char c;
		m_In->read(&c,1);
		len++;
	}
	return len;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorLWO::skipChunk(int size) {
	m_In->skip(size + size % 2);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorLWO::read_LAYR(int size) {
	m_CurrentLayer = &m_Layers[m_Layers.append()];
	
	readShort();				// number
	readShort();				// flags
	readFloat();				// pivot
	readFloat();
	readFloat();
	size -= 2 + 2 + 4 * 3;
	size -= readString(m_CurrentLayer->m_Name);
	if(size > 0) 
		readShort();			// parent
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorLWO::read_PNTS(int size) {
	if (m_CurrentLayer == NULL) {
		skipChunk(size);
		return;
	}

	m_CurrentLayer->m_Vertices.setSize(size / 12);
	Vector3<float>* vertices = &m_CurrentLayer->m_Vertices[0];
	for(int i = 0; i < size / 12; i++) {
		vertices[i].x = readFloat();
		vertices[i].y = readFloat();
		vertices[i].z = readFloat();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorLWO::read_VMAP(int size) {
	if (m_CurrentLayer == NULL) {
		skipChunk(size);
		return;
	}
	
	int chunk = readInt();
	int dimension = readShort();
	size -= 6;
	
	String name;
	size -= readString(name);
	
	// vertex texture coordinates
	if(chunk == LWO_ID_TXUV && dimension == 2) {
		VTexCoord* vtexcoord = & * m_CurrentLayer->m_VTexCoords.insert(name);
		
		int num_texcoords = size / 10;
		int num_vertices = m_CurrentLayer->m_Vertices.getSize();
		vtexcoord->m_Coords.setSize(num_vertices);
		Vector2<float>* texcoords = &vtexcoord->m_Coords[0];
		memset(texcoords,0,sizeof(Vector2<float>) *num_vertices);
		
		for(int i = 0; i < num_texcoords; i++) {
			int vertex = readShort();
			float u = readFloat();
			float v = readFloat();
			if(vertex < num_vertices) {
				texcoords[vertex].x = u;
				texcoords[vertex].y = v;
			}
		}
	} else {
		skipChunk(size);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorLWO::read_POLS(int size) {
	if (m_CurrentLayer == NULL) {
		skipChunk(size);
		return;
	}
	Layer *layer = m_CurrentLayer;
	
	int chunk = readInt();
	size -= 4;
	if(chunk != LWO_ID_FACE) {
		skipChunk(size);
		return;
	}
	
	// load m_Polygons
	for(int s = size; s > 0; ) {
		Polygon* p = &layer->m_Polygons[layer->m_Polygons.append()];
		memset(p, 0, sizeof(Polygon));
		p->m_Size = readShort() & 0x03ff;
		ASSERT(p->m_Size <= NUM_VERTEX);
		s -= 2;
		for(unsigned int i = 0; i < p->m_Size; i++) {
			p->m_Vertex[i] = readShort();
		}

		// calculate m_Polygons normal
		p->m_Normal = ((layer->m_Vertices[p->m_Vertex[1]] - layer->m_Vertices[p->m_Vertex[0]])
			% (layer->m_Vertices[p->m_Vertex[2]] - layer->m_Vertices[p->m_Vertex[0]])).normalise();
		
		// Adjust remaining size
		s -= p->m_Size * 2;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorLWO::read_PTAG(int size) {
	if (m_CurrentLayer == NULL) {
		skipChunk(size);
		return;
	}
	Layer *layer = m_CurrentLayer;
	
	int chunk = readInt();
	size -= 4;
	
	// polygonal surface
	if(chunk == LWO_ID_SURF) {
		int num_m_Polygons = layer->m_Polygons.getSize();
		layer->m_PolygonTags.setSize(num_m_Polygons);
		int* polytags = &layer->m_PolygonTags[0];
		
		for(int i = 0; i < num_m_Polygons; i++) {
			polytags[i] = -1;
		}
		for(int i = 0; i < size / 4; i++) {
			int polygon = readShort();
			int tag = readShort();
			if(polygon < num_m_Polygons) {
				polytags[polygon] = tag;
			}
		}
	} else {
		skipChunk(size);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorLWO::read_VMAD(int size) {
	if (m_CurrentLayer == NULL) {
		skipChunk(size);
		return;
	}
	Layer *layer = m_CurrentLayer;
	
	int chunk = readInt();
	int dimension = readShort();
	size -= 6;
	
	String name;
	size -= readString(name);
	
	// polygon texture coordinates
	if(chunk == LWO_ID_TXUV && dimension == 2) {
		PTexCoord* ptexcoord = & * m_CurrentLayer->m_PTexCoords.insert(name);

		int num_texcoords = size / 12;
		int num_vertices = m_CurrentLayer->m_Vertices.getSize();

		ptexcoord->m_Polygons.setSize(num_texcoords);
		ptexcoord->m_Vertices.setSize(num_texcoords);
		ptexcoord->m_Coords.setSize(num_texcoords);
		for(int i = 0; i < num_texcoords; i++) {
			ptexcoord->m_Vertices[i] = readShort();
			ptexcoord->m_Polygons[i] = readShort();
			ptexcoord->m_Coords[i].x = readFloat();
			ptexcoord->m_Coords[i].y = readFloat();
		}
	} else {
		skipChunk(size);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorLWO::read_BLOK(int size) {
	while(size > 0) {
		int sub_chunk = readInt();
		int sub_size = readShort();
		size -= 6;
		switch(sub_chunk) {
			case LWO_ID_VMAP:
				size -= readString(m_CurrentSurface->m_Texture);
				break;
			default:
				skipChunk(sub_size);
				size -= sub_size + sub_size % 2;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorLWO::read_SURF(int size) {
	if (m_CurrentLayer == NULL) {
		skipChunk(size);
		return;
	}
	Layer *layer = m_CurrentLayer;
	
	m_CurrentSurface = &layer->m_Surfaces[layer->m_Surfaces.append()];
	m_CurrentSurface->m_Smoothing = -1.0f;
	m_CurrentSurface->m_Appearance = new Appearance();
	m_CurrentSurface->m_Appearance->attachMaterial(m_Material);
	m_CurrentSurface->m_Appearance->attachShader(m_PlainShader);

	size -= readString(m_CurrentSurface->m_Name);
	size -= readString(m_CurrentSurface->m_Source);
	
	// read mapping and smooth angle
	while(size > 0) {
		int sub_chunk = readInt();
		int sub_size = readShort();
		size -= 6;
		switch (sub_chunk) {
			case LWO_ID_BLOK:
				read_BLOK(sub_size);
				size -= sub_size + sub_size % 2;
				break;
			case LWO_ID_SMAN:
				m_CurrentSurface->m_Smoothing = readFloat();
				size -= 4;
				break;
			default:
				skipChunk(sub_size);
				size -= sub_size + sub_size % 2;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorLWO::read_LWO2(int size) {
	while(size > 0) {
		int sub_chunk = readInt();
		int sub_size = readInt();
		size -= 8;
		
		switch (sub_chunk) {
			case LWO_ID_LAYR:
				read_LAYR(sub_size);
				break;
			case LWO_ID_PNTS: 
				read_PNTS(sub_size);
				break;
			case LWO_ID_VMAP: 
				read_VMAP(sub_size);
				break;
			case LWO_ID_POLS:	
				read_POLS(sub_size);
				break;
			case LWO_ID_PTAG: 
				read_PTAG(sub_size);
				break;
			case LWO_ID_VMAD: 
				read_VMAD(sub_size);
				break;
			case LWO_ID_SURF: 
				read_SURF(sub_size);
				break;
			case LWO_ID_BBOX: 
				skipChunk(sub_size);
				break;
			case LWO_ID_CLIP: 
				skipChunk(sub_size);
				break;
			case LWO_ID_TAGS: 
				skipChunk(sub_size);
				break;
			default:
				skipChunk(sub_size);
				break;
		}
		
		size -= sub_size + sub_size % 2;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorLWO::createSurfaces(TranslatorLWO::Layer* layer) {
	if(layer->m_PolygonTags.isEmpty())
		return;
		
	// m_Polygons
	unsigned int num_polys = layer->m_Polygons.getSize();
	unsigned int current_surface = layer->m_Surfaces.getSize();
	for(unsigned int i = 0; i < num_polys; i++) {
		unsigned int surface = layer->m_PolygonTags[i];
		if (surface >= layer->m_Surfaces.getSize())
			continue;
		layer->m_Surfaces[surface].m_Polygons.insert(i);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorLWO::applyMapping(Layer *layer,Surface *surface) {
	// vertex texture coordinates
	Map<String,VTexCoord>::Iterator vtiter = layer->m_VTexCoords.at(surface->m_Name);
	if (vtiter != layer->m_VTexCoords.end()) {
		VTexCoord *vtexcoord = &(*vtiter); 
		for(unsigned int j = 0; j < surface->m_Polygons.getSize(); j++) {
			Polygon *p = &layer->m_Polygons[surface->m_Polygons[j]];
			for(unsigned int k = 0; k < p->m_Size; k++) {
				p->m_TexCoords[k] = vtexcoord->m_Coords[p->m_Vertex[k]];
			}
		}
	}
	
	// polygon texture coordinates
	Map<String,PTexCoord>::Iterator ptiter = layer->m_PTexCoords.at(surface->m_Name);
	if (ptiter != layer->m_PTexCoords.end()) {
		PTexCoord *ptexcoord = &(*ptiter);
		for(unsigned int j = 0; j < ptexcoord->m_Polygons.getSize(); j++) {
			Polygon *p = &layer->m_Polygons[ptexcoord->m_Polygons[j]];
			for(unsigned int k = 0; k < p->m_Size; k++) {
				if(ptexcoord->m_Vertices[j] == p->m_Vertex[k])
					p->m_TexCoords[k] = ptexcoord->m_Coords[j];
			}
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorLWO::calculateNormals(Layer *layer,Surface *surface) {
	unsigned int num_polygons = layer->m_Polygons.getSize();
	unsigned int num_vertices = layer->m_Vertices.getSize();

	// flat normals
	if(surface->m_Smoothing < 0.0f) {
		for(unsigned int i = 0; i < num_polygons; i++) {
			Polygon *p = &layer->m_Polygons[i];
			for(unsigned int j = 0; j < p->m_Size; j++) {
				p->m_Normals[j] = p->m_Normal;
			}
		}
		return;
	}
	
	// Create a multimap that contains all polygons used by a
	// vertex
	Multimap<int,int> vertexmap;
	for(unsigned int i = 0; i < num_polygons; i++) {
		Polygon *p = &layer->m_Polygons[i];
		for(unsigned int j = 0; j < p->m_Size; j++)
			vertexmap.insert(p->m_Vertex[j],i);
	}
	
	for(unsigned int i = 0; i < num_polygons; i++) {
		Polygon *p = &layer->m_Polygons[i];
		for(unsigned int j = 0; j < p->m_Size; j++) {
			int v = p->m_Vertex[j];
			Multimap<int,int>::Iterator iter = vertexmap.at(v);
			ASSERT(iter != vertexmap.end());
			while(iter != vertexmap.end() && iter.key() == v) {
				Vector3<float> normal = layer->m_Polygons[iter.item()].m_Normal;
				float angle = acos(-(p->m_Normal * normal));
				if(angle > surface->m_Smoothing)
					p->m_Normals[j] += normal;
				++iter;
			}
		}
	}
	
	for(unsigned int i = 0; i < num_polygons; i++) {
		Polygon *p = &layer->m_Polygons[i];
		for(unsigned int j = 0; j < p->m_Size; j++)
			p->m_Normals[j].normalise();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Mesh* TranslatorLWO::createMesh(Layer* layer, Surface* surface) {
	// calculate surface mapping and normals
	applyMapping(layer,surface);
	calculateNormals(layer,surface);
	
	// mesh surface
	//Surface *s = new Surface;
	//strcpy(s->name,surface->m_Name);
	
	// create mesh vertexes
	unsigned int num_tris = 0;
	unsigned int num_polys = surface->m_Polygons.getSize();
	for(unsigned int k = 0; k < num_polys; k++) {
		Polygon *p = &layer->m_Polygons[surface->m_Polygons[k]];
		num_tris += p->m_Size - 2;
	}

	// Create vertex buffer
	VertexFormat fmt;
	VertexBuffer* vb;
	fmt.set(0, Types::Float3, Attributes::Vertex);
	fmt.set(1, Types::Float3, Attributes::Normal);
	fmt.set(2, Types::Float2, Attributes::TextureAll);
	vb = VertexBuffer::create(fmt, num_tris*3, VertexBuffer::Static);

	Vertex* vertex = (Vertex*)vb->lock(VertexBuffer::Discard);
	for(unsigned int k = 0; k < num_polys; k++) {
		Polygon *p = &layer->m_Polygons[surface->m_Polygons[k]];
		for(unsigned int l = 2; l < p->m_Size; l++) {
			vertex[0].m_Vertex = layer->m_Vertices[p->m_Vertex[0]];
			vertex[1].m_Vertex = layer->m_Vertices[p->m_Vertex[l - 1]];
			vertex[2].m_Vertex = layer->m_Vertices[p->m_Vertex[l - 0]];
			vertex[0].m_Normal = p->m_Normals[0];
			vertex[1].m_Normal = p->m_Normals[l - 1];
			vertex[2].m_Normal = p->m_Normals[l - 0];
			vertex[0].m_Texture = Vector2<float>(p->m_TexCoords[0].x,1.0f - p->m_TexCoords[0].y);
			vertex[1].m_Texture = Vector2<float>(p->m_TexCoords[l - 1].x,1.0f - p->m_TexCoords[l - 1].y);
			vertex[2].m_Texture = Vector2<float>(p->m_TexCoords[l - 0].x,1.0f - p->m_TexCoords[l - 0].y);
			vertex += 3;
		}
	}
	vb->unlock();
	
	VertexMesh* mesh = new VertexMesh();
	mesh->setName(surface->m_Name);
	mesh->attachVertexBuffer(vb);
	mesh->setPrimitive(Mesh::TriangleList);
	mesh->attachAppearance(surface->m_Appearance);
	mesh->updateBounds();
	return mesh;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Model* TranslatorLWO::createModel() {
	StaticModel* model = new StaticModel();
	for(unsigned int i = 0; i < m_Layers.getSize(); i++) {
		Layer *layer = &m_Layers[i];
		createSurfaces(layer);
		for(unsigned int j = 0; j < layer->m_Surfaces.getSize(); j++) {
			Surface *surface = &layer->m_Surfaces[j];
			Mesh* mesh = createMesh(layer, surface);			
			model->insertMesh(mesh);
			surface->m_Appearance->release();
		}
	}
	
	model->finish();
	
	return model;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorLWO::TranslatorLWO() {
	// Create some resouces
	m_PlainShader = new PlainShader();
	m_TextureShader = new TextureShader();
	m_Material = new ColourMaterial();
	m_Material->setAmbient(Colour<float>(0.2f, 0.2f, 0.2f));
	m_Material->setDiffuse(Colour<float>(1.0f, 1.0f, 1.0f));
	m_Material->setSpecular(Colour<float>(1.0f, 1.0f, 1.0f));
	m_Material->setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorLWO::~TranslatorLWO() {
	// Release resources
	m_PlainShader->release();
	m_TextureShader->release();
	m_Material->release();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Model* TranslatorLWO::onDecode(InputStream* in,size_t ) {
	m_In = in;
	int chunk = readInt();
	if(chunk != LWO_ID_FORM)
		throw magnum::io::TranslatorFormatException();
	
	int size = readInt();
	size -= 4;
	
	chunk = readInt();
	if(chunk != LWO_ID_LWO2)
		throw magnum::io::TranslatorFormatException();
	
	// Clear layer list
	m_Layers.clear();
	m_CurrentLayer = NULL;
	
	read_LWO2(size);
	
	return createModel();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorLWO::onEncode(OutputStream* ,const Model* ) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorLWO::onIdentify(File* in) {
	dword form;
	dword size;
	dword chunk;

	try {
		int64 oldpos;
		oldpos = in->getPosition();	
		in->read(&form, sizeof(form));
		in->read(&size, sizeof(size));
		in->read(&chunk, sizeof(chunk));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if ((form != dword2msb(LWO_ID_FORM)) || (chunk != dword2msb(LWO_ID_LWO2)))
		return false;

	return true;
}

}; };

