/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "io.exception.h"
#include "io.inputstream.h"
#include "io.sysfile.h"
#include "container.heap.h"
#include "graphics3d.context.h"
#include "graphics3d.plainshader.h"
#include "graphics3d.textureshader.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.material.h"
#include "graphics3d.texture2d.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.meshgroup.h"
#include "graphics3d.framemodel.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.renderstream.h"
#include "graphics3d.appearance.h"
#include "graphics3d.texturemanager.h"
#include "graphics3d.shadermanager.h"
#include "graphics3d.frameanimation.h"
#include "graphics3d.translator.md2.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "io.inputstream.inl"
#include "io.translator.inl"
#include "io.file.inl"
#include "container.heap.inl"
#include "container.map.inl"
#include "container.array.inl"
#include "container.list.inl"
#include "container.iterator.inl"
#include "container.pair.inl"
#include "container.chain.inl"
#include "container.cache.inl"
#include "container.allocator.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.context.inl"
#include "graphics3d.meshgroup.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.model.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.framemodel.inl"
#include "graphics3d.animation.inl"
#include "graphics3d.frameanimation.inl"
#include "graphics3d.keyframe.inl"
#include "graphics3d.translator.inl"
#include "graphics3d.translator.md2.inl"

#define MD2_MAGIC		0x32504449


namespace magnum { namespace graphics3d {
using magnum::container::Heap;
using magnum::object::AutoPointer;

static float vertexnormals[][3] = {
#include "graphics3d.translator.md2normals.h"
};


const TranslatorMD2::AnimationInfo TranslatorMD2::m_Animations[] = {
    // first, last, fps, looping, name
    {   0,  39,  9, true, "stand" },			// STAND
    {  40,  45, 10, true, "run" },				// RUN
    {  46,  53, 10, false, "attack" },			// ATTACK
    {  54,  57,  7, false, "pain_a" },			// PAIN_A
    {  58,  61,  7, false, "pain_b" },			// PAIN_B
    {  62,  65,  7, false, "pain_c" },			// PAIN_C
    {  66,  71,  7, false, "jump" },			// JUMP_DOWN
    {  72,  83,  7, false, "flip" },			// FLIP
    {  84,  94,  7, false, "salute" },			// SALUTE
    {  95, 111, 10, false, "fallback" },		// FALLBACK
    { 112, 122,  7, false, "wave" },			// WAVE
    { 123, 134,  6, false, "point" },			// POINT
    { 135, 153, 10, true, "crouch_stand" },		// CROUCH_STAND
    { 154, 159,  7, true, "crouch_walk" },		// CROUCH_WALK
    { 160, 168, 10, false, "crouch_attack" },   // CROUCH_ATTACK
    { 169, 172,  7, false, "crouch_pain" },		// CROUCH_PAIN
    { 173, 177,  5, false, "crouch_death" },	// CROUCH_DEATH
    { 178, 183,  7, false, "death_fallback" },  // DEATH_FALLBACK
    { 184, 189,  7, false, "death_fallforward" }, // DEATH_FALLFORWARD
    { 190, 197,  7, false, "death_fallbackslow" }, // DEATH_FALLBACKSLOW
    // JUMP is not in the table; it is handled specially
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorMD2::TranslatorMD2() {
	m_Triangles = NULL;
	m_Vertices = NULL;
	m_TexCoords = NULL;
	m_In = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Model* TranslatorMD2::onDecode(InputStream* in,size_t size) {
	if (size == Autosize)
		size = (size_t)in->available();

	m_Frames.clear();

	m_Triangles = NULL;
	m_Vertices = NULL;
	m_TexCoords = NULL;
	m_Skin = NULL;
	m_In = in;

	// Read in Header
	Header head;

	if (in->read(&head, sizeof(head)) != sizeof(head))
		throw magnum::io::TranslatorFormatException();
	if (head.m_Magic != Magic)
		throw magnum::io::TranslatorFormatException();

	size_t position = sizeof(head);

	m_TriangleCount = head.m_NumTriangles;
	m_FrameSize = head.m_FrameSize;
	m_SkinWidth = head.m_SkinWidth;
	m_SkinHeight = head.m_SkinHeight;

	int offsetSkins = head.m_OffsetSkins; 
	int offsetTexCoords = head.m_OffsetTexCoords;
	int offsetTriangles = head.m_OffsetTriangles;
	int offsetFrames = head.m_OffsetFrames;
	int offsetGlCommands = head.m_OffsetGlCommands;
	size_t skip;

	// Get next chunk to read
	Heap<int> ofHeap;
	ofHeap.insert(-offsetSkins);
	ofHeap.insert(-offsetTexCoords);
	ofHeap.insert(-offsetTriangles);
	ofHeap.insert(-offsetFrames);
	ofHeap.insert(-offsetGlCommands);
	for (unsigned int i= 0; i < 5; i++) {
		size_t nextPos = -ofHeap.remove();
		if (nextPos == 0)
			continue;
		if (nextPos < position)
			throw magnum::io::TranslatorFormatException();

		m_In->skip(nextPos - position);
		position += nextPos - position;

		if (nextPos == offsetSkins) {
			skip = readSkins(head.m_NumSkins);	
			offsetSkins = 0;
		} else if (nextPos == offsetTexCoords) {
			skip = readTexCoords(head.m_NumTexCoords);
			offsetTexCoords = 0;
		} else if (nextPos == offsetTriangles) {
			skip = readTriangles(head.m_NumTriangles);
			offsetTriangles = 0;
		} else if (nextPos == offsetFrames) {
			skip = readFrames(head.m_NumFrames);
			offsetFrames = 0;
		} else if (nextPos == offsetGlCommands) {
			skip = readGlCommands(head.m_OffsetGlCommands);
			offsetGlCommands = 0;
		}

		position += skip;
	}

	// Create Appearance etc
	AutoPointer<Appearance> appearance = new Appearance();
	if (m_Skin != NULL) {
		AutoPointer<Shader> shader = new TextureShader();
		appearance->attachShader(shader);
		appearance->attachTexture(m_Skin);
	} else {
		AutoPointer<Shader> shader = new PlainShader();
		appearance->attachShader(shader);
	}
	AutoPointer<ColourMaterial> material = new ColourMaterial();
	material->setAmbient(Colour<float>(0.2f, 0.2f, 0.2f));
	material->setDiffuse(Colour<float>(1.0f, 1.0f, 1.0f));
	material->setSpecular(Colour<float>(1.0f, 1.0f, 1.0f));
	material->setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
	material->setShininess(20.0f);
	appearance->attachMaterial(material);
	
	// Create Model etc
	FrameModel* model = new FrameModel();
	AutoPointer<MeshGroup> meshgroup = new MeshGroup();
	AutoPointer<IndexMesh> mesh = new IndexMesh();
	mesh->setPrimitive(Mesh::TriangleList);
	mesh->attachAppearance(appearance);

	meshgroup->insert(mesh);
	model->insertGroup(meshgroup);

	// Process model data
	createIndexMapping();
	createTextureCoords(mesh);
	createIndices(mesh);
	createFrames(model, meshgroup);
	createAnimations(model);

	// finish Model
	mesh->updateBounds();
	model->finish();

	// Release memory
	m_VertexSource.clear();
	m_TextureSource.clear();
	m_VertexMapping.clear();
	for (unsigned int i = 0; i < m_Frames.getSize(); i++)
		delete[] (byte*)m_Frames[i];
	m_Frames.clear();
	if (m_TexCoords != NULL) {
		delete[] m_TexCoords;
		m_TexCoords = NULL;
	}
	if (m_Triangles != NULL) {
		delete[] m_Triangles;
		m_Triangles = NULL;
	}
	if (m_Vertices != NULL) {
		delete[] m_Vertices;
		m_Vertices = NULL;
	}
	if (m_Skin != NULL) {
		m_Skin->release();
		m_Skin = NULL;
	}

	return model;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD2::readSkins(unsigned int count) {
	char skin[64];
	size_t read = 0;

	for (unsigned int i = 0; i < count; i++) {
		read += m_In->read(skin, sizeof(skin));
		if (m_Skin != NULL)
			continue;

		// Try to load skin
		m_Skin = dynamic_cast<Texture2D*>(createTexture(skin));
		if (m_Skin != NULL)
			continue;
			
		// Try to cut off path information
		if (strchr(skin,'/') != NULL) {
			const char* s1 = skin;
			const char* s2 = strchr(s1,'/');
			do {
				s1 = s2+1;
				s2 = strchr(s1,'/');
			} while(s2 != NULL);
			m_Skin = dynamic_cast<Texture2D*>(createTexture(s1));
			if (m_Skin != NULL)
				continue;
		}
	}

	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD2::readTexCoords(unsigned int count) {
	size_t read;

	m_TexCoords = new TextureCoordinate[count];
	read = m_In->read(m_TexCoords, sizeof(TextureCoordinate)*count);

	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD2::readTriangles(unsigned int count) {
	size_t read;

	m_Triangles = new Triangle[count];
	read = m_In->read(m_Triangles, sizeof(Triangle)*count);

	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD2::readFrames(unsigned int count) {
	size_t read = 0;

	for (unsigned int i = 0; i < count; i++) {
		Frame* frame = (Frame *)new byte[m_FrameSize];
		read += m_In->read(frame, m_FrameSize);
		m_Frames.insert(frame);
	}

	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int TranslatorMD2::readGlCommands(unsigned int count) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMD2::createIndexMapping() {
	m_VertexSource.clear();
	m_TextureSource.clear();
	m_VertexMapping.clear();

	unsigned int vertex = 0;

	for (unsigned int i = 0; i < m_TriangleCount; i++) {
		IPair ip;
		for (unsigned int j = 0; j < 3; j++) {
			ip = IPair(m_Triangles[i].m_VertexIndices[j],m_Triangles[i].m_TextureIndices[j]);
			if (m_VertexMapping.contains(ip))
				continue;
			m_VertexMapping.insert(ip, vertex);
			m_VertexSource.insert(m_Triangles[i].m_VertexIndices[j]);
			m_TextureSource.insert(m_Triangles[i].m_TextureIndices[j]);
			vertex++;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMD2::createTextureCoords(IndexMesh* mesh) {
	unsigned int count = m_TextureSource.getSize();

	// Allocate Vertex Buffer
	VertexFormat fmt;
	fmt.set(0, Types::Float2, Attributes::TextureAll);
	AutoPointer<VertexBuffer> vb = VertexBuffer::create(fmt, count, VertexBuffer::Static);
	mesh->attachVertexBuffer(vb, 0);

	Vector2<float>* texcords = (Vector2<float>*)vb->lock(VertexBuffer::Discard);
	unsigned int texsrc;
	float u,v;

	for (unsigned int i = 0; i < count; i++) {
		texsrc = m_TextureSource[i];
		u = (float)m_TexCoords[texsrc].s / (float)m_SkinWidth;
		v = (float)m_TexCoords[texsrc].t / (float)m_SkinHeight;
		texcords[i].x = u;
		texcords[i].y = v;
	}

	vb->unlock();
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMD2::createFrames(FrameModel* model, MeshGroup *group) {
	unsigned int vc = m_VertexSource.getSize();
	unsigned int fc = m_Frames.getSize();

	VertexFormat fmt;
	fmt.set(0, Types::Float3, Attributes::Vertex);
	fmt.set(1, Types::Float3, Attributes::Normal);
	Frame* frame;
	unsigned int idx;

	for (unsigned int f = 0; f < fc; f++) {
		frame = (Frame*)m_Frames[f];
		AutoPointer<VertexBuffer> vb = VertexBuffer::create(fmt, vc, VertexBuffer::System);
		vb->setName(frame->m_Name);
		model->insertFrame(group, vb, f);

		void* buf = vb->lock(VertexBuffer::Discard);
		Vector3<float>* vptr = (Vector3<float>*)buf;
		Vector3<float>* nptr = vptr + 1;
		for (unsigned int i = 0; i < vc; i++) {
			idx = m_VertexSource[i];
			vptr->x = (float)frame->vertices[idx].m_Vertex[1] * frame->m_Scale[1] + frame->m_Translate[1];
			vptr->y = (float)frame->vertices[idx].m_Vertex[2] * frame->m_Scale[2] + frame->m_Translate[2];
			vptr->z = -(float)frame->vertices[idx].m_Vertex[0] * frame->m_Scale[0] - frame->m_Translate[0];
			nptr->x = vertexnormals[frame->vertices[idx].m_LightNormalIndex][1];
			nptr->y = vertexnormals[frame->vertices[idx].m_LightNormalIndex][2];
			nptr->z = -vertexnormals[frame->vertices[idx].m_LightNormalIndex][0];
			vptr += 2;
			nptr += 2;
		}
		vb->unlock();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMD2::createIndices(IndexMesh* mesh) {
	unsigned int tc = m_TriangleCount;
	IPair ip;

	AutoPointer<IndexBuffer> ib = IndexBuffer::create(IndexBuffer::Index16, tc*3, IndexBuffer::Static);
	mesh->attachIndexBuffer(ib);

	word *indices = (word*)ib->lock(IndexBuffer::Discard);
	for (unsigned int i = 0; i < m_TriangleCount; i++) {
		ip = IPair(m_Triangles[i].m_VertexIndices[0],m_Triangles[i].m_TextureIndices[0]);
		indices[i*3] = m_VertexMapping[ip];
		ip = IPair(m_Triangles[i].m_VertexIndices[1],m_Triangles[i].m_TextureIndices[1]);
		indices[i*3 + 1] = m_VertexMapping[ip];
		ip = IPair(m_Triangles[i].m_VertexIndices[2],m_Triangles[i].m_TextureIndices[2]);
		indices[i*3 + 2] = m_VertexMapping[ip];
	}
	ib->unlock();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMD2::createAnimations(FrameModel* model) {
	for (unsigned int i = 0; i < sizeof(m_Animations)/sizeof(AnimationInfo); ++i) {
		AutoPointer<FrameAnimation> anim = new FrameAnimation();
		anim->setName(m_Animations[i].m_Name);
		anim->setRepeat(m_Animations[i].m_Loop ? Animation::Loop : Animation::Single);
		float dist = 1.0f / (float)m_Animations[i].m_Fps;
		float time = 0.0f;
		for (unsigned int j = m_Animations[i].m_First; j <= m_Animations[i].m_Last; ++j) {
			anim->insert(time, j);
			time += dist;
		}
		anim->insert(time, m_Animations[i].m_First);
		anim->finish();
		model->insertAnimation(anim);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD2::onEncode(OutputStream* ,const Model* ) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorMD2::onIdentify(File* in) {
	dword id;

	try {
		int64 oldpos;
		oldpos = in->getPosition();	
		in->read(&id, sizeof(id));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if (id != MD2_MAGIC)
		return false;

	return true;
}


}; };
