/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "io.exception.h"
#include "io.inputstream.h"
#include "io.sysfile.h"
#include "container.heap.h"
#include "graphics3d.context.h"
#include "graphics3d.plainshader.h"
#include "graphics3d.textureshader.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.material.h"
#include "graphics3d.texture.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.meshgroup.h"
#include "graphics3d.framemodel.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.renderstream.h"
#include "graphics3d.appearance.h"
#include "graphics3d.texturemanager.h"
#include "graphics3d.shadermanager.h"
#include "graphics3d.frameanimation.h"
#include "graphics3d.translator.md3.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "io.inputstream.inl"
#include "io.file.inl"
#include "io.translator.inl"
#include "container.heap.inl"
#include "container.array.inl"
#include "container.list.inl"
#include "container.iterator.inl"
#include "container.chain.inl"
#include "container.allocator.inl"
#include "container.map.inl"
#include "container.cache.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.context.inl"
#include "graphics3d.meshgroup.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.model.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.framemodel.inl"
#include "graphics3d.animation.inl"
#include "graphics3d.frameanimation.inl"
#include "graphics3d.keyframe.inl"
#include "graphics3d.translator.inl"
#include "graphics3d.translator.md3.inl"


namespace magnum { namespace graphics3d {
using magnum::container::Heap;
using magnum::object::AutoPointer;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorMD3::TranslatorMD3() {
	m_In = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Model* TranslatorMD3::onDecode(InputStream* in,size_t size) {
	m_Frames.clear();
	m_Textures.clear();
	m_Vertices.clear();
	m_Surfaces.clear();
	m_Tags.clear();
	m_In = in;

	// Read in Header
	Header head;
	if (in->read(&head, sizeof(head)) != sizeof(head))
		throw magnum::io::TranslatorFormatException();
	if ((head.m_Magic != Magic1) && (head.m_Magic != Magic2))
		throw magnum::io::TranslatorFormatException();


	int offsetTags = head.m_OffsetTags;
	int offsetSurfaces = head.m_OffsetSurfaces;
	int offsetFrames = head.m_OffsetFrames;

	// Create empty model
	AutoPointer<FrameModel> model = new FrameModel();
	AutoPointer<MeshGroup> group = new MeshGroup();
	model->insertGroup(group);

	// Get next chunk to read
	Heap<int> ofHeap;
	ofHeap.insert(-offsetTags);
	ofHeap.insert(-offsetSurfaces);
	ofHeap.insert(-offsetFrames);
	size_t position = sizeof(head);
	size_t skip;
	for (unsigned int i= 0; i < 3; i++) {
		size_t nextPos = -ofHeap.remove();
		if (nextPos == 0)
			continue;
		if (nextPos < position)
			throw magnum::io::TranslatorFormatException();

		m_In->skip(nextPos - position);
		position += nextPos - position;

		if (nextPos == offsetTags) {
			skip = readTags(head.m_NumTags);
			offsetTags = 0;	
		} else if (nextPos == offsetSurfaces) {
			skip = readSurfaces(group, head.m_NumSurfaces);
			offsetSurfaces = 0;
		} else if (nextPos == offsetFrames) {
			skip = readFrames(head.m_NumFrames);
			offsetFrames = 0;
		}

		position += skip;
	}

	// Create frames
	createFrames(model, group);

	// Create animations
	createAnimations(model);
	
	// Create joints
	createJoints(model);

	// finish Model
	model->finish();
	model->refer();

	// Release memory
	m_Frames.clear();
	m_Vertices.clear();
	m_Textures.clear();
	m_Surfaces.clear();
	m_Tags.clear();

	return model;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD3::readTags(unsigned int count) {
	size_t read = 0;

	for (unsigned int i = 0; i < count; i++) {
		Tag* tag = m_Tags.append();
		read += m_In->read(tag, sizeof(Tag));
	}

	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD3::readFrames(unsigned int count) {
	size_t read = 0;

	for (unsigned int i = 0; i < count; i++) {
		Frame* frame = m_Frames.append();
		read += m_In->read(frame, sizeof(Frame));
	}

	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD3::readSurfaces(MeshGroup* group, unsigned int count) {
	AutoPointer<PlainShader> shader = new PlainShader();
	AutoPointer<ColourMaterial> material = new ColourMaterial();
	material->setAmbient(Colour<float>(0.1f, 0.1f, 0.1f));
	material->setDiffuse(Colour<float>(0.41f, 0.41f, 0.52f));
	material->setSpecular(Colour<float>(1.0f, 1.0f, 1.0f));
	material->setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
	material->setShininess(20.0f);
	size_t total = 0;
	
	for (unsigned int i = 0; i < count; i++) {
		Surface *surf = m_Surfaces.append();
		size_t read = m_In->read(surf, sizeof(Surface));
		if (read != sizeof(Surface))
			throw magnum::io::TranslatorFormatException();
		total += read;

		// Create empty mesh			
		AutoPointer<IndexMesh> mesh = new IndexMesh();
		mesh->setName(surf->m_Name);
		AutoPointer<Appearance> app = new Appearance();
		app->attachShader(shader);
		app->attachMaterial(material);
		mesh->attachAppearance(app);
		group->insert(mesh);

		// Allocate space for vertex- and texture data
		unsigned int vstart = m_Vertices.append(surf->m_NumVertices * surf->m_NumFrames);
		Vertex* vcoords = &m_Vertices[vstart];

		unsigned int tstart = m_Textures.append(surf->m_NumVertices);
		Vector2<float>* tcoords = &m_Textures[tstart];

		// Get next chunk to read
		Heap<int> ofHeap;
		ofHeap.insert(-(int)surf->m_OffsetShaders);
		ofHeap.insert(-(int)surf->m_OffsetTextures);
		ofHeap.insert(-(int)surf->m_OffsetTriangles);
		ofHeap.insert(-(int)surf->m_OffsetVertices);
		size_t position = sizeof(Surface);
		size_t skip = 0;
		for (unsigned int i= 0; i < 4; i++) {
			size_t nextPos = -ofHeap.remove();
			if (nextPos == 0)
				continue;
			if (nextPos < position)
				throw magnum::io::TranslatorFormatException();

			m_In->skip(nextPos - position);
			position += nextPos - position;

			if (nextPos == surf->m_OffsetShaders) {
				skip = readShaders(surf->m_NumShaders);	
				surf->m_OffsetShaders = 0;
			} else if (nextPos == surf->m_OffsetTextures) {
				skip = readTextures(tcoords, surf->m_NumVertices);	
				surf->m_OffsetTextures = 0;
			} else if (nextPos == surf->m_OffsetTriangles) {
				skip = readTriangles(mesh, tstart, surf->m_NumTriangles);
				surf->m_OffsetTriangles = 0;
			} else if (nextPos == surf->m_OffsetVertices) {
				skip = readVertices(vcoords, surf->m_NumVertices * surf->m_NumFrames);
				surf->m_OffsetVertices = 0;
			}

			total += skip;
			position += skip;
		}
	}
	
	return total;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD3::readShaders(unsigned int count) {
	Shader shader;
	size_t read = 0;

	for (unsigned int i = 0; i < count; i++) {
		read += m_In->read(&shader, sizeof(Shader));
	}

	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD3::readTextures(Vector2<float>* data, unsigned int count) {
	return m_In->read(data, count*sizeof(Vector2<float>));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD3::readTriangles(IndexMesh* mesh, unsigned int offset, unsigned int count) {
	Array<unsigned int> indices;
	indices.setSize(3*count);
	size_t read;
	read = m_In->read(&indices[0], 3*count*sizeof(int));
	
	AutoPointer<IndexBuffer> ib = IndexBuffer::create(IndexBuffer::Index32, 3*count, IndexBuffer::Static);
	mesh->attachIndexBuffer(ib);
	unsigned int* dst = (unsigned int*)ib->lock(IndexBuffer::Discard);
	unsigned int* src = &indices[0];
	for (unsigned int i = 0; i < count; i++) {
		dst[0] = src[0] + offset;
		dst[2] = src[1] + offset;
		dst[1] = src[2] + offset;
		dst += 3;
		src += 3;
	}
	ib->unlock();
	
	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD3::readVertices(TranslatorMD3::Vertex* data, unsigned int count) {
	return m_In->read(data, count*sizeof(Vertex));
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMD3::createFrames(FrameModel* model, MeshGroup *group) {
	unsigned int vc = m_Textures.getSize();
	unsigned int fc = m_Frames.getSize();

	VertexFormat fmt;
	fmt.set(0, Types::Float3, Attributes::Vertex);
	fmt.set(1, Types::Float3, Attributes::Normal);
	
	unsigned int f = 0;
	for (List<Frame>::Iterator frame = m_Frames.begin(); frame != m_Frames.end(); frame++, f++) {
		// Create vertex buffer for this frame
		AutoPointer<VertexBuffer> vb = VertexBuffer::create(fmt, vc, VertexBuffer::System);
		vb->setName(frame->m_Name);
		model->insertFrame(group, vb, f);

		// Lock vertex buffer of current frame
		void* buf = vb->lock(VertexBuffer::Discard);
		Vector3<float>* vptr = (Vector3<float>*)buf;
		Vector3<float>* nptr = vptr + 1;

		// Loop over all surfaces
		unsigned start = 0;
		for (List<Surface>::Iterator iter = m_Surfaces.begin(); iter != m_Surfaces.end(); ++iter) {
			// Get starting address in source vertex buffer
			unsigned int idx = start + f*iter->m_NumVertices;
			Vertex* src = &m_Vertices[idx];
			
			// Copy vertex buffer
			for (unsigned int i = 0; i < iter->m_NumVertices; i++) {
				vptr->x = (float)src->m_Vertex[0] / 64.0f;
				vptr->y = (float)src->m_Vertex[1] / 64.0f;
				vptr->z = (float)src->m_Vertex[2] / 64.0f;
				float lng = (float)(src->m_Normal & 0xff) * (2.0f * (float)M_PI / 255.0f);
				float lat = (float)((src->m_Normal >> 8) & 0xff) * (2.0f * (float)M_PI / 255.0f);
				nptr->x = cos(lat)*sin(lng);
				nptr->y = sin(lat)*sin(lng);
				nptr->z = cos(lng);
				vptr += 2;
				nptr += 2;
				src++;
			}
		
			// Go to next surface
			start += iter->m_NumVertices * iter->m_NumFrames;
		}
		
		// Unlock vertex buffer
		vb->unlock();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMD3::createAnimations(FrameModel* model) {
	Map<String, FrameAnimation*> anims;
	unsigned int f = 0;
	for (List<Frame>::Iterator frame = m_Frames.begin(); frame != m_Frames.end(); frame++, f++) {
		// Get prefix of animation
		String name = frame->m_Name;

		FrameAnimation* anim;
		if (anims.contains(name)) {
			anim = anims[name];
		} else {
			anim = new FrameAnimation();
			anim->setName(name);
			anim->setRepeat(Animation::Loop);
			anims.insert(name, anim);
		}
		
		anim->insert(anim->getSize() * 0.1f, f);
	}

	for (Map<String, FrameAnimation*>::Iterator anim = anims.begin(); anim != anims.end(); ++anim) {
		(*anim)->finish();
		model->insertAnimation(*anim);
		(*anim)->release();		
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMD3::createJoints(FrameModel* model) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD3::onEncode(OutputStream* ,const Model* ) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorMD3::onIdentify(File* in) {
	dword id;

	try {
		int64 oldpos;
		oldpos = in->getPosition();	
		in->read(&id, sizeof(id));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if ((id != Magic1) && (id != Magic2))
		return false;

	return true;
}


}; };
