/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "io.exception.h"
#include "io.inputstream.h"
#include "io.sysfile.h"
#include "container.heap.h"
#include "graphics3d.context.h"
#include "graphics3d.plainshader.h"
#include "graphics3d.textureshader.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.material.h"
#include "graphics3d.texture.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.meshgroup.h"
#include "graphics3d.framemodel.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.renderstream.h"
#include "graphics3d.appearance.h"
#include "graphics3d.texturemanager.h"
#include "graphics3d.shadermanager.h"
#include "graphics3d.frameanimation.h"
#include "graphics3d.translator.md4.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "io.inputstream.inl"
#include "io.file.inl"
#include "io.translator.inl"
#include "container.heap.inl"
#include "container.array.inl"
#include "container.list.inl"
#include "container.iterator.inl"
#include "container.chain.inl"
#include "container.allocator.inl"
#include "container.cache.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.context.inl"
#include "graphics3d.meshgroup.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.model.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.framemodel.inl"
#include "graphics3d.animation.inl"
#include "graphics3d.frameanimation.inl"
#include "graphics3d.keyframe.inl"
#include "graphics3d.translator.inl"
#include "graphics3d.translator.md4.inl"


namespace magnum { namespace graphics3d {
using magnum::container::Heap;
using magnum::object::AutoPointer;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorMD4::TranslatorMD4() {
	m_In = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Model* TranslatorMD4::onDecode(InputStream* in,size_t size) {
	m_Frames.clear();
	m_LODs.clear();
	m_In = in;

	// Read in Header
	Header head;
	if (in->read(&head, sizeof(head)) != sizeof(head))
		throw magnum::io::TranslatorFormatException();
	if (head.m_Magic != Magic)
		throw magnum::io::TranslatorFormatException();

	// Create empty model
	AutoPointer<FrameModel> model = new FrameModel();
	AutoPointer<MeshGroup> group = new MeshGroup();
	model->insertGroup(group);

	// Get next chunk to read
	Heap<int> ofHeap;
	ofHeap.insert(-(int)head.m_OffsetFrames);
	ofHeap.insert(-(int)head.m_NumLODs);
	size_t position = sizeof(head);
	size_t skip;
	for (unsigned int i= 0; i < 2; i++) {
		size_t nextPos = -ofHeap.remove();
		if (nextPos == 0)
			continue;
		if (nextPos < position)
			throw magnum::io::TranslatorFormatException();

		m_In->skip(nextPos - position);
		position += nextPos - position;

		if (nextPos == head.m_OffsetFrames) {
			skip = readFrames(head.m_NumFrames, head.m_NumBones);
			head.m_OffsetFrames = 0;	
		} else if (nextPos == head.m_OffsetLODs) {
			skip = readLODs(head.m_NumLODs);
			head.m_OffsetLODs = 0;
		}

		position += skip;
	}

	// Create frames
//	createFrames(model, group);

	// finish Model
	model->finish();
	model->refer();

	// Release memory
	m_Frames.clear();
	m_LODs.clear();

	return model;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD4::readFrames(unsigned int count,unsigned int bones) {
	size_t read = 0;

	for (unsigned int i = 0; i < count; i++) {
		FrameExt* frame = m_Frames.append();
		read += m_In->read(frame, sizeof(Frame));
		read += readBones(frame, bones);
	}

	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD4::readBones(FrameExt* frame,unsigned int count) {
	size_t read = 0;

	for (unsigned int i = 0; i < count; i++) {
		Bone* bone = frame->m_Bones.append();
		read += m_In->read(bone, sizeof(Bone));
	}

	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD4::readLODs(unsigned int count) {
	size_t read = 0;

	for (unsigned int i = 0; i < count; i++) {
		LODExt* lod = m_LODs.append();
		read += m_In->read(lod, sizeof(LOD));
		read += (size_t)m_In->skip(lod->m_OffsetSurfaces - sizeof(LOD));
		read += readSurfaces(lod, lod->m_NumSurfaces);
	}
	
	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD4::readSurfaces(LODExt* lod, unsigned int count) {
	size_t total = 0;

	for (unsigned int i = 0; i < count; i++) {
		Surface* surf = lod->m_Surfaces.append();
		size_t read = m_In->read(surf, sizeof(Surface));
		if (read != sizeof(Surface))
			throw magnum::io::TranslatorFormatException();
		total += read;
		
		// Get next chunk to read
		Heap<int> ofHeap;
		ofHeap.insert(-(int)surf->m_OffsetBoneReferences);
		ofHeap.insert(-(int)surf->m_OffsetTriangles);
		ofHeap.insert(-(int)surf->m_OffsetVertices);
		size_t position = sizeof(Surface);
		size_t skip = 0;
		for (unsigned int i= 0; i < 4; i++) {
			size_t nextPos = -ofHeap.remove();
			if (nextPos == 0)
				continue;
			if (nextPos < position)
				throw magnum::io::TranslatorFormatException();

			m_In->skip(nextPos - position);
			position += nextPos - position;

			if (nextPos == surf->m_OffsetBoneReferences) {
				skip = readBoneReferences(surf->m_NumBoneReferences);	
				surf->m_OffsetBoneReferences = 0;
			} else if (nextPos == surf->m_OffsetTriangles) {
				skip = readTriangles(surf->m_NumTriangles);
				surf->m_OffsetTriangles = 0;
			} else if (nextPos == surf->m_OffsetVertices) {
				skip = readVertices(surf->m_NumVertices);
				surf->m_OffsetVertices = 0;
			}

			total += skip;
			position += skip;
		}
	}
	
	return total;	
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD4::readTriangles(unsigned int count) {
	return (size_t)m_In->skip(count*3*sizeof(int));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD4::readBoneReferences(unsigned int count) {
	return (size_t)m_In->skip(count*sizeof(int));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD4::readVertices(unsigned int count) {
	size_t read = 0;

	for (unsigned int i = 0; i < count; i++) {
		Vertex v;
		read += m_In->read(&v, sizeof(Vertex));
		read += (size_t)m_In->skip(v.m_NumWeights * sizeof(Weight));
	}
	
	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMD4::onEncode(OutputStream* ,const Model* ) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorMD4::onIdentify(File* in) {
	dword id;

	try {
		int64 oldpos;
		oldpos = in->getPosition();	
		in->read(&id, sizeof(id));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if (id != Magic)
		return false;

	return true;
}

}; };

