/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "io.exception.h"
#include "io.inputstream.h"
#include "io.sysfile.h"
#include "container.heap.h"
#include "geometry.transform.h"
#include "graphics3d.context.h"
#include "graphics3d.plainshader.h"
#include "graphics3d.textureshader.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.material.h"
#include "graphics3d.texture.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.weightbuffer.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.skeleton.h"
#include "graphics3d.bonemodel.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.renderstream.h"
#include "graphics3d.appearance.h"
#include "graphics3d.texturemanager.h"
#include "graphics3d.shadermanager.h"
#include "graphics3d.translator.mds.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "math.vector.inl"
#include "math.vector3.inl"
#include "math.matrix.inl"
#include "math.quaternion.inl"
#include "io.inputstream.inl"
#include "io.file.inl"
#include "io.translator.inl"
#include "container.heap.inl"
#include "container.array.inl"
#include "container.list.inl"
#include "container.iterator.inl"
#include "container.chain.inl"
#include "container.allocator.inl"
#include "container.map.inl"
#include "container.cache.inl"
#include "geometry.transform.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.context.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.weightbuffer.inl"
#include "graphics3d.model.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.skeleton.inl"
#include "graphics3d.bonemodel.inl"
#include "graphics3d.animation.inl"
#include "graphics3d.boneanimation.inl"
#include "graphics3d.keyframe.inl"
#include "graphics3d.translator.inl"
#include "graphics3d.translator.mds.inl"


namespace magnum { namespace graphics3d {
using magnum::container::Heap;
using magnum::object::AutoPointer;
using magnum::geometry::Transform3;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorMDS::TranslatorMDS() {
	m_In = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMDS::onEncode(OutputStream* ,const Model* ) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorMDS::onIdentify(File* in) {
	dword id;

	try {
		int64 oldpos;
		oldpos = in->getPosition();	
		in->read(&id, sizeof(id));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if (id != Magic)
		return false;

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Model* TranslatorMDS::onDecode(InputStream* in,size_t size) {
	m_Frames.clear();
	m_Surfaces.clear();
	m_Tags.clear();
	m_In = in;

	// Read in Header
	Header head;
	if (in->read(&head, sizeof(head)) != sizeof(head))
		throw magnum::io::TranslatorFormatException();
	if (head.magic != Magic)
		throw magnum::io::TranslatorFormatException();

	int offsetTags = head.offsetTags;
	int offsetSurfaces = head.offsetSurfaces;
	int offsetFrames = head.offsetFrames;
	int offsetBones = head.offsetBones;

	// Create empty model
	AutoPointer<BoneModel> model = new BoneModel();
	model->setName(head.name);
	AutoPointer<WeightBuffer> weights = new WeightBuffer();
	model->attachWeightBuffer(weights);
	AutoPointer<Skeleton> skeleton = new Skeleton();
	model->attachSkeleton(skeleton);

	// Get next chunk to read
	Heap<int> ofHeap;
	ofHeap.insert(-offsetTags);
	ofHeap.insert(-offsetSurfaces);
	ofHeap.insert(-offsetFrames);
	ofHeap.insert(-offsetBones);
	size_t position = sizeof(head);
	size_t skip;
	for (unsigned int i= 0; i < 3; i++) {
		size_t nextPos = -ofHeap.remove();
		if (nextPos == 0)
			continue;
		if (nextPos < position)
			throw magnum::io::TranslatorFormatException();

		m_In->skip(nextPos - position);
		position += nextPos - position;

		if (nextPos == offsetTags) {
			skip = readTags(head.numTags);
			offsetTags = 0;	
		} else if (nextPos == offsetBones) {
			skip = readBones(skeleton, head.numBones);
			offsetBones = 0;
		} else if (nextPos == offsetSurfaces) {
			skip = readSurfaces(model, weights, head.numSurfaces);
			offsetSurfaces = 0;
		} else if (nextPos == offsetFrames) {
			skip = readFrames(head.numBones, head.numFrames);
			offsetFrames = 0;
		}

		position += skip;    
	}

	// Create vertices
	createVertices(model);

	// Create animations
	createAnimations(model);

	// finish Model
	weights->finish();
	model->finish();
	model->refer();

	// Release memory
	m_Frames.clear();
	m_Surfaces.clear();
	m_Tags.clear();

	return model;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMDS::readTags(unsigned int count) {
	size_t read = 0;

	for (unsigned int i = 0; i < count; i++) {
		read += m_In->read((Tag*)m_Tags.append(), sizeof(Tag));
	}

	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMDS::readBones(Skeleton* skeleton,unsigned int count) {
	size_t read = 0;
	Array<magnum::graphics3d::Bone*> bones;

	for (unsigned int i = 0; i < count; i++) {
		Bone b;
		read += m_In->read(&b, sizeof(Bone));
		
		if (b.parentIndex == -1)
			bones.insert(skeleton->insert(b.name));
		else
			bones.insert(skeleton->insert(b.name, bones[b.parentIndex]));

		magnum::graphics3d::Bone* bone = *(bones.end() - 1);
		bone->setDistance(b.parentDist);
		bone->setTorsoWeight(b.torsoWeight);
		bone->setIndex(i);
	}

	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMDS::readSurfaces(BoneModel* model,WeightBuffer* weights,unsigned int count) {
	AutoPointer<PlainShader> shader = new PlainShader();
	AutoPointer<ColourMaterial> material = new ColourMaterial();
	material->setAmbient(Colour<float>(0.1f, 0.1f, 0.1f));
	material->setDiffuse(Colour<float>(0.41f, 0.41f, 0.52f));
	material->setSpecular(Colour<float>(1.0f, 1.0f, 1.0f));
	material->setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
	material->setShininess(20.0f);

	size_t total = 0;
	for (unsigned int i = 0; i < count; i++) {
		Surface *surf = m_Surfaces.append();
		size_t read = m_In->read(surf, sizeof(Surface));
		if (read != sizeof(Surface))
			throw magnum::io::TranslatorFormatException();
		total += read;

		// Create empty mesh			
		AutoPointer<IndexMesh> mesh = new IndexMesh();
		AutoPointer<Appearance> app = new Appearance();
		app->attachShader(shader);
		app->attachMaterial(material);
		mesh->setName(surf->name);
		mesh->attachAppearance(app);
		model->insertMesh(mesh);

		// Allocate space for vertex- and texture data
		unsigned int vstart = m_Vertices.getSize();

		// Get next chunk to read
		Heap<int> ofHeap;
		ofHeap.insert(-(int)surf->offsetTriangles);
		ofHeap.insert(-(int)surf->offsetVertices);
		ofHeap.insert(-(int)surf->offsetCollapseMap);
		ofHeap.insert(-(int)surf->offsetBoneRefs);
		size_t position = sizeof(Surface);
		size_t skip = 0;
		for (unsigned int i= 0; i < 4; i++) {
			size_t nextPos = -ofHeap.remove();
			if (nextPos == 0)
				continue;
			if (nextPos < position)
				throw magnum::io::TranslatorFormatException();

			m_In->skip(nextPos - position);
			position += nextPos - position;

			if (nextPos == surf->offsetCollapseMap) {
				skip = readCollapseMap(surf->numVertices);	
				surf->offsetCollapseMap = 0;
			} else if (nextPos == surf->offsetBoneRefs) {
				skip = readBoneRefs(surf->numBoneRefs);	
				surf->offsetBoneRefs = 0;
			} else if (nextPos == surf->offsetTriangles) {
				skip = readTriangles(mesh, vstart, surf->numTriangles);
				surf->offsetTriangles = 0;
			} else if (nextPos == surf->offsetVertices) {
				skip = readVertices(weights, surf->numVertices);
				surf->offsetVertices = 0;
			}

			total += skip;
			position += skip;
		}
	}
	
	return total;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMDS::readBoneRefs(unsigned int count) {
	return (size_t)m_In->skip(count * sizeof(int32));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMDS::readCollapseMap(unsigned int count) {
	return (size_t)m_In->skip(count * sizeof(int32));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMDS::readFrames(unsigned int numbones,unsigned int count) {
	size_t total = 0;
	
	for (unsigned int i = 0; i < count; i++) {
		// Read frame header
		FrameExt* frame = m_Frames.append();
		size_t read = m_In->read(frame, sizeof(Frame));
		if (read != sizeof(Frame))
			throw magnum::io::TranslatorFormatException();
		total += read;
		
		// Read compressed bone data
		for (unsigned int j = 0; j < numbones; j++) {
			CompressedBoneFrame* bone = &frame->bones[frame->bones.append()];
			read = m_In->read(bone, sizeof(CompressedBoneFrame));
			if (read != sizeof(CompressedBoneFrame))
				throw magnum::io::TranslatorFormatException();
			total += read;
		}
	}
	
	return total;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMDS::readTriangles(IndexMesh* mesh, unsigned int offset, unsigned int count) {
	Array<unsigned int> indices;
	indices.setSize(3*count);
	size_t read;
	read = m_In->read(&indices[0], 3*count*sizeof(int));
	if (read != 3*count*sizeof(int))
		throw magnum::io::TranslatorFormatException();
	
	AutoPointer<IndexBuffer> ib = IndexBuffer::create(IndexBuffer::Index32, 3*count, IndexBuffer::Static);
	mesh->attachIndexBuffer(ib);
	unsigned int* dst = (unsigned int*)ib->lock(IndexBuffer::Discard);
	unsigned int* src = &indices[0];
	for (unsigned int i = 0; i < count; i++) {
		dst[0] = src[0] + offset;
		dst[1] = src[1] + offset;
		dst[2] = src[2] + offset;
		dst += 3;
		src += 3;
	}
	ib->unlock();
	
	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorMDS::readVertices(WeightBuffer* wb,unsigned int count) {
	size_t total = 0;
	unsigned int vstart = m_Vertices.append(count);
	StaticVertex* vcoords = &m_Vertices[vstart];
	
	for (unsigned int i = 0; i < count; ++i) {
		Vertex v;
		size_t read = m_In->read(&v, sizeof(Vertex));
		if (read != sizeof(Vertex))
			throw magnum::io::TranslatorFormatException();
		
		// Save static part	
		*vcoords++ = v;
		
		// Read in weights
		Weight w[20];
		ASSERT(v.numWeights <= 20);
		read = m_In->read(w, v.numWeights*sizeof(Weight));
		if (read != v.numWeights*sizeof(Weight))
			throw magnum::io::TranslatorFormatException();
		
		// Rotate into correct coordinate system
		Vector3<float> normal(-v.normal.y, v.normal.z, v.normal.x);
		for (unsigned int i = 0; i < v.numWeights; i++) {
			float tmp = w[i].position.x;
			w[i].position.x = -w[i].position.y;
			w[i].position.y = w[i].position.z;
			w[i].position.z = tmp;
		}
		
		// Insert weights into weight buffer
		wb->insert(v.numWeights, normal, w);
			
		total += read + sizeof(Vertex);;
	}
	
	return total;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMDS::createAnimations(BoneModel* model) {
	AutoPointer<BoneAnimation> anim = new BoneAnimation(model->getSkeleton());
	Skeleton* skel = model->getSkeleton();
	Vector3<float> poffset[256];
	
	unsigned int time = 0;
	for (List<FrameExt>::Iterator iter = m_Frames.begin(); iter != m_Frames.end(); ++iter,++time) {
		BoneAnimation::Frame frame;
		frame.setSize(iter->bones.getSize());
		frame.setTime((float)time / 10.0f);
		for (Skeleton::Iterator bone = skel->begin(); bone != skel->end(); ++bone) {
			unsigned int i = bone->getIndex();
		    const float scaleRad = 2.0f * (float)M_PI / 65536.0f; // in Radiants
			float pitch = iter->bones[i].angles[0] * scaleRad;
			float yaw   = iter->bones[i].angles[1] * scaleRad;
			float roll  = iter->bones[i].angles[2] * scaleRad;

			Transform3<float> trans;
			trans.setRotationY(yaw);
			trans.rotateX(pitch);
			trans.rotateZ(roll);
			frame[i].rotation.fromMatrix(trans.getRotation());

			Transform3<float> mtx;
			mtx.setRotationX(iter->bones[i].offsetAngles[0] * scaleRad);
			mtx.rotateY(-iter->bones[i].offsetAngles[1] * scaleRad);

			Vector3<float> vec = bone->getDistance() * Vector3<float>(mtx[0][2], -mtx[1][2], mtx[2][2]);
			if (bone->getParent() == NULL) {
				vec += Vector3<float>(-iter->parentOffset.y,iter->parentOffset.z,iter->parentOffset.x);
			} else {
				vec += poffset[bone->getParent()->getIndex()];
			}
			poffset[i] = vec;
			frame[i].offset = vec;
		}
		anim->insert(frame);
	}
	
	anim->finish();
	model->insertAnimation(anim);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMDS::createVertices(BoneModel* model) {
	// Setup vertex format
	VertexFormat fmt;
	fmt.set(0, Types::Float2, Attributes::TextureAll);

	// Create vertex buffer
	AutoPointer<VertexBuffer> vb = VertexBuffer::create(fmt, m_Vertices.getSize(), VertexBuffer::Static);
	
	// Copy vertices into vertex buffer
	float* dst = (float*)vb->lock(VertexBuffer::Discard);
	for (unsigned int i = 0; i < m_Vertices.getSize(); i++) {
		*dst++ = m_Vertices[i].texCoords.x;
		*dst++ = m_Vertices[i].texCoords.y;
	}
	vb->unlock();
	
	// Attach vertex buffer to all meshes
	MeshList mlist = model->getMeshes();
	for (MeshList::Iterator iter = mlist.begin(); iter != mlist.end(); ++iter)
		((IndexMesh*)*iter)->attachVertexBuffer(vb);
}


}; };
