/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "io.datainput.h"
#include "io.exception.h"
#include "io.sysfile.h"
#include "io.inputbuffer.h"
#include "graphics3d.context.h"
#include "graphics3d.exception.h"
#include "graphics3d.vertexformat.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.appearance.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.plainshader.h"
#include "graphics3d.colourshader.h"
#include "graphics3d.textureshader.h"
#include "graphics3d.colourmaterial.h"
#include "graphics3d.staticmodel.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.shadermanager.h"
#include "graphics3d.texturemanager.h"
#include "graphics3d.translator.mobj.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "base.string.inl"
#include "math.vector2.inl"
#include "math.vector3.inl"
#include "container.iterator.inl"
#include "container.array.inl"
#include "container.map.inl"
#include "container.list.inl"
#include "container.cache.inl"
#include "container.allocator.inl"
#include "io.datainput.inl"
#include "io.translator.inl"
#include "io.stream.inl"
#include "io.inputstream.inl"
#include "io.inputbuffer.inl"
#include "io.file.inl"
#include "io.sysfile.inl"
#include "graphics.colour.inl"
#include "graphics3d.context.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.material.inl"
#include "graphics3d.colourmaterial.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.model.inl"
#include "graphics3d.texturemanager.inl"
#include "graphics3d.texturefactory.inl"
#include "graphics3d.translator.inl"
#include "graphics3d.staticmodel.inl"


namespace magnum { namespace graphics3d {
using magnum::object::AutoPointer;
using magnum::io::DataInput;
using magnum::io::SystemFile;
using magnum::io::InputBuffer;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void TranslatorOBJ::adjustVertex(int& i) const {
	if (i > 0)
		i--; 
	else
		i += m_Vertices.getSize();	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void TranslatorOBJ::adjustTexture(int& i) const {
	if (i > 0)
		i--; 
	else
		i += m_Textures.getSize();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline void TranslatorOBJ::adjustNormal(int& i) const {
	if (i > 0)
		i--; 
	else
		i += m_Normals.getSize();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::mtlReadMaterial(const String& line) {
	char name[256];
	sscanf(line,"newmtl %s",name);
	m_Material = new ColourMaterial();
	m_Material->setEmission(Colour<float>(0,0,0));
	m_Appearance = new Appearance();
	m_Appearance->attachMaterial(m_Material);
	m_Materials.insert(name, m_Appearance);
	m_Material->release();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::mtlReadIllum(const String& line) {	
	int d;
	sscanf(line,"illum %d",&d);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::mtlReadDiffuse(const String& line) {	
	float r,g,b;
	sscanf(line,"Kd %f %f %f",&r,&g,&b);
	m_Material->setDiffuse(Colour<float>(r,g,b));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::mtlReadSpecular(const String& line) {	
	float r,g,b;
	sscanf(line,"Ks %f %f %f",&r,&g,&b);
	m_Material->setSpecular(Colour<float>(r,g,b));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::mtlReadAmbient(const String& line) {
	float r,g,b;
	sscanf(line,"Ka %f %f %f",&r,&g,&b);
	m_Material->setAmbient(Colour<float>(r,g,b));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::mtlReadHighlight(const String& line) {
	float h;
	sscanf(line,"Ns %f",&h);
	m_Material->setShininess(h);;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::mtlReadDiffuseMap(const String& line) {
	if (!m_Appearance)
		return;
		
	char name[256];
	sscanf(line,"map_Kd %s",name);
	
	try {
		AutoPointer<Texture> texture = createTexture(name);
		if (texture == NULL)
			return;

		texture->setFilter(Texture::LinearMipmapLinear, Texture::Linear);
		m_Appearance->attachTexture(texture);
	} catch (Exception ex) {

	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::mtlReadMain(DataInput& input) {
	while(input.available() > 0) {
		String line = input.readLine();
		if (line.getLength() == 0)
			continue;
			
		String word;
		const char* ptr = &line[0];
		while((*ptr == ' ') || (*ptr  == '\t'))
			ptr++;
		line = ptr;
		ptr = &line[0];
		while(*ptr != 0) {
			if ((*ptr == ' ') || (*ptr == '\t'))
				break;
			word += *ptr++;
		}
		
		if (word.getLength() == 0)
			continue;
		if (word[0] == '#')
			continue;
		else if (word == "newmtl")
			mtlReadMaterial(line);
		else if (word == "illum")
			mtlReadIllum(line);
		else if (word == "Kd")
			mtlReadDiffuse(line);
		else if (word == "Ks")
			mtlReadSpecular(line);
		else if (word == "Ka")
			mtlReadAmbient(line);
		else if (word == "map_Kd")
			mtlReadDiffuseMap(line);
		else if (word == "Ns")
			mtlReadHighlight(line);
		else if (word == "d")
			continue;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::readVertex(const String& line) {
	float x,y,z;
	sscanf(line, "v %f %f %f", &x, &y, &z);
	m_Vertices.insert(Vector3<float>(x,y,z));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::readFace(const String& line) {
	if (!m_Triangles)
		return;
	Triangle* tri = &(*m_Triangles)[m_Triangles->append()];
	tri->m_Appearance = m_Appearance;
	
	int slashes = 0;
	const char* ptr = line;
	while(*ptr != 0) {
		if (*ptr++ == '/')
			slashes++;
	}
	
	if (slashes == 0) {
		sscanf(line, "f %d %d %d", &tri->m_v1, &tri->m_v2, &tri->m_v3);

		// Adjust vertex indices
		adjustVertex(tri->m_v1);
		adjustVertex(tri->m_v2);
		adjustVertex(tri->m_v3);

		// Set other indices to -1
		tri->m_n1 = tri->m_n2 = tri->m_n3 = -1;
		tri->m_t1 = tri->m_t2 = tri->m_t3 = -1;
	} else if (slashes == 3) {
		sscanf(line, "f %d/%d %d/%d %d/%d", 
			&tri->m_v1, &tri->m_t1,
			&tri->m_v2, &tri->m_t2,
			&tri->m_v3, &tri->m_t3);

		// Adjust vertex indices
		adjustVertex(tri->m_v1);
		adjustVertex(tri->m_v2);
		adjustVertex(tri->m_v3);

		// Adjust texture indices
		adjustTexture(tri->m_t1);
		adjustTexture(tri->m_t2);
		adjustTexture(tri->m_t3);

		// Set other indices to -1
		tri->m_n1 = tri->m_n2 = tri->m_n3 = -1;
	} else if (strstr(line,"//")) {
		sscanf(line, "f %d//%d %d//%d %d//%d", 
			&tri->m_v1, &tri->m_n1,
			&tri->m_v2, &tri->m_n2,
			&tri->m_v3, &tri->m_n3);

		// Adjust vertex indices
		adjustVertex(tri->m_v1);
		adjustVertex(tri->m_v2);
		adjustVertex(tri->m_v3);

		// Adjust normal indices
		adjustNormal(tri->m_n1);
		adjustNormal(tri->m_n2);
		adjustNormal(tri->m_n3);

		// Set other indices to -1
		tri->m_t1 = tri->m_t2 = tri->m_t3 = -1;
	} else {
		sscanf(line, "f %d/%d/%d %d/%d/%d %d/%d/%d", 
			&tri->m_v1, &tri->m_t1, &tri->m_n1,
			&tri->m_v2, &tri->m_t2, &tri->m_n2,
			&tri->m_v3, &tri->m_t3, &tri->m_n3);

		// Adjust vertex indices
		adjustVertex(tri->m_v1);
		adjustVertex(tri->m_v2);
		adjustVertex(tri->m_v3);

		// Adjust texture indices
		adjustTexture(tri->m_t1);
		adjustTexture(tri->m_t2);
		adjustTexture(tri->m_t3);

		// Adjust normal indices
		adjustNormal(tri->m_n1);
		adjustNormal(tri->m_n2);
		adjustNormal(tri->m_n3);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::readUV(const String& line) {
	float x,y,z;
	sscanf(line, "vt %f %f %f", &x, &y, &z);
	m_Textures.insert(Vector3<float>(x,y,z));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::readNormal(const String& line) {
	float x,y,z;
	sscanf(line, "vn %f %f %f", &x, &y, &z);
	m_Normals.insert(Vector3<float>(x,y,z));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::readObject(const String& line) {
	m_Object = m_Objects.append();
	m_Triangles = &m_Object->m_Triangles;
	m_Object->m_Name = &line[2];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::readGroup(const String& line) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::readMaterial(const String& line) {
	if (!m_Object)
		return;
	char name[255];	
	sscanf(line, "usemtl %s", name);
	Map<String, Appearance*>::Iterator iter = m_Materials.at(name);
	if (iter == m_Materials.end()) 
		m_Appearance = NULL;
	else
		m_Appearance = *iter;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::readMaterialLibrary(const String& line) {
	char fname[255];	
	sscanf(line, "mtllib %s", fname);

	File* file;
	try {
		if (getTextureManager() != NULL)
			file = getTextureManager()->getFileFactory()->open(fname);
		else
			file = SystemFile::open(fname);
	} catch(...) {
		return;
	}
	if (file == NULL)
		return;
	
	InputBuffer buffer(file->getInputStream());
	DataInput input(buffer);
	mtlReadMain(input);
	file->release();

	// Attach appropriate shaders to all appearances
	AutoPointer<PlainShader> plainShader = new PlainShader();
	AutoPointer<TextureShader> textureShader = new TextureShader();
	for(Map<String, Appearance*>::Iterator iter = m_Materials.begin(); iter != m_Materials.end(); ++iter) {
		if (*iter == NULL)
			continue;
		if ((*iter)->getTexture() != NULL)
			(*iter)->attachShader(textureShader);
		else
			(*iter)->attachShader(plainShader);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::readMain(DataInput& input) {
	while(input.available() > 0) {
		String line = input.readLine();
		if (line.getLength() == 0)
			continue;
			
		String word;
		const char* ptr = &line[0];
		while((*ptr == ' ') || (*ptr  == '\t'))
			ptr++;
		line = ptr;
		ptr = &line[0];
		while(*ptr != 0) {
			if ((*ptr == ' ') || (*ptr == '\t'))
				break;
			word += *ptr++;
		}
		
		if (word.getLength() == 0)
			continue;
		if (word[0] == '#')
			continue;
		else if (word == "v")
			readVertex(line);
		else if (word == "f")
			readFace(line);
		else if (word == "vt")
			readUV(line);
		else if (word == "vn")
			readNormal(line);
		else if (word == "o")
			readObject(line);
		else if (word == "g")
			readGroup(line);
		else if (word == "s")
			continue;
		else if (word == "usemtl")
			readMaterial(line);
		else if (word == "mtllib")
			readMaterialLibrary(line);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::createNormals() {
	Map<unsigned int,unsigned int> normals;
	int firstnormal = m_Normals.getSize();
	
	// Iterate over all objects
	for (List<Object>::Iterator obj = m_Objects.begin(); obj != m_Objects.end(); ++obj) {
		// Iteratate over all triangles in the current object
		for (Array<Triangle>::Iterator tri = obj->m_Triangles.begin(); tri != obj->m_Triangles.end(); ++tri) {
			// Insert vertex indices into maps
			if (tri->m_n1 == -1) {
				Map<unsigned int,unsigned int>::Iterator n = normals.at(tri->m_v1);
				if (n == normals.end()) {
					tri->m_n1 = m_Normals.insert(Vector3<float>::Null);
					normals.insert(tri->m_v1, tri->m_n1);
				} else {
					tri->m_n1 = *n;
				}
			}

			if (tri->m_n2 == -1) {
				Map<unsigned int,unsigned int>::Iterator n = normals.at(tri->m_v2);
				if (n == normals.end()) {
					tri->m_n2 = m_Normals.insert(Vector3<float>::Null);
					normals.insert(tri->m_v2, tri->m_n2);
				} else {
					tri->m_n2 = *n;
				}
			}

			if (tri->m_n3 == -1) {
				Map<unsigned int,unsigned int>::Iterator n = normals.at(tri->m_v3);
				if (n == normals.end()) {
					tri->m_n3 = m_Normals.insert(Vector3<float>::Null);
					normals.insert(tri->m_v3, tri->m_n3);
				} else {
					tri->m_n3 = *n;
				}
			}
		}
	}
	
	// Create normals
	for (List<Object>::Iterator obj = m_Objects.begin(); obj != m_Objects.end(); ++obj) {
		// Iteratate over all triangles in the current object
		for (Array<Triangle>::Iterator tri = obj->m_Triangles.begin(); tri != obj->m_Triangles.end(); ++tri) {
			Vector3<float> v1 = m_Vertices[tri->m_v1];
			Vector3<float> v2 = m_Vertices[tri->m_v2];
			Vector3<float> v3 = m_Vertices[tri->m_v3];
			Vector3<float> normal = (v2-v1) % (v3-v1);

			if (tri->m_n1 >= firstnormal)
				m_Normals[tri->m_n1] += normal;
			if (tri->m_n2 >= firstnormal)
				m_Normals[tri->m_n2] += normal;
			if (tri->m_n3 >= firstnormal)
				m_Normals[tri->m_n3] += normal;
		}
	}
	
	// Normalize normals
	for (unsigned int i = firstnormal; i < m_Normals.getSize(); i++) {
		float l = m_Normals[i].length();
		if (l > 0.0f)
			m_Normals[i] /= l;
	}
	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOBJ::createVertexMapping(Map<Vertex,unsigned int>& simpMap, Map<Vertex,unsigned int>& texMap) {
	// Iterate over all objects
	for (List<Object>::Iterator obj = m_Objects.begin(); obj != m_Objects.end(); ++obj) {
		// Iteratate over all triangles in the current object
		for (Array<Triangle>::Iterator tri = obj->m_Triangles.begin(); tri != obj->m_Triangles.end(); ++tri) {
			// Insert vertex indices into maps
			Vertex vtx;
			if (tri->m_t1 == -1) {
				vtx.x = tri->m_v1;
				vtx.y = tri->m_n1;
				vtx.z = -1;
				if (!simpMap.contains(vtx))
					simpMap.insert(vtx, simpMap.getSize());

				vtx.x = tri->m_v2;
				vtx.y = tri->m_n2;
				vtx.z = -1;
				if (!simpMap.contains(vtx))
					simpMap.insert(vtx, simpMap.getSize());

				vtx.x = tri->m_v3;
				vtx.y = tri->m_n3;
				vtx.z = -1;
				if (!simpMap.contains(vtx))
					simpMap.insert(vtx, simpMap.getSize());
			} else {
				vtx.x = tri->m_v1;
				vtx.y = tri->m_n1;
				vtx.z = tri->m_t1;
				if (!texMap.contains(vtx))
					texMap.insert(vtx, texMap.getSize());

				vtx.x = tri->m_v2;
				vtx.y = tri->m_n2;
				vtx.z = tri->m_t2;
				if (!texMap.contains(vtx))
					texMap.insert(vtx, texMap.getSize());

				vtx.x = tri->m_v3;
				vtx.y = tri->m_n3;
				vtx.z = tri->m_t3;
				if (!texMap.contains(vtx))
					texMap.insert(vtx, texMap.getSize());
			}
		}
	}
}


		
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
VertexBuffer* TranslatorOBJ::createSimpVertexBuffer(const Map<Vertex,unsigned int>& simpMap) {
	VertexBuffer* simpBuf = NULL;
	if (simpMap.getSize() > 0) {
		// Create versimp and index buffers
		VertexFormat fmt;
		fmt.set(0, Types::Float3, Attributes::Vertex);
		fmt.set(1, Types::Float3, Attributes::Normal);
		simpBuf = VertexBuffer::create(fmt, simpMap.getSize(), VertexBuffer::Static);
		if (simpBuf == NULL)
			throw MemoryException();
			
		// Fill versimp buffer
		float* buf = (float*)simpBuf->lock(VertexBuffer::Discard);
		for (Map<Vertex,unsigned int>::Iterator iter = simpMap.begin(); iter != simpMap.end(); ++iter) {
			float* ptr = &buf[6*iter.item()];
			ptr[0] = m_Vertices[iter.key().x].x;
			ptr[1] = m_Vertices[iter.key().x].y;
			ptr[2] = m_Vertices[iter.key().x].z;
			ptr[3] = m_Normals[iter.key().y].x;
			ptr[4] = m_Normals[iter.key().y].y;
			ptr[5] = m_Normals[iter.key().y].z;
		}
		simpBuf->unlock();
	}
	return simpBuf;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
VertexBuffer* TranslatorOBJ::createTexVertexBuffer(const Map<Vertex,unsigned int>& texMap) {
	VertexBuffer* texBuf = NULL;
	if (texMap.getSize() > 0) {
		// Create vertex and index buffers
		VertexFormat fmt;
		fmt.set(0, Types::Float3, Attributes::Vertex);
		fmt.set(1, Types::Float3, Attributes::Normal);
		fmt.set(2, Types::Float2, Attributes::TextureAll);
		texBuf = VertexBuffer::create(fmt, texMap.getSize(), VertexBuffer::Static);
		if (texBuf == NULL)
			throw MemoryException();
			
		// Fill vertex buffer
		float* buf = (float*)texBuf->lock(VertexBuffer::Discard);
		for (Map<Vertex,unsigned int>::Iterator iter = texMap.begin(); iter != texMap.end(); ++iter) {
			float* ptr = &buf[8*iter.item()];
			ptr[0] = m_Vertices[iter.key().x].x;
			ptr[1] = m_Vertices[iter.key().x].y;
			ptr[2] = m_Vertices[iter.key().x].z;
			ptr[3] = m_Normals[iter.key().y].x;
			ptr[4] = m_Normals[iter.key().y].y;
			ptr[5] = m_Normals[iter.key().y].z;
			ptr[6] = m_Textures[iter.key().z].x;
			ptr[7] = 1.0f - m_Textures[iter.key().z].y;
		}
		texBuf->unlock();
	}
	return texBuf;
}


void TranslatorOBJ::createIndices(Map<Appearance*,Array<unsigned int> >& smap, const Map<Vertex,unsigned int>& simpMap,
			Map<Appearance*,Array<unsigned int> >& tmap, const Map<Vertex,unsigned int>& texMap) {	
	for (List<Object>::Iterator obj = m_Objects.begin(); obj != m_Objects.end(); ++obj) {
		// Iteratate over all triangles in the current object
		for (Array<Triangle>::Iterator tri = obj->m_Triangles.begin(); tri != obj->m_Triangles.end(); ++tri) {
			// Insert vertex indices into maps
			if (tri->m_t1 == -1) {
				Array<unsigned int>* indices;
				Map<Appearance*,Array<unsigned int> >::Iterator it = smap.at(tri->m_Appearance);
				if (it == smap.end()) {
					indices = smap.insert(tri->m_Appearance);
				} else {
					indices = it;
				}
				
				Vertex vtx;
				vtx.x = tri->m_v1;
				vtx.y = tri->m_n1;
				vtx.z = -1;
				indices->insert(simpMap[vtx]);

				vtx.x = tri->m_v2;
				vtx.y = tri->m_n2;
				vtx.z = -1;
				indices->insert(simpMap[vtx]);

				vtx.x = tri->m_v3;
				vtx.y = tri->m_n3;
				vtx.z = -1;
				indices->insert(simpMap[vtx]);
			} else {
				Array<unsigned int>* indices;
				Map<Appearance*,Array<unsigned int> >::Iterator it = tmap.at(tri->m_Appearance);
				if (it == tmap.end()) {
					indices = tmap.insert(tri->m_Appearance);
				} else {
					indices = it;
				}

				Vertex vtx;
				vtx.x = tri->m_v1;
				vtx.y = tri->m_n1;
				vtx.z = tri->m_t1;
				indices->insert(texMap[vtx]);

				vtx.x = tri->m_v2;
				vtx.y = tri->m_n2;
				vtx.z = tri->m_t2;
				indices->insert(texMap[vtx]);

				vtx.x = tri->m_v3;
				vtx.y = tri->m_n3;
				vtx.z = tri->m_t3;
				indices->insert(texMap[vtx]);
			}
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
StaticModel* TranslatorOBJ::createModel() {
	Map<Vertex,unsigned int> simpMap;
	Map<Vertex,unsigned int> texMap;

	createNormals();
	createVertexMapping(simpMap, texMap);
	AutoPointer<VertexBuffer> simpBuf = createSimpVertexBuffer(simpMap);
	AutoPointer<VertexBuffer> texBuf = createTexVertexBuffer(texMap);

	// Iterate over all objects
	Map<Appearance*,Array<unsigned int> > tmap;
	Map<Appearance*,Array<unsigned int> > smap;
	createIndices(smap, simpMap, tmap, texMap);
	
	// Create default appearances for mesh
	AutoPointer<Appearance> papp = new Appearance();
	AutoPointer<Shader> pshader = new PlainShader();
	papp->attachShader(pshader);

	AutoPointer<Appearance> tapp = new Appearance();
	AutoPointer<Shader> tshader = new TextureShader();
	tapp->attachShader(tshader);
	
	// Create material
	AutoPointer<ColourMaterial> material = new ColourMaterial();
	material->setAmbient(Colour<float>(0.1f, 0.1f, 0.1f));
	material->setDiffuse(Colour<float>(0.41f, 0.41f, 0.52f));
	material->setSpecular(Colour<float>(1.0f, 1.0f, 1.0f));
	material->setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
	material->setShininess(20.0f);
	papp->attachMaterial(material);
	tapp->attachMaterial(material);

	// Create meshes
	StaticModel* model = new StaticModel();
	
	// Insert simple meshes
	for (Map<Appearance*,Array<unsigned int> >::Iterator it = smap.begin(); it != smap.end(); ++it) {
		AutoPointer<IndexBuffer> ib = IndexBuffer::create(IndexBuffer::Index32, it->getSize(), IndexBuffer::Static);
		if (ib == NULL)
			throw MemoryException();
		ib->setBuffer(0, &(*it)[0], it->getSize());
		AutoPointer<IndexMesh> mesh = new IndexMesh();
		mesh->setPrimitive(Mesh::TriangleList);
		if (it.key())
			mesh->attachAppearance(it.key());
		else
			mesh->attachAppearance(papp);
		mesh->attachVertexBuffer(simpBuf);
		mesh->attachIndexBuffer(ib);
		mesh->updateBounds();
		model->insertMesh(mesh);
	}

	// Insert textured meshes
	for (Map<Appearance*,Array<unsigned int> >::Iterator it = tmap.begin(); it != tmap.end(); ++it) {
		AutoPointer<IndexBuffer> ib = IndexBuffer::create(IndexBuffer::Index32, it->getSize(), IndexBuffer::Static);
		if (ib == NULL)
			throw MemoryException();
		ib->setBuffer(0, &(*it)[0], it->getSize());
		AutoPointer<IndexMesh> mesh = new IndexMesh();
		mesh->setPrimitive(Mesh::TriangleList);
		if (it.key())
			mesh->attachAppearance(it.key());
		else
			mesh->attachAppearance(tapp);
		mesh->attachVertexBuffer(texBuf);
		mesh->attachIndexBuffer(ib);
		mesh->updateBounds();
		model->insertMesh(mesh);
	}
	
	model->finish();
	
	return model;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Model* TranslatorOBJ::onDecode(InputStream* in,size_t size) {
	if (size == Autosize)
		size = (size_t)in->available();
		
	InputBuffer buffer(in);
	DataInput input(buffer);

	// Create dummy object
	m_Object = m_Objects.append();
	m_Triangles = &m_Object->m_Triangles;
	m_Appearance = new Appearance();
	m_Appearance->attachShader(new PlainShader());
	m_Appearance->getShader()->release();
	m_Materials.insert("!magnum-default",m_Appearance);
	
	// Read file
	readMain(input);
	
	// Create model
	StaticModel* model = createModel();

	// Attach appropriate shaders to all appearances
	for(Map<String, Appearance*>::Iterator iter = m_Materials.begin(); iter != m_Materials.end(); ++iter) {
		if (*iter == NULL)
			continue;
		(*iter)->release();
	}
	
	// Release memory
	m_Materials.clear();
	m_Vertices.clear();
	m_Normals.clear();
	m_Textures.clear();
	m_Objects.clear();
	m_Triangles = NULL;	
	m_Object = NULL;
	m_Appearance = NULL;

	return model;	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorOBJ::onEncode(OutputStream* ,const Model* ) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorOBJ::onIdentify(File* ) {	
	return false;
}


}; };
