/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include <stdlib.h>

#include "io.exception.h"
#include "io.inputbuffer.h"
#include "graphics3d.translator.off.h"
#include "graphics3d.staticmodel.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.context.h"
#include "graphics3d.appearance.h"
#include "graphics3d.plainshader.h"
#include "graphics3d.colourmaterial.h"
#include "graphics3d.texturemanager.h"
#include "graphics3d.shadermanager.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "base.string.inl"
#include "container.array.inl"
#include "io.inputstream.inl"
#include "io.inputbuffer.inl"
#include "io.file.inl"
#include "io.translator.inl"
#include "math.vector3.inl"
#include "graphics3d.staticmodel.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.context.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.colourmaterial.inl"

namespace magnum { namespace graphics3d {
using magnum::object::AutoPointer;
using magnum::io::InputBuffer;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorOFF::TranslatorOFF() {
	m_VertexCount = 0;
	m_FaceCount = 0;
	
	m_Input = NULL;
	m_ReadColours = false;
	m_ReadNormals = false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
const char* TranslatorOFF::readToken() {
	m_LineFinished = false;
	char* cur = m_Token;
	int size = 0;
	for (;;) {
		char c;
		int ret = (int)m_Input->read(&c, 1);
		if (ret <= 0)
			break;
		if (c == '#') {
			finishLine();
			if (size > 0)
				break;
			continue;
		}
		if (c == ' ' || c == 0) {
			if (size > 0)
				break;
			continue;
		}
		if (c == '\n' || c == '\r') {
			if (size > 0) {
				m_LineFinished = true;
				break;
			}
			continue;
		}
		if (size >= sizeof(m_Token)-1)
			continue;
		*cur++ = c;
		size++;
	}

	m_Token[size] = 0;
	if (size > 0)
		return m_Token;
	throw magnum::io::TranslatorFormatException();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOFF::finishLine() {
	if (m_LineFinished) {
		m_LineFinished = false;
		return;
	}
	for (;;) {
		char c;
		size_t ret = m_Input->read(&c, 1);
		if (ret <= 0)
			break;
		if (c == '\n' || c == '\r')
			break;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOFF::readHeader() {
	// Read header
	readToken();
	if (strcmp(m_Token,"OFF"))
		throw magnum::io::TranslatorFormatException();

	readToken();
	m_VertexCount = atoi(m_Token);
	readToken();
	m_FaceCount = atoi(m_Token);
	readToken();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOFF::readVertices() {
	m_Vertices.setSize(m_VertexCount);
	if (m_ReadColours)
		m_Colours.setSize(m_VertexCount);
	if (m_ReadNormals)
		m_Normals.setSize(m_VertexCount);
		
	for (unsigned int i = 0; i < m_VertexCount; ++i) {
		readToken();
		float x = (float)atof(m_Token);
		readToken();
		float y = (float)atof(m_Token);
		readToken();
		float z = (float)atof(m_Token);
		m_Vertices[i] = Vector3<float>(x,y,z);		
		
		if (m_ReadNormals) {
			readToken();
			x = (float)atof(m_Token);
			readToken();
			y = (float)atof(m_Token);
			readToken();
			z = (float)atof(m_Token);
			m_Normals[i] = Vector3<float>(x,y,z);
		}
		if (m_ReadColours) {
			readToken();
			x = (float)atof(m_Token);
			readToken();
			y = (float)atof(m_Token);
			readToken();
			z = (float)atof(m_Token);
			m_Colours[i] = Colour<float>(x,y,z);
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOFF::readIndices() {
	m_Indices.setCapacity(3*m_FaceCount);

	for (unsigned int i = 0; i < m_FaceCount; ++i) {
		readToken();
		int count = atoi(m_Token);
		if (count < 3)
			throw magnum::io::TranslatorFormatException();

		readToken();
		int a = atoi(m_Token);
		m_Indices.insert(a);
		readToken();
		int b = atoi(m_Token);
		m_Indices.insert(b);
		readToken();
		int c = atoi(m_Token);
		m_Indices.insert(c);
		for (int i = 3; i < count; i++) {
			readToken();
			m_Indices.insert(a);
			m_Indices.insert(c);
			c = atoi(m_Token);
			m_Indices.insert(c);
		}
		finishLine();
	}	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorOFF::createNormals() {
	// Create normals
	m_Normals.setSize(m_VertexCount);
	unsigned int i;
	for(i = 0; i < m_VertexCount; i++) 
		m_Normals[i].null();
		
	for (i = 0; i < m_Indices.getSize() / 3; i++) {
		int vi1,vi2,vi3;
		vi1 = m_Indices[3*i];
		vi2 = m_Indices[3*i+1];
		vi3 = m_Indices[3*i+2];

		Vector3<float> v1 = m_Vertices[vi1];
		Vector3<float> v2 = m_Vertices[vi2];
		Vector3<float> v3 = m_Vertices[vi3];
		Vector3<float> normal = (v2-v1) % (v3-v1);

		m_Normals[vi1] += normal;
		m_Normals[vi2] += normal;
		m_Normals[vi3] += normal;
	}
	for (i = 0; i < m_VertexCount; i++) {
		float l = m_Normals[i].length();
		if (l > 0.0f)
			m_Normals[i] /= l;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Model* TranslatorOFF::onDecode(InputStream* in,size_t size) {
	m_Input = new InputBuffer(in);
	
	readHeader();
	readVertices();
	readIndices();
	
	m_Input->release();
	m_Input = NULL;
	
	if (!m_ReadNormals)
		createNormals();
	
	// Create vertex and index buffers
	VertexFormat fmt;
	fmt.set(0, Types::Float3, Attributes::Vertex);
	fmt.set(1, Types::Float3, Attributes::Normal);
	if (m_ReadColours)
		fmt.set(2, Types::Float3, Attributes::Colour);
	AutoPointer<VertexBuffer> vb = VertexBuffer::create(fmt, m_VertexCount, VertexBuffer::Static);
	Vector3<float>* vbmem = (Vector3<float>*)vb->lock(VertexBuffer::Discard);
	if (m_ReadColours) {
		for (unsigned int i = 0; i < m_VertexCount; i++) {
			vbmem[3*i] = m_Vertices[i];
			vbmem[3*i + 1] = m_Normals[i];
			vbmem[3*i + 2] = Vector3<float>(m_Colours[i].R, m_Colours[i].G, m_Colours[i].B);
		}
	} else {
		for (unsigned int i = 0; i < m_VertexCount; i++) {
			vbmem[2*i] = m_Vertices[i];
			vbmem[2*i + 1] = m_Normals[i];
		}
	}
	vb->unlock();
	
	// Create Tri-Mesh (if tris are present)
	AutoPointer<IndexBuffer> ib = IndexBuffer::create(IndexBuffer::Index32, m_Indices.getSize(), IndexBuffer::Static);
	ib->setBuffer(0, m_Indices.getArray(), m_Indices.getSize());

	// Create appearance for mesh
	AutoPointer<Appearance> appearance = new Appearance();
	AutoPointer<Shader> shader = new PlainShader();
	appearance->attachShader(shader);
	
	// Create material
	AutoPointer<ColourMaterial> material = new ColourMaterial();
	material->setAmbient(Colour<float>(0.1f, 0.1f, 0.1f));
	material->setDiffuse(Colour<float>(0.41f, 0.41f, 0.52f));
	material->setSpecular(Colour<float>(1.0f, 1.0f, 1.0f));
	material->setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
	material->setShininess(20.0f);
	appearance->attachMaterial(material);

	// Create new mesh
	AutoPointer<IndexMesh> mesh = new IndexMesh();
	mesh->setPrimitive(Mesh::TriangleList);
	mesh->attachAppearance(appearance);
	mesh->attachIndexBuffer(ib);
	mesh->attachVertexBuffer(vb,0);
	mesh->updateBounds();

	// Create model
	StaticModel* model = new StaticModel();
	model->insertMesh(mesh);
	model->finish();

	// Release memory
	m_Vertices.clear();
	m_Indices.clear();
	m_Normals.clear();
	
	return model;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorOFF::onEncode(OutputStream* ,const Model* ) {
	ASSERT(false);
	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorOFF::onIdentify(File* in) {
	return false;
}

}; };
