/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "io.datainput.h"
#include "io.exception.h"
#include "graphics3d.context.h"
#include "graphics3d.exception.h"
#include "graphics3d.vertexformat.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.appearance.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.plainshader.h"
#include "graphics3d.colourshader.h"
#include "graphics3d.colourmaterial.h"
#include "graphics3d.staticmodel.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.translator.ply.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "base.string.inl"
#include "math.vector2.inl"
#include "math.vector3.inl"
#include "container.iterator.inl"
#include "container.array.inl"
#include "container.allocator.inl"
#include "container.list.inl"
#include "io.datainput.inl"
#include "io.translator.inl"
#include "io.file.inl"
#include "io.stream.inl"
#include "io.inputstream.inl"
#include "io.inputbuffer.inl"
#include "graphics.colour.inl"
#include "graphics3d.context.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.colourmaterial.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.joint.inl"
#include "graphics3d.model.inl"
#include "graphics3d.staticmodel.inl"
#include "graphics3d.translator.ply.inl"

#include <stdlib.h>


namespace magnum { namespace graphics3d {
using magnum::object::AutoPointer;
using magnum::io::DataInput;
using magnum::types::msb2word;
using magnum::types::msb2dword;
using magnum::types::msb2qword;
using magnum::types::lsb2word;
using magnum::types::lsb2dword;
using magnum::types::lsb2qword;

const char* TranslatorPLY::m_TypeNames[] = {
	"invalid",
	"char", 
	"short", 
	"int",
	"uchar", 
	"ushort", 
	"uint",
	"float", 
	"double",
};

const unsigned int TranslatorPLY::m_TypeSizes[] = {
  0, 1, 2, 4, 1, 2, 4, 4, 8
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Model* TranslatorPLY::onDecode(InputStream* in,size_t size) {
	if (size == Autosize)
		size = (size_t)in->available();
		
	m_In.setStream(in);

	readHeader();
	installHandlers();
	readData();
	createNormals();
	
	AutoPointer<VertexBuffer> vb = createVertexBuffer();
	AutoPointer<IndexBuffer> ib = createIndexBuffer();
	
	// Create appearance for mesh
	AutoPointer<Appearance> appearance = new Appearance();
	if (m_Colours.getSize() > 0) {
		Shader* shader = new ColourShader();
		appearance->attachShader(shader);
		shader->release();
	} else {
		Shader* shader = new PlainShader();
		appearance->attachShader(shader);
		shader->release();
	}
	
	// Create material
	AutoPointer<ColourMaterial> material = new ColourMaterial();
	if (m_Colours.getSize() > 0) {
		material->setAmbient(Colour<float>(0.2f, 0.2f, 0.2f));
		material->setDiffuse(Colour<float>(1.0f, 1.0f, 1.0f));
		material->setSpecular(Colour<float>(0.2f, 0.2f, 0.2f));
		material->setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
		material->setShininess(20.0f);
	} else {
		material->setAmbient(Colour<float>(0.2f, 0.2f, 0.2f));
		material->setDiffuse(Colour<float>(0.7f, 0.7f, 0.5f));
		material->setSpecular(Colour<float>(1.0f, 1.0f, 1.0f));
		material->setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
		material->setShininess(20.0f);
	}
	appearance->attachMaterial(material);

	// Create new mesh
	AutoPointer<IndexMesh> mesh = new IndexMesh();
	mesh->setPrimitive(Mesh::TriangleList);
	mesh->attachAppearance(appearance);
	mesh->attachIndexBuffer(ib);
	mesh->attachVertexBuffer(vb,0);
	mesh->updateBounds();

	// Create model
	StaticModel* model = new StaticModel();
	//model->setName(m_Name);
	model->insertMesh(mesh);
	model->finish();

	// Release memory
	m_Vertices.clear();
	m_Indices.clear();
	m_Normals.clear();
	m_Colours.clear();

	return model;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
VertexBuffer* TranslatorPLY::createVertexBuffer() {
	VertexBuffer* vb;

	// Do we have colours?
	if (m_Colours.getSize() > 0) {
		// Create vertex and index buffers
		VertexFormat fmt;
		fmt.set(0, Types::Float3, Attributes::Vertex);
		fmt.set(1, Types::Float3, Attributes::Normal);
		fmt.set(2, Types::Float3, Attributes::Colour);
		vb = VertexBuffer::create(fmt, m_Vertices.getSize(), VertexBuffer::Static);
		if (vb == NULL)
			vb = VertexBuffer::create(fmt, m_Vertices.getSize(), VertexBuffer::System);
		if (vb == NULL)
			throw MemoryException();		

		Vector3<float>* vbmem = (Vector3<float>*)vb->lock(VertexBuffer::Discard);
		for (unsigned int i = 0; i < m_Vertices.getSize(); i++) {
			vbmem[3*i] = m_Vertices[i];
			vbmem[3*i + 1] = m_Normals[i];
			vbmem[3*i + 2] = Vector3<float>((float)m_Colours[3*i]/255.0f, (float)m_Colours[3*i+1]/255.0f, (float)m_Colours[3*i+2]/255.0f);
		}
		vb->unlock();
	} else {
		// Create vertex and index buffers
		VertexFormat fmt;
		fmt.set(0, Types::Float3, Attributes::Vertex);
		fmt.set(1, Types::Float3, Attributes::Normal);
		vb = VertexBuffer::create(fmt, m_Vertices.getSize(), VertexBuffer::Static);
		if (vb == NULL)
			vb = VertexBuffer::create(fmt, m_Vertices.getSize(), VertexBuffer::System);
		if (vb == NULL)
			throw MemoryException();		

		Vector3<float>* vbmem = (Vector3<float>*)vb->lock(VertexBuffer::Discard);
		for (unsigned int i = 0; i < m_Vertices.getSize(); i++) {
			vbmem[2*i] = m_Vertices[i];
			vbmem[2*i + 1] = m_Normals[i];
		}
		vb->unlock();
	}	
	
	return vb;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
IndexBuffer* TranslatorPLY::createIndexBuffer() {
	IndexBuffer* ib = IndexBuffer::create(IndexBuffer::Index32, m_Indices.getSize(), IndexBuffer::Static);
	if (ib == NULL)
		throw MemoryException();		
		
	ib->setBuffer(0, m_Indices.getArray(), m_Indices.getSize());
	return ib;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t TranslatorPLY::onEncode(OutputStream* ,const Model* ) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorPLY::onIdentify(File* ) {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPLY::createNormals() {
	// Create normals
	m_Normals.setSize(m_Vertices.getSize());
	unsigned int i;
	for(i = 0; i < m_Vertices.getSize(); i++) 
		m_Normals[i].null();
		
	for (i = 0; i < m_Indices.getSize() / 3; i++) {
		int vi1,vi2,vi3;
		vi1 = m_Indices[3*i];
		vi2 = m_Indices[3*i+1];
		vi3 = m_Indices[3*i+2];

		Vector3<float> v1 = m_Vertices[vi1];
		Vector3<float> v2 = m_Vertices[vi2];
		Vector3<float> v3 = m_Vertices[vi3];
		Vector3<float> normal = (v2-v1) % (v3-v1);

		m_Normals[vi1] += normal;
		m_Normals[vi2] += normal;
		m_Normals[vi3] += normal;
	}
	for (i = 0; i < m_Vertices.getSize(); i++) {
		float l = m_Normals[i].length();
		if (l > 0.0f)
			m_Normals[i] /= l;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPLY::readVertex(const String& str,DataType d) {
	*m_CurrentVertex++ = getItem<float>(str, d);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPLY::readColour(const String& str,DataType d) {
	*m_CurrentColour++ = getItem<byte>(str, d);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPLY::readFace(unsigned int size) {
	m_CurrentIndex = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPLY::readFaceIndex(const String& str,DataType d) {
	if (m_CurrentIndex == 0) {
		m_FirstIndex = getItem<unsigned int>(str, d);
		m_CurrentIndex++;
	} else if (m_CurrentIndex == 1) {
		m_LastIndex = getItem<unsigned int>(str, d);
		m_CurrentIndex++;
	} else {
		m_Indices.insert(m_FirstIndex);
		m_Indices.insert(m_LastIndex);
		m_LastIndex = getItem<unsigned int>(str, d);
		m_Indices.insert(m_LastIndex);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPLY::installHandlers() {
	addPropertyHandler("vertex", "x", &TranslatorPLY::readVertex);
	addPropertyHandler("vertex", "y", &TranslatorPLY::readVertex);
	addPropertyHandler("vertex", "z", &TranslatorPLY::readVertex);
	addListHandler("face", "vertex_indices", &TranslatorPLY::readFace);
	addPropertyHandler("face", "vertex_indices", &TranslatorPLY::readFaceIndex);

	if (findProperty("vertex","red") != NULL 
	&& findProperty("vertex","green") != NULL
	&& findProperty("vertex","blue") != NULL) {
		addPropertyHandler("vertex", "red", &TranslatorPLY::readColour);
		addPropertyHandler("vertex", "green", &TranslatorPLY::readColour);
		addPropertyHandler("vertex", "blue", &TranslatorPLY::readColour);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPLY::readHeader() {
	DataInput input(m_In);
	Array<String> words;
	String line;
	Element* elem = NULL;
	Property* prop = NULL;
	
	m_Elements.clear();
	m_Comments.clear();
	m_Infos.clear();
	
	line = input.readLine();
	if (line != "ply")
		throw magnum::io::TranslatorFormatException();

	while (true) {
		line = input.readLine();
		if (line.getLength() == 0)
			continue;
		getWords(line, words);
	
		/* parse words */
		if (words[0] == "format") {
			if (words.getSize() != 3)
				throw magnum::io::TranslatorFormatException();
			if (words[1] == "ascii")
				m_Type = ASCII;
			else if (words[1] == "binary_big_endian")
				m_Type = BinaryBE;
			else if (words[1] == "binary_little_endian")
				m_Type = BinaryLE;
			else
				throw magnum::io::TranslatorFormatException();
			m_Version = (float)atof (words[2]);
		}
		else if (words[0] == "element")
			elem = addElement(words);
		else if (words[0] == "property")
			prop = addProperty(elem, words);
		else if (words[0] == "comment")
			m_Comments.insert(line);
		else if (words[0] == "obj_info")
			m_Infos.insert(line);
		else if (words[0] == "end_header")
			break;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPLY::readData() {
	m_Normals.clear();
	m_Vertices.clear();
	m_Indices.clear();
	m_Colours.clear();
	
	// Get number of vertices
	Element* vertices = findElement("vertex");
	if (vertices != NULL) {
		m_Vertices.setSize(vertices->m_Size);
		m_CurrentVertex = &m_Vertices[0].x;
	} else {
		throw magnum::io::TranslatorFormatException();
	}
		
	// Get number of faces
	Element* indices = findElement("face");
	if (indices != NULL) {
		m_Indices.setCapacity(indices->m_Size*3);
	} else {
		throw magnum::io::TranslatorFormatException();
	}

	// Check if model has colour information
	if (findProperty("vertex","red") != NULL 
	&& findProperty("vertex","green") != NULL
	&& findProperty("vertex","blue") != NULL) {
		m_Colours.setSize(3 * vertices->m_Size);
		m_CurrentColour = &m_Colours[0];
	}
		
	// Iterate over all elements
	for (List<Element>::Iterator li = m_Elements.begin(); li != m_Elements.end(); ++li) {
		if (m_Type == ASCII) {
			DataInput input(m_In);
			Array<String> words;
			String line;
			for (unsigned int i = 0; i < li->m_Size; ++i) {
				line = input.readLine();
				getWords(line, words);
				getElementASCII((Element*)li, words);
			}
		} else {
			for (unsigned int i = 0; i < li->m_Size; ++i)
				getElementBinary((Element*)li);
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorPLY::Element* TranslatorPLY::findElement(const char* elem) const {
	for (List<Element>::Iterator iter = m_Elements.begin(); iter != m_Elements.end(); ++iter) {
		if (iter->m_Name == elem) {
			return (Element*)iter;
		}
	}
	
	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorPLY::Property* TranslatorPLY::findProperty(const char* elem,const char* prop) const {
	for (List<Element>::Iterator iter = m_Elements.begin(); iter != m_Elements.end(); ++iter) {
		if (iter->m_Name != elem)
			continue;
		for (Array<Property>::Iterator piter = iter->m_Properties.begin(); piter != iter->m_Properties.end(); ++piter) {
			if (piter->m_Name == prop) {
				return (Property*)piter;
			}		
		}
	}

	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPLY::addElementHandler(const char* elem, ElementHandler hnd) {
	for (List<Element>::Iterator iter = m_Elements.begin(); iter != m_Elements.end(); ++iter) {
		if (iter->m_Name == elem) {
			iter->m_Handler = hnd;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPLY::addPropertyHandler(const char* elem, const char* prop, PropertyHandler hnd) {
	for (List<Element>::Iterator iter = m_Elements.begin(); iter != m_Elements.end(); ++iter) {
		if (iter->m_Name != elem)
			continue;
		for (Array<Property>::Iterator piter = iter->m_Properties.begin(); piter != iter->m_Properties.end(); ++piter) {
			if (piter->m_Name == prop) {
				piter->m_Handler = hnd;
			}		
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPLY::addListHandler(const char* elem, const char* prop, ListHandler hnd) {
	for (List<Element>::Iterator iter = m_Elements.begin(); iter != m_Elements.end(); ++iter) {
		if (iter->m_Name != elem)
			continue;
		for (Array<Property>::Iterator piter = iter->m_Properties.begin(); piter != iter->m_Properties.end(); ++piter) {
			if (piter->m_Name == prop) {
				piter->m_Lister = hnd;
			}		
		}
	}
}


/*----------------------------------------------------------------------------
Return the type of a property, given the name of the property.

Entry:
  name - name of property type

Exit:
  returns integer code for property, or 0 if not found
----------------------------------------------------------------------------*/
TranslatorPLY::DataType TranslatorPLY::getPropertyType(const String& type_name) {
	for (int i = TypeNone + 1; i < TypeEnd; i++)
		if (type_name == m_TypeNames[i])
			return (DataType)i;

	/* if we get here, we didn't find the type */
	return TypeNone;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorPLY::Element* TranslatorPLY::addElement(const Array<String>& words) {
	Element* elem = &*m_Elements.append();
	elem->m_Name = words[1];
	elem->m_Size = atoi(words[2]);
	elem->m_Handler = NULL;
	return elem;
}


/*----------------------------------------------------------------------------
Add a property to a PLY file descriptor.

Entry:
  plyfile - PLY file descriptor
  words   - list of words describing the property
  nwords  - number of words in the list
----------------------------------------------------------------------------*/
TranslatorPLY::Property* TranslatorPLY::addProperty(TranslatorPLY::Element* elem, const Array<String>& words) {
	/* create the new property */
	Property* prop = &elem->m_Properties[elem->m_Properties.append()];
	if (words[1] == "list") {       /* is a list */
		prop->m_ListType = getPropertyType(words[2]);
		prop->m_Type = getPropertyType(words[3]);
		prop->m_Name = words[4];
		prop->m_Handler = NULL;
		prop->m_Lister = NULL;
	} else {                                        /* not a list */
		prop->m_ListType = TypeNone;
		prop->m_Type = getPropertyType(words[1]);
		prop->m_Name = words[2];
		prop->m_Handler = NULL;
		prop->m_Lister = NULL;
	}
	return prop;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorPLY::getWords(const char* in, Array<String>& words) {
	/* convert line-feed and tabs into spaces */
	/* (this guarentees that there will be a space before the */
	/*  null character at the end of the string) */

	String str;
	words.setSize(0);
	for (const char* ptr = in; *ptr != '\0'; ptr++) {
		if (*ptr == '\t')
			str += ' ';
		else if (*ptr == '\n')
			break;
		else
			str += *ptr;
	}
	str += ' ';
	
	/* find the words in the line */
	char* sptr = str;
	while (*sptr != '\0') {
	    while (*sptr == ' ')
			sptr++;

		/* break if we reach the end */
		if (*sptr == '\0')
			break;

		/* save pointer to beginning of word */
		const char* start = sptr;

		/* jump over non-spaces */
		while (*sptr != ' ')
			sptr++;

		*sptr++ = '\0';
		words.insert(String(start));
	}
}


/*----------------------------------------------------------------------------
Get the value of an item from a binary file, and place the result
into an integer, an unsigned integer and a double.
----------------------------------------------------------------------------*/
template<class T>
inline T TranslatorPLY::getItem(const String& str,DataType data) {
	if (m_Type == ASCII)
		return getItemASCII<T>(str, data);
	else
		return getItemBinary<T>(data);
}


/*----------------------------------------------------------------------------
Get the value of an item from a binary file, and place the result
into an integer, an unsigned integer and a double.
----------------------------------------------------------------------------*/
template<class T>
T TranslatorPLY::getItemBinary(DataType type) {
	char c[8],tmp;
	void *ptr = (void *) c;

	switch (type) {
    	case TypeChar:
    	  	m_In.read(ptr, 1);
    	  	return (T)*((char *) ptr);
    	case TypeUChar:
    	  	m_In.read(ptr, 1);
    	  	return (T)*((unsigned char *) ptr);
    	case TypeShort:
     	  	m_In.read(ptr, 2);
			#if defined(__LSBFIRST__)
     	  		if (m_Type == BinaryBE) {
     	  			tmp = c[1];
     	  			c[1] = c[0];
     	  			c[0] = tmp;
     	  		}
      		#elif defined(__MSBFIRST__)
     	  		if (m_Type == BinaryLE) {
     	  			tmp = c[1];
     	  			c[1] = c[0];
     	  			c[0] = tmp;
     	  		}
     	  	#endif
   	  		return (T)*((short int*)ptr);
    	case TypeUShort:
    	  	m_In.read(ptr, 2);
			#if defined(__LSBFIRST__)
     	  		if (m_Type == BinaryBE) {
     	  			tmp = c[1];
     	  			c[1] = c[0];
     	  			c[0] = tmp;
     	  		}
      		#elif defined(__MSBFIRST__)
     	  		if (m_Type == BinaryLE) {
     	  			tmp = c[1];
     	  			c[1] = c[0];
     	  			c[0] = tmp;
     	  		}
     	  	#endif
   	  		return (T)*((unsigned short int*)ptr);
    	case TypeInt:
    	  	m_In.read(ptr, 4);
			#if defined(__LSBFIRST__)
     	  		if (m_Type == BinaryBE) {
     	  			tmp = c[3];
     	  			c[3] = c[0];
     	  			c[0] = tmp;
     	  			tmp = c[1];
     	  			c[1] = c[2];
     	  			c[2] = tmp;
     	  		}
      		#elif defined(__MSBFIRST__)
     	  		if (m_Type == BinaryLE) {
     	  			tmp = c[3];
     	  			c[3] = c[0];
     	  			c[0] = tmp;
     	  			tmp = c[1];
     	  			c[1] = c[2];
     	  			c[2] = tmp;
     	  		}
     	  	#endif
   	  		return (T)*((int*)ptr);
    	case TypeUInt:
    	  	m_In.read(ptr, 4);
			#if defined(__LSBFIRST__)
     	  		if (m_Type == BinaryBE) {
     	  			tmp = c[3];
     	  			c[3] = c[0];
     	  			c[0] = tmp;
     	  			tmp = c[1];
     	  			c[1] = c[2];
     	  			c[2] = tmp;
     	  		}
      		#elif defined(__MSBFIRST__)
     	  		if (m_Type == BinaryLE) {
     	  			tmp = c[3];
     	  			c[3] = c[0];
     	  			c[0] = tmp;
     	  			tmp = c[1];
     	  			c[1] = c[2];
     	  			c[2] = tmp;
     	  		}
     	  	#endif
   	  		return (T)*((unsigned int*)ptr);
    	case TypeFloat:
    	  	m_In.read(ptr, 4);
			#if defined(__LSBFIRST__)
     	  		if (m_Type == BinaryBE) {
     	  			tmp = c[3];
     	  			c[3] = c[0];
     	  			c[0] = tmp;
     	  			tmp = c[1];
     	  			c[1] = c[2];
     	  			c[2] = tmp;
     	  		}
      		#elif defined(__MSBFIRST__)
     	  		if (m_Type == BinaryLE) {
     	  			tmp = c[3];
     	  			c[3] = c[0];
     	  			c[0] = tmp;
     	  			tmp = c[1];
     	  			c[1] = c[2];
     	  			c[2] = tmp;
     	  		}
     	  	#endif
    	  	return (T)*((float *) ptr);
    	case TypeDouble:
    	  	m_In.read(ptr, 4);
			#if defined(__LSBFIRST__)
     	  		if (m_Type == BinaryBE) {
     	  			tmp = c[7];
     	  			c[7] = c[0];
     	  			c[0] = tmp;
     	  			tmp = c[1];
     	  			c[1] = c[6];
     	  			c[6] = tmp;
     	  			tmp = c[2];
     	  			c[2] = c[5];
     	  			c[5] = tmp;
     	  			tmp = c[3];
     	  			c[3] = c[4];
     	  			c[4] = tmp;
     	  		}
      		#elif defined(__MSBFIRST__)
     	  		if (m_Type == BinaryLE) {
     	  			tmp = c[7];
     	  			c[7] = c[0];
     	  			c[0] = tmp;
     	  			tmp = c[1];
     	  			c[1] = c[6];
     	  			c[6] = tmp;
     	  			tmp = c[2];
     	  			c[2] = c[5];
     	  			c[5] = tmp;
     	  			tmp = c[3];
     	  			c[3] = c[4];
     	  			c[4] = tmp;
     	  		}
     	  	#endif
    	  	return (T)*((double *) ptr);
    	default:
    		ASSERT(false);
    		return T(0);
	}
}


/*----------------------------------------------------------------------------
Extract the value of an item from an ascii word, and place the result
into an integer, an unsigned integer and a double.
----------------------------------------------------------------------------*/
template<class T>
T TranslatorPLY::getItemASCII(const String& word,DataType type) {
	switch (type) {
    	case TypeChar:
    	case TypeUChar:
    	case TypeShort:
    	case TypeUShort:
    	case TypeInt:
    	  	return (T)atoi(word);
    	case TypeUInt:
    	  	return (T)strtoul (word, (char **) NULL, 10);
    	case TypeFloat:
    	case TypeDouble:
    	  	return (T)atof (word);
    	default:
    		ASSERT(false);
    		return T(0);
	}
}


/*----------------------------------------------------------------------------
Read an element from an ascii file.
----------------------------------------------------------------------------*/
void TranslatorPLY::getElementASCII(const TranslatorPLY::Element* elem, const Array<String>& words) {
	/* read in the element */
	if (words.getSize() == 0)
		throw magnum::io::TranslatorFormatException();

	unsigned int which_word = 0;
	for (Array<Property>::Iterator iter = elem->m_Properties.begin(); iter != elem->m_Properties.end(); ++iter) {
		Property& prop = *iter;

		if (prop.m_ListType != TypeNone) {       /* a list */
			/* get and store the number of items in the list */
			int count = getItemASCII<int>(words[which_word++], prop.m_ListType);
			if (prop.m_Lister != NULL)
				(this->*prop.m_Lister)(count);

			/* read items and store them into the array */
			for (int k = 0; k < count; k++) {
				if (prop.m_Handler != NULL)
					(this->*prop.m_Handler)(words[which_word], prop.m_Type);
				which_word++;
			}
		} else {                     /* not a list */
			if (prop.m_Handler != NULL)
				(this->*prop.m_Handler)(words[which_word], prop.m_Type);
			which_word++;
		}
	}
}


/*----------------------------------------------------------------------------
Read an element from a binary file.
----------------------------------------------------------------------------*/
void TranslatorPLY::getElementBinary(const TranslatorPLY::Element* elem) {
	for (Array<Property>::Iterator iter = elem->m_Properties.begin(); iter != elem->m_Properties.end(); ++iter) {
		Property& prop = *iter;

		if (prop.m_ListType != TypeNone) {       /* a list */
			/* get and store the number of items in the list */
			int count = getItemBinary<int>(prop.m_ListType);
			if (prop.m_Lister != NULL)
				(this->*prop.m_Lister)(count);

			/* read items and store them into the array */
			for (int k = 0; k < count; k++) {
				if (prop.m_Handler != NULL)
					(this->*prop.m_Handler)(m_Dummy, prop.m_Type);
				else
					m_In.skip(m_TypeSizes[prop.m_Type]);
			}
		} else {                     /* not a list */
			if (prop.m_Handler != NULL)
				(this->*prop.m_Handler)(m_Dummy, prop.m_Type);
			else
				m_In.skip(m_TypeSizes[prop.m_Type]);
		}
	}
}

}; };

