/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "geometry.meshcollider.h"
#include "graphics3d.vertexmesh.h"
#include "graphics3d.context.h"
#include "graphics3d.material.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.shader.h"
#include "graphics3d.blendmode.h"
#include "graphics3d.stencilmode.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.appearance.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "base.string.inl"
#include "system.thread.inl"
#include "math.vector.inl"
#include "math.vector3.inl"
#include "math.matrix.inl"
#include "math.range.inl"
#include "geometry.trianglemesh.inl"
#include "geometry.bounds.inl"
#include "geometry.sphere.inl"
#include "geometry.box.inl"
#include "geometry.alignedbox.inl"
#include "geometry.collider.inl"
#include "geometry.meshcollider.inl"
#include "geometry.intersection.inl"
#include "geometry.linear.inl"
#include "geometry.line.inl"
#include "geometry.ray.inl"
#include "geometry.segment.inl"
#include "container.array.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.context.inl"
#include "graphics3d.material.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.pixelprogram.inl"
#include "graphics3d.vertexprogram.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.appearance.inl"


namespace magnum { namespace graphics3d {
using magnum::object::AutoPointer;
using magnum::geometry::Bounds;
using magnum::geometry::AlignedBox3;
using magnum::geometry::MeshCollider3;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
VertexMesh::VertexMesh() {
	m_Range.setStart(0);
	m_Range.setEnd(0);
	m_BufferCount = 0;
	m_Attributes = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexMesh::attachVertexBuffer(VertexBuffer* vb,unsigned int i) {
	ASSERT(i < VertexBuffer::MaxBuffers);
	ASSERT(vb != NULL);

	m_VertexBuffer[i] = vb;
	updateBufferInfo();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexMesh::detachVertexBuffer(unsigned int i) {
	ASSERT(i < VertexBuffer::MaxBuffers);

	m_VertexBuffer[i] = NULL;
	updateBufferInfo();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexMesh::releaseVertexBuffers() {
	for (int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		m_VertexBuffer[i] = NULL;
	}
	m_BufferCount = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexMesh::onRender() const {
	selectVertexBuffers();

	if (getAppearance() != NULL)
		getAppearance()->select();

	if (m_Range.getSize()) {
		getContext()->renderVertices(getPrimitive(), m_Range.getStart(), m_Range.getSize());
	} else {
		unsigned int size = 0;
		for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
			if (m_VertexBuffer[i] != NULL) {
				size = m_VertexBuffer[i]->getSize();
				break;
			}
		}
		getContext()->renderVertices(getPrimitive(), size);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool VertexMesh::onUpdateCollider() {
	// Find correct attribute and offset
	VertexBuffer* vb;
	unsigned int offset;
	unsigned int type;
	unsigned int count;
	if (!searchVertices(&vb, &offset, &count, &type))
		return false;

	// Get offset, stride etc	
	unsigned int numVertices = vb->getSize();
	unsigned int incVertices = vb->getStride();
	void* vertices = vb->lock(VertexBuffer::ReadOnly);
	if (vertices == NULL)
		return false;

	// Adjust buffer for offset and range	
	vertices = (void*)((byte *)vertices + offset);
	if (m_Range.getSize() > 0) {
		vertices = (void*)((byte *)vertices + m_Range.getStart()*incVertices);
		numVertices = m_Range.getSize();	
	}
	
	AutoPointer<MeshCollider3<float> > collider = new MeshCollider3<float>();
	collider->setPrimitive((MeshCollider3<float>::Primitive)getPrimitive());

	float x,y,z;
#define copyVertex(T) \
		x = (float)((T*)vertices)[0]; \
		y = (count >= 2) ? (float)((T*)vertices)[1] : 0.0f; \
		z = (count >= 3) ? (float)((T*)vertices)[2] : 0.0f; \
		vertices = (void*)((byte *)vertices + incVertices); \
		collider->setVertex(i, Vector3<float>(x,y,z));
#define copyVertices(T) \
		collider->setVertices(numVertices); \
		for (unsigned int i = 0; i < numVertices; i++) { \
			copyVertex(T); \
		}
	switch(type) {
		case Types::Byte3:
		case Types::Byte4:
			copyVertices(int8);
			break;
		case Types::UByte3:
		case Types::UByte4:
			copyVertices(uint8);
			break;
		case Types::Short3:
		case Types::Short4:
			copyVertices(int16);
			break;
		case Types::UShort3:
		case Types::UShort4:
			copyVertices(uint16);
			break;
		case Types::Int3:
		case Types::Int4:
			copyVertices(int32);
			break;
		case Types::UInt3:
		case Types::UInt4:
			copyVertices(uint32);
			break;
		case Types::Float3:
		case Types::Float4:
			copyVertices(float);
			break;
		case Types::Double3:
		case Types::Double4:
			copyVertices(double);
			break;
		default:
			break;
	}
	vb->unlock();

#undef copyVertex
#undef copyVertices

	attachCollider(collider);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool VertexMesh::onUpdateBounds() {
	// Find correct attribute and offset
	VertexBuffer* vb;
	unsigned int offset;
	unsigned int type;
	unsigned int count;
	if (!searchVertices(&vb, &offset, &count, &type))
		return false;
	
	// Get offset, stride etc	
	unsigned int numVertices = vb->getSize();
	unsigned int incVertices = vb->getStride();
	void* vertices = vb->lock(VertexBuffer::ReadOnly);
	if (vertices == NULL)
		return NULL;

	// Adjust buffer for offset and range	
	vertices = (void*)((byte *)vertices + offset);
	if (m_Range.getSize() > 0) {
		vertices = (void*)((byte *)vertices + m_Range.getStart()*incVertices);
		numVertices = m_Range.getSize();	
	}
	
	switch(type) {
		case Types::Byte3:
		case Types::Byte4: {
			AlignedBox3<byte> box;
			box.bound((Vector<byte,3>*)vertices, incVertices, numVertices);
			setBounds(AlignedBox3<float>(Vector<float,3>(box.getMinimum()), Vector<float,3>(box.getMaximum())));
			break;		
		}
		case Types::Short3:
		case Types::Short4: {
			AlignedBox3<short> box;
			box.bound((Vector<short,3>*)vertices, incVertices, numVertices);
			setBounds(AlignedBox3<float>(Vector<float,3>(box.getMinimum()), Vector<float,3>(box.getMaximum())));
			break;		
		}
		case Types::Integer3:
		case Types::Integer4: {
			AlignedBox3<int> box;
			box.bound((Vector<int,3>*)vertices, incVertices, numVertices);
			setBounds(AlignedBox3<float>(Vector<float,3>(box.getMinimum()), Vector<float,3>(box.getMaximum())));
			break;		
		}
		case Types::Float3:
		case Types::Float4: {
			AlignedBox3<float> box;
			box.bound((Vector<float,3>*)vertices, incVertices, numVertices);
			setBounds(AlignedBox3<float>(box.getMinimum(), box.getMaximum()));
			break;		
		}
		case Types::Double3:
		case Types::Double4: {
			AlignedBox3<double> box;
			box.bound((Vector<double,3>*)vertices, incVertices, numVertices);
			setBounds(AlignedBox3<float>(Vector<float,3>(box.getMinimum()), Vector<float,3>(box.getMaximum())));
			break;		
		}
		
		default:
			vb->unlock();
			return false;
	}
	
	vb->unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TriangleMesh3<float>* VertexMesh::onTriangulate() const {
	// Find correct attribute and offset
	VertexBuffer* vb;
	unsigned int offset;
	unsigned int type;
	unsigned int count;
	if (!searchVertices(&vb, &offset, &count, &type))
		return NULL;

	// Copy vertices
	unsigned int numVertices = vb->getSize();
	unsigned int incVertices = vb->getStride();
	void* vertices = vb->lock(VertexBuffer::ReadOnly);
	if (vertices == NULL)
		return NULL;
		
	// Adjust buffer for offset and range	
	vertices = (void*)((byte *)vertices + offset);
	if (m_Range.getSize() > 0) {
		vertices = (void*)((byte *)vertices + m_Range.getStart()*incVertices);
		numVertices = m_Range.getSize();	
	}

	// Create Triangle mesh
	TriangleMesh3<float>* mesh = new TriangleMesh3<float>();
	float x,y,z;

#define copyVertex(T) \
		x = (float)((T*)vertices)[0]; \
		y = (count >= 2) ? (float)((T*)vertices)[1] : 0.0f; \
		z = (count >= 3) ? (float)((T*)vertices)[2] : 0.0f; \
		vertices = (void*)((byte *)vertices + incVertices); \
		mesh->setVertex(i, Vector3<float>(x,y,z));
#define copyVertices(T) \
		mesh->setVertices(numVertices); \
		for (unsigned int i = 0; i < numVertices; i++) { \
			copyVertex(T); \
		}
	switch(type) {
		case Types::Byte3:
		case Types::Byte4:
			copyVertices(int8);
			break;
		case Types::UByte3:
		case Types::UByte4:
			copyVertices(uint8);
			break;
		case Types::Short3:
		case Types::Short4:
			copyVertices(int16);
			break;
		case Types::UShort3:
		case Types::UShort4:
			copyVertices(uint16);
			break;
		case Types::Int3:
		case Types::Int4:
			copyVertices(int32);
			break;
		case Types::UInt3:
		case Types::UInt4:
			copyVertices(uint32);
			break;
		case Types::Float3:
		case Types::Float4:
			copyVertices(float);
			break;
		case Types::Double3:
		case Types::Double4:
			copyVertices(double);
			break;
		default:
			break;
	}
	vb->unlock();

	// Create Triangles
	switch(getPrimitive()) {
		case TriangleList:
			for (unsigned int i = 0; i < numVertices/3; i++) {
				mesh->setTriangle(i, i*3, i*3+1, i*3+2);
			}
			break;
		case TriangleStrip:
			for (unsigned int i = 0; i < numVertices - 2; i++) {
				if (i & 1)
					mesh->setTriangle(i, i, i+1, i+2);
				else
					mesh->setTriangle(i, i+1, i, i+2);
			}
			break;
		case TriangleFan:
			for (unsigned int i = 0; i < numVertices - 2; i++) {
				mesh->setTriangle(i, 0, i+1, i+2);
			}
			break;
		case QuadStrip:
			ASSERT(false);
			break;
		case QuadList:
			for (unsigned int i = 0; i < numVertices/4; i++) {
				mesh->setTriangle(2*i, i*4, i*4+1, i*4+2);
				mesh->setTriangle(2*i+1, i*4, i*4+2, i*4+3);
			}
			break;
		default:
			break;
	}


#undef copyVertex
#undef copyVertices

	return mesh;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexMesh::updateBufferInfo() {
	m_VertexCount = 0;
	m_BufferCount = 0;
	m_Attributes = 0;
	for (int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (m_VertexBuffer[i] == NULL)
			continue;

		m_BufferCount = i+1;
		m_Attributes |= m_VertexBuffer[i]->getAttributes();

		if (m_VertexCount == 0)
			m_VertexCount = m_VertexBuffer[i]->getSize();
		else
			m_VertexCount = min(m_VertexCount, m_VertexBuffer[i]->getSize());
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool VertexMesh::searchVertices(VertexBuffer** vbp,unsigned int* offset, unsigned int* count, unsigned int* type) const {
	ASSERT(vbp != NULL);
	ASSERT(offset != NULL);
	ASSERT(count != NULL);
	ASSERT(type != NULL);

	// Search appropriate vertex buffer
	unsigned int i;
	for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (getVertexBuffer(i) == NULL)
			continue;
		if (getVertexBuffer(i)->getAttributes() & Attributes::Vertex) {
			*vbp = getVertexBuffer(i);
			break;
		}
	}
	if (i == VertexBuffer::MaxBuffers)
		return false;

	// Find correct attribute and offset
	VertexBuffer* vb = *vbp;
	for (i = 0; i < VertexFormat::MaxComponents; i++) {
		if (vb->getAttribute(i) & Attributes::Vertex) {
			*offset = vb->getOffset(i);
			*type = vb->getType(i);
			*count = vb->getCount(i);
			return true;
		}
	}

	return false;
}

}; };
