/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "gui.view.h"
#include "gui.window.h"
#include "gui.container.h"
#include "gui.context.h"
#include "gui.bitmap.h"
#include "container.chain.h"
#include "graphics.dimension.h"
#include "graphics.rectangle.h"
#include "application.message.h"

#include "base.string.inl"
#include "object.class.inl"
#include "object.inl"
#include "gui.view.inl"
#include "gui.window.inl"
#include "gui.context.inl"
#include "container.chain.inl"
#include "container.iterator.inl"
#include "graphics.transform.inl"
#include "graphics.dimension.inl"
#include "graphics.rectangle.inl"
#include "graphics.point.inl"


#ifndef __FORWINDOWS__
#error "This file is only for Windows"
#else
#include <commctrl.h>
#endif


namespace magnum { namespace gui {

char View::windowClass[] = "magnum::gui::view::class";

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void View::registerClass() {
	static bool registered = false;
	if (registered)
		return;

	// Register Window Class - normally this has only to be done once
	WNDCLASSEX WndClass;
	memset(&WndClass, 0, sizeof(WNDCLASSEX));
	WndClass.cbSize		    = sizeof(WNDCLASSEX);
	WndClass.style 		    = CS_OWNDC;
	WndClass.lpfnWndProc    = windowProc;
	WndClass.hInstance 	    = GetModuleHandle(NULL);
	WndClass.hIcon 		    = LoadIcon(NULL, IDI_APPLICATION);
	WndClass.hCursor 	    = LoadCursor(NULL, IDC_ARROW);
	WndClass.hbrBackground  = (HBRUSH)(COLOR_WINDOW);
	WndClass.lpszMenuName 	= NULL;
	WndClass.lpszClassName  = windowClass;
	WndClass.cbWndExtra		= sizeof(intptr_t);
	RegisterClassEx(&WndClass);
	registered = true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
LRESULT CALLBACK View::windowProc(HWND hwnd, UINT Message,WPARAM wParam, LPARAM lParam) {
	View* view = (View *)(intptr_t)GetWindowLong(hwnd,0);
	int ret;
	int modifiers;

	switch(Message) {
		case WM_CREATE:
			view = (View*)(((CREATESTRUCT *)lParam)->lpCreateParams);
			SetWindowLongPtr(hwnd,0,(LONG_PTR)view);
			return 0;
		case WM_ERASEBKGND:
			return 1;
		case WM_PALETTECHANGED:
			if (hwnd == (HWND)wParam)
				break;
			/* fall through to WM_QUERYNEWPALETTE */
		case WM_QUERYNEWPALETTE:
			return FALSE;
		case WM_PAINT:
			PAINTSTRUCT PaintStruct;
			BeginPaint(hwnd, &PaintStruct);
PRINT4("View::windowProc: WM_PAINT: (%d,%d)-(%d,%d)",PaintStruct.rcPaint.left,PaintStruct.rcPaint.top, PaintStruct.rcPaint.right,PaintStruct.rcPaint.bottom);
			view->onPaint(&view->m_Context,
						Rectangle<float>(
						(float)PaintStruct.rcPaint.left, 
						(float)PaintStruct.rcPaint.top, 
						(float)(PaintStruct.rcPaint.right - PaintStruct.rcPaint.left + 1), 
						(float)(PaintStruct.rcPaint.bottom - PaintStruct.rcPaint.top + 1)));
			EndPaint(hwnd, &PaintStruct);
			return 0;

		case WM_MOVE:
PRINT("View::windowProc: WM_MOVE");
			if (view == NULL)
				break;
			view->m_Frame.setX(LOWORD(lParam));
			view->m_Frame.setY(HIWORD(lParam));
			view->onResize(view->m_Frame);
			return 0;

		case WM_SIZE:
PRINT("View::windowProc: WM_SIZE");
			if (view == NULL)
				break;
			view->m_Frame.setWidth((float)LOWORD(lParam));
			view->m_Frame.setHeight((float)HIWORD(lParam));
			view->onResize(view->m_Frame);
			return 0;

		case WM_CHAR:
PRINT1("View::windowProc: WM_CHAR: %c",wParam);
			modifiers = 0;
			if (GetKeyState(VK_LSHIFT))
				modifiers |= Character::ShiftLeft;
			if (GetKeyState(VK_RSHIFT))
				modifiers |= Character::ShiftRight;
			if (GetKeyState(VK_LCONTROL))
				modifiers |= Character::ControlLeft;
			if (GetKeyState(VK_RCONTROL))
				modifiers |= Character::ControlRight;
			if (GetKeyState(VK_LMENU))
				modifiers |= Character::AltLeft;
			if (GetKeyState(VK_RMENU))
				modifiers |= Character::AltRight;
			if (GetKeyState(VK_APPS))
				modifiers |= Character::Meta;
			if (GetKeyState(VK_LWIN))
				modifiers |= Character::MetaLeft;
			if (GetKeyState(VK_RWIN))
				modifiers |= Character::MetaRight;
			ret = view->onChar((Character::Code)wParam,(Character::Modifier)modifiers);
			return ret ? 0 : -1;
			
		case WM_KEYDOWN:
PRINT("View::windowProc: WM_KEYDOWN");
			if ((lParam & 0x40000000) == 0) {
				ret = view->onKeyDown((Keyboard::Key)wParam);
				return ret ? 0 : -1;
			} else {
				return -1;
			}

		case WM_KEYUP:
PRINT("View::windowProc: WM_KEYUP");
			ret = view->onKeyUp((Keyboard::Key)wParam);
			return ret ? 0 : -1;
		
		case WM_MOUSELEAVE:
PRINT("View::windowProc: WM_MOUSELEAVE");
			view->m_MouseTracking = false;
			view->onMouseLeave();
			return 0;
			
		case WM_MOUSEMOVE:
PRINT2("View::windowProc: WM_MOUSEMOVE: (%d,%d)",LOWORD(lParam),HIWORD(lParam));
			modifiers = 
				((wParam & MK_LBUTTON) ? Mouse::LeftDown : Mouse::LeftUp) |
				((wParam & MK_MBUTTON) ? Mouse::MiddleDown : Mouse::MiddleUp) |
				((wParam & MK_RBUTTON) ? Mouse::RightDown : Mouse::RightUp) |
				((wParam & MK_SHIFT) ? Mouse::Shift : 0) |
				((wParam & MK_CONTROL) ? Mouse::Control : 0);
			if (!view->m_MouseTracking) {
				PRINT("View::windowProc: WM_MOUSEMOVE -> onMouseEnter");
				view->m_MouseTracking = true;
				view->onMouseEnter(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)), (Mouse::Button)modifiers);
				TRACKMOUSEEVENT tme;
				tme.cbSize = sizeof(tme);
				tme.dwFlags = TME_LEAVE;
				tme.dwHoverTime = 0;
				tme.hwndTrack = view->m_WindowHandle;
				_TrackMouseEvent(&tme);
			}
			ret = view->onMouseMove(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)), (Mouse::Button)modifiers);
			return ret ? 0 : -1;

		case WM_MBUTTONDOWN:
PRINT("View::windowProc: WM_MBUTTONDOWN");
			ret = view->onMouseButton(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)),Mouse::MiddleDown);
			return ret ? 0 : -1;

		case WM_MBUTTONUP:
PRINT("View::windowProc: WM_MBUTTONUP");
			ret = view->onMouseButton(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)),Mouse::MiddleUp);
			return ret ? 0 : -1;

		case WM_LBUTTONDOWN:
PRINT("View::windowProc: WM_LBUTTONDOWN");
			ret = view->onMouseButton(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)),Mouse::LeftDown);
			return ret ? 0 : -1;

		case WM_LBUTTONUP:
PRINT("View::windowProc: WM_LBUTTONUP");
			ret = view->onMouseButton(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)),Mouse::LeftUp);
			return ret ? 0 : -1;

		case WM_RBUTTONDOWN:
PRINT("View::windowProc: WM_RBUTTONDOWN");
			ret = view->onMouseButton(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)),Mouse::RightDown);
			return ret ? 0 : -1;

		case WM_RBUTTONUP:
PRINT("View::windowProc: WM_RBUTTONUP");
			ret = view->onMouseButton(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)),Mouse::RightUp);
			return ret ? 0 : -1;
			
		case WM_MOUSEWHEEL:
PRINT("View::windowProc: WM_MOUSEWHEEL");
			modifiers = 
				((wParam & MK_LBUTTON) ? Mouse::LeftDown : Mouse::LeftUp) |
				((wParam & MK_MBUTTON) ? Mouse::MiddleDown : Mouse::MiddleUp) |
				((wParam & MK_RBUTTON) ? Mouse::RightDown : Mouse::RightUp) |
				((wParam & MK_SHIFT) ? Mouse::Shift : 0) |
				((wParam & MK_CONTROL) ? Mouse::Control : 0);
			return view->onMouseWheel(HIWORD(wParam), (Mouse::Button)modifiers);
	}

	return DefWindowProc(hwnd, Message, wParam, lParam);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onAttachWindow() {
	// Get parent
	HWND parent;
	Rectangle<float> frame = convertToParent(getBounds());;
	if(m_Parent != NULL) {
		parent = m_Parent->m_Context.m_WindowHandle;
	} else {
		parent = getWindow()->getWindowHandle();
	}
	
	// Create a static control...
	registerClass();
  	m_WindowHandle = CreateWindowEx(0,windowClass,
							   "view", WS_VISIBLE | WS_CHILD,
							   (int)frame.getX(), (int)frame.getY(),
							   (int)frame.getWidth(), (int)frame.getHeight(),
							   parent, NULL,
							   GetModuleHandle(NULL), (LPVOID)this);
	ASSERT(m_WindowHandle != NULL);

  	m_DeviceHandle = GetDC(m_WindowHandle);
	ASSERT(m_DeviceHandle != NULL); 
	
	// Initialize graphics context
  	SetMapMode(m_DeviceHandle,MM_TEXT);
	SetGraphicsMode(m_DeviceHandle,GM_ADVANCED);
	
	// Init context
	m_Context.m_WindowHandle = m_WindowHandle;
	m_Context.m_DeviceHandle = m_DeviceHandle;

	m_MouseTracking = false;
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onDetachWindow() {
	if (m_WindowHandle == NULL)
		return false;
	if (m_DeviceHandle == NULL)
		return false;
	ReleaseDC(m_WindowHandle, m_DeviceHandle);
	DestroyWindow(m_WindowHandle);
	m_WindowHandle = NULL;
	m_DeviceHandle = NULL;
	m_Context.m_WindowHandle = NULL;
	m_Context.m_DeviceHandle = NULL;
	return true;
}

}; };
