/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "gui.view.h"
#include "gui.window.h"
#include "gui.context.h"
#include "gui.bitmap.h"
#include "application.message.h"
#include "system.thread.h"

#include "base.string.inl"
#include "gui.view.inl"
#include "gui.window.inl"
#include "gui.context.inl"
#include "system.thread.inl"
#include "system.signal.inl"
#include "object.inl"
#include "object.class.inl"
#include "graphics.point.inl"
#include "graphics.dimension.inl"
#include "graphics.rectangle.inl"
#include "graphics.transform.inl"
#include "application.looper.inl"
#include "application.message.inl"
#include "application.messenger.inl"
#include "container.iterator.inl"
#include "container.chain.inl"


#include <stdio.h>

namespace magnum { namespace gui {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Window::Window() {
	m_Focus = NULL;
	m_Open = false;

	m_MinSize = Dimension<float>(2,2);
	m_MaxSize = Dimension<float>(4096,4096);

#if defined(__FORWINDOWS__)
	m_WindowHandle = NULL;
	m_WindowStyle = NULL;
#elif defined(__FORX11__)
    m_WindowHandle = 0;
	m_DeviceHandle = 0;
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Window::~Window() {
	if (isAlive()) {
		close();
	}
	clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::open() {
	if (isAlive() || isOpen())
		return false;
	start();
	sendMessage(this, MSG_OPEN_WINDOW);
	m_Signal.wait();
	m_Signal.reset();
	return isOpen();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::close() {
	if (!isAlive() || !isOpen())
		return false;
	sendMessage(this, MSG_CLOSE_WINDOW);
	m_Signal.wait();
	m_Signal.reset();
	stop();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Window::setLocation(const Point<float>& point) {
	m_Frame.setOrigin(point);
	if (!isOpen()) {
		onMove(m_Frame.getOrigin());
		return;
	}

#if defined(__FORWINDOWS__)
	RECT rect;
  	rect.left = (int)m_Frame.getX();
  	rect.top = (int)m_Frame.getY();
  	rect.right = (int)m_Frame.getX() + (int)m_Frame.getWidth();
  	rect.bottom = (int)m_Frame.getY() + (int)m_Frame.getHeight();
  	AdjustWindowRectEx(&rect, m_WindowStyle, FALSE, WS_EX_CLIENTEDGE);
	SetWindowPos(m_WindowHandle, NULL, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, SWP_NOZORDER | SWP_NOCOPYBITS | SWP_NOSIZE);
#elif defined(__FORX11__)
	XMoveWindow(m_Display, m_WindowHandle, m_Bounds.getX(), m_Bounds.getY());
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Window::setSize(const Dimension<float>& dim) {
	m_Frame.setSize(dim);
	if (!isOpen()) {
		onResize(m_Frame);
		return;	
	}

#if defined(__FORWINDOWS__)
	RECT rect;
  	rect.left = (int)m_Frame.getX();
  	rect.top = (int)m_Frame.getY();
  	rect.right = (int)m_Frame.getX() + (int)m_Frame.getWidth();
  	rect.bottom = (int)m_Frame.getY() + (int)m_Frame.getHeight();
  	AdjustWindowRectEx(&rect, m_WindowStyle, FALSE, WS_EX_CLIENTEDGE);
	SetWindowPos(m_WindowHandle, NULL, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, SWP_NOZORDER | SWP_NOCOPYBITS | SWP_NOMOVE);
#elif defined(__FORX11__)
	XResizeWindow(m_Display, m_WindowHandle, m_Bounds.getWidth(), m_Bounds.getHeight());
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Window::setName(const char* title) {
	m_Title = title;

#if defined(__FORWINDOWS__)
	if (isOpen())
		SetWindowText(m_WindowHandle, title);
#elif defined(__FORX11__)
	if (isOpen())
		XStoreName(m_Display, m_WindowHandle, title);
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::invalidate(const Rectangle<float>& rect) {
	if (!isOpen())
		return false;

#if defined(__FORWINDOWS__)
	RECT r;
	r.left = (int)rect.getX();
	r.top = (int)rect.getY(),
	r.right = (int)rect.getX() + (int)rect.getWidth() - 1;
	r.bottom = (int)rect.getY() + (int)rect.getHeight() - 1;
	InvalidateRect(m_WindowHandle, &r, TRUE);
#elif defined(__FORX11__)
	XClearArea(m_Display, m_WindowHandle, rect.getX(), rect.getY(), rect.getWidth(), rect.getHeight(), True);
	XSync(m_Display, False);
#endif

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::invalidate(void) {
	if (!isOpen())
		return false;

#if defined(__FORWINDOWS__)
	InvalidateRect(m_WindowHandle, NULL, TRUE);
#elif defined(__FORX11__)
	XClearArea(m_Display, m_WindowHandle, 0, 0, m_Bounds.getWidth(), m_Bounds.getHeight(), True);
	XSync(m_Display, False);
#endif

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::insert(View* view) {
	if (view->getWindow() != NULL)
		return false;
	if (view->getParent() != NULL)
		return false;

	m_Views.insert(view);
	insertHandler(view);
	view->m_Window = this;

	if (isOpen())
		view->onAttachWindow();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::remove(View *view) {
	if (view->getWindow() != this)
		return false;
	if (view->getParent() != NULL)
		return false;

	if (m_Focus == view) 
		releaseFocus();

	if (isOpen())
		view->onDetachWindow();

	view->m_Window = NULL;
	m_Views.remove(view);
	removeHandler(view);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::clear() {
	View *view;

	releaseFocus();

	while(m_Views.begin() != m_Views.end()) {
		view = &(*m_Views.begin());

		if (isOpen())
			view->onDetachWindow();

		view->m_Window = NULL;
		m_Views.remove(view);
		removeHandler(view);
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::setFocus(View* focus) {
	if ((focus != NULL) && (focus->m_Window != this))
		return false;
	if (m_Focus == focus)
		return true;

	if (m_Focus != NULL) {
		m_Focus->onReleaseFocus();
	}
	m_Focus = focus;
	if (m_Focus != NULL) {
		m_Focus->onGetFocus();
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::releaseFocus(View* focus) {
	if (focus == NULL)
		return true;
	if (focus->m_Window != this)
		return false;
	if (m_Focus != focus)
		return false;

	focus->onReleaseFocus();
	m_Focus = NULL;

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::releaseFocus() {
	if (m_Focus != NULL) {
		m_Focus->onReleaseFocus();
	}
	m_Focus = NULL;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::setPalette(const Palette<byte>& ) {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::getPalette(Palette<byte>& ) const {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::realizePalette(void) {
	if (m_PaletteHandle == NULL)
		return false;
	SelectPalette(m_DeviceHandle, m_PaletteHandle,FALSE);
	if(::RealizePalette(m_DeviceHandle) == GDI_ERROR) 
		return false;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::unrealizePalette(void) {
	if (m_PaletteHandle == NULL)
		return false;
	UnrealizeObject(m_PaletteHandle);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onHandleMessage(const Message& msg) {
	View *view;

	switch(msg.getCommand()) {
		case MSG_OPEN_WINDOW:
			if (sysWindowCreate() == false) {
				stop();
				m_Signal.raise();
				return false;
			}
			m_Open = true;
			m_Focus = NULL;
			m_Signal.raise();
			onCreate();
			sysWindowShow();
			return true;

		case MSG_CLOSE_WINDOW:
			onDestroy();
			sysWindowDestroy();
			m_Open = false;
			m_Focus = NULL;
			m_Signal.raise();
			return true;

		case MSG_INVALIDATE_VIEW:
#if defined(__FORWINDOWS__)
			if (msg.getPointer("view",(void **)&view)) {
				PAINTSTRUCT PaintStruct;
				BeginPaint(m_WindowHandle, &PaintStruct);
				view->onPaint(&m_Context, view->getBounds());
				EndPaint(m_WindowHandle, &PaintStruct);
			}
#elif defined(__FORX11__)
			if (msg.getPointer("view",(void **)&view)) {
				view->onPaint(view->getBounds());
			}
#endif
			return true;

		case MSG_RESIZE_VIEW:
			if (msg.getPointer("view",(void **)&view)) {
				view->onResize(view->getFrame());
			}
			return true;

		default:
			return Looper::handleMessage(msg);
	}

	return true;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onCreate(void) {
	// attach all children to window
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;

	while (iter != m_Views.end()) {
		view = &(*iter);
		view->onAttachWindow();
		iter++;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onDestroy(void) {
	// attach all children to window
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;

	while (iter != m_Views.end()) {
		view = &(*iter);
		view->onDetachWindow();
		iter++;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onRequestClose(void) {
	close();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onPaint(Context* context, const Rectangle<float>& rect) {
#if defined(__FORWINDOWS__)
	return true;
#else
	Chain<View>::Iterator iter = m_Views.begin();
	Rectangle<float> lrect;
	View* view;

	while (iter != m_Views.end()) {
		view = &(*iter);
		if (rect.intersects(view->getFrame())) {
			view->onPaint(context, rect);
		}
		iter++;
	}

	return true;
#endif	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onResize(const Rectangle<float>& ) {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;
	float dw, dh;

	dw = m_Frame.getWidth() - m_OldFrame.getWidth();
	dh = m_Frame.getHeight() - m_OldFrame.getHeight();

	if ((dw == 0)  && (dh == 0)) {
		m_OldFrame = m_Frame;
		return true;
	}

	// Now resize all childs (if needed)
	while (iter != m_Views.end()) {
		view = &(*iter);
		float x,y,w,h;

		if (view->m_Alignement == 0) {
			iter++;
			continue;
		}

		if ((view->m_Alignement & View::FollowLeftRight) == View::FollowLeftRight) {
			x = view->m_Frame.getX();
			w = view->m_Frame.getWidth() + dw;
		} else if (view->m_Alignement & View::FollowLeft) {
			x = view->m_Frame.getX();
			w = view->m_Frame.getWidth();
		} else if (view->m_Alignement & View::FollowRight) {
			x = view->m_Frame.getX() + dw;
			w = view->m_Frame.getWidth();
		} else {
			x = view->m_Frame.getX();
			w = view->m_Frame.getWidth();
		}

		if ((view->m_Alignement & View::FollowTopBottom) == View::FollowTopBottom) {
			y = view->m_Frame.getY();
			h = view->m_Frame.getHeight() + dh;
		} else if (view->m_Alignement & View::FollowTop) {
			y = view->m_Frame.getY();
			h = view->m_Frame.getHeight();
		} else if (view->m_Alignement & View::FollowBottom) {
			y = view->m_Frame.getY() + dh;
			h = view->m_Frame.getHeight();
		} else {
			y = view->m_Frame.getY();
			h = view->m_Frame.getHeight();
		}

		view->setFrame(Rectangle<float>(x, y, w, h));
		iter++;
	}

	m_OldFrame = m_Frame;

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onMove(const Point<float>& ) {
	m_OldFrame = m_Frame;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onKeyDown(Keyboard::Key ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onKeyUp(Keyboard::Key ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onChar(Character::Code ,Character::Modifier ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onMouseMove(const Point<float>& pos ,Mouse::Button button) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onMouseButton(const Point<float>& pos ,Mouse::Button button) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onMouseWheel(float angle ,Mouse::Button mod) {
	if (m_Focus == NULL)
		return false;
	return m_Focus->onMouseWheel(angle, mod);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onActivate(void) {
	// activate all children
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;

	while (iter != m_Views.end()) {
		view = &(*iter);
		view->onActivateWindow();
		iter++;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::onDeactivate(void) {
	// deactivate all children
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;

	while (iter != m_Views.end()) {
		view = &(*iter);
		view->onDeactivateWindow();
		iter++;
	}

	return true;
}


}; };
