/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "gui.view.h"
#include "gui.window.h"
#include "gui.context.h"
#include "gui.bitmap.h"
#include "container.chain.h"
#include "system.thread.h"
#include "application.looper.h"
#include "application.message.h"

#include "base.string.inl"
#include "object.inl"
#include "gui.view.inl"
#include "gui.window.inl"
#include "gui.context.inl"
#include "system.thread.inl"
#include "application.looper.inl"
#include "application.message.inl"
#include "graphics.point.inl"
#include "graphics.dimension.inl"
#include "graphics.rectangle.inl"
#include "container.iterator.inl"
#include "container.chain.inl"

#ifndef __FORWINDOWS__
#	error "This file is only for Windows"
#endif


namespace magnum { namespace gui {
using namespace magnum::system;

static char windowClass[] = "magnum::gui::window::class";


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
static void registerClass() {
	static bool registered = false;
	if (registered)
		return;

	// Register Window Class - normally this has only to be done once
	WNDCLASSEX WndClass;
	memset(&WndClass, 0, sizeof(WNDCLASSEX));
	WndClass.cbSize		    = sizeof(WNDCLASSEX);
	WndClass.style 		    = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	WndClass.lpfnWndProc    = Window::windowProc;
	WndClass.hInstance 	    = GetModuleHandle(NULL);
	WndClass.hIcon 		    = LoadIcon(NULL, IDI_APPLICATION);
	WndClass.hCursor 	    = LoadCursor(NULL, IDC_ARROW);
	WndClass.hbrBackground  = (HBRUSH)(COLOR_WINDOW);
	WndClass.lpszMenuName 	= NULL;
	WndClass.lpszClassName  = windowClass;
	RegisterClassEx(&WndClass);
	registered = true;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::sysWindowCreate() {
	registerClass();

	RECT rect;
	POINT point;

	m_WindowStyle = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MAXIMIZEBOX | WS_SIZEBOX;

  	rect.left = (int)m_Frame.getX();
  	rect.top = (int)m_Frame.getY();
  	rect.right = (int)m_Frame.getX() + (int)m_Frame.getWidth();
  	rect.bottom = (int)m_Frame.getY() + (int)m_Frame.getHeight();
  	AdjustWindowRectEx(&rect, m_WindowStyle, FALSE, WS_EX_CLIENTEDGE);

	// Create Window
  	m_WindowHandle = CreateWindowEx(WS_EX_CLIENTEDGE,windowClass,
							   m_Title, m_WindowStyle,
							   CW_USEDEFAULT, CW_USEDEFAULT,
							   rect.right - rect.left, rect.bottom - rect.top,
							   NULL, NULL, GetModuleHandle(NULL), (LPVOID)this);
	if (m_WindowHandle == NULL) 
		return false;

	// Get device handle
 	m_DeviceHandle = GetDC(m_WindowHandle);
 	if (m_DeviceHandle == NULL)
 		return false;
 		
  	SetMapMode(m_DeviceHandle,MM_TEXT);
	SetGraphicsMode(m_DeviceHandle,GM_ADVANCED);
	
	// Set palette stuff if needed
  	if (GetDeviceCaps(m_DeviceHandle,RASTERCAPS) & RC_PALETTE) {
		HPALETTE oldPal;

		m_Palette.palVersion = 0x300;
		m_Palette.palNumEntries = 256;
		m_PaletteHandle = CreatePalette(&m_Palette);
		oldPal = SelectPalette(m_DeviceHandle, m_PaletteHandle, FALSE);
		::DeleteObject(oldPal);
   	}


	// Get client coordinates and save them
	GetClientRect(m_WindowHandle, &rect);
	point.x = rect.left;
	point.y = rect.top;
	ClientToScreen(m_WindowHandle, &point);
	m_Frame.setX((float)point.x);
	m_Frame.setY((float)point.y);
	m_Frame.setWidth((float)(rect.right - rect.left));
	m_Frame.setHeight((float)(rect.bottom - rect.top));

	// Init context
	m_Context.m_WindowHandle = m_WindowHandle;
	m_Context.m_DeviceHandle = m_DeviceHandle;

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::sysWindowDestroy() {
	if (m_WindowHandle == NULL) 
		return false;

	if (m_PaletteHandle != NULL) {
		::UnrealizeObject(m_PaletteHandle);
		DeleteObject(m_PaletteHandle);
	}
	if (m_DeviceHandle != NULL) {
		ReleaseDC(m_WindowHandle, m_DeviceHandle);
		m_DeviceHandle = NULL;
	}

	if(DestroyWindow(m_WindowHandle) == FALSE) 
		return false;

	m_WindowHandle = NULL;
	m_PaletteHandle = NULL;

	// CLear Context	
	m_Context.m_WindowHandle = NULL;
	m_Context.m_DeviceHandle = NULL;
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Window::sysWindowShow() {
	if (m_WindowHandle == NULL)
		return false;

	ShowWindow(m_WindowHandle, SW_SHOW);
  	UpdateWindow(m_WindowHandle);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
LRESULT CALLBACK Window::windowProc(HWND hwnd, UINT msg,WPARAM wParam, LPARAM lParam) {
	Window* win = (Window *)(intptr_t)GetWindowLong(hwnd,GWL_USERDATA);
	RECT rect;
	POINT point;
	int buttons;
	int modifiers;

	switch(msg) {
		case WM_CREATE:
PRINT("Window::windowProc: WM_CREATE");
			win = (Window*)(((CREATESTRUCT *)lParam)->lpCreateParams);
			SetWindowLong(hwnd,GWL_USERDATA,(LONG)(intptr_t)win);
			return 0;

		case WM_CLOSE:
PRINT("Window::windowProc: WM_CLOSE");
			win->onRequestClose();
			return 0;

//		case WM_DESTROY:
//			PostQuitMessage(0);
//			return 0;

		case WM_GETMINMAXINFO:
PRINT("Window::windowProc: WM_GETMINMAXINFO");
			MINMAXINFO* minmaxinfo;
			if (win == NULL)
				return 1;
			minmaxinfo = (MINMAXINFO *)lParam;
			minmaxinfo->ptMaxTrackSize.x = (long)win->m_MaxSize.getWidth();
			minmaxinfo->ptMaxTrackSize.y = (long)win->m_MaxSize.getHeight();
			minmaxinfo->ptMinTrackSize.x = (long)win->m_MinSize.getWidth();
			minmaxinfo->ptMinTrackSize.y = (long)win->m_MinSize.getHeight();
			return 0;


		case WM_KEYDOWN:
PRINT("Window::windowProc: WM_KEYDOWN");
			if ((lParam & 0x40000000) != 0)
				return 0;
			if (win->m_Focus == NULL)
				win->onKeyDown((Keyboard::Key)wParam);
			else
				win->m_Focus->onKeyDown((Keyboard::Key)wParam);
			return 0;

		case WM_KEYUP:
PRINT("Window::windowProc: WM_KEYUP");
			if (win->m_Focus == NULL) {
				win->onKeyUp((Keyboard::Key)wParam);
			} else {
				win->m_Focus->onKeyUp((Keyboard::Key)wParam);
			}
			return 0;

		case WM_CHAR:
PRINT1("Window::windowProc: WM_CHAR: %c",wParam);
			modifiers = 0;
			if (GetKeyState(VK_LSHIFT))
				modifiers |= Character::ShiftLeft;
			if (GetKeyState(VK_RSHIFT))
				modifiers |= Character::ShiftRight;
			if (GetKeyState(VK_LCONTROL))
				modifiers |= Character::ControlLeft;
			if (GetKeyState(VK_RCONTROL))
				modifiers |= Character::ControlRight;
			if (GetKeyState(VK_LMENU))
				modifiers |= Character::AltLeft;
			if (GetKeyState(VK_RMENU))
				modifiers |= Character::AltRight;
			if (GetKeyState(VK_APPS))
				modifiers |= Character::Meta;
			if (GetKeyState(VK_LWIN))
				modifiers |= Character::MetaLeft;
			if (GetKeyState(VK_RWIN))
				modifiers |= Character::MetaRight;
			if (win->m_Focus == NULL)
				win->onChar((Character::Code)wParam, (Character::Modifier)modifiers);
			else
				win->m_Focus->onChar((Character::Code)wParam, (Character::Modifier)modifiers);
			return 0;

		case WM_PAINT:
			PAINTSTRUCT PaintStruct;
			BeginPaint(hwnd, &PaintStruct);
PRINT4("Window::windowProc: WM_PAINT: (%d,%d)-(%d,%d)",PaintStruct.rcPaint.left,PaintStruct.rcPaint.top, PaintStruct.rcPaint.right,PaintStruct.rcPaint.bottom);
			win->onPaint(&win->m_Context,
						Rectangle<float>(
						(float)PaintStruct.rcPaint.left, 
						(float)PaintStruct.rcPaint.top, 
						(float)(PaintStruct.rcPaint.right - PaintStruct.rcPaint.left + 1), 
						(float)(PaintStruct.rcPaint.bottom - PaintStruct.rcPaint.top + 1)));
			XFORM worldTransform;
			worldTransform.eM11 = 
			worldTransform.eM22 = 1.0f;
			worldTransform.eM12 = 
			worldTransform.eM21 = 
			worldTransform.eDx = 
			worldTransform.eDy = 0.0f;
			SetWorldTransform(win->m_Context.m_DeviceHandle, &worldTransform);
			EndPaint(hwnd, &PaintStruct);
			return 0;

		case WM_MOUSEMOVE:
PRINT2("Window::windowProc: WM_MOUSEMOVE: (%d,%d)",LOWORD(lParam),HIWORD(lParam));
			buttons = 
				((wParam & MK_LBUTTON) ? Mouse::LeftDown : Mouse::LeftUp) |
				((wParam & MK_MBUTTON) ? Mouse::MiddleDown : Mouse::MiddleUp) |
				((wParam & MK_RBUTTON) ? Mouse::RightDown : Mouse::RightUp) |
				((wParam & MK_SHIFT) ? Mouse::Shift : 0) |
				((wParam & MK_CONTROL) ? Mouse::Control : 0);
			win->onMouseMove(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)), (Mouse::Button)buttons);
			return 0;

		case WM_MBUTTONDOWN:
PRINT("Window::windowProc: WM_MBUTTONDOWN");
			win->onMouseButton(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)),Mouse::MiddleDown);
			return 0;

		case WM_MBUTTONUP:
PRINT("Window::windowProc: WM_MBUTTONUP");
			win->onMouseButton(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)),Mouse::MiddleUp);
			return 0;

		case WM_LBUTTONDOWN:
PRINT("Window::windowProc: WM_LBUTTONDOWN");
			win->onMouseButton(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)),Mouse::LeftDown);
			return 0;

		case WM_LBUTTONUP:
PRINT("Window::windowProc: WM_LBUTTONUP");
			win->onMouseButton(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)),Mouse::LeftUp);
			return 0;

		case WM_RBUTTONDOWN:
PRINT("Window::windowProc: WM_RBUTTONDOWN");
			win->onMouseButton(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)),Mouse::RightDown);
			return 0;

		case WM_RBUTTONUP:
PRINT("Window::windowProc: WM_RBUTTONUP");
			win->onMouseButton(Point<float>((float)LOWORD(lParam),(float)HIWORD(lParam)),Mouse::RightUp);
			return 0;
			
		case WM_MOUSEWHEEL:
PRINT("Window::windowProc: WM_MOUSEWHEEL");
			buttons = 
				((wParam & MK_LBUTTON) ? Mouse::LeftDown : Mouse::LeftUp) |
				((wParam & MK_MBUTTON) ? Mouse::MiddleDown : Mouse::MiddleUp) |
				((wParam & MK_RBUTTON) ? Mouse::RightDown : Mouse::RightUp) |
				((wParam & MK_SHIFT) ? Mouse::Shift : 0) |
				((wParam & MK_CONTROL) ? Mouse::Control : 0);
			return win->onMouseWheel(HIWORD(wParam) ,(Mouse::Button)buttons);

		case WM_MOVE:
PRINT("Window::windowProc: WM_MOVE");
			GetClientRect(win->m_WindowHandle, &rect);
			point.x = rect.left;
			point.y = rect.top;
			ClientToScreen(win->m_WindowHandle, &point);
			win->m_Frame.setX((float)point.x);
			win->m_Frame.setY((float)point.y);
			win->onMove(win->m_Frame.getOrigin());
			return 0;

		case WM_SIZE:
PRINT("Window::windowProc: WM_SIZE");
			GetClientRect(win->m_WindowHandle, &rect);
			win->m_Frame.setWidth((float)(rect.right - rect.left));
			win->m_Frame.setHeight((float)(rect.bottom - rect.top));
			win->onResize(win->m_Frame);
			return 0;

		case WM_ACTIVATE:
PRINT("Window::windowProc: WM_ACTIVATE");
			if ((LOWORD(wParam) == WA_ACTIVE) || (LOWORD(wParam) == WA_CLICKACTIVE)) {
				win->onActivate();
			} else {
				win->onDeactivate();
			}
			return 0;

		case WM_ERASEBKGND:
PRINT("Window::windowProc: WM_ERASEBKGND");
			return 1;

		case WM_QUERYNEWPALETTE:
PRINT("Window::windowProc: WM_QUERYNEWPALETTE");
			if (win->m_PaletteHandle == NULL) 
				return 0;
  			SelectPalette(win->m_DeviceHandle,win->m_PaletteHandle,FALSE);
  			::RealizePalette(win->m_DeviceHandle);
			return 1;

		case WM_PALETTECHANGED:
PRINT("Window::windowProc: WM_PALETTECHANGED");
			if ((hwnd != (HWND)wParam) && (win->m_PaletteHandle != NULL)) {
 				SelectPalette(win->m_DeviceHandle,win->m_PaletteHandle,TRUE);
	  			::RealizePalette(win->m_DeviceHandle);
				return 1;
			}
			return 0;

		default:
			return DefWindowProc(hwnd, msg, wParam, lParam);
	}
}

}; };
