/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "gui3d.context.h"
#include "gui3d.bitmap.h"
#include "gui3d.fontmap.h"
#include "graphics3d.shader.h"
#include "graphics3d.material.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"

#include "gui3d.context.inl"
#include "gui3d.fontmap.inl"
#include "object.pointer.inl"
#include "container.map.inl"
#include "container.allocator.inl"
#include "graphics.dimension.inl"
#include "graphics.rectangle.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.depthmode.inl"
#include "graphics3d.polygonmode.inl"
#include "graphics3d.blendmode.inl"
#include "graphics3d.stencilmode.inl"
#include "graphics3d.matrixstack.inl"
#include "graphics3d.lightstack.inl"
#include "graphics3d.context.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.texture2d.inl"


namespace magnum { namespace gui3d {
using magnum::graphics3d::MatrixStack;
using magnum::graphics3d::Shader;
using magnum::graphics3d::Material;
using magnum::graphics3d::FrameBuffer;
using magnum::graphics3d::DepthMode;
using magnum::graphics3d::PolygonMode;
using magnum::graphics3d::BlendMode;
using magnum::graphics3d::StencilMode;
using magnum::graphics3d::Texture;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Context::Context(magnum::graphics3d::Context* c) {
	ASSERT(c != NULL);
	m_Context = c;
	m_MatrixStack = c->getMatrixStack();
	m_Size = c->getSize();
	m_Mode = None;
	m_Bitmap = NULL;
	m_Font = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::start() {
	// Reset render context
	Material::selectNone();
	Shader::selectNone();
	DepthMode::selectNone();
	PolygonMode::selectNone();
	BlendMode::selectNone();
	StencilMode::selectNone();

	m_Mode = None;
	m_Size = m_Context->getSize();
	
	// Save current matrices and setup new
	m_MatrixStack->selectMode(MatrixStack::Projection);
	m_MatrixStack->push();
	m_MatrixStack->identity();
	m_MatrixStack->ortho(0.0f, 0.0f, (float)m_Size.getWidth(), (float)m_Size.getHeight(), -1.0f, 1.0f);
	m_MatrixStack->selectMode(MatrixStack::Texture);
	m_MatrixStack->push();
	m_MatrixStack->identity();
	m_MatrixStack->selectMode(MatrixStack::ModelView);
	m_MatrixStack->push();
	m_MatrixStack->identity();

	// Disable culling
	m_Context->setCulling(magnum::graphics3d::Context::CullNone);
	
	// Disable Z-Buffer
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);
	
	// Disable lighting and materials
	glDisable(GL_LIGHTING);	
	glDisable(GL_COLOR_MATERIAL);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_TEXTURE_3D);
	glDisable(GL_TEXTURE_CUBE_MAP);
	
	// Enable Clipping
	m_Context->setClipRectangle(Rectangle<int>(0, 0, m_Size.getWidth(), m_Size.getHeight()));
	m_Context->setClipping(true);

	glShadeModel(GL_FLAT);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::finish() {
	// CHeck if we need to end a call
	resetMode();

	// Disable Clipping
	m_Context->setClipping(false);

	// Reset smooth shading
	glShadeModel(GL_SMOOTH);
	
	// Restore matrices
	m_MatrixStack->selectMode(MatrixStack::ModelView);
	m_MatrixStack->pop();
	m_MatrixStack->selectMode(MatrixStack::Texture);
	m_MatrixStack->pop();
	m_MatrixStack->selectMode(MatrixStack::Projection);
	m_MatrixStack->pop();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::drawRectangle(const Rectangle<float>& r) {
	if (m_Mode != None)
		resetMode();
	glBegin(GL_LINE_LOOP);
	glVertex2f(r.getX(), r.getY());
	glVertex2f(r.getX(), r.getY() + r.getHeight());
	glVertex2f(r.getX() + r.getWidth(), r.getY() + r.getHeight());
	glVertex2f(r.getX() + r.getWidth(), r.getY());
	glEnd();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::fillRectangle(const Rectangle<float>& r) {
	if (m_Mode != Rectangles) {
		resetMode();
		glBegin(GL_QUADS);
		m_Mode = Rectangles;
	}
	glVertex2f(r.getX(), r.getY());
	glVertex2f(r.getX(), r.getY() + r.getHeight());
	glVertex2f(r.getX() + r.getWidth(), r.getY() + r.getHeight());
	glVertex2f(r.getX() + r.getWidth(), r.getY());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::drawPoint(const Point<float>& p) {
	if (m_Mode != Points) {
		resetMode();
		glBegin(GL_POINTS);
		m_Mode = Points;
	}
	glVertex2f(p.getX(), p.getY());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::drawPoints(const Point<float>* p, unsigned int n) {
	if (m_Mode != Points) {
		resetMode();
		glBegin(GL_POINTS);
		m_Mode = Points;
	}
	const Point<float>* end = p + n;
	while(p < end) {
		glVertex2f(p->getX(), p->getY());
		p++;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::drawLine(const Point<float>& a, const Point<float>& b) {
	if (m_Mode != Lines) {
		resetMode();
		glBegin(GL_LINES);
		m_Mode = Lines;
	}
	glVertex2f(a.getX(), a.getY());
	glVertex2f(b.getX(), b.getY());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::drawLines(const Point<float>* p, unsigned int n) {
	if (m_Mode != Lines) {
		resetMode();
		glBegin(GL_LINES);
		m_Mode = Lines;
	}
	const Point<float>* end = p + n;
	while(p < end) {
		glVertex2f(p->getX(), p->getY());
		p++;
		glVertex2f(p->getX(), p->getY());
		p++;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::drawPolygon(const Point<float>* p, unsigned int n) {
	if (m_Mode != None)
		resetMode();

	glBegin(GL_LINE_LOOP);
	const Point<float>* end = p + n;
	while(p < end) {
		glVertex2f(p->getX(), p->getY());
		p++;
	}
	glEnd();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::fillPolygon(const Point<float>* p, unsigned int n) {
	if (m_Mode != None)
		resetMode();

	glBegin(GL_POLYGON);
	const Point<float>* end = p + n;
	while(p < end) {
		glVertex2f(p->getX(), p->getY());
		p++;
	}
	glEnd();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::drawBitmap(const Point<float>& p) {
	ASSERT(m_Bitmap != NULL);
	if (m_Mode != Textures) {
		resetMode();
		m_Mode = Textures;
		m_Bitmap->select(0);
		glEnable(GL_TEXTURE_2D);
		glColor3f(1.0f, 1.0f, 1.0f);
		glBegin(GL_QUADS);
	}
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(p.getX(), p.getY());
	glTexCoord2f(1.0f, 0.0f);
	glVertex2f(p.getX() + m_Bitmap->getWidth(), p.getY());
	glTexCoord2f(1.0f, 1.0f);
	glVertex2f(p.getX() + m_Bitmap->getWidth(), p.getY() + m_Bitmap->getHeight());
	glTexCoord2f(0.0f, 1.0f);
	glVertex2f(p.getX(), p.getY() + m_Bitmap->getHeight());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::drawBitmap(const Point<float>& p,const Rectangle<float>& r) {
	ASSERT(m_Bitmap != NULL);
	if (m_Mode != Textures) {
		resetMode();
		m_Mode = Textures;
		m_Bitmap->select(0);
		glEnable(GL_TEXTURE_2D);
		glColor3f(1.0f, 1.0f, 1.0f);
		glBegin(GL_QUADS);
	}
	
	float w = (float)m_Bitmap->getWidth();
	float h = (float)m_Bitmap->getHeight();
	float bx0 = (float)r.getX() / w;
	float by0 = (float)r.getY() / h;
	float bx1 = (float)(r.getX() + r.getWidth()) / w;
	float by1 = (float)(r.getY() + r.getHeight()) / h;
	
	glTexCoord2f(bx0, by0);
	glVertex2f(p.getX(), p.getY());
	glTexCoord2f(bx1, bx0);
	glVertex2f(p.getX() + r.getWidth(), p.getY());
	glTexCoord2f(bx1, by1);
	glVertex2f(p.getX() + r.getWidth(), p.getY() + r.getHeight());
	glTexCoord2f(bx0, by1);
	glVertex2f(p.getX(), p.getY() + r.getHeight());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::drawBitmap(const Rectangle<float>& r) {
	ASSERT(m_Bitmap != NULL);
	if (m_Mode != Textures) {
		resetMode();
		m_Mode = Textures;
		m_Bitmap->select(0);
		glEnable(GL_TEXTURE_2D);
		glColor3f(1.0f, 1.0f, 1.0f);
		glBegin(GL_QUADS);
	}
	
	float w = r.getWidth();
	float h = r.getHeight();
	float bx1 = w / (float)m_Bitmap->getWidth();
	float by1 = h / (float)m_Bitmap->getHeight();
	
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(r.getX(), r.getY());
	glTexCoord2f(bx1, 0.0f);
	glVertex2f(r.getX() + w, r.getY());
	glTexCoord2f(bx1, by1);
	glVertex2f(r.getX() + w, r.getY() + h);
	glTexCoord2f(0.0f, by1);
	glVertex2f(r.getX(), r.getY() + h);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::drawBitmap(const Rectangle<float>& r,const Point<float>& p) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::drawString(const Point<float>& p,const char* str) {
	ASSERT(str != NULL);
	ASSERT(m_Font != NULL);
	if (m_Mode != Strings) {
		resetMode();
		m_Mode = Strings;
		m_Font->getBitmap()->select(0);
		glEnable(GL_TEXTURE_2D);
		glColor3f(m_Colour.R, m_Colour.G, m_Colour.B);
		glEnable(GL_BLEND);
		glBegin(GL_QUADS);
	}
	
	float x,y, xx;
	x = p.getX();
	y = p.getY();
	
	while(*str != 0) {
		const FontMap::Glyph* glyph = &m_Font->m_Glyphs[*str];
		xx = floor(x + glyph->bh);
		glTexCoord2f(glyph->tx0, glyph->ty0);
		glVertex2f(xx, y - glyph->bv);
		glTexCoord2f(glyph->tx1, glyph->ty0);
		glVertex2f(xx + glyph->width, y - glyph->bv);
		glTexCoord2f(glyph->tx1, glyph->ty1);
		glVertex2f(xx + glyph->width, y + glyph->depth);
		glTexCoord2f(glyph->tx0, glyph->ty1);
		glVertex2f(xx, y + glyph->depth);
		x += glyph->advance;
		str++;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Rectangle<float> Context::getStringExtents(const char* str) const {
	float x = 0.0f,y = 0.0f,w = 0.0f,h = 0.0f;
	while(*str != 0) {
		const FontMap::Glyph* glyph = &m_Font->m_Glyphs[*str];
		w += glyph->advance;
		if (y < glyph->bv)
			y = glyph->bv;
		if (h < glyph->depth)
			h = glyph->depth;
		str++;
	};
	
	return Rectangle<float>(x,-y,w,h+y);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Context::resetMode() {
	if (m_Mode == None)
		return;
	glEnd();
	switch (m_Mode) {
		case Textures:
			glDisable(GL_TEXTURE_2D);
			break;
		case Strings:
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_BLEND);
			break;
	}
	m_Mode = None;
}

}; };
