/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "gui3d.root.h"
#include "gui3d.view.h"

#include "system.thread.inl"
#include "container.chain.inl"
#include "container.iterator.inl"
#include "container.queue.inl"
#include "application.message.inl"
#include "application.messagequeue.inl"
#include "graphics.dimension.inl"
#include "graphics.rectangle.inl"
#include "graphics.point.inl"
#include "graphics.transform.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.framebuffer.inl"
#include "graphics3d.context.inl"
#include "object.pointer.inl"
#include "gui3d.root.inl"
#include "gui3d.view.inl"
#include "gui3d.context.inl"


namespace magnum { namespace gui3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Root::Root(magnum::graphics3d::Context* c, bool di) : m_Context(c) {
	m_Focus = NULL;
	m_MouseView = NULL;
	m_KeyboardModifiers = 0;
	m_Size = m_OldSize = c->getFrameBuffer()->getSize();
	attachThread(Thread::getCurrent());
#if defined(__FORWINDOWS__)
	m_WindowHandle = c->getWindowHandle();
	m_DeviceContext = c->getDeviceHandle();
	m_WindowProc = (WNDPROC)GetWindowLongPtr(m_WindowHandle, GWL_WNDPROC);
	ASSERT(m_WindowProc != NULL);
	SetWindowLongPtr(m_WindowHandle, GWL_USERDATA, (LONG_PTR)this);
	if (di)
		SetWindowLongPtr(m_WindowHandle, GWL_WNDPROC, (LONG_PTR)inputWindowProc);
	else
		SetWindowLongPtr(m_WindowHandle, GWL_WNDPROC, (LONG_PTR)stdWindowProc);
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Root::~Root() {
	clear();
	detachThread();
#if defined(__FORWINDOWS__)
	if (m_WindowProc != NULL) {
		SetWindowLongPtr(m_WindowHandle, GWL_WNDPROC, (LONG_PTR)m_WindowProc);	
		m_WindowProc = NULL;
	}
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::insert(View* view) {
	if (view->getRoot() != NULL)
		return false;
	if (view->getParent() != NULL)
		return false;

	m_Views.insertFirst(view);
	insertHandler(view);
	view->m_Root = this;
	view->m_Context = &m_Context;

	view->onAttach();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::remove(View *view) {
	if (view->getRoot() != this)
		return false;
	if (view->getParent() != NULL)
		return false;

	if (m_Focus == view) 
		releaseFocus();
	if (m_MouseView == view)
		m_MouseView = NULL;

	view->onDetach();

	view->m_Root = NULL;
	view->m_Context = NULL;
	m_Views.remove(view);
	removeHandler(view);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::clear() {
	View *view;

	releaseFocus();
	m_MouseView = NULL;

	while(m_Views.begin() != m_Views.end()) {
		view = &(*m_Views.begin());

		view->onDetach();

		view->m_Root = NULL;
		view->m_Context = NULL;
		m_Views.remove(view);
		removeHandler(view);
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::setFocus(View* focus) {
	if ((focus != NULL) && (focus->m_Root != this))
		return false;
	if (m_Focus == focus)
		return true;
	if (!focus->isEnabled())
		return false;

	if (m_Focus != NULL) {
		m_Focus->onReleaseFocus();
	}
	m_Focus = focus;
	if (m_Focus != NULL) {
		m_Focus->onGetFocus();
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::releaseFocus(View* focus) {
	if (focus == NULL)
		return true;
	if (focus->m_Root != this)
		return false;
	if (m_Focus != focus)
		return false;

	focus->onReleaseFocus();
	m_Focus = NULL;

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::releaseFocus() {
	if (m_Focus != NULL) {
		m_Focus->onReleaseFocus();
	}
	m_Focus = NULL;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Root::setSize(const Dimension<float>& dim) {
	m_Size = dim;
	onResize(m_Size);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Root::attachDirectInput() {
	m_MouseTracking = false;
	SetWindowLongPtr(m_WindowHandle, GWL_WNDPROC, (LONG_PTR)inputWindowProc);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Root::detachDirectInput() {
	m_MouseTracking = false;
	SetWindowLongPtr(m_WindowHandle, GWL_WNDPROC, (LONG_PTR)stdWindowProc);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::getDirectInput() const {
#if defined(__FORWINDOWS__)
	WNDPROC wp = (WNDPROC)GetWindowLongPtr(m_WindowHandle, GWL_WNDPROC);
	return (wp == &inputWindowProc);
#else
	return false;
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Root::process() {
	Message* msg;
	while(true) {
		m_MessageMutex.lock();
		msg = m_MessageQueue.removeHead();
		m_MessageMutex.unlock();
		if (msg == NULL)
			return;
		msg->getTarget()->handleMessage(*msg);
		msg->release();
	};
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Root::paint() {
	m_Context.start();
	m_Context.setClipping(Rectangle<int>(Point<int>(0,0),m_Size));
	
	// Paint children
	Chain<View>::Iterator iter = m_Views.rbegin();
	while (iter != m_Views.rend()) {
		m_Context.setTransform(iter->m_AbsTransform);
		iter->onPaint();
		iter--;
	}
	
	m_Context.finish();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onResize(const Dimension<float>& ) {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;
	float dw, dh;

	dw = m_Size.getWidth() - m_OldSize.getWidth();
	dh = m_Size.getHeight() - m_OldSize.getHeight();

	if ((dw == 0)  && (dh == 0)) {
		m_OldSize = m_Size;
		return true;
	}

	// Now resize all childs (if needed)
	while (iter != m_Views.end()) {
		view = &(*iter);
		float x,y,w,h;

		if (view->m_Alignement == 0) {
			iter++;
			continue;
		}

		if ((view->m_Alignement & View::FollowLeftRight) == View::FollowLeftRight) {
			x = view->m_Frame.getX();
			w = view->m_Frame.getWidth() + dw;
		} else if (view->m_Alignement & View::FollowLeft) {
			x = view->m_Frame.getX();
			w = view->m_Frame.getWidth();
		} else if (view->m_Alignement & View::FollowRight) {
			x = view->m_Frame.getX() + dw;
			w = view->m_Frame.getWidth();
		} else {
			x = view->m_Frame.getX();
			w = view->m_Frame.getWidth();
		}

		if ((view->m_Alignement & View::FollowTopBottom) == View::FollowTopBottom) {
			y = view->m_Frame.getY();
			h = view->m_Frame.getHeight() + dh;
		} else if (view->m_Alignement & View::FollowTop) {
			y = view->m_Frame.getY();
			h = view->m_Frame.getHeight();
		} else if (view->m_Alignement & View::FollowBottom) {
			y = view->m_Frame.getY() + dh;
			h = view->m_Frame.getHeight();
		} else {
			y = view->m_Frame.getY();
			h = view->m_Frame.getHeight();
		}

		view->setFrame(Rectangle<float>(x, y, w, h));
		iter++;
	}

	m_OldSize = m_Size;

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onKeyDown(Keyboard::Key key) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onKeyUp(Keyboard::Key key) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onChar(Character::Code code, Character::Modifier mod) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onMouseMove(const Point<float>& pos, Mouse::Button button) {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;

	for (;iter != m_Views.end(); ++iter) {
		view = &(*iter);
		if (view->getFrame().isInside(pos)) {
			if (!view->isEnabled())
				continue;
			if (m_MouseView != view) {
				if (m_MouseView != NULL)
					m_MouseView->onMouseLeave();
				m_MouseView = view;
				view->onMouseEnter(pos, button);
			}
			if (view->onMouseMove(view->convertFromParent(pos), button))
				return true;
		}
	}
	if (m_MouseView != NULL) {
		m_MouseView->onMouseLeave();
		m_MouseView = NULL;
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onMouseButton(const Point<float>& pos, Mouse::Button button) {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;

	for (;iter != m_Views.end(); ++iter) {
		view = &(*iter);
		if (view->getFrame().isInside(pos)) {
			if (!view->isEnabled())
				continue;
			if (view->onMouseButton(view->convertFromParent(pos), button))
				return true;
		}
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onMouseWheel(float ,Mouse::Button ) {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onMouseEnter(const Point<float>& pos, Mouse::Button button) {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;

	for (;iter != m_Views.end(); ++iter) {
		view = &(*iter);
		if (view->getFrame().isInside(pos)) {
			if (!view->isEnabled())
				continue;
			if (view->onMouseEnter(view->convertFromParent(pos), button))
				return true;
		}
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onMouseLeave() {
	if (m_MouseView == NULL)
		return true;
	m_MouseView->onMouseLeave();
	m_MouseView = NULL;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onActivate(void) {
	// activate all children
	Chain<View>::Iterator iter = m_Views.begin();

	while (iter != m_Views.end()) {
		iter->onActivate();
		iter++;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onDeactivate(void) {
	// deactivate all children
	Chain<View>::Iterator iter = m_Views.begin();

	while (iter != m_Views.end()) {
		iter->onDeactivate();
		iter++;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onSendMessage(Message* msg) {
	m_MessageMutex.lock();
	m_MessageQueue.insertHead(msg);
	m_MessageMutex.unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onPostMessage(Message* msg) {
	m_MessageMutex.lock();
	m_MessageQueue.insert(msg);
	m_MessageMutex.unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Root::onSynchronize() {
	ASSERT(false);
	return false;
}


}; };
