/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "gui3d.slider.h"
#include "gui3d.context.h"
#include "gui3d.root.h"

#include "object.pointer.inl"
#include "graphics.dimension.inl"
#include "math.range.inl"
#include "gui3d.slider.inl"
#include "gui3d.control.inl"
#include "gui3d.context.inl"
#include "gui3d.view.inl"


namespace magnum { namespace gui3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Slider::Slider(const Rectangle<float>& r,const Message& msg,const Range<float>& rng) : Control(r,msg) {
	m_Range = rng;
	m_Step = 0.0f;
	m_Value = m_Range.getStart();
	init();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Slider::Slider(const Rectangle<float>& r,const Message& msg) : Control(r,msg) {
	m_Range.setStart(0.0f);
	m_Range.setEnd(1.0f);
	m_Step = 0.0f;
	m_Value = 0.0f;
	init();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Slider::init() {
	m_State = NormalState;

	m_SliderColour = Colour<float>(0.9f, 0.9f, 0.9f);
	m_BorderColour = Colour<float>(0.3f, 0.3f, 0.3f);

	m_HoverSliderColour = Colour<float>(0.9f, 0.9f, 0.9f);
	m_HoverBorderColour = Colour<float>(0.4f, 0.4f, 0.4f);

	m_PressedSliderColour = Colour<float>(1.0f, 1.0f, 1.0f);
	m_PressedBorderColour = Colour<float>(1.0f, 0.8f, 0.2f);

	m_DisabledSliderColour = Colour<float>(0.4f, 0.4f, 0.4f);
	m_DisabledBorderColour = Colour<float>(0.3f, 0.3f, 0.3f);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Slider::onPaint() {
	Context* context = getContext();
	
	const Colour<float>* borderColour;
	const Colour<float>* sliderColour;

	if (m_State == NormalState) {
		borderColour = &m_BorderColour;
		sliderColour = &m_SliderColour;
	} else if (m_State == PressedState) {
		borderColour = &m_PressedBorderColour;
		sliderColour = &m_PressedSliderColour;
	} else if (m_State == HoveringState) {
		borderColour = &m_HoverBorderColour;
		sliderColour = &m_HoverSliderColour;
	} else {
		borderColour = &m_DisabledBorderColour;
		sliderColour = &m_DisabledSliderColour;
	}

	context->setColour(*borderColour);
	Point<float> p = getBounds().getOrigin();
	p.setY(p.getY() + getBounds().getHeight()*0.5f - 1.0f);
	Rectangle<float> rect;
	rect.setOrigin(p);
	rect.setHeight(3);
	rect.setWidth(getBounds().getWidth());
	context->fillRectangle(rect);
	
	context->setColour(*sliderColour);
	p.setX(p.getX() + (getBounds().getWidth()-4) / m_Range.getSize() * (m_Value - m_Range.getStart()));
	p.setY(getBounds().getOrigin().getY()+1.0f);
	rect.setOrigin(p);
	rect.setHeight(getBounds().getHeight()-1.0f);
	rect.setWidth(4);
	context->fillRectangle(rect);
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Slider::onEnable() {
	m_State = NormalState;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Slider::onDisable() {
	m_State = DisabledState;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Slider::onMouseMove(const Point<float>& pos, Mouse::Button button) {
	if (m_State == PressedState) {
		// Recalc slider position
		m_Value = pos.getX()*m_Range.getSize()/(getBounds().getWidth()-4) + m_Range.getStart(); 
		m_Value = m_Range.clamp(m_Value);
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Slider::onMouseEnter(const Point<float>&, Mouse::Button buttons) {
	m_State = HoveringState;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Slider::onMouseButton(const Point<float>& pos, Mouse::Button buttons) {
	if (buttons & Mouse::LeftDown) {
		// Recalc slider position
		m_Value = pos.getX()*m_Range.getSize()/(getBounds().getWidth()-4) + m_Range.getStart(); 
		m_Value = m_Range.clamp(m_Value);

		if (m_State != PressedState) {
			m_State = PressedState;
			onPush();
		}
		return true;
	}
	if (buttons & Mouse::LeftUp) {
		if (m_State == PressedState) {
			onRelease();
			onAction();
		}
		m_State = HoveringState;
		return true;
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Slider::onMouseLeave() {
	if (m_State == PressedState)
		onRelease();
	m_State = NormalState;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Slider::onPush() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Slider::onRelease() {
	return true;
}

}; };
