/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "system.thread.h"
#include "geometry.transform.h"
#include "input.p5glove.h"

#include "math.vector3.inl"
#include "math.matrix.inl"
#include "system.thread.inl"
#include "container.allocator.inl"
#include "container.chain.inl"
#include "container.list.inl"
#include "geometry.transform.inl"
#include "input.device.inl"
#include "input.marker.inl"
#include "input.body.inl"
#include "input.glove.inl"
#include "input.p5glove.inl"

#include <P5dll.h>


namespace magnum { namespace input {
using namespace magnum::math;
using magnum::system::Thread;
using magnum::geometry::Transform3;

static CP5DLL	GloveDll;
int P5Glove::m_References = 0;


/*--------------------------------------------------------------------------*/
/**
 */
unsigned int P5Glove::getGloves() {
	dllRefer();
	return GloveDll.m_nNumP5;
	dllRelease();
}


/*--------------------------------------------------------------------------*/
/**
 */
void P5Glove::dllRefer() {
	if (m_References == 0) {
		GloveDll.P5_Init();
		Thread::sleep(500);
	}
	m_References++;
}


/*--------------------------------------------------------------------------*/
/**
 */
void P5Glove::dllRelease() {
	if(m_References > 0) {
		m_References--;
		if (m_References == 0) {
			GloveDll.P5_Close();	
		}
	}
}


/*--------------------------------------------------------------------------*/
/**
 */
P5Glove::P5Glove(unsigned int idx) {
	dllRefer();
	GloveDll.P5_SetMouseState(idx, 0);
	m_Glove = idx;
	m_FilterIndex = 0;
	for (unsigned int i = 0; i < FilterSize; ++i)
		onLoop();
}


/*--------------------------------------------------------------------------*/
/**
 */
P5Glove::~P5Glove() {
	dllRelease();
}


/*--------------------------------------------------------------------------*/
/**
 */
void P5Glove::onLoop() {
	if ((m_Glove >= GloveDll.m_nNumP5) || (GloveDll.m_P5Devices == NULL))
		return;
		
	m_FilterIndex = (m_FilterIndex + 1) % FilterSize;
	
	P5Data* data = &GloveDll.m_P5Devices[m_Glove];
	m_RawPositions[m_FilterIndex].x = data->m_fx / 2.54f;
	m_RawPositions[m_FilterIndex].y = data->m_fy / 2.54f;
	m_RawPositions[m_FilterIndex].z = data->m_fz / 2.54f;
	m_RawRotations[m_FilterIndex].x = data->m_fyaw;
	m_RawRotations[m_FilterIndex].x = data->m_fpitch;
	m_RawRotations[m_FilterIndex].x = data->m_froll;
	
	updateFinger(Thumb, float(data->m_byBendSensor_Data[0]) / 63.0f);
	updateFinger(Index, float(data->m_byBendSensor_Data[1]) / 63.0f);
	updateFinger(Middle, float(data->m_byBendSensor_Data[2]) / 63.0f);
	updateFinger(Ring, float(data->m_byBendSensor_Data[3]) / 63.0f);
	updateFinger(Pinky, float(data->m_byBendSensor_Data[4]) / 63.0f);
	
	filterXYZ();
	filterYPR();
	
	Transform3<float> trans;
	trans.setRotationY(m_Rotation.x);
	trans.rotateX(m_Rotation.y);
	trans.rotateZ(m_Rotation.z);
	trans.setTranslation(m_Position);
	updateTransform(trans);
	
	fireMove(getTransform(), (Button)0);
}


/*--------------------------------------------------------------------------*/
/**
 */
void P5Glove::onCalibrate() {
	GloveDll.P5_CalibrateBendSensors(m_Glove);
	GloveDll.P5_CalibratePositionData(m_Glove);
}


/*--------------------------------------------------------------------------*/
/**
 */
void P5Glove::filterXYZ() {
	//We are going to implement a dynamic filter, which will flush the filter array values at different rates
	//based on the rate of change of the users hand.  This will allow for greater latency of motion.
	//The setpoint determines the number of pixel motion that will flush the entire filter.
	//The idea is when the user doensnt move much, we average alot of frames, but during fast motion, we
	//average fewer and fewer frames in order to reduce latency
	#define FLUSH_SETPOINT	4.0f

	unsigned int len;
	float val, flen;
	Vector3<float> curvec = m_RawPositions[m_FilterIndex % FilterSize];
	Vector3<float> lastvec = m_RawPositions[(m_FilterIndex-1) % FilterSize];
	
	flen = floor(FilterSize * FLUSH_SETPOINT / abs(curvec.x - lastvec.x) + 2.0f);
	if (flen > FilterSize)
		len = FilterSize;
	else
		len = unsigned int(flen);
	val = m_RawPositions[m_FilterIndex % FilterSize].x;
	for (unsigned int i = 1; i < len; i++)
		val += m_RawPositions[(m_FilterIndex - i) % FilterSize].x;
	m_Position.x = val / len;

	flen = floor(FilterSize * FLUSH_SETPOINT / abs(curvec.y - lastvec.y) + 2.0f);
	if (flen > FilterSize)
		len = FilterSize;
	else
		len = unsigned int(flen);
	val = m_RawPositions[m_FilterIndex % FilterSize].y;
	for (unsigned int i = 1; i < len; i++)
		val += m_RawPositions[(m_FilterIndex - i) % FilterSize].y;
	m_Position.y = val / len;

	flen = floor(FilterSize * FLUSH_SETPOINT / abs(curvec.z - lastvec.z) + 2.0f);
	if (flen > FilterSize)
		len = FilterSize;
	else
		len = unsigned int(flen);
	val = m_RawPositions[m_FilterIndex % FilterSize].z;
	for (unsigned int i = 1; i < len; i++)
		val += m_RawPositions[(m_FilterIndex - i) % FilterSize].z;
	m_Position.z = val / len;
}



/*--------------------------------------------------------------------------*/
/**
 */
void P5Glove::filterYPR() {
	Vector3<float> curvec = m_RawRotations[m_FilterIndex % FilterSize];
	Vector3<float> lastvec = m_RawRotations[(m_FilterIndex-1) % FilterSize];
	Vector3<float> val;
	
	val = m_RawRotations[m_FilterIndex % FilterSize];
	for (unsigned int i = 1; i < FilterSize; i++)
		val += m_RawRotations[(m_FilterIndex - i) % FilterSize];
	m_Rotation = val / FilterSize;
}

}; };
