/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "math.consts.h"
#include "input3d.flightcontroller.h"

#include "object.inl"
#include "base.string.inl"
#include "math.matrix.inl"
#include "math.vector.inl"
#include "math.vector2.inl"
#include "math.vector3.inl"
#include "geometry.transform.inl"
#include "container.list.inl"
#include "container.allocator.inl"
#include "application.listener.inl"
#include "graphics.point.inl"
#include "graphics.dimension.inl"
#include "input3d.controller.inl"
#include "input3d.flightcontroller.inl"

namespace magnum { namespace input3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
FlightController::FlightController() {
	m_Position.setX(0.0f);
	m_Position.setY(0.0f);
	m_Center.setX(0.0f);
	m_Center.setY(0.0f);
	m_Size.setWidth(2.0f);
	m_Size.setHeight(2.0f);
	m_Motion = 1.0f;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FlightController::onMouseArea(const Rectangle<float>& rect) {
	m_Center = rect.getOrigin() + 0.5f*rect.getSize();
	m_Size = 0.5f*rect.getSize();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FlightController::onMouseMove(const Point<float>& pnt,Mouse::Button buttons) {
	Point<float> newpos((pnt.x - m_Center.x) / m_Size.getWidth(), -(pnt.y - m_Center.y) / m_Size.getHeight());
	
    if (buttons % Mouse::AnyDown) {
		if (buttons & Mouse::LeftDown)
			m_Motion = -0.1f;
		else if (buttons & Mouse::RightDown)
			m_Motion = 0.1f;    
	    float distanceX = -(m_Position.x - newpos.x);
	    float distanceY =  (m_Position.y - newpos.y);
	    rotate(distanceX, distanceY);
	}
    m_Position = newpos;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FlightController::onMouseButton(const Point<float>& pnt,Mouse::Button button) {
	m_Position = Point<float>((pnt.x - m_Center.x) / m_Size.getWidth(), -(pnt.y - m_Center.y) / m_Size.getHeight());
    if (button == Mouse::LeftDown)
		m_Motion = -0.1f;
    else if (button == Mouse::RightDown) 
		m_Motion = 0.1f;    
    else if ((button == Mouse::LeftUp) || (button == Mouse::RightUp))
		m_Motion = 0.0f;    

}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FlightController::onMouseWheel(float f) {
	forward(f*m_Speed);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FlightController::onChar(Character::Code ,Character::Modifier ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FlightController::onKeyDown(Keyboard::Key key) {
	switch(key) {
		case Keyboard::Left:
			right(m_Speed);   
			break;
		case Keyboard::Right:
			right(-m_Speed);
			break;
		case Keyboard::Up:
			forward(-m_Speed); 
			break;
		case Keyboard::Down:
			forward(m_Speed);  
			break;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FlightController::onKeyUp(Keyboard::Key ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FlightController::onUpdate(float time) {
	forward(time*m_Motion*m_Speed);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FlightController::onModify() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FlightController::onReset() {
	Transform3<float> t;
	t.setAxis(0,m_Right);
	t.setAxis(1,m_Up);
	t.setAxis(2,m_Direction);
	t.setTranslation(Vector3<float>::Null);
	set(t);
    m_Motion = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FlightController::rotate(float deltaX, float deltaY) {
    // rotate around the up vector
    Transform3<float> mtx;

    // rotate around the side vector
	Vector3<float> right = getAxis(0);
    Vector3<float> dir = getAxis(2);
    mtx.identity();
    mtx.setRotation(-deltaY,right);
	dir = mtx * dir;

    mtx.setRotation(-deltaX, m_Up);
	dir = mtx * dir;
	right = mtx * right;
    
    setAxis(0,dir);
	setAxis(1,dir % right);
	setAxis(2,right);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FlightController::forward(float step) {
    Vector3<float> lv = getAxis(2);
    lv.normalise();
    translate(step * lv);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FlightController::right(float step) {
    float angle = -0.19634954f / 8.0f;
    rotate(step*angle, 0);
}


}; };
