/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "math.consts.h"
#include "math.linearsolver.h"
#include "input3d.trackballcontroller.h"

#include "object.inl"
#include "base.string.inl"
#include "math.matrix.inl"
#include "math.vector.inl"
#include "math.vector2.inl"
#include "math.vector3.inl"
#include "math.linearsolver.inl"
#include "geometry.transform.inl"
#include "container.list.inl"
#include "container.allocator.inl"
#include "application.listener.inl"
#include "graphics.point.inl"
#include "graphics.dimension.inl"
#include "input3d.controller.inl"
#include "input3d.trackballcontroller.inl"

namespace magnum { namespace input3d {
using namespace magnum::math;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TrackballController::TrackballController() {
	m_Radius = 0.8f;
	m_Radius2 = 0.64f;
	m_Position.setX(0.0f);
	m_Position.setY(0.0f);
	m_MouseCenter.setX(0.0f);
	m_MouseCenter.setY(0.0f);
	m_MouseSize.setWidth(2.0f);
	m_MouseSize.setHeight(2.0f);
	m_Speed = 1.0f;
	m_Rotation.identity();
	m_Translation.null();
	m_Center.null();
	m_AutoCenter = false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TrackballController::rotate(const Point<float>& mfrom, const Point<float>& mto) {
	Point<float> from((mfrom.getX() - m_MouseCenter.getX()) / m_MouseSize.getWidth(), (mfrom.getY() - m_MouseCenter.getY()) / m_MouseSize.getHeight());
	Point<float> to((mto.getX() - m_MouseCenter.getX()) / m_MouseSize.getWidth(), (mto.getY() - m_MouseCenter.getY()) / m_MouseSize.getHeight());

	Vector3<float> p1(from.getX(), from.getY(), projectToSphere(from));
    Vector3<float> p2(to.getX(), to.getY(), projectToSphere(to));

    Vector3<float> axis = p2 % p1;
    axis.y = -axis.y;
    Vector3<float> diff = p1 - p2;
    if (axis.length() < 0.001f)
		return;

    float tmp = min(diff.length() / (2.0f * m_Radius), 1.0f);
    float phi = -2.0f * asin(tmp);
    m_Rotation.rrotate(phi, axis);
    m_Rotation.setTranslation(0.5f*(m_Center - m_Rotation.getRotation() * m_Center));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TrackballController::translateXY(const Point<float>& mfrom, const Point<float>& mto) {
	Vector2<float> from(mfrom.getX(), mfrom.getY());
	Vector2<float> to(mto.getX(), mto.getY());
	Vector2<float> diff = m_Speed * (from - to);
	
	if (m_AutoCenter)
		m_Translation += m_Rotation * Vector3<float>(diff.x, -diff.y, 0);
	else
		m_Translation += Vector3<float>(diff.x, -diff.y, 0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TrackballController::translateZ(const Point<float>& mfrom, const Point<float>& mto) {
	float diff = m_Speed *(mto.getY() - mfrom.getY());	

	if (m_AutoCenter)
		m_Translation += diff * m_Rotation.getAxis(2);
	else
		m_Translation.z += diff;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TrackballController::setAutoCenter(bool c) {
	if (c == m_AutoCenter)
		return;
	m_AutoCenter = c;
	if (m_AutoCenter)
		m_Translation = m_Rotation * m_Translation;
	else
		m_Translation = m_Rotation.getInverse() * m_Translation;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TrackballController::onMouseArea(const Rectangle<float>& rect) {
	m_MouseCenter = rect.getOrigin() + 0.5f*rect.getSize();
	m_MouseSize = 0.5f*rect.getSize();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TrackballController::onMouseMove(const Point<float>& p,Mouse::Button b) {
	if (b & Mouse::LeftDown)
		rotate(m_Position, p);
	if (b & Mouse::RightDown) 
		translateXY(m_Position, p);
	if (b & Mouse::MiddleDown) 
		translateZ(m_Position, p);
	m_Position = p;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TrackballController::onMouseButton(const Point<float>& ,Mouse::Button ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TrackballController::onMouseWheel(float v) {
	if (m_AutoCenter)
		m_Translation += m_Speed*v*m_Rotation.getAxis(2);
	else
		m_Translation.z += m_Speed*v;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TrackballController::onUpdate(float ) {
#if 0
	PRINT("TrackballController::onUpdate()");
	PRINT4("  Rotation -> [%f, %f, %f, %f]",m_Rotation[0][0],m_Rotation[0][1],m_Rotation[0][2],m_Rotation[0][3]);
	PRINT4("  Rotation -> [%f, %f, %f, %f]",m_Rotation[1][0],m_Rotation[1][1],m_Rotation[1][2],m_Rotation[1][3]);
	PRINT4("  Rotation -> [%f, %f, %f, %f]",m_Rotation[2][0],m_Rotation[2][1],m_Rotation[2][2],m_Rotation[2][3]);
	PRINT4("  Rotation -> [%f, %f, %f, %f]",m_Rotation[3][0],m_Rotation[3][1],m_Rotation[3][2],m_Rotation[3][3]);
	PRINT3("  Translation -> [%f, %f, %f]",m_Translation[0],m_Translation[1],m_Translation[2]);
#endif
	set(m_Rotation);
	if (m_AutoCenter)
		translate(m_Translation);
	else
		translate(m_Rotation * m_Translation);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TrackballController::onModify() {
	m_Rotation.setRotation(getTransform().getRotation());
    m_Rotation.setTranslation(m_Center - m_Rotation.getRotation() * m_Center);
    
    Transform3<float> mtx = m_Rotation;
    mtx.invert();
	m_Translation = mtx * getTransform().getTranslation();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TrackballController::onReset() {
	identity();
	m_Rotation.identity();
	m_Translation = Vector3<float>(0,0,0);
}

}; };
