/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "io.exception.h"
#include "io.archive.pak.h"

#include "io.file.inl"
#include "io.fileinfo.inl"
#include "io.directory.inl"
#include "io.archive.inl"
#include "io.archive.pak.inl"

namespace magnum { namespace io {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ArchivePAK::ArchivePAK(File* f) : Archive(f) {
	// Check id
	char id[4];
	if (rawRead(id, 0, sizeof(id)) != sizeof(id))
		throw ArchiveFormatException();
	if (((id[0] != 'p') && (id[0] != 'P'))
	 || ((id[1] != 'a') && (id[1] != 'A'))
	 || ((id[2] != 'c') && (id[2] != 'C'))
	 || ((id[3] != 'k') && (id[3] != 'K')))
		throw ArchiveFormatException();
	
	dword position;
	if (rawRead(&position, 4, sizeof(position)) != sizeof(position))
		throw ArchiveFormatException();
	
	dword length;	
	if (rawRead(&length, 4, sizeof(length)) != sizeof(length))
		throw ArchiveFormatException();

	// Read directory		
	m_Entries.clear();
	m_Entries.setSize(length / 64);
	for (unsigned int i = 0; i < length/64; i++) {
		Entry* e = &m_Entries[i];
		if (rawRead(e, position + 64*i, sizeof(Entry)) != sizeof(Entry))
			throw ArchiveFormatException();
		strlwr(e->m_Name);
	}	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool ArchivePAK::onExists(const char* n) {
	String name = String(name).toLower();
	for (Array<Entry>::Iterator iter = m_Entries.begin(); iter != m_Entries.end(); ++iter) {
		if (name == iter->m_Name)
			return true;
	}
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ArchiveFile* ArchivePAK::onCreate(const char* ) {
	ASSERT(false);
	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ArchiveFile* ArchivePAK::onOpen(const char* n,File::OpenFlags f) {
	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ArchiveDirectory* ArchivePAK::onList() {
	return new DirectoryPAK(this);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void DirectoryPAK::onRefresh() {
	ArchivePAK* arch = (ArchivePAK*)getArchive();
	for (Array<ArchivePAK::Entry>::Iterator iter = arch->m_Entries.begin(); iter != arch->m_Entries.end(); ++iter) {
		insert(iter->m_Name, iter->m_Size, FileInfo::Readable);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void DirectoryPAK::onOpen(const char* ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
FilePAK::FilePAK(ArchivePAK* arch,const ArchivePAK::Entry* e) : ArchiveFile(arch) {
	m_Start = e->m_Position;
	m_Size = e->m_Size;
	m_Position = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t FilePAK::onWrite(const void *buf, size_t size) {
	ASSERT(false);
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t FilePAK::onRead(void *buf, size_t size) {
	if (size == 0)
		return 0;
	if (m_Position >= m_Size)
		return EndOfFile;

	size_t s2 = min(size, (size_t)(m_Size - m_Position));
	s2 = rawRead(buf, m_Position, s2);
	m_Position += s2;
	return s2;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int64 FilePAK::onSetPosition(int64 where,PositionMode mode) {
	switch(mode) {
		case Absolute:
			if (where < 0)
				throw IllegalPositionException();
			if (where > m_Size)
				throw IllegalPositionException();
			m_Position = where;
			break;
		case Current:
			if (m_Position + where < 0)
				throw IllegalPositionException();
			if (m_Position + where > m_Size)
				throw IllegalPositionException();
			m_Position += where;
			break;
		case End:
			if (where > 0)
				throw IllegalPositionException();
			if (m_Size + where < 0)
				throw IllegalPositionException();
			m_Position = m_Size + where;
		default:
			throw IllegalPositionModeException();
	}

	return m_Position;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int64 FilePAK::onGetPosition() const {
	return m_Position;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool FilePAK::onRewind() {
	m_Position = 0;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool FilePAK::onFlush() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int64 FilePAK::onGetSize() const {
	return m_Size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int64 FilePAK::onSetSize(int64 newSize) {
	ASSERT(false);
	return m_Size;
}

}; };

