/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "io.inputbuffer.h"

#include "object.inl"
#include "object.softpointer.inl"
#include "base.string.inl"
#include "io.stream.inl"
#include "io.inputstream.inl"
#include "io.inputbuffer.inl"


namespace magnum { namespace io {
using magnum::math::min;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
InputBuffer::InputBuffer(size_t bufsize) : m_BufferSize(bufsize) {
	ASSERT(bufsize > 0);
	m_Buffer = new byte[m_BufferSize];
	m_BufferLeft = 0;
	m_BufferCurrent = m_Buffer;
	m_Stream = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
InputBuffer::InputBuffer(InputStream* in, size_t bufsize) : m_Stream(in), m_BufferSize(bufsize) {
	ASSERT(bufsize > 0);
	ASSERT(in != NULL);
	m_Buffer = new byte[m_BufferSize];
	m_BufferLeft = 0;
	m_BufferCurrent = m_Buffer;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
InputBuffer::InputBuffer(InputStream& in, size_t bufsize) : m_Stream(&in), m_BufferSize(bufsize) {
	ASSERT(bufsize > 0);
	m_Buffer = new byte[m_BufferSize];
	m_BufferLeft = 0;
	m_BufferCurrent = m_Buffer;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
InputBuffer::~InputBuffer() {
	if (m_Buffer) {
		delete[] m_Buffer;
		m_Buffer = NULL;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void InputBuffer::setStream(InputStream* in) {
	ASSERT(in != NULL);
	m_BufferLeft = 0;
	m_BufferCurrent = m_Buffer;
	m_Stream = in;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void InputBuffer::setStream(InputStream& in) {
	m_BufferLeft = 0;
	m_BufferCurrent = m_Buffer;
	m_Stream = &in;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t InputBuffer::onRead(void *buffer, size_t size) {
	ASSERT(m_Stream != NULL);

	byte* dst = (byte *)buffer;
	size_t total = 0;
	
	// Perform a direct read for large requests
	if (size > m_BufferSize) {
		// Copy current buffer to destination
		if (m_BufferLeft > 0) {
			memcpy(dst, m_BufferCurrent, m_BufferLeft);
			dst += m_BufferLeft;
			size -= m_BufferLeft;
			total = m_BufferLeft;
			m_BufferLeft = 0;
		}
		
		// Directly read from input stream
		while (size > 0) {
			size_t read = m_Stream->read(dst, size);
			if ((int)read <= 0) {
				if (total > 0)
					return total;
				else
					return EndOfFile;
			}
			dst += read;
			size -= read;
			total += read;
		}
		return total;
	}
	
	// Shortcut for single bytes, as needed by parsers
	if (size == 1) {
		// Refill buffer if it is empty
		if (m_BufferLeft == 0) {
			m_BufferCurrent = m_Buffer;
			m_BufferLeft = m_Stream->read(m_Buffer, m_BufferSize);
			if ((int)m_BufferLeft <= 0) {
				m_BufferLeft = 0;
				return EndOfFile;
			}
		}
	
		*dst = *m_BufferCurrent;
		m_BufferCurrent++;
		m_BufferLeft--;
		return 1;
	}

	// This is the normal buffered case for small requests
	while(size > 0) {
		// Refill buffer if it is empty
		if (m_BufferLeft == 0) {
			m_BufferCurrent = m_Buffer;
			m_BufferLeft = m_Stream->read(m_Buffer, m_BufferSize);
			if ((int)m_BufferLeft <= 0) {
				m_BufferLeft = 0;
				if (total > 0)
					return total;
				else
					return EndOfFile;
			}
		}
		
		// Copy from internal buffer to destination buffer
		size_t read = min(size, m_BufferLeft);
		memcpy(dst, m_BufferCurrent, read);
		m_BufferCurrent += read;
		m_BufferLeft -= read;
		dst += read;
		size -= read;
		total += read;
	}

	return total;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t InputBuffer::onSkip(size_t size) {
	ASSERT(m_Stream != NULL);

	size_t total = 0;

	while(size > 0) {
		if (m_BufferLeft == 0) {
			m_BufferCurrent = m_Buffer;
			m_BufferLeft = m_Stream->read(m_Buffer, m_BufferSize);
			if ((int)m_BufferLeft <= 0) {
				m_BufferLeft = 0;
				if (total > 0)
					return total;
				else
					return EndOfFile;
			}
		}
		size_t read = min(size, m_BufferLeft);
		m_BufferCurrent += read;
		m_BufferLeft -= read;
		size -= read;
		total += read;
	}

	return total;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int64 InputBuffer::onAvailable() {
	return m_BufferLeft + m_Stream->available();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool InputBuffer::onClose() {
	if (m_Stream == NULL)
		return true;

	m_Stream = NULL;
	m_BufferLeft = 0;
	return true;
}


}; };
