/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "io.outputbuffer.h"

#include "object.inl"
#include "object.softpointer.inl"
#include "base.string.inl"
#include "io.stream.inl"
#include "io.outputstream.inl"
#include "io.outputbuffer.inl"


namespace magnum { namespace io {
using magnum::math::min;

/*--------------------------------------------------------------------------*/
/**
 */
OutputBuffer::OutputBuffer(size_t bufsize) : m_BufferSize(bufsize) {
	ASSERT(bufsize > 0);
	m_Buffer = new byte[m_BufferSize];
	m_BufferUsed = 0;
	m_BufferCurrent = m_Buffer;
	m_Stream = NULL;
}


/*--------------------------------------------------------------------------*/
/**
 */
OutputBuffer::OutputBuffer(OutputStream* out, size_t bufsize) : m_Stream(out), m_BufferSize(bufsize) {
	ASSERT(bufsize > 0);
	ASSERT(out != NULL);
	m_Buffer = new byte[m_BufferSize];
	m_BufferUsed = 0;
	m_BufferCurrent = m_Buffer;
}


/*--------------------------------------------------------------------------*/
/**
 */
OutputBuffer::OutputBuffer(OutputStream& out, size_t bufsize) : m_Stream(&out), m_BufferSize(bufsize) {
	ASSERT(bufsize > 0);
	m_Buffer = new byte[m_BufferSize];
	m_BufferUsed = 0;
	m_BufferCurrent = m_Buffer;
}


/*--------------------------------------------------------------------------*/
/**
 */
OutputBuffer::~OutputBuffer() {
	if (m_Stream != NULL) {
		if (m_BufferUsed > 0)
			onFlush();
		m_Stream = NULL;
	}
	if (m_Buffer) {
		delete[] m_Buffer;
		m_Buffer = NULL;
	}
}


/*--------------------------------------------------------------------------*/
/**
 */
void OutputBuffer::setStream(OutputStream* out) {
	ASSERT(out != NULL);
	if (m_Stream != NULL)
		onFlush();
	m_BufferUsed = 0;
	m_BufferCurrent = m_Buffer;
	m_Stream = out;
}


/*--------------------------------------------------------------------------*/
/**
 */
void OutputBuffer::setStream(OutputStream& out) {
	if (m_Stream != NULL)
		onFlush();
	m_BufferUsed = 0;
	m_BufferCurrent = m_Buffer;
	m_Stream = &out;
}


/*--------------------------------------------------------------------------*/
/**
 */
size_t OutputBuffer::onWrite(const void *buffer, size_t size) {
	ASSERT(m_Stream != NULL)
	size_t total = 0;
	const byte* src = (const byte*)buffer;
	
	// Special case for large write requests
	if (size > m_BufferSize) {
		// First write out pending buffers
		m_BufferCurrent = m_Buffer;
		while(m_BufferUsed > 0) {	
			size_t written = m_Stream->write(m_BufferCurrent, m_BufferUsed);
			m_BufferUsed -= written;
			m_BufferCurrent += written;
		}
		m_BufferCurrent = m_Buffer;
		
		// Now write out new buffers
		while (size > 0) {
			size_t written = m_Stream->write(src, size);
			size -= written;
			src += written;
			total += written;
		}
		
		return total;
	}
	
	// This is the general case for small blocks
	while (size > 0) {
		// Do we need to flush the output buffer?
		if (m_BufferUsed >= m_BufferSize) {
			m_BufferCurrent = m_Buffer;
			while(m_BufferUsed > 0) {	
				size_t written = m_Stream->write(m_BufferCurrent, m_BufferUsed);
				m_BufferUsed -= written;
				m_BufferCurrent += written;
			}

			m_BufferCurrent = m_Buffer;
		}

		size_t write = min(size, m_BufferSize - m_BufferUsed);
		memcpy(m_BufferCurrent, src, write);
		m_BufferCurrent += write;
		m_BufferUsed += write;
		src += write;
		total += write;
		size -= write;
	}
	
	return total;
}


/*--------------------------------------------------------------------------*/
/**
 */
bool OutputBuffer::onFlush() {
	ASSERT(m_Stream != NULL)
	
	const byte* src = m_Buffer;
	while(m_BufferUsed > 0) {	
		size_t written = m_Stream->write(src, m_BufferUsed);
		m_BufferUsed -= written;
		src += written;
	}

	m_BufferCurrent = m_Buffer;
	return m_Stream->flush();
}


/*--------------------------------------------------------------------------*/
/**
 */
bool OutputBuffer::onClose() {
	if (m_Stream == NULL)
		return true;

	onFlush();
	m_Stream = NULL;
	return true;
}

}; };
