/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "io.streamtokeniser.h"

#include "object.inl"
#include "object.softpointer.inl"
#include "base.string.inl"
#include "io.inputstream.inl"
#include "io.streamtokeniser.inl"

namespace magnum { namespace io {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
StreamTokeniser::StreamTokeniser() {
	m_LastByte = 0;
	m_Type = Unknown;
	resetSyntax();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
StreamTokeniser::StreamTokeniser(InputStream* in) : m_Input(in) {
	m_LastByte = 0;
	m_Type = Unknown;
	resetSyntax();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void StreamTokeniser::resetSyntax() {
	memset(m_TokenTable, Unknown, sizeof(m_TokenTable));
	m_Whitespaces = " \t";
	mapChars(m_Whitespaces, Whitespace);
	m_Ordinaries = "!#$%&()*+.,-/<=>;:?@[\\]^{|}~";
	mapChars(m_Ordinaries, Ordinary);
	m_Words = "01234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_";
	mapChars(m_Words, Word);
	m_Quotes ="'\"";
	mapChars(m_Quotes, Quote);
	m_Comments = "";
	mapChars("\n\r", EndOfLine);
	mapChar(0, Unknown);
	m_SlashSlashComments = false;
	m_SlashStarComments = false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void StreamTokeniser::mapChar(char c,Token t) {
	m_TokenTable[(unsigned char)c] = t;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void StreamTokeniser::mapChars(const String& tokens, Token type) {
	for (unsigned int i = 0; i < tokens.getLength(); ++i)
		m_TokenTable[(unsigned char)tokens[i]] = type;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
StreamTokeniser::Token StreamTokeniser::readQuote(char marker) {
	m_String.clear();
	for (;;) {
		int read = (int)m_Input->read(&m_LastByte,1);
		if (read <= 0)
			break;
		if (m_LastByte == marker)
			break;
		m_String += m_LastByte;
	}
	m_LastByte = 0;
	m_Type = Quote;
	
	return Quote;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
StreamTokeniser::Token StreamTokeniser::readComment() {
	m_String.clear();
	for (;;) {
		int read = (int)m_Input->read(&m_LastByte,1);
		if (read <= 0)
			break;
		if ((m_LastByte == '\n') || (m_LastByte == '\r'))
			break;
		m_String += m_LastByte;
	}
	m_Type = Comment;
	
	return Comment;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
StreamTokeniser::Token StreamTokeniser::readSlashSlashComment() {
	m_String.clear();
	for (;;) {
		int read = (int)m_Input->read(&m_LastByte,1);
		if (read <= 0)
			break;
		if ((m_LastByte == '\n') || (m_LastByte == '\r'))
			break;
		m_String += m_LastByte;
	}
	m_Type = Comment;
	m_LastByte = 0;
	
	return Comment;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
StreamTokeniser::Token StreamTokeniser::readSlashStarComment() {
	m_String.clear();
	
	m_Input->read(&m_LastByte,1);
	for (;;) {
		char c;
		int read = (int)m_Input->read(&c,1);
		if (read <= 0)
			break;
		if ((c == '/') && (m_LastByte == '*'))
			break;
		m_String += m_LastByte;
		m_LastByte = c;
	}
	m_Type = Comment;
	m_LastByte = 0;
	
	return Comment;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
StreamTokeniser::Token StreamTokeniser::nextToken() {
	// Process any leading whitespaces
	while (m_TokenTable[m_LastByte] == Whitespace) {
		int read = (int)m_Input->read(&m_LastByte,1);
		if (read <= 0) {
			m_Type = EndOfFile;
			return EndOfFile;
		}
	}
	
	// Now check last valid byte
	m_Type = (Token)m_TokenTable[m_LastByte];
	ASSERT(m_Type < 10);
	if (m_Type == Quote)
		return readQuote(m_LastByte);
	if (m_Type == Comment)
		return readComment();
		
	// CLear string if needed
	if (m_Type != Unknown)
		m_String = m_LastByte;
	else
		m_String.clear();

	// Check if we reached the end of file
	if (m_Input->available() == 0) {
		if (m_LastByte == 0)
			return EndOfFile;
		m_LastByte = 0;
		return m_Type;
	}

	// Read some more bytes
	for(;;) {
		unsigned char c;
		int read = (int)m_Input->read(&c,1);
		if (read <= 0)
			break;
		if (c == '\r') {
			continue;
		} else if ((c == '/') && (m_LastByte == '/') && (m_SlashSlashComments)) {
			return readSlashSlashComment();
		} else if ((c == '*') && (m_LastByte == '/') && (m_SlashStarComments)) {
			return readSlashStarComment();
		}
		
		// Save last byte
		m_LastByte = c;
		
		// Check for ordinaries
		if (m_Type == Ordinary)
			break;
		if (m_Type == EndOfLine)
			break;
		
		// Get token of current char
		Token ct = (Token)m_TokenTable[c];
		ASSERT(ct < 10);
		if (m_Type == Unknown) {
			m_Type = ct;
			m_String = c;
			continue;
		} else if (ct != m_Type) {
			break;
		}
			
		// Check for end of line etc			
		if (c == 0)
			break;
		m_String += c;
	}
	
	return m_Type;
}

}; };

