/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "io.sysfile.h"
#include "io.exception.h"

#include "object.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "system.timestamp.inl"
#include "io.file.inl"
#include "io.filestream.inl"
#include "io.sysfile.inl"
#include "io.inputstream.inl"
#include "io.outputstream.inl"

#ifndef __FORWINDOWS__
#error "This file is only for Windows"
#endif


namespace magnum { namespace io {
using magnum::types::uint64;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SystemFile::SystemFile(unsigned int flags) : File(flags) {
	m_FileHandle = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SystemFile::~SystemFile() {
	if (m_FileHandle ==NULL )
		return;

	if (CloseHandle(m_FileHandle) == 0)
		throwException();

	m_FileHandle = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t SystemFile::onWrite(const void *buf, size_t size) {
	DWORD written;
	WriteFile(m_FileHandle, buf, (DWORD)size, &written, NULL);
	return written;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
size_t SystemFile::onRead(void *buf, size_t size) {
	if (size == 0)
		return 0;
	DWORD read;
	ReadFile(m_FileHandle, buf, (DWORD)size, &read, NULL);
	if (read == 0)
		return EndOfFile;
	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::onFlush() {
	return (FlushFileBuffers(m_FileHandle) == TRUE);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int64 SystemFile::onSetPosition(int64 where, PositionMode mode) {
	DWORD method;
	switch (mode) {
		case Absolute:
			method = FILE_BEGIN;
			break;
		case Current:
			method = FILE_CURRENT;
			break;
		case End:
			method = FILE_END;
			break;
		default:
			throw IllegalPositionModeException();
	}
	LONG low = (DWORD)where;
	LONG hi = (DWORD)(where >> 32);
	low = SetFilePointer(m_FileHandle, low, &hi, method);
	if (low == 0xffffffff)
		throwException();
	return (int64)low + ((int64)hi << 32);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::onRewind() {
	if (SetFilePointer(m_FileHandle, 0, NULL, FILE_BEGIN) != 0)
		throwException();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int64 SystemFile::onGetPosition() const {
	LONG lo,hi=0;
	lo = SetFilePointer(m_FileHandle, 0, &hi, FILE_CURRENT);
	return (int64)lo + ((int64)hi << 32);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int64 SystemFile::onGetSize() const {
	LONG curlo,curhi=0,sizelo,sizehi=0;
	curlo = SetFilePointer(m_FileHandle, 0, &curhi, FILE_CURRENT);
	sizelo = SetFilePointer(m_FileHandle, 0, &sizehi, FILE_END);
	SetFilePointer(m_FileHandle, curlo, &curhi, FILE_BEGIN);
	return (int64)sizelo + ((int64)sizehi << 32);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int64 SystemFile::onSetSize(int64 newSize) {
	LONG curlo, curhi=0;
	curlo = SetFilePointer(m_FileHandle, 0, &curhi, FILE_CURRENT);
	LONG newlo = (LONG)newSize;
	LONG newhi = (LONG)(newSize >> 32);
	SetFilePointer(m_FileHandle, newlo, &newhi, FILE_BEGIN);
	SetEndOfFile(m_FileHandle);
	SetFilePointer(m_FileHandle, curlo, &curhi, FILE_BEGIN);
	return newSize;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Timestamp SystemFile::getCreationTime() const {
	ASSERT(m_FileHandle!=NULL);

	FILETIME ftime;
	GetFileTime(m_FileHandle,&ftime,NULL,NULL);
	return Timestamp(((((uint64)ftime.dwHighDateTime) << 32) + ftime.dwLowDateTime) / 10);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Timestamp SystemFile::getModificationTime() const {
	ASSERT(m_FileHandle!=NULL);

	FILETIME ftime;
	GetFileTime(m_FileHandle,NULL,NULL,&ftime);
	return Timestamp(((((uint64)ftime.dwHighDateTime) << 32) + ftime.dwLowDateTime) / 10);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Timestamp SystemFile::getAccessTime() const {
	ASSERT(m_FileHandle!=NULL);

	FILETIME ftime;
	GetFileTime(m_FileHandle,NULL,&ftime,NULL);
	return Timestamp(((((uint64)ftime.dwHighDateTime) << 32) + ftime.dwLowDateTime) / 10);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SystemFile* SystemFile::open(const char *fname, OpenFlags flags) {
	DWORD sysFlags = 0;

	if (flags & WriteOnly)
		sysFlags |= GENERIC_WRITE;
	if (flags & ReadOnly)
		sysFlags |= GENERIC_READ;

	HANDLE handle = CreateFile(fname, sysFlags, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (handle == INVALID_HANDLE_VALUE) 
		throwException();

	SystemFile* file = new SystemFile(flags);
	file->m_FileHandle = handle;

	return file;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SystemFile* SystemFile::create(const char *fname) {
	HANDLE handle = CreateFile(fname, GENERIC_READ | GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (handle == INVALID_HANDLE_VALUE)
		throwException();

	SystemFile* file =new SystemFile(Writeable | Readable);
	file->m_FileHandle = handle;
	return file;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::remove(const char *fname) {
	return (DeleteFile(fname)==0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::exists(const char *fname) {
	return (GetFileAttributes(fname) != INVALID_FILE_ATTRIBUTES);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void SystemFile::throwException() {
	DWORD err = GetLastError();
	switch(err) {
		case ERROR_READ_FAULT:
			throw FileReadException("The system cannot read from the specified device.");
		case ERROR_WRITE_FAULT:
			throw FileWriteException("The system cannot write to the specified device.");
		case ERROR_WRITE_PROTECT:
			throw FileWriteException("The media is write protected.");
		case ERROR_WRONG_DISK:
			throw IOException("The wrong diskette is in the drive.");
		case ERROR_UNRECOGNIZED_VOLUME:
			throw IOException("The volume does not contain a recognized file system.");
		case ERROR_UNRECOGNIZED_MEDIA:
			throw IOException("The disk media is not recognized. It may not be formatted.");
		case ERROR_TOO_MANY_OPEN_FILES:
			throw FileOpenException("The system cannot open the file.");
		case ERROR_OPEN_FAILED:
			throw FileOpenException("The system cannot open the device or file specified.");
		case ERROR_SHARING_VIOLATION:
			throw IOException("The process cannot access the file because it is being used by another process.");
		case ERROR_SEEK:
			throw IOException("The drive cannot locate a specific area or track on the disk.");
		case ERROR_SEEK_ON_DEVICE:
			throw FileSeekException("The file pointer cannot be set on the specified device or file.");
		case ERROR_FILE_NOT_FOUND:
			throw FileNotFoundException("The system cannot find the file specified.");
		case ERROR_FILE_EXISTS:
			throw FileExistsException("The file exists.");
		case ERROR_FILE_INVALID:
			throw IOException("The volume for a file has been externally altered so that the opened file is no longer valid.");
		case ERROR_DISK_FULL:
			throw DiskFullException("There is not enough space on the disk.");
		default:
			throw IOException();
	}
}

}; };
