/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "media.pin.h"
#include "media.graph.h"
#include "media.listener.h"

#include "object.inl"
#include "base.string.inl"
#include "application.listener.inl"
#include "container.iterator.inl"
#include "container.allocator.inl"
#include "container.chain.inl"
#include "container.list.inl"
#include "system.signal.inl"
#include "media.node.inl"
#include "media.graph.inl"
#include "media.looper.inl"
#include "media.message.inl"
#include "media.pin.inl"

namespace magnum { namespace media {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Graph::Graph() : Looper(&m_Thread), m_Thread(&Graph::threadFunc, this) {
	m_State = Node::Stopped;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Graph::~Graph() {
	while (m_Nodes.begin() != m_Nodes.end()) {
		Node* node = m_Nodes.begin();
		node->m_Graph = NULL;
		m_Nodes.remove(node);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::insert(Node* node) {
	ASSERT(m_State == Node::Stopped);
	if (m_State != Node::Stopped)
		return false;
		
	ASSERT(node->getGraph() == NULL);
	if (node->getGraph() != NULL)
		return false;
	
	m_Nodes.insert(node);
	node->m_Graph = this;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::remove(Node* node) {
	ASSERT(m_State == Node::Stopped);
	if (m_State != Node::Stopped)
		return false;
		
	ASSERT(node->getGraph() == this);
	if (node->getGraph() != this)
		return false;
	
	node->clearMessages();
	m_Nodes.remove(node);
	node->m_Graph = NULL;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::clear() {
	ASSERT(m_State == Node::Stopped);
	if (m_State != Node::Stopped)
		return false;

	while (m_Nodes.begin() != m_Nodes.end()) {
		Node* node = m_Nodes.begin();
		node->m_Graph = NULL;
		m_Nodes.remove(node);
	}
	
	clearMessages();
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::connect(OutputPin* src, InputPin* dst) {
	ASSERT(m_State == Node::Stopped);
	if (m_State != Node::Stopped)
		return false;

	// Precheck some conditions like ownership and connectivity		
	ASSERT(src->getNode()->getGraph() == this);
	ASSERT(dst->getNode()->getGraph() == this);
	if (src->getNode()->getGraph() != this)
		return false;
	if (dst->getNode()->getGraph() != this)
		return false;
	ASSERT(src->getConnection() == NULL);
	ASSERT(dst->getConnection() == NULL);
	if (src->getConnection() != NULL)	
		return false;
	if (dst->getConnection() != NULL)
		return false;

	// Try to ge a common format for both pins
	if (!negotiateFormat(src, dst))
		return false;
	
	// Setup connection pointers
	src->m_Connection = dst;
	dst->m_Connection = src;
	
	// Try to connect both pins
	if (!src->getNode()->onConnect(src)) {
		src->m_Connection = NULL;
		dst->m_Connection = NULL;
		return false;
	}
	if (!dst->getNode()->onConnect(dst)) {
		src->getNode()->onDisconnect(src);
		src->m_Connection = NULL;
		dst->m_Connection = NULL;
		return false;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::connect(OutputPin* src, InputPin* dst, const Format& fmt) {
	ASSERT(m_State == Node::Stopped);
	if (m_State != Node::Stopped)
		return false;
		
	// Precheck some conditions like ownership and connectivity		
	ASSERT(src->getNode()->getGraph() == this);
	ASSERT(dst->getNode()->getGraph() == this);
	if (src->getNode()->getGraph() != this)
		return false;
	if (dst->getNode()->getGraph() != this)
		return false;
	ASSERT(src->getConnection() == NULL);
	ASSERT(dst->getConnection() == NULL);
	if (src->getConnection() != NULL)	
		return false;
	if (dst->getConnection() != NULL)
		return false;

	// Try to ge a common format for both pins
	if (!src->setFormat(fmt))
		return false;
	if (!dst->setFormat(fmt))
		return false;
	
	// Setup connection pointers
	src->m_Connection = dst;
	dst->m_Connection = src;
	
	// Try to connect both pins
	if (!src->getNode()->onConnect(src)) {
		src->m_Connection = NULL;
		dst->m_Connection = NULL;
		return false;
	}
	if (!dst->getNode()->onConnect(dst)) {
		src->getNode()->onDisconnect(src);
		src->m_Connection = NULL;
		dst->m_Connection = NULL;
		return false;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::disconnect(OutputPin* src, InputPin* dst) {
	ASSERT(m_State == Node::Stopped);
	if (m_State != Node::Stopped)
		return false;

	// Precheck some conditions like ownership and connectivity		
	ASSERT(src->getNode()->getGraph() == this);
	ASSERT(dst->getNode()->getGraph() == this);
	if (src->getNode()->getGraph() != this)
		return false;
	if (dst->getNode()->getGraph() != this)
		return false;
	ASSERT(src->getConnection() == dst);
	ASSERT(dst->getConnection() == src);
	if (src->getConnection() != dst)	
		return false;
	if (dst->getConnection() != src)
		return false;

	// Disconnect both pins		
	src->getNode()->onDisconnect(src);
	dst->getNode()->onDisconnect(dst);
	src->m_Connection = NULL;
	dst->m_Connection = NULL;

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::start(bool blocking) {
	if (m_State == Node::Starting)
		return true;
	if (m_State == Node::Stopping)
		return false;
	if (m_State == Node::Running)
		return true;
	if (m_State == Node::Paused)
		return resume();

	m_State = Node::Starting;
	m_Thread.start();

	Thread::cycle();
	Thread::cycle();
	Thread::cycle();

	sendMessage(Message::Start);
	if (!blocking)
		return true;
		
	// Wait for graph to be started
	for (;;) {
		if (m_State == Node::Running)
			return true;
		else if (m_State == Node::Stopped)
			return false;		
		Thread::sleep(50);
	}
	
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::stop(bool blocking) {
	if (m_State == Node::Starting)
		return false;
	if (m_State == Node::Stopping)
		return true;
	if (m_State == Node::Stopped)
		return true;

	sendMessage(Message::Stop);
	sendMessage(Message::Quit);
	
	if (!blocking)
		return  true;

	// Wait for thread to be stopped	
	m_Thread.join();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::reset() {
	if (m_State == Node::Paused)
		return false;

	sendMessage(Message::Reset);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::pause() {
	if (m_State == Node::Paused)
		return true;
	if (m_State != Node::Running)
		return false;
			
	sendMessage(Message::Pause);
	for (unsigned int i = 0; i < 12; ++i) {
		if (m_State == Node::Paused)
			return true;		
		Thread::sleep(60);
	}
	
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::resume() {
	if (m_State == Node::Running)
		return true;
	if (m_State != Node::Paused)
		return false;

	sendMessage(Message::Resume);
	for (unsigned int i = 0; i < 12; ++i) {
		if (m_State == Node::Running)
			return true;		
		Thread::sleep(60);
	}
	
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::onLoop() {
	bool ret = processMessages();
		
	for (Iterator iter = m_Nodes.begin(); iter != m_Nodes.end(); ++iter) {
		iter->loop();
	}
	
	return ret;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Graph::onHandleMessage(const Message& msg) {
	switch (msg.getCommand()) {
		case Message::Start:
			onStart();
			return;
		case Message::Stop:
			onStop();
			return;
		case Message::Pause:
			onPause();
			return;
		case Message::Resume:
			onResume();
			return;
		case Message::Reset:
			onReset();
			return;
		case Message::EndOfInput:
			ASSERT(msg.getPin() != NULL);
			onEndOfInput((InputPin*)msg.getPin());
			return;
		case Message::EndOfOutput:
			ASSERT(msg.getPin() != NULL);
			onEndOfOutput((OutputPin*)msg.getPin());
			return;
		default:
			return;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::onStart() {
	m_State = Node::Starting;
	
	for (Iterator iter = m_Nodes.begin(); iter != m_Nodes.end(); ++iter) {
		if (!iter->start()) {
			sendMessage(Message::Stop);
			sendMessage(Message::Quit);
		}
	}

	waitForState(Node::Running);
	m_State = Node::Running;
	m_Dispatcher.fire(&MediaListener::onStart, this);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::onStop() {
	m_State = Node::Stopping;
	
	for (Iterator iter = m_Nodes.begin(); iter != m_Nodes.end(); ++iter) {
		iter->stop();
	}
	
	waitForState(Node::Stopped);
	m_State = Node::Stopped;
	m_Dispatcher.fire(&MediaListener::onStop, this);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::onReset() {
	for (Iterator iter = m_Nodes.begin(); iter != m_Nodes.end(); ++iter) {
		if (!iter->reset())
			return false;
	}

	m_Dispatcher.fire(&MediaListener::onReset, this);
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::onPause() {
	for (Iterator iter = m_Nodes.begin(); iter != m_Nodes.end(); ++iter) {
		if (!iter->pause())
			return false;
	}
	
	waitForState(Node::Paused);
	m_State = Node::Paused;
	m_Dispatcher.fire(&MediaListener::onPause, this);
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::onResume() {
	for (Iterator iter = m_Nodes.begin(); iter != m_Nodes.end(); ++iter) {
		if (!iter->resume())
			return false;
	}
	
	waitForState(Node::Running);
	m_State = Node::Running;
	m_Dispatcher.fire(&MediaListener::onResume, this);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Graph::onEndOfInput(InputPin* pin) {
	m_Dispatcher.fire(&MediaListener::onEndOfInput, this, pin);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Graph::onEndOfOutput(OutputPin* pin) {
	m_Dispatcher.fire(&MediaListener::onEndOfOutput, this, pin);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::negotiateFormat(OutputPin* src, InputPin* dst) {
	Format fmt = src->getFormat();
	if (!dst->adjustFormat(fmt)) {
		fmt = dst->getFormat();
		if (!src->adjustFormat(fmt))
			return false;
		if (!dst->adjustFormat(fmt))
			return false;
	}  else {
		if (!src->adjustFormat(fmt))
			return false;
	}
	
	if (!dst->setFormat(fmt))
		return false;
	if (!src->setFormat(fmt))
		return false;
		
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::waitForState(Graph::State state) {
	// Wait for all nodes till they are running
	for (unsigned int i = 0; i < 10; i++) {
		bool allok = true;
		for (Iterator iter = m_Nodes.begin(); iter != m_Nodes.end(); ++iter) {
			if (iter->getState() != state)
				allok = false;
		}
		
		if (allok)
			return true;

		for (Iterator iter = m_Nodes.begin(); iter != m_Nodes.end(); ++iter)
			iter->loop();
			
		Thread::sleep(50);
	}
	
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Graph::threadFunc(void* parm) {
	Graph* graph = static_cast<Graph*>(parm);
	for (;;) {
		if (!graph->loop())
			break;
			
		graph->waitMessage(100);
	}
	
	return true;
}

}; };
