/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "media.node.h"
#include "media.device.h"
#include "media.pin.h"
#include "media.graph.h"
#include "media.buffer.h"
#include "media.listener.h"

#include "base.string.inl"
#include "object.class.inl"
#include "object.inl"
#include "system.thread.inl"
#include "system.signal.inl"
#include "system.mutex.inl"
#include "application.listener.inl"
#include "container.allocator.inl"
#include "container.list.inl"
#include "container.chain.inl"
#include "container.iterator.inl"
#include "media.message.inl"
#include "media.looper.inl"
#include "media.node.inl"
#include "media.device.inl"
#include "media.pin.inl"
#include "media.graph.inl"
#include "media.buffer.inl"


namespace magnum { namespace media {

const char Node::DefaultInput[] = "DefaultInput";
const char Node::DefaultOutput[] = "DefaultOutput";
const char Node::DefaultFilter[] = "DefaultFilter";
const char Node::DefaultNode[] = "DefaultNode";


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Node::Node(const char* name, Type t) : Looper(NULL) {
	m_Name = name;
	m_Graph = NULL;
	m_Device = NULL;
	m_State = Stopped;
	m_Type = t;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Node::Node(Device* dev, const char* name, Type t) : Looper(NULL) {
	m_Name = name;
	m_Graph = NULL;
	m_Device = dev;
	m_State = Stopped;
	m_Type = t;

	if (m_Device != NULL)
		m_Device->refer();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Node::~Node() {
	if (m_Device != NULL)
		m_Device->release();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::sendBuffer(OutputPin* pin,Buffer* buffer) {
	ASSERT(pin->getNode() == this);
	if (!pin->isConnected()) {
		buffer->release();
		return true;	
	}

	InputPin* cpin = (InputPin *)pin->getConnection();
	Node* cnode = cpin->getNode();

	buffer->m_Producer = pin;
	buffer->m_Consumer = cpin;
	
	if (cnode->getThread() == NULL) {
		cnode->onReceiveBuffer(cpin, buffer);
		return true;
	}

	cnode->sendMessage(Message::ReceiveBuffer, cpin, buffer);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::postBuffer(OutputPin* pin,Buffer* buffer) {
	ASSERT(pin->getNode() == this);
	if (!pin->isConnected()) {
		buffer->release();
		return true;	
	}

	InputPin* cpin = (InputPin *)pin->getConnection();
	Node* cnode = cpin->getNode();

	buffer->m_Producer = pin;
	buffer->m_Consumer = cpin;

	cnode->sendMessage(Message::ReceiveBuffer, cpin, buffer);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
InputPin* Node::getInputPin(const char* which) {
	if (which == NULL) {
		return m_DefaultInput;
	} else if (strcmp(which, Pin::DefaultInput) == 0) {
		return m_DefaultInput;
	}

	for (Chain<InputPin>::Iterator iter = m_InputPins.begin(); iter != m_InputPins.end(); ++iter) {
		if (iter->getName() == which)
			return iter;	
	}
	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
OutputPin* Node::getOutputPin(const char* which) {
	if (which == NULL) {
		return m_DefaultOutput;
	} else if (strcmp(which, Pin::DefaultOutput) == 0) {
		return m_DefaultOutput;
	}

	for (Chain<OutputPin>::Iterator iter = m_OutputPins.begin(); iter != m_OutputPins.end(); ++iter) {
		if (iter->getName() == which)
			return iter;	
	}
	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::start() {
	if (getState() == Node::Starting)
		return true;
	if (getState() == Node::Stopping)
		return false;
	if (getState() == Node::Running)
		return true;
	if (getState() == Node::Paused)
		return resume();
		
	sendMessage(Message::Start);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::stop() {
	if (getState() == Node::Starting)
		return false;
	if (getState() == Node::Stopping)
		return true;
	if (getState() == Node::Stopped)
		return true;
		
	sendMessage(Message::Stop);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::reset() {
	if (getState() == Node::Paused)
		return false;
		
	sendMessage(Message::Reset);
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::pause() {
	if (getState() == Node::Paused)
		return true;
	if (getState() != Node::Running)
		return false;
		
	sendMessage(Message::Pause);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::resume() {
	if (getState() == Node::Running)
		return true;
	if (getState() != Node::Paused)
		return false;
		
	sendMessage(Message::Resume);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::endOfInput(InputPin* pin) {
	ASSERT(pin != NULL);
	ASSERT(pin->getNode() == this);
	getGraph()->sendMessage(Message::EndOfInput, pin);
	OutputPin* con = pin->getConnection();
	if (con == NULL)
		return;
	Node* node = con->getNode();
	ASSERT(node != NULL);
	node->sendMessage(Message::EndOfOutput, con);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::endOfOutput(OutputPin* pin) {
	ASSERT(pin != NULL);
	ASSERT(pin->getNode() == this);
	getGraph()->sendMessage(Message::EndOfOutput, pin);
	InputPin* con = pin->getConnection();
	if (con == NULL)
		return;
	Node* node = con->getNode();
	ASSERT(node != NULL);
	node->sendMessage(Message::EndOfInput, con);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::waitForState(Node::State state) {
	for (unsigned int i = 0; i < 10; i++) {
		loop();
		if (getState() == state)
			return true;
		if (getThread() != NULL)
			Thread::sleep(50);	
	}
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::onLoop() {
	// Process any messages
	return processMessages();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onHandleMessage(const Message& msg) {
	switch (msg.getCommand()) {
		case Message::Connect:
			ASSERT(msg.getPin() != NULL);
			onConnect(msg.getPin());
			return;
		case Message::Disconnect:
			ASSERT(msg.getPin() != NULL);
			onDisconnect(msg.getPin());
			return;
		case Message::ReceiveBuffer:
			ASSERT(msg.getPin() != NULL);
			ASSERT(msg.getPointer() != NULL);
			onReceiveBuffer((InputPin*)msg.getPin(), (Buffer*)msg.getPointer());
			return;
		case Message::RequestBuffer:
			ASSERT(msg.getPin() != NULL);
			onRequestBuffer((OutputPin*)msg.getPin(), (Buffer*)msg.getPointer());
			return;
		case Message::Start:
			m_State = Starting;
			if (onStart())
				m_State = Running;
			return;
		case Message::Stop:
			m_State = Stopping;
			if (onStop())
				m_State = Stopped;
			return;
		case Message::Pause:
			if (onPause())
				m_State = Paused;
			return;
		case Message::Resume:
			if (m_State == Stopped)
				return;
			if (onResume())
				m_State = Running;
			return;
		case Message::Reset:
			onReset();
			return;
		case Message::EndOfInput:
			ASSERT(msg.getPin() != NULL);
			onEndOfInput((InputPin*)msg.getPin());
			return;
		case Message::EndOfOutput:
			ASSERT(msg.getPin() != NULL);
			onEndOfOutput((OutputPin*)msg.getPin());
			return;
		default:
			return;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::onConnect(Pin* ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::onDisconnect(Pin* ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::onReceiveBuffer(InputPin* ,Buffer* ) {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::onRequestBuffer(OutputPin* ,Buffer* buf) {
	if (buf != NULL)
		buf->release();
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onDestroyBuffer(Buffer* ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::onStart() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::onStop() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::onPause() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::onResume() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::onReset() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::onSetFormat(Pin* ,const Format& ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::onAdjustFormat(Pin* ,Format& ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onEndOfInput(InputPin* ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onEndOfOutput(OutputPin* ) {
}

}; };
