/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "io.exception.h"
#include "media.exception.h"
#include "media.buffer.h"
#include "media.reader.ogg.h"

#include "object.pointer.inl"
#include "io.file.inl"
#include "io.stream.inl"
#include "io.inputstream.inl"
#include "media.node.inl"
#include "media.pin.inl"
#include "media.format.inl"
#include "media.reader.inl"
#include "media.buffer.inl"

namespace magnum { namespace media {
using magnum::io::FileReadException;

/*--------------------------------------------------------------------------*/
/**
 */
ReaderOGG::ReaderOGG(File* file) : Reader(file) {
	m_FileStart = file->getPosition();
	openFile();
	setupFormat();
}


/*--------------------------------------------------------------------------*/
/**
 */
ReaderOGG::ReaderOGG(InputStream* in) : Reader(in) {
	openFile();
	setupFormat();
}


/*--------------------------------------------------------------------------*/
/**
 */
ReaderOGG::~ReaderOGG() {
	ov_clear(&m_File);
}


/*--------------------------------------------------------------------------*/
/**
 */
bool ReaderOGG::openFile() {
	// Setup callback structures
	ov_callbacks cb;
	void* file;
	if (getFile() != NULL) {
		file = getFile();	
		cb.read_func = &file_read_func;
		cb.seek_func = &file_seek_func;
		cb.close_func = &file_close_func;
		cb.tell_func = &file_tell_func;
	} else {
		file = getStream();
		cb.read_func = &stream_read_func;
		cb.seek_func = &stream_seek_func;
		cb.close_func = &stream_close_func;
		cb.tell_func = &stream_tell_func;
	}
	
	// open file
	int ret = ov_open_callbacks(file, &m_File, NULL, 0, cb);
	switch (ret) {
		case OV_EREAD:
			throw FileReadException();
		case OV_ENOTVORBIS:
			throw MediaFileFormatException();
		case OV_EVERSION:
			throw MediaFileFormatException();
		case OV_EBADHEADER:
			throw MediaFileFormatException();
		case OV_EFAULT:
			throw MediaFileException();
	}

	return true;
}

	
/*--------------------------------------------------------------------------*/
/**
 */
bool ReaderOGG::setupFormat() {
	// Get file info
	vorbis_info* info = ov_info(&m_File, -1);

	// Create format for pin according to file info
	Format fmt(Format::Audio);
	fmt.setCodec("raw");
	fmt.audio().setChannels(info->channels);
	fmt.audio().setSamplerate(info->rate);
	fmt.audio().setFormat(AudioFormat::Signed16);
	fmt.setBitrate(info->channels * info->rate * 16);
	fmt.setBufferSize(fmt.getBitrate() / 16);
	
	return setFormat(fmt);
}


/*--------------------------------------------------------------------------*/
/**
 */
bool ReaderOGG::onSetFormat(const Format& fmt) {
	if (fmt.getType() != Format::Audio)
		return false;
	if (fmt.getCodec() != "raw")
		return false;

	vorbis_info* info = ov_info(&m_File, -1);
	if (fmt.audio().getSamplerate() != info->rate)
		return false;
	if (fmt.audio().getChannels() != info->channels)
		return false;
	if (fmt.getBufferSize() < 8000)
		return false;
		
	int wordsize, sign;	
	switch(fmt.audio().getFormat()) {
		case AudioFormat::Signed8:
			wordsize = 1;
			sign = 1;
			break;
		case AudioFormat::Unsigned8:
			wordsize = 1;
			sign = 0;
			break;
		case AudioFormat::Signed16:
			wordsize = 2;
			sign = 1;
			break;
		case AudioFormat::Unsigned16:
			wordsize = 2;
			sign = 0;
			break;
		default:
			return false;		
	};
	
	m_Wordsize = wordsize;
	m_Signed = sign;
	return true;
}


/*--------------------------------------------------------------------------*/
/**
 */
bool ReaderOGG::onAdjustFormat(Format& fmt) {
	if (fmt.getType() != Format::Audio)
		return false;
	if (fmt.getBufferSize() < 8000)
		fmt.setBufferSize(8000);

	vorbis_info* info = ov_info(&m_File, -1);
	fmt.audio().setChannels(info->channels);
	fmt.audio().setSamplerate(info->rate);
	fmt.setCodec("raw");
	switch (fmt.audio().getFormat()) {
		case AudioFormat::Signed8:
			fmt.setBitrate(info->channels * info->rate * 8);
			break;
		case AudioFormat::Unsigned8:
			fmt.setBitrate(info->channels * info->rate * 8);
			break;
		case AudioFormat::Signed16:
			fmt.setBitrate(info->channels * info->rate * 16);
			break;
		case AudioFormat::Unsigned16:
			fmt.setBitrate(info->channels * info->rate * 16);
			break;
		default:
			return false;		
	}
	
	return true;
}


/*--------------------------------------------------------------------------*/
/**
 */
bool ReaderOGG::onRead(Buffer* buf) {
	int stream;
	char* data = (char*)buf->getBuffer();
	int bytes_left = buf->getMaxSize();
	int bytes_read = 0;
	
	for (;;) {
		long ret = ov_read(&m_File, data, bytes_left, 0, m_Wordsize, m_Signed, &stream);
		if (ret > 0) {
			if (stream != 0)
				continue;
			bytes_read += ret;
			bytes_left -= ret;
			if (bytes_left < ret)
				break;
			data += ret;
			continue;
		} else if (ret ==  0) {
			if (bytes_read == 0) {
				endOfOutput(getPin());
				return false;
			}
			break;
		} else {
			if (bytes_read == 0)
				return false;
			break;
		}
	}

	buf->setSize(bytes_read);
	return true;
}


/*--------------------------------------------------------------------------*/
/**
 */
bool ReaderOGG::onReset() {
	if (getFile() == NULL)
		return false;

	getFile()->setPosition(m_FileStart);
	try {
		ov_clear(&m_File);
		openFile();
	} catch(...) {
		return false;
	}
	
	return true;	
}


/*--------------------------------------------------------------------------*/
/**
 */
size_t ReaderOGG::file_read_func(void *ptr, size_t size, size_t nmemb, void *datasource) {
	 size_t bytes = static_cast<File*>(datasource)->read(ptr, size*nmemb);
	 if ((int)bytes <= 0)
		return 0;
	 return bytes / size;
}


/*--------------------------------------------------------------------------*/
/**
 */
int ReaderOGG::file_seek_func(void *datasource, ogg_int64_t offset, int whence) {
	switch(whence) {
		case SEEK_SET:
			static_cast<File*>(datasource)->setPosition(offset, File::Absolute);
			return 0;
		case SEEK_CUR:
			static_cast<File*>(datasource)->setPosition(offset, File::Current);
			return 0;
		case SEEK_END:
			static_cast<File*>(datasource)->setPosition(offset, File::End);
			return 0;
		default:
			return -1;
	}
}


/*--------------------------------------------------------------------------*/
/**
 */
int ReaderOGG::file_close_func(void* ) {
	return 0;
}


/*--------------------------------------------------------------------------*/
/**
 */
long ReaderOGG::file_tell_func(void* datasource) {
	return (long)static_cast<File*>(datasource)->getPosition();
}


/*--------------------------------------------------------------------------*/
/**
 */
size_t ReaderOGG::stream_read_func(void *ptr, size_t size, size_t nmemb, void *datasource) {
	 size_t bytes = static_cast<InputStream*>(datasource)->read(ptr, size*nmemb);
	 if ((int)bytes <= 0)
		return 0;
	 return bytes / size;
}


/*--------------------------------------------------------------------------*/
/**
 */
int ReaderOGG::stream_seek_func(void *datasource, ogg_int64_t offset, int whence) {
	return -1;
}


/*--------------------------------------------------------------------------*/
/**
 */
int ReaderOGG::stream_close_func(void* ) {
	return 0;
}


/*--------------------------------------------------------------------------*/
/**
 */
long ReaderOGG::stream_tell_func(void* ) {
	return 0;
}


}; };
