/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "io.exception.h"
#include "media.exception.h"
#include "media.buffer.h"
#include "media.reader.wav.h"

#include "object.pointer.inl"
#include "io.file.inl"
#include "io.stream.inl"
#include "io.inputstream.inl"
#include "media.node.inl"
#include "media.pin.inl"
#include "media.format.inl"
#include "media.reader.inl"
#include "media.buffer.inl"

namespace magnum { namespace media {
using magnum::io::FileReadException;
using magnum::types::byte;

#define MCHAR4(a, b, c, d) (a | (b << 8) | (c << 16) | (d << 24))


/*--------------------------------------------------------------------------*/
/**
 */
ReaderWAV::ReaderWAV(File* file) : Reader(file) {
	m_Format = NULL;
	m_FileStart = file->getPosition();
	openFile();
	setupFormat();
}


/*--------------------------------------------------------------------------*/
/**
 */
ReaderWAV::ReaderWAV(InputStream* in) : Reader(in) {
	m_Format = NULL;
	openFile();
	setupFormat();
}


/*--------------------------------------------------------------------------*/
/**
 */
ReaderWAV::~ReaderWAV() {
	if (m_Format != NULL) {
		delete[] (byte*)m_Format;
		m_Format = NULL;
	}
}


/*--------------------------------------------------------------------------*/
/**
 */
bool ReaderWAV::openFile() {
	InputStream* in = getStream();

	// Release old format specs
	if (m_Format != NULL) {
		delete[] (byte*)m_Format;
		m_Format = NULL;
	}

	// Read RIFF header
	RIFFHeader rhead;
	if (in->read(&rhead, sizeof(rhead)) != sizeof(rhead))
		throw MediaFileFormatException();
	if ((rhead.m_ChunkId != MCHAR4('R','I','F','F'))
		|| (rhead.m_Format != MCHAR4('W','A','V','E')))
		throw MediaFileFormatException();

	// Read format header
	ChunkHeader chunk;
	if (in->read(&chunk, sizeof(chunk)) != sizeof(chunk))
		throw MediaFileFormatException();
	if (chunk.m_ChunkId != MCHAR4('f','m','t',' '))
		throw MediaFileFormatException();
	m_Format = (FormatHeader*)new byte[chunk.m_ChunkSize];
	if (in->read(m_Format, chunk.m_ChunkSize) != chunk.m_ChunkSize)
		throw MediaFileFormatException();
	
	// Find data chunk
	for (;;) {
		if (in->read(&chunk, sizeof(chunk)) != sizeof(chunk))
			throw MediaFileFormatException();
		if (chunk.m_ChunkId = MCHAR4('d','a','t','a')) {
			m_ListLeft = 0;
			m_DataLeft = chunk.m_ChunkSize;
			break;
		}
		if (chunk.m_ChunkId = MCHAR4('w','a','v','l')) {
			m_ListLeft = chunk.m_ChunkSize;
			m_DataLeft = 0;
			break;
		}
		in->skip(chunk.m_ChunkSize);
	}

	return true;
}


/*--------------------------------------------------------------------------*/
/**
 */
bool ReaderWAV::setupFormat() {	
	Format fmt(Format::Audio);
	fmt.setBitrate(m_Format->m_ByteRate*8);
	fmt.setBufferSize(m_Format->m_ByteRate);
	if (m_Format->m_Format != 1)
		throw UnsupportedCodecException();		
	fmt.setCodec("raw");
	fmt.audio().setChannels(m_Format->m_Channels);
	fmt.audio().setSamplerate(m_Format->m_SampleRate);
	if (m_Format->m_BitDepth == 8)
		fmt.audio().setFormat(AudioFormat::Signed8);
	else if (m_Format->m_BitDepth == 16)
		fmt.audio().setFormat(AudioFormat::Signed16);
	else
		throw UnsupportedCodecException();		
		
	setFormat(fmt);

	return true;
}


/*--------------------------------------------------------------------------*/
/**
 */
bool ReaderWAV::onSetFormat(const Format& fmt) {
	if (fmt.getType() != Format::Audio)
		return false;
	if (fmt.getCodec() != "raw")
		return false;
	if (m_Format == NULL)
		return true;
		
	if (fmt.audio().getSamplerate() != m_Format->m_SampleRate)
		return false;
	if (fmt.audio().getChannels() != m_Format->m_Channels)
		return false;
	if (fmt.getBufferSize() < 8000)
		return false;
		
	switch (fmt.audio().getFormat()) {
		case AudioFormat::Signed8:
			if (m_Format->m_BitDepth != 8)
				return false;
			break;
		case AudioFormat::Signed16:
			if (m_Format->m_BitDepth != 16)
				return false;
			break;
		default:
			return false;
	}

	return true;	
}


/*--------------------------------------------------------------------------*/
/**
 */
bool ReaderWAV::onAdjustFormat(Format& fmt) {
	if (fmt.getType() != Format::Audio)
		return false;
	if (m_Format == NULL)
		return true;
		
	fmt.setCodec("raw");
	fmt.audio().setSamplerate(m_Format->m_SampleRate);
	fmt.audio().setChannels(m_Format->m_Channels);
	if (fmt.getBufferSize() < 8000)
		fmt.setBufferSize(8000);
		
	if (m_Format->m_BitDepth == 8)
		fmt.audio().setFormat(AudioFormat::Signed8);
	else if (m_Format->m_BitDepth == 16)
		fmt.audio().setFormat(AudioFormat::Signed16);

	return true;	
}


/*--------------------------------------------------------------------------*/
/**
 */
bool ReaderWAV::onRead(Buffer* buf) {
	InputStream* in = getStream();
	char* data = (char*)buf->getBuffer();
	unsigned int bytes_left = buf->getMaxSize();
	bytes_left -= bytes_left % 16;
	unsigned int bytes_read = 0;

	for (;;) {
		if (m_DataLeft > 0) {
			unsigned int toread = min(m_DataLeft, bytes_left);
			toread = (unsigned int)in->read(data, toread);
			bytes_left -= toread;
			bytes_read += toread;
			m_DataLeft -= toread;
			if (bytes_left == 0)
				break;
			continue;
		} else if (m_ListLeft > 0) {
			for(;;) {
				if (m_ListLeft == 0)
					break;
				ChunkHeader chunk;
				in->read(&chunk, sizeof(chunk));
				m_ListLeft -= sizeof(chunk) + chunk.m_ChunkSize;
				if (chunk.m_ChunkId != MCHAR4('d','a','t','a')) {
					in->skip(chunk.m_ChunkSize);
					continue;
				}
				m_DataLeft = chunk.m_ChunkSize;
				break;
			}
		} else {
			break;
		}
	}	

	if (bytes_read == 0) {
		endOfOutput(getPin());
		return false;
	}
	buf->setSize(bytes_read);
	return true;
}


/*--------------------------------------------------------------------------*/
/**
 */
bool ReaderWAV::onReset() {
	if (getFile() == NULL)
		return false;

	getFile()->setPosition(m_FileStart);
	try {
		openFile();
	} catch(...) {
		return false;
	}
	
	return true;	
}

}; };
