/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "media.buffer.h"
#include "media.wavedevice.h"
#include "media.waveinput.h"

#include "base.string.inl"
#include "object.class.inl"
#include "object.inl"
#include "container.allocator.inl"
#include "container.iterator.inl"
#include "container.list.inl"
#include "media.looper.inl"
#include "media.node.inl"
#include "media.pin.inl"
#include "media.buffer.inl"
#include "media.audio.inl"
#include "media.wavedevice.inl"
#include "media.waveinput.inl"


namespace magnum { namespace media { 


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
WaveInput::WaveInput(WaveDevice* dev) : AudioProducer(dev, Node::DefaultInput) , m_OutputPin(this) {
	m_BufferCount = 3;
	m_BufferSize = 0x4000;
	m_BufferQueue = 0;
	
	m_Format.wFormatTag = WAVE_FORMAT_PCM;
	m_Format.nChannels = 2;
	m_Format.nSamplesPerSec = 44100;
	m_Format.nAvgBytesPerSec = 44100*2*2;
	m_Format.nBlockAlign = 4;	
	m_Format.wBitsPerSample = 16;
	m_Format.cbSize = 0;

	Format fmt(Format::Audio);
	fmt.setBitrate(44100*2*16);
	fmt.setBufferSize(44100*2*2);
	fmt.setCodec("raw");
	fmt.audio().setChannels(2);
	fmt.audio().setSamplerate(44100);
	fmt.audio().setFormat(AudioFormat::Signed16);
	m_OutputPin.requestFormat(fmt);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
WaveInput::~WaveInput() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Buffer* WaveInput::createBuffer() {
	m_BufferMutex.lock();
	WAVEHDR* hdr = m_Buffers.append();
	Buffer* buf = new Buffer(this, hdr, m_BufferSize);
	hdr->lpData = (LPSTR)buf->getBuffer();
	hdr->dwBufferLength = 0;
	hdr->dwUser = NULL;
	hdr->dwFlags = 0;
	hdr->dwLoops = 0;
	m_BufferMutex.unlock();

	if (waveInPrepareHeader(m_WaveHandle, hdr, sizeof(WAVEHDR)) != MMSYSERR_NOERROR)
		return NULL;

	return buf;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveInput::releaseBuffer(Buffer* buf) {
	WAVEHDR* hdr;
	
	if (buf->getOwner() == this) {
		hdr = (WAVEHDR*)buf->getOwnerData();
		buf->resetOwner();
	} else {
		hdr = (WAVEHDR*)buf->getUserData();
	}

	// Unprepare header
	waveInUnprepareHeader(m_WaveHandle, hdr, sizeof(WAVEHDR));
	
	// The buffer was only queued, if the length is greater than 0
	buf->release();

	// Remove wave header
	m_BufferMutex.lock();
	m_Buffers.remove(hdr);
	m_BufferMutex.unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveInput::onStart() {
	MMRESULT result;

	result = waveInOpen(&m_WaveHandle, WAVE_MAPPER, &m_Format, (DWORD_PTR)(intptr_t)&WaveInput::callback, (DWORD_PTR)(intptr_t)this, CALLBACK_FUNCTION | WAVE_FORMAT_DIRECT);
	if (result != MMSYSERR_NOERROR) {
		result = waveInOpen(&m_WaveHandle, WAVE_MAPPER, &m_Format, (DWORD_PTR)(intptr_t)&WaveInput::callback, (DWORD_PTR)(intptr_t)this, CALLBACK_FUNCTION);
		if (result != MMSYSERR_NOERROR)
			return false;
	}

	// Add some buffers
	m_BufferQueue = 0;
	for (unsigned int i = 0; i < m_BufferCount ;i++) {
		Buffer* buffer = createBuffer();
		WAVEHDR* hdr = (WAVEHDR*)buffer->getOwnerData();
		hdr->dwBufferLength = buffer->getMaxSize();
		if (waveInAddBuffer(m_WaveHandle, hdr, sizeof(WAVEHDR)) != MMSYSERR_NOERROR)
			return false;
		m_BufferQueue++;
	}

	// Start recording
	if (waveInStart(m_WaveHandle) != MMSYSERR_NOERROR)
		goto errorStart;

	return true;

errorStart:
	waveInClose(m_WaveHandle);
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveInput::onStop() {
	// Stop recording
	if (waveInStop(m_WaveHandle) != MMSYSERR_NOERROR)
		return false;
	if (waveInReset(m_WaveHandle) != MMSYSERR_NOERROR)
		return false;

	// Release all pending buffers
	processMessages();
		
	// Close device
	if (waveInClose(m_WaveHandle) != MMSYSERR_NOERROR)
		return false;

	return true;	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveInput::onReset() {
	// Reset recording
	if (waveInReset(m_WaveHandle) != MMSYSERR_NOERROR)
		return false;
	
	// Add some buffers
	for (unsigned int i = m_BufferQueue; i < m_BufferCount ;i++) {
		Buffer* buffer = createBuffer();
		WAVEHDR* hdr = (WAVEHDR*)buffer->getOwnerData();
		hdr->dwBufferLength = buffer->getMaxSize();
		if (waveInAddBuffer(m_WaveHandle, hdr, sizeof(WAVEHDR)) != MMSYSERR_NOERROR)
			return false;
		m_BufferQueue++;
	}

	return true;	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveInput::onPause() {
	return (waveInStop(m_WaveHandle) == MMSYSERR_NOERROR);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveInput::onResume() {
	return (waveInStart(m_WaveHandle) == MMSYSERR_NOERROR);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveInput::onSetFormat(Pin* pin, const Format& fmt) {
	if (pin != &m_OutputPin)
		return false;
	if (fmt.getType() != Format::Audio)
		return false;
	if (fmt.getCodec() != "raw")
		return false;
	if (fmt.getBufferSize() < 8000)
		return false;
		
	WAVEFORMATEX wfmt;
	wfmt.wFormatTag = WAVE_FORMAT_PCM;
	wfmt.nChannels = fmt.audio().getChannels();
	wfmt.nSamplesPerSec = fmt.audio().getSamplerate();
	switch (fmt.audio().getFormat()) {
		case AudioFormat::Signed8:
			wfmt.nBlockAlign = m_Format.nChannels;
			wfmt.nAvgBytesPerSec = wfmt.nSamplesPerSec * wfmt.nBlockAlign;
			wfmt.wBitsPerSample = 8;
			wfmt.cbSize = 0;
			break;
		case AudioFormat::Signed16:
			wfmt.nBlockAlign = m_Format.nChannels*2;	
			wfmt.nAvgBytesPerSec = wfmt.nSamplesPerSec * wfmt.nBlockAlign;
			wfmt.wBitsPerSample = 16;
			wfmt.cbSize = 0;
			break;
		default:
			return false;
	}
	
	if (waveInOpen(NULL, WAVE_MAPPER, &wfmt, NULL, NULL, WAVE_FORMAT_DIRECT | WAVE_FORMAT_QUERY) != MMSYSERR_NOERROR)
		if (waveInOpen(NULL, WAVE_MAPPER, &wfmt, NULL, NULL, WAVE_FORMAT_QUERY) != MMSYSERR_NOERROR)
			return false;

	m_Format = wfmt;
	m_BufferSize = fmt.getBufferSize();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveInput::onAdjustFormat(Pin* pin, Format& fmt) {
	if (pin != &m_OutputPin)
		return false;
	if (fmt.getType() != Format::Audio)
		return false;
	if (fmt.getCodec() != "raw")
		fmt.setCodec("raw");
	if (fmt.getBufferSize() < 8000)
		fmt.setBufferSize(8000);
		
	WAVEFORMATEX wfmt;
	wfmt.wFormatTag = WAVE_FORMAT_PCM;
	if (fmt.audio().getChannels() > 2)
		fmt.audio().setChannels(2);
	wfmt.nChannels = fmt.audio().getChannels();
	wfmt.nSamplesPerSec = fmt.audio().getSamplerate();
	switch (fmt.audio().getFormat()) {
		case AudioFormat::Unsigned8:
			fmt.audio().setFormat(AudioFormat::Signed8);
		case AudioFormat::Signed8:
			fmt.setBitrate(m_Format.nChannels*8*fmt.audio().getSamplerate());
			wfmt.nAvgBytesPerSec = fmt.getBitrate() / 8;
			wfmt.nBlockAlign = m_Format.nChannels;	
			wfmt.wBitsPerSample = 8;
			wfmt.cbSize = 0;
			break;
		case AudioFormat::Unsigned16:
			fmt.audio().setFormat(AudioFormat::Signed16);
		case AudioFormat::Signed16:
			fmt.setBitrate(m_Format.nChannels*16*fmt.audio().getSamplerate());
			wfmt.nAvgBytesPerSec = fmt.getBitrate() / 16;
			wfmt.nBlockAlign = m_Format.nChannels*2;	
			wfmt.wBitsPerSample = 16;
			wfmt.cbSize = 0;
			break;
		default:
			return false;
	}
	
	if (waveInOpen(NULL, WAVE_MAPPER, &wfmt, NULL, NULL, WAVE_FORMAT_DIRECT | WAVE_FORMAT_QUERY) != MMSYSERR_NOERROR)
		if (waveInOpen(NULL, WAVE_MAPPER, &wfmt, NULL, NULL, WAVE_FORMAT_QUERY) != MMSYSERR_NOERROR)
			return false;

	return true;		
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void WaveInput::onDestroyBuffer(Buffer* buf) {
	ASSERT(buf->getOwner() == this);
	WAVEHDR* hdr = (WAVEHDR*)buf->getOwnerData();
	ASSERT(hdr != NULL);
	
	if (getState() != Stopped)
		waveInUnprepareHeader(m_WaveHandle, hdr, sizeof(WAVEHDR));

	m_BufferMutex.lock();
	m_Buffers.remove(hdr);
	m_BufferMutex.unlock();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void WaveInput::onEndOfOutput(OutputPin* ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveInput::onRequestBuffer(OutputPin* pin,Buffer* buf) {
	if (pin != &m_OutputPin)
		return false;
		
	if (buf == NULL) {
		if (m_BufferQueue >= m_BufferCount)
			return true;
		buf = createBuffer();
		WAVEHDR* hdr = (WAVEHDR*)buf->getOwnerData();
		hdr->dwBufferLength = buf->getMaxSize();
		waveInAddBuffer(m_WaveHandle, hdr, sizeof(WAVEHDR));
		m_BufferQueue++;
		return true;
	} else if (buf->getOwner() == this) {
		WAVEHDR* hdr = (WAVEHDR*)buf->getOwnerData();
		hdr->dwBufferLength = buf->getMaxSize();
		waveInAddBuffer(m_WaveHandle, hdr, sizeof(WAVEHDR));
		m_BufferQueue++;
		return true;
	} else {
		m_BufferMutex.lock();
		WAVEHDR* hdr = m_Buffers.append();
		hdr->lpData = (LPSTR)buf->getBuffer();
		hdr->dwBufferLength = buf->getMaxSize();
		hdr->dwUser = (DWORD_PTR)buf;
		hdr->dwFlags = 0;
		hdr->dwLoops = 0;
		buf->setUserData(hdr);
		m_BufferMutex.unlock();
		waveInPrepareHeader(m_WaveHandle, hdr, sizeof(WAVEHDR));
		waveInAddBuffer(m_WaveHandle, hdr, sizeof(WAVEHDR));
		m_BufferQueue++;
		return true;
	}	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void CALLBACK WaveInput::callback(HWAVEIN wh,UINT uMsg,DWORD dwInstance,DWORD dwParam1,DWORD dwParam2) {
	if (uMsg != WIM_DATA)	
		return;

	WaveInput* node = (WaveInput*)(intptr_t)dwInstance;
	WAVEHDR* hdr = (WAVEHDR *)(intptr_t)dwParam1;
	Buffer* buf = (Buffer *)hdr->dwUser;
	buf->setSize(hdr->dwBytesRecorded);
	
	if (buf->getOwner() != node) {
		waveInUnprepareHeader(wh, hdr, sizeof(WAVEHDR));
		node->m_BufferMutex.lock();
		node->m_Buffers.remove(hdr);
		node->m_BufferMutex.unlock();
	}
	node->postBuffer(&node->m_OutputPin, buf);
	node->m_BufferQueue--;
	if ((node->m_BufferQueue < node->m_BufferSize) && (node->getState() == Node::Running))
		node->requestBuffer(&node->m_OutputPin);
}

}; };
