/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "media.buffer.h"
#include "media.wavedevice.h"
#include "media.waveoutput.h"

#include "base.string.inl"
#include "object.class.inl"
#include "object.inl"
#include "container.allocator.inl"
#include "container.iterator.inl"
#include "container.list.inl"
#include "system.atom.inl"
#include "media.looper.inl"
#include "media.node.inl"
#include "media.pin.inl"
#include "media.audio.inl"
#include "media.buffer.inl"
#include "media.wavedevice.inl"
#include "media.waveoutput.inl"


namespace magnum { namespace media {


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
WaveOutput::WaveOutput(WaveDevice* dev) : AudioConsumer(dev, Node::DefaultOutput) , m_InputPin(this) {
	m_BufferCount = 3;
	m_BufferQueue = 0;
	m_BufferLength = 800;
	m_BufferSize = m_BufferLength*44100*4/1000;
	
	m_WaveHandle = NULL;
	m_Format.wFormatTag = WAVE_FORMAT_PCM;
	m_Format.nChannels = 2;
	m_Format.nSamplesPerSec = 44100;
	m_Format.nAvgBytesPerSec = 44100*2*2;
	m_Format.nBlockAlign = 4;
	m_Format.wBitsPerSample = 16;
	m_Format.cbSize = 0;
	
	Format fmt(Format::Audio);
	fmt.setBitrate(44100*2*16);
	fmt.setBufferSize(44100*2*2);
	fmt.setCodec("raw");
	fmt.audio().setChannels(2);
	fmt.audio().setSamplerate(44100);
	fmt.audio().setFormat(AudioFormat::Signed16);
	m_InputPin.requestFormat(fmt);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
WaveOutput::~WaveOutput() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Buffer* WaveOutput::createBuffer() {
	m_BufferMutex.lock();
	WAVEHDR* hdr = m_Buffers.append();
	Buffer* buf = new Buffer(this, hdr, m_BufferSize);
	hdr->lpData = (LPSTR)buf->getBuffer();
	hdr->dwBufferLength = buf->getMaxSize();
	hdr->dwUser = (DWORD_PTR)buf;
	hdr->dwFlags = 0;
	hdr->dwLoops = 0;
	m_BufferMutex.unlock();

	if (waveOutPrepareHeader(m_WaveHandle, hdr, sizeof(WAVEHDR)) != MMSYSERR_NOERROR)
		return NULL;

	return buf;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveOutput::releaseBuffer(Buffer* buf) {
	WAVEHDR* hdr;
	
	if (buf->getOwner() == this) {
		hdr = (WAVEHDR*)buf->getOwnerData();
		buf->resetOwner();
	} else {
		hdr = (WAVEHDR*)buf->getUserData();
	}
		
	// Unprepare header
	waveOutUnprepareHeader(m_WaveHandle, hdr, sizeof(WAVEHDR));

	buf->release();

	// Remove wave header
	m_BufferMutex.lock();
	m_Buffers.remove(hdr);
	m_BufferMutex.unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void WaveOutput::insertPreBuffer(Buffer* buf) {
	if (buf->getOwner() == this) {
		// Get WAVE header an fill in information
		WAVEHDR* hdr = (WAVEHDR*)buf->getOwnerData();
		hdr->dwBufferLength = buf->getSize();
		m_BufferMutex.lock();
		m_PreBuffers.insert(buf);
		m_BufferMutex.unlock();
	} else {
		// Create new WAVE header structure
		m_BufferMutex.lock();
		WAVEHDR* hdr = m_Buffers.append();
		hdr->lpData = (LPSTR)buf->getBuffer();
		hdr->dwBufferLength = buf->getSize();
		hdr->dwUser = (DWORD_PTR)buf;
		hdr->dwFlags = 0;
		hdr->dwLoops = 0;
		buf->setUserData(hdr);
		m_PreBuffers.insert(buf);
		m_BufferMutex.unlock();
		waveOutPrepareHeader(m_WaveHandle, hdr, sizeof(WAVEHDR));
	}		
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void WaveOutput::playPreBuffers() {
	m_BufferMutex.lock();
	for (;;) {
		// Get buffer pointer
		List<Buffer*>::Iterator iter = m_PreBuffers.begin();
		if (iter == m_PreBuffers.end())
			break;
		Buffer* buf = *iter;
		m_PreBuffers.remove(iter);
		
		// Playback buffer
		WAVEHDR* hdr;
		if (buf->getOwner() == this)
			hdr = (WAVEHDR*)buf->getOwnerData();
		else
			hdr = (WAVEHDR*)buf->getUserData();
		waveOutWrite(m_WaveHandle, hdr, sizeof(WAVEHDR));
		m_BufferQueue++;
	}
	m_BufferMutex.unlock();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void WaveOutput::releasePreBuffers() {
	m_BufferMutex.lock();
	for (;;) {
		// Get buffer pointer
		List<Buffer*>::Iterator iter = m_PreBuffers.begin();
		if (iter == m_PreBuffers.end())
			break;
		Buffer* buf = *iter;
		m_PreBuffers.remove(iter);
		
		// Playback buffer
		WAVEHDR* hdr;
		if (buf->getOwner() == this) {
			hdr = (WAVEHDR*)buf->getOwnerData();
			buf->resetOwner();
		} else {
			hdr = (WAVEHDR*)buf->getUserData();
		}
		// Unprepare header
		waveOutUnprepareHeader(m_WaveHandle, hdr, sizeof(WAVEHDR));
		buf->release();
	}
	m_BufferMutex.unlock();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveOutput::onStart() {
	MMRESULT result;
	
	result = waveOutOpen(&m_WaveHandle, WAVE_MAPPER, &m_Format, (DWORD_PTR)(intptr_t)&WaveOutput::callback, (DWORD_PTR)(intptr_t)this, CALLBACK_FUNCTION | WAVE_FORMAT_DIRECT);
	if (result != MMSYSERR_NOERROR) {
		result = waveOutOpen(&m_WaveHandle, WAVE_MAPPER, &m_Format, (DWORD_PTR)(intptr_t)&WaveOutput::callback, (DWORD_PTR)(intptr_t)this, CALLBACK_FUNCTION);
		if (result != MMSYSERR_NOERROR)
			return false;
	}

	// Request some buffers
	if (m_InputPin.getConnection() == NULL)
		return true;

	m_NoRequests = false;
	m_BufferQueue = 0;		
	for (unsigned int i = 0; i < m_BufferCount ;i++) {
		Buffer* buf = createBuffer(); 
		requestBuffer(&m_InputPin, buf);
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveOutput::onStop() {
	// Stop playback
	m_NoRequests = true;
	if (waveOutReset(m_WaveHandle) != MMSYSERR_NOERROR)
		return false;

	processMessages();
	releasePreBuffers();
	
	if (waveOutClose(m_WaveHandle) != MMSYSERR_NOERROR)
		return false;

	return true;	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveOutput::onReset() {
	if (getState() != Node::Running)
		return true;
		
	// Restart playback
	m_NoRequests = false;
	if (waveOutReset(m_WaveHandle) != MMSYSERR_NOERROR)
		return false;

	releasePreBuffers();
		
	// Request some additional buffers buffers
	if (m_InputPin.getConnection() == NULL)
		return true;

	for (unsigned int i = m_BufferQueue; i < m_BufferCount ;i++) {
		Buffer* buf = createBuffer(); 
		requestBuffer(&m_InputPin, buf);
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveOutput::onPause() {
	return (waveOutPause(m_WaveHandle) == MMSYSERR_NOERROR);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveOutput::onResume() {
	return (waveOutRestart(m_WaveHandle) == MMSYSERR_NOERROR);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveOutput::onSetFormat(Pin* pin, const Format& fmt) {
	if (pin != &m_InputPin)
		return false;
	if (fmt.getType() != Format::Audio)
		return false;
	if (fmt.getCodec() != "raw")
		return false;
		
	WAVEFORMATEX wfmt;
	wfmt.wFormatTag = WAVE_FORMAT_PCM;
	wfmt.nChannels = fmt.audio().getChannels();
	wfmt.nSamplesPerSec = fmt.audio().getSamplerate();
	switch (fmt.audio().getFormat()) {
		case AudioFormat::Signed8:
			wfmt.nBlockAlign = m_Format.nChannels;
			wfmt.nAvgBytesPerSec = wfmt.nSamplesPerSec * wfmt.nBlockAlign;
			wfmt.wBitsPerSample = 8;
			wfmt.cbSize = 0;
			break;
		case AudioFormat::Signed16:
			wfmt.nBlockAlign = m_Format.nChannels*2;	
			wfmt.nAvgBytesPerSec = wfmt.nSamplesPerSec * wfmt.nBlockAlign;
			wfmt.wBitsPerSample = 16;
			wfmt.cbSize = 0;
			break;
		default:
			return false;
	}
	
	unsigned int bsize = m_BufferLength * wfmt.nAvgBytesPerSec / 1000;
	if (fmt.getBufferSize() < bsize)
		return false;

	if (waveOutOpen(NULL, WAVE_MAPPER, &wfmt, NULL, NULL, WAVE_FORMAT_DIRECT | WAVE_FORMAT_QUERY) != MMSYSERR_NOERROR)
		if (waveOutOpen(NULL, WAVE_MAPPER, &wfmt, NULL, NULL, WAVE_FORMAT_QUERY) != MMSYSERR_NOERROR)
			return false;

	m_Format = wfmt;
	m_BufferSize = fmt.getBufferSize();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveOutput::onAdjustFormat(Pin* pin, Format& fmt) {
	if (pin != &m_InputPin)
		return false;
	if (fmt.getType() != Format::Audio)
		return false;
	if (fmt.getCodec() != "raw")
		fmt.setCodec("raw");
		
	WAVEFORMATEX wfmt;
	wfmt.wFormatTag = WAVE_FORMAT_PCM;
	if (fmt.audio().getChannels() > 2)
		fmt.audio().setChannels(2);
	wfmt.nChannels = fmt.audio().getChannels();
	wfmt.nSamplesPerSec = fmt.audio().getSamplerate();
	switch (fmt.audio().getFormat()) {
		case AudioFormat::Unsigned8:
			fmt.audio().setFormat(AudioFormat::Signed8);
		case AudioFormat::Signed8:
			fmt.setBitrate(m_Format.nChannels*8*fmt.audio().getSamplerate());
			wfmt.nAvgBytesPerSec = fmt.getBitrate() / 8;
			wfmt.nBlockAlign = m_Format.nChannels;	
			wfmt.wBitsPerSample = 8;
			wfmt.cbSize = 0;
			break;
		case AudioFormat::Unsigned16:
			fmt.audio().setFormat(AudioFormat::Signed16);
		case AudioFormat::Signed16:
			fmt.setBitrate(m_Format.nChannels*16*fmt.audio().getSamplerate());
			wfmt.nAvgBytesPerSec = fmt.getBitrate() / 8;
			wfmt.nBlockAlign = m_Format.nChannels*2;	
			wfmt.wBitsPerSample = 16;
			wfmt.cbSize = 0;
			break;
		default:
			return false;
	}

	unsigned int bsize = m_BufferLength * wfmt.nAvgBytesPerSec / 1000;
	if (fmt.getBufferSize() < bsize)
		fmt.setBufferSize(bsize);
	
	if (waveOutOpen(NULL, WAVE_MAPPER, &wfmt, NULL, NULL, WAVE_FORMAT_DIRECT | WAVE_FORMAT_QUERY) != MMSYSERR_NOERROR)
		if (waveOutOpen(NULL, WAVE_MAPPER, &wfmt, NULL, NULL, WAVE_FORMAT_QUERY) != MMSYSERR_NOERROR)
			return false;

	return true;		
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void WaveOutput::onDestroyBuffer(Buffer* buf) {
	ASSERT(buf->getOwner() == this);
	WAVEHDR* hdr = (WAVEHDR*)buf->getOwnerData();
	ASSERT(hdr != NULL);
	
	if (getState() != Stopped)
		waveOutUnprepareHeader(m_WaveHandle, hdr, sizeof(WAVEHDR));

	// Remove wave header
	m_BufferMutex.lock();
	m_Buffers.remove(hdr);
	m_BufferMutex.unlock();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void WaveOutput::onEndOfInput(InputPin* ) {
	m_NoRequests = true;
	playPreBuffers();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveOutput::onReceiveBuffer(InputPin* ,Buffer* buf) {
	m_NoRequests = false;
	
	// Did we enqueue the first buffer?
	if (m_BufferQueue == 0) {
		insertPreBuffer(buf);
		if (m_PreBuffers.getSize() >= m_BufferCount) {
			playPreBuffers();
		} else {
			Buffer* buf = createBuffer(); 
			requestBuffer(&m_InputPin, buf);
		}
		return true;	
	}

	// Enqueue buffers
	m_BufferQueue++;
	if (buf->getOwner() == this) {
		// Get WAVE header an fill in information
		WAVEHDR* hdr = (WAVEHDR*)buf->getOwnerData();
		hdr->dwBufferLength = buf->getSize();
		waveOutWrite(m_WaveHandle, hdr, sizeof(WAVEHDR));
		return true;
	} else {
		// Create new WAVE header structure
		m_BufferMutex.lock();
		WAVEHDR* hdr = m_Buffers.append();
		hdr->lpData = (LPSTR)buf->getBuffer();
		hdr->dwBufferLength = buf->getSize();
		hdr->dwUser = (DWORD_PTR)buf;
		hdr->dwFlags = 0;
		hdr->dwLoops = 0;
		buf->setUserData(hdr);
		m_BufferMutex.unlock();
		waveOutPrepareHeader(m_WaveHandle, hdr, sizeof(WAVEHDR));
		waveOutWrite(m_WaveHandle, hdr, sizeof(WAVEHDR));
		return true;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool WaveOutput::onLoop() {
	// Request some new buffers (if needed)
	if (getState() == Running && !m_NoRequests) {
	}
	
	// Process any messages
	return processMessages();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void WaveOutput::onHandleMessage(const Message& msg) {
	switch (msg.getCommand()) {
		case Message::ReceiveBuffer:
			ASSERT(msg.getPin() != NULL);
			ASSERT(msg.getPointer() != NULL);
			onReceiveBuffer((InputPin*)msg.getPin(), (Buffer*)msg.getPointer());
			return;
		case Message::RequestBuffer:
			ASSERT(msg.getPin() != NULL);
			onRequestBuffer((OutputPin*)msg.getPin(), (Buffer*)msg.getPointer());
			return;
		case ReleaseBuffer:
			releaseBuffer(static_cast<Buffer*>(msg.getPointer()));
			return;
		default:
			Node::onHandleMessage(msg);
			return;
	}
}			



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void CALLBACK WaveOutput::callback(HWAVEIN ,UINT uMsg,DWORD dwInstance,DWORD dwParam1,DWORD dwParam2) {
	if (uMsg != WOM_DONE)	
		return;

	WaveOutput* node = (WaveOutput*)(intptr_t)dwInstance;
	node->m_BufferQueue--;

	WAVEHDR* hdr = (WAVEHDR *)(intptr_t)dwParam1;
	Buffer* buf = (Buffer *)hdr->dwUser;
		
	// If the node is about to be stopped or received an
	// end of input, we should no longer request new
	// buffers
	if (node->m_NoRequests) {
		node->sendMessage(ReleaseBuffer, NULL, buf);
		return;	
	}
	
	if (buf->getOwner() != node) {
		node->sendMessage(ReleaseBuffer, NULL, buf);
		if ((node->m_BufferQueue < node->m_BufferCount) && (node->getState() == Node::Running))
			node->requestBuffer(&node->m_InputPin);
	} else {
		if ((node->m_BufferQueue < node->m_BufferCount) && (node->getState() == Node::Running))
			node->requestBuffer(&node->m_InputPin, buf);
		else
			node->sendMessage(ReleaseBuffer, NULL, buf);
	}
}

}; };
