/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "io.outputbuffer.h"
#include "io.datainput.h"
#include "net.exception.h"
#include "net.url.h"
#include "net.httprequest.h"
#include "net.httpresponse.h"
#include "net.httpclientconnection.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "object.softpointer.inl"
#include "base.string.inl"
#include "container.chain.inl"
#include "container.list.inl"
#include "container.allocator.inl"
#include "container.iterator.inl"
#include "container.pair.inl"
#include "io.datainput.inl"
#include "io.stream.inl"
#include "io.inputstream.inl"
#include "io.outputstream.inl"
#include "io.outputbuffer.inl"
#include "net.mimeheaders.inl"
#include "net.socket.inl"
#include "net.streamsocket.inl"
#include "net.url.inl"
#include "net.httpversion.inl"
#include "net.httprequest.inl"
#include "net.httpresponse.inl"
#include "net.httpclient.inl"
#include "net.httpinputstream.inl"
#include "net.httpclientconnection.inl"


namespace magnum { namespace net {
using magnum::object::AutoPointer;
using magnum::io::OutputBuffer;
using magnum::io::DataInput;

/*--------------------------------------------------------------------------*/
/**
 */
HttpClientConnection::HttpClientConnection() {
	m_Client = NULL;
	m_Stream = new HttpInputStream(this);
}


/*--------------------------------------------------------------------------*/
/**
 */
HttpClientConnection::HttpClientConnection(HttpClient* c)  {
	m_Client = c;
	m_Stream = new HttpInputStream(this);
}
	

/*--------------------------------------------------------------------------*/
/**
 */
HttpClientConnection::~HttpClientConnection() {
	m_Client = NULL;
	m_Stream->detach();
	m_Stream->release();
	m_Stream = NULL;
}


/*--------------------------------------------------------------------------*/
/**
 */
bool HttpClientConnection::connect(const URL& url) {
	unsigned int port = url.getPort();
	if (port == 0)
		port = 80;
	return m_Socket.connect(url.getHost(), port);
}


/*--------------------------------------------------------------------------*/
/**
 */
bool HttpClientConnection::disconnect() {
	m_Stream->close();
	return m_Socket.disconnect();
}


/*--------------------------------------------------------------------------*/
/**
 */
HttpResponse HttpClientConnection::request(const HttpRequest& request) {
	const URL& url = request.getURL();
	
	// Open socket
	if (!m_Socket.isConnected()) {
		connect(url);
	}

	// Use output buffering
	AutoPointer<OutputStream> out = new OutputBuffer(m_Socket.getOutputStream());
	
	// Build request string
	String cmd;
/*	if (request.getVersion() == HttpVersion::Http_1_0)
		cmd = request.getMethod() + " " + url.getQuery() + " HTTP/1.0\r\n";
	else if (request.getVersion() == HttpVersion::Http_1_1)
		cmd = request.getMethod() + " " + url.getFile() + " HTTP/1.1\r\nHost: "+url.getHost()+
		"\r\nAccept-Encoding:\r\n"; */
	if (url.isHostNumeric())
		cmd = request.getMethod() + " " + url.getFile() + " HTTP/1.0\r\n"; 
	else
		cmd = request.getMethod() + " " + url.getQuery() + " HTTP/1.0\r\n";

	// Send request to HTTP server
	out->write((const char*)cmd, cmd.getLength());
	request.getHeaders().writeTo(out);
	out->write("\r\n",3);
	out->flush();
	
	
	// Prepare input parser
	m_Stream->m_StreamLeft = -1;
	DataInput in(getInputStream());
	String line;
	
	// Read response
	line = in.readLine();
	if (line.getLength() < 12)
		throw HttpResponseException();
	if (line[0] != 'H' || line[1] != 'T' || line[2] != 'T' || line[3] != 'P' || line[4] != '/')
		throw HttpResponseException();

	// Read version string
	const char* c = &line[5];
	unsigned int vermajor = 0;	
	unsigned int verminor = 0;
	while(*c != '.') {
		if (*c < '0' || *c > '9')
			throw HttpResponseException();
		vermajor = 10*vermajor + (*c - '0');
		c++;
	}
	c++;
	while(*c != ' ') {
		if (*c < '0' || *c > '9')
			throw HttpResponseException();
		verminor = 10*verminor + (*c - '0');
		c++;
	}
	c++;
	
	unsigned int retcode = 0;
	while(*c != ' ') {
		if (*c < '0' || *c > '9')
			throw HttpResponseException();
		retcode = 10*retcode + (*c - '0');
		c++;
	}
	c++;

	// Build response
	HttpResponse response(request);
	response.setCode((HttpResponse::ResponseCode)retcode);
	if (vermajor == 0 && verminor == 9)
		response.setVersion(HttpVersion::Http_0_9);
	else if (vermajor == 1 && verminor == 0)
		response.setVersion(HttpVersion::Http_1_0);
	else if (vermajor == 1 && verminor == 1)
		response.setVersion(HttpVersion::Http_1_1);
	else
		response.setVersion(HttpVersion::Unknown);
		
	// Scan mime-headers
	MimeHeaders& headers = response.getHeaders();
	for (;;) {
		line = in.readLine();
		if (line.getLength() == 0)
			break;
		unsigned int sep = line.indexOf(':');
		if (sep == -1)
			continue;
		line[sep] = 0;
		if (line.getLength() < sep+2)
			headers.insert(line, "");
		else
			headers.insert(line, &line[sep+2]);
	}
	// HACK: Read last \n
	in.readByte();	
	
	// Attach input stream to response
	response.setInputStream(getInputStream());
	
	return response;
}

}; };
