/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "net.httpclientconnection.h"
#include "net.httpinputstream.h"

#include "object.inl"
#include "io.stream.inl"
#include "io.inputstream.inl"
#include "net.httpclientconnection.inl"
#include "net.httpinputstream.inl"

namespace magnum { namespace net {
using magnum::math::min;

/*--------------------------------------------------------------------------*/
/**
 */
HttpInputStream::HttpInputStream(HttpClientConnection* con) {
	ASSERT(con != NULL);
	m_Connection = con;
	m_Stream = m_Connection->getSocket()->getInputStream();
	m_StreamLeft = 0;
	m_BufferLeft = 0;
	m_BufferCurrent = m_Buffer;
}


/*--------------------------------------------------------------------------*/
/**
 */
HttpInputStream::~HttpInputStream() {
	m_Stream = NULL;
	m_Connection = NULL;
	m_StreamLeft = 0;
	m_BufferLeft = 0;
	m_BufferCurrent = m_Buffer;
}


/*--------------------------------------------------------------------------*/
/**
 */
size_t HttpInputStream::onRead(void *buffer, size_t size) {
	if (m_StreamLeft != -1)
		size = min(size, m_StreamLeft + m_BufferLeft);
		
	size_t total = 0;

	// Perform a direct read for large requests
	if (size > BufferSize) {
		// Copy current buffer to destination
		if (m_BufferLeft > 0) {
			memcpy(buffer, m_BufferCurrent, m_BufferLeft);
			buffer = (byte*)buffer + m_BufferLeft;
			size -= m_BufferLeft;
			total = m_BufferLeft;
			m_BufferLeft = 0;
		}
		
		// Directly read from input stream
		while (size > 0) {
			size_t read = m_Stream->read(buffer, size);
			if ((int)read <= 0) {
				if (total > 0)
					return total;
				else
					return EndOfFile;
			}
			buffer = (byte*)buffer + read;
			size -= read;
			total += read;
		}
		return total;
	}	
	
	// Shortcut for single bytes, as needed by parsers
	if (size == 1) {
		// Refill buffer if it is empty
		if (m_BufferLeft == 0) {
			if (!fillBuffer())
				return EndOfFile;
		}
	
		*((byte*)buffer) = *m_BufferCurrent;
		m_BufferCurrent++;
		m_BufferLeft--;
		return 1;
	}

	// This is the normal buffered case for small requests
	while (size > 0) {
		// Do we need to fill the buffers?
		if (m_BufferLeft == 0) {
			if (!fillBuffer()) {
				if (total > 0)
					return total;
				else
					return EndOfFile;
			}
		}
		
		// Copy from internal buffer into dest buffer
		size_t bsize = min(m_BufferLeft, size);
		memcpy(buffer, m_BufferCurrent, bsize);
		m_BufferLeft -= bsize;
		m_BufferCurrent += bsize;
		size -= bsize;
		total += bsize;
		buffer = (byte*)buffer + bsize;
	}
	
	return total;
}


/*--------------------------------------------------------------------------*/
/**
 */
size_t HttpInputStream::onSkip(size_t size) {
	if (m_StreamLeft != -1)
		size = min(size, m_StreamLeft + m_BufferLeft);

	size_t total = 0;
	while (size > 0) {
		// Do we need to fill the buffers?
		if (m_BufferLeft == 0) {
			if (!fillBuffer()) {
				if (total > 0)
					return total;
				else
					return EndOfFile;
			}
		}

		// Skip buffer
		size_t bsize = min(m_BufferLeft, size);
		m_BufferLeft -= bsize;
		m_BufferCurrent += bsize;
		size -= bsize;
		total += bsize;
	}

	return total;
}


/*--------------------------------------------------------------------------*/
/**
 */
int64 HttpInputStream::onAvailable() {
	if (m_StreamLeft == -1)
		return 0x7fffffff;
	return m_StreamLeft + m_BufferLeft;
}


/*--------------------------------------------------------------------------*/
/**
 */
bool HttpInputStream::onClose() {
	while (m_StreamLeft > 0) {		
		int skipped = (int)m_Stream->skip(m_StreamLeft);
		if (skipped <= 0)
			break;
		m_StreamLeft -= skipped;
	}
	
	m_StreamLeft = 0;
	m_BufferLeft = 0;
	return true;
}


/*--------------------------------------------------------------------------*/
/**
 */
void HttpInputStream::detach() {
	m_Stream = NULL;
	m_Connection = NULL;
	m_StreamLeft = 0;
	m_BufferLeft = 0;
	m_BufferCurrent = m_Buffer;
}


/*--------------------------------------------------------------------------*/
/**
 */
bool HttpInputStream::fillBuffer() {
	size_t toread;
	if (m_StreamLeft == -1)
		toread = BufferSize;
	else
		toread = min(m_StreamLeft, (size_t)BufferSize);
		
	int ret = (int) m_Stream->read(m_Buffer, toread);
	if (ret <= 0) {
		m_StreamLeft = 0;
		return false;
	}
	m_BufferCurrent = m_Buffer;
	m_BufferLeft = ret;
	m_StreamLeft -= ret;
	return true;
}


}; };
