/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "net.serversocket.h"
#include "net.streamsocket.h"
#include "net.address.h"
#include "net.exception.h"

#include "base.string.inl"
#include "object.class.inl"
#include "object.inl"
#include "object.exception.inl"
#include "net.socket.inl"
#include "net.serversocket.inl"
#include "net.streamsocket.inl"
#include "net.address.inl"

namespace magnum { namespace net {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ServerSocket::ServerSocket() {
#if defined(__FORWINDOWS__)
	m_Socket = WSASocket(AF_INET, SOCK_STREAM, IPPROTO_TCP, NULL, 0, 0);
	m_SelectIndex = 0;
#else
	m_Socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	m_SelectIndex = getdtablehi();
#endif
	m_Port = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ServerSocket::~ServerSocket() {
	if (m_Socket != NULL) {
#if defined(__FORWINDOWS__)
		closesocket(m_Socket);
#else
		close(m_Socket);
#endif
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool ServerSocket::bind(const Address& localAdr, unsigned int localPort) {
	sockaddr_in adr;

	m_Address = localAdr;
	m_Port = htons((unsigned short)localPort);

	adr.sin_family = AF_INET;
	adr.sin_port = m_Port;
	adr.sin_addr.s_addr = m_Address.getAddressNO();

	if (::bind(m_Socket, (sockaddr*)&adr, sizeof(adr)) != 0)
		Socket::throwBindException();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool ServerSocket::bind(const String& localAdr, unsigned int localPort) {
	sockaddr_in adr;

	m_Address.setHostname(localAdr);
	m_Port = htons((unsigned short)localPort);

	adr.sin_family = AF_INET;
	adr.sin_port = m_Port;
	adr.sin_addr.s_addr = m_Address.getAddressNO();

	if (::bind(m_Socket, (sockaddr*)&adr, sizeof(adr)) != 0)
		Socket::throwBindException();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool ServerSocket::bind(unsigned int localPort) {
	int size;
	sockaddr_in adr;

	m_Port = htons((unsigned short)localPort);
	m_Address.setAddress(0);

	adr.sin_family = AF_INET;
	adr.sin_port = m_Port;
	adr.sin_addr.s_addr = INADDR_ANY;

	if (::bind(m_Socket, (sockaddr*)&adr, sizeof(adr)) != 0)
		Socket::throwBindException();

	size = sizeof(sockaddr_in);
	getsockname(m_Socket, (sockaddr *)&adr, &size);
	m_Address.setAddressNO(adr.sin_addr.s_addr);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
StreamSocket* ServerSocket::accept() {
	sockaddr_in adr;
	SOCKET socket;
	int adrSize = sizeof(adr);

#if defined(__FORWINDOWS__)
	socket = WSAAccept(m_Socket, (sockaddr*)&adr, &adrSize, NULL, NULL);
#elif defined(__FORUNIX__)
	socket = ::accept(m_Socket, (sockaddr*)&adr, &adrSize);
#endif
	if (socket == -1) {
		Socket::throwSocketException();
	}

	StreamSocket* streamSocket = new StreamSocket();

	streamSocket->m_Socket = socket;
	streamSocket->m_LocalAddress = m_Address;
	streamSocket->m_LocalPort = m_Port;
	streamSocket->m_RemoteAddress.setAddressNO(adr.sin_addr.s_addr);
	streamSocket->m_RemotePort = adr.sin_port;
	return streamSocket;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool ServerSocket::wait() {
	fd_set read;

#if defined(__FORWINDOWS__)
	read.fd_count = 1;
	read.fd_array[0] = m_Socket;
#else
	FD_ZERO(&read);
	FD_SET(m_Socket, &read);
#endif

	if (select(m_SelectIndex, &read, NULL, NULL, NULL) == SOCKET_ERROR)
		Socket::throwSocketException();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool ServerSocket::wait(unsigned int millis) {
	fd_set read;
	timeval time;
	int ret;

	time.tv_usec = 1000*millis % 1000000;
	time.tv_sec = millis / 1000;

#if defined(__FORWINDOWS__)
	read.fd_count = 1;
	read.fd_array[0] = m_Socket;
#else
	FD_ZERO(&read);
	FD_SET(m_Socket, &read);
#endif

	ret =  select(m_SelectIndex, &read, NULL, NULL, &time);
	if (ret == SOCKET_ERROR)
		Socket::throwSocketException();

	return (ret > 0);
}

}; };
