/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "net.socket.h"
#include "net.exception.h"

#include "io.stream.inl"
#include "io.inputstream.inl"
#include "io.outputstream.inl"
#include "base.string.inl"
#include "object.inl"
#include "object.exception.inl"
#include "object.class.inl"
#include "net.socket.inl"
#include "net.socketstream.inl"
#include "net.address.inl"


namespace magnum { namespace net {

const char* const Socket::ExceptionUnknown = "Unknown Socket error." ;
const char* const Socket::ExceptionNotInitialised = "Sockets not initialised." ;
const char* const Socket::ExceptionNetDown = "The network subsystem has failed."; 
const char* const Socket::ExceptionNotConnected = "The socket is not connected.";
const char* const Socket::ExceptionInterrupted = "The (blocking) call was canceled.";
const char* const Socket::ExceptionBlocked = "A blocking call is in progress.";
const char* const Socket::ExceptionNetReset = "The connection has been broken.";
const char* const Socket::ExceptionDescriptor = "The descriptor is not a socket.";
const char* const Socket::ExceptionAddressFault = "The Buffer is not completely contained in a valid part of the user address space."; 
const char* const Socket::ExceptionShutdown = "The socket has been shut down.";
const char* const Socket::ExceptionWouldBlock = "There are too many outstanding overlapped I/O requests.";
const char* const Socket::ExceptionMessageSize = "The message was too large to fit into the specified buffer.";
const char* const Socket::ExceptionNotBound = "The socket has not been bound.";
const char* const Socket::ExceptionAddressInUse = "The local address of the socket is already in use.";
const char* const Socket::ExceptionNoRouteToHost = "The network cannot be reached from this host at this time.";
const char* const Socket::ExceptionConnectionRefused = "The attempt to connect was rejected.";
const char* const Socket::ExceptionAlreadyConnected = "The socket is already connected.";
const char* const Socket::ExceptionConnectionAborted = "The virtual circuit was terminated due to a time-out or other failure.";
const char* const Socket::ExceptionConnectionReset = "The virtual circuit was reset by the remote side.";
const char* const Socket::ExceptionConnectionClosed = "Socket is message oriented and the virtual circuit was gracefully closed by the remote side."; 
const char* const Socket::ExceptionIOPending = "An overlapped operation was successfully initiated and completion will be indicated at a later time."; 
const char* const Socket::ExceptionIOAborted = "The overlapped operation has been canceled due to the closure of the socket."; 
const char* const Socket::ExceptionTimedOut = "Attempt to connect timed out without establishing a connection.";
const char* const Socket::ExceptionUnknownHost = "Unknown Host.";
const char* const Socket::ExceptionInvalidAddress = "Invalid Host-Address.";
const char* const Socket::ExceptionAddressNotAvailable = "Address not available.";
const char* const Socket::ExceptionBrokenPipe = "Broken Pipe.";




/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Socket::Socket(unsigned int protocol) : m_InputStream(this), m_OutputStream(this) {
#if defined(__FORWINDOWS__)
	m_Socket = WSASocket(AF_INET, protocol == IPPROTO_TCP ? SOCK_STREAM : SOCK_DGRAM, protocol, NULL, 0, 0);
	if (m_Socket == INVALID_SOCKET)
		Socket::throwSocketException();
	m_SelectIndex = 0;
#else
	m_Socket = socket(AF_INET, protocol == IPPROTO_TCP ? SOCK_STREAM : SOCK_DGRAM, protocol);
	m_SelectIndex = getdtablehi();
#endif
	m_LocalPort = 0;
	m_RemotePort = 0;
	m_Connected = false;
	m_Bound = false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Socket::Socket() : m_InputStream(this), m_OutputStream(this) {
	m_Socket = NULL;
	m_LocalPort = 0;
	m_RemotePort = 0;
	m_Connected = false;
	m_Bound = false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Socket::~Socket() {
	if (m_Socket != NULL) {
#if defined(__FORWINDOWS__)
		closesocket(m_Socket);
#else
		close(m_Socket);
#endif
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Socket::bind(const Address& localAdr, unsigned int localPort) {
	sockaddr_in adr;

	m_LocalAddress = localAdr;
	m_LocalPort = htons((unsigned short)localPort);

	adr.sin_family = AF_INET;
	adr.sin_port = (unsigned short)m_LocalPort;
	adr.sin_addr.s_addr = m_LocalAddress.getAddressNO();

	if (::bind(m_Socket, (sockaddr*)&adr, sizeof(adr)) != 0)
		Socket::throwBindException();

	m_Bound = true;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Socket::bind(const char* localAdr, unsigned int localPort) {
	sockaddr_in adr;

	m_LocalAddress.setHostname(localAdr);
	m_LocalPort = htons((unsigned short)localPort);

	adr.sin_family = AF_INET;
	adr.sin_port = m_LocalPort;
	adr.sin_addr.s_addr = m_LocalAddress.getAddressNO();

	if (::bind(m_Socket, (sockaddr*)&adr, sizeof(adr)) != 0)
		Socket::throwBindException();

	m_Bound = true;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Socket::bind(const Address& localAdr) {
	sockaddr_in adr;

	m_LocalAddress = localAdr;
	m_LocalPort = 0;

	adr.sin_family = AF_INET;
	adr.sin_port = 0;
	adr.sin_addr.s_addr = m_LocalAddress.getAddressNO();

	if (::bind(m_Socket, (sockaddr*)&adr, sizeof(adr)) != 0)
		Socket::throwBindException();

	m_Bound = true;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Socket::bind(const char* localAdr) {
	sockaddr_in adr;

	m_LocalAddress.setHostname(localAdr);
	m_LocalPort = 0;

	adr.sin_family = AF_INET;
	adr.sin_port = 0;
	adr.sin_addr.s_addr = m_LocalAddress.getAddressNO();

	if (::bind(m_Socket, (sockaddr*)&adr, sizeof(adr)) != 0)
		Socket::throwBindException();

	m_Bound = true;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Socket::bind(unsigned int localPort) {
	sockaddr_in adr;

	m_LocalPort = htons((unsigned short)localPort);
	m_LocalAddress.setAddress(0);

	adr.sin_family = AF_INET;
	adr.sin_port = m_LocalPort;
	adr.sin_addr.s_addr = INADDR_ANY;

	if (::bind(m_Socket, (sockaddr*)&adr, sizeof(adr)) != 0)
		Socket::throwBindException();

	m_Bound = true;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Socket::wait(Socket::WaitMode mode) {
	fd_set set;

#if defined(__FORWINDOWS__)
	set.fd_count = 1;
	set.fd_array[0] = m_Socket;
#else
	FD_ZERO(&set);
	FD_SET(m_Socket, &set);
#endif
	if (select(m_SelectIndex, (mode & WaitRead) ? &set : NULL, (mode & WaitWrite) ? &set : NULL, NULL, NULL) == SOCKET_ERROR)
		Socket::throwSocketException();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Socket::wait(Socket::WaitMode mode, unsigned int millis) {
	int ret;
	fd_set set;
	timeval time;

	time.tv_usec = 1000*millis % 1000000;
	time.tv_sec = millis / 1000;

#if defined(__FORWINDOWS__)
	set.fd_count = 1;
	set.fd_array[0] = m_Socket;
#else
	FD_ZERO(&set);
	FD_SET(m_Socket, &set);
#endif
	ret = select(m_SelectIndex, (mode & WaitRead) ? &set : NULL, (mode & WaitWrite) ? &set : NULL, NULL, &time);
	if (ret == SOCKET_ERROR)
		Socket::throwSocketException();

	return (ret > 0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Socket::connect(const Address& remoteAdr, unsigned int port) {
	int size;
	sockaddr_in adr;

	// Connect socket to remote address
	m_RemoteAddress = remoteAdr;
	m_RemotePort = htons((unsigned short)port);

	adr.sin_family = AF_INET;
	adr.sin_port = m_RemotePort;
	adr.sin_addr.s_addr = m_RemoteAddress.getAddressNO();

#if defined(__FORWINDOWS__)
	if (WSAConnect(m_Socket, (sockaddr *)&adr, sizeof(adr), NULL, NULL, NULL, NULL) != 0) {
		throwSocketException();
	}
#else
	if (::connect(m_Socket, (sockaddr *)&adr, sizeof(adr)) != 0) {
		throwSocketException();
	}
#endif

	// Get socket information...
	size = sizeof(sockaddr_in);
	getsockname(m_Socket, (sockaddr *)&adr, &size);
	if (!m_LocalAddress.isValid()) {
		m_LocalAddress.setAddressNO(adr.sin_addr.s_addr);
	} else {
		if (m_LocalAddress.getAddressNO() != adr.sin_addr.s_addr)
			m_LocalAddress.setAddressNO(adr.sin_addr.s_addr);
	}
	m_LocalPort = adr.sin_port;

	m_Connected = true;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Socket::disconnect() {
	m_RemoteAddress.setAddress(0);
	m_RemotePort = 0;
	shutdown(m_Socket, SHUT_RDWR);
	m_Connected = false;
	return true;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Socket::shutdownInput() {
	shutdown(m_Socket, SHUT_RD);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Socket::shutdownOutput() {
	shutdown(m_Socket, SHUT_WR);
}

}; };
