/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "geometry.alignedbox.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.appearance.h"
#include "graphics3d.context.h"
#include "graphics3d.plainshader.h"
#include "plots3d.exception.h"
#include "plots3d.lineplot3d.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "base.string.inl"
#include "math.vector.inl"
#include "math.vector3.inl"
#include "math.matrix.inl"
#include "geometry.alignedbox.inl"
#include "geometry.bounds.inl"
#include "container.array.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.context.inl"
#include "graphics3d.matrixstack.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.fixedshader.inl"
#include "plots3d.plot.inl"
#include "plots3d.plot3d.inl"
#include "plots3d.lineplot3d.inl"

namespace magnum { namespace plots3d {
using magnum::object::AutoPointer;
using magnum::graphics3d::Context;
using magnum::graphics3d::MatrixStack;
using magnum::graphics3d::VertexFormat;
using magnum::graphics3d::Types;
using magnum::graphics3d::Mesh;
using magnum::graphics3d::Shader;
using magnum::graphics3d::PlainShader;
using magnum::graphics3d::FixedShader;
using magnum::graphics3d::Attributes;
using magnum::geometry::AlignedBox3;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
LinePlot3d::LinePlot3d() {
	m_Resolution = 100;
	m_Mesh.setPrimitive(Mesh::LineStrip);
	AutoPointer<Shader> shader = new PlainShader(FixedShader::Plain);
	getAppearance()->attachShader(shader);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void LinePlot3d::onClear() {
	m_Mesh.detachVertexBuffer();
	m_Vertices.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void LinePlot3d::onRender(const Matrix4<float>& mtx) const {
	if (m_Mesh.getVertexBuffer() == NULL)
		return;
		
	Context* c = getContext();
	ASSERT(c != NULL);
	c->getMatrixStack()->setModel(mtx);
	
	if (getAppearance() != NULL)
		getAppearance()->select();
	m_Mesh.render();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void LinePlot3d::beginPlot() {
	// Check if resolution is acceptable
	if (m_Resolution < 2)
		throw InvalidOperationException();

	// Check if current vertex buffer is acceptable
	if (m_Mesh.getVertexBuffer() != NULL) {
		if (m_OldResolution == m_Resolution)
			return;
	}

	// Allocate space for temp vertices and normals
	m_Vertices.setSize(m_Resolution);

	// Create vertex format
	VertexFormat fmt;
	fmt.insert(Types::Vector3f, Attributes::Vertex);
	
	// Allocate buffer
	AutoPointer<VertexBuffer> vb = VertexBuffer::create(fmt, m_Resolution, VertexBuffer::Static);
	if (vb == NULL)
		throw MemoryException();

	m_Mesh.attachVertexBuffer(vb);
	m_OldResolution = m_Resolution;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void LinePlot3d::finishPlot() {
	unsigned int count = m_Resolution;

	// Collect min/max info
	Vertex* vertex = &m_Vertices[0];
	Vertex* vend = vertex + count;
	m_Minimum.x = m_Maximum.x = m_Vertices[0].position.x;
	m_Minimum.y = m_Maximum.y = m_Vertices[0].position.y;
	m_Minimum.z = m_Maximum.z = m_Vertices[0].position.z;
	while (vertex < vend) {
		if (vertex->position.x <= m_Minimum.x)
			m_Minimum.x = vertex->position.x;
		else if (vertex->position.x >= m_Maximum.x)
			m_Maximum.x = vertex->position.x;
		if (vertex->position.y <= m_Minimum.y)
			m_Minimum.y = vertex->position.y;
		else if (vertex->position.y >= m_Maximum.y)
			m_Maximum.y = vertex->position.y;
		if (vertex->position.z <= m_Minimum.z)
			m_Minimum.z = vertex->position.z;
		else if (vertex->position.z >= m_Maximum.z)
			m_Maximum.z = vertex->position.z;
		vertex++;
	}
	
	publishBounds(Bounds3<float>(AlignedBox3<float>(m_Minimum,m_Maximum)));
	
	// Copy vertices and normals into vertex buffer
	m_Mesh.getVertexBuffer()->setBuffer(0, &m_Vertices[0], count);
}

}; };

