/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "object.autopointer.h"
#include "geometry.alignedbox.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.appearance.h"
#include "graphics3d.context.h"
#include "plots3d.exception.h"
#include "plots3d.surfaceplot3d.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "base.string.inl"
#include "math.vector.inl"
#include "math.vector2.inl"
#include "math.vector3.inl"
#include "math.matrix.inl"
#include "geometry.alignedbox.inl"
#include "geometry.bounds.inl"
#include "container.array.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.context.inl"
#include "graphics3d.matrixstack.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.vertexformat.inl"
#include "plots3d.plot.inl"
#include "plots3d.plot3d.inl"
#include "plots3d.surfaceplot3d.inl"

namespace magnum { namespace plots3d {
using magnum::object::AutoPointer;
using magnum::graphics3d::Context;
using magnum::graphics3d::MatrixStack;
using magnum::graphics3d::VertexFormat;
using magnum::graphics3d::Types;
using magnum::graphics3d::Mesh;
using magnum::graphics3d::Attributes;
using magnum::geometry::AlignedBox3;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SurfacePlot3d::SurfacePlot3d() {
	m_Resolution.x = 100;
	m_Resolution.y = 100;
	m_Mesh.setPrimitive(Mesh::TriangleStrip);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void SurfacePlot3d::onClear() {
	m_Mesh.detachIndexBuffer();
	m_Mesh.detachVertexBuffer();
	m_Vertices.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void SurfacePlot3d::onRender(const Matrix4<float>& mtx) const {
	if ((m_Mesh.getIndexBuffer() == NULL) || (m_Mesh.getVertexBuffer() == NULL))
		return;
		
	Context* c = getContext();
	ASSERT(c != NULL);
	c->getMatrixStack()->setModel(mtx);
	
	if (getAppearance() != NULL)
		getAppearance()->select();
	m_Mesh.render();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void SurfacePlot3d::beginPlot() {
	unsigned int width = m_Resolution.x;
	unsigned int height = m_Resolution.y;
	
	// Check if resolution is acceptable
	if ((width < 2) || (height < 2))
		throw InvalidOperationException();

	// Check if current vertex buffer is acceptable
	if ((m_Mesh.getVertexBuffer() != NULL) && (m_Mesh.getIndexBuffer() != NULL)) {
		if (m_OldResolution == m_Resolution)
			return;
	}

	// Allocate space for temp vertices and normals
	m_Vertices.setSize(width * height);

	// Create vertex format
	VertexFormat fmt;
	fmt.insert(Types::Vector3f, Attributes::Vertex);
	fmt.insert(Types::Vector3f, Attributes::Normal);
	
	// Allocate buffer
	AutoPointer<VertexBuffer> vb = VertexBuffer::create(fmt, width * height, VertexBuffer::Static);
	if (vb == NULL)
		throw MemoryException();

	// Create index buffer	
	unsigned int size = 2*(height-1)*width + (height-2)*2;
	AutoPointer<IndexBuffer> ib = IndexBuffer::create(IndexBuffer::Index32, size, IndexBuffer::Static);
	if (ib == NULL)
		throw MemoryException();
		
	// Create a grid-like triangle-strip
	unsigned int* buffer = (unsigned int*)ib->lock(IndexBuffer::Discard);
	for (unsigned int i = 0; i < height-1; i++) {
		unsigned int idx = i*width;
		for (unsigned int j = 0; j < width; j++) {
			*buffer++ = idx;
			*buffer++ = idx+width;
			idx++;
		}
		if (i < height-2) {
			*buffer++ = idx+width-1;
			*buffer++ = idx;
		}
	}
	ib->unlock();
	
	m_Mesh.attachIndexBuffer(ib);
	m_Mesh.attachVertexBuffer(vb);
	m_OldResolution = m_Resolution;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void SurfacePlot3d::finishPlot() {
	unsigned int width = m_Resolution.x;
	unsigned int height = m_Resolution.y;
	unsigned int count = m_Resolution.x * m_Resolution.y;

	// Loop over all triangles
	for (unsigned int i = 0; i < height-1; i++) {
		unsigned int idx = i*width;
		for (unsigned int j = 0; j < width-1; j++) {
			Vector3<float> normal;
			
			unsigned int i0 = idx;
			unsigned int i1 = idx + width;
			unsigned int i2 = idx + 1;
			normal = (m_Vertices[i1].position - m_Vertices[i0].position) % (m_Vertices[i2].position - m_Vertices[i0].position);
			m_Vertices[i0].normal += normal;
			m_Vertices[i1].normal += normal;
			m_Vertices[i2].normal += normal;

			unsigned int j0 = idx + width;
			unsigned int j1 = idx + width + 1;
			unsigned int j2 = idx + 1;
			normal = (m_Vertices[j1].position - m_Vertices[j0].position) % (m_Vertices[j2].position - m_Vertices[j0].position);
			m_Vertices[j0].normal += normal;
			m_Vertices[j1].normal += normal;
			m_Vertices[j2].normal += normal;

			idx++;
		}
	}
	
	// Renormalise all normals and collect min/max info
	Vertex* vertex = &m_Vertices[0];
	Vertex* vend = vertex + count;
	m_Minimum.x = m_Maximum.x = m_Vertices[0].position.x;
	m_Minimum.y = m_Maximum.y = m_Vertices[0].position.y;
	m_Minimum.z = m_Maximum.z = m_Vertices[0].position.z;
	while (vertex < vend) {
		if (vertex->position.x <= m_Minimum.x)
			m_Minimum.x = vertex->position.x;
		else if (vertex->position.x >= m_Maximum.x)
			m_Maximum.x = vertex->position.x;
		if (vertex->position.y <= m_Minimum.y)
			m_Minimum.y = vertex->position.y;
		else if (vertex->position.y >= m_Maximum.y)
			m_Maximum.y = vertex->position.y;
		if (vertex->position.z <= m_Minimum.z)
			m_Minimum.z = vertex->position.z;
		else if (vertex->position.z >= m_Maximum.z)
			m_Maximum.z = vertex->position.z;
		vertex->normal.normalise();
		vertex++;
	}
	
	publishBounds(Bounds3<float>(AlignedBox3<float>(m_Minimum,m_Maximum)));
	
	// Copy vertices and normals into vertex buffer
	m_Mesh.getVertexBuffer()->setBuffer(0, &m_Vertices[0], count);
}

}; };

