/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "scene3d.modelnode.h"
#include "scene3d.modifier.h"
#include "scene3d.locale.h"
#include "scene3d.audio.h"
#include "scene3d.behaviour.h"
#include "scene3d.graphicscontext.h"
#include "scene3d.universe.h"
#include "scene3d.intersection.h"
#include "graphics3d.parameter.h"
#include "graphics3d.parameterbuffer.h"
#include "graphics3d.parameterformat.h"
#include "graphics3d.types.h"
#include "graphics3d.camera.h"
#include "graphics3d.appearance.h"
#include "graphics3d.context.h"
#include "graphics3d.shader.h"
#include "graphics3d.texture.h"
#include "graphics3d.light.h"
#include "graphics3d.material.h"
#include "graphics3d.framebuffer.h"

#include "object.pointer.inl"
#include "container.iterator.inl"
#include "container.array.inl"
#include "container.allocator.inl"
#include "math.vector.inl"
#include "math.matrix.inl"
#include "geometry.space.inl"
#include "geometry.sphere.inl"
#include "geometry.bounds.inl"
#include "geometry.transform.inl"
#include "graphics3d.parameterbuffer.inl"
#include "graphics3d.parameterformat.inl"
#include "graphics3d.model.inl"
#include "graphics3d.accelerator.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.light.inl"
#include "graphics3d.material.inl"
#include "graphics3d.context.inl"
#include "graphics3d.resource.inl"
#include "scene3d.intersection.inl"
#include "scene3d.modelnode.inl"
#include "scene3d.leaf.inl"
#include "scene3d.node.inl"
#include "scene3d.locale.inl"
#include "scene3d.graphics.inl"
#include "scene3d.geometry.inl"
#include "scene3d.modifier.inl"
#include "scene3d.universe.inl"
#include "scene3d.graphicscontext.inl"


namespace magnum {namespace scene3d {
using magnum::graphics3d::Parameters;
using magnum::graphics3d::ParameterBuffer;
using magnum::graphics3d::ParameterFormat;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ModelNode::ModelNode() {
	m_Parameters = NULL;
	m_ModelMatrix = NULL;
	m_ViewMatrix = NULL;
	m_ModelViewMatrix = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ModelNode::ModelNode(const ModelNode& m) : Geometry(m) {
	m_Parameters = NULL;
	if (m.m_Model != NULL) {
		attachModel(m.m_Model);
	} else {
		m_ModelMatrix = NULL;
		m_ViewMatrix = NULL;
		m_ModelViewMatrix = NULL;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ModelNode::ModelNode(Model* m) {
	ASSERT(m != NULL);
	m_Parameters = NULL;
	attachModel(m);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ModelNode::~ModelNode() {
	if (m_Parameters != NULL)
		m_Parameters->release();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void ModelNode::attachModel(Model* m) {
	// Initialize some member vars
	if (m_Parameters != NULL)
		m_Parameters->release();
	m_Parameters = NULL;
	m_ModelMatrix = NULL;
	m_ViewMatrix = NULL;
	m_ModelViewMatrix = NULL;
	
	m_Model = m;
	setBounds(m_Model->getBounds());
	enqueueTransform();

	// Try to get the parameter format
	const ParameterFormat* fmt = &m_Model->getParameters();
	if (fmt == NULL)
		return;

	// Create a parameter buffer
	m_Parameters = ParameterBuffer::create(*fmt);

	// Get Pointer to transformation matrix
	ParameterFormat::Iterator iter;
	iter = fmt->at(Parameters::ModelView);
	if (iter != fmt->end())
		m_ModelViewMatrix = (Matrix4<float>*)m_Parameters->getPointer(iter->getLocation());
	iter = fmt->at(Parameters::Model);
	if (iter != fmt->end())
		m_ModelMatrix = (Matrix4<float>*)m_Parameters->getPointer(iter->getLocation());
	iter = fmt->at(Parameters::View);
	if (iter != fmt->end())
		m_ViewMatrix = (Matrix4<float>*)m_Parameters->getPointer(iter->getLocation());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void ModelNode::detachModel() {
	if (m_Parameters != NULL)
		m_Parameters->release();
	m_Model = NULL;
	m_Parameters = NULL;
	m_ModelMatrix = NULL;
	m_ViewMatrix = NULL;
	m_ModelViewMatrix = NULL;
	enqueueTransform();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
ModelNode* ModelNode::onClone() const {
	return new ModelNode(*this);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void ModelNode::onRender(GraphicsContext* context,const Transform3<float>& camera) const {
	ASSERT(m_Model != NULL);
	if (m_ModelViewMatrix != NULL)
		*m_ModelViewMatrix = (Matrix4<float>)(camera * getWorldTransform());
	if (m_ModelMatrix != NULL)
		*m_ModelMatrix = (Matrix4<float>)(getWorldTransform());
	if (m_ViewMatrix != NULL)
		*m_ViewMatrix = (Matrix4<float>)(camera);

	m_Model->render(m_Parameters);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool ModelNode::onIntersect(const Ray<float,3>& ray,Intersection* is) const {
	if (m_Model == NULL)
		return false;
	if (m_ModelViewMatrix != NULL)
		*m_ModelViewMatrix = (Matrix4<float>)(getWorldTransform());
	if (m_ModelMatrix != NULL)
		*m_ModelMatrix = (Matrix4<float>)(getWorldTransform());
	if (m_ViewMatrix != NULL)
		m_ViewMatrix->identity();
		
	if (m_Model->intersect(m_Parameters, ray, is)) {
		is->setNode(this);
		return true;
	}
	
	return false;
}

}; };
