/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "scene3d.node.h"
#include "scene3d.group.h"
#include "scene3d.root.h"
#include "scene3d.locale.h"
#include "scene3d.universe.h"

#include "object.inl"
#include "object.attachment.inl"
#include "base.string.inl"
#include "scene3d.node.inl"
#include "scene3d.locale.inl"

#include "typeinfo.h"

namespace magnum { namespace scene3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Node::Node(Node::Type t) {
	m_Parent = NULL;
	m_Next = m_Prev = NULL;
	m_NextDirty = m_PrevDirty = NULL;
	m_Type = t;
	m_Locale = NULL;
	m_Flags = Enabled | AutoBounds;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Node::Node(const Node& n) : m_Name(n.m_Name) {
	m_Parent = NULL;
	m_Next = m_Prev = NULL;
	m_NextDirty = m_PrevDirty = NULL;
	m_Type = n.m_Type;
	m_Locale = NULL;
	m_Flags = Enabled | AutoBounds;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Node::~Node() {
	ASSERT(m_Parent == NULL);
	ASSERT(m_Next == NULL);
	ASSERT(m_Prev == NULL);
	dequeueTransform();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::enable() {
	if (m_Flags & Enabled)
		return;
	m_Flags |= Enabled;
	onEnable();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::disable() {
	if ((m_Flags & Enabled) == 0)
		return;
	m_Flags &= ~Enabled;
	onDisable();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::enqueueTransform() {
	// Set dirty flag
	m_Flags |= DirtyTransform;

	// m_PrevDirty is the indicator wether the node is already queued
	// for updating
	if (m_PrevDirty != NULL)
		return;

	// No universe means that we do not have acces to a dirty queue
	if (m_Locale == NULL)
		return;

	// No parent - no updating needed (?)
	if (m_Parent == NULL)
		return;
	
	Universe* universe = m_Locale->getUniverse();
	ASSERT(universe != NULL);
		
	Node* anchor;

	// Get suitable node to insert after
	if (((Node*)m_Parent)->m_PrevDirty != NULL)
		anchor = m_Parent;
	else
		anchor = &universe->m_DirtyNodes;

	// Link this node behind anchor
	m_NextDirty = anchor->m_NextDirty;
	m_PrevDirty = anchor;
	anchor->m_NextDirty = this;

	// Enqueue all children
	if (m_Type & GroupType) {
		Node* node = ((Group*)this)->m_First;
		while(node != NULL) {
			node->m_Flags |= DirtyTransform;
			if (!node->m_PrevDirty)
				node->enqueueTransform();
			node = node->m_Next;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::dequeueTransform() {
	// Is this node really enqueued?
	if (m_PrevDirty == NULL)
		return;

	// Parentless nodes won't be enqueued
	ASSERT(m_Parent != NULL);

	// Check if parent is enqueued...
	if (((Node*)m_Parent)->m_PrevDirty != NULL) {
		m_Flags |= DirtyTransform;
		return;
	}		

	m_Flags &= ~DirtyTransform;

	m_PrevDirty->m_NextDirty = m_NextDirty;
	if (m_NextDirty != NULL)
		m_NextDirty->m_PrevDirty = m_PrevDirty;

	m_NextDirty = m_PrevDirty = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Node* Node::getRoot() const {
	const Node* node = this;
	while (node->m_Parent != NULL) {
		node = node->m_Parent;
	}

	return (Node*)node;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Node* Node::onClone() const {
	return new Node(*this);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onAttachLocale(Locale* l) {
	ASSERT(l != NULL);
	ASSERT(m_Locale == NULL);
	m_Locale = l;
	if (m_Parent != NULL) {
		if ((m_Parent->getType() & TransformType) || m_Parent->testFlag(ParentTransform))
			setFlag(ParentTransform);
		enqueueTransform();
	} else {
		ASSERT(typeid(*this) == typeid(Root));	
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onDetachLocale(Locale* l) {
	ASSERT(l != NULL);
	ASSERT(m_Locale == l);
	clearFlag(ParentTransform);
	dequeueTransform();
	m_Locale = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onEnable() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onDisable() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onTransform(const Transform3<float>& t) {
}

}; };
