/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/


#include "system.assembler.h"
#include "system.program.h"

#include "object.inl"
#include "base.string.inl"
#include "system.assembler.inl"
#include "system.program.inl"
#include "container.array.inl"

namespace magnum {namespace system {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Assembler::Assembler() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Assembler::~Assembler() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::clear() {
	m_Commands.clear();
	m_Labels.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::finish() {
	m_Commands.shrink();
	m_Labels.shrink();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Program* Assembler::compile() {
	Program* prg = new Program();
	Command* cmd;

	unsigned int i,j;
	unsigned int param[5];
	unsigned int pcount = 0;
	void* obj;
	for (i = 0; i < m_Commands.getSize(); i++) {
		cmd = &m_Commands[i];
		switch(cmd->m_Opcode) {
			case opThisCall:
				prg->addByte(0xb8);
				prg->addDword((dword)(intptr_t)cmd->m_Pointer[0]);
				prg->addByte(0xb9);
				prg->addDword((dword)(intptr_t)obj);
				while(pcount > 0) {
					pcount--;
					prg->addByte(0x68);
					prg->addDword(param[pcount]);
				};
				prg->addByte(0xff);
				prg->addByte(0xd0);
				break;
			case opStdCall:
				prg->addByte(0xb8);
				prg->addDword((dword)(intptr_t)cmd->m_Pointer[0]);
				while(pcount > 0) {
					pcount--;
					prg->addByte(0x68);
					prg->addDword(param[pcount]);
				};
				prg->addByte(0xff);
				prg->addByte(0xd0);
				break;
			case opCCall:
				prg->addByte(0xb8);
				prg->addDword((dword)(intptr_t)cmd->m_Pointer[0]);
				j = pcount;
				while(pcount > 0) {
					pcount--;
					prg->addByte(0x68);
					prg->addDword(param[pcount]);
				};
				prg->addByte(0xff);
				prg->addByte(0xd0);
				if (j > 0) {
					prg->addByte(0x83);
					prg->addByte(0xc4);
					prg->addByte(j*4);
				}
				break;
			case opReturn:
				prg->addByte(0xc3);
				break;
			case opParameter:
				param[pcount] = cmd->m_Integer[0];
				pcount++;
				break;
			case opThis:
				obj = cmd->m_Pointer[0];
				break;
			case opPush:
				prg->addByte(0x50 + cmd->m_Integer[0]);
				break;
			case opPop:
				prg->addByte(0x58 + cmd->m_Integer[0]);
				break;
			case opMove:
				prg->addByte(0x89);
				prg->addByte(0xc0 + cmd->m_Integer[0] + (cmd->m_Integer[1] << 3));
				break;
			case opLoadImm:
				prg->addByte(0xb8 + cmd->m_Integer[0]);
				prg->addDword(cmd->m_Integer[1]);
				break;
			case opLoadPtr:
				prg->addByte(0x8b);
				prg->addByte(5 + (cmd->m_Integer[0] << 3));
				prg->addDword(cmd->m_Integer[1]);
				break;
			case opLoadInd:
				prg->addByte(0x8b);
				prg->addByte((cmd->m_Integer[0] << 3) + cmd->m_Integer[1]);
				break;
			case opLoadIndOff:
				prg->addByte(0x8b);
				prg->addByte(0x80 + (cmd->m_Integer[0] << 3) + cmd->m_Integer[1]);
				prg->addDword(cmd->m_Integer[2]);
				break;
			case opStorePtr:
				prg->addByte(0x89);
				prg->addByte(5 + (cmd->m_Integer[1] << 3));
				prg->addDword(cmd->m_Integer[0]);
				break;
			case opStoreInd:
				prg->addByte(0x89);
				prg->addByte((cmd->m_Integer[1] << 3) + cmd->m_Integer[0]);
				break;
			case opStoreIndOff:
				prg->addByte(0x89);
				prg->addByte(0x80 + (cmd->m_Integer[2] << 3) + cmd->m_Integer[0]);
				prg->addDword(cmd->m_Integer[1]);
				break;
			case opLabel:
				prg->addLabel(m_Labels[cmd->m_Integer[0]]);
				break;

		}
	}

	prg->finish();

	return prg;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmCall(void (__stdcall *adr)()) {
	Command cmd;
	cmd.m_Opcode = opStdCall;
	cmd.m_Pointer[0] = adr;
	m_Commands.insert(cmd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmCall(void (__cdecl *adr)()) {
	Command cmd;
	cmd.m_Opcode = opCCall;
	cmd.m_Pointer[0] = adr;
	m_Commands.insert(cmd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmReturn() {
	Command cmd;
	cmd.m_Opcode = opReturn;
	m_Commands.insert(cmd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmPush(Assembler::Register reg) {
	Command cmd;
	cmd.m_Opcode = opPush;
	cmd.m_Integer[0] = (int)reg;
	m_Commands.insert(cmd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmPop(Assembler::Register reg) {
	Command cmd;
	cmd.m_Opcode = opPop;
	cmd.m_Integer[0] = (int)reg;
	m_Commands.insert(cmd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmLoad(Assembler::Register reg, const void* adr) {
	Command cmd;
	cmd.m_Opcode = opLoadPtr;
	cmd.m_Integer[0] = (int)reg;
	cmd.m_Pointer[1] = (void *)adr;
	m_Commands.insert(cmd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmLoad(Assembler::Register reg, int data) {
	Command cmd;
	cmd.m_Opcode = opLoadImm;
	cmd.m_Integer[0] = (int)reg;
	cmd.m_Integer[1] = (int)data;
	m_Commands.insert(cmd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmLoad(Assembler::Register reg, Assembler::Register src) {
	Command cmd;
	cmd.m_Opcode = opLoadInd;
	cmd.m_Integer[0] = (int)reg;
	cmd.m_Integer[1] = (int)src;
	m_Commands.insert(cmd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmLoad(Assembler::Register reg, Assembler::Register src, int off) {
	Command cmd;
	cmd.m_Opcode = opLoadIndOff;
	cmd.m_Integer[0] = (int)reg;
	cmd.m_Integer[1] = (int)src;
	cmd.m_Integer[2] = off;
	m_Commands.insert(cmd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmMove(Assembler::Register reg, Register reg2) {
	Command cmd;
	cmd.m_Opcode = opMove;
	cmd.m_Integer[0] = (int)reg;
	cmd.m_Integer[1] = (int)reg2;
	m_Commands.insert(cmd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmStore(void* adr, Assembler::Register reg) {
	Command cmd;
	cmd.m_Opcode = opStorePtr;
	cmd.m_Pointer[0] = adr;
	cmd.m_Integer[1] = (int)reg;
	m_Commands.insert(cmd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmStore(Assembler::Register tg, Assembler::Register reg) {
	Command cmd;
	cmd.m_Opcode = opStoreInd;
	cmd.m_Integer[0] = (int)tg;
	cmd.m_Integer[1] = (int)reg;
	m_Commands.insert(cmd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmStore(Assembler::Register tg, int off, Assembler::Register reg) {
	Command cmd;
	cmd.m_Opcode = opStoreIndOff;
	cmd.m_Integer[0] = (int)tg;
	cmd.m_Integer[1] = (int)off;
	cmd.m_Integer[2] = reg;
	m_Commands.insert(cmd);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Assembler::asmLabel(const char* label) {
	Command cmd;
	cmd.m_Opcode = opLabel;
	cmd.m_Integer[0] = m_Labels.getSize();
	m_Commands.insert(cmd);
	m_Labels.insert(String(label));
}

}; };
