/* 
  Magnum 
  Copyright (C) 2002-2006 Kaya Kupferschmidt. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Kupferschmidt  (k.kupferschmidt@dimajix.de)
*/

#include "system.thread.h"
#include "object.class.h"
#include "object.exception.h"

#include "base.string.inl"
#include "object.class.inl"
#include "object.inl"
#include "object.exception.inl"
#include "system.thread.inl"
#include "system.mutex.inl"
#include "system.signal.inl"
#include "container.chain.inl"
#include "container.iterator.inl"

#include <stdlib.h>
#include <stdio.h>

namespace magnum { namespace system {
using namespace magnum::container;
using magnum::object::Exception;

Chain<Thread> Thread::g_Threads;
Mutex Thread::g_Mutex;
Thread::Storage Thread::g_Current;
Thread::Startup Thread::g_Startup;
#if defined(__FORWINDOWS__)
FARPROC Thread::g_SwitchToThread;
#endif


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Thread::Startup::Startup() {
	g_Current = createTLS();
	g_SwitchToThread = GetProcAddress(GetModuleHandle("kernel32"), "SwitchToThread");
	createCurrent();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
#if defined(__FORWINDOWS__)
DWORD WINAPI Thread::ThreadProc(LPVOID param) {
#else
void* Thread::ThreadProc(void* param) {
#endif
	Thread* thread = (Thread *)param;

#if defined(__FORWINDOWS__)
	MSG msg;
	PeekMessage(&msg, NULL, WM_USER, WM_USER, PM_NOREMOVE);
#endif

	// Set Thread pointer into TLS
	setTLS(g_Current, thread);

	// Raise signal 
	thread->m_Signal.raise(false);

	try {
		thread->run();
	} catch (Exception ex) {
		printf(ex.toString());
		::exit(-1);	
	} catch (...) {
		::exit(-1);
	}

	// Reset signal
	thread->m_Signal.reset();

#if defined(__FORWINDOWS__)
	thread->m_ThreadId = 0;
	thread->m_ThreadHandle = NULL;
#elif defined(__FORUNIX__)
	memset(&thread->m_ThreadId, 0, sizeof(thread->m_ThreadId));
#endif

	g_Mutex.lock();
	g_Threads.remove(thread);
	g_Mutex.unlock();
	thread->release();

	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Thread::Thread(bool (*fn)()) {
	m_Function = fn;
	m_Function2 = NULL;
	m_Function2Data = NULL;
#if defined(__FORWINDOWS__)
	m_ThreadId = 0;
	m_ThreadHandle = NULL;
#elif defined(__FORUNIX__)
	memset(&m_ThreadId, 0, sizeof(m_ThreadId));
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Thread::Thread(bool (*fn)(void* ),void* data) {
	m_Function = NULL;
	m_Function2 = fn;
	m_Function2Data = data;
#if defined(__FORWINDOWS__)
	m_ThreadId = 0;
	m_ThreadHandle = NULL;
#elif defined(__FORUNIX__)
	memset(&m_ThreadId, 0, sizeof(m_ThreadId));
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Thread::~Thread() {

}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Thread::run() {
	if (m_Function != NULL)
		return m_Function();
	else if (m_Function2 != NULL)
		return m_Function2(m_Function2Data);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Thread::start(bool sync) {
#if defined(__FORWINDOWS__)
	if (m_ThreadHandle != NULL)
		return true;
#elif defined(__FORUNIX__)
	if (m_ThreadId != 0)
		return true;
#endif

	refer();
	g_Mutex.lock();
	g_Threads.insert(this);
	g_Mutex.unlock();
	m_Signal.reset();

#if defined(__FORWINDOWS__)
	m_ThreadHandle = CreateThread(NULL,0,ThreadProc,(LPVOID)this, CREATE_SUSPENDED, &m_ThreadId);
	ResumeThread(m_ThreadHandle);
#elif defined(__FORUNIX__)
	pthread_create(&m_ThreadId, NULL, ThreadProc, (void *)this); 
#endif

	if (sync) {
		m_Signal.wait();
	}

	return true;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Thread::join() const {
	if (getCurrent() == this)
		return false;
	if (!isAlive())
		return false;

#if defined(__FORWINDOWS__)
	return (WaitForSingleObject(m_ThreadHandle, INFINITE) == WAIT_OBJECT_0);
#elif defined(__FORUNIX__)
	return (pthread_join(m_ThreadId, NULL) == 0);
#else
	return false;
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Thread::join(unsigned int millis) const {
	if (getCurrent() == this)
		return false;
	if (!isAlive())
		return false;

#if defined(__FORWINDOWS__)
	return (WaitForSingleObject(m_ThreadHandle, millis) == WAIT_OBJECT_0);
#elif defined(__FORUNIX__)
	return (pthread_join(m_ThreadId, NULL) == 0);
#else
	return false;
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Thread::suspend() {
#if defined(__FORWINDOWS__)
	SuspendThread(m_ThreadHandle);
	return true;
#else
	return false;
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Thread::resume() {
#if defined(__FORWINDOWS__)
	ResumeThread(m_ThreadHandle);
	return true;
#else
	return false;
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Thread::sleep(unsigned int millis) {
#if defined(__FORWINDOWS__)
	Sleep(millis);
	return true;
#else
	::sleep(millis / 1000);
	return false;
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Thread::exit() {
	Thread *thread = getCurrent();

#if defined(__FORWINDOWS__)
	thread->m_ThreadHandle = NULL;
	thread->m_ThreadId = 0;
#elif defined(__FORUNIX__)
	memset(&thread->m_ThreadId, 0, sizeof(thread->m_ThreadId)); 
#endif

	g_Mutex.lock();
	g_Threads.remove(thread);
	g_Mutex.unlock();
	thread->release();

#if defined(__FORWINDOWS__)
	ExitThread(0);
#elif defined(__FORUNIX__)
	pthread_exit(NULL);
#endif
}


/*--------------------------------------------------------------------------*/
/** Causes the calling thread to yield execution to another thread that is 
 * ready to run on the current processor. The operating system selects the 
 * thread to yield to.
 */
void Thread::cycle() {
#if defined(__FORWINDOWS__)
	if (g_SwitchToThread)
		g_SwitchToThread();
	else
		Sleep(0);
#else
	::sleep(1);
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Thread::Storage Thread::createTLS() {
#if defined(__FORWINDOWS__)
	Storage s = TlsAlloc();
	if (s == 0xFFFFFFFF)
		return StorageNone;
	else
		return s;
#elif defined(__FORUNIX__)
	Storage s;
	int ret = pthread_key_create(&s, NULL);
	if (ret != 0)
		return StorageNone;
	else
		return s;
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Thread::releaseTLS(Thread::Storage s) {
#if defined(__FORWINDOWS__)
	TlsFree(s);
#elif defined(__FORUNIX__)
	pthread_key_delete(s);
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Thread* Thread::searchCurrent() {
	Chain<Thread>::Iterator iter = g_Threads.begin();
	Chain<Thread>::Iterator end = g_Threads.end();

#if defined(__FORWINDOWS__)
	DWORD id = GetCurrentThreadId();
#elif defined(__FORUNIX__)
	pthread_t id = pthread_self(); 
#else
	int id = 0;
#endif

	// Search the thread
	g_Mutex.lock();
	while (iter != end) {
#if defined(__FORUNIX__)
		if (pthread_equal((iter)->m_ThreadId, id) == TRUE)
			break;
#else
		if ((iter)->m_ThreadId == id)
			break;
#endif
		iter++;
	}
	g_Mutex.unlock();

	// Didn't we find the thread?
	if (iter == end) {
		return createCurrent();
	}

	setTLS(g_Current, &(*iter));
	return &(*iter);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Thread* Thread::createCurrent() {
	// Create a new Thread object for the current thread...
	Thread *thread = new Thread();

#if defined(__FORWINDOWS__)	
	HANDLE handle = GetCurrentThread();
	HANDLE phandle = GetCurrentProcess();
	HANDLE thandle;
	DuplicateHandle(phandle, handle, phandle, &thandle, NULL, FALSE,DUPLICATE_SAME_ACCESS);
	thread->m_ThreadHandle = thandle;
	thread->m_ThreadId = GetCurrentThreadId();
#elif defined(__FORUNIX__)
	thread->m_ThreadId = pthread_self();
#endif

	g_Mutex.lock();
	g_Threads.insert(thread);
	g_Mutex.unlock();

	setTLS(g_Current, thread);

	return thread;
}

}; };
