/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "magnum.stats.h"
#include "graphics.h"
#include "gui.h"

#include "graphics.inl"
#include "gui.inl"



namespace magnum { namespace sample {
using namespace magnum::types;
using magnum::graphics::Colour;
using magnum::graphics3d::BlendMode;
using magnum::graphics3d::DepthMode;
using magnum::graphics3d::DirectionalLight;
using magnum::graphics3d::Context;
using magnum::graphics3d::Shader;
using magnum::graphics3d::PhongShader;
using magnum::graphics3d::FixedShader;
using magnum::graphics3d::Material;
using magnum::graphics3d::ColourMaterial;
using magnum::graphics3d::GeometryRenderer;
using magnum::geometry::Sphere3;
using magnum::geometry::Cube3;
using magnum::geometry::AlignedBox3;
using magnum::geometry::Box3;
using magnum::geometry::Frustum3;
using magnum::graphics::Colour;
using magnum::stats::Uniform;
using magnum::algorithm::swap;

static DirectionalLight*	g_Light;
static StringView*			g_StringView;
static Cube3<float>			g_GeomCube;
static PhongShader			g_Shader((FixedShader::Features)(FixedShader::TwoSided | FixedShader::Lighting));
static ColourMaterial		g_MatCube;
static Box3<float>			g_GeomBox;
static ColourMaterial		g_MatBox;
static Sphere3<float>		g_GeomSphere;
static ColourMaterial		g_MatSphere;
static AlignedBox3<float>	g_GeomAlignedBox;
static ColourMaterial		g_MatAlignedBox;
static Frustum3<float>		g_GeomFrustum;
static ColourMaterial		g_MatFrustum;
static Uniform				g_RandCoord(-200,200);
static Uniform				g_RandRadius(20,150);
static Uniform				g_RandAngle(0, 360);

enum Command {
	RandSphere,
	RandCube,
	RandAlignedBox,
	RandBox,
	RandFrustum,
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void testIntersections() {
	String text;
	if (g_GeomCube.intersect(g_GeomAlignedBox))
		text += "Cube+AlignedBox  ";
	if (g_GeomCube.intersect(g_GeomBox))
		text += "Cube+Box  ";
	if (g_GeomAlignedBox.intersect(g_GeomBox))
		text += "AlignedBox+Box  ";
	if (g_GeomCube.intersect(g_GeomSphere))
		text += "Cube+Sphere  ";
	if (g_GeomAlignedBox.intersect(g_GeomSphere))
		text += "AlignedBox+Sphere  ";
	if (g_GeomBox.intersect(g_GeomSphere))
		text += "Box+Sphere  ";
	if (g_GeomFrustum.intersect(g_GeomSphere))
		text += "Frustum+Sphere  ";
	if (g_GeomFrustum.intersect(g_GeomCube))
		text += "Frustum+Cube  ";
	if (g_GeomFrustum.intersect(g_GeomAlignedBox))
		text += "Frustum+AlignedBox  ";
	if (g_GeomFrustum.intersect(g_GeomBox))
		text += "Frustum+Box  ";
		
	g_StringView->setText(text);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void randSphere(Sphere3<float>& sphere) {
	sphere.setOrigin(Vector3<float>(g_RandCoord(),g_RandCoord(),g_RandCoord()));
	sphere.setRadius(g_RandRadius());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void randCube(Cube3<float>& cube) {
	cube.setOrigin(Vector3<float>(g_RandCoord(),g_RandCoord(),g_RandCoord()));
	cube.setRadius(g_RandRadius());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void randAlignedBox(AlignedBox3<float>& abox) {
	Vector3<float> mi, ma;
	mi.x = g_RandCoord();
	mi.y = g_RandCoord();
	mi.z = g_RandCoord();
	ma.x = g_RandCoord();
	ma.y = g_RandCoord();
	ma.z = g_RandCoord();
	if (mi.x > ma.x)
		swap(mi.x, ma.x);
	if (mi.y > ma.y)
		swap(mi.y, ma.y);
	if (mi.z > ma.z)
		swap(mi.z, ma.z);
	
	abox.setMinimum(mi);
	abox.setMaximum(ma);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void randBox(Box3<float>& box) {
	Transform3<float> t;
	t.identity();
    t.rotate((float)(g_RandAngle()/180*M_PI), Vector3<float>(1.0f, 0.0f, 1.0f));
    t.rotate((float)(g_RandAngle()/180*M_PI), Vector3<float>(0.0f, 1.0f, 0.0f));
    box.setOrigin(Vector3<float>(g_RandCoord(),g_RandCoord(),g_RandCoord()));
    box.setAxis(0, t.getRotation().getColumn(0) * g_RandRadius());
    box.setAxis(1, t.getRotation().getColumn(1) * g_RandRadius());
    box.setAxis(2, t.getRotation().getColumn(2) * g_RandRadius());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void randFrustum(Frustum3<float>& frst) {
	Transform3<float> t;
	t.identity();
    t.rotate((float)(g_RandAngle()/180*M_PI), Vector3<float>(1.0f, 0.0f, 1.0f));
    t.rotate((float)(g_RandAngle()/180*M_PI), Vector3<float>(0.0f, 1.0f, 0.0f));
    frst.setOrigin(Vector3<float>(g_RandCoord(),g_RandCoord(),g_RandCoord()));
    frst.setDirection(t.getRotation().getColumn(0));
    frst.setAxis(0, t.getRotation().getColumn(1) * g_RandRadius());
    frst.setAxis(1, t.getRotation().getColumn(2) * g_RandRadius());
    frst.setNear(80);
    frst.setFar(400);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onInit() {
	m_Camera.setNearPlane(10);
	m_Camera.setFarPlane(2000);

	g_Light = new DirectionalLight();
	g_Light->setDirection(Vector3<float>(0.1f,0.2f,1.0f));
	g_Light->setAmbient(Colour<float>(0.2f,0.2f,0.2f));
	g_Light->setDiffuse(Colour<float>(1.0f,1.0f,1.0f));
	g_Light->setSpecular(Colour<float>(0.7f,0.7f,0.7f));
	g_Light->select();

	// Create GUI
	Button* button;
	button = new Button(Rectangle<float>(10,10,170,40), Message(this, RandCube), "Rand Cube");
	button->setFont(m_Gui->getFont());
	getGui()->insert(button);

	button = new Button(Rectangle<float>(10,60,170,40), Message(this, RandAlignedBox), "Rand Aligned Box");
	button->setFont(m_Gui->getFont());
	getGui()->insert(button);

	button = new Button(Rectangle<float>(10,110,170,40), Message(this, RandBox), "Rand Box");
	button->setFont(m_Gui->getFont());
	getGui()->insert(button);

	button = new Button(Rectangle<float>(10,160,170,40), Message(this, RandSphere), "Rand Sphere");
	button->setFont(m_Gui->getFont());
	getGui()->insert(button);

	button = new Button(Rectangle<float>(10,210,170,40), Message(this, RandFrustum), "Rand Frustum");
	button->setFont(m_Gui->getFont());
	getGui()->insert(button);
	
	StringView* sview;
	sview = new StringView(Rectangle<float>(10,470,170,40), "Models: 0");
	sview->setAlignement(View::FollowBottom);
	sview->setFont(m_Gui->getFont());
	getGui()->insert(sview);
	g_StringView = sview;

	randCube(g_GeomCube);
	g_MatCube.setAmbient(Colour<float>(0.2f, 0.2f, 0.2f));
	g_MatCube.setDiffuse(Colour<float>(1.0f, 0.2f, 0.2f));
	g_MatCube.setSpecular(Colour<float>(1.0f, 0.8f, 0.7f));
	g_MatCube.setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
	g_MatCube.setShininess(20.0f);

	randAlignedBox(g_GeomAlignedBox);
	g_MatAlignedBox.setAmbient(Colour<float>(0.2f, 0.2f, 0.2f));
	g_MatAlignedBox.setDiffuse(Colour<float>(0.2f, 1.0f, 0.2f));
	g_MatAlignedBox.setSpecular(Colour<float>(0.8f, 1.0f, 0.7f));
	g_MatAlignedBox.setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
	g_MatAlignedBox.setShininess(20.0f);

	randBox(g_GeomBox);
	g_MatBox.setAmbient(Colour<float>(0.2f, 0.2f, 0.2f));
	g_MatBox.setDiffuse(Colour<float>(0.2f, 0.2f, 1.0f));
	g_MatBox.setSpecular(Colour<float>(0.7f, 0.8f, 1.0f));
	g_MatBox.setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
	g_MatBox.setShininess(20.0f);

	randSphere(g_GeomSphere);
	g_MatSphere.setAmbient(Colour<float>(0.2f, 0.2f, 0.2f));
	g_MatSphere.setDiffuse(Colour<float>(0.2f, 1.0f, 1.0f));
	g_MatSphere.setSpecular(Colour<float>(1.0f, 1.0f, 1.0f));
	g_MatSphere.setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
	g_MatSphere.setShininess(20.0f);

	randFrustum(g_GeomFrustum);
	g_MatFrustum.setAmbient(Colour<float>(0.2f, 0.2f, 0.2f));
	g_MatFrustum.setDiffuse(Colour<float>(1.0f, 0.2f, 1.0f));
	g_MatFrustum.setSpecular(Colour<float>(1.0f, 1.0f, 0.7f));
	g_MatFrustum.setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
	g_MatFrustum.setShininess(20.0f);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onExit() {
	Shader::selectNone();
	Material::selectNone();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onHandleMessage(const Message& msg) {
	switch (msg.getCommand()) {
		case RandSphere:
			randSphere(g_GeomSphere);
			return true;
		case RandCube:
			randCube(g_GeomCube);
			return true;
		case RandAlignedBox:
			randAlignedBox(g_GeomAlignedBox);
			return true;
		case RandBox:
			randBox(g_GeomBox);
			return true;
		case RandFrustum:
			randFrustum(g_GeomFrustum);
			return true;
	};
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GraphicsView::onRender() {
	BlendMode::selectDefault();
	DepthMode::selectDefault();
	m_Context->clear(Context::ClearColour | Context::ClearDepth | Context::ClearStencil);
	m_Context->setCulling(Context::CullNone);

	m_Camera.select();
	m_MatrixStack->selectMode(MatrixStack::ModelView);
	m_MatrixStack->identity();
	g_Light->select();
	m_MatrixStack->set((Matrix4<float>)getTransform().getInverse());
	
	g_Shader.select();
	g_MatCube.select();
	GeometryRenderer::render(g_GeomCube);
	g_MatAlignedBox.select();
	GeometryRenderer::render(g_GeomAlignedBox);
	g_MatBox.select();
	GeometryRenderer::render(g_GeomBox);
	g_MatSphere.select();
	GeometryRenderer::render(g_GeomSphere, 40);
	g_MatFrustum.select();
	GeometryRenderer::render(g_GeomFrustum);

	testIntersections();
	
	m_Gui->process();
	m_Gui->paint();
}

}; };

